################################################################################
# Base class for all EMBL mergers (one per organism).
##
package EMBLMerger::Merger;

use strict;

##########################################################################################
#
##
sub new {
    my ($class, $options) = @_;

    my $self  =  {
        'CVSROOT'               => '/nfs/pathdata/repository/',
        'annotation.merged.dir' => "annotation.merged",
    };

    if (defined $options) {
        for my $key (keys %{$options}) {
            $self->{$key} = $options->{$key} if(defined $options->{$key});
        }
    }

    bless($self, $class);

    $self->initialize();;

    return $self;
}

################################################################################
# The inheriting objects use this to deal with org specific things at object
# creation time.
##
sub initialize {
    my ($self) = @_;

    # print "I should only run once!";
}

##########################################################################################
#
##
sub set_cvs_root {
    my ($self) = @_;

    $ENV{'CVSROOT'} = $self->{'CVSROOT'};
}


##########################################################################################
# merge_two_files
#
# Assumes we are in the correct directory
##
sub merge_two_files {
    my ($self, $org, $sparse_file, $cvs_file, $new_file) = @_;

    unless (-e "./annotation_map.properties"){
        `rcp /nfs/pathdata/conf/annotation_map.properties.$org ./annotation_map.properties`;
    }


    ##
    # tigrmerge complains that the sparse files features are out-of-bounds
    # I have made a start on this in TigrMerge.java, probably have to subclass SymbolTokenization
    # Bodge the sequence into the second file
    ##
    $self->bodge_sequence($cvs_file, $sparse_file);


    ##
    # Create directory if is does not already exist (better done in tigrmerge)
    ##
    $self->create_directories_for($new_file);


    my $setup = "PSU_CONFIG_DIR=/nfs/pathsoft/prod/config; export PSU_CONFIG_DIR";

    # ************************************************************************************
    # USE pjmtigrmerge = WHEN TESTING ONLY

    print `$setup ; tigrmerge  $cvs_file $sparse_file $new_file yes 2>&1`; # Note redirection of STDERR
    print "tigrmering complete\n";
}

##########################################################################################
#
# bodge_sequence
# 
# Take sequence form first file (cvs versioN) and insert into the new sparse file that
# has no sequence.
# 
##
sub bodge_sequence {
    my ($self, $cvs_file, $new_file) = @_;

    print "bodge_sequence()\n";

    ##
    # Grab sequence from first (cvs) file
    ##
    open ONE, "<$cvs_file";

    my $grab_seq = 0;
    my $seq      = undef;

    while(<ONE>){
        if (m|//|) {
            last;
        }
        elsif (/^SQ /) {
            $grab_seq = 1;
            $seq = $_;
        }
        elsif ($grab_seq) {
            $seq .= $_;
        }
    }

    close ONE;

    ##
    # Insert sequence into second (new) file
    ##

    open TWO, "<$new_file" or die "ERROR: could not open for reading $new_file";

    my @new_file_lines;

    while (<TWO>){
        if (m|//|) {
            push @new_file_lines, "//\n";
            last;
        }
        elsif (/^SQ /) {
            push @new_file_lines, $seq;
        }
        else {
            push @new_file_lines, $_;
        }
    }

    close TWO;

    print "bodge_sequence(): read new_file: $new_file\n";

    ##
    # Move new file side ways and leave modified version with sequence in it (name) place
    ##

    print "bodge_sequence(): mv ${new_file} ${new_file}.no_sequence about to be run\n";

    print `mv ${new_file} ${new_file}.no_sequence`;

    open  TWO, ">$new_file" or die "ERROR: could not open for writing $new_file";;
    print TWO @new_file_lines;
    close TWO;

    print "bodge_sequence(): $new_file written to\n";
}

################################################################################
#
# create_directories_for
#
# For a given path to a file, it checks the directory structure specified
# exists. If it does not is creates it.
#
# It can cope with these;
#         filename_on_its_own
#         ./filename
#         /tmp/filename <--- SPECAIL CASE
#         annotation/Trypanosoma/brucei/chr02/Tb927_02_mV2.embl
##

sub create_directories_for {
    my ($self, $new_file) = @_;
    
    my @dirs = split('/', $new_file);
    pop @dirs;
    #print "dirs = ", join(", ", @dirs), "\n";


    ##
    # SPECIAL CASE: Fix path that has slash as first char i.e. /tmp/thing
    ##
    if ($new_file =~ m|^/|) {
        shift @dirs;
        $dirs[0] = "/".$dirs[0];
    }

    ##
    # Create each dir as need be
    ##
    my $prev_dirs = "";

    foreach my $dir (@dirs) {
        next if ($dir eq '.');

        my $whole_dir = "${prev_dirs}${dir}";

        unless (-d $whole_dir) {
            #print "Making directory: $whole_dir\n";
            mkdir "$whole_dir" or die "ERROR: Unable to create directory $whole_dir";
        }
        else {
            #print "$whole_dir already exists\n";
        }

        $prev_dirs .= "${dir}/";
    }

}


1;
