/*
 * This file is part of Gemini. Copyright (c) 2001, 2002 Genome
 * Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.psu.gemini;

import java.io.*;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.awt.Component;
import javax.swing.ProgressMonitor;
import javax.swing.ProgressMonitorInputStream;

import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import org.biojava.bio.BioError;
import org.biojava.bio.BioException;
import org.biojava.bio.program.sax.BlastLikeSAXParser;
import org.biojava.bio.program.sax.FastaSearchSAXParser;
import org.biojava.bio.program.ssbind.SeqSimilarityAdapter;
import org.biojava.bio.program.ssbind.SimilarityPairBuilder;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.db.IllegalIDException;
import org.biojava.bio.seq.db.SequenceDB;
import org.biojava.bio.seq.db.SequenceDBInstallation;
import org.biojava.bio.seq.io.EmblLikeFormat;
import org.biojava.bio.seq.io.FastaFormat;
import org.biojava.bio.seq.io.SeqIOTools;
import org.biojava.bio.seq.io.SequenceBuilderFactory;
import org.biojava.bio.seq.io.SequenceFormat;
import org.biojava.utils.ChangeVetoException;

import uk.ac.sanger.util.seq.SequenceFormatTester;
import uk.ac.sanger.util.seq.SequenceFormatTesterDefaultImpl;

/**
 * <code>DataHandler</code> centralises opening, reading and parsing
 * sequence and search result files.
 *
 * @author Keith James
 */
public final class DataHandler
{
    /**
     * <code>BLAST</code> search type Blast (NCBI- or Wu-Blast 2).
     */
    public static final int BLAST = 0;

    /**
     * <code>FASTA</code> search type Fasta.
     */
    public static final int FASTA = 1;

    static final String   QUERY = "QUERY";
    static final String SUBJECT = "SUBJECT";

    private Component              monitor;
    private Sequence               querySequence;
    private Sequence               subjectSequence;
    private SequenceDB             queryDB;
    private SequenceDB             subjectDB;
    private SequenceDBInstallation subjectDBHolder;
    private SequenceFormatTester   formatTester;

    private ContentHandler         handler;
    private Sequence               annotatedQuery;
    private int                    searchType;

    /**
     * Creates a new <code>DataHandler</code>.
     *
     * @param monitor a <code>Component</code> to be the parent of the
     * loading progress monitor.
     */
    public DataHandler(Component monitor)
    {
        this.monitor = monitor;
        formatTester = SequenceFormatTesterDefaultImpl.getInstance();
    }

    /**
     * <code>getQuerySequence</code> returns the current query
     * sequence.
     *
     * @return a <code>Sequence</code>.
     */
    public Sequence getQuerySequence()
    {
        return querySequence;
    }

    /**
     * <code>getSubjectSequence</code> returns the current subject
     * sequence.
     *
     * @return a <code>Sequence</code>.
     */
    public Sequence getSubjectSequence()
    {
        return subjectSequence;
    }

    /**
     * <code>getAnnotatedQuerySequence</code> returns a view of the
     * query sequence annotated with features representing similarity
     * with the query.
     *
     * @return a <code>Sequence</code>.
     */
    public Sequence getAnnotatedQuerySequence()
    {
        return annotatedQuery;
    }

    /**
     * <code>setSearchType</code> sets the search type to be expected.
     *
     * @param searchType an <code>int</code>, one of
     * <code>DataHandler.BLAST</code> or
     * <code>DataHandler.FASTA</code>.
     */
    public void setSearchType(int searchType)
    {
        switch (searchType)
        {
            case (BLAST):
            case (FASTA):
                this.searchType = searchType;
                break;

            default:
                throw new IllegalArgumentException("Unknown search type requested");
        }
    }

    /**
     * <code>setQuerySeqFromFile</code> opens and reads the query
     * sequence and database from a file.
     *
     * @param file a <code>File</code>.
     *
     * @exception IOException if an IO error occurs.
     * @exception BioException if an error occurs.
     */
    public void setQuerySeqFromFile(File file) throws IOException, BioException
    {
        querySequence = readSequenceFromFile(file);
        queryDB = new SingleSequenceDB(QUERY, querySequence);
    }

    /**
     * <code>setSubjectSeqFromFile</code> opens and reads the subject
     * sequence and database from a file.
     *
     * @param file a <code>File</code>.
     *
     * @exception IOException if an IO error occurs.
     * @exception BioException if an error occurs.
     */
    public void setSubjectSeqFromFile(File file) throws IOException, BioException
    {
        subjectSequence = readSequenceFromFile(file);
        subjectDB = new SingleSequenceDB(SUBJECT, subjectSequence);

        subjectDBHolder = new SequenceDBInstallation()
        {
            public SequenceDB getSequenceDB(String id)
            {
                return DataHandler.this.subjectDB;
            }

            public Set getSequenceDBs()
            {
                return null;
            }
        };
    }

    /**
     * <code>setSubjectSeqFromInstance</code> sets the subject
     * sequence and database from a sequence instance.
     *
     * @param seq a <code>Sequence</code>.
     */
    public void setSubjectSeqFromInstance(Sequence seq)
    {
        subjectDB = new SingleSequenceDB(SUBJECT, seq);

        subjectDBHolder = new SequenceDBInstallation()
        {
            public SequenceDB getSequenceDB(String id)
            {
                return DataHandler.this.subjectDB;
            }

            public Set getSequenceDBs()
            {
                return null;
            }
        };
    }

    /**
     * <code>readCompFromFile</code> reads a sequence comparison data
     * from a named file.
     *
     * @param file a <code>File</code>.
     * @param chunkSize an <code>int</code> indicating the length of
     * query chunks.
     *
     * @exception IOException if an IO error occurs.
     * @exception SAXException if a SAX error occurs.
     */
    public void readCompFromFile(File file, int chunkSize)
        throws IOException, SAXException
    {
        String m = "Reading comparison results from " + file.getName();

        ProgressMonitorInputStream pmis =
            new ProgressMonitorInputStream(monitor, m, new FileInputStream(file));

        BufferedReader reader = new BufferedReader(new InputStreamReader(pmis));

        XMLReader parser = null;
        if (searchType == BLAST)
        {
            parser = new BlastLikeSAXParser();
            ((BlastLikeSAXParser) parser).setModeLazy();
        }
        else if (searchType == FASTA)
            parser = new FastaSearchSAXParser();
        else
            throw new Error("Internal error: an unknown search type was selected");

        IncrementalSimilarityPairBuilder builder =
            new IncrementalSimilarityPairBuilder(0, chunkSize);

        builder.setQuerySeqHolder(queryDB);
        builder.setSubjectDBInstallation(subjectDBHolder);

        SeqSimilarityAdapter adapter = new SeqSimilarityAdapter();
        adapter.setSearchContentHandler(builder);

        parser.setContentHandler(adapter);
        parser.parse(new InputSource(reader));

        try
        {
            annotatedQuery = builder.getAnnotatedQuerySeq(QUERY);
        }
        catch (IllegalIDException iie)
        {
            throw new BioError(iie, "An error occured while retrieving the annotated query sequence");
        }
    }

    /**
     * <code>readSequenceFromFile</code> reads a sequence from a named
     * file.
     *
     * @param file a <code>File</code>.
     *
     * @return a <code>Sequence</code>.
     *
     * @exception IOException if an IO error occurs.
     * @exception BioException if an error occurs.
     */
    Sequence readSequenceFromFile(File file) throws IOException, BioException
    {
        SequenceFormat format = formatTester.getSequenceFormat(file);
        String m = "Reading sequence from " + file.getName();

        ProgressMonitorInputStream pmis =
            new ProgressMonitorInputStream(monitor, m, new FileInputStream(file));

        BufferedReader br = new BufferedReader(new InputStreamReader(pmis));

        SequenceIterator si;

        if (EmblLikeFormat.class.isInstance(format))
            si = SeqIOTools.readEmbl(br);
        else if (FastaFormat.class.isInstance(format))
            si = SeqIOTools.readFastaDNA(br);
        else
            throw new BioException("Sequence file '"
                                   + file
                                   + "' is in an unknown format (Fasta and EMBL supported)");

        return si.nextSequence();
    }
}
