/*
 * This file is part of Gemini. Copyright (c) 2001, 2002 Genome
 * Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.psu.gemini;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.biojava.bio.Annotation;
import org.biojava.bio.BioException;
import org.biojava.bio.BioError;
import org.biojava.bio.BioRuntimeException;
import org.biojava.bio.SmallAnnotation;
import org.biojava.bio.program.ssbind.AlphabetResolver;
import org.biojava.bio.program.ssbind.AnnotationFactory;
import org.biojava.bio.search.SearchContentHandler;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.StrandedFeature.Strand;
import org.biojava.bio.seq.StrandedFeature;
import org.biojava.bio.seq.ViewSequence;
import org.biojava.bio.seq.db.IllegalIDException;
import org.biojava.bio.seq.db.SequenceDB;
import org.biojava.bio.seq.db.SequenceDBInstallation;
import org.biojava.bio.seq.homol.SimilarityPairFeature;
import org.biojava.bio.seq.io.SymbolTokenization;
import org.biojava.bio.symbol.Alignment;
import org.biojava.bio.symbol.FiniteAlphabet;
import org.biojava.bio.symbol.RangeLocation;
import org.biojava.bio.symbol.SimpleAlignment;
import org.biojava.utils.ChangeVetoException;

/**
 * <p><code>IncrementalSimilarityPairBuilder</code> annotates query
 * and subject <code>Sequence</code> with
 * <code>SimilarityPairFeature</code>s created from SAX events
 * supplied via a <code>SeqSimilarityAdapter</code>. The SAX events
 * should describe elements conforming to the BioJava
 * BlastLikeDataSetCollection DTD. Suitable sources are
 * <code>BlastLikeSAXParser</code> or <code>FastaSAXParser</code>.</p>
 *
 * <p>Annotated <code>ViewSequence</code>s wrapping both query and
 * subject sequences are created.</p>
 *
 * <p><code>IncrementalSimilarityPairBuilder</code> is so called
 * because it can annotate the <code>ViewSequence</code>s in stages
 * from a BlastLikeDataSetCollection which contains one or more
 * BlastLikeDataSets (each represents one search). This feature is
 * used where a query sequence is very big and has been broken down
 * into fragments, which are searched individually, in order to reduce
 * compute time. The arguments supplied to the constructor are the
 * position in the query sequence to start annotating and the
 * length.</p>
 *
 * @author Keith James
 */
public class IncrementalSimilarityPairBuilder implements SearchContentHandler
{
    /**
     * Constant <code>SIMILARITY_PAIR_FEATURE_TYPE</code> the type
     * String used by <code>IncrementalSimilarityPairBuilder</code>
     * when creating <code>SimilarityPairFeature</code>s. This is the
     * String which is returned when an
     * <code>SimilarityPairFeature</code>'s <code>getType()</code>
     * method is called.
     */
    public static final String SIMILARITY_PAIR_FEATURE_TYPE = "similarity";

    // Supplier of instances of searched databases
    private SequenceDBInstallation subjectDBs;
    // The specific database searched
    private SequenceDB subjectDB;
    // Holder for all query sequences
    private SequenceDB querySeqHolder;
    // View of query sequence instance
    private Sequence queryView;
    // View of subject sequence instance
    private Sequence subjectView;

    // Sequence index offset to add to the coordinates of annotated
    // features. This is added to the currentOffset with each
    // BlastLikeDataset. The assumption is that the query sequence was
    // broken into chunks of this size before searching and the
    // corresponding searches are in order in the
    // BlastLikeDatasetCollection.
    private int offsetIncrement;
    // Current running total offset (equal to offsetIncrement *
    // BlastLikeDataset count)
    private int currentOffset;

    // Data holders for search result properties
    private Map resultData;
    private Map hitData;
    private Map subHitData;

    private AlphabetResolver   alphaResolver;
    private SymbolTokenization tokenParser;
    private StringBuffer       tokenBuffer;

    // Flag indicating whether there are more results in the stream
    private boolean moreSearchesAvailable = false;

    public IncrementalSimilarityPairBuilder(int startOffset,
                                            int offsetIncrement)
    {
        currentOffset = startOffset;
        this.offsetIncrement = offsetIncrement;

        resultData       = new HashMap();
        hitData          = new HashMap();
        subHitData       = new HashMap();
        alphaResolver    = new AlphabetResolver();
        tokenBuffer      = new StringBuffer(1024);
    }

    public Sequence getAnnotatedQuerySeq(String queryId)
        throws IllegalIDException
    {
        return queryView;
    }

    public Sequence getAnnotatedSubjectSeq(String subjectId)
        throws IllegalIDException
    {
        return subjectView;
    }

    /**
     * <code>getQuerySeqHolder</code> returns the database of query
     * sequences used to retrieve sequences for creation of the
     * various result objects.
     *
     * @return a <code>SequenceDB</code> value.
     */
    public SequenceDB getQuerySeqHolder()
    {
        return querySeqHolder;
    }

    /**
     * <code>setQuerySeqHolder</code> sets the query sequence holder
     * to a specific database.
     *
     * @param querySeqHolder a <code>SequenceDB</code> containing the
     * query sequence(s).
     */
    public void setQuerySeqHolder(SequenceDB querySeqHolder)
    {
        this.querySeqHolder = querySeqHolder;
    }

    /**
     * <code>getSubjectDBInstallation</code> returns the installation
     * in which all the databases searched may be
     * found. <code>SequenceDB</code>s are retrieved for creation of
     * the various result objects.
     *
     * @return a <code>SequenceDBInstallation</code> containing the
     * subject database(s).
     */
    public SequenceDBInstallation getSubjectDBInstallation()
    {
        return subjectDBs;
    }

    /**
     * <code>setSubjectDBInstallation</code> sets the subject database
     * holder to a specific installation.
     *
     * @param subjectDBs a <code>SequenceDBInstallation</code>
     * containing the subject database(s)
     */
    public void setSubjectDBInstallation(SequenceDBInstallation subjectDBs)
    {
        this.subjectDBs = subjectDBs;
    }

    public void setQuerySeq(String querySeqId)
        throws BioException
    {
        if (querySeqHolder == null)
            throw new BioException("Running IncrementalSimilarityPairBuilder with null query SequenceDB");

        if (queryView == null)
        {
            Sequence temp = querySeqHolder.getSequence(querySeqId);

            // It shouldn't happen, but it can with some implementations
            // of SequenceDB
            if (temp == null)
                throw new BioException("Failed to retrieve query sequence from holder using ID '"
                                       + querySeqId
                                       + "' (sequence was null)");

            queryView = new ViewSequence(temp, DataHandler.QUERY);
        }
    }

    public void setSubjectDB(String subjectDBName)
        throws BioException
    {
        if (subjectDBs == null)
            throw new BioException("Running IncrementalSimilarityPairBuilder with null subject SequenceDBInstallation");

        subjectDB = subjectDBs.getSequenceDB(subjectDBName);

	if (subjectDB == null)
	    throw new BioException("Failed to retrieve database from installation using ID '"
				   + subjectDBName
                                   + "'");
    }

    public boolean getMoreSearches()
    {
        return moreSearchesAvailable;
    }

    public void setMoreSearches(boolean value)
    {
        moreSearchesAvailable = value;
    }

    public void startSearch() { }

    public void endSearch()
    {
        currentOffset += offsetIncrement;
    }

    public void startHeader()
    {
        resultData.clear();
    }

    public void endHeader() { }

    public void startHit()
    {
        hitData.clear();
        subHitData.clear();
    }

    public void endHit() { }

    public void startSubHit() { }

    public void endSubHit()
    {
        try
        {
            makeSimilarity();
        }
        catch (BioException be)
        {
            throw new BioRuntimeException(be, "Failed to build similarity from search data");
        }
    }

    public void addSearchProperty(Object key, Object value)
    {
        resultData.put(key, value);
    }

    public void addHitProperty(Object key, Object value)
    {
        hitData.put(key, value);
    }

    public void addSubHitProperty(Object key, Object value)
    {
        subHitData.put(key, value);
    }

    private void makeSimilarity() throws BioException
    {
        subHitData.putAll(resultData);
        subHitData.putAll(hitData);

        // Try to get a valid TokenParser
        if (tokenParser == null)
        {
            String identifier;
            // Try explicit sequence type first
            if (subHitData.containsKey("hitSequenceType"))
                identifier = (String) subHitData.get("hitSequenceType");
            // Otherwise try to resolve from the program name (only
            // works for Blast)
            else if (subHitData.containsKey("program"))
                identifier = (String) subHitData.get("program");
            else
                throw new BioException("Failed to determine sequence type");

            FiniteAlphabet alpha = alphaResolver.resolveAlphabet(identifier);
            tokenParser = alpha.getTokenization("token");
        }

        // Set strands of hit on query and subject
        Strand qStrand = StrandedFeature.POSITIVE;
        Strand sStrand = StrandedFeature.POSITIVE;

        // In cases where an explicit strand is given (FASTA DNA, BLASTN)
        if (subHitData.containsKey("queryStrand") &&
            subHitData.get("queryStrand").equals("minus"))
            qStrand = StrandedFeature.NEGATIVE;

        if (subHitData.containsKey("subjectStrand") &&
            subHitData.get("subjectStrand").equals("minus"))
            sStrand = StrandedFeature.NEGATIVE;

        // In cases where a frame is given as this contains strand
        // information (TBLASTN for hit, TBLASTX for both query and
        // hit)
        if (subHitData.containsKey("queryFrame") &&
            ((String) subHitData.get("queryFrame")).startsWith("minus"))
            qStrand = StrandedFeature.NEGATIVE;

        if (subHitData.containsKey("subjectFrame") &&
            ((String) subHitData.get("subjectFrame")).startsWith("minus"))
            sStrand = StrandedFeature.NEGATIVE;

        // Get start/end
        int qStart = Integer.parseInt((String) subHitData.get("querySequenceStart"));
        int   qEnd = Integer.parseInt((String) subHitData.get("querySequenceEnd"));
        int sStart = Integer.parseInt((String) subHitData.get("subjectSequenceStart"));
        int   sEnd = Integer.parseInt((String) subHitData.get("subjectSequenceEnd"));

        // Apply offset to query
        qStart += currentOffset;
        qEnd   += currentOffset;

        // The start/end coordinates from BioJava XML don't follow the
        // BioJava paradigm of start < end, with orientation given by
        // the strand property. Rather, they present start/end as
        // displayed in BLAST output, with the coordinates being
        // inverted on the reverse strand. We account for this here.
        if (qStrand == StrandedFeature.NEGATIVE)
        {
            int swap = qStart;
            qStart = qEnd;
            qEnd   = swap;
        }

        if (sStrand == StrandedFeature.NEGATIVE)
        {
            int swap = sStart;
            sStart = sEnd;
            sEnd   = swap;
        }

        String subjectSeqId = (String) subHitData.get("HitId");

        // If we have already instantiated a subjectView for this sequence
        if (subjectView == null)
        {
            Sequence temp = subjectDB.getSequence(subjectSeqId);

            if (temp == null)
                throw new BioException("Failed to retrieve subject sequence from subjectDB using ID '"
                                       + subjectSeqId
                                       + "' (sequence was null)");

            subjectView = new ViewSequence(temp, DataHandler.SUBJECT);
        }

        // Map of Alignment sequences
        Map labelMap = new HashMap();

        try
        {
            // Set source to the program name
            String source = "unknown";
            if (subHitData.containsKey("program"))
                source = (String) subHitData.get("program");

            double score = 0.0;
            if (subHitData.containsKey("score"))
            {
                score = Double.parseDouble((String) subHitData.get("score"));
            }

            // Query sequence feature
            SimilarityPairFeature.Template qt = new SimilarityPairFeature.Template();
            qt.type       = SIMILARITY_PAIR_FEATURE_TYPE;
            qt.source     = source;
            qt.location   = new RangeLocation(qStart, qEnd);
            qt.strand     = qStrand;
            qt.score      = score;
            qt.annotation = AnnotationFactory.makeAnnotation(subHitData);

            // Subject sequence feature
            SimilarityPairFeature.Template st = new SimilarityPairFeature.Template();
            st.type       = SIMILARITY_PAIR_FEATURE_TYPE;
            st.source     = source;
            st.location   = new RangeLocation(sStart, sEnd);
            st.strand     = sStrand;
            st.score      = score;
            st.annotation = AnnotationFactory.makeAnnotation(subHitData);

            qt.alignment = SimilarityPairFeature.EMPTY_PAIRWISE;
            st.alignment = SimilarityPairFeature.EMPTY_PAIRWISE;

            qt.sibling = (SimilarityPairFeature) subjectView.createFeature(st);
            st.sibling = (SimilarityPairFeature) queryView.createFeature(qt);
        }
        catch (ChangeVetoException cve)
        {
            throw new BioError(cve, "Assertion failure creating SimilarityPairFeature. Template modification vetoed");
        }
    }
}
