/*
 * This file is part of Gemini. Copyright (c) 2001, 2002 Genome
 * Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.psu.gemini;

import java.util.Collections;
import java.util.Set;

import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.db.IllegalIDException;
import org.biojava.bio.seq.db.SequenceDB;
import org.biojava.utils.ChangeListener;
import org.biojava.utils.ChangeType;
import org.biojava.utils.ChangeVetoException;
import org.biojava.utils.Changeable;

/**
 * <code>SingleSequenceDB</code> is <code>SequenceDB</code>
 * implementation which contains only one sequence. This sequence is
 * returned on request, regardless of the sequence ID used. These
 * objects are immutable; sequences may not be added or removed.
 *
 * @author Keith James
 */
public final class SingleSequenceDB implements SequenceDB
{
    private String   name;
    private Sequence sequence;

    public SingleSequenceDB(String name, Sequence sequence)
    {
        this.name     = name;
        this.sequence = sequence;
    }

    public String getName()
    {
        return name;
    }

    public Set ids()
    {
        return Collections.EMPTY_SET;
    }

    public SequenceIterator sequenceIterator()
    {
        return new SequenceIterator()
        {
            boolean hasNext = true;

            public boolean hasNext()
            {
                return hasNext;
            }

            public Sequence nextSequence()
            {
                if (hasNext)
                {
                    hasNext = false;
                    return SingleSequenceDB.this.sequence;
                }
                else
                {
                    return null;
                }
            }
        };
    }

    public Sequence getSequence(String id)
        throws IllegalIDException, BioException
    {
        return sequence;
    }

    public void addSequence(Sequence id)
        throws IllegalIDException, BioException, ChangeVetoException
    {
         throw new ChangeVetoException("Failed to add sequence to SingleSequenceDB; the object is immutable");
    }

    public void removeSequence(String id)
        throws IllegalIDException, BioException, ChangeVetoException
    {
        throw new ChangeVetoException("Failed to remove sequence from SingleSequenceDB; the object is immutable");
    }

    public boolean isUnchanging(ChangeType type)
    {
        return true;
    }

    public void addChangeListener(ChangeListener l) { }
    public void addChangeListener(ChangeListener l, ChangeType t) { }
    public void removeChangeListener(ChangeListener l) { }
    public void removeChangeListener(ChangeListener l, ChangeType t) { }
}
