/*
 * This file is part of Gemini. Copyright (c) 2001, 2002 Genome
 * Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.psu.gemini.gui;

import java.io.File;

import java.awt.Graphics2D;

import org.biojava.bio.gui.sequence.PairwiseSequencePanel;
import org.biojava.bio.gui.sequence.TranslatedSequencePanel;

/**
 * <code>AbstractExporter</code> is an abstract base class for objects
 * which paint onto a <code>Graphics2D</code> to generate an image and
 * then export that image to a file.
 *
 * @author Keith James
 */
abstract class AbstractExporter implements Runnable
{
    private Thread   thread;

    protected GeminiUI geminiUI;
    protected File     exportFile;

    /**
     * Creates a new <code>AbstractExporter</code> for the specified
     * UI.
     *
     * @param geminiUI a <code>GeminiUI</code>.
     */
    AbstractExporter(GeminiUI geminiUI)
    {
        thread = new Thread(this);
        this.geminiUI = geminiUI;
    }

    /**
     * <code>setExportFile</code> sets the file to which the image
     * will be exported.
     *
     * @param exportFile a <code>File</code>.
     */
    void setExportFile(File exportFile)
    {
        this.exportFile = exportFile;
    }

    /**
     * <code>doExport</code> performs the export.
     */
    void doExport()
    {
        thread.start();
    }

    /**
     * <code>run</code> is the exporter thread run method which must
     * be implemented by concrete subclasses.
     */
    public abstract void run();

    /**
     * <code>paintSections</code> performs the translations and paint
     * operations.
     *
     * @param g2 a <code>Graphics2D</code>.
     */
    protected void paintSections(Graphics2D g2)
    {
        PairwiseSequencePanel    plotPanel = geminiUI.getPairwisePanel();
        TranslatedSequencePanel horizPanel = geminiUI.getXSequencePanel();
        TranslatedSequencePanel  vertPanel = geminiUI.getYSequencePanel();

        int plotWidth = plotPanel.getWidth();
        int vertWidth = vertPanel.getWidth();
        int  plotHeight = plotPanel.getHeight();
        int horizHeight = horizPanel.getHeight();

        g2.translate(0, horizHeight);
        vertPanel.paint(g2);
        g2.translate(0, - horizHeight);
        g2.translate(vertWidth, 0);
        horizPanel.paint(g2);
        g2.translate(0, horizHeight);
        plotPanel.paint(g2);
    }
}
