/*
 * This file is part of Gemini. Copyright (c) 2001, 2002 Genome
 * Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.psu.gemini.gui;

import javax.swing.JLabel;
import javax.swing.JSlider;
import java.text.NumberFormat;
import java.text.FieldPosition;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * <p><code>AbstractSliderLabel</code> is an abstract base class for
 * labels which update their text when a slider's value changes.</p>
 *
 * <p>The label format created is (with the prefix flag set true):</p>
 *
 * <p>"&lt;invariant label&gt; &lt;transformed slider value&gt;"</p>
 * <p>or</p>
 * <p>"&lt;transformed slider value&gt; &lt;invariant label&gt;"</p>
 *
 * <p>(with the prefix flag set false).</p>
 *
 * <p>The integer value obtained from the slider's
 * <code>getValue()</code> method is transformed by the
 * <code>transformValue()</code> method and added to the label,
 * separated by a single space.</p>
 *
 * @author Keith James
 */
public abstract class AbstractSliderLabel extends JLabel
    implements ChangeListener
{
    private String        label;
    private boolean       prefix;
    private NumberFormat  nf;
    private FieldPosition fp;

    /**
     * <code>sb</code> is the buffer in which the updated label is
     * created.
     */
    protected StringBuffer sb;

    /**
     * Creates a new <code>AbstractSliderLabel</code>.
     *
     * @param label a <code>String</code> to be used as the invariant
     * label.
     * @param prefix a <code>boolean</code> indicating whether the
     * label is to be applied as a prefix to the variable value.
     * @param nf a <code>NumberFormat</code> used to format the
     * transformed value.
     * @param fp a <code>FieldPosition</code> used to format the
     * transformed value. 
     */
    public AbstractSliderLabel(String        label,
                               boolean       prefix,
                               NumberFormat  nf,
                               FieldPosition fp)
    {
        super(label);

        this.label  = label;
        this.prefix = prefix;
        this.nf     = nf;
        this.fp     = fp;

        sb = new StringBuffer(128);
    }

    public void stateChanged(ChangeEvent ce)
    {
        JSlider source = (JSlider) ce.getSource();
        int val = source.getValue();

        sb.setLength(0);

        if (! label.equals(""))
        {
            if (prefix)
            {
                sb.append(label);
                sb.append(" ");
                nf.format(transformValue(val), sb, fp);
            }
            else
            {
                nf.format(transformValue(val), sb, fp);
                sb.append(" ");
                sb.append(label);
            }
        }
        else
        {
            nf.format(transformValue(val), sb, fp);
        }

        setText(sb.toString());
    }

    /**
     * <code>transformValue</code>.
     *
     * @param value an <code>int</code> which is obtained from the
     * slider's <code>getValue()</code> method when the slider is
     * moved.
     *
     * @return an <code>Object</code> to format into text is the label
     * prefix
     */
    protected abstract Object transformValue(int value);
}
