/*
 * This file is part of Gemini. Copyright (c) 2001, 2002 Genome
 * Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.psu.gemini.gui;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.io.File;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;

import org.biojava.bio.gui.sequence.PairwiseSequencePanel;
import org.biojava.bio.gui.sequence.TranslatedSequencePanel;

import uk.ac.sanger.psu.gemini.Gemini;
import uk.ac.sanger.psu.gemini.DataHandler;
import uk.ac.sanger.psu.gemini.gui.GeminiUI;

/**
 * <code>ActionFactory</code> creates <code>Action</code> objects
 * suitable for use in menus or toolbars.
 *
 * @author Keith James
 */
final class ActionFactory
{
    static final int          OPEN = 0;
    static final int    EXPORT_SVG = 1;
    static final int   EXPORT_JPEG = 2;
    static final int          QUIT = 3;
    static final int         ABOUT = 4;
    static final int ANTIALIAS_SEQ = 5;
    static final int ANTIALIAS_CMP = 6;
    static final int      CONTROLS = 7;
    static final int DONE_CONTROLS = 8;

    private GeminiUI       geminiUI;
    private RenderingHints nullHint;

    private FileExtFilter jpegFilter;
    private FileExtFilter pngFilter;
    private FileExtFilter svgFilter;

    /**
     * Creates a new <code>ActionFactory</code>.
     *
     * @param geminiUI a <code>GeminiUI</code>.
     */
    ActionFactory(GeminiUI geminiUI)
    {
        this.geminiUI = geminiUI;
        nullHint = new RenderingHints(null);
        jpegFilter = new FileExtFilter(new String [] { "jpg", "jpeg" },
                                       "JPEG images (*.jpg, *.jpeg)");
        svgFilter = new FileExtFilter(new String [] { "svg" },
                                      "SVG images (*.svg)");
    }

    /**
     * <code>createAction</code> creates a new action for the parent UI.
     *
     * @param type an <code>int</code> indicating the type of action
     * required.
     *
     * @return an <code>Action</code>.
     *
     * @exception Exception if an error occurs.
     */
    Action createAction(int type) throws Exception
    {
        switch (type)
        {
            // Open dialog
            case (OPEN):
                return new AbstractAction("Open ...")
                {
                    public void actionPerformed(ActionEvent ae)
                    {
                        geminiUI.getOpenDialog().show();
                    }
                };

            // Export SVG
            case (EXPORT_SVG):
                return new AbstractAction("SVG format")
                {
                    public void actionPerformed(ActionEvent ae)
                    {
                        final SVGExporter exporter = geminiUI.getSVGExporter();
                        JFileChooser chooser = geminiUI.getFileChooser();
                        chooser.addChoosableFileFilter(jpegFilter);
                        chooser.addChoosableFileFilter(svgFilter);
                        chooser.setFileFilter(svgFilter);
                        chooser.setCurrentDirectory(new File("."));
                        if (chooser.showDialog(geminiUI, "Save") != 
                            JFileChooser.APPROVE_OPTION)
                            return;

                        File file = chooser.getSelectedFile();
                        exporter.setExportFile(file);

                        Thread export = new Thread()
                            {
                                public void run()
                                {
                                    exporter.doExport();
                                }
                            };

                        SwingUtilities.invokeLater(export);
                    }
                };

            // Export JPEG
            case (EXPORT_JPEG):
                return new AbstractAction("JPEG format")
                {
                    public void actionPerformed(ActionEvent ae)
                    {
                        final JPEGExporter exporter = geminiUI.getJPEGExporter();
                        JFileChooser chooser = geminiUI.getFileChooser();
                        chooser.addChoosableFileFilter(jpegFilter);
                        chooser.addChoosableFileFilter(svgFilter);
                        chooser.setFileFilter(jpegFilter);
                        chooser.setCurrentDirectory(new File("."));
                        if (chooser.showDialog(geminiUI, "Save") != 
                            JFileChooser.APPROVE_OPTION)
                            return;

                        File file = chooser.getSelectedFile();
                        exporter.setExportFile(file);

                        Thread export = new Thread()
                        {
                            public void run()
                            {
                                exporter.doExport();
                            }
                        };

                        SwingUtilities.invokeLater(export);
                    }
                };

            // Quit application
            case (QUIT):
                return new AbstractAction("Quit")
                {
                    public void actionPerformed(ActionEvent ae)
                    {
                        System.exit(0);
                    }
                };

            // About dialog
            case (ABOUT):
                return new AbstractAction("About Gemini")
                {
                    public void actionPerformed(ActionEvent ae)
                    {
                        geminiUI.getAboutDialog().show();
                    }
                };

            // Toggle antialiasing of sequence renderers
            case (ANTIALIAS_SEQ):
                return new AbstractAction("Sequence antialiasing")
                {
                    public void actionPerformed(ActionEvent ae)
                    {
                        JCheckBoxMenuItem source =
                            (JCheckBoxMenuItem) ae.getSource();

                        TranslatedSequencePanel horizPanel =
                            geminiUI.getXSequencePanel();
                        TranslatedSequencePanel vertPanel =
                            geminiUI.getYSequencePanel();

                        RenderingHints antiAliasHint = geminiUI.getRenderingHints();

                        if (vertPanel.getRenderingHints() == antiAliasHint)
                        {
                            vertPanel.setRenderingHints(nullHint);
                            horizPanel.setRenderingHints(nullHint);
                            source.setSelected(false);
                        }
                        else
                        {
                            vertPanel.setRenderingHints(antiAliasHint);
                            horizPanel.setRenderingHints(antiAliasHint);
                            source.setSelected(true);
                        }
                    }
                };

            // Toggle antialiasing of pairwise renderer
            case (ANTIALIAS_CMP):
                return new AbstractAction("Comparison antialiasing")
                {
                    public void actionPerformed(ActionEvent ae)
                    {
                        JCheckBoxMenuItem source =
                            (JCheckBoxMenuItem) ae.getSource();

                        PairwiseSequencePanel plotPanel =
                            geminiUI.getPairwisePanel();

                        RenderingHints antiAliasHint = geminiUI.getRenderingHints();

                        if (plotPanel.getRenderingHints() == antiAliasHint)
                        {
                            plotPanel.setRenderingHints(nullHint);
                            source.setSelected(false);
                        }
                        else
                        {
                            plotPanel.setRenderingHints(antiAliasHint);
                            source.setSelected(true);
                        }
                    }
                };

            // Display control panel
            case (CONTROLS):
                return new AbstractAction("Control panel ...")
                {
                    public void actionPerformed(ActionEvent ae)
                    {
                        geminiUI.getControlPanel().setVisible(true);
                    }
                };

            // Done with control panel
            case (DONE_CONTROLS):
                return new AbstractAction("Done")
                {
                    public void actionPerformed(ActionEvent ae)
                    {
                        geminiUI.getControlPanel().setVisible(false);
                    }
                };

            // Bad argument
            default:
                throw new Exception("Invalid Action type requested: " + type);
        }
    }

    private static class FileExtFilter extends FileFilter
    {
        private String description = null;
        private String [] extensions = null;

        public FileExtFilter(String [] extensions, String description)
        {
            this.extensions = new String [extensions.length];

            for (int i = 0; i < extensions.length; i++)
                this.extensions[i] = "." + extensions[i].toLowerCase();
            this.description = description;
        }

        public String getDescription()
        {
            return description;
        }

        public boolean accept(File file)
        {
            if (file == null) 
                return false;
            else if (file.isDirectory())
                return true;

            String name = file.getName().toLowerCase();

            for (int i = 0; i < extensions.length; i++)
                if (name.endsWith(extensions[i]))
                    return true;

            return false;
        }
    }
}
