/*
 * This file is part of Gemini. Copyright (c) 2001, 2002 Genome
 * Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.psu.gemini.gui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.Image;
import java.text.FieldPosition;
import java.text.NumberFormat;
import java.util.Hashtable;
import javax.swing.*;
import javax.swing.ImageIcon;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.biojava.bio.seq.FeatureFilter;

import uk.ac.sanger.util.gui.ImageFactory;
import uk.ac.sanger.util.gui.ImageFactoryDefaultImpl;

/**
 * <code>ControlPanel</code> is a component which contains controls
 * for setting the main display's scale and score filters.
 *
 * @author Keith James
 */
final class ControlPanel extends JFrame
{
    static final int MIN_SCORE = 0;
    static final int MAX_SCORE = 1;

    private static Dimension RIGID_DIM = new Dimension(1, 1);

    private GeminiUI      geminiUI;
    private ActionFactory acFactory;

    private JSlider scaleSlider;
    private JLabel  scaleLabel;
    private JSlider maxScoreSlider;
    private JLabel  maxScoreLabel;
    private JSlider minScoreSlider;
    private JLabel  minScoreLabel;

    private JButton doneButton;

    private ImageFactory imFactory;
    private Image        magPlusIcon;
    private Image        magMinusIcon;
    private Image        filterIcon;

    /**
     * Creates a new <code>ControlPanel</code>.
     *
     * @param geminiUI a <code>GeminiUI</code>.
     *
     * @exception Exception if an error occurs.
     */
    ControlPanel(GeminiUI geminiUI) throws Exception
    {
        super("Control Panel");
        this.geminiUI = geminiUI;
        acFactory = new ActionFactory(geminiUI);

        initComponents();
    }

    /**
     * <code>getScaleSlider</code> returns a reference to the scale
     * slider.
     *
     * @return a <code>JSlider</code>.
     */
    JSlider getScaleSlider()
    {
        return scaleSlider;
    }

    /**
     * <code>getScaleLabel</code> returns a reference to the label
     * which displays the current scale.
     *
     * @return a <code>JLabel</code>.
     */
    JLabel getScaleLabel()
    {
        return scaleLabel;
    }

    /**
     * <code>getMaxScoreSlider</code> returns a reference to the
     * maximum score slider.
     *
     * @return a <code>JSlider</code>.
     */
    JSlider getMaxScoreSlider()
    {
        return maxScoreSlider;
    }

    /**
     * <code>getMinScoreSlider</code> returns a reference to the
     * minimum score slider.
     *
     * @return a <code>JSlider</code>.
     */
    JSlider getMinScoreSlider()
    {
        return minScoreSlider;
    }

    /**
     * <code>getMaxScoreLabel</code>returns a reference to the label
     * which displays the current maximum score.
     *
     * @return a <code>JLabel</code>.
     */
    JLabel getMaxScoreLabel()
    {
        return maxScoreLabel;
    }

    /**
     * <code>getMinScoreLabel</code> returns a reference to the label
     * which displays the current minimum score.
     *
     * @return a <code>JLabel</code>.
     */
    JLabel getMinScoreLabel()
    {
        return minScoreLabel;
    }

    private void initComponents() throws Exception
    {
        imFactory = ImageFactoryDefaultImpl.getInstance();
        magPlusIcon  = imFactory.createImage("viewmag+_22x22");
        magMinusIcon = imFactory.createImage("viewmag-_22x22");
        filterIcon   = imFactory.createImage("filter_22x22");

        setSize(250, 350);
        setResizable(false);

        // Fonts
        Font font = new Font("sanserif", Font.BOLD, 11);

        // Formatter
        NumberFormat nFormat = NumberFormat.getInstance();
        nFormat.setMaximumFractionDigits(1);

        // Containing panel
        JPanel controls = new JPanel();
        controls.setLayout(new GridLayout(4, 1));

        // Scale label
        String scSuffix = "bp/pixel";
        scaleLabel = new ScaleSliderLabel(geminiUI,
                                          scSuffix, false,
                                          nFormat,
                                          new FieldPosition(NumberFormat.INTEGER_FIELD));
        double s = geminiUI.calculateScale(GeminiUI.INITIAL_SCALE);
        scaleLabel.setFont(font);
        scaleLabel.setText(nFormat.format(1 / s) + " " + scSuffix);
        scaleLabel.setAlignmentX(Component.CENTER_ALIGNMENT);

        // Min score label
        minScoreLabel = new ScoreSliderLabel("Scores >=", true,
                                             nFormat, new FieldPosition(NumberFormat.INTEGER_FIELD));
        minScoreLabel.setFont(font);
        minScoreLabel.setAlignmentX(Component.CENTER_ALIGNMENT);

        // Max score label
        maxScoreLabel = new ScoreSliderLabel("Scores <=", true,
                                             nFormat, new FieldPosition(NumberFormat.INTEGER_FIELD));
        maxScoreLabel.setFont(font);
        maxScoreLabel.setAlignmentX(Component.CENTER_ALIGNMENT);

        JPanel p = new JPanel();
        p.setBorder(new TitledBorder(new EtchedBorder(), "Selected display"));
        p.setLayout(new BoxLayout(p, BoxLayout.Y_AXIS));

        p.add(scaleLabel);
        p.add(minScoreLabel);
        p.add(maxScoreLabel);
        controls.add(p);

        // Scale slider
        scaleSlider = new JSlider(SwingConstants.HORIZONTAL,
                                  5, 100, GeminiUI.INITIAL_SCALE);

        Hashtable scaleTable = new Hashtable();
        scaleTable.put(new Integer(10), new JLabel(new ImageIcon(magPlusIcon)));
        scaleTable.put(new Integer(90), new JLabel(new ImageIcon(magMinusIcon)));
        scaleSlider.setLabelTable(scaleTable);
        scaleSlider.setPaintLabels(true);
        scaleSlider.addChangeListener(new ScaleSliderListener());
        scaleSlider.addChangeListener((ChangeListener) scaleLabel);

        p = new JPanel();
        p.setBorder(new TitledBorder(new EtchedBorder(), "Scale"));
        p.setLayout(new BoxLayout(p, BoxLayout.Y_AXIS));
        p.add(Box.createRigidArea(RIGID_DIM));
        p.add(scaleSlider);
        p.add(Box.createRigidArea(RIGID_DIM));
        controls.add(p);

        // Min score slider
        minScoreSlider = new JSlider(SwingConstants.HORIZONTAL, 0, 0, 0);
        minScoreSlider.setPaintLabels(true);
        minScoreSlider.addChangeListener(new ScoreSliderListener(MIN_SCORE));
        minScoreSlider.addChangeListener((ChangeListener) minScoreLabel);

        p = new JPanel();
        p.setBorder(new TitledBorder(new EtchedBorder(), "Minimum score displayed"));
        p.setLayout(new BoxLayout(p, BoxLayout.Y_AXIS));
        p.add(Box.createRigidArea(RIGID_DIM));
        p.add(minScoreSlider);
        p.add(Box.createRigidArea(RIGID_DIM));
        controls.add(p);

        // Max score slider
        maxScoreSlider = new JSlider(SwingConstants.HORIZONTAL, 0, 0, 0);
        maxScoreSlider.setPaintLabels(true);
        maxScoreSlider.addChangeListener(new ScoreSliderListener(MAX_SCORE));
        maxScoreSlider.addChangeListener((ChangeListener) maxScoreLabel);

        p = new JPanel();
        p.setBorder(new TitledBorder(new EtchedBorder(), "Maximum score displayed"));
        p.setLayout(new BoxLayout(p, BoxLayout.Y_AXIS));
        p.add(Box.createRigidArea(RIGID_DIM));
        p.add(maxScoreSlider);
        p.add(Box.createRigidArea(RIGID_DIM));
        controls.add(p);

        getContentPane().add(controls, BorderLayout.CENTER);

        doneButton = new JButton(acFactory.createAction(ActionFactory.DONE_CONTROLS));
        getContentPane().add(doneButton, BorderLayout.SOUTH);
    }

    /**
     * <code>ScaleSliderListener</code> which updates the sequence
     * panels when the slider is used to change the scale.
     */
    private class ScaleSliderListener implements ChangeListener
    {
	public void stateChanged(ChangeEvent ce)
	{
	    JSlider source = (JSlider) ce.getSource();
	    if (! source.getValueIsAdjusting())
	    {
		int val = source.getValue();
		double s = geminiUI.calculateScale(val);

                geminiUI.getXSequencePanel().setScale(s);
                geminiUI.getYSequencePanel().setScale(s);
                geminiUI.getPairwisePanel().setScale(s);

                geminiUI.updateScrollBars();
            }
        }
    }

    /**
     * <code>ScoreSliderListener</code> which updates the score
     * filter.
     */
    private class ScoreSliderListener implements ChangeListener
    {
        private int type;

        public ScoreSliderListener(int type)
        {
            this.type = type;
        }

        public void stateChanged(ChangeEvent ce)
        {
            JSlider source = (JSlider) ce.getSource();
            if (! source.getValueIsAdjusting())
            {
                geminiUI.setFeatureFilter(new FeatureFilter.ByPairwiseScore((double) minScoreSlider.getValue(),
                                                                            (double) maxScoreSlider.getValue()));
            }
        }
    }
}
