/*
 * This file is part of Gemini. Copyright (c) 2001, 2002 Genome
 * Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.psu.gemini.gui;

import java.awt.BorderLayout;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import org.biojava.bio.Annotation;
import org.biojava.bio.seq.Feature;

/**
 * <code>FeatureViewer</code> is a simple tree viewer for displaying
 * feature annotation.
 *
 * @author Keith James
 */
public class FeatureViewer extends JFrame
{
    private DefaultTreeModel       model;
    private DefaultMutableTreeNode root;
    private JTree                  tree;
    private JScrollPane            treePane;
    private Feature                feature;

    private TreeModelListener repaintListener = new TreeModelListener()
    {
        public void treeNodesChanged(TreeModelEvent tme)
        {
            repaint();
        }

        public void treeNodesInserted(TreeModelEvent tme)
        {
            treeNodesChanged(tme);
        }
 
        public void treeNodesRemoved(TreeModelEvent tme)
        {
            treeNodesChanged(tme);
        }

        public void treeStructureChanged(TreeModelEvent tme)
        {
            treeNodesChanged(tme);
        }
    };

    public FeatureViewer()
    {
        initComponents();
    }

    private void initComponents()
    {
        root  = new DefaultMutableTreeNode();
        model = new DefaultTreeModel(root);
        tree  = new JTree(model);

        treePane = new JScrollPane(tree);

        tree.putClientProperty("JTree.lineStyle", "Angled");
        tree.getSelectionModel().setSelectionMode
            (TreeSelectionModel.SINGLE_TREE_SELECTION);

        // Listen for changes to the TreeModel
        model.addTreeModelListener(repaintListener);

        getContentPane().add(treePane, BorderLayout.CENTER);
    }

    /**
     * <code>getFeature</code> returns the feature whose annotation is
     * currently being displayed.
     *
     * @return a <code>Feature</code>.
     */
    public Feature getFeature()
    {
        return feature;
    }

    /**
     * <code>setFeature</code> sets the feature whose annotation is to
     * be displayed.
     *
     * @param feature a <code>Feature</code> object.
     */
    public void setFeature(final Feature feature)
    {
        this.feature = feature;

        setTitle("View of Feature "
                 + feature.getType()
                 + " ("
                 + feature.getSource()
                 + ")");

        // Clear the root node's children
        root.removeAllChildren();
        // Set the root node's user object to this feature
        root.setUserObject(feature);
        // Add the new child nodes
        populateTree(root, feature.getAnnotation());
        // Fire an event to indicate a total tree change
        model.nodeStructureChanged(root);
        // Expand the tree
        for (int i = tree.getRowCount(); i-- >= 0;)
            tree.expandRow(i);
    }

    private void populateTree(final DefaultMutableTreeNode root,
                              final Annotation             annotation)
    {
        Map m = annotation.asMap();

        for (Iterator ki = m.keySet().iterator(); ki.hasNext();)
        {
            Object key = ki.next();

            // Skip the internal data
            if (key == Feature.PROPERTY_DATA_KEY)
                continue;

            Object val = m.get(key);

            DefaultMutableTreeNode node = new DefaultMutableTreeNode(key);
            populateNode(node, val);

            root.add(node);
        }
    }

    private void populateNode(final DefaultMutableTreeNode node,
                              final Object                 o)
    {
        if (o instanceof String)
            node.add(new DefaultMutableTreeNode((String) o));

        if (o instanceof Collection)
        {
            for (Iterator ki = ((Collection) o).iterator(); ki.hasNext();)
            {
                populateNode(node, ki.next());
            }
        }
    }
}
