/*
 * This file is part of Gemini. Copyright (c) 2001, 2002 Genome
 * Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.psu.gemini.gui;

import java.awt.*;
import java.awt.event.*;
import java.beans.PropertyChangeEvent;
import java.net.URL;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import javax.swing.*;
import javax.swing.border.BevelBorder;
import javax.swing.border.EtchedBorder;

import org.biojava.bio.BioError;
import org.biojava.bio.gui.sequence.*;
import org.biojava.bio.seq.Feature;
import org.biojava.bio.seq.FeatureFilter;
import org.biojava.bio.seq.FeatureHolder;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.homol.SimilarityPairFeature;
import org.biojava.utils.ChangeVetoException;

import uk.ac.sanger.util.gui.ImageFactory;
import uk.ac.sanger.util.gui.ImageFactoryDefaultImpl;

import uk.ac.sanger.psu.gemini.DataHandler;

/**
 * <code>GeminiUI</code> is the main user interface. It contains the
 * pair plot itself, the X and Y feature displays, the main menus and
 * scrollbars for navigation.
 *
 * @author Keith James
 */
public final class GeminiUI extends JFrame
{
    // The starting scale for all sequence panels
    static final int INITIAL_SCALE = 40;

    // The scroll block size for both X and Y
    private static final int SCROLL_BLOCK_INC = 100;

    // Filter to exclud all features except SimilarityPairFeatures
    private static FeatureFilter spf;

    static
    {
        String className =
            "org.biojava.bio.seq.homol.SimilarityPairFeature";

        try
        {
            spf = new FeatureFilter.ByClass(Class.forName(className));
        }
        catch (Exception e)
        {
            throw new BioError(e, "Failed to load Class for " + className);
        }
    }

    // The main menus
    private JMenuBar menuBar;
    private JMenu fileMenu;
    private JMenu exportMenu;
    private JMenu optionMenu;
    private JMenu aboutMenu;

    // Action factory
    private ActionFactory acFactory;
    // Sequence renderer factory
    private SequenceRendererFactory srFactory;
    // SequenceViewerListener factory
    private SequenceViewerListenerFactory svFactory;
    // Toolbar factory
    private JToolBarFactory tbFactory;

    // Panel holding main GridBagged interface inside a BorderLayout
    // for the JToolBar
    private JPanel innerPanel;

    // Sequence panels and corresponding scrollbars
    private PairwiseSequencePanel   plotPanel;
    private TranslatedSequencePanel vertPanel;
    private TranslatedSequencePanel horizPanel;
    private JScrollBar vertBar;
    private JScrollBar horizBar;

    // Filtering renderer
    private PairwiseFilteringRenderer filteringRenderer;

    // Scale for all sequence panels
    private double basesPerPixel;

    // Rendering hints for all sequence panels
    private RenderingHints antiAliasHint;

    // The feature viewer
    private FeatureViewer fViewer;

    // The control panel
    private ControlPanel conPanel;

    // Corner info panel
    private InfoPanel infoPanel;

    // Dialogs
    private OpenDialog openDialog;
    private AboutDialog aboutDialog;
    private JDialog loadProgress;

    // The main toolbar
    private JToolBar mainBar;

    // JPEG, PNG and SVG exporters
    private JPEGExporter jpegExporter;
    private SVGExporter svgExporter;

    // Datahandler
    private DataHandler dataHandler;

    // File chooser used throughout
    private JFileChooser fileChooser;

    // Fonts
    private Font panelFont;

    // Images from central image factory
    private ImageFactory imFactory;
    private Image dragCursor;
    private Image openIcon;
    private Image quitIcon;
    private Image exportIcon;
    private Image controlIcon;
    private Image aboutIcon;

    // Enabled (data loaded) flag
    private boolean isEnabled;

    /**
     * Creates a new <code>GeminiUI</code> object.
     *
     * @param title a <code>String</code>.
     *
     * @exception Exception if an error occurs.
     */
    public GeminiUI(String title) throws Exception
    {
        super(title);
        initComponents();
    }

    /**
     * <code>setXSequence</code> sets the X-axis sequence for the
     * sequence panel.
     *
     * @param seq a <code>Sequence</code>.
     */
    public void setXSequence(Sequence seq) throws Exception
    {
        try
        {
            removeSourceFeatures(seq);
            horizPanel.setSequence(seq);
            horizPanel.setRenderer(srFactory.createRenderer(SequenceRendererFactory.BASIC_HORIZONTAL));
        }
        catch (ChangeVetoException cve)
        {
            throw new BioError(cve, "Assertion failure: unexpected locking encountered on horizontal sequence or panel");
        }
    }

    /**
     * <code>setYSequence</code> sets the Y-axis sequence for the
     * sequence panel.
     *
     * @param seq a <code>Sequence</code>.
     */
    public void setYSequence(Sequence seq) throws Exception
    {
        try
        {
            removeSourceFeatures(seq);
            vertPanel.setSequence(seq);
            vertPanel.setRenderer(srFactory.createRenderer(SequenceRendererFactory.BASIC_VERTICAL));
        }
        catch (ChangeVetoException cve)
        {
            throw new BioError(cve, "Assertion failure: unexpected locking encountered on vertical sequence or panel");
        }
    }

    /**
     * <code>setPlotXSequence</code> sets the X sequence of the
     * plot panel.
     *
     * @param seq a <code>Sequence</code>.
     */
    public void setPlotXSequence(Sequence seq)
    {
        plotPanel.setSequence(seq);

        // Calculate the maximum and minimum scores observed
        FeatureHolder simPairFeatures = seq.filter(spf, false);

        double maxHitScore = 0;
        double minHitScore = 0;

        for (Iterator spfi = simPairFeatures.features(); spfi.hasNext();)
        {
            SimilarityPairFeature f = (SimilarityPairFeature) spfi.next();

            double score = f.getScore();

            maxHitScore = Math.max(maxHitScore, score);
            minHitScore = Math.min(minHitScore, score);
        }

        Hashtable labels = new Hashtable(2);
        Integer intLabel;
        intLabel = new Integer(0);
        labels.put(intLabel, new JLabel(intLabel.toString()));
        intLabel = new Integer((int) maxHitScore);
        labels.put(intLabel, new JLabel(intLabel.toString()));

        JSlider slider;
        slider = conPanel.getMaxScoreSlider();
        slider.setMaximum((int) maxHitScore);
        slider.setValue((int) maxHitScore);
        slider.setLabelTable(labels);

        slider = conPanel.getMinScoreSlider();
        slider.setMaximum((int) maxHitScore);
        slider.setValue(0);
        slider.setLabelTable(labels);

        conPanel.getMaxScoreLabel().setText("Scores <=" + (int) maxHitScore);
        conPanel.getMinScoreLabel().setText("Scores >=" + (int) minHitScore);
    }

    /**
     * <code>setPlotYSequence</code> sets the Y sequence of the
     * plot panel.
     *
     * @param seq a <code>Sequence</code>.
     */
    public void setPlotYSequence(Sequence seq)
    {
        plotPanel.setSecondarySequence(seq);
    }

    public RenderingHints getRenderingHints()
    {
        return antiAliasHint;
    }

    /**
     * <code>getPairwisePanel</code> returns the plot panel.
     *
     * @return a <code>PairwiseSequencePanel</code>.
     */
    public PairwiseSequencePanel getPairwisePanel()
    {
        return plotPanel;
    }

    /**
     * <code>getFeatureFilter</code> returns the filter currently in
     * use by the plot panel.
     *
     * @return a <code>FeatureFilter</code>.
     */
    public FeatureFilter getFeatureFilter()
    {
        return filteringRenderer.getFilter();
    }

    /**
     * <code>setFeatureFilter</code> sets the filter to be used
     * by the plot panel.
     *
     * @param filter a <code>FeatureFilter</code>.
     */
    public void setFeatureFilter(FeatureFilter filter)
    {
        try
        {
            filteringRenderer.setFilter(filter);
        }
        catch (ChangeVetoException cve)
        {
            throw new BioError(cve, "Assertion failure: unexpected locking encountered on filtering renderer");
        }
    }

    /**
     * <code>getControlPanel</code> returns the control panel.
     *
     * @return a <code>ControlPanel</code>.
     */
    public ControlPanel getControlPanel()
    {
        return conPanel;
    }

    /**
     * <code>getInfoPanel</code> returns the info panel.
     *
     * @return an <code>InfoPanel</code> value.
     */
    public InfoPanel getInfoPanel()
    {
        return infoPanel;
    }

    /**
     * <code>getXSequencePanel</code> returns the X-axis panel.
     *
     * @return a <code>TranslatedSequencePanel</code>.
     */
    public TranslatedSequencePanel getXSequencePanel()
    {
        return horizPanel;
    }

    /**
     * <code>getYSequencePanel</code> returns the Y-axis panel.
     *
     * @return a <code>TranslatedSequencePanel</code>.
     */
    public TranslatedSequencePanel getYSequencePanel()
    {
        return vertPanel;
    }

    /**
     * <code>getOpenDialog</code> returns the current open dialog.
     *
     * @return an <code>OpenDialog</code>.
     */
    public OpenDialog getOpenDialog()
    {
        return openDialog;
    }

    /**
     * <code>getComponentsEnabled</code> returns whether the plotting
     * areas, scrollbars and data-related widgets are enabled.
     *
     * @return a <code>boolean</code> value.
     */
    public boolean getComponentsEnabled()
    {
        return isEnabled;
    }

    /**
     * <code>setComponentsEnabled</code> enables or disables the
     * plotting areas, scrollbars and data-related widgets.
     *
     * @param value a <code>boolean</code> value.
     */
    public void setComponentsEnabled(boolean value)
    {
        isEnabled = value;

        infoPanel.setVisible(value);
        plotPanel.setVisible(value);

        horizBar.setEnabled(value);
        vertBar.setEnabled(value);
        // Export menu item
        fileMenu.getItem(1).setEnabled(value);
        // Option menu
        optionMenu.setEnabled(value);

        if (isEnabled)
            updateScrollBars();
    }

    /**
     * <code>updateScrollBars</code> updates the scrollbars to reflect
     * the current viewing offsets and sequences lengths.
     */
    public void updateScrollBars()
    {
        if (horizPanel.getSequence() == null ||
            vertPanel.getSequence() == null)
            return;

        int seqVis, seqTot, extent;
        int barMin, barMax, barVal;

        // Make the horiz bar reflect the sequence length and scale
        seqVis = horizPanel.getVisibleSymbolCount() -
            horizPanel.getSymbolTranslation();
        seqTot = horizPanel.getSequence().length();

        extent = Math.min(seqVis, seqTot - 1);

        barMin = horizBar.getMinimum();
        barMax = horizBar.getMaximum();
        barVal = horizBar.getValue();

        if ((extent + barVal) >= barMax)
            barVal = Math.max(0, barMax - extent);

        horizBar.setValues(barVal, extent, 0, seqTot);

         // Make the vert bar reflect the sequence length and scale
        seqVis = vertPanel.getVisibleSymbolCount() -
            vertPanel.getSymbolTranslation();
        seqTot = vertPanel.getSequence().length();

        extent = Math.min(seqVis, seqTot - 1);

        barMin = vertBar.getMinimum();
        barMax = vertBar.getMaximum();
        barVal = vertBar.getValue();

        if ((extent + barVal) >= barMax)
            barVal = Math.max(0, barMax - extent);

        vertBar.setValues(barVal, extent, 0, seqTot);
        vertBar.repaint();
        horizBar.repaint();
    }

    /**
     * <code>getXScrollBar</code> returns the X-axis scrollbar.
     *
     * @return a <code>JScrollBar</code>.
     */
    JScrollBar getXScrollBar()
    {
        return horizBar;
    }

    /**
     * <code>getYScrollBar</code> returns the Y-axis scrollbar.
     *
     * @return a <code>JScrollBar</code>.
     */
    JScrollBar getYScrollBar()
    {
        return vertBar;
    }

    /**
     * <code>getFeatureViewer</code> returns the current feature
     * viewer component.
     *
     * @return a <code>FeatureViewer</code>.
     */
    FeatureViewer getFeatureViewer()
    {
        return fViewer;
    }

    /**
     * <code>getJPEGExporter</code> returns the current JPEG exporter.
     *
     * @return a <code>JPEGExporter</code>.
     */
    JPEGExporter getJPEGExporter()
    {
        return jpegExporter;
    }

    /**
     * <code>getSVGExporter</code> returns the current SVG exporter.
     *
     * @return a <code>SVGExporter</code>.
     */
    SVGExporter getSVGExporter()
    {
        return svgExporter;
    }

    /**
     * <code>getDataHandler</code> returns the current sequence and
     * comparison data file handler.
     *
     * @return a <code>DataHandler</code>.
     */
    DataHandler getDataHandler()
    {
        return dataHandler;
    }

    /**
     * <code>getFileChooser</code> returns the current file chooser.
     *
     * @return a <code>JFileChooser</code>.
     */
    JFileChooser getFileChooser()
    {
        return fileChooser;
    }

    /**
     * <code>getAboutDialog</code> returns the current about dialog.
     *
     * @return an <code>AboutDialog</code>.
     */
    AboutDialog getAboutDialog()
    {
        return aboutDialog;
    }

    /**
     * <code>calculateScale</code> calculates a scale value in pixels
     * per base from an integer.
     *
     * @param val an <code>int</code>.
     *
     * @return a <code>double</code>.
     */
    double calculateScale(int val)
    {
        return Math.exp(-val / 7.0) * 20;
    }

    /**
     * <code>initComponents</code> creates and initialises GUI
     * components.
     *
     * @exception Exception if an error occurs.
     */
    private void initComponents() throws Exception
    {
        imFactory = ImageFactoryDefaultImpl.getInstance();
        dragCursor  = imFactory.createImage("drag_cursor");
        openIcon    = imFactory.createImage("fileopen_22x22");
        quitIcon    = imFactory.createImage("exit_22x22");
        exportIcon  = imFactory.createImage("fileexport_22x22");
        controlIcon = imFactory.createImage("gear_22x22");
        aboutIcon   = imFactory.createImage("info_22x22");

        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        acFactory = new ActionFactory(this);
        srFactory = new SequenceRendererFactory(this);
        svFactory = new SequenceViewerListenerFactory(this);
        tbFactory = new JToolBarFactory(this);

        conPanel   = new ControlPanel(this);
        infoPanel  = new InfoPanel(this);
        menuBar    = new JMenuBar();
        innerPanel = new JPanel(new GridBagLayout());

        plotPanel  = new PairwiseSequencePanel();
        vertBar    = new JScrollBar(SwingConstants.VERTICAL);
        horizBar   = new JScrollBar(SwingConstants.HORIZONTAL);
        vertPanel  = new TranslatedSequencePanel();
        horizPanel = new TranslatedSequencePanel();

        loadProgress = new JDialog(this, "Loading file");
        dataHandler = new DataHandler(loadProgress);

        openDialog  = new OpenDialog(this);
        aboutDialog = new AboutDialog(this);

        panelFont = new Font("courier", Font.PLAIN, 9);

        innerPanel.setOpaque(false);
        innerPanel.setBorder(new EtchedBorder());
        getContentPane().add(innerPanel, BorderLayout.CENTER);

        Map hints = new HashMap();
        hints.put(RenderingHints.KEY_ANTIALIASING,
                  RenderingHints.VALUE_ANTIALIAS_ON);
        antiAliasHint = new RenderingHints(hints);

        Thread createViewer = new Thread()
        {
            public void run()
            {
                fViewer = new FeatureViewer();
                fViewer.setSize(350, 300);
                fViewer.setVisible(false);
            }
        };

        SwingUtilities.invokeLater(createViewer);

        // File menu
        fileMenu = new JMenu();
        fileMenu.setText("File");
        fileMenu.setMnemonic('f');
        fileMenu.add(new JMenuItem(acFactory.createAction(ActionFactory.OPEN)));
        fileMenu.getItem(0).setIcon(new ImageIcon(openIcon));
        fileMenu.getItem(0).setMnemonic('o');
        fileMenu.getItem(0).setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O,
                                                                  InputEvent.CTRL_MASK));

        // Export menu
        exportMenu = new JMenu();
        exportMenu.setText("Export ...");
        exportMenu.setMnemonic('e');
        jpegExporter = new JPEGExporter(this);
        exportMenu.add(new JMenuItem(acFactory.createAction(ActionFactory.EXPORT_JPEG)));

        // Check for availability of Batik
        try
        {
            svgExporter = new SVGExporter(this);
            exportMenu.add(new JMenuItem(acFactory.createAction(ActionFactory.EXPORT_SVG)));
        }
        catch (NoClassDefFoundError ncdfe)
        {
            exportMenu.add(new JMenuItem("SVG format"));
            exportMenu.getItem(1).setEnabled(false);
        }

        fileMenu.add(exportMenu);
        fileMenu.getItem(1).setIcon(new ImageIcon(exportIcon));
        fileMenu.getItem(1).setMnemonic('e');

        fileMenu.add(new JMenuItem(acFactory.createAction(ActionFactory.QUIT)));
        fileMenu.getItem(2).setIcon(new ImageIcon(quitIcon));
        fileMenu.getItem(2).setMnemonic('q');
        fileMenu.getItem(2).setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q,
                                                                  InputEvent.CTRL_MASK));

        Thread createChooser = new Thread()
        {
            public void run()
            {
                fileChooser = new JFileChooser();
            }
        };

        SwingUtilities.invokeLater(createChooser);

        // Option menu
        optionMenu = new JMenu();
        optionMenu.setText("Options");
        optionMenu.setMnemonic('o');
        JCheckBoxMenuItem antiAliasSeq =
            new JCheckBoxMenuItem(acFactory.createAction(ActionFactory.ANTIALIAS_SEQ));
        JCheckBoxMenuItem antiAliasCmp =
            new JCheckBoxMenuItem(acFactory.createAction(ActionFactory.ANTIALIAS_CMP));
        JMenuItem conPanelVis =
            new JMenuItem(acFactory.createAction(ActionFactory.CONTROLS));

        // Default is not to antialias
        antiAliasSeq.setState(false);
        antiAliasCmp.setState(false);
        optionMenu.add(antiAliasSeq);
        optionMenu.add(antiAliasCmp);
        optionMenu.addSeparator();
        optionMenu.add(conPanelVis);
        optionMenu.getItem(0).setMnemonic('e');
        optionMenu.getItem(0).setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_E,
                                                                    InputEvent.CTRL_MASK));
        optionMenu.getItem(1).setMnemonic('m');
        optionMenu.getItem(1).setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_M,
                                                                    InputEvent.CTRL_MASK));

        optionMenu.getItem(3).setIcon(new ImageIcon(controlIcon));
        optionMenu.getItem(3).setMnemonic('n');
        optionMenu.getItem(3).setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_N,
                                                                    InputEvent.CTRL_MASK));

        // About menu
        aboutMenu = new JMenu();
        aboutMenu.setText("About");
        aboutMenu.setMnemonic('a');
        aboutMenu.add(new JMenuItem(acFactory.createAction(ActionFactory.ABOUT)));
        aboutMenu.getItem(0).setIcon(new ImageIcon(aboutIcon));

        // MenuBar
        menuBar.setBorder(new BevelBorder(BevelBorder.RAISED));
        menuBar.add(fileMenu);
        menuBar.add(optionMenu);
        menuBar.add(aboutMenu);
        setJMenuBar(menuBar);

        // Main toolbar
        mainBar = tbFactory.makeJToolBar(JToolBarFactory.MAIN);
        getContentPane().add(mainBar, BorderLayout.NORTH);

        GridBagConstraints constraints = new GridBagConstraints();

        // Info panel
        constraints.gridx = 0;
        constraints.gridy = 0;
        constraints.fill   = GridBagConstraints.BOTH;
        constraints.anchor = GridBagConstraints.NORTHWEST;
        constraints.weightx = 0;
        constraints.weighty = 0;
        constraints.gridheight = 2;
        constraints.gridwidth  = 2;

        infoPanel.setOpaque(true);
        infoPanel.setBackground(Color.white);
        infoPanel.setBorder(new EtchedBorder());

        // Innerpanel contains the infoPanel
        innerPanel.add(infoPanel, constraints);

        // Horiz scrollbar
        constraints.gridx = 2;
        constraints.gridy = 0;
        constraints.fill   = GridBagConstraints.HORIZONTAL;
        constraints.anchor = GridBagConstraints.NORTH;
        constraints.weightx = 0.0;
        constraints.weighty = 0.0;
        constraints.gridheight = 1;
        constraints.gridwidth  = 1;

        // Innerpanel contains the horizBar
        innerPanel.add(horizBar, constraints);

        horizBar.setBlockIncrement(SCROLL_BLOCK_INC);
        horizBar.addAdjustmentListener(new BarListener(horizPanel));
        horizBar.addAdjustmentListener(new PairwiseBarListener(plotPanel));

        // Horiz panel
        constraints.gridx = 2;
        constraints.gridy = 1;
        constraints.fill   = GridBagConstraints.BOTH;
        constraints.anchor = GridBagConstraints.SOUTH;
        constraints.weightx = 1.0;
        constraints.weighty = 0.0;

        horizPanel.setOpaque(true);
        horizPanel.setBackground(Color.white);
        horizPanel.setFont(panelFont);

        // Innerpanel contains the horizPanel
        innerPanel.add(horizPanel, constraints);

        horizPanel.setScale(calculateScale(INITIAL_SCALE));
        horizPanel.setDirection(SequenceRenderContext.HORIZONTAL);
        horizPanel.addSequenceViewerListener(svFactory.createListener(SequenceViewerListenerFactory.FEATURE_VIEWER));

        // Vert scrollbar
        constraints.gridx = 0;
        constraints.gridy = 2;
        constraints.fill   = GridBagConstraints.VERTICAL;
        constraints.anchor = GridBagConstraints.NORTH;
        constraints.weightx = 0.0;
        constraints.weighty = 0.0;

        // Innerpanel contains the vertBar
        innerPanel.add(vertBar, constraints);

        vertBar.setBlockIncrement(SCROLL_BLOCK_INC);
        vertBar.addAdjustmentListener(new BarListener(vertPanel));
        vertBar.addAdjustmentListener(new PairwiseBarListener(plotPanel));

        // Vert panel
        constraints.gridx = 1;
        constraints.gridy = 2;
        constraints.fill   = GridBagConstraints.BOTH;
        constraints.anchor = GridBagConstraints.EAST;
        constraints.weightx = 0.0;
        constraints.weighty = 1.0;

        vertPanel.setOpaque(true);
        vertPanel.setBackground(Color.white);
        vertPanel.setFont(panelFont);

        // Innerpanel contains the vertPanel
        innerPanel.add(vertPanel, constraints);

        vertPanel.setScale(calculateScale(INITIAL_SCALE));
        vertPanel.setDirection(SequenceRenderContext.VERTICAL);
        vertPanel.addSequenceViewerListener(svFactory.createListener(SequenceViewerListenerFactory.FEATURE_VIEWER));

        // Plot panel
        constraints.gridx = 2;
        constraints.gridy = 2;
        constraints.fill   = GridBagConstraints.BOTH;
        constraints.anchor = GridBagConstraints.NORTHWEST;
        constraints.weightx = 1.0;
        constraints.weighty = 1.0;

        plotPanel.setOpaque(true);
        plotPanel.setBackground(Color.white);
        plotPanel.setFont(panelFont);
        plotPanel.setBorder(new EtchedBorder());

        // Innerpanel contains the plotPanel
        innerPanel.add(plotPanel, constraints);

        // Primary sequence will be on the X-axis
        plotPanel.setScale(calculateScale(INITIAL_SCALE));
        plotPanel.setDirection(SequenceRenderContext.HORIZONTAL);

        try
        {
            PairwiseOverlayRenderer overlay = new PairwiseOverlayRenderer();
            overlay.addRenderer(new CrosshairRenderer());

            filteringRenderer =
                new PairwiseFilteringRenderer(new PairwiseDiagonalRenderer(),
                                              new FeatureFilter.ByPairwiseScore(0, Double.MAX_VALUE),
                                              false);
            overlay.addRenderer(filteringRenderer);
            plotPanel.setRenderer(overlay);
        }
        catch (ChangeVetoException cve)
        {
            throw new BioError(cve, "Assertion failure: unexpected locking encountered on pairwise renderer or panel");
        }

        // Add a listener which keeps the scrollbars updated for the
        // whole application during window resizes
        plotPanel.addHierarchyBoundsListener(new HierarchyBoundsListener()
         {
            public void ancestorResized(HierarchyEvent he)
            {
                updateScrollBars();
            }

            public void ancestorMoved(HierarchyEvent he) { }
        });

        plotPanel.addSequenceViewerListener(svFactory.createListener(SequenceViewerListenerFactory.DEBUG_PRINT));

        // Listener which responds to middle-button drag MouseEvents
        // by scrolling the plot panel
        SequenceViewerAdapter dragScroll = new SequenceViewerAdapter()
        {
            Cursor norm = Cursor.getDefaultCursor();
            Cursor drag =
                Toolkit.getDefaultToolkit().createCustomCursor(dragCursor,
                                                               new Point(16, 16),
                                                               "DRAG_CURSOR");
            Point start, current;

            int priDelta, secDelta;
            int priTrans, secTrans;

            public void mousePressed(SequenceViewerEvent sve)
            {
                MouseEvent me = sve.getMouseEvent();
                if (! SwingUtilities.isMiddleMouseButton(me))
                    return;
                start = me.getPoint();
                plotPanel.setCursor(drag);
            }

            public void mouseDragged(SequenceViewerEvent sve)
            {
                MouseEvent me = sve.getMouseEvent();
                if (! SwingUtilities.isMiddleMouseButton(me))
                    return;
                current = sve.getMouseEvent().getPoint();

                priDelta = plotPanel.graphicsToSequence(start) -
                    plotPanel.graphicsToSequence(current);

                secDelta = plotPanel.graphicsToSecondarySequence(start) -
                    plotPanel.graphicsToSecondarySequence(current);

                priTrans = plotPanel.getSymbolTranslation() + priDelta;
                secTrans = plotPanel.getSecondarySymbolTranslation() + secDelta;

                // Constrain values to within sequences
                priTrans = Math.min(priTrans, plotPanel.getSequence().length() - 1);
                priTrans = Math.max(priTrans, 0);
                secTrans = Math.min(secTrans, plotPanel.getSecondarySequence().length() - 1);
                secTrans = Math.max(secTrans, 0);

                plotPanel.setSymbolTranslation(priTrans);
                plotPanel.setSecondarySymbolTranslation(secTrans);

                horizPanel.setSymbolTranslation(priTrans);
                vertPanel.setSymbolTranslation(secTrans);
                horizBar.setValue(priTrans);
                vertBar.setValue(secTrans);

                // delta Point
                start = current;
            }

            public void mouseReleased(SequenceViewerEvent sve)
            {
                plotPanel.setCursor(norm);
            }
        };

        plotPanel.addSequenceViewerListener(dragScroll);
        plotPanel.addSequenceViewerMotionListener(dragScroll);

        setComponentsEnabled(false);
    }

    /**
     * <code>removeSourceFeatures</code> deletes the EMBL "source"
     * features.
     *
     * @param seq a <code>Sequence</code>.
     *
     * @exception ChangeVetoException if the features cannot be
     * removed.
     */
    private void removeSourceFeatures(Sequence seq)
        throws ChangeVetoException
    {
        FeatureHolder sources =
            seq.filter(new FeatureFilter.ByType("source"), false);

        for (Iterator fi = sources.features(); fi.hasNext();)
        {
            seq.removeFeature((Feature) fi.next());
        }
    }

    /**
     * <code>BarListener</code> for passing scrollbar motion to the
     * sequence panel.
     */
    private class BarListener implements AdjustmentListener
    {
        private TranslatedSequencePanel p;

        public BarListener(TranslatedSequencePanel p)
        {
            this.p = p;
        }

        public void adjustmentValueChanged(AdjustmentEvent ae)
        {
            JScrollBar source = (JScrollBar) ae.getAdjustable();

            // Set the new symbol translation
            p.setSymbolTranslation(source.getValue());
        }
    }

    /**
     * <code>PairwiseBarListener</code> for passing scrollbar motion
     * to the pairwise sequence panel.
     */
    private class PairwiseBarListener implements AdjustmentListener
    {
        private PairwiseSequencePanel p;

        public PairwiseBarListener(PairwiseSequencePanel p)
        {
            this.p = p;
        }

        public void adjustmentValueChanged(AdjustmentEvent ae)
        {
            JScrollBar source = (JScrollBar) ae.getAdjustable();

            // Set the new symbol translation on the correct sequence
            if (source.getOrientation() == SequenceRenderContext.HORIZONTAL)
                p.setSymbolTranslation(source.getValue());
            else
                p.setSecondarySymbolTranslation(source.getValue());
        }
    }
}
