/*
 * This file is part of Gemini. Copyright (c) 2001, 2002 Genome
 * Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.psu.gemini.gui;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridLayout;
import java.text.FieldPosition;
import java.text.NumberFormat;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.event.ChangeListener;

/**
 * <code>InfoPanel</code> displays the current scale and score filters
 * in the main UI so that they are still visible when the control
 * panel is closed.
 *
 * @author Keith James
 */
final class InfoPanel extends JPanel
{
    private GeminiUI     geminiUI;
    private ControlPanel conPanel;

    private JLabel scaleLabel;
    private JLabel maxScoreLabel;
    private JLabel minScoreLabel;

    /**
     * Creates a new <code>InfoPanel</code>.
     *
     * @param geminiUI a <code>GeminiUI</code>.
     */
    InfoPanel(GeminiUI geminiUI)
    {
        super(new GridLayout(4, 1));
        this.geminiUI = geminiUI;
        this.conPanel = geminiUI.getControlPanel();

        initComponents();
    }

    private void initComponents()
    {
        // Fonts
        Font font = new Font("sanserif", Font.PLAIN, 9);

        // Formatter
        NumberFormat nFormat = NumberFormat.getInstance();
        nFormat.setMaximumFractionDigits(1);

        // Info label
        JLabel infoLabel = new JLabel("Display settings:");
        infoLabel.setFont(font);

        // Scale label
        String scSuffix = "bp/pixel";
        scaleLabel = new ScaleSliderLabel(geminiUI,
                                          scSuffix, false,
                                          nFormat,
                                          new FieldPosition(NumberFormat.INTEGER_FIELD));
        double s = geminiUI.calculateScale(GeminiUI.INITIAL_SCALE);
        scaleLabel.setFont(font);
        scaleLabel.setText(nFormat.format(1 / s) + " " + scSuffix);
        scaleLabel.setAlignmentX(Component.CENTER_ALIGNMENT);

         // Min score label
        minScoreLabel = new ScoreSliderLabel("Scores >=", true,
                                             nFormat, new FieldPosition(NumberFormat.INTEGER_FIELD));
        minScoreLabel.setFont(font);
        minScoreLabel.setAlignmentX(Component.CENTER_ALIGNMENT);

        // Max score label
        maxScoreLabel = new ScoreSliderLabel("Scores <=", true,
                                             nFormat, new FieldPosition(NumberFormat.INTEGER_FIELD));
        maxScoreLabel.setFont(font);
        maxScoreLabel.setAlignmentX(Component.CENTER_ALIGNMENT);

        add(infoLabel);
        add(scaleLabel);
        add(minScoreLabel);
        add(maxScoreLabel);

        conPanel.getScaleSlider().addChangeListener((ChangeListener) scaleLabel);
        conPanel.getMaxScoreSlider().addChangeListener((ChangeListener) maxScoreLabel);
        conPanel.getMinScoreSlider().addChangeListener((ChangeListener) minScoreLabel);
    }
}
