/*
 * This file is part of Gemini. Copyright (c) 2001, 2002 Genome
 * Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.psu.gemini.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.File;
import java.io.IOException;
import javax.swing.*;

import org.xml.sax.SAXException;

import uk.ac.sanger.psu.gemini.DataHandler;

/**
 * <code>OpenDialog</code> handles opening of sequence and comparison
 * files.
 *
 * @author Keith James
 */
final class OpenDialog extends JDialog
{
    private GeminiUI geminiUI;
    private JFileChooser chooser;

    private File queryFile;
    private File subjectFile;
    private File comparisonFile;
    private boolean selfMatch;

    /**
     * Creates a new <code>OpenDialog</code>.
     *
     * @param geminiUI a <code>GeminiUI</code> parent.
     *
     * @exception Exception if an error occurs.
     */
    public OpenDialog(GeminiUI geminiUI) throws Exception
    {
        super(geminiUI, "Open", true);
        this.geminiUI = geminiUI;
        initComponents();
    }

    private void initComponents() throws Exception
    {
        JPanel p = new JPanel(new GridBagLayout());
        Dimension d = new Dimension(200, 20);

        final JTextField queryField = new JTextField();
        queryField.setPreferredSize(d);
        final JTextField subjectField = new JTextField();
        subjectField.setPreferredSize(d);
        final JTextField comparisonField = new JTextField();
        comparisonField.setPreferredSize(d);

        // Button to call query FileChooser
        JButton queryButton = new JButton("...");
        queryButton.addActionListener(new ActionListener()
            {
                public void actionPerformed(ActionEvent ae)
                {
                    chooser = geminiUI.getFileChooser();
                    chooser.setCurrentDirectory(new File("."));
                    if (chooser.showDialog(geminiUI, "OK") != 
                        JFileChooser.APPROVE_OPTION)
                        return;

                    queryFile = chooser.getSelectedFile();

                    try
                    {
                        queryField.setText(queryFile.getCanonicalPath());
                    }
                    catch (IOException ioe)
                    {
                        JOptionPane.showMessageDialog(OpenDialog.this,
                                                      "Unable to read file: " + ioe.getMessage(),
                                                      "Unable to read file",
                                                      JOptionPane.ERROR_MESSAGE);
                    }
                }
            });

        // Button to call subject FileChooser
        final JButton subjectButton = new JButton("...");
        subjectButton.addActionListener(new ActionListener()
            {
                public void actionPerformed(ActionEvent ae)
                {
                    chooser = geminiUI.getFileChooser();
                    chooser.setCurrentDirectory(new File("."));
                    if (chooser.showDialog(geminiUI, "OK") != 
                        JFileChooser.APPROVE_OPTION)
                        return;

                    subjectFile = chooser.getSelectedFile();

                    try
                    {
                        subjectField.setText(subjectFile.getCanonicalPath());
                    }
                    catch (IOException ioe)
                    {
                        JOptionPane.showMessageDialog(OpenDialog.this,
                                                      "Unable to read file: " + ioe.getMessage(),
                                                      "Unable to read file",
                                                      JOptionPane.ERROR_MESSAGE);
                    }
                }
            });

        // Button to call comparison FileChooser
        JButton comparisonButton = new JButton("...");
        comparisonButton.addActionListener(new ActionListener()
            {
                public void actionPerformed(ActionEvent ae)
                {
                    chooser = geminiUI.getFileChooser();
                    chooser.setCurrentDirectory(new File("."));
                    if (chooser.showDialog(geminiUI, "OK") != 
                        JFileChooser.APPROVE_OPTION)
                        return;

                    comparisonFile = chooser.getSelectedFile();

                    try
                    {
                        comparisonField.setText(comparisonFile.getCanonicalPath());
                    }
                    catch (IOException ioe)
                    {
                        JOptionPane.showMessageDialog(OpenDialog.this,
                                                      "Unable to read file: " + ioe.getMessage(),
                                                      "Unable to read file",
                                                      JOptionPane.ERROR_MESSAGE);
                    }
                }
            });

        GridBagConstraints constraints = new GridBagConstraints();

        // Query label, field and button
        constraints.insets = new Insets(2, 2, 2, 2);
        constraints.gridx = 0;
        constraints.gridy = 0;
        constraints.fill   = GridBagConstraints.HORIZONTAL;
        constraints.anchor = GridBagConstraints.WEST;
        constraints.weightx = 0.0;
        constraints.weighty = 0.0;
        constraints.gridheight = 1;
        constraints.gridwidth  = 1;
        p.add(new JLabel("Query sequence"), constraints);

        constraints.gridx = 1;
        constraints.weightx = 1.0;
        p.add(queryField, constraints);

        constraints.gridx = 2;
        constraints.weightx = 0.0;
        p.add(queryButton, constraints);

        // Subject label, field and button
        constraints.gridx = 0;
        constraints.gridy = 1;
        p.add(new JLabel("Subject sequence"), constraints);

        constraints.gridx = 1;
        constraints.weightx = 1.0;
        p.add(subjectField, constraints);

        constraints.gridx = 2;
        constraints.weightx = 0.0;
        p.add(subjectButton, constraints);

        // Comparison label, field and button
        constraints.gridx = 0;
        constraints.gridy = 2;
        p.add(new JLabel("Comparison"), constraints);

        constraints.gridx = 1;
        constraints.weightx = 1.0;
        p.add(comparisonField, constraints);

        constraints.gridx = 2;
        constraints.weightx = 0.0;
        p.add(comparisonButton, constraints);

        getContentPane().add(p, BorderLayout.NORTH);

        p = new JPanel(new GridBagLayout());

        // Self match and program type
        constraints.gridx = 0;
        constraints.gridy = 0;
        constraints.weightx = 0.0;
        constraints.weighty = 0.0;

        JCheckBox selfCompBox = new JCheckBox("Self comparison");
        selfCompBox.addItemListener(new ItemListener()
            {
                public void itemStateChanged(ItemEvent ie)
                {
                    if (ie.getStateChange() == ItemEvent.SELECTED)
                    {
                        selfMatch = true;
                        subjectButton.setEnabled(false);
                        subjectField.setEnabled(false);
                    }
                    else
                    {
                        selfMatch = false;
                        subjectButton.setEnabled(true);
                        subjectField.setEnabled(true);
                    }
                }
            });
        p.add(selfCompBox, constraints);

        constraints.insets = new Insets(2, 20, 2, 2);
        constraints.gridx = 1;
        constraints.weightx = 1.0;
        p.add(new JLabel("Program type"), constraints);

        // These indices must match the DataHandler search types
        JComboBox searchTypes = new JComboBox(new String [] { "NCBI/Wu Blast 2",
                                                              "Fasta" });
        searchTypes.addActionListener(new ActionListener()
            {
                public void actionPerformed(ActionEvent ae)
                {
                    JComboBox cb = (JComboBox) ae.getSource();
                    int searchType = cb.getSelectedIndex();
                    geminiUI.getDataHandler().setSearchType(searchType);
                }
            });

        searchTypes.setSelectedIndex(DataHandler.BLAST);

        constraints.insets = new Insets(2, 2, 2, 2);
        constraints.gridx = 2;
        constraints.weightx = 0.0;
        p.add(searchTypes, constraints);
 
        // Query fragment
        final IntegerField integerField = new IntegerField();
        integerField.setPreferredSize(new Dimension(30, 20));
        integerField.setEnabled(false);

        JCheckBox fragmentedBox = new JCheckBox("Fragmented query");
        fragmentedBox.addItemListener(new ItemListener()
            {
                public void itemStateChanged(ItemEvent ie)
                {
                    if (ie.getStateChange() == ItemEvent.SELECTED)
                    {
                        integerField.setEnabled(true);
                    }
                    else
                    {
                        integerField.setEnabled(false);
                    }
                }
            });

        constraints.gridx = 0;
        constraints.gridy = 1;
        constraints.weightx = 0.0;
        p.add(fragmentedBox, constraints);

        JLabel fragmentLabel = new JLabel("Query fragment size");

        constraints.insets = new Insets(2, 20, 2, 2);
        constraints.gridx = 1;
        constraints.weightx = 1.0;
        p.add(fragmentLabel, constraints);

        constraints.insets = new Insets(2, 2, 2, 2);
        constraints.gridx = 2;
        constraints.weightx = 0.0;
        p.add(integerField, constraints);
        getContentPane().add(p, BorderLayout.CENTER);

        p = new JPanel();
        JButton okOpen = new JButton("OK");
        okOpen.addActionListener(new ActionListener()
            {
                public void actionPerformed(ActionEvent ae)
                {
                    int fragmentSize = 0;
                    if (integerField.isEnabled())
                        fragmentSize = integerField.getValue();

                    final int value = fragmentSize;
                    checkTextFields(queryField, subjectField, comparisonField);

                    Runnable loader = new Runnable()
                        {
                            public void run()
                            {
                                OpenDialog.this.openData(queryFile,
                                                         subjectFile,
                                                         comparisonFile,
                                                         value);
                            }
                        };

                    Thread t = new Thread(loader);
                    t.start();
                }
            });

        p.add(okOpen);

        JButton cancelOpen = new JButton("Cancel");
        cancelOpen.addActionListener(new ActionListener()
            {
                public void actionPerformed(ActionEvent ae)
                {
                    setVisible(false);
                }
            });
        p.add(cancelOpen);

        getContentPane().add(p, BorderLayout.SOUTH);

        pack();
        setResizable(false);

        // Centre on screen
        Rectangle screen = getGraphicsConfiguration().getBounds();
        setLocation((screen.width - getWidth()) / 2,
                    (screen.height - getHeight()) / 2);
    }

    /**
     * <code>checkTextFields</code> updates the <code>File</code>s
     * which may have changed if the text fields have been edited.
     *
     * @param queryField a <code>JTextField</code>.
     * @param subjectField a <code>JTextField</code>.
     * @param comparisonField a <code>JTextField</code>.
     */
    private void checkTextFields(JTextField queryField,
                                 JTextField subjectField,
                                 JTextField comparisonField)
    {
        queryFile = createFile(queryField.getText());

        if (selfMatch)
            subjectFile = null;
        else
            subjectFile = createFile(subjectField.getText());

        comparisonFile = createFile(comparisonField.getText());
    }

    /**
     * <code>createFile</code> creates a <code>File</code> from a file
     * name.
     *
     * @param fileName a <code>String</code>.
     *
     * @return a <code>File</code> or null if the file name was null
     * or empty.
     */
    private File createFile(String fileName)
    {
        File file = null;

        if (fileName == null || fileName.equals(""))
            return file;

        return new File(fileName);
    }

    /**
     * <code>openData</code> reads all data files.
     *
     * @param queryFile a <code>File</code>.
     * @param subjectFile a <code>File</code>.
     * @param comparisonFile a <code>File</code>.
     * @param fragmentSize an <code>int</code> indicating the query
     * fragment size.
     */
    private void openData(File queryFile, File subjectFile, File comparisonFile,
                          int fragmentSize)
    {
        if (queryFile == null)
        {
            JOptionPane.showMessageDialog(this,
                                          "Please select a query sequence file",
                                          "Sequence file required",
                                          JOptionPane.WARNING_MESSAGE);
        }
        else if (comparisonFile == null)
        {
            JOptionPane.showMessageDialog(this,
                                          "Please select a comparison data file",
                                          "Comparison file required",
                                          JOptionPane.WARNING_MESSAGE);
        }
        else if (subjectFile == null && (! selfMatch))
        {
            JOptionPane.showMessageDialog(this,
                                          "Please select a subject sequence file",
                                          "Sequence file required",
                                          JOptionPane.WARNING_MESSAGE);
        }
        else
        {
            try
            {
                DataHandler dataHandler = geminiUI.getDataHandler();
                dataHandler.setQuerySeqFromFile(queryFile);

                if (selfMatch)
                    dataHandler.setSubjectSeqFromInstance(dataHandler.getQuerySequence());
                else
                    dataHandler.setSubjectSeqFromFile(subjectFile);

                int queryLength = dataHandler.getQuerySequence().length();

                // fragmentSize will be 0 if the query is not fragmented
                if (fragmentSize < 1)
                    fragmentSize = queryLength;

                if (fragmentSize > queryLength)
                    throw new IllegalArgumentException("The query fragment size ("
                                                       + fragmentSize
                                                       + ") may not be larger than the query sequence length ("
                                                       + queryLength
                                                       + ")");

                dataHandler.readCompFromFile(comparisonFile, fragmentSize);
                geminiUI.setXSequence(dataHandler.getQuerySequence());
                geminiUI.setPlotXSequence(dataHandler.getAnnotatedQuerySequence());

                if (selfMatch)
                {
                    geminiUI.setYSequence(dataHandler.getQuerySequence());
                    geminiUI.setPlotYSequence(dataHandler.getQuerySequence());
                }
                else
                {
                    geminiUI.setYSequence(dataHandler.getSubjectSequence());
                    geminiUI.setPlotYSequence(dataHandler.getSubjectSequence());
                }

                if (! geminiUI.getComponentsEnabled())
                    geminiUI.setComponentsEnabled(true);

                setVisible(false);
            }
            catch (IOException ioe)
            {
                setVisible(false);
                JOptionPane.showMessageDialog(null,
                                              "An error occurred reading an input file:\n"
                                              + ioe.getMessage(),
                                              "Error reading file",
                                              JOptionPane.ERROR_MESSAGE);
            }
            catch (SAXException se)
            {
                setVisible(false);
                JOptionPane.showMessageDialog(null,
                                              "An error occurred reading the comparison file:\n"
                                              + se.getMessage(),
                                              "Error reading comparison file",
                                              JOptionPane.ERROR_MESSAGE);
            }
            catch (Exception e)
            {
                setVisible(false);
                JOptionPane.showMessageDialog(null,
                                              "An error occurred reading data:\n"
                                              + e.getMessage(),
                                              "Error reading data:",
                                              JOptionPane.ERROR_MESSAGE);
            }
            catch (Error e)
            {
                setVisible(false);
                JOptionPane.showMessageDialog(null,
                                              "A serious internal error occurred:\n"
                                              + e.getMessage(),
                                              "Serious internal error",
                                              JOptionPane.ERROR_MESSAGE);
                e.printStackTrace();
                System.exit(0);
            }
        }
    }
}
