/*
 * This file is part of Gemini. Copyright (c) 2001, 2002 Genome
 * Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.psu.gemini.gui;

import java.awt.Color;
import java.awt.BasicStroke;
import java.awt.Paint;

import org.biojava.bio.gui.sequence.*;
import org.biojava.bio.seq.FeatureFilter;
import org.biojava.bio.seq.OptimizableFilter;
import org.biojava.bio.seq.StrandedFeature;
import org.biojava.utils.ChangeVetoException;

/**
 * <code>SequenceRendererFactory</code> creates various types of
 * <code>SequenceRenderer</code> with different rendering
 * characteristics.
 *
 * @author Keith James
 */
final class SequenceRendererFactory
{
    /**
     * <code>BASIC_HORIZONTAL</code> a non-configurable,
     * horizontally-oriented renderer.
     */
    static final int BASIC_HORIZONTAL = 0;

    /**
     * <code>BASIC_VERTICAL</code> a non-configurable,
     * vertically-oriented renderer.
     */
    static final int BASIC_VERTICAL = 1;

    private FeatureFilter filtForCDS = new FeatureFilter.And(new
            FeatureFilter.ByType("CDS"), new
                FeatureFilter.StrandFilter(StrandedFeature.POSITIVE));

    private FeatureFilter filtRevCDS = new FeatureFilter.And(new
            FeatureFilter.ByType("CDS"), new
                FeatureFilter.StrandFilter(StrandedFeature.NEGATIVE));

    private OptimizableFilter filtMisc = new FeatureFilter.ByType("misc_feature");
    private OptimizableFilter filtRev = new FeatureFilter.StrandFilter(StrandedFeature.NEGATIVE);

    private RectangularBeadRenderer rrBlue =
        new RectangularBeadRenderer(6.0f, 0.0f,
                                    Color.black, Color.blue,
                                    new BasicStroke());

    private RectangularBeadRenderer rrGreen =
        new RectangularBeadRenderer(6.0f, 6.0f,
                                    Color.black, Color.green,
                                    new BasicStroke());

    private GeminiUI geminiUI;

    /**
     * Creates a new <code>SequenceRendererFactory</code>.
     *
     * @param geminiUI a <code>GeminiUI</code> parent.
     */
    SequenceRendererFactory(GeminiUI geminiUI)
    {
        this.geminiUI = geminiUI;
    }

    SequenceRenderer createRenderer(int type) throws Exception
    {
          switch (type)
          {
              case BASIC_HORIZONTAL:
                  return createHorizRenderer();

              case BASIC_VERTICAL:
                  return createVertRenderer();

              // Bad argument
              default:
                throw new Exception("Invalid SequenceRenderer type requested: " + type);
          }
    }

    private SequenceRenderer createHorizRenderer() throws ChangeVetoException
    {
        MultiLineRenderer multi = new MultiLineRenderer();

        double blockDepth = 6.0;
        Paint forFill = Color.yellow;
        Paint revFill = Color.red;

        ZiggyFeatureRenderer forZiggyCDS = new ZiggyFeatureRenderer();
        ZiggyFeatureRenderer revZiggyCDS = new ZiggyFeatureRenderer();

        forZiggyCDS.setBlockDepth(blockDepth);
        forZiggyCDS.setFill(forFill);
        revZiggyCDS.setBlockDepth(blockDepth);
        revZiggyCDS.setFill(revFill);

        rrBlue.setDelegateRenderer(filtRev, rrGreen);

        FeatureBlockSequenceRenderer  forCDS = new FeatureBlockSequenceRenderer();
        FeatureBlockSequenceRenderer  revCDS = new FeatureBlockSequenceRenderer();
        FeatureBlockSequenceRenderer fbrMisc = new FeatureBlockSequenceRenderer();

        forCDS.setFeatureRenderer(forZiggyCDS);
        revCDS.setFeatureRenderer(revZiggyCDS);
        fbrMisc.setFeatureRenderer(rrBlue);

        // For horiz, for/rev are oppostite way round to vert
        multi.addRenderer(new FilteringRenderer(forCDS, filtForCDS, false));
        multi.addRenderer(new FilteringRenderer(revCDS, filtRevCDS, false));
        multi.addRenderer(new FilteringRenderer(fbrMisc, filtMisc, false));
        multi.addRenderer(new SymbolSequenceRenderer());
        multi.addRenderer(new RulerRenderer(RulerRenderer.TICKS_UP));

        return multi;
    }

    private SequenceRenderer createVertRenderer() throws ChangeVetoException
    {
        MultiLineRenderer multi = new MultiLineRenderer();

        double blockDepth = 6.0;
        Paint forFill = Color.yellow;
        Paint revFill = Color.red;

        ZiggyFeatureRenderer forZiggyCDS = new ZiggyFeatureRenderer();
        ZiggyFeatureRenderer revZiggyCDS = new ZiggyFeatureRenderer();

        forZiggyCDS.setBlockDepth(blockDepth);
        forZiggyCDS.setFill(forFill);
        revZiggyCDS.setBlockDepth(blockDepth);
        revZiggyCDS.setFill(revFill);

        rrBlue.setDelegateRenderer(filtRev, rrGreen);

        FeatureBlockSequenceRenderer  forCDS = new FeatureBlockSequenceRenderer();
        FeatureBlockSequenceRenderer  revCDS = new FeatureBlockSequenceRenderer();
        FeatureBlockSequenceRenderer fbrMisc = new FeatureBlockSequenceRenderer();

        forCDS.setFeatureRenderer(forZiggyCDS);
        revCDS.setFeatureRenderer(revZiggyCDS);
        fbrMisc.setFeatureRenderer(rrBlue);

        // For vertical, for/rev are oppostite way round to horiz
        multi.addRenderer(new FilteringRenderer(revCDS, filtRevCDS, false));
        multi.addRenderer(new FilteringRenderer(forCDS, filtForCDS, false));
        multi.addRenderer(new FilteringRenderer(fbrMisc, filtMisc, false));
        multi.addRenderer(new SymbolSequenceRenderer());
        multi.addRenderer(new RulerRenderer(RulerRenderer.TICKS_UP));

        return multi;
    }
}
