/*
 * This file is part of Gemini. Copyright (c) 2001, 2002 Genome
 * Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.psu.gemini.gui;

import java.util.Iterator;

import org.biojava.bio.gui.sequence.SequenceViewerEvent;
import org.biojava.bio.gui.sequence.SequenceViewerListener;
import org.biojava.bio.seq.Feature;
import org.biojava.bio.seq.FeatureHolder;

/**
 * <code>SequenceViewerListenerFactory</code> creates
 * <code>SequenceViewerListener</code>s of various types. These
 * respond in different ways to mouse gestures and clicks on the
 * <code>SequenceViewer</code> they are registered with.
 *
 * @author Keith James
 */
final class SequenceViewerListenerFactory
{
    /**
     * <code>FEATURE_VIEWER</code> a listener which updates a viewer
     * with the <code>Feature</code> clicked on.
     */
    static final int FEATURE_VIEWER = 0;

    /**
     * <code>DEBUG_PRINT</code> a listener which prints the
     * <code>Feature</code> clicked on to STDOUT.
     */
    static final int DEBUG_PRINT = 1;

    private GeminiUI geminiUI;

    /**
     * Creates a new <code>SequenceViewerListenerFactory</code>.
     *
     * @param geminiUI a <code>GeminiUI</code> parent.
     */
    SequenceViewerListenerFactory(GeminiUI geminiUI)
    {
        this.geminiUI = geminiUI;
    }

    SequenceViewerListener createListener(int type) throws Exception
    {
        switch (type)
        {
            case FEATURE_VIEWER:
                return new SequenceViewerAdapter()
                {
                    public void mouseClicked(SequenceViewerEvent sve)
                    {
                        Object t = sve.getTarget();

                        if (t instanceof FeatureHolder)
                        {
                            FeatureHolder fh = (FeatureHolder) t;
                            for (Iterator fi = fh.features(); fi.hasNext();)
                            {
                                Feature f = (Feature) fi.next();

                                FeatureViewer fv = geminiUI.getFeatureViewer();
                                fv.setFeature(f);
                                if (! fv.isVisible())
                                    fv.setVisible(true);
                            }
                        }
                    }
                };

            case DEBUG_PRINT:
                return new SequenceViewerAdapter()
                {
                    public void mouseClicked(SequenceViewerEvent sve)
                    {
                        Object t = sve.getTarget();

                        if (t instanceof FeatureHolder)
                        {
                            FeatureHolder fh = (FeatureHolder) t;
                            for (Iterator fi = fh.features(); fi.hasNext();)
                            {
                                Feature f = (Feature) fi.next();
                                System.out.println(f);
                            }
                        }
                    }
                };

            // Bad argument
            default:
                throw new Exception("Invalid SequenceViewerListener type requested: " + type);
        }
    }
}
