#!/usr/local/bin/perl5.6.1 -w

#
# this script to predict gene structures from data ESTs
#
# to do so, use of exonerate - Guy Slater - guy@ebi.ac.uk
# and Exon Pair Merging to predict both transcripts and gene structure
# run genomewise on the predicted genes to work out the ORF, the start and stop codons and the UTRs
#

# output files : 
#   * sequence.exonerate.tab
#   * sequence.exonerate.genomewise.tab
#   * sequence.tab with the gene features predicted by exonerate and the CDS features predicted by genomewise

#
# e.g. perl ESTs2Genes.pl -i gDNA.ESTs.exonerate.tab >! gDNA.ESTs.exonerate.merged.tab
#

# process :
# 1/ Exon Pair Merging to generate transcripts and genes features
# 2/ write the predicted features into an EMBL file
# 3/ write the gene features into a GFF file in order to a Gene Prediction tool, geneid, giving it a set of predicted exons. - optional
# 4/ write geneid results into an EMBL format file.

# External Modules

# Constraint variables declaration before using them
use strict;
##########################################

use Data::Dumper;

# eval "require Bio::PSU::Utils::ExonPairMerger";
use Bio::PSU::Utils::ExonPairMerger;

# PSU modules for EMBL writing
use Bio::PSU::SeqFactory;
use Bio::PSU::Seq;
use Bio::PSU::Feature;
use Bio::PSU::Range;

##########################################
=pod

=item * Configuration loading from the command line

=cut
##########################################

my $_verbose = 0;

my $_help          = "-help";
my $_input         = "-i";

# no input per default
my $input      = "-1";
my $max_intron_size = 400;
my $geneid = 0;

# Check if in the command line there is an input file name (-i option)
# the other options are not compulsory (set of options per default)
my $lgth = @ARGV + 0;
# if no args, display the help message
if ($lgth == 0) {
  display_help ();
  exit 1;
}
elsif ($lgth >= 1) {
  my $i = 0;
  foreach my $element (@ARGV) {
    if ($element eq $_help) {
      display_help ();
      exit 1;
    }
    elsif ($element eq $_input) {
      $input = $ARGV[$i+1];
    }
    elsif ($element eq "_geneid") {
      $geneid = 1;
    }
    # if at least one of the options is not recognized, display the help message and exit
    elsif ($element =~ /^-/) {
      print STDERR "option not recognized: $element\n";
      display_help ();
      exit 1;
    }
    $i ++;
  }
}

##########################################
=pod

=item * Exon Pair Merging

=cut
##########################################

if ($_verbose) {
  print STDERR "processing Exon Pair Merging...\n";
}

my $exonPairMerger = Bio::PSU::Utils::ExonPairMerger->new (
							   input           => $input,
							   max_intron_size => $max_intron_size,
							   verbose         => $_verbose
							  );

# STDOUT
my $outtab_obj = Bio::PSU::SeqFactory->make(
					    -fh     => \*STDOUT,
					    -format => 'embl'
					   );

while (my ($sequenceid, $merged_features) = $exonPairMerger->next_sequence()) {

  if (defined $sequenceid) {

    if ($_verbose) {
      print STDERR "merging process for sequence, $sequenceid, done.\n";
    }

    if (@$merged_features > 0) {

      # Process the mRNAs to get rid of intron of 1bp length
      $merged_features = process_introns_size ($merged_features);

      my $newseq_obj = Bio::PSU::Seq->new (
  					   -id   => $sequenceid,
					   -type => "dna",
					   -desc => "Exonerate predictions - ESTs2Genes post processing"
					  );
      $newseq_obj->features (@$merged_features); 
      # copy this seq obj into a new file
      $outtab_obj->write_seq($newseq_obj);
    }

  }
  else {
    # no more sequence to process !! - should be tested in the while test !!!
    last;
  }
}

if ($_verbose) {
  print STDERR "process done\n";
}

###################################################################
# The end
###################################################################

 # display the help message

sub display_help {
  print "ESTs2Genes.pl -help -i [exonerate tab output file] -geneid\n\n";
  print "\t-help a help message\n";
  print "\t-i The exonerate predictions in embl format (exonerate2art.pl script) - (no default, always specify it)\n";
  print "\t-geneid run geneid on the set of predicted exons - (default: deactivated)\n\n";
}


sub runGeneid {
  my ($gDNA, $exons_gff_file_name, $geneid_output) = @_;
  
  my $geneid_parameter_file = "/nfs/disk222/yeastpub/databases/dicty/geneid_table/dicty.param";

  print STDOUT "\nRunning geneid using the predicted exons...\n";
  print STDOUT "command: lsrun -v /bin/csh -c \"unlimit;ulimit -a;/nfs/disk222/yeastpub/bio-soft/geneid_v1.0/bin/geneid -GP $geneid_parameter_file -R $exons_gff_file_name $gDNA >! $geneid_output\"\n";
  
  my @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit -a;/nfs/disk222/yeastpub/bio-soft/geneid_v1.0/bin/geneid -GP $geneid_parameter_file -R $exons_gff_file_name $gDNA >! $geneid_output\"");
  system (@args);
}


sub runGeneid2Tab {
  my ($geneid_output_file) = @_;

  print STDOUT "command: lsrun -v /bin/csh -c \"/nfs/disk222/yeastpub/bin/geneid2tab.pl $geneid_output_file\"\n";

  my @args = ("lsrun -v /bin/csh -c \"/nfs/disk222/yeastpub/bin/geneid2tab.pl $geneid_output_file\"");
  system (@args);
}

# $type => gene or transcript
# if the gene is on the reverse strand, map the exons coordinates from the forward strand to the reverse one.

sub generateEvidenceFiles {
  my ($type, $path, $seq_length, @exonFeaturesPerGene) = @_;
  my @evi_files_forward = ();
  my @evi_files_reverse = ();
  my $i = 0;

  # print STDERR "dump in generateEvidenceFiles: " . Dumper (@exonFeaturesPerGene) . "\n";
  
  foreach my $hash_ref (@exonFeaturesPerGene) {
    my %hash = %$hash_ref;
    my @exons = @{$hash{exons}};
    my $evi_file = "$path/evi_file.$type.$i";
    push (@evi_files_forward, "$evi_file.forward");
    push (@evi_files_reverse, "$evi_file.reverse");
    
    my @new_exons_reverse = ();
    
    open EVI_FORWARD, ">$evi_file.forward" or die "can't open evidence file, $evi_file.forward\n";
    open EVI_REVERSE, ">$evi_file.reverse" or die "can't open evidence file, $evi_file.reverse\n";
    
    foreach my $exon (@exons) {
      
      # Forward s strand processing
      
      print EVI_FORWARD $exon;
      
      # Reverse strand processing
      
      $exon =~ /exon\t(\d+)\t(\d+)\n/;
      my $start = mapCoordinate ($1, $seq_length);
      my $end   = mapCoordinate ($2, $seq_length);
      # reverse start and end !!
      $exon = "exon\t$end\t$start\n";
      push (@new_exons_reverse, $exon);
    }
    
    # last element => first entry in the file
    my $j = @new_exons_reverse-1;
    while ($j>=0) {
      my $exon = $new_exons_reverse[$j];
      print EVI_REVERSE $exon;
      $j--;
    }
    
    close EVI_FORWARD;
    close EVI_REVERSE;
    $i++;
  }
  
  return (\@evi_files_forward, \@evi_files_reverse);
}


# $type => gene or transcript
# if the gene is on the reverse strand, map the exons coordinates from the forward strand to the reverse one.

sub generateEvidenceFiles_backup {
  my ($type, $path, $seq_length, @exonFeaturesPerGene) = @_;
  my @evi_files_forward = ();
  my @evi_files_reverse = ();
  my $i = 0;

  # print STDERR "dump in generateEvidenceFiles: " . Dumper (@exonFeaturesPerGene) . "\n";
  
  foreach my $hash_ref (@exonFeaturesPerGene) {
    my %hash = %$hash_ref;
    my $strand = $hash{strand};
    my @exons = @{$hash{exons}};
    my $evi_file = "$path/evi_file.$type.$i";
    if ($strand == 1) {
      push (@evi_files_forward, $evi_file);
    }
    elsif ($strand == -1) {
      push (@evi_files_reverse, $evi_file);
    }
    else {
      print STDERR "error, can't work out the strand of the current gene - exons: @exons\nwon't run genomewise on it";
    }

    my @new_exons = ();

    open FILE, ">$evi_file" or die "can't open evidence file, $evi_file\n";
    foreach my $exon (@exons) {
      if ($strand == -1) {
	# my $exonFeature = "exon\t$start\t$end\n";
	# print STDERR "exon: $exon\n";
	$exon =~ /exon\t(\d+)\t(\d+)\n/;
	my $start = mapCoordinate ($1, $seq_length);
	my $end   = mapCoordinate ($2, $seq_length);
	# reverse start and end !!
	$exon = "exon\t$end\t$start\n";
	# print STDERR "exon after processing: $exon\n";
      }
      push (@new_exons, $exon);
    }

    if ($strand == 1) {
      # first element => first entry in the file
      my $i = 0;
      while ($i<@new_exons) {
	my $exon = $new_exons[$i];
	print FILE $exon;
	$i++;
      }
    }
    else {
      # last element => first entry in the file
      my $i = @new_exons-1;
      while ($i>=0) {
	my $exon = $new_exons[$i];
	print FILE $exon;
	$i--;
      }
    }
    close FILE;
    $i++;
  }

  return (\@evi_files_forward, \@evi_files_reverse);
}

sub runGenomewise {
  my ($gdna, @evi_files) = @_;

  my @genomewiseResultFiles = ();

  print "running genomewise ...\n";

  # print STDERR "evi files: @evi_files\n";

  foreach my $evi_file (@evi_files) {

    my @tmp = split ('/', $evi_file);
    my $evi_file_name = pop @tmp;

    print  "lsrun -v /bin/csh -c \"unlimit;ulimit -c 0;/nfs/team81/axk/tools/genomewise/genomewise $gdna $evi_file -notrans -nogff -nogenes -gene 500 -kbyte 20000 >! $gdna.$evi_file_name.genomewise\"\n";

    my @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit -c 0;/nfs/team81/axk/tools/genomewise/genomewise $gdna $evi_file -notrans -nogff -nogenes >! $gdna.$evi_file_name.genomewise\"");
    system (@args);

    push (@genomewiseResultFiles, "$gdna.$evi_file_name.genomewise");
  }

  return @genomewiseResultFiles;
}

# need seq length to remap on the forward strand the reverse features !!!!

sub genomewise2art {
  my ($gDNA, $strand, $seq_length, @genomewiseResultFiles) = @_;
  my @features = ();
  my $tab_file = undef;
  if ($strand == 1) {
    $tab_file = "$gDNA.forward.genomewise.tab";
  }
  else {
    $tab_file = "$gDNA.reverse.genomewise.tab";
  }

  # one gene per file

  foreach my $genomewise_file (@genomewiseResultFiles) {
    my $feature = undef;
    my @ranges = ();
    my $firstExon = 1;
    my $utr5 = undef;
    my $utr3 = undef;
    my $codon_start = undef;

    open (GENOMEWISE, "<$genomewise_file") or die "can't open file, $genomewise_file\n";

    while (<GENOMEWISE>) {
      my $line = $_;
      if ($line =~ /gene/i) {
	# new gene
      }
      elsif ($line =~ /end/i) {
	# parsing ended for this gene
	# create the new feature object
	$feature = Bio::PSU::Feature->new(
					  -key    => 'CDS',
					  -ranges => [@ranges],
					 );
	
	if (defined $utr5) {
	  # $feature->qadd ('5UTR', $utr5);
	  $feature->qadd ('note', "5UTR - ($utr5)");
	}
	if (defined $utr3) {
	  # $feature->qadd ('3UTR', $utr3);
	  $feature->qadd ('note', "3UTR - ($utr3)");
	}
	$feature->qadd ('codon_start', $codon_start);
	push (@features, $feature);
	# reinit the ranges object
	@ranges = ();
      }
      elsif ($line =~ /utr5/i) {
	$line =~ /.+utr5\s+(\d+)\s+(\d+)/;
	my $start = $1;
	my $end = $2;
	if ($strand == -1) {
	   # reverse start and end
	  $start = mapCoordinate ($start, $seq_length);
	  $end = mapCoordinate ($end, $seq_length);
	  my $tmp = $start;
	  $start = $end;
	  $end = $tmp;
	}

	$utr5  = "$start..$end";
	# print STDERR "5UTR - start, end: $start, $end\n";
      }
      elsif ($line =~ /utr3/i) {
	$line =~ /.+utr3\s+(\d+)\s+(\d+)/;
	my $start = $1;
	my $end = $2;
	if ($strand == -1) {
	  # reverse start and end
	  $start = mapCoordinate ($start, $seq_length);
	  $end = mapCoordinate ($end, $seq_length);
	  my $tmp = $start;
	  $start = $end;
	  $end = $tmp;
	}
	$utr3  = "$start..$end";
	
	# print STDERR "3UTR - start, end: $start, $end\n";

      }
      elsif ($line =~ /cds/i) {
	$line =~ /.+cds\s+(\d+)\s+(\d+)\s+phase\s+(\d)/;
	my $start = $1;
	my $end   = $2;
	if ($strand == -1) {
	  # reverse start and end !!
	  $start = mapCoordinate ($start, $seq_length);
	  $end   = mapCoordinate ($end, $seq_length);
	  my $tmp = $start;
	  $start = $end;
	  $end = $tmp;
	}
	my $phase = -1;

	# print STDERR "start, end, phase: $start, $end, $phase\n";

	if ($firstExon) {
	  # parse the phase => codon_start qualifier
	  $phase = $3;
	  $firstExon = 0;
	}
	my $range = Bio::PSU::Range->new (-start       => $start,
					  -end         => $end,
					  -strand      => $strand);
	push (@ranges, $range);
	if ($phase != -1) {
	  $codon_start = processPhase ($phase);
	}
      }
    }
    close GENOMEWISE;
  }

  # write features in an Embl file

  my $new_seq_obj = Bio::PSU::Seq->new (
					-type => 'dna'
				       );
  $new_seq_obj->features (@features);
  my $new_output_obj = Bio::PSU::SeqFactory->make(
						  -file   => ">$tab_file",
						  -format => 'embl'
						 );
  $new_output_obj->write_seq($new_seq_obj);
}


sub getGenomewisePredictions {
  my ($strand, $seq_length, @genomewiseResultFiles) = @_;
  my @features = ();
  
  # one gene per file

  foreach my $genomewise_file (@genomewiseResultFiles) {
    my $feature = undef;
    my @ranges  = ();
    my $firstExon = 1;
    my $utr5 = undef;
    my $utr3 = undef;
    my $codon_start = undef;
    
    open (GENOMEWISE, "<$genomewise_file") or die "can't open file, $genomewise_file\n";
    
    while (<GENOMEWISE>) {
      my $line = $_;
      if ($line =~ /gene/i) {
	# new gene
      }
      elsif ($line =~ /end/i) {
	# parsing ended for this gene
	# create the new feature object
	$feature = Bio::PSU::Feature->new(
					  -key    => 'CDS',
					  -ranges => [@ranges],
					 );
	
	if (defined $utr5) {
	  # $feature->qadd ('5UTR', $utr5);
	  $feature->qadd ('note', "5UTR - ($utr5)");
	}
	if (defined $utr3) {
	  # $feature->qadd ('3UTR', $utr3);
	  $feature->qadd ('note', "3UTR - ($utr3)");
	}
	$feature->qadd ('codon_start', $codon_start);
	push (@features, $feature);
	# reinit the ranges object
	@ranges = ();
      }
      elsif ($line =~ /utr5/i) {
	$line =~ /.+utr5\s+(\d+)\s+(\d+)/;
	my $start = $1;
	my $end = $2;
	if ($strand == -1) {
	  # reverse start and end
	  $start = mapCoordinate ($start, $seq_length);
	  $end = mapCoordinate ($end, $seq_length);
	  my $tmp = $start;
	  $start = $end;
	  $end = $tmp;
	}

	$utr5  = "$start..$end";
	# print STDERR "5UTR - start, end: $start, $end\n";
      }
      elsif ($line =~ /utr3/i) {
	$line =~ /.+utr3\s+(\d+)\s+(\d+)/;
	my $start = $1;
	my $end = $2;
	if ($strand == -1) {
	  # reverse start and end
	  $start = mapCoordinate ($start, $seq_length);
	  $end = mapCoordinate ($end, $seq_length);
	  my $tmp = $start;
	  $start = $end;
	  $end = $tmp;
	}
	$utr3  = "$start..$end";
	
	# print STDERR "3UTR - start, end: $start, $end\n";
	
      }
      elsif ($line =~ /cds/i) {
	$line =~ /.+cds\s+(\d+)\s+(\d+)\s+phase\s+(\d)/;
	my $start = $1;
	my $end   = $2;
	if ($strand == -1) {
	  # reverse start and end !!
	  $start = mapCoordinate ($start, $seq_length);
	  $end   = mapCoordinate ($end, $seq_length);
	  my $tmp = $start;
	  $start = $end;
	  $end = $tmp;
	}
	my $phase = -1;
	
	# print STDERR "start, end, phase: $start, $end, $phase\n";

	if ($firstExon) {
	  # parse the phase => codon_start qualifier
	  $phase = $3;
	  $firstExon = 0;
	}
	my $range =  Bio::PSU::Range->new(-start       => $start,
					  -end         => $end,
					  -strand      => $strand);
	push (@ranges, $range);
	if ($phase != -1) {
	  $codon_start = processPhase ($phase);
	}
      }
    }
    close GENOMEWISE;
  }
  
  return @features;
}


sub writeEmblFeatures {
  my ($emblFile, $seq_str, @features) = @_;

   # write features in an Embl file

  my $new_seq_obj = Bio::PSU::Seq->new (
					-type => 'dna'
				       );
  $new_seq_obj->str ($seq_str);
  $new_seq_obj->features (@features);
  my $new_output_obj = Bio::PSU::SeqFactory->make(
						  -file   => ">$emblFile",
						  -format => 'embl'
						 );
  $new_output_obj->write_seq($new_seq_obj);
}


sub processPhase {
  my ($phase) = @_;
  my $codon_start = undef;

  if ($phase =~ /0/) {
    $codon_start = 1;
  }
  elsif ($phase =~ /1/) {
    $codon_start = 2;
  }
  elsif ($phase =~ /2/) {
    $codon_start = 3;
  }
  
  return $codon_start;
}


# map a corrdinate from a strand to the other one

sub mapCoordinate {
  my ($previous_coordinate, $seq_length) = @_;
  my $coordinate = undef;
  
  $coordinate = $seq_length - $previous_coordinate + 1;
  
  return $coordinate;
}


sub getSequenceLength {
  my ($gDNA) = @_;
  my $seq_length = -1;
  
  my $new_input_obj = Bio::PSU::SeqFactory->make(
						 -file   => "<$gDNA",
						 -format => 'fasta'
						);
  my $seq = $new_input_obj->next_seq;
  $seq_length = length ($seq->str);

  # print STDERR "length seq: $seq_length\n";

  return $seq_length;
}


sub process_introns_size {
  my ($merged_features) = @_;
  my @new_features = ();

  foreach my $feature (@$merged_features) {
    if ($feature->key eq "mRNA") {
      my @ranges = $feature->ranges;
      my $i = 0;
      my $range1 = $ranges[$i];
      $i++;
      while ($i < @ranges) {
	my $range2 = $ranges[$i];
	if ($range2->start <= ($range1->end + 1)) {
	  # Merge them
	  my $new_range = Bio::PSU::Range->new (
						-start  => $range1->start,
						-end    => $range2->end,
						-strand => $range1->strand,
					       );
	  # Get rid of range 1 and 2 and replace them by new_range
	  splice (@ranges, $i-1, 2, $new_range);
	  $range1 = $new_range;
	}
	else {
	  $range1 = $range2;
	  $i++;
	}
	#$i++;
      }
      my $new_feature = $feature->clone (-ranges => [@ranges]);
      push (@new_features, $new_feature);
    }
    else {
      push (@new_features, $feature);
    }
  }

  return \@new_features;
}
