#!/usr/local/bin/perl5.6.1

=head1 NAME

ESTs_annotation2embl.pl

=head1 SYNOPSIS

ESTs_annotation2ga.pl -help -fa [ESTs file] -droso [blastx output file against Flybase database] -swall [blastx output file against Swall database] -e [ESTScan peptides file in FASTA format] -ga [gene assocs file]

        -help  a help message
        -fa    the sequences in fasta format - (no default)
        -droso the output of the blastx search against Flybase database - (no default)
        -swall the output of the blastx search against Swall database - (no default)
        -e     the fasta translated sequences made by ESTScan
        -ga    the gene assocs file - default is fb

=head1 DESCRIPTION

  This script to generate an annotation file in the EMBL format from ESTs annotation data (blastx and ESTScan data)

  @author Arnaud Kerhornou - axk@sanger.ac.uk

=cut

#############################
#
# GO Database Access set up
#
#############################

BEGIN {

  my $configFile = "/nfs/pathdb/dev/go-cgi/config.pl";

  if (-f $configFile) {
    require $configFile;
  }

}

# External Modules

# Constraint variables declaration before using them
use strict;
##########################################

use Data::Dumper;

# Benchmark Module
use Benchmark;
##################################################################

###################
# PSU Bio Modules
##
use Bio::PSU::Feature;
use Bio::PSU::SeqFactory;
use Bio::PSU::Range;

# Bioperl ESTScan parser
# Not the original one, made some modifications to make it work properly
# with the given translations fasta file

use Bio::Tools::ESTScan;

# eval "require Bio::PSU::Utils::GeneAssociationParser";
use Bio::PSU::Utils::GeneAssociationParser;

# eval "require Bio::PSU::Utils::BlastParsing";
use Bio::PSU::Utils::BlastParsing;

##########################################
=pod

=item * Configuration loading from the command line

=cut
##########################################

$|=1;

my $t1 = Benchmark->new ();

#####################################################

#
# GO Database access to get the GO Mapping information
#
# e.g Get term name and aspect giving the GO accession number
#

use GO::AppHandle;

my $dbname = $ENV{GO_DBNAME};
my $dbport = $ENV{GO_DBPORT};
my $dbhost = $ENV{GO_DBHOST};
my $dbuser = $ENV{GO_DBUSER};

if (not $dbname =~ /go\d/) {
  print STDERR "GO database name uncorrect, must be go1 or go2 !!";
  exit 1;
}

my $apph = GO::AppHandle->connect (
				   -dbname => $dbname,
				   -dbport => $dbport,
				   -dbhost => $dbhost,
				   -dbuser => $dbuser,
				  ) 
or die "can't connect to GO database, $dbname!!!\n";

####################################################

############################
# Debug boolean
##

my $debug = 0;

my $_help           = "-help";
my $_fasta          = "-fa";
my $_droso          = "-droso";
my $_swall          = "-swall";
my $_estscan        = "-e";
my $_ga             = "-ga";
my $_output         = "-o";
my $_tissue_type    = "-t";
 
# fasta file per default
my $fasta = "-1";
my $droso = "-1";
my $swall = "-1";
my $estscan = "-1";
# gene assocs file default is FLybase
# Make sure it is ungzipped !!!!
my $ga = "/nfs/pathsoft/databases/go_cvs/gene-associations/gene_association.fb";
my $output = undef;
my $tissue_type = undef;
my $note_tissue = undef;

my $i = 0;
foreach my $element (@ARGV) {
  if ($element eq $_help) {
    display_help ();
    exit 1;
  }
  elsif ($element eq $_fasta) {
    $fasta = $ARGV[$i+1];
  }
  elsif ($element eq $_droso) {
    $droso = $ARGV[$i+1];
  }
  elsif ($element eq $_swall) {
    $swall = $ARGV[$i+1];
  }
  elsif ($element eq $_estscan) {
    $estscan = $ARGV[$i+1];
  }
  elsif ($element eq $_ga) {
    $ga = $ARGV[$i+1];
  }
  elsif ($element eq $_output) {
    $output = $ARGV[$i+1];
  }
  elsif ($element eq $_tissue_type) {
    $tissue_type = $ARGV[$i+1];
  }
  # if at least one of the options is not recognized, display the help message and exit
  elsif ($element =~ /^-/) {
    print STDERR "option not recognized: $element\n";
    display_help ();
    exit 1;
  }
  $i++;
}

if (($fasta eq "-1") || ($droso eq "-1") || ($swall eq "-1") || ($estscan eq "-1")) {
  print STDERR "option(s) not specified!!\n";
  display_help ();
  exit 1;
}

if (not defined $tissue_type) {
  print STDERR "The tissue type has not been defined...\n";
}
else {
  $note_tissue = "cDNA from infected $tissue_type";
}

my $pepfile = "$fasta.pep";
my $pepseqs_factory = Bio::PSU::SeqFactory->make(-file   => ">$pepfile",
				                 -format => 'fasta',
                                                 -type   => 'protein',
                                                );


###############################
# Parse Drosophila blast file output
# Get only the 3 best hits
##

my $blast_parser = Bio::PSU::Utils::BlastParsing->new (
						       blastOutput => $droso,
						       max_hits => 3,
						      );

print STDERR "parsing $droso blast output...\n";
my %results_droso_blast = $blast_parser->parse;

$blast_parser->setBlastOutput ($swall);

print STDERR "parsing $swall blast output...\n";
my %results_swall_blast = $blast_parser->parse;

###################
# Parse Flybase Gene Association File
##

my $ga_parser = Bio::PSU::Utils::GeneAssociationParser->new (
							      geneAssociationFile => $ga,
							    );
  
print STDERR "parsing gene association file...\n";
my %ga_entries;
%ga_entries = $ga_parser->parse;

###################
# Parse ESTScan translated sequences
##

print STDERR "parsing ESTScan translated sequences FASTA file...\n";

#my $estscanfile_factory = Bio::PSU::SeqFactory->make(-file   => "<$estscan",
#					             -format => 'fasta',
#                                                     -type   => 'protein',);
#my %ESTScan_seqs;
#while (my $seq_obj = $estscanfile_factory->next_seq) {
#    my $seq_name = $seq_obj->id;
#    my $seq_str  = $seq_obj->str;
#    $ESTScan_seqs{$seq_name} = $seq_str;
#}

# Bioperl

my %ESTScan_seqs;
my $estscan_parser = Bio::Tools::ESTScan->new(-file => $estscan);
while(my $gene = $estscan_parser->next_prediction()) {

  # $gene is an instance of Bio::Tools::Prediction::Gene

  foreach my $orf ($gene->exons()) {
    # $orf is an instance of Bio::Tools::Prediction::Exon
    $ESTScan_seqs{$orf->seq_id} = $orf;
  }
}
$estscan_parser->close();

#########################
# Generate EMBL output
##

generateEMBLOutput ($fasta, \%ga_entries, \%results_droso_blast, \%results_swall_blast, \%ESTScan_seqs, $output, $tissue_type);

print STDERR "generation done.\n";

my $t2 = Benchmark->new ();

print STDERR "\nTotal : ", timestr (timediff ($t2, $t1)), "\n";

#################
# end
##

#############################
# display the help message
##

sub display_help {
  print "ESTs_annotation2embl.pl $_help $_fasta [ESTs file] $_droso [blastx output file against Flybase database] $_swall [blastx output file against swall database] $_estscan [ESTScan translation sequences in FASTA] $_ga [gene assocs file] $_tissue_type [tissue type]\n\n";
  print "\t$_help a help message\n";
  print "\t$_fasta the sequences in fasta format - (no default)\n";
  print "\t$_droso the output of the blastx search against Flybase database - (no default)\n";
  print "\t$_swall the output of the blastx search against Swall database - (no default)\n";
  print "\t$_estscan the translation sequences predicted by ESTScan - (no default)\n";
  print "\t$_ga the gene assocs file to inferred the gene assocs from - (optional - the default is Flybase)\n";
  print "\t$_output the annotated EMBL output file - (optional - default is STDOUT)\n";
  print "\t$_tissue_type the tissue type the EST library is made from - (no default)\n\n";
}

####################################
# Generate EMBL file
##

sub generateEMBLOutput {
  my ($fasta, $ga_entries_ref, $results_droso_blast_ref, $results_swall_blast_ref, $ESTScan_seqs_ref, $output, $tissue_type) = @_;
  my %results_droso_blastx = %{$results_droso_blast_ref};
  my %results_swall_blastx = %{$results_swall_blast_ref};
  my %ga_entries           = %{$ga_entries_ref};
  my %ESTScan_seqs         = %{$ESTScan_seqs_ref};

  ###############################
  # Get the EST sequence objects
  ##

  print STDERR "parsing fasta ESTs database...\n";

  my $seqs_factory = Bio::PSU::SeqFactory->make(-file   => "<$fasta",
						-format => 'fasta');
  my %seqs;
  while (my $seq_obj = $seqs_factory->next_seq) {
    my $seq_name = $seq_obj->id;
    $seqs{$seq_name} = $seq_obj;
  }

  my $embl_obj = undef;
  if (defined $output) {
    $embl_obj = Bio::PSU::SeqFactory->make(-file   => ">$output",
					   -format => 'embl');
  }
  else {
    $embl_obj = Bio::PSU::SeqFactory->make(-fh   => \*STDOUT,
					   -format => 'embl');
  }

  ##########################
  # EMBL file generation
  ##

  print STDERR "generating embl output ...\n";

  my $colour       = 7;
  my $search_type  = "blastx";
  my $feature_type = "mRNA";
  my $method       = "Automatic annotation";

  my @est_accs     = sort_per_key (keys (%seqs));

  if ($debug) {
    print STDERR "EST accs: " . join (', ', @est_accs) . "\n";
  }

  foreach my $est_acc (@est_accs) {
    print STDERR "current EST accession: $est_acc\n";
    if ($debug) {
      print STDERR "get the Sequence object for EST, $est_acc...\n";
    }

    my $seq_obj         = $seqs{$est_acc};
    my $seq_length      = length ($seq_obj->str);
    my $seq_description = $seq_obj->desc;

    if ($debug) {
      print STDERR "Sequence object done, " . Dumper ($seq_obj) . "\n";
    }

    # The Feature object

    my $feature  = undef;
    my @notes = (); 
    my @similarities = (); 
    my $GOs;
    my $product     = undef; 
    my $orf_frame   = 1;
    my $orf_strand  = undef; 
    my $orf_start   = undef; 
    my $orf_end     = undef; 
    my $translation = undef; 
    my $range       = undef;
    my $systematic_id = $est_acc;

    # Primary Name

    # The Systematic Id or the EST name if singlet.
    my $primary_name = $systematic_id;
    my $note = $seq_description;
    if ($seq_description =~ /singlet/i) {
      $seq_description =~ /^([^\s]+)\s(.+)/;

      $primary_name = $1;
      
      # Get rid of the primary name in the note qualifier
      $note = $2;
    }
    
    push (@notes, $note);
    
    if ($debug) {
      print STDERR "primary name: $primary_name\nnote: $note\n";
    }

    ##################################
    # incorporate similarity data
    ##

    my @results_droso_blastx = ();
    my @results_swall_blastx = ();
    if (exists $results_droso_blastx{$est_acc}) {

      if ($debug) {
	print STDERR "Flybase hits parsing...\n";
      }

      @results_droso_blastx = @{$results_droso_blastx{$est_acc}};
      my $i = 0;
      while ($i<@results_droso_blastx) {
	my $blast_hit = $results_droso_blastx[$i];
	my $q_strand = $blast_hit->getQueryStrand;
	my $s_acc = $blast_hit->getSubjectAccession;
 
        # Get the Flybase ID from the description line
	# in order to fix the subject accession number to be the Flybase Identifier
        my $s_desc = $blast_hit->getSubjectDescription;
	my $flybase_id = getFlybaseId ($s_desc);
	$s_acc = $flybase_id;
	my $fb_gene_symbol = "";

        if ($debug) {
          print STDERR "matching Flybase id, $flybase_id\n";
        }

	if ($i == 0) {

          # product from the best Droso hit (or also Anopheles ????)
	  # Using SRS
	  $product = getProduct ($s_acc);

          if ($debug) {
            print STDERR "got product, $product\n";
          }

	  #########################################
	  # Set ORF information if the best Droso hit
	  # => /frame
	  # => /strand - NO See $range->strand
	  # => /orf_start
	  # => /orf_end
	  # => /translation
	  #########################################

	  ############################
	  #
	  # Map the frame into start_codon qualifier...
	  #
	  
	  # ...
	  
	  ############################
	  
	  $orf_frame  = $blast_hit->getQueryFrame;
          $orf_strand = $blast_hit->getQueryStrand;
          $orf_start  = $blast_hit->getQueryStart;
          $orf_end    = $blast_hit->getQueryEnd;

	  my $orf_sequence = $seq_obj->subseq ($orf_start, $orf_end, 0);
	  if ($orf_strand == -1) {
	    my $orf_sequence_rev = $orf_sequence->revcom;
	    $orf_sequence = $orf_sequence_rev;
	  }
          my $protein_sequence = $orf_sequence->translate (1);
	  my $blastx_translation = $protein_sequence->str;

          if ($debug) {
	    print STDERR "blastx translation seq: $blastx_translation\n\n";
          }
	  
	  my $s_start = $blast_hit->getSubjectStart;
	  if ($s_start == 1) {
	    # sequence starts with a Met so don't try to maximize it
	    $translation = $blastx_translation;
	  } 
	  else {
	    # Best ORF
	    
	    my $orf_feature = Bio::PSU::Feature->new (
						      -key    => 'CDS',
						      -start  => 1,
						      -end    => length ($seq_obj->str),
						      -strand => $orf_strand,
						     );
	    my $tmpseq_obj = $seq_obj->clone;
	    $tmpseq_obj->features ($orf_feature);
	    
	    my $whole_orf_sequence  = $orf_feature->translate (1, $orf_frame);
	    my $whole_orfs_str      = $whole_orf_sequence->str;
	    
	    # print STDERR "ORF seq: $whole_orfs_str\n\n";
	    
	    my $max_orf = getMaxOrf ($whole_orfs_str);
	    
	    if ($debug) {
	      print STDERR "max length, orf: " . length ($max_orf) . ", $max_orf\n";
	    }
	    
	    if (index ($max_orf, $blastx_translation) > -1) {
	      print STDERR "\n\t$est_acc - within !!!\n\n";
	      $translation = $max_orf;
	    }
	    elsif (length ($max_orf) < length ($blastx_translation)) {
              print STDERR "\tmax_orf < blastx_translation !!\n";
	      $translation = $blastx_translation;
              # get rid of the stop codons then...
	      $translation =~ s/\*//g;
	    }
	    else {
	      print STDERR "\n\t$est_acc - not within !!!\n\n";
	      # keep the max_orf though, probably due to a stop codon in the blastx alignment !!
	      $translation = $max_orf;
	    }
	  }

	  ##############################
	  #
	  # GO annotations transfert
	  #
	  ##############################
	  
	  my $go_accs;

          print STDERR "s_acc before ga_entries\n";

	  if (exists ($ga_entries{$s_acc})) {
	    my $ga_entries = $ga_entries{$s_acc};
	    ($GOs, $go_accs, $fb_gene_symbol) = getGOannotations ($apph, $ga_entries, $s_acc);
	  }
	}
	
	######################
	# add similarity data
	##

	my $similarity_value = setSimilarityQualifierValue ($blast_hit, $fb_gene_symbol, "Drosophila melanogaster", $s_acc);
	push (@similarities, $similarity_value);

	$i++;
      }
    }
    if (exists $results_swall_blastx{$est_acc}) {

      if ($debug) {
	print STDERR "swall hits parsing...\n";
      }

      @results_swall_blastx = @{$results_swall_blastx{$est_acc}};
      my $i = 0;
      while ($i<@results_swall_blastx) {
	my $blast_hit = $results_swall_blastx[$i];
        my $s_acc = $blast_hit->getSubjectAccession;
	my $q_strand = $blast_hit->getQueryStrand;
        my $q_frame  = $blast_hit->getQueryFrame;

        if (($q_strand != $orf_strand) || ($q_frame != $orf_frame)) {
          if ($q_strand != $orf_strand) {
            print STDERR "Swall hit on a different strand for EST, $est_acc...\n";
          }
          else {
            print STDERR "Swall hit on a different frame for EST, $est_acc...\n";
          }
        }

	my $similarity_value = setSimilarityQualifierValue ($blast_hit, "", "", $s_acc);
	if (defined $similarity_value) {
	  push (@similarities, $similarity_value);
	}
	else {
          if ($debug) {
	    print STDERR "undefined value => swall hit not reported\n";
          }
	}

	$i++;
      }
    }

    # Always generate a mRNA Feature

    $range = Bio::PSU::Range->new(-start     => 1,
				  -end       => $seq_length,
				  -strand    => $orf_strand,
				  -no_5prime => 1,
				  -no_3prime => 1,
				 );

    $feature = Bio::PSU::Feature->new(
				      -key    => $feature_type,
				      -ranges => $range,
				     );
    $feature->qadd ('systematic_id', $systematic_id);
    $feature->qadd ('primary_name', $primary_name);
    if (defined $note_tissue) {
      push (@notes, $note_tissue);
    }
    $feature->qadd ('note', @notes);
    $feature->qadd ('method', $method);
    if (not defined $product) {
      $product = "hypothetical protein";
    }
    $feature->qadd ('product', $product);
    if (@similarities > 0) {
      $feature->qadd ('similarity', @similarities);
    }

    if (defined $GOs && (@$GOs > 0)) {
      $feature->qadd ('GO', @$GOs);
    }
    if (defined $tissue_type) {
      $feature->qadd ('tissue_type', $tissue_type);
    }

    if (defined $orf_frame && defined $orf_start && defined $orf_end && defined $translation) {
      my $note = "translation sequence determined from sequence similarity";
      $feature->qadd ('note', $note);
      $feature->qadd ('orf_frame', $orf_frame);
      # $feature->qadd ('orf_start', $orf_start);
      # $feature->qadd ('orf_end', $orf_end);
    }

    # Get the translation using ESTscan unless there is a BLASTX hit Vs Dmel

    if (not defined $translation) {
      if (not defined $ESTScan_seqs{$est_acc}) {
        print STDERR "No translation sequence generated by ESTScan for EST, $est_acc!!\n";
      }
      else {
        my $estscan_pred  = $ESTScan_seqs{$est_acc};
        my $predicted_cds = $estscan_pred->predicted_cds;
        $translation      = $predicted_cds->seq;
        my $score         = $estscan_pred->score;
        $orf_strand       = $estscan_pred->strand;
        $orf_start        = $estscan_pred->start;
        $orf_end          = $estscan_pred->end;
        my $note          = "translation sequence generated by ESTScan";
        $feature->qadd ('note', $note);
        $feature->qadd ('score', $score);
        $feature->qadd ('orf_start', $orf_start);
        $feature->qadd ('orf_end', $orf_end);
        $feature->strand ($orf_strand);
      }
    }

    if (defined $translation) {
      $feature->qadd ('translation', $translation);
      my $pepobj = Bio::PSU::Seq->new (
                                       -id => $est_acc,
                                       -str => $translation);
      $pepseqs_factory->write_seq ($pepobj);
    }
    
    if (defined $feature) {
      $seq_obj->features ($feature);
    }
    $embl_obj->write_seq ($seq_obj);

  } # end foreach EST key
}

sub setSimilarityQualifierValue {
  my ($blast_hit, $gene_name, $organism, $s_acc) = @_;

  # Give in parameter the s_acc, this way we make sure we get the proper ID
  # for flybase, (set the Flysbase Id) !!!

  my $search_type = $blast_hit->getSearchType;
  my $product = "";

  # get the organism if from Swall, ie if $organism is empty
  if ($organism eq "") {
    # my @fields = qx(getz '[SWALL-acc:$s_acc]|[SWALL-id:$s_acc]' -f "org des gen");
    my @fields = qx(getz '[swall-ALLTEXT:$s_acc]' -f "org des gen");

    if ($debug) {
      print STDERR "org, des, gen: @fields\n";
    }

    $organism = $fields[0];
    if ($debug) {
      print STDERR "getz organism: $organism from swall accession number, $s_acc\n";
    }
    $organism =~ /OS\s+(.+)\./;
    $organism = $1;
    if ($organism =~ /(.+)\s+\(.+/) {
      $organism = $1;
    }
    if ($debug) {
      print STDERR "parsed organism: $organism\n";
    }
    # if it's Drosophila, do not report it as already Flybase report
    if ($organism eq "Drosophila melanogaster") {
      if ($debug) {
        print STDERR "Swall hits Drosophila -> not reported!\n";
      }
      return undef;
 }
    if (@fields >= 2) {
      $product = $fields[1];
      $product =~ /DE\s+(.+)\./;
      $product = $1;
    }
    if (@fields >= 3 && $fields[2] =~ /GN/) {
      $gene_name = $fields[2];
      $gene_name =~ /GN\s+(.+)\./;
      $gene_name = $1;
    }
  }
  else {
    if ($gene_name eq "") {
      # this a hit Vs Flybase but the gene_name is unknown
      $gene_name = $blast_hit->getSubjectGeneName;
    }
    # get the product from Swall description
    # beware sometimes, one flybase accession number goes to several swall entry !
    $product = qx(getz '[swall-ALLTEXT:$s_acc]' -f "des");
    $product =~ /DE\s+(.+)\./;
    $product = $1;
  }

  my $s_length = $blast_hit->getSubjectLength;
  my $percent_id = $blast_hit->getPercent;
  my $score    = $blast_hit->getScore;
  my $evalue   = $blast_hit->getEValue;
  my $q_strand = $blast_hit->getQueryStrand;
  my $q_frame  = $blast_hit->getQueryFrame;

  my $db_name = "SWALL";
  if ($organism =~ /Drosophila/) {
    $db_name = "FB";
  }

  my $similarity_value = "$search_type; $db_name:$s_acc; $organism; $product; $gene_name; $s_length; id=$percent_id%; E()=$evalue; score=$score; query frame = $q_frame, query strand = $q_strand; ; ;";

  if ($debug) {
    print STDERR "similarity  value: $similarity_value\n";
  }  
  return $similarity_value;
}

sub sort_per_key {
  my (@keys) = @_;
  my @singlet_keys = ();
  my @indexes = ();

  foreach my $key (@keys) {
    # it is a contig
    if ($key =~ /CONTIG(\d+)/i) {
      my $index = $1;
      push (@indexes, $index);
    }
    # it is a singlet
    else {
      push (@singlet_keys, $key);
    }
  }

  @indexes = sort { $a <=> $b } @indexes;

  my $i = 0;
  while ($i<@indexes) {
    my $index = $indexes[$i];
    $indexes[$i] = "CONTIG$index";
    $i++;
  }
  @singlet_keys = sort (@singlet_keys);
  push (@indexes, @singlet_keys);
  return @indexes;
}

sub is_in {
  my ($key, @all_keys) = @_;
  
  foreach my $key_tmp (@all_keys) {
    if ($key =~ /^$key_tmp$/) {
      return 1;
    }
  }
  return 0;
}

sub getMaxOrf {
  my ($whole_orfs_str) = @_;
  
  $whole_orfs_str =~ s/\+/\*/g;
  $whole_orfs_str =~ s/#/\*/g;

  my @orfs = split '\*', $whole_orfs_str;
  my $max_length  = 0;
  my $max_orf     = "";

  foreach my $orf (@orfs) {
    if (length ($orf) > $max_length) {
      $max_length = length ($orf);
      $max_orf = $orf;
    }
  }

  return $max_orf;

}

# Get the Flybase Id from the subject sequence description line

sub getFlybaseId {
  my ($s_desc) = @_;

  if ($debug) {
    print STDERR "subject descr: $s_desc\n";
  }
  
  my $pattern = 'FB\w+';
  $s_desc =~ /.+($pattern).+/;
  my $flybase_id = $1;
  
  if ($debug) {
    print STDERR "Flybase Id: $flybase_id\n";
  }

  return $flybase_id;
}


sub getProduct {
  my ($s_acc) = @_;

  # get the product from the swall description field
  my @products = qx(getz '[swall-ALLTEXT:$s_acc]' -f "des");
  
  if ($debug) {
    print STDERR "product array:\n@products\n";
  }
  
  my $product = $products[0];
  $product =~ /DE\s+(.+)/;	  
  $product = $1;
  
  # some cleaning...

  if ($product =~ /\(/) {
    $product =~ /([^\(]+)\(.+/;
    $product = $1;
    # get rid of the ending space
    $product =~ s/\s$//;
  }
  
  if ($product =~ /\.\s$/) {
    $product =~ s/\.\s//;
  }
  
  if ($debug) {
    print STDERR "product from swall, $product.\n";
  }

  # add putative or if no match return "hypothetical protein"
  
  if ($product =~ /\w+/) {
    if (not $product =~ /putative/) {
      $product = "putative $product";
    }
  }
  else {
    # didn't get a product from Flybase
    $product = "hypothetical protein";
  }

  return $product;
}


sub getGOannotations {
  
  my ($apph, $ga_entries, $s_acc) = @_;
  
  my $with    = "FB:$s_acc";
  my $evcode  = "IEA";
  my $dbref   = "GOC:unpublished";
  my $date    = undef;
  my $fb_gene_symbol = "";
  my @go_accs = ();
  my @GOs     = ();

  {
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime;
    $year += 1900;
    $date = sprintf "%s%2.2d%2.2d", $year, $mon, $mday;
  }
  
  # get the GO ids list
  
  foreach my $ga_entry (@$ga_entries) {
    
    next if $ga_entry->getEvidence eq "IEA";
    
    my $go_acc         = $ga_entry->getGOid;
    $fb_gene_symbol    = $ga_entry->getDbObjectSymbol;
    my $ontology_name  = $ga_entry->getAspect;
    if ($ontology_name =~ /^p$/i) {
      $ontology_name   = "process";
    }
    elsif ($ontology_name =~ /^c$/i) {
      $ontology_name  = "component";
    }
    elsif ($ontology_name =~ /^f$/i) {
      $ontology_name  = "function";
    }
    else {
      print STDERR "GO term type unknown, $ontology_name!!!\n";
      print STDERR "Not supposed to get here !!!\n";
      $ontology_name  = "unknown";
    }
    
    # check if the gene association line has not already been printed
    if (! is_in ($go_acc, @go_accs)) {
      
      my $go_term = "unknown";
      
      # Instantiate a term object using GO perl API and the GO MySQL database
      
      my $termobj = $apph->get_term ({acc=>$go_acc});
      if (not defined $termobj) {
	# Probably an obsolete term which has not been updated
	# it is then a synonym which is stored in another table...
	print STDERR "Problem with querying the GO database, can't instantiate GO term object for go acc, $go_acc...\n";
	print STDERR "GO Term name set up as unknown!\n";
	$go_term = "unknown";
      }
      else {
	$go_term = $termobj->name;
      }
		
      my $go_qualifier = "aspect=$ontology_name; GOid=$go_acc; term=$go_term; evidence=$evcode; db_xref=$dbref; with=$with; date=$date";
      
      push (@GOs, $go_qualifier);		
      push (@go_accs, $go_acc);
    }
  }
  
  return (\@GOs, \@go_accs, $fb_gene_symbol);
}
