#!/usr/local/bin/perl5.6.0 -w

=head1 NAME

blast2tab

=head1 SYNOPSIS

 blast2tab < filename

 blast_program | blast2tab

=head1 DESCRIPTION

A simple Blast output filter which presents Blast output as an EMBL
feature table while trapping and discarding HSPs below a threshold
(Score, E-value, P-value, % identity). It applies filters to each HSP
in isolation.

The filter may be supplied with an offset with it will apply to all
hits using the formula actual_offset = (report count *
supplied_offset). This allows a large query to be broken up into
pieces and the hit coordinates reconstructed after searching. The
offset supplied should be equal to the length of the query pieces and
the reports in the same order as the query pieces. If a base filename
is supplied with the -c option then tab file reports will be created
using this name, one tab file per blast report.

=head1 AUTHOR

Keith James (kdj@sanger.ac.uk)

=cut

use strict;
use Getopt::Std;
use Bio::PSU::SearchFactory;
use Bio::PSU::SeqFactory;

my (%opts, $efilter, $ifilter, $pfilter, $sfilter, $offset, $chunk, $key);
my $multiple = 0;
my $usage = "Usage: blast2tab [-eipsm] [-k <key>] [-c <>] ][-o int] < blast_output

\t-e Remove HSPs with E-value > arg
\t-i Remove HSPs with percentage identity < arg
\t-p Remove HSPs with P-value < arg
\t-s Remove HSPs with score < arg

\t-k Key of feature to create, overriding default of (T)BLAST(PNX)_HIT
\t-o Integer offset increment per report
\t-c Create chunked output using this base name

\t -m Multiple Query Sequences BLAST Output
";

getopts('c:e:hmi:k:o:p:s:', \%opts);

defined $opts{h} and die $usage;
defined $opts{e} and $efilter = $opts{e};
defined $opts{i} and $ifilter = $opts{i};
defined $opts{p} and $pfilter = $opts{p};
defined $opts{s} and $sfilter = $opts{s};
defined $opts{k} and $key     = $opts{k};
defined $opts{o} and $offset  = $opts{o};
defined $opts{c} and $chunk   = $opts{c};

defined $opts{m} and $multiple++;

$offset ||= 0;

my $blast = Bio::PSU::SearchFactory->make(-fh      => \*STDIN,
					  -program => 'blast');

my $current_offset = 0;

my $sembl;
my $fth;
if ($multiple) {
  $sembl = Bio::PSU::SeqFactory->make(-fh => \*STDOUT, -format => 'embl');
}
else {
  $fth = Bio::PSU::IO::FTHandler->new;
}
my $seqobj = undef;
#my @features = ();

while (my $result = $blast->next_result)
{

  my $q_id = $result->q_id;

  # print STDERR "Parsing results for query, $q_id...\n";

  if ($multiple) {
    #if (defined $seqobj && @features > 0) {

      # print STDERR "got " . @features . " features\n";

      #$seqobj->features ([@features]);
      #@features = ();
      #$sembl->write_seq ($seqobj);
    #}
    $seqobj = Bio::PSU::Seq->new (-id   => $q_id,
				  -type => 'dna');
  }

  my $chunk_filename;
  if ((not $multiple) && defined $chunk)
    {
      $chunk_filename = sprintf("%s_%d-%d.tab",
				$chunk, $current_offset + 1,
				$current_offset + $offset);
      open(CHUNK, ">$chunk_filename")
	or die "Failed to open $chunk_filename for writing: $!\n";
    }
  
  my $type = $result->type;
  my $db   = $result->database;
  $type =~ s/^(\S+)\s+.*/$1/;
  
  while (my $hit = $result->next_hit)
    {
      
    HSP: while (my $hsp = $hit->next_hsp)
      {
	my $e = $hsp->expect;
	my $i = $hsp->percent;
	my $p = $hsp->pval;
	my $s = $hsp->score;
	
	if (defined $efilter)
	  {
	    next HSP unless $e < $efilter
	  }
	
	if (defined $ifilter)
	  {
	    next HSP unless $i > $ifilter
	  }
	
	if (defined $pfilter)
	  {
	    next HSP unless $p < $pfilter
	  }
	
	if (defined $sfilter)
	  {
	    next HSP unless $s > $sfilter
	  }
	
	my $q_begin  = $hsp->q_begin;
	my $q_end    = $hsp->q_end;
	my $q_strand = $hsp->q_strand;
	my $s_begin  = $hsp->s_begin;
	my $s_end    = $hsp->s_end;
	my $s_strand = $hsp->s_strand;
	my $s_desc   = $hit->s_desc;
	$s_desc =~ s/\"/\"\"/g;
	
	my $location = "";
	
	# strand will be undef for proteins
	if (defined $s_strand)
	  {
	    $location = " strand $s_strand";
	  }
	
	my $note = sprintf("%s similarity (percentage match=$i) to %s %s, %d..%d%s, expect %s, score %s",
			   $type,
			   $hit->s_id,
			   $s_desc,
			   $hsp->s_begin,
			   $hsp->s_end,
			   $location,
			   $e, $s);
	
	if ($q_end < $q_begin)
	  {
	    ($q_end, $q_begin) = ($q_begin, $q_end);
	  }
	
	if ($s_end < $s_begin)
	  {
	    ($s_end, $s_begin) = ($s_begin, $s_end);
	  }
	
	$key ||= $type. "_HIT";
	
	my $annotation =
	  Bio::PSU::Feature->new(-start  => $q_begin + $current_offset,
				 -end    => $q_end + $current_offset,
				 -strand => $q_strand,
				 -key    => $key);
	
	$annotation->qadd('type', $result->type);
	$annotation->qadd('database', $result->database);
	$annotation->qadd('temporary_systematic_id', $hit->q_id);
	$annotation->qadd('note', $note);
	$annotation->qadd('acc', $hit->s_id);
	$annotation->qadd('desc', $s_desc);
	$annotation->qadd('evalue', $e);
	$annotation->qadd('identity', $i);
	$annotation->qadd('score',  $s);
	$annotation->qadd('subject_start', $hsp->s_begin);
	$annotation->qadd('subject_end', $hsp->s_end);
	$annotation->qadd('subject_strand', $hsp->s_strand);
	
	if (not $multiple) {
	  if (defined $chunk)
	    {
	      print CHUNK $fth->make_feature_block($annotation);
	    }
	  else
	    {
	      print $fth->make_feature_block($annotation);
	    }
	}
	else {
	  # print STDERR "add annotation feature object...\n";
	  # push (@features, $annotation);
	  $seqobj->features ($annotation);
	}
      }
    }
  
  $current_offset += $offset;
  
  if ((not $multiple) && defined $chunk)
    {
      close(CHUNK);
    }
  elsif ($multiple) {
    $sembl->write_seq ($seqobj);
  }
}
