#!/usr/local/bin/perl -w

=head1 NAME

blastdb_sync - Synchronizes local versions of Blast databases with the
distributed copies supplied by Blast database tracking system

=head1 SYNOPSIS

 blastdbsync

=head1 DESCRIPTION

Synchronizes local versions of Blast databases with the distributed
copies supplied by Blast database tracking system. Carries out
symlinking or modification of local Blast databases. Simply run on
each cluster to perform an update.

=head1 METHODS

None

=head1 AUTHOR

Keith James (kdj@sanger.ac.uk)

=head1 COPYRIGHT

Copyright (C) 2002 Genome Research Limited. All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind. It may
be used, redistributed and/or modified under the same conditions as
Perl itself.

=cut

use strict;
use File::stat;
use Blastdb::System;

my $pressdb_command = "/usr/local/pubseq/bin/pressdb";
my $setdb_command = "/usr/local/pubseq/bin/setdb";
my $formatdb_command = "/usr/local/pubseq/bin/formatdb";

my $blastdb_main  = "/data/blastdb";
my $blastdb_local = "$blastdb_main/psu";

my $log_home = $blastdb_local;
my $log_file = "$blastdb_local/local_update.log";

my $exit_value = 0;

my $system_name = `/nfs/pathsoft/bin/clu_get_name`;
chomp $system_name;

my $system = Blastdb::System->new({'system' => $system_name});

eval {
    foreach my $db ($system->available_dbs)
    {
        my @components = $db->components;

        foreach my $comp (@components)
        {
            my @files = $comp->files;

            foreach my $file (@files)
            {
                my $filename = $file->filename;

                if (! -e "$blastdb_main/$filename")
                {
                    write_log("Warning:\tFile $blastdb_main/$filename is missing",
                              $log_file);
                }
                elsif (! -e "$blastdb_local/$filename")
                {
                    write_log("   Info:\tCreating new symlink for $blastdb_main/$filename",
                              $log_file);
                    symlink "$blastdb_main/$filename", "$blastdb_local/$filename";
                    next;
                }
                else
                {
                    my  $main_age = stat("$blastdb_main/$filename")->mtime;
                    my $local_age = stat("$blastdb_local/$filename")->mtime;

                    if ($main_age > $local_age)
                    {
                        # Commented out by kdj as SWALL now arrives in this form
                        # update_local($db->name, $blastdb_main, $blastdb_local, $filename);
                    }
                }
            }
        }
    }
};

if ($@)
{
    write_log("   Error: " . $@, $log_file);
    $exit_value = -1;
}

# Delete any dangling symlinks left where files have been removed from
# $blastdb_main after they were made
my @all_files = glob("$blastdb_local/*");

foreach my $file (@all_files)
{
    if (! -e $file)
    {
        write_log("   Info:\tRemoving old symlink for $file", $log_file);
        unlink $file;
    }
}

write_log("   DONE", $log_file);

exit($exit_value);

=head2 update_local

 Title   : update_local
 Usage   : update_local($db_name, $main_dir, $local_dir, $filename)
 Function: Updates local version of a Blast database under the
         : Blast database tracking system
 Returns : void
 Args    : Database name e.g. "swall", path of unmodified dir,
           path of dir for local mods, filename

=cut

sub update_local
{
    my ($db_name, $main_dir, $local_dir, $filename) = @_;

    my $tmp_dir = "$local_dir/tmp";
    mkdir $tmp_dir or die "Failed to create temporary directory $tmp_dir: $!\n";

    my $main_filename = "$main_dir/$filename";
    my $tmp_filename = "$tmp_dir/$filename";
    my $local_filename = "$local_dir/$filename";

    if ($db_name eq 'swall')
    {
        if ($main_filename !~ /[ahd|atb|bsq|phr|pin|psq]$/)
        {
            modify_swall_header($main_filename, $tmp_filename);
            create_peptide_blast_indices($tmp_filename);
        }

        # Now move the new indices into place
        system "mv -f $tmp_dir/* $local_dir";
    }

    rmdir "$tmp_dir" or warn "Failed to remove temporary directory $tmp_dir: $!\n";
}

=head2 modify_swall_header

 Title   : modify_swall_header
 Usage   : modify_swall_header($main_filename, $tmp_filename)
 Function: Creates a swall database with PSU style Fasta header
 Returns : void
 Args    : filename of unmodified file, filename of locally modified copy

=cut

sub modify_swall_header
{
    my ($main_filename, $tmp_filename) = @_;

    open(IN, "<$main_filename") or die "Failed to open $main_filename for reading: $!\n";
    open(OUT, ">$tmp_filename") or die "Failed to open $tmp_filename for writing: $!\n";

    while (<IN>)
    {
        s/^(\S+)\s+(\S+)\s+(.*?)Desc:\s+(.*)/$1 $2 $4/;
        print OUT $_;
    }

    close(OUT);
    close(IN);
}

=head2 create_dna_blast_indices

 Title   : create_dna_blast_indices
 Usage   : create_dna_blast_indices($file)
 Function: Indexes a Fasta file for WU and NCBI Blast
 Returns : void
 Args    : Fasta filename

=cut

sub create_dna_blast_indices
{
    my $file = shift;

    my $return;
    my $error_message = "create_dna_blast_indices failed";

    $return = 0xffff & system "$pressdb_command $file";
    if ($return)
    {
        $return && write_log("Warning:\tpressdb failed on $file",
                             $log_file);
        die $error_message;
    }
    else
    {
        write_log("   Info:\tpressdb run on $file", $log_file);
    }

    $return = 0xffff & system "$formatdb_command -p F -i $file";
    if ($return)
    {
        $return && write_log("Warning:\tformatdb failed on $file",
                             $log_file);
        die $error_message;
    }
    else
    {
        write_log("Info:\tformatdb run on $file",
                  $log_file);
    }
}

=head2 create_peptide_blast_indices

 Title   : create_peptide_blast_indices
 Usage   : create_peptide_blast_indices($file)
 Function: Indexes a Fasta file for WU and NCBI Blast
 Returns : void
 Args    : Fasta filename

=cut

sub create_peptide_blast_indices
{
    my $file = shift;

    my $return;
    my $error_message = "create_peptide_blast_indices failed";

    $return = 0xffff & system "$setdb_command $file";
    if ($return)
    {
        write_log("Warning:\tsetdb failed on $file",
                  $log_file);
        die $error_message;
    }
    else
    {
        write_log("   Info:\tsetdb run on $file",
                  $log_file);
    }

    $return = 0xffff & system "$formatdb_command -p T -i $file";
    if ($return)
    {
        $return && write_log("Warning:\tformatdb failed on $file",
                             $log_file);
        die $error_message;
    }
    else
    {
        write_log("   Info:\tformatdb run on $file",
                  $log_file);
    }
}

=head2 write_log

 Title   : write_log
 Usage   : write_log($statement, $log_file)
 Function: Writes a log statement to a log file with a timestamp
 Returns : void
 Args    : Statement to write, log file to write to

=cut

sub write_log
{
    my ($statement, $log_file) = @_;

    my $undef_log_statement = "<UNDEFINED LOG STATEMENT>";
    my $empty_log_statement = "<EMPTY LOG STATEMENT>";

    if (! defined $statement)
    {
        $statement = $undef_log_statement
    }
    elsif ($statement =~ /^\s*$/)
    {
        $statement = $empty_log_statement
    }

    my $time = get_time();

    open(LOG, ">>$log_file") or warn "Failed to open log $log_file\n";
    print LOG "<$time> $statement\n";
    close(LOG);
}

=head2 get_time

 Title   : get_time
 Usage   : get_time()
 Function: Returns current time formatted for human readability
 Returns : Time string
 Args    : None

=cut

sub get_time
{
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) =
        localtime(time);

    my @wdays = qw(Sun Mon Tue Wed Thur Fri Sat);
    my @mons  = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);

    $year += 1900;

    $mon  = $mons[$mon];
    $wday = $wdays[$wday];

    return sprintf("%02d:%02d:%02d %02d %s %s",
                   $hour, $min, $sec,
                   $mday, $mon, $year);
}
