#!/usr/local/bin/perl5.6.1 -w

=head1 NAME

blastwrap - Simple Blast search wrapper for aliasing databases

=head1 SYNOPSIS

 blastwrap -p

 blastwrap blastn %emnew queryseq.nt

 blastwrap wublastp %swall queryseq.aa

=head1 DESCRIPTION

This script is a wrapper for blast searches allowing a single database
alias for split databases. To use it precede your normal Blast command
(any version of Blast) with 'blastwrap' and substitute the database
name for an alias with a leading '%' e.g. %embl_other, %embl_htg,
%swissprot.

To view the current aliases and database path use 'blastwrap -p'. The
script checks that the database is present and readable before
commencing the search. If it is not, or if the Blast program exits
with an error, the script will terminate.

The script will treat commands starting with 'blastall' as if they are
using NCBI Blast 2 and will generate a command line which searches a
virtual database consisting of all the parts combined and gives a
single result.

=head1 METHODS

None

=head1 AUTHOR

Keith James (kdj@sanger.ac.uk)

=head1 COPYRIGHT

Copyright (C) 2001 Genome Research Limited. All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind. It may
be used, redistributed and/or modified under the same conditions as
Perl itself.

=cut

use strict;
use Getopt::Std;
use File::Basename;
use Blastdb::System;

my $BLASTDB = "";
my $BLASTMAT = "";
my $WUBLASTMAT = "";

# This is passed to the system() call before calling the blast
# executable to maximise the memory available for blast
my $magic_ulimit_string = "ulimit -d unlimited;";

my (%opts, $db_alias, $print_db, $verbose);
my $usage = "Usage: blastwrap [-pv] <blast command>

\t-p print out the known databases and aliases
\t-v verbose mode; prints command lines executed to STDERR
";

getopts('hpv', \%opts);

defined $opts{h} and die $usage;
defined $opts{p} and $print_db = 1;
defined $opts{v} and $verbose = 1;

unless (defined $verbose) { $verbose = 0 }

# Try ulimit and if it fails don't do it later
if (system "$magic_ulimit_string")
{
    print STDERR "ulimit failed - continuing anyway\n";
    $magic_ulimit_string = "";
}

my $command = join(" ", @ARGV);

my $system_name = `/nfs/pathsoft/bin/clu_get_name`;
chomp $system_name;
print STDERR "[$system_name]\n" if $verbose;

if (defined $print_db)
{
    my $system = Blastdb::System->new({'system' => $system_name});
    my @available_dbs = $system->available_dbs;
    print_db(\@available_dbs);
    exit;
}

if ($command =~ / %(\S+)/)
{
    $db_alias = $1;
}
else
{
    print STDERR "Executing: $command\n" if $verbose;
    print STDERR "Setting BLASTMAT to /data/blastdb/aa\n" if $verbose;
    print STDERR "Setting WUBLASTMAT to /data/blastdb\n" if $verbose;

    local $ENV{BLASTMAT} = "/data/blastdb/aa";
    local $ENV{WUBLASTMAT} = "/data/blastdb";

    my $return = 0xffff & system "$magic_ulimit_string $command";
    if ($return)
    {
        trap("<user database>", $command,
             "search exited with return code: $return")
    }

    exit;
}

my $system = Blastdb::System->new({'system' => $system_name});

my $db = $system->current_db_by_name($db_alias);

if (! defined $db)
{
    trap($db_alias, $command,
         "database object could not be retrieved");
}

check_db($db, $command);

print STDERR "Setting BLASTDB to '", $db->path, "'\n" if $verbose;
print STDERR "Setting BLASTMAT to '", $db->path, "/aa'\n" if $verbose;
print STDERR "Setting WUBLASTMAT to '", $db->path, "'\n" if $verbose;

$BLASTDB = $db->path;
$BLASTMAT = $db->path. "/aa";
$WUBLASTMAT = $db->path;

my $comp_names = join(" ", $db->component_names);
$command =~ s/%\S+/\"$comp_names\"/;

print STDERR "Executing: $command\n" if $verbose;

{
    local $ENV{BLASTDB} = $BLASTDB;
    local $ENV{BLASTMAT} = $BLASTMAT;
    local $ENV{WUBLASTMAT} = $WUBLASTMAT;

    my $return = 0xffff & system "$magic_ulimit_string $command";
    if ($return)
    {
        trap($db_alias, $command,
             "search exited with return code: $return")
    }
}

sub check_db
{
    my ($db, $command) = @_;

    my $name = $db->name;
    my $path = $db->path;
    my @components = $db->components;

    foreach my $comp (@components)
    {
        my @files = $comp->files;

        foreach my $file (@files)
        {
            my $filename = "$path/" . $file->filename;

            unless (-e "$filename")
            {
                trap($db->name, $command,
                     "database file $filename not found")
            }

            unless (-r "$filename")
            {
                trap($db->name, $command,
                     "database file $filename not readable")
            }
        }
    }
}

sub print_db
{
    my ($dbs) = @_;

    foreach my $db (@$dbs)
    {
        printf "  %s - version %s [%s]\n",
            $db->name, $db->version, $db->path;

        my @components = $db->components;

        foreach my $comp (@components)
        {
            print "    ", $comp->name, "\n";
        }

        print "\n";
    }
}

sub trap
{
    my ($db_alias, $command, $message) = @_;

    die "Search of $db_alias failed: $message when running: $command";
}

