#!/usr/local/bin/perl5.6.1 -w

=head1 NAME

cat_embl - concatenates EMBL files

=head1 SYNOPSIS

 cat_embl < *.embl > out

=head1 DESCRIPTION

This script takes any number of EMBL files on STDIN and concatenates
them in that order, writing the new file to STDOUT. The headers of the
first entry are retained, except for the accession number, which is
set to unknown.

=head1 METHODS

None

=head1 AUTHOR

Keith James (kdj@sanger.ac.uk)

=head1 COPYRIGHT

Copyright (C) 2002 Genome Research Limited. All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind. It may
be used, redistributed and/or modified under the same conditions as
Perl itself.

=cut

use strict;
use Getopt::Std;
use Bio::SeqIO;
use Bio::Location::SplitLocationI;

my $ei = Bio::SeqIO->new(-fh     => \*STDIN,
                         -format => 'embl');

my $eo = Bio::SeqIO->new(-fh     => \*STDOUT,
                         -format => 'embl');

my $offset = 0;

my $cat_seq;

while (my $seq = $ei->next_seq)
{
    if (defined $cat_seq)
    {
        my $str = $cat_seq->seq . $seq->seq;
        my $pri =
            Bio::PrimarySeq->new(-seq      => $str,
                                 -id       => $cat_seq->id,
                                 -alphabet => $cat_seq->alphabet);
        $cat_seq->primary_seq($pri);

        my @features = $seq->all_SeqFeatures;
        $seq->flush_SeqFeatures;

        foreach my $f (@features)
        {
            my $loc = $f->location;

            if ($loc->isa('Bio::Location::SplitLocationI'))
            {
                foreach my $subloc ($loc->sub_Location)
                {
                    offset_location($subloc, $offset);
                }
            }
            else
            {
                offset_location($loc, $offset);
            }

            $f->attach_seq($cat_seq->primary_seq);
            $cat_seq->add_SeqFeature($f);
        }
    }
    else
    {
        $cat_seq = $seq;
    }

    $offset += $seq->length;
}

$eo->write_seq($cat_seq);

sub offset_location
{
    my ($loc, $offset) = @_;

    my $max_start = $loc->max_start;
    my $min_start = $loc->min_start;
    my $max_end = $loc->max_end;
    my $min_end = $loc->min_end;

    defined $max_start && $loc->max_start($max_start + $offset);
    defined $min_start && $loc->min_start($min_start + $offset);
    defined $max_end && $loc->max_end($max_end + $offset);
    defined $min_end && $loc->min_end($min_end + $offset);

    return $loc;
}
