#!/usr/local/bin/perl

# External Modules

# Constraint variables declaration before using them
use strict;

use Cwd;

# Bio modules
# PSU
use Bio::PSU::SeqFactory;

# Benchmark Module
use Benchmark;
##################################################################

=head1 NAME

 - clusters_with_ESTs.pl [-help] -cl [clusters File] -est [ESTs File] -gss [GSSs File]

=head1 SYNOPSIS


=head1 DESCRIPTION

  Parse the clusters data set generated by a clustering process with phrap on GSSs and ESTs sequences.
  The clusters containing ESTs members are added to a new clusters File.
  
  @author axk

=cut

my $t1 = Benchmark->new ();

my $clusters_file = "";
my $ESTs_file     = "";
my $GSSs_file     = "";

my $_help         = "-help";
my $_cluster      = "-cl";
my $_est          = "-est";
my $_gss          = "-gss";

my $lgth = @ARGV + 0;
# if no args, display the help message
if (($lgth<6) || ($lgth>6)) {
  display_help ();
  exit 1;
}
else {
  my $i = 0;
  foreach my $element (@ARGV) {
    if ($element eq $_help) {
      display_help ();
      exit 1;
    }
    elsif ($element eq $_cluster) {
      $clusters_file = $ARGV[$i+1];
    }
    elsif ($element eq $_est) {
      $ESTs_file = $ARGV[$i+1];
    }
    elsif ($element eq $_gss) {
      $GSSs_file = $ARGV[$i+1];
    }
    # if at least one of the options is not recognized, display the help message and exit
    elsif ($element =~ /^-/) {
      print STDERR "option not recognized: $element\n";
      display_help ();
      exit 1;
    }
    $i ++;
  }
}

$clusters_file =~ /(\w+)\..+/;
my $output_file = $1 . ".clusters_with_ESTs_members.fasta";
my $curdir = cwd;

# create a hashtable with ESTs ACC and GSSs ACC
print STDERR "processing...\n";

my %acc_hash = ();

open (EST, "< $ESTs_file") or die "can't open the ESTs file, $ESTs_file !!!\n";

while (<EST>) {
  my $line = $_;
  $line =~ />(\w+:\w+)\s+.+/;
  my $acc_number = $1;
  $acc_hash{$acc_number} = "est";
}
close (EST);

open (GSS, "< $GSSs_file") or die "can't open the GSSs file, $GSSs_file !!!\n";

while (<GSS>) {
  my $line = $_;
  $line =~ />(\w+:\w+)\s+.+/;
  my $acc_number = $1;
  $acc_hash{$acc_number} = "gss";
}
close (GSS);

print STDERR ("hashtables done\n");
print STDERR "Filtering the clusters file in order to keep only the clusters containing ESTs...\n";

# foreach sequence in the clusters File, if with one GSSs member, copy the contig into the output

# open (CLUSTERS, "< $clusters_file") or die "can't open the clusters file, $clusters_file !!!\n";
my $clusters_file_path = $curdir . "/$clusters_file";
# open (FILEOUT, ">$output_file") or die "can't open the output file, $output_file !!!\n";

my $seqi = Bio::PSU::SeqFactory->make(-file   => "<$clusters_file_path",
				      -format => 'fasta');
open (FILEOUT, "> $output_file") or die "can't create this file: $!";
my $seqo = Bio::PSU::SeqFactory->make(-fh     => \*FILEOUT, 
				      -format => 'fasta');

while (my $sequence = $seqi->next_seq) {
  my $desc_seq = $sequence->desc;
  # print STDERR ("desc: $desc_seq\n");
  # filtering the singlets sequences
  if ($desc_seq =~ /\s*(\d+)\s+members, constructed from\s+(.+)/) {
    # print STDERR ("nb_mb - mb_list : $1 - $2\n");
    my $nb_members = $1;
    my $members_list = $2;
    while ($members_list =~ /(\w+:\w+)\,\s+(.+)/) {
      # print STDERR ("members_list - acc_nb mb_list : $members_list - $1 - $2\n");
      my $acc_nb = $1;
      $members_list = $2;
      if ($acc_hash{$acc_nb} eq "est") {
	$seqo->write_seq($sequence);
      }
    }
    chomp ($members_list);
    my $acc_nb = $members_list;
    if ($acc_nb eq $acc_hash{$acc_nb}) {
      $seqo->write_seq($sequence);
    }
  }
}

close FILEOUT;

print STDERR ("filtering done.\n");
print STDERR ("the clusters can be found in the $output_file file\n");

my $t2 = Benchmark->new ();

print STDERR "\nprocessing completed : ", timestr (timediff ($t2, $t1)), "\n";


###################################################################
# The end
###################################################################

 # display the help message

sub display_help {
  print "command: clusters_with_ESTs.pl [-help] -cl [clusters File] -est [ESTs File] -gss [GSSs File]\n";
  print "\t-help a help message\n";
  print "\t-cl the clusters File\n";
  print "\t-est the ESTs File\n";
  print "\t-gss the GSSs File\n";
}
