#!/usr/local/bin/perl5.6.1 -w

=head1 NAME

dgpi2embl - Generate EMBL annotation of GPI-anchored signal sequences predicted
by DGPI

=head1 SYNOPSIS

Examples:

 dgpi2embl -t 11 -y gram- embl_file
 dgpi2embl -t 1 -y euk -c 0 embl_file

The EMBL entry is written to STDOUT.

=head1 DESCRIPTION

This program requires an EMBL file containing annotated CDS
features. The features will be translated using the specified
translation table (default is table 1; standard), except for those
with /pseudo, /partial or /transl_except qualifiers.

The output is parsed by the script to produce an EMBL entry on
STDOUT.

Setting the -y switch (to euk, gram+ or gram- is important, as is
choosing the correct translation table with the -t switch. The -c
switch sets the colour of the features (using the Artemis colour
codes).

=head1 METHODS

None

=head1 AUTHOR

Arnaud Kerhornou (axk@sanger.ac.uk)
Initially adapted from signalp2embl.pl from Keith James (kdj@sanger.ac.uk)

=head1 COPYRIGHT

Copyright (C) 2003 Genome Research Limited. All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind. It may
be used, redistributed and/or modified under the same conditions as
Perl itself.

=cut

use strict;
use File::Temp qw/tempfile/;
use Getopt::Std;

# BioPSU
use Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper;
use Bio::PSU::SeqFactory;
use Bio::PSU::FeatureFactory;
use Bio::PSU::Utils::Dgpi;

my $debug = 0;

# Configuration of SignalP variables
my $bindir     = "/nfs/pathsoft/external/bio-soft/dgpi";
my $dgpi_bin   = "$bindir/dgpi";
my $dgpi_args  = "-output=text -q";

my $unknown_id = "Unknown_CDS";

my (%opts, $colour, $regex, $table, $type);
my $annot = 0;
my $output_file = undef;
my $usage = "Usage: dgpi2embl [-a] ][-c <colour>] [-t <table>] -y <type> -o <output file> <dgpi options>

\t-a annotate original CDS with results
\t-c Colour code of features  (default  0)

\t-r specify regex used to recognise a systematic gene name
\t-t Translation table number (default 1; standard numbering)
\t-y Organism type            (gram+, gram-, euk)
\t-o Output EMBL file         (default is standard output)
\tTables available are:

\t 1) Standard
\t 2) Vertebrate Mitochondrial
\t 3) Yeast Mitochondrial
\t11) Bacterial
";

# Process options
getopts('ac:hp:r:t:y:o:', \%opts);

defined $opts{h} and die $usage;
defined $opts{a} and $annot++;
defined $opts{c} and $colour      = $opts{c};
defined $opts{r} and $regex       = $opts{r};
defined $opts{t} and $table       = $opts{t};
defined $opts{y} and $type        = $opts{y};
defined $opts{o} and $output_file = $opts{o};

# Default colour is white
$colour = 0 unless defined $colour;

# Default translation table is universal (1)
$table ||= '1';

unless (defined $type && $type =~ /^(gram\+$|gram-|euk)/) { die $usage }

my $feature_type = "GPI-Anchor";
my $algorithm = "DGPI v2.04";

my $file = shift;
unless (defined $file) { die $usage }

my $id_helper =
  Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper->new(-selection_regex => $regex);

# Read EMBL stream and get entry
my $es = Bio::PSU::SeqFactory->make(-file   => "<$file",
                                    -format => 'embl');

my $eo;
if (defined $output_file) {
  $eo = Bio::PSU::SeqFactory->make(-file   => ">$output_file",
                                   -format => 'embl');
}
else {
  $eo = Bio::PSU::SeqFactory->make(-fh     => \*STDOUT,
                                   -format => 'embl');
}

while (my $entry = $es->next_seq)
{
    my %id_to_cds;

    foreach my $f ($entry->features)
    {
        # Discard non-CDS features
        next unless $f->key eq 'CDS';

        # Discard any CDS with /pseudo or /partial qualifier
        next if $f->qexists('partial');
        next if $f->qexists('pseudo');

        # Discard CDS with partial start
        if ((${$f->{ranges}}[0]->fuzzy_start != 0) &&
            ($f->strand == 1))
        {
            next;
        }

        if ((${$f->{ranges}}[0]->fuzzy_end != 0) &&
            ($f->strand == -1))
        {
            next;
        }

        my ($fh, $tempfile) = tempfile("DGPI.XXXXXX", UNLINK => 1);
        close($fh);

        # Create Fasta output stream
        my $fo = Bio::PSU::SeqFactory->make(-file   => ">$tempfile",
                                            -format => 'fasta');

        my $id = $id_helper->get_systematic_id($f);
        $id_to_cds{$id} = $f;

        if ($debug) {
	  print STDERR "\n";
          print STDERR "processing CDS, $id...\n";
        }

        # Translate with specified translation table

        my $frame = 0;

        if ($f->qexists('codon_start'))
        {
            my @codon_start = $f->qvalues('codon_start');
            $frame = $codon_start[0] - 1;
        }

        my $f_pep = $f->translate($table, $frame);
        my $f_pep_str = $f_pep->str;

        $f_pep_str =~ s/\*$//;

        # Translation exceptions need to be skipped for now
        if (grep /\*/, $f_pep_str)
        {
            if ($f->qexists('transl_except'))
            {
                next;
            }
            else
            {
                warn "Stop codon present in ID $id. Skipping this CDS\n";
                next;
            }
        }

        # Set the id for the Fasta header
        $f_pep = Bio::PSU::Seq->new(-str      => $f_pep_str,
                                    -id       => $id,
                                    -type     => 'protein');

        # Write translation to temp file as Fasta
        $fo->write_seq($f_pep);

	# Run DGPI
	
	if ($debug) {
	  print STDERR "Running DGPI with the following command:\n";
	  print STDERR "$dgpi_bin $dgpi_args -file=$tempfile\n";
	}
	
        my @dgpi_errput = qx/$dgpi_bin $dgpi_args -file=$tempfile/;

        my $dgpi_output = "$tempfile.txt";

	# Check if the file exists, if it doesn't, 
	# that means DGPI didn't work properly and was not able to generate an output.

	if (-f $dgpi_output) {

	  if ($debug) {
	    print STDERR "parsing DGPI output...\n";
	  }

	  my $dgpi_parser = Bio::PSU::Utils::Dgpi->new (
							file => $dgpi_output,
							debug => $debug
						       );
	  
	  while (my $dgpi_prediction = $dgpi_parser->next_prediction()) {
	    
            if ($dgpi_prediction->{has_cleavage_site}) {
	    
	      if ($debug) {
		print STDERR "GPI-anchored predicted as well as the cleavage site\n";
	      }

	      my $cprob    = $dgpi_prediction->{cleavage};
	      my $coords   = $dgpi_prediction->{coords};
	      my $position = $dgpi_prediction->{position};
	      
	      if ($debug) {
	        print STDERR "coords: $coords\n";
	      }
	      
	      $coords =~ /([^,]+),(.+)/;
	      my $cpos1 = $1;
	      my $cpos2 = $2;
	      
	      my $note   = "$feature_type Signal predicted for $id by $algorithm with cleavage site probability $cprob near $position";
	      
	      if ($debug) {
	        print STDERR "cpos1, cpos2: $cpos1, $cpos2\n";
	      }
	      
	      my $cds = $id_to_cds{$id};
	      my ($cds_start, $cds_end) = ($cds->start, $cds->end);
	      my $cds_strand = $cds->strand;
	    
	      my ($nuc_start, $nuc_end);
	    
	      if ($cds_strand == 1)
	      {
		$nuc_start = $cds_start + (3 * ($cpos1 - 1));
		$nuc_end   = $cds_start + (3 * ($cpos2 - 1)) + 2;
	      }
	      else
	      {
		$nuc_start = $cds_end - (3 * ($cpos2 + 0)) + 1;
		$nuc_end   = $cds_end - (3 * ($cpos1 + 0)) + 1 + 2;
	      }
	    
	      if ($cpos1 < 1 or $cpos2 < 0)
	      {
		print STDERR "Prediction was outside of CDS - skipping addition of feature\n"
	      }
	      else
	      {
		my $misc = Bio::PSU::Feature->new(-key => 'misc_feature',
					          -start   => $nuc_start,
					          -end     => $nuc_end,
					          -strand  => $cds_strand);
		
		$misc->qadd('id', $id);
		$misc->qadd('note', $note);
		$misc->qadd('colour', $colour);
		
		$misc->qadd('type', $feature_type);
		$misc->qadd('algorithm', $algorithm);
		$misc->qadd('coord', "$coords");
		$misc->qadd('cleavage', $cprob);
		
		if ($debug) {
		  print STDERR "adding misc_feature...\n";
		}

		$entry->features($misc);
	      }
	    
	      # Added facility for annotation of original features
	      if ($annot)
		{
		$f->qadd('note', $note);
	      }
            }
            else {
              # GPI-anchored predicted but no cleavage site
	      # just a note
	      
	      if ($debug) {
		print STDERR "GPI-anchored predicted but no cleavage site\n";
	      }
	      
	      if ($annot)
	      {
		my $note   = "$feature_type Signal predicted for $id by $algorithm, no cleavage site predicted";
		$f->qadd('note', $note);
	      }

            }
	  } # end DGPI output parsing
	  
	  $dgpi_parser->close_fh();
	  unlink "$tempfile";
	  unlink "$tempfile.txt";

	}
	else {
	  print STDERR "DGPI could not predict anything for this CDS, $id!!\n";
	}
    } # end foreach feature

    $eo->write_seq($entry);

} # end foreach sequence
