#!/usr/local/bin/perl5.6.1 -w

=head1 NAME

embl2KEGGlinks - create links to KEGG summarising EC number data
from an EMBL file

=head1 SYNOPSIS

Examples:

  embl2KEGGlinks s_typhi S_typhi.embl

=head1 DESCRIPTION

This script reads EC numbers from the EC_number qualifier of all CDS
features in an EMBL file. It queries KEGG to find the KEGG maps which
include these enzymes and then constructs the following web pages; a
summary page linking each CDS to one or more KEGG maps, a KEGG map
page which lists all maps referenced and finally, one page for each
KEGG maps referenced, which contains links to all CDS which occur on
that map.

The script requires a base name for the generated files and an EMBL
file (which may contain more than one entry). The files created are:

1 x <base name>.gene.html           (gene list)
1 x <base name>.maps.html           (KEGG map summary)
n x <base name>.<KEGG map ID>.html  (one for each map)

=head1 METHODS

None

=head1 AUTHOR

Keith James (kdj@sanger.ac.uk)

=head1 COPYRIGHT

Copyright (C) 2002 Genome Research Limited. All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind. It may
be used, redistributed and/or modified under the same conditions as
Perl itself.

=cut


use strict;
use LWP::UserAgent;
use Bio::PSU::SeqFactory;
use Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper;

my $usage = "Usage: embl2KEGGlinks.pl <base name> <regex> <embl file>

\t <base name> specifies the base name for generated files
\t <regex> specifies a regular expression to distinguish systematic ID
\t <embl file> specifies the EMBL file to process
";

my $basename = shift;
unless (defined $basename) { die $usage; }

my $regex = shift;
unless (defined $regex) { die $usage; }

my $file = shift;
unless (defined $file) { die $usage; }

my $genepage = "gene";
my $mapspage = "maps";

my $filename = "$basename.$genepage.html";

unless (defined $file) { die $usage }

my $id_helper =
  Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper->new(-selection_regex => $regex);

# Read EMBL stream and get entries
my $es = Bio::PSU::SeqFactory->make(-file => "<$file");

my $agent = LWP::UserAgent->new;
$agent->agent('Wellcome Trust Sanger Institute (Pathogen Sequencing)) ' . $agent->agent);
$agent->proxy('http', 'http://wwwcache.sanger.ac.uk:3128/');

my %maps_to_enzymes;
my %maps_to_descs;
my %maps_to_cds;
my %ec_to_maps;

my @page_lines;

my $sleep_time = 5;

while (my $e = $es->next_seq)
{
 FEAT: foreach my $f ($e->features)
    {
        # Discard non-CDS features
		next unless $f->key eq 'CDS';

        my @ec_numbers = $f->EC_number;
        my $id = $id_helper->get_systematic_id($f);
        my $product = $f->product;
		$product ||= "undefined product";

        my $cds_seen = 0;

        push(@page_lines, "      <tr>");
        push(@page_lines, "        <td>$id</td>");
        push(@page_lines, "        <td>$product</td>");

        foreach my $ec_number (@ec_numbers)
        {
            unless ($ec_number =~ /^[\d+-\.?]{4}/)
            {
                print STDERR "Warning: skipping invalid EC number $ec_number\n";
                next;
            }

            my $maps;

            # Check cache for maps for this EC number
            if (exists $ec_to_maps{$ec_number})
            {
                $maps = $ec_to_maps{$ec_number};
            }
            else
            {
                print STDERR "Sleeping for $sleep_time seconds...\n";
                sleep($sleep_time);
                $maps = get_maps($ec_number);
            }

            # If a CDS has >1 EC number we only report the CDS once
            if (! $cds_seen)
            {
                $cds_seen++;

                push(@page_lines, "<td>");

                foreach my $map_id (keys %$maps)
                {
                    push(@page_lines, "<a href=\"./$basename.$map_id.html\">$map_id</a>");
                    push(@page_lines, "<a name=\"$id\">");

                    push(@{$maps_to_cds{$map_id}}, $f);
                }

                push(@page_lines, "</td>");
            }

            # If maps were cached for this EC number then we are done
            if (exists $ec_to_maps{$ec_number})
            {
                next FEAT;
            }
            else
            {
                $ec_to_maps{$ec_number} = $maps;
            }

            foreach my $map_id (keys %$maps)
            {
                if (! exists $maps_to_descs{$map_id})
                {
                    $maps_to_descs{$map_id} = $maps->{$map_id};
                }

                if (! exists $maps_to_enzymes{$map_id})
                {
                    $maps_to_enzymes{$map_id} = [];
                }

                push(@{$maps_to_enzymes{$map_id}}, $ec_number);
            }
        }

        push(@page_lines, "      </tr>");

    }
}

# Print gene list page

open(OUT, ">$filename") or die "Failed to open $filename for writing: $!\n";

print OUT "<html>\n";
print OUT "  <body>\n";
print OUT "    <h1>Gene list</h1>\n";
print OUT "    <table border=\"1\">\n";
print OUT "      <tr><th>Gene ids(s)</th> <th>Product</th> <th>KEGG map(s)</th></tr>\n";

print OUT join("\n", @page_lines);

print OUT "    </table>\n";
print OUT "  </body>\n";
print OUT "</html>\n";
close(OUT);

# Print pathway summary page

$filename = "$basename.$mapspage.html";

open(OUT, ">$filename") or die "Failed to open $filename for writing: $!\n";

print OUT "<html>\n";
print OUT "  <body>\n";
print OUT "    <h1>KEGG Pathway Maps</h1>\n";
print OUT "    <table border=\"1\">\n";
print OUT "      <tr><th>KEGG Map ID</th> <th>Description</th></tr>\n";

foreach my $map_id (keys %maps_to_descs)
{
    print OUT "      <tr>\n";
    print OUT "        <td>";
    print OUT "<a href=\"./$basename.$map_id.html\">$map_id</a>";
    print OUT "        </td>\n";
    print OUT "        <td>";
    print OUT $maps_to_descs{$map_id};
    print OUT "        </td>\n";
    print OUT "        </tr>\n";
}

print OUT "    </table>\n";
print OUT "  </body>\n";
print OUT "</html>\n";

close(OUT);

# Print pathway pages

foreach my $map_id (keys %maps_to_cds)
{
    my $filename = "$basename.$map_id.html";
    my $ec_string = join("+", @{$maps_to_enzymes{$map_id}});

    open(OUT, ">$filename") or die "Failed to open $filename for writing: $!\n";

    print OUT "<html>\n";
    print OUT "  <body>\n";
    print OUT "    <h1>", $maps_to_descs{$map_id}, "</h1>\n";
    print OUT "    <h3>View <a href=\"$basename.$mapspage.html\">pathway list</a></h3>";
    print OUT "    <h3>View this pathway at <a href=\"http://www.genome.ad.jp/dbget-bin/show_pathway?$map_id+$ec_string\">KEGG</a></h3>";
    print OUT "    <table border=\"1\">\n";
    print OUT "      <tr><th>Genes represented in this pathway</th></tr>\n";

    foreach my $f (@{$maps_to_cds{$map_id}})
    {
        my $id = $id_helper->get_systematic_id($f);
        print OUT "      <tr>\n";
        print OUT "        <td>";
        print OUT "<a href=\"./$basename.$genepage.html#$id\">$id</a>";
        print OUT "        </td>\n";
        print OUT "      <tr>\n";
    }

    print OUT "    </table>\n";
    print OUT "  </body>\n";
    print OUT "</html>\n";

    close(OUT);
}

sub get_maps
{
    my $ec_number = shift;

    my $maps = {};

    my $request = HTTP::Request->new(GET => "http://www.genome.ad.jp/dbget-bin/www_bget?enzyme+$ec_number");
    print STDERR "Trying $ec_number ...\n";

    my $response = $agent->request($request);

    if ($response->is_success)
    {
        my @lines = split("\n", $response->content);

        foreach my $line (@lines)
        {
            if ($line =~ /PATH: +<A HREF[^>]+>(MAP\d+)<\/A> (.+)$/)
            {
                my $map_id   = $1;
                my $map_desc = $2;

                $maps->{$map_id} = $map_desc;

				print STDERR "Got reference to map $map_id: $map_desc\n";
            }
        }
    }
	else
	{
		print STDERR "Request failed\n";
	}

    return $maps;
}
