#!/usr/local/bin/perl5.6.1 -w

#
# this script to wrap an ESTs Clustering process
# Input:   one Fasta File with the all set of sequences
# Ouptput: a unique fasta file with a the contigs and singlets.
#

#
# e.g. perl -w est_clustering.pl -i [input_file] -o [output_directory] -c [config_file] -r true -e estate -m babel -s true -n DICTY.CONTIG.
#

# process : 
# 1/ Run RepeatMasker - optional => -repeatMasker true/false 
# 2/ Run a clustering tool ESTate or Phrap or cap3 or tigr => -estTool estate/phrap/cap3/tigr
# 3/ If ESTate, run phrap or cap3 to generate the consensus sequence for each cluster
# 4/ Run a perl program who generate the unique fasta file with contigs and singlets information

# External Modules

# Constraint variables declaration before using them
use strict;
##########################################

use Data::Dumper;

# Benchmark Module
use Benchmark;
##################################################################

# Bio modules
# PSU
use Bio::PSU::SeqFactory;
use Bio::PSU::Seq;

BEGIN {
  my $_path="/nfs/pathsoft/prod/perl";
  unshift (@INC, $_path);
}

# Module for managing global application configuration information
use ConfigReader::Simple;
##########################################

=head1 NAME

est_clustering.pl - a ESTs Clustering wrapper script

=head1 SYNOPSIS

  perl -w est_clustering.pl -help -i [input_file|directory_input_file] -o [output_directory] -c [config_file] -r [true|false] -e [estate|phrap|cap3|tigr] -m [machine] -singlets [true|false] -n [name of the contigs]

  -help a help message
  -i the input
  -o the output
  -c the config file
  -r repeats masking process
  -e the ESTs clustering tool
  -m the machine where the process will be done
  -singlets add the ESTs singletons into the output fasta file
  -n name of the contigs
=head1 DESCRIPTION

  This script to wrap an ESTs Clustering process

  @author Arnaud Kerhornou - axk@sanger.ac.uk

=cut

=head1 ALGORITHM

=cut

##########################################
=pod

=item * Configuration loading from the command line

=cut
##########################################

my $t1 = Benchmark->new ();

my $debug = 0;

my $_help          = "-help";
my $_input         = "-i";
my $_output        = "-o";
my $_config        = "-c";
my $_repeatMasking = "-r";
my $_estTool       = "-e";
my $_machine       = "-m";
my $_singlets      = "-singlets";
my $_name          = "-n";

# no input per default
my $input_name     = "-1";

# The options per default

my $output_name    = "clusters";
my $config_name    = "/nfs/pathsoft/prod/perl/est_clustering.cfg";
my $repeat_masking = "false";
my $est_tool       = "phrap";
my $machine        = "pcs3";
my $singlets       = "true";
my $contigName     = "CONTIG";

my $cur_dir = qx/pwd/;
chomp ($cur_dir);

# Check if in the command line there is an input file name (-i option)
# the other options are not compulsory (set of options per default)
my $lgth = @ARGV + 0;
# if no args, display the help message
if ($lgth == 0) {
  print "specify at least an input file with the option \"-i\"\n";
  display_help ();
  exit 1;
}
elsif ($lgth >= 1) {
  my $i = 0;
  foreach my $element (@ARGV) {
    if ($element eq $_help) {
      display_help ();
      exit 1;
    }
    elsif ($element eq $_input) {
      $input_name = $ARGV[$i+1];
    }
    elsif ($element eq $_output) {
      $output_name = $ARGV[$i+1];
    }
    elsif ($element eq $_config) {
      $config_name = $ARGV[$i+1];
    }
    elsif ($element eq $_repeatMasking) {
      $repeat_masking = $ARGV[$i+1];
    }
    elsif ($element eq $_estTool) {
      $est_tool = $ARGV[$i+1];
    }
    elsif ($element eq $_machine) {
      $machine = $ARGV[$i+1];
    }
    elsif ($element eq $_singlets) {
      $singlets = $ARGV[$i+1];
    }
    elsif ($element eq $_name) {
      $contigName = $ARGV[$i+1];
    }
    # if at least one of the options is not recognized, display the help message and exit
    elsif ($element =~ /^-/) {
      print STDERR "option not recognized: $element\n";
      display_help ();
      exit 1;
    }
    $i ++;
  }
}
if ($input_name eq "-1") {
  print STDERR "ERROR - specify at least the name of an input file with the -i option\n";
  print STDERR "Syntax : est_clustering.pl -i [input_file]\n";
  exit 1;
}

####################################################
# INITIALISATION
####################################################

my $_singletons_name_file = undef;
my $_singletons_name_suffix = undef;
if (lc ($est_tool) eq "estate") {
  $_singletons_name_file = "$output_name/singletons.fasta";
}
elsif (lc ($est_tool) eq "phrap") {
  $_singletons_name_suffix = ".singlets";
}
elsif (lc ($est_tool) eq "cap3") {
  $_singletons_name_suffix = ".cap.singlets";
}
elsif (lc ($est_tool) eq "tigr") {
  $_singletons_name_suffix = ".singlets";
}

if ($est_tool =~ /cap3|tigr|phrap/i) {
  $_singletons_name_file = $input_name . $_singletons_name_suffix;
}
my $singletons_file_path = $cur_dir . "/" . $_singletons_name_file;  

my $phrap_version = "phrap.manyreads";
# my $phrap_version = "phrap";

my %ace_parsing_options;
if ($est_tool eq "phrap" && $phrap_version eq "phrap") {
  %ace_parsing_options = (
			  'ace_name_suffix'  => '.ace',
			  'contig_pattern'   => 'DNA\s(\S+)',
			  'AF_test'          => '^Assembled_from',
			  'AF_pattern'       => 'Assembled_from\*?\s+(\S+)\s+-?\d+\s+-?\d+',
			  'member_sequence'  => 'DNA\s(\S+)',
			 );
}
elsif ($est_tool eq "phrap" && $phrap_version eq "phrap.manyreads") {
  %ace_parsing_options = (
			  'ace_name_suffix' => '.ace',
			  'contig_pattern'  => 'CO\s(\S+)\s+.*',
			  'AF_test'         => '^AF',
			  'AF_pattern'      => 'AF\s+(\S+)\s+(C|U)\s+-?\d+.*',
			  'member_sequence' => 'RD\s+(\S+).*',
			 );
}
elsif ($est_tool eq "cap3") {
  %ace_parsing_options = (
			  'ace_name_suffix' => '.cap.ace',
			  'contig_pattern'  => 'CO\s(\S+)\s+.*',
			  'AF_test'         => '^AF',
			  'AF_pattern'      => 'AF\s+(\S+)\s+(C|U)\s+-?\d+.*',
			  'member_sequence' => 'RD\s+(\S+).*',
			 );
}
elsif ($est_tool eq "tigr") {
  %ace_parsing_options = (
			  'ace_name_suffix'  => '.ace',
			  'contig_pattern'   => 'DNA\s(\S+)',
			  'AF_test'          => '^Assembled_from',
			  'AF_pattern'       => 'Assembled_from\*?\s+(\S+)\s+-?\d+\s+-?\d+',
			  'member_sequence'  => 'DNA\s(\S+)',
			 );
}

my $input_name_base         = "$input_name.results";
my $output_name_base        = "$output_name";

# remove previous output directories
  
my @args = "rm -rf $output_name_base*";
system (@args);

##########################################
=pod

=item * Config File Loading

=cut
##########################################

# phrap parameters
  
my $_word        = "word";
my $_display     = "display";
my $_min_overlap = "min_overlap";
my $_node_space  = "node_space";

# cap3 parameters

my $_overlap_pid = "p";

my $cfg = ConfigReader::Simple->new ("$config_name", ["$_word $_display $_min_overlap $_node_space $_overlap_pid"]) or die "No such file or directory";
$cfg->parse ();



##########################################
=pod

=item * Repeat Masking

=cut
  ##########################################
  
if (lc ($repeat_masking) eq "true") {
  $input_name = repeatMasking ($input_name);
}

##########################################
=pod

=item * ESTs clustering

=cut
##########################################

if (not $est_tool =~ /estate/i) {
  @args = "mkdir $output_name";
  system (@args);
}

if (lc ($est_tool) eq "estate") {
  runESTate ($cfg, $input_name, $output_name, $_word, $_display, $_min_overlap);
}
elsif (lc ($est_tool) eq "phrap") {
  runPhrap ($cfg, $input_name, $_node_space, $phrap_version);
}
elsif ($est_tool =~ /cap3/i) {
  runCap3 ($cfg, $_overlap_pid, $input_name);
}
elsif ($est_tool =~ /tigr/i) {
  runTigr ($input_name);
  print STDERR "Generating the singlets fasta file...\n";
  generateSingletsFile ($input_name, "$input_name.fasta", $singletons_file_path);
  print STDERR "Generation done!...\n";
}
else {
  print STDERR "can not process the clustering with the following tool, $est_tool\n";
  exit 0;
}

my @a_info = ();
my @clusters_info = ();

if ($est_tool =~ /phrap|cap3|tigr/i) {

  # ace parsing
  
  @clusters_info = getClustersInfoFromAce ($cur_dir, $input_name, %ace_parsing_options);
  
  # generate the "clusters" directory with a set of cluster files
  
  ####
  ## problem with phrap - get rid of the clusters of one member
  ## leave them at this point, just generate a new set on the same dir
  ## in part 3, using @a_info
  ####

  cluster_files_generation ($cur_dir, $output_name, @clusters_info);
}

##################################
# part 3
##################################

# 3/ Run the perl script to generate a unique fasta file with contigs and (optionally) singlets information

my %mappingId2Description = getMappingId2Description ($input_name);

@a_info = gather_contig_info ($cur_dir, $output_name, $contigName, %mappingId2Description);

if ($est_tool =~ /phrap/i) {
  # get the ESTs sequence objects

  print STDERR "Generating an ESTs sequence objects hash...\n";

  my %ESTs = getESTsSequenceObjects ($input_name);

  print STDERR "Generation done!\n";
  print STDERR "Reprocessing the clusters directory because phrap generates one-member clusters and we are considering them as singlets...\n";

  cluster_files_generation_phrap ($cur_dir, $output_name, \%ESTs, @a_info);

  print STDERR "cluster files generation done!\n";

}

if ($singlets =~ /true/i) {
  my @a_info_tmp = gather_singlets_info ($singletons_file_path);
  @a_info = concat (\@a_info, \@a_info_tmp);
}

####
## Here is the process where clusters of one member become singlets
####
unique_fasta_file_generation ($cur_dir, $input_name_base, @a_info);

print "process done!\n";

my $t2 = Benchmark->new ();
print  "\nTotal : ", timestr (timediff ($t2, $t1)), "\n";

###################################################################
# The end
###################################################################

 # display the help message

sub display_help {
  print "est_clustering.pl $_help $_input [input_file] $_output [output_directory] $_config [config_file] $_repeatMasking [true|false] $_estTool [estate|phrap|cap3|tigr] $_machine [machine] $_singlets [true|false] $_name [Contigs name]\n\n";
  print "\t$_help a help message\n";
  print "\t$_input the input       - (no default, always specify it)\n";
  print "\t$_output the output      - (default: \"clusters\" directory)\n";
  print "\t$_config the config file - (default file: /nfs/pathsoft/prod/perl/est_clustering.cfg)\n";
  print "\t$_repeatMasking repeats masking process  - (no repeats masking process per default)\n";
  print "\t$_estTool the ESTs clustering tool - (default: phrap)\n";
  print "\t$_machine the machine where the process will be done, via lsrun - (default: babel)\n";
  print "\t$_singlets add the ESTs singletons sequences into the output fasta file - (default: true)\n";
print "\t$_name Prefix Name of the contigs - (default: CONTIG)\n\n";
}


sub repeatMasking {
  my ($input_name) = @_;
  
  print STDOUT "\nrepeats masking processing ...\n";
  print STDOUT "command: lsrun -v /bin/csh -c \"unlimit;ulimit -a;RepeatMasker $input_name\"\n";
  
  my @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit -a;RepeatMasker $input_name\"");
  system (@args);
  
  # the input is now $input concatenated with ".masked"
  $input_name = $input_name . ".masked";

   print STDOUT "\ndust processing ...\n";

  @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit -a;dust $input_name > $input_name.dust\"");
  system (@args);

  # the input is now $input concatenated with ".dust"
  $input_name = $input_name . ".dust";
  
  print STDOUT "repeats masking done\n";
  print STDOUT "the input file is now $input_name\n\n";

  return $input_name;
}


sub runESTate {
  my ($cfg, $input_name, $output_name, $_word, $_display, $_min_overlap) = @_;
  
  print STDOUT "ESTs clustering processing using ESTate ...\n";
  my $input_name_ava = $input_name . ".ava";
  my $input_name_avarc = $input_name . ".avarc";
  my $word = $cfg->get($_word);
  
  print STDOUT "command: lsrun -v precluster -d $input_name -A $input_name_ava -R $input_name_avarc -w $word\n";
  @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit;/nfs/disk222/yeastpub/bio-soft/estate/bin/precluster -d $input_name -A $input_name_ava -R $input_name_avarc -w $word\"");
  # @args = ("unlimit;ulimit;/nfs/disk222/yeastpub/bio-soft/estate/bin/precluster -d $input_name -A $input_name_ava -R $input_name_avarc -w $word");
  system (@args);
  
  # display ON/OFF
  my $display = $cfg->get($_display);
  # minimum overlap alignment score
  my $min_overlap = $cfg->get($_min_overlap);
  
  print STDOUT "command: lsrun -v estcluster -d $input_name -A $input_name_ava -R $input_name_avarc -o $output_name -t $display -j $min_overlap\n";
  @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit -a;/nfs/disk222/yeastpub/bio-soft/estate/bin/estcluster -d $input_name -A $input_name_ava -R $input_name_avarc -o $output_name -t $display -j $min_overlap\"");
  # @args = ("unlimit;ulimit -a;/nfs/disk222/yeastpub/bio-soft/estate/bin/estcluster -d $input_name -A $input_name_ava -R $input_name_avarc -o $output_name -t $display -j $min_overlap");
  system (@args);
  
  print STDOUT "ESTs clustering using ESTate is done\n";  
}


sub runPhrap {
  my ($cfg, $input_name, $_node_space, $phrap_version) = @_;
  
  print STDOUT "ESTs clustering processing using Phrap ...\n";
  
  my $node_space = $cfg->get($_node_space);

  # if the database is pretty big, you can use phrap.manyreads
  
  if ($phrap_version eq "phrap") {
    my @args = ("lsrun -m $machine -v /bin/csh -c \"unlimit;ulimit -a;phrap $input_name -ace -node_space $node_space \"");
    system (@args);
  }
  else {
    my @args = ("lsrun -m $machine -v /bin/csh -c \"unlimit;ulimit -a;phrap.manyreads $input_name -ace -node_space $node_space \"");
    system (@args);
  }
  
  print STDOUT "\nESTs clustering using Phrap is done\n";
  
}


sub runCap3 {
  my ($cfg, $_overlap_pid, $input_name) = @_;
  
  print STDOUT "ESTs clustering processing using cap3 ...\n";
  
  my $overlap_pid = $cfg->get($_overlap_pid);

  print STDOUT "Overlap percentage identity cutoff of $overlap_pid\n";

  my @args = ("lsrun -m $machine -v /bin/csh -c \"unlimit;ulimit -a;cap3 $input_name -$_overlap_pid $overlap_pid >! $input_name.align \"");
  system (@args);
}


sub runTigr {
  my ($input_name) = @_;
  
  print STDOUT "ESTs clustering processing using Tigr ...\n";
  
  my @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit -a; TIGR_Assembler -A $input_name.ace -n $input_name -g 8 -l 40 -e 15 -p 97.5 -s -a $input_name.align -f $input_name.fasta $input_name.scratch < $input_name >! $input_name.asm\"");
  system (@args);
}


sub generateSingletsFile {
  my ($input_name, $tigr_results_file, $singletons_file_path) = @_;

  my $input_seqF = Bio::PSU::SeqFactory->make (
					       -file   => "<$input_name",
					       -format => 'fasta',
					       -type   => 'dna'
					      );
  my $tigr_seqF = Bio::PSU::SeqFactory->make (
					       -file   => "<$tigr_results_file",
					       -format => 'fasta',
					       -type   => 'dna'
					      );
  my $singlets_seqF = Bio::PSU::SeqFactory->make (
						  -file => ">$singletons_file_path",
						  -format => 'fasta'
						 );
  
  my @tigr_ids = ();
  while (my $tigr_seq_obj = $tigr_seqF->next_seq) {
    my $tigr_seq_id = $tigr_seq_obj->id;
    if (not $tigr_seq_id =~ /^$input_name/) {
      # this is a singlet EST
      push (@tigr_ids, $tigr_seq_id);
    }
  }

  while (my $input_seq_obj = $input_seqF->next_seq) {
    my $input_seq_id = $input_seq_obj->id;
    if (is_in ($input_seq_id, \@tigr_ids)) {
      $singlets_seqF->write_seq ($input_seq_obj);
    }
  }
}


sub getClustersInfoFromAce {
  my ($cur_dir, $input_name, %ace_parsing_options) = @_;

  my @clusters_info = parse_ace_file ($cur_dir, $input_name, %ace_parsing_options);

  return @clusters_info;
}


sub parse_ace_file { 
  my ($cur_dir, $input_name, %ace_parsing_options) = @_;
  
  my $ace_name = "$input_name" . $ace_parsing_options{ace_name_suffix};

  print STDERR "parsing the ace file, $ace_name...\n";
  
  # open the ace file
  
  open (ACE, "< $ace_name") or die "can't open the ace file, $ace_name !!!\n";
  
  # parse it
  
  # this data structure stores everything !!!!
  # this array will be returned to main
  my @clusters_list = ();
  
  while (<ACE>) {
    if ($_ =~ /$ace_parsing_options{contig_pattern}/) {
      my $contig_name = $1;
      if ($debug) {
	print STDERR "new cluster, contig name : $contig_name\n";
      }
      # retrieve Contig sequence
      my $sequence_contig = "";
      if ($debug) {
	print STDERR "parsing contig sequence...\n";
      }
      my $new_line = <ACE>;
      while (lc ($new_line) =~ /(a|c|t|g|n)\S+/) {
	chomp ($new_line);
	$sequence_contig = $sequence_contig . $new_line;
	$new_line = <ACE>;
      }
      if ($debug) {
	print STDERR "sequence_contig : $sequence_contig\n";
      }

      $sequence_contig =~ s/\*//g;
      
      # then Base quality
      while (not $new_line =~ /$ace_parsing_options{AF_test}/) {
	$new_line = <ACE>;
      }
      my @members = ();
      # then information about the number of members within the current cluster
      while ($new_line =~ /$ace_parsing_options{AF_test}/) {
	$new_line =~ /$ace_parsing_options{AF_pattern}/;
	my $member_name = $1;
	if ($debug) {
	  print STDERR "found member $member_name, members list : @members, new_line : $new_line\n";
	}
	if (not is_in ($member_name, \@members)) {
	  push (@members, $member_name);
	}
	$new_line = <ACE>;
      }
      my $members_nb = (@members+0);
      my $cpt = 0;
      my @members_list = ();

      if ($debug) {
	print STDERR "found $members_nb members for the current cluster\n";
      }

      # Retrieve the name and the sequence of each member
      
      while ($cpt < $members_nb) {
	if ($new_line =~ /$ace_parsing_options{member_sequence}/) {
	  my $member_name = $1;
	  if ($debug) {
	    print STDERR "member name : $member_name\n";
	  }
	  my $new_line = <ACE>;
	  my $sequence = "";
	  while (lc ($new_line) =~ /(a|c|t|g|n)\S+/) {
	    chomp ($new_line);
	    $sequence = $sequence . $new_line;
	    $new_line = <ACE>;
	  }
	  if ($debug) {
	    print STDERR "member sequence : $sequence\n";
	  }
	  
	  # the sequence may contain * which don't fit with the fasta format, so remove them from the sequence
	  # this * is for a shift in the alignment

	  $sequence =~ s/\*//g;

	  my %h_member = (line_info => "$member_name",
			  sequence  => "$sequence",
			 );
	  $cpt ++;
	  push (@members_list, \%h_member);
	}
	$new_line = <ACE>;
      }

      my $line_info = $contig_name . " of $members_nb members : " . join ( ", ", @members);

      my %h_contig = (line_info => "$line_info",
		      sequence  => "$sequence_contig",
		     );

      # this data structure stores all the information about one contig and its members
      my %cluster_info = (cluster_info => \%h_contig,
			  members_info => \@members_list,
			 );
      push (@clusters_list, \%cluster_info);
      
      if ($debug) {
	print STDERR "Contig done - current line : $new_line\n";
      }
      
    }
  }
  close (ACE);
  print STDERR "parsing $ace_name done - " . (@clusters_list+0) . " clusters found\n";

  return @clusters_list;
}


sub cluster_files_generation {
  my ($cur_dir, $output_name, @clusters_list) = @_;
  my $output_name_path = $cur_dir . "/" . $output_name;

  print STDERR "generating " . (@clusters_list+0) . " cluster files\n";

  foreach my $cluster_info_ref (@clusters_list) {
    my %cluster_info = %$cluster_info_ref;
    my $h_contig_ref = $cluster_info{cluster_info};
    my %h_contig = %$h_contig_ref;

    my $members_list_ref = $cluster_info{members_info};
    my @members_list = @$members_list_ref;
    my $line_info = $h_contig{line_info};
    $line_info =~ /Contig(\d+).+/;
    my $contig_index = $1;

    my $cluster_file_name = "cluster" . $contig_index . ".fasta";
    my $cluster_file_name_path = $output_name_path . "/" . $cluster_file_name;
    
    my $seqF = Bio::PSU::SeqFactory->make (
					  -file => ">$cluster_file_name_path",
					  -format => 'fasta'
					 );

    # write members information

    foreach my $h_member_ref (@members_list) {
      my %h_member = %$h_member_ref;
      my $sequence = $h_member{sequence};
      $sequence = uc ($sequence);

      my $seq = Bio::PSU::Seq->new (
				    -id   => $h_member{line_info},
				    -str  => $sequence,
				    -type => 'dna'
				   );
      $seqF->write_seq ($seq);
    }

    my $sequence = $h_contig{sequence};
    $sequence = uc ($sequence);
    
    my $seq = Bio::PSU::Seq->new (
				  -id   => $h_contig{line_info},
				  -str  => $sequence,
				  -type => 'dna'
				 );
    $seqF->write_seq ($seq);
  }
}


sub cluster_files_generation_phrap {
  my ($cur_dir, $output_name, $ESTs, @a_info) = @_;
  my $output_name_path = $cur_dir . "/" . $output_name;

  print STDERR "removing old cluster directory, $output_name_path...";
  qx/rm -rf $output_name_path/;
  qx/mkdir $output_name_path/;

  foreach my $h_element (@a_info) {

      my @members = @{$h_element->{members}};
      if (@members == 1) {
        # cluster of one member, this is therefore a singlet !!
        next;
      }

      my $contig_name = $h_element->{contig_name};
      my $fasta_file_name = "$contig_name.fasta";
      my $fasta_file_path = "$output_name_path" . "/" . "$fasta_file_name";
      my $seqF = Bio::PSU::SeqFactory->make (
					 -file => ">$fasta_file_path",
					 -format => 'fasta'
					);
 
      # The members of the cluster

      foreach my $member_id (@members) {
        my $seqobj_member = $ESTs->{$member_id};
        $seqF->write_seq ($seqobj_member);
      }

      # The contig sequence

      my $contig_sequence = $h_element->{sequence};
      $contig_sequence    = uc ($contig_sequence);

      my $seqobj = Bio::PSU::Seq->new (
		  		       -id   => $h_element->{line_info},
				       -str  => $contig_sequence,
				       -type => 'dna'
				      );
      $seqF->write_seq ($seqobj);
  }
}


sub gather_contig_info {
  my ($cur_dir, $output_name, $contigPrefixName, %mappingId2Description) = @_;

  my @a_info = ();
  my $index = 1;

  my $output_name_path = $cur_dir . "/" . $output_name;
  # print "output_path: $output_name_path.\n";
  
  # output directory files
  opendir (THISDIR, $output_name_path) or die "can not read this directory $output_name_path ...";
  my @files = grep {$_ =~ /^cluster.*fasta/} readdir THISDIR;
  closedir THISDIR;

  # print STDOUT "files: \n@files\n";
  
  foreach my $file_name (@files) {
    my $file_name_path = $output_name_path . "/" . $file_name;
    # print "file: $file_name_path\n";
    my @tmp = split (/\./, $file_name);
    my $cluster_name = $tmp[0];
    # print "cl name: $cluster_name\n";
    my $cluster_number = substr ($cluster_name, 7, length ($cluster_name)) + 0;
    # print STDOUT "cluster_number: $cluster_number\n";
    
    # my $contig_name = $contigPrefixName . $cluster_number;
    my $contig_name = $contigPrefixName . $index;
 
    my $contig_info = $contig_name;
    my @members_list = ();
    my $sequence = undef;

    my $seqF = Bio::PSU::SeqFactory->make (
					   -file   => "<$file_name_path",
					   -format => 'fasta',
					   -type   => 'dna'
					  );

    while (my $seq_obj = $seqF->next_seq) {
      my $seq_name = $seq_obj->id;
      my $description = $seq_obj->desc;

      if (not ($description =~ /MEMBERS/i)) {
	push (@members_list, $seq_name);
      }
      else {
	$sequence = $seq_obj->str;
      }
    }

    my $nb_members = @members_list;
    # print "found $nb_members members for contig $contig_name\n";
    my $line_info = "";
    if ($nb_members > 1) {
      $line_info = $contig_name . " " . $nb_members . " members";
      my $i = 0;
      my $member = $members_list[$i];
      $line_info = $line_info . ", constructed from $member";
      $i ++;
      while ($i < $nb_members) {
	$member = $members_list[$i];
	$line_info = $line_info . ", $member";
	$i ++;
      }
      $index++;
    }
    elsif ($nb_members == 1) {

      # only one member => it is a singlet !!

      my $description = $mappingId2Description{$members_list[0]};
      $contig_name    = $members_list[0];
      $line_info      = $members_list[0] . " $description SINGLET";
      # don't incremente the index as it is a singlet !!
    }
     
    # build the hash element
    
    my %h_element = (contig_name => $contig_name,
		     members     => [@members_list],
		     line_info   => $line_info,
		     sequence    => $sequence,
		    );
    
    # print "members list       = ", join(', ', @members_list), "\n";
    # print "h_element{members} = ", Data::Dumper::Dumper ($h_element{'members'});
    
    # add data
    
    push (@a_info, \%h_element);
  }
  
  return @a_info;
}


sub gather_singlets_info {
  my ($singletons_file_path) = @_;

  my @a_info = ();

  print STDERR "parsing singlets file, $singletons_file_path...\n";

  my $seqF = Bio::PSU::SeqFactory->make (
					 -file   => "<$singletons_file_path",
					 -format => 'fasta',
					 -type   => 'dna'
					);
  while (my $seq_obj = $seqF->next_seq) {
    my $line_info = $seq_obj->id . " " . $seq_obj->desc . " SINGLET";
    my %h_element = (
		     line_info => $line_info,
		     sequence  => $seq_obj->str,
		    );
    push (@a_info, \%h_element);
  }
  return @a_info;
}


sub unique_fasta_file_generation {
  my ($cur_dir, $input_file_base, @a_info) = @_;
  
  my $fasta_file_name = $input_file_base;
  
  print "generation of fasta file $fasta_file_name .. - " . @a_info . " entries\n";
  
  my $seqF = Bio::PSU::SeqFactory->make (
					 -file => ">$fasta_file_name",
					 -format => 'fasta'
					);

  foreach my $href (@a_info) {
    my $sequence = $href->{sequence};
    $sequence = uc ($sequence);

    my $seq = Bio::PSU::Seq->new (
				  -id   => $href->{line_info},
				  -str  => $sequence,
				  -type => 'dna'
				 );
    $seqF->write_seq ($seq);
  }
  print "generation of fasta file $fasta_file_name done!\n";
}  


sub getMappingId2Description {
  my ($input_name) = @_;
  
  my %mappingId2Description;
  
  my $seqF = Bio::PSU::SeqFactory->make (
					 -file   => "<$input_name",
					 -format => 'fasta',
					 -type   => 'dna',
					);
  
  while (my $seq_obj = $seqF->next_seq) {
    $mappingId2Description{$seq_obj->id} = $seq_obj->desc;
  }
  
  return %mappingId2Description;
}


sub concat {
  my ($aref1, $aref2) = @_;
  my @array1 = @$aref1;
  my @array2 = @$aref2;
  
  foreach my $element (@array2) {
    push (@array1, $element);
  }
  
  return @array1;
}


sub is_in {
  my ($element, $aref) = @_;
  my @list = @$aref;

  foreach my $e2 (@list) {
    if ($element eq $e2) {
      return 1;
    }
  }
  return 0;
}


sub getESTsSequenceObjects {
  my ($input_name) = @_;
  my %ESTs;

  my $seqF = Bio::PSU::SeqFactory->make (
                                         -file   => "$input_name",
				         -format => "fasta",
					 -type   => 'dna'
                                        );

  while (my $seqobj = $seqF->next_seq) {
    my $seqid = $seqobj->id;
    $ESTs{$seqid} = $seqobj;
  }
 
  return %ESTs;
}
