#!/usr/local/bin/perl5.6.0 -w

# External Modules

# Constraint variables declaration before using them
use strict;
##########################################

use Data::Dumper;

# Benchmark Module
use Benchmark;
##################################################################

eval "require Bio::PSU::Utils::EstwiseParsing";
# use Bio::PSU::Utils::EstwiseParsing;

##########################################
=pod

=item * Configuration loading from the command line

=cut
##########################################

my $t1 = Benchmark->new ();


############################
# Debug boolean
##

my $debug = 0;

my $_help           = "-help";
my $_estwise        = "-estwise";
my $_sequences_url  = "-urlfasta";
my $_title          = "-title";

# no estwise ouput per default
my $estwise = "-1";

######################
# HTML Page Title
##
my $title = "Glossina morsitans morsitans ESTs Annotation report";

######################
# Sequence URL
##
my $sequences_url = "ftp://ftp.sanger.ac.uk/pub/pathogens/Glossina/ESTs";

######################
# Mapping Files
##
my $pfam_mapping_file = "/nfs/pathsoft/external/bio-soft/proteinMappingFiles/pfamName2Accession.mapping";
my $ipro_mapping_file = "/nfs/pathsoft/external/bio-soft/proteinMappingFiles/protein2ipr.dat";
my $go_mapping_file   = "/nfs/pathsoft/external/bio-soft/proteinMappingFiles/interpro2go.txt";

#######################
# OTHER URLs (Pfam, Interpro, GO)
##
my $_pfam_url = "http://www.sanger.ac.uk/cgi-bin/Pfam/getacc?";
my $_ipro_url = "http://www.ebi.ac.uk/interpro/IEntry?ac=";
my $_go_url = "http://golgi.ebi.ac.uk/ego/QuickGO?mod=display&query=";
my $_go_url_2 = "&entry=";

my $i = 0;
foreach my $element (@ARGV) {
  if ($element eq $_help) {
    display_help ();
    exit 1;
  }
  elsif ($element eq $_estwise) {
    $estwise = $ARGV[$i+1];
  }
  elsif ($element eq $_sequences_url) {
    $sequences_url = $ARGV[$i+1];
  }
  elsif ($element eq $_tiltle) {
    $title = $ARGV[$i+1];
  }
  # if at least one of the options is not recognized, display the help message and exit
  elsif ($element =~ /^-/) {
    print STDERR "option not recognized: $element\n";
    display_help ();
    exit 1;
  }
  $i++;
}

if (($estwise eq "-1") || ($sequences_url eq "-1") || ($title eq "-1")) {
  print STDERR "option(s) not specified!!\n";
  display_help ();
  exit 1;
}


####################
# Parse estwisedb file output
##

my $estwise_parser = Bio::PSU::Utils::EstwiseParsing->new (
							   estwiseOutput => $estwise,
							   pfamMapping => $pfam_mapping_file,
							   iproMapping => $ipro_mapping_file,
							   goMapping   => $go_mapping_file,
							  );

if ($debug) {
  print STDERR "parsing object: " . Dumper ($estwise_parser) . "\n";
}

print STDERR "parsing estwisedb output...\n";
my %results_estwise = $estwise_parser->parse;

if ($debug) {
  print STDERR "#############\nestwise results object: " . Dumper (%results_estwise) . "\n#############\n";
}

#########################
# Generate HTML output
##
my $html_estwise_name = "$estwise.html";
generateEstwiseHTMLOutput ($html_estwise_name, \%results_estwise);


#########################
# Generate The Gene Association File
##
my $assocFileName = "gene_association";
generateGeneAssociationFile ($assocFileName, \%results_estwise);

print STDERR "generation done.\n";

my $t2 = Benchmark->new ();

print STDERR "\nTotal : ", timestr (timediff ($t2, $t1)), "\n";

#################
# end
##

#############################
# display the help message
##

sub display_help {
  print "estwise2html.pl $_help $_estwise [ESTs File] $_sequences_url [ESTs file URL] $_title [HTML Page Title]\n\n";
  print "\t$_help a help message\n";
  print "\t$_estwise the estwise output file or directory - (no default, always specify it)\n";
  print "\t$_sequences_url the sequences URL - (default is Glossina ESTs ftp URL)\n";
  print "\t$_title the HTML Page Title\n\n";
}

###################################
# Generate a gene association file
# a set of GO IDs for each gene sequence (here ESTs)
##

sub generateGeneAssociationFile {
  my ($assocFileName, $results_estwise_ref) = @_;
  my %results_estwise = %{$results_estwise_ref};

  my @keys = keys (%results_estwise);
  @keys = sort_per_key (@keys);
  my $db = "GeneDB";
  my $evidence = "IEA";
  my $taxon = "taxon:37546";
  my $db_object_type = "transcript";
  my $date = qx/date +%Y-%m-%d/;
  my $db_ref = "PubMed:unpublished";

  open FILE, ">$assocFileName" or die "can't open file, assocFileName!\n";
  
  foreach my $seq (@keys) {
    my @GOs = ();
    if (defined $results_estwise{$seq}) {
      my @results_estwise = @{$results_estwise{$seq}};
      foreach my $estwise_hit (@results_estwise) {
	my @go_accs = $estwise_hit->getGOAccessions;
	push (@GOs, @go_accs);
      }
    }

    if ($debug) {
      print STDERR "GO IDs: " . Dumper (@GOs) . "\n";
    }
    
    foreach my $go_ref (@GOs) {
      my %go = %$go_ref;
      # my $go_acc = $go{go_acc};
      print FILE "$db\t$seq\t$seq\t$go{go_acc}\t$db_ref\t$evidence\t$seq\t$db_object_type\t$taxon\t$date\n";
    }
  }
  close FILE;
}

####################################
# Generate Estwise HTML page
##

sub generateEstwiseHTMLOutput {
  my ($html_estwise_name, $results_estwise_ref) = @_;
  my %results_estwise = %{$results_estwise_ref};

  print STDERR "generating estwise html output ...\n";

  open HTML_ESTWISE, "> $html_estwise_name" or die "can't open file, $html_estwise_name\n";

  print HTML_ESTWISE "<html><head></head><body>\n<center><h2>$title</h2></center>\n";
  # print HTML_ESTWISE "\n<a target=\"new\" href=\"$sequences_url\">ESTs sequences</a><br><br>";
  print HTML_ESTWISE "<table border><tr><th colspan=6>Pfam predictions using Estwise</th></tr><tr><th>EST</th><th>Strand</th><th>Pfam Name</th><th>Bits</th><th>INTERPRO</th><th>GO</th></tr>\n";

  my @keys_estwise = keys (%results_estwise);

  print STDOUT "\t* " . @keys_estwise . " sequences have at least one Pfam domain prediction.\n";

  @keys_estwise = sort_per_key (@keys_estwise);

  #########
  # Estwise html generation
  ##

  my $color = "silver";

  foreach my $key (@keys_estwise) {
    if ($debug) {
      print STDERR "current key: $key\n";
    }

    my $est_acc = $key;
    my $i = 0;

    my @results_estwise = ();
    if (defined $results_estwise{$key}) {
      @results_estwise = @{$results_estwise{$key}};
    }

    while ($i<@results_estwise) {
      
      print HTML_ESTWISE "<tr bgcolor=$color>\n";
      
      if ($debug) {
	print STDERR "#############\narray results_estwise object: " . Dumper (@results_estwise) . "\n#############\n";
      }

      if ($i<@results_estwise) {
	
	my $estwise_hit = $results_estwise[$i];
	
	# my $est_acc   = $estwise_hit->getESTAccession;
	my $strand    = $estwise_hit->getStrand;
	my $pfam_name = $estwise_hit->getPfamName;
	my $pfam_acc  = $estwise_hit->getPfamAccession;
	my $bits      = $estwise_hit->getBits;
	my $ipro_acc  = $estwise_hit->getInterproAccession;
	my $ipro_name = $estwise_hit->getInterproName;
	my @go_accs   = $estwise_hit->getGOAccessions;
	  
	print HTML_ESTWISE "<td><a href=\"$sequences_url/$est_acc.fasta\">$est_acc</a></td>";
	print HTML_ESTWISE "<td align=\"center\">$strand</td>";
	
	if (not $pfam_acc =~ /undefined/i) {
	  print HTML_ESTWISE "<td><a href=\"$_pfam_url$pfam_acc\">$pfam_name</a></td>";
	}
	else {
	  print HTML_ESTWISE "<td>$pfam_name</td>";
	}
	print HTML_ESTWISE "<td>$bits</td>";
	if (not $ipro_acc =~ /undefined/i) {
	  print HTML_ESTWISE "<td><a href=\"$_ipro_url$ipro_acc\">$ipro_name</a></td>";
	}
	else {
	  print HTML_ESTWISE "<td>$ipro_name</td>";
	}
	
	if (@go_accs == 0) {
	  print HTML_ESTWISE "<td>\"undefined\"</td>"; 
	}
	else {
	  if ($debug) {
	    print STDERR "go list dump: " . Dumper (@go_accs) . "\n";
	  }

	  my $go_ref = pop (@go_accs);
	  my %go = %$go_ref;
	  print HTML_ESTWISE "<td><a href=\"$_go_url$go{go_acc}$_go_url_2$go{go_acc}\">$go{go_term}</a>";
	  while (my $go_ref = pop (@go_accs)) {
	    my %go = %$go_ref;
	    print HTML_ESTWISE ", <a href=\"$_go_url$go{go_acc}$_go_url_2$go{go_acc}\">$go{go_term}</a>";
	  }
	  print HTML_ESTWISE "</td>";
	}
      }
      $i++;
      print HTML_ESTWISE "</tr>\n";
    }
    if ($color =~ /silver/ ) {
      $color = "white";
    }
    else {
      $color = "silver";
    }
  } # end foreach key
  print HTML_ESTWISE "</table>\n</body></html>\n";
  
  close HTML_ESTWISE;  
}


sub sort_per_key {
  my (@keys) = @_;
  my @singlet_keys = ();
  my @indexes = ();

  foreach my $key (@keys) {
    if ($debug) {
      print STDERR "key to sort: $key\n";
    }
    # it is a contig
    if ($key =~ /CONTIG(\d+)/i) {
      my $index = $1;
      push (@indexes, $index);
    }
    # it is a singlet
    else {
      push (@singlet_keys, $key);
    }
  }

  if ($debug) {
    print STDERR "indexes before sorting: @indexes\n";
  }

  @indexes = sort { $a <=> $b } @indexes;

  if ($debug) {
    print STDERR "indexes after sorting: @indexes\n";
  }

  my $i = 0;
  while ($i<@indexes) {
    my $index = $indexes[$i];
    $indexes[$i] = "CONTIG$index";
    $i++;
  }
  @singlet_keys = sort (@singlet_keys);
  push (@indexes, @singlet_keys);
  return @indexes;
}
