#!/usr/local/bin/perl -w

use strict;
use Getopt::Std;

# $Header: /nfs/pathsoft/repository/psu/genexec/perl/src/exblx_to_tab.pl,v 1.7 2003/10/02 14:47:51 kmr Exp $

# purpose: This script will convert the output from MSPcrunch -x into tab file
# purpose: format.  Hits to the same protein will be collected together with a
# purpose: join ().

my $min_score = -999;

sub usage
{
  die <<EOF;
usage: $0 -h (for help)
       $0 [-m min_score] [-k output_key] [-d max_intron_size]
EOF
}

if (!@ARGV || ($ARGV[0] eq "-h")) {
  usage;
}

my %options = ();

getopt ("m:k:d:", \%options);

if (exists $options{'m'}) {
  $min_score = $options{'m'};

  if ($min_score !~ /^[\d\.]+$/) {
    die "$0: the -m option need a numeric argument\n";
  }
}

# arbitrarily large by default
my $max_intron_size = 99999999;

if (exists $options{'d'}) {
  $max_intron_size = $options{'d'};

  if ($max_intron_size !~ /^[\d\.]+$/) {
    die "$0: the -d option need a numeric argument\n";
  }
}

my $output_key = "BLASTCDS";

if (exists $options{k}) {
  $output_key = $options{k};
}

sub print_group
{
  my ($note_field, $complement_flag, @hit_array) = @_;

  if (@hit_array == 0) {
    return;
  }

  # sort the hit into left-right order
  @hit_array = sort {
    $a->{start} <=> $b->{start}
  } @hit_array;

  my $max_score = -99999;

  if (
      (grep {
        if ($_->{score} > $max_score) {
          $max_score = $_->{score};
        }
        $_->{score} > $min_score
      } @hit_array) == 0) {
    print STDERR "no hits > $min_score for:\n  $note_field\n";
    return;
  }

#  my $prev_last_base;

  my @orig_hit_array = @hit_array;

  while (@hit_array) {
    my @local_hits = ();
    my @location_bits = ();

    while (@hit_array) {
      my $hit = shift @hit_array;

      push @local_hits, $hit;

      push @location_bits, $hit->{start} . ".." . $hit->{stop};

      if (!@hit_array ||
          $hit_array[0]{start} - $hit->{stop} + 1 > $max_intron_size) {
        last;
      }
    }

    my $location = "";

    # new feature
    if ($complement_flag) {
      $location .= "complement(";
    }

    if (@local_hits > 1) {
      $location .= "join(";
    }

    $location .= join ",", @location_bits;

    if (@local_hits > 1) {
      $location .= ")";
    }

    if ($complement_flag) {
      $location .= ")";
    }

    my $small_score = int($max_score/10);

    $note_field =~ s/\"/""/g;

    printf "FT   %-15.15s $location\n", $output_key;
    print "FT                   /note=\"max score = $max_score  $note_field\"\n";
    print "FT                   /score=$small_score\n";

    my $colour_index = $small_score;

    if ($colour_index > 99) {
      $colour_index = 99;
    }

    my $colour = (17,16,15,2)[$colour_index / 25];

    print "FT                   /colour=$colour\n";

    for my $hit (@local_hits) {
      my $hit_score = $hit->{score};
      my $start = $hit->{start};
      my $end = $hit->{stop};

      my $protein_start = $hit->{protein_start};
      my $protein_stop = $hit->{protein_stop};

      if ($complement_flag) {
        my $note = "score = $hit_score  protein: $protein_start..$protein_stop  at complement($start..$end)";
        print "FT                   /note=\"$note\"\n";
      } else {
        my $note = "score = $hit_score  protein: $protein_start..$protein_stop  at $start..$end";
        print "FT                   /note=\"$note\"\n";
      }
    }
  }
}

sub print_hits
{
  my ($note_field, @hit_array) = @_;

  my @forward_strand_hits = grep { ! $_->{complement_flag} } @hit_array;
  my @reverse_strand_hits = grep { $_->{complement_flag} }   @hit_array;

  print_group $note_field, 1, @reverse_strand_hits;
  print_group $note_field, 0, @forward_strand_hits;
}

my %hit_hash = ();

while (<>) {

  my ($score, $frame, $start, $stop, $protein_start, $protein_stop, $note);

  if (/(\d+)\s+\((.)\S+\)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(.*)/ ||
     /(\d+)\s+([\d\.]+)\s+(\d+)\s+(\d+)\s+(\S+)\s+(\d+)\s+(\d+)\s+(.*)/) {
    if ($1 eq "-" || $2 eq "+") {
      ($score, $frame, $start, $stop, $protein_start, $protein_stop, $note) =
        ($1, $2, $3, $4, $5, $6, $7);
    } else {
      ($score, $start, $stop, $protein_start, $protein_stop, $note) =
        ($1, $3, $4, $6, $7, $8);
      if ($start > $stop) {
        $frame = "-";
        ($start, $stop) = ($stop, $start);
      } else {
        $frame = "+";
      }
    }

    my $complement_flag = 0;

    if ($frame eq '-') {
      $complement_flag = 1;
    } else {
      $complement_flag = 0;
    }

    if (! exists $hit_hash{$note}) {
      $hit_hash{$note} = [];
    }

    my $array_ref = $hit_hash{$note};

    push @$array_ref, {
      score => $score,
      complement_flag => $complement_flag,
      start => $start,
      stop => $stop,
      protein_start => $protein_start,
      protein_stop => $protein_stop
    };
  } else {
    die "error reading while reading MSPcrunch output:\n   $_\n";
  }
}

for my $hash_key (keys %hit_hash) {
  my @hit_array = @{$hit_hash{$hash_key}};
  print_hits $hash_key, @hit_array;
}
