#!/usr/local/bin/perl --  # -*-Perl-*-

# $Header: /nfs/pathsoft/repository/psu/genexec/perl/src/exon_base_count.pl,v 1.4 2003/05/19 13:05:51 kdj Exp $

# purpose: This script counts gc bases and all bases of CDS features and for
# purpose: the whole sequence.  It prints a base count for each CDS feature,
# purpose: summary for each input file and a summary of the all the input
# purpose: files.  The -exons flag will cause the base count/gc count for
# purpose: each exon to be printed rather than the totals for each CDS.

use strict;

use Embl;
use EMBL::Parser;

print STDERR "#" x 70, "\n";
print STDERR "This script has not been updated to process files which
use (temporary) systematic_ids and may not give accurate results on
such data. If you need this feature please contact a programmer as it
will be more reliable and efficient to create a completely new script.\n";
print STDERR "#" x 70, "\n";

sleep(2);

my $show_exons = 0;

if ($ARGV[0] eq '-exons') {
  $show_exons = 1;
  shift;
}

my $print_duplicates = 0;

if ($ARGV[0] eq '-dup') {
  $print_duplicates = 1;
  shift;
}

if (@ARGV == 0) {
  die <<END;
usage: $0 [-exons] embl_file_name another_embl_file ...

options:

   -exons will cause the base count for each exon to be printed rather
          than the totals for each CDS
   
   -dup   will print out duplicate gene names 
END
}

open ERRORS, ">>exon_base_count_errors";

my @submission_files = @ARGV;

my %all_gene_names = ();

# my @submission_files = glob "chr1/c24C9*/c24*_submission";
{
  my $total_cds_bases = 0;
  my $total_cds_gc_bases = 0;
  my $total_bases = 0;
  my $total_gc_bases = 0;
  my $total_cds_count = 0;

  for my $submission_file (@submission_files) {
    print "file: $submission_file\n";

    my ($cds_bases, $cds_gc_count, $bases, $gc_count, $cds_count) =
      count_bases_and_gc ($submission_file);
    $total_cds_bases += $cds_bases;
    $total_cds_gc_bases += $cds_gc_count;
    $total_bases += $bases;
    $total_gc_bases += $gc_count;
    $total_cds_count += $cds_count;
  }

  print 
    "total CDS bases: $total_cds_bases,  CDS GC bases $total_cds_gc_bases,  " .
    "percent CDS gc: ", 100 * $total_cds_gc_bases/$total_cds_bases, "\n" .
    "total bases: $total_bases,  total gc bases: $total_gc_bases,  " .
    "percent gc: ", 100 * $total_gc_bases/$total_bases, "\n";
  print "total number of CDS features:  $total_cds_count\n";
}

sub count_bases_and_gc
{
  my $cds_bases = 0;
  my $cds_gc_bases = 0;
  my $entry_bases = 0;
  my $entry_gc_bases = 0;
  my $cds_count = 0;

  my $file_name = $_[0];

  open IN_FILE, "$file_name" || die "$0: can't open $file_name\n";

  # read in some embl files from a stream
  my $set = Read Embl(\*IN_FILE); # read all of the embl entries from that stream.

  # get an array of all of the EMBL::Entry objects within $set
  my @entries = $set->entries();
  while (my $entry = pop @entries) {
    my $ft = $entry->FT ();

    my @features = $ft->features ();

    my $sequence = $entry->SQ ();

    my $length = $sequence->length ();

    $entry_bases += $length;

    my @entry_gc_bases = $sequence->seq =~ m/(g|c)/g;

    $entry_gc_bases += @entry_gc_bases;

    my $gene_counter = 1;

    #  print "sequence length = $length\n";

    # loop over the features in the current entry
  FEATURE: for my $f (@features) {
      my %feature_hash = %$f;

      my $this_cds_bases = 0;

      if ($feature_hash{key} ne "CDS") {
        # go straight to the next feature - we are only interested in
        # CDS features
        next;
      }

      $cds_count++;

      #    print "feature: $$f{key}   $$f{location}\n";

      my $location = $feature_hash{location};
      my $orig_location = $feature_hash{location};

      my $complement_flag = 0;

      if ($location =~ s/complement\(([^\)]+)\)/$1/g) {
        $complement_flag = 1;
      }

      $location =~ s/join\(([^\)]*)\)/$1/g;
      
      # this is a spliced gene so write out the exons

      my @join_bits = split (',',$location);

      if ($complement_flag) {
        @join_bits = reverse @join_bits;
      }

      my $exon_id = "a";

      # write out each exon

      my @gene_names = ();

      for my $qual (@{$feature_hash{qualifierList}}) {
        if ($qual->{name} eq "gene") {
          my $new_gene_name = $qual->{value};

          $new_gene_name =~ s/"(.*)"/$1/;

          push @gene_names, $new_gene_name;

          if (exists $all_gene_names{$new_gene_name} && $print_duplicates) {
            if ($file_name eq $all_gene_names{$new_gene_name}) {
              print "duplicated gene name: $new_gene_name in $file_name\n";
            } else  {
              print "duplicated gene name: $new_gene_name in $file_name " .
              "(and $all_gene_names{$new_gene_name})\n";
            }
          } else {
            $all_gene_names{$new_gene_name} = $file_name;
          }
        }
      }

      if (!@gene_names) {
        print ERRORS "CDS with no /gene qualifier in input file $file_name\n";
      }

      if (!@gene_names) {
        # use the accession number instead
              
        my $id = $entry->ID ();

        @gene_names = "*$id.$gene_counter";
      }

      if (@join_bits == 1) {
        $exon_id = "";
      }

      for (my $i = 0 ; $i < @join_bits ; ++$i) {
        my $exon_range = $join_bits[$i];

        my $start_of_exon;
        my $end_of_exon;

        if ($exon_range =~
            m/<?(\d+)
            (?:\.\.
             >?
             (\d+))?/x) {

          $start_of_exon = $1;
          if (defined $2) {
            $end_of_exon = $2;
          } else {
            $end_of_exon = $1;
          }

          
          my @exon_names = 
            map {
              $_ . (scalar (@join_bits) == 1 ? "" : "_") . $exon_id
            } @gene_names;

          my $exon_length = $end_of_exon - $start_of_exon + 1;

          $cds_bases += $exon_length;

          $this_cds_bases += $exon_length;

          if ($show_exons) {
            print ">@exon_names  ", $exon_length, "\n";
          }
          #                  "  $start_of_exon $end_of_exon\n";

          my $subseq_str = lc $sequence->subseq ($start_of_exon - 1,
                                                 $end_of_exon);

          # count the gc content
          my @gc_bases = $subseq_str =~ m/(g|c)/g;

          $cds_gc_bases += @gc_bases;

        } else {
          # the right hand side exon doesn't look like a range
          warn "Failed to understand this location: $orig_location";
          next;
        }

        ++$exon_id;
      }

      if (!$show_exons) {
        print ">@gene_names  ", $this_cds_bases, "\n";
      }

      #  print "$feature_hash{gene}\n";

      ++$gene_counter;
    }
  }

  if ($cds_bases > 0) {
    print "$file_name:  CDS bases: $cds_bases,  GC bases $cds_gc_bases  " .
          "percent gc: ", 100 * $cds_gc_bases/$cds_bases, "\n";
  } else {
    "$file_name: no CDS features\n";
  }

  if ($entry_bases > 0) {

    print "$file_name:  bases: $entry_bases,  GC bases $entry_gc_bases  " .
          "percent gc: ", 100 * $entry_gc_bases/$entry_bases, "\n";
  } else {
    "$file_name: no bases\n";
  }

  print "$file_name:  number of CDS features:  $cds_count\n";

  return ($cds_bases, $cds_gc_bases, $entry_bases, $entry_gc_bases, 
          $cds_count);
}
