#!/usr/local/bin/perl5.6.1 -w

use strict;
use File::Temp qw/tempfile/;
use Getopt::Std;
use Benchmark;

# GFF Exonerate parser
use Bio::PSU::Utils::ExonerateGFF;

# PSU modules for EMBL writing
use Bio::PSU::SeqFactory;
use Bio::PSU::Seq;

=head1 NAME

exonerate2art.pl 

=head1 SYNOPSIS

  exonerate2art.pl [-r] [-c <cDNA input>] [-g <gDNA input>] [-i <exonerate results file>] [-o <output>] [-t <sequence type>] <exonerate options>
    
      \t-r mask the genomic sequence
      \t-o Output EMBL file         (default is standard output)
      \t-t Sequence Type (EST, Unigene, cDNA cluster), that will be reported in a note field


=head1 DESCRIPTION

  This script can run RepeatMasker and Exonerate
  And then it converts the GFF Exonerate output into an EMBL file.
  The default output is Standard Output (STDOUT)
  The output file is generated in such a way that the results can be displayed into the sequence viewer software, Artemis.

  @author Arnaud Kerhornou - axk@sanger.ac.uk

=cut

my $t1 = Benchmark->new ();
my $debug = 0;

# Exonerate
my $exoneratebindir    = "/nfs/pathsoft/external/bio-soft/exonerate";
my $exonerate_bin   = "$exoneratebindir/exonerate";
my $exonerate_args  = "--model est2genome --dnahspthreshold 75 --score 500 --fsmmemory 1024 --showtargetgff --softmasktarget --saturatethreshold 10 -V 1 --forcescan target --maxintron 1000";

# RepeatMasker
my $rmbindir   = "/usr/local/pubseq/scripts";
my $rm_bin     = "$rmbindir/RepeatMasker";
my $maskingdir = "masking";
my $rm_args    = "-dir $maskingdir";

# Dust
my $dustbindir = "/usr/local/pubseq/bin";
my $dust_bin   = "$dustbindir/dust";
my $dust_args  = "";

my (%opts, $cDNA_input, $gDNA_input, $exonerate_input, $output, $mask, $type);
my $usage = "Usage: exonerate2art.pl [-r] [-c <cDNA input>] [-g <gDNA input>] [-i <exonerate results file>] [-o <output>] <exonerate options>
\t-r mask the genomic sequence
\t-o Output EMBL file         (default is standard output)
\t-t Sequence Type (EST, Unigene, cDNA cluster)
";

# Process options
getopts('h:rc:g:i:o:t:', \%opts);

defined $opts{h} and die $usage;
defined $opts{r} and $mask++;
defined $opts{c} and $cDNA_input = $opts{c};
defined $opts{g} and $gDNA_input = $opts{g};
defined $opts{i} and $exonerate_input = $opts{i}; 
defined $opts{o} and $output = $opts{o}; 
defined $opts{t} and $type   = $opts{t};

unless ($exonerate_input || ($cDNA_input && $gDNA_input)) { 
  print STDERR "Specify either an exonerate output file or a cDNA database and a sequence file\n";
  die $usage; 
}

unless ($type) {
  print STDERR "Specify a sequence type (EST, Unigene, cDNA cluster)!\n";
  die $usage;
}

# If exonerate file given, no need to run exonerate

if (defined $exonerate_input) {
  if ($debug) {
    print STDERR "given exonerate file is directly processed\n";
  }
}
else {
  my $gDNA_inputForExonerateRun = $gDNA_input;
  if ($mask) {

    # RepeatMasking
    qx/rm -rf $maskingdir; mkdir $maskingdir/;
    qx/$rm_bin $rm_args $gDNA_input/;
    $gDNA_inputForExonerateRun = "$maskingdir/$gDNA_input.masked";
 
    # dust
    qx/$dust_bin $dust_args $gDNA_inputForExonerateRun > $gDNA_inputForExonerateRun.dust/;
    $gDNA_inputForExonerateRun .= ".dust";
  }

  # Run exonerate

  my ($fh, $tempExonerateOutput) = tempfile ("Exonerate.XXXXXX", UNLINK => 1);
  close ($fh);

  qx/$exonerate_bin --target $gDNA_inputForExonerateRun --query $cDNA_input $exonerate_args > $tempExonerateOutput/;

  # Clean Exonerate Output 
  
  my $tempExonerateOutputGFF = "$tempExonerateOutput.gff";

  if (! -f $tempExonerateOutput) {
    print STDERR "ERROR - can't file exonerate output file, $tempExonerateOutputGFF!!\n";
  }

  qx/filter_exonerate_output.pl $tempExonerateOutput > $tempExonerateOutputGFF/;
  $exonerate_input = $tempExonerateOutputGFF;
}

##################################
#
# Parsing Exonerate output
#
##################################

if ($debug) {
  print STDERR "instanciating ExonerateGFF object for parsing file, $exonerate_input...\n";
}

my $exonerate_parser = Bio::PSU::Utils::ExonerateGFF->new (
							   file => $exonerate_input,
							   debug => $debug,
							   type  => $type,
							  );
my %featuresPerSequenceID;
while (my ($sequenceid, $gene_features) = $exonerate_parser->next_gene_prediction ()) {
  
  if (defined $sequenceid) {
    if (exists $featuresPerSequenceID{$sequenceid}) {
      my $features = $featuresPerSequenceID{$sequenceid};
      push (@$features, @$gene_features);
      $featuresPerSequenceID{$sequenceid} = $features;
    }
    else {
      $featuresPerSequenceID{$sequenceid} = $gene_features;
    }
  }
  else {
    last;
  }
}

if ($debug) {
  print STDERR "Parsing done!\n";
}

########################
#
# Writing Exonerate Features into a tab file...
#
########################

my $tab_obj;
if (defined $output) {
  $tab_obj = Bio::PSU::SeqFactory->make(-file   => ">$output",
					-format => 'embl');
}
else {
  # STDOUT
  $tab_obj = Bio::PSU::SeqFactory->make(-file   => \*STDOUT,
					-format => 'embl');
}

my @sequenceids = keys (%featuresPerSequenceID);
foreach my $sequenceid (@sequenceids) {
  # my $output = "$sequenceid.exonerate.tab";

  if ($debug) {
    print STDERR "Writing Features into a new file, $output...\n";
  }

  my $seq_obj = Bio::PSU::Seq->new (
				    -id => $sequenceid,
				    -type => "dna",
				    -desc => "Exonerate predictions"
				   );
  my $features = $featuresPerSequenceID{$sequenceid};
  $seq_obj->features (@$features);
  
  # copy this seq obj into a new file
  
  $tab_obj->write_seq($seq_obj);

  if ($debug) {
    print STDERR "file, $output, done\n";
  }

}

if ($debug) {
  print STDERR "process complete\n";
}

my $t2 = Benchmark->new ();
print STDERR "\nTotal : ", timestr (timediff ($t2, $t1)), "\n";

#############################################
# The End
#############################################
