#!/usr/local/bin/perl5.6.1 -w

use strict;
use Getopt::Std;

use Bio::PSU::Range;
use Bio::PSU::Seq;
use Bio::PSU::Feature;
use Bio::PSU::IO::EMBL::Stream;
use Bio::PSU::IO::BufferFH;

my $debug = 0;

if ((@ARGV+0) == 0) {
  print STDERR "specify the name of a geneid output file\n";
  exit 0;
}

my (%opts, $cdsFile, $otherFeatFile);
getopts('c:o', \%opts);
defined $opts{c} and $cdsFile = $opts{c};
defined $opts{o} and $otherFeatFile = $opts{o};

my $gffFile = $ARGV[0];
my @tmp = split ('\.', $gffFile);
$cdsFile       ||= $tmp[0] . ".geneid.cds.tab";
$otherFeatFile ||= $tmp[0] . ".geneid.rest.tab";
open GENEID,  "<$gffFile" or die "can't open file, $gffFile\n";
open CDS,     ">$cdsFile" or die "can't open file, $cdsFile\n";

my $bfh      = Bio::PSU::IO::BufferFH->new(-fh => \*CDS);
my $stream   = Bio::PSU::IO::EMBL::Stream->new(-bfh => $bfh);

my @features = ();
my $seqobj;
my $seqName  = undef;

while (<GENEID>) {
  next if /^##/;
  my $line = $_;
  
  if ($debug) {
    print STDERR "new line : $line";
  }

  # extracting the number of exons for the current gene

  if ($line =~ /# Gene \d+\s*\(\w+\)\. (\d+) exons\..+/) {

    my $nb_exons = $1;
    my $i = 1;
      
    if ($debug) {
       print STDERR "new gene of $nb_exons exons found\n";
    }

    # CDS ranges
    my @ranges = ();
    # exon scores
    my @scores = ();
    my $cds_score   = 0;
    my $codon_start = undef;

    # build the cds feature by joining the exon together

    while ($i <= $nb_exons) {
      $line = <GENEID>;
      # remove the white spaces just keep the tabulations
      $line =~ s/ //g;

      if ($debug) {
	print STDERR "feature line : $line\n";
      }

      $line =~ /^([^\t]+)\t+(\S+)\t+(\w+)\t+(\d+)\t+(\d+)\t+(\S+)\t+(\+|-)\t+(\d+)\t+.+/;
      my $seqName_tmp = $1;
      my $nameFeature = $3;
      my $start       = $4;
      my $end         = $5;
      my $score       = $6;
      my $strand      = 1;
      if ($7 =~ /-/) {
        $strand = -1;
      }
      my $frame       = $8;
      # parsing problem when the strand is "-", can not get the frame in that case !!!
      # so substitute - by + to make the parsing work
      if (not defined $frame) {
        $line =~ s/-/\+/g;
        $line =~ /(\w+)\t+(\S+)\t+(\w+)\t+(\d+)\t+(\d+)\t+(\S+)\t+(\+|-)\t+(\d+)\t+.+/;
        $frame = $8;
      }
      # specify the frame for the first exon, which might not start with a MET
      if (($strand==1 && $i==1) || ($strand==-1 && $i==$nb_exons)) {
        if ($frame =~ /0/) {
          $codon_start = 1;
        }
        elsif ($frame =~ /1/) {
          $codon_start = 2;
        }
        elsif ($frame =~ /2/) {
          $codon_start = 3;
        }
      }

      if (not defined $seqName) {
	$seqName = $seqName_tmp;
	$seqobj = Bio::PSU::Seq->new (-type => 'dna',
				      -id => $seqName);
      }
      elsif ($seqName ne $seqName_tmp) {
	$seqobj->features ([@features]);
	@features = ();
	$stream->write_seq ($seqobj);
	$seqName = $seqName_tmp;
	$seqobj = Bio::PSU::Seq->new (-type => 'dna',
				      -id => $seqName);
      }

      my $range = Bio::PSU::Range->new(
				       -start       => $start,
                                       -end         => $end,
                                       -strand      => $strand
				      );
      push (@ranges, $range);
      push (@scores, $score);
      $cds_score = $cds_score + $score;

      $i++;
    }

    # build the feature object

    my $feature = Bio::PSU::Feature->new(-key    => 'CDS',
                                         -ranges => \@ranges);
    my $note = "GeneID - exon scores: " . join (', ', @scores);
    $feature->qadd('note', $note);
    $feature->qadd('method', 'GENEID');
    $feature->qadd('colour', '3');
    $feature->qadd('score', $cds_score);
    # $feature->qadd('note', 'exon scores, associated with the location ranges - always from left to right');
    $feature->qadd('codon_start', $codon_start);

    push (@features, $feature);
  }
  else {
    if ($debug) {
      print STDERR "next execution shouldn't occur\n";
    }
  }
}

$seqobj->features ([@features]);
$stream->write_seq($seqobj);
