#!/usr/local/bin/perl5.6.1 -w

use strict;
use Getopt::Std;

use Bio::PSU::SeqFactory;

my $_debug = 0;

my (%opts);
my $prompt = 0;
my ($input, $output, $digits) = undef;
my $id_qualifier = "systematic_id";
getopts('pi:o:d:q:', \%opts);
defined $opts{i} and $input   = $opts{i};
defined $opts{o} and $output  = $opts{o};
defined $opts{d} and $digits  = $opts{d};
defined $opts{q} and $id_qualifier = $opts{q};
defined $opts{p} and $prompt++;

if (not (defined $input && defined $digits)) {
  die "Usage :

This script is for generating a set of unique systematic identifiers for CDS, rRNA and tRNA features...

generateSytematicId.pl -i [tab input file] [-o [tab output file]] -d [Number of digits the index is made of] [-p [prompt the Contig Name] -q [systematic id qualifier]]
    -q systematic id qualifier, e.g. temporary_systematic_id or systematic_id - Default is systematic_id
";
}

my $sin   = Bio::PSU::SeqFactory->make(
				       -file   => "$input",
				      );
my $sout;
if (defined $output) {
  $sout = Bio::PSU::SeqFactory->make(
				     -file   => ">$output",
				     -format => 'embl'
				    );
}
else {
  $sout = Bio::PSU::SeqFactory->make(
				     -fh   => \*STDOUT,
				     -format => 'embl',
				    );
}

#########################
#
# Index Generation
#
#########################

# Tab file processing

if ($_debug) {
  print STDERR "Tab file processing...\n";
}

while (my $seqobj = $sin->next_seq ()) {

  my $index = 1;

  my $seqName = $seqobj->id;
  my $prefix = "$seqName" . "_";

  if ($prompt) {
    print STDERR "contig name: $seqName\n";
    print STDERR "Systematic Id prefix: \n";
    $prefix = readline (*STDIN);
    chomp $prefix;
  }

  print STDERR "\nprefix: $prefix\n\n";

  my @new_features = ();
  foreach my $feature ($seqobj->features) {

    if ($feature->key =~ /^CDS$|rRNA|tRNA|rDNA_Palindrome/) {

      if ($_debug) {
	print STDERR "Feature key, " . $feature->key . " is having a systematic id...\n";
      }

      my $systematic_id = set_systematic_id ($prefix, $index, $digits);
      my $newfeature = $feature->clone ();
      $newfeature->qremove ('gene');
      $newfeature->qremove ('systematic_id');
      $newfeature->qremove ('temporary_systematic_id');
      $newfeature->qadd ($id_qualifier, $systematic_id);
      push (@new_features, $newfeature);
      $index++;
    }
    else {
      my $newfeature = $feature->clone ();
      push (@new_features, $newfeature);
    }
  }
  my $newseqobj = $seqobj->clone (-features => [@new_features]);
  $sout->write_seq ($newseqobj);
}

if ($_debug) {
  print STDERR "tab file processing done!\n";
}

# Done

sub set_systematic_id {
  my ($prefix, $index, $digits) = @_;

  if (($digits - length ($index)) < 0) {
    print STDERR "ERROR - Can't make any index, the number of digits you have given is lower than the number of genes to index !!!\n";
    exit 1;
  }
  elsif (length ($index) == $digits) {
    # that's fine !
  }
  else {
    my $predigits = $digits - length ($index);
    $index = doIndex ($predigits, $index);
  }

  my $systematic_id = "$prefix$index";

  return $systematic_id;
}

sub doIndex {
  my ($predigits, $index) = @_;

  for (my $i=0; $i<$predigits; $i++) {
    $index = "0" . $index;
  }

  return $index;
}
