#!/usr/local/bin/perl5.6.1 -w

=head1 NAME

glimmermx2tab

=head1 SYNOPSIS

 glimmermx2tab < file

=head1 DESCRIPTION

Converts the output of GlimmerExon into an EMBL feature table.

=head1 AUTHOR

Keith James (kdj@sanger.ac.uk)

=cut

use strict;
use Getopt::Std;
use Bio::PSU::Feature;
use Bio::PSU::Range;
use Bio::PSU::IO::FTHandler;

my %opts;
my $usage = "Usage: glimmermx2tab < file\n";

getopts('h', \%opts);

defined $opts{h} and die $usage;

my $fth = Bio::PSU::IO::FTHandler->new;

my $genes = {};
my $seq_name = "unknown";

while (<>)
{
    if (/^Sequence name: (\S+)/)
    {
        $seq_name = $1;
        next;
    }

    if (/^\s+(\d+)\s+(\d+)\s+([+-])\s+(\S+)\s+(\d+)\s+(\d+)\s+\d+\s+(\d+)/)
    {
        add_exon($genes, $1, $2, $3, $4, $5, $6, $7);
    }
}

foreach my $gene_id (sort keys %$genes)
{
    # Exon types are Single, Initial and Terminal
    my %exon_types;
    my $exon_frame = 0;
    my $truncated = 0;
    my @ranges;

    foreach my $exon_id (sort keys %{$genes->{$gene_id}})
    {
        my ($exon_type, $exon_start, $exon_end, $strand, $frame) =
            @{%{$genes->{$gene_id}}->{$exon_id}};

        $exon_types{$exon_type}++;

        if (($exon_id == 1) &&
            ($exon_type eq "Internal" || $exon_type eq "Terminal"))
        {
            $exon_frame = $frame;
        }

        my $exon_strand = $strand eq '+' ? 1 : -1;
        my $range = Bio::PSU::Range->new(-start  => $exon_start,
                                         -end    => $exon_end,
                                         -strand => $exon_strand);

        push(@ranges, $range);
    }

    my $feature = Bio::PSU::Feature->new(-key    => 'CDS',
                                         -ranges => \@ranges);

    $feature->qadd('temporary_systematic_id', "$seq_name.$gene_id");

    if ($exon_frame)
    {
        $feature->qadd('codon_start', $exon_frame);
    }

    print $fth->make_feature_block($feature);
}

sub add_exon
{
    my ($genes, $gene_id, $exon_id, $strand,
        $exon_type, $exon_start, $exon_end, $frame) = @_;

    if (exists $genes->{$gene_id})
    {
        $genes->{$gene_id}->{$exon_id} = [$exon_type, $exon_start, $exon_end,
                                          $strand, $frame];
    }
    else
    {
        $genes->{$gene_id} = {};
        $genes->{$gene_id}->{$exon_id} = [$exon_type, $exon_start, $exon_end,
                                          $strand, $frame];
    }
}
