#!/usr/local/bin/perl5.6.0 -w

##########################################################################
#
# This script converts HMMgene output into an annotation file file, in Embl format for display into Artemis
#
# @Param  The HMMgene Output in GFF format
# @Param  The fasta sequence file, needed to resolve the ORF when the first exon is missing
# @Return A Tab file, with the suffix '.cds.tab'
#
##########################################################################

use strict;
use Getopt::Std;

use Data::Dumper;

# PSU modules

use Bio::PSU::Range;
use Bio::PSU::Seq;
use Bio::PSU::Feature;
use Bio::PSU::IO::EMBL::Stream;
use Bio::PSU::IO::BufferFH;
use Bio::PSU::SeqFactory;

#################
# Debugging mode
##

my $debug = 1;

if ((@ARGV+0) == 0) {
  print STDERR "specify the name of a hmmgene output file and the name of the fasta sequence file\n";
  exit 0;
}

my (%opts, $cdsFile, $otherFeatFile);
getopts('c:o:', \%opts);
defined $opts{c} and $cdsFile = $opts{c};
defined $opts{o} and $otherFeatFile = $opts{o};

my $gffFile       = $ARGV[0];
my $seqFile       = $ARGV[1];
my @tmp           = split ('\.', $gffFile);
$cdsFile       ||= $tmp[0] . ".hmmgene.cds.tab";
$otherFeatFile ||= $tmp[0] . ".hmmgene.rest.tab";

open HMMGENE, "<$gffFile"       or die "can't open file, $gffFile\n";
open CDS,     ">$cdsFile"       or die "can't open file, $cdsFile\n";

my @psu_features = ();
my $seq      = Bio::PSU::Seq->new (-type => 'dna');

my $seqi     = Bio::PSU::SeqFactory->make(-file => $seqFile,    -format => 'fasta');
my $seq_obj  = $seqi->next_seq;

my $cds_bfh    = Bio::PSU::IO::BufferFH->new(-fh => \*CDS);
my $cds_stream = Bio::PSU::IO::EMBL::Stream->new(-bfh => $cds_bfh);

my $other_bfh    = Bio::PSU::IO::BufferFH->new(-fh => \*CDS);
my $other_stream = Bio::PSU::IO::EMBL::Stream->new(-bfh => $other_bfh);

my @forward_features = ();
my @reverse_features = ();

# -1 => undefined status
#  1 => forward strand features
#  0 => reverse strand features
my $is_forward_feature = -1;

#####################################################
#
# Get The GFF HMMGENE Features into two separate arrays, 
# one for the forward features, one for the reverse features
#
#####################################################

while (<HMMGENE>) {
  next if /^#/;

  my $line  = $_;
  if ($line =~ /\w+\t+\S+\t+SEQUENCE\t+\d+\t+\d+\t+\S+\t+(-|\+)\t+\S+/) {
    my $strand = $1;
    if ($strand eq "+") {
      $is_forward_feature = 1;
    }
    elsif ($strand eq "-") {
      $is_forward_feature = 0;
    }
  }
  elsif ($is_forward_feature == 1) {
    push (@forward_features, $line);
  }
  elsif ($is_forward_feature == 0) {
    push (@reverse_features, $line);
  }
}

close HMMGENE;

#####################################################
#
# Processing the forward features
#
#####################################################

my $gene_feature = undef;
my $cds_feature  = undef;

my $index = 0;
while ($index < @forward_features) {
  my $line = $forward_features[$index];
  $line    =~ /\w+\t+\S+\t+(\w+)\t+\d+\t+\d+\t+\S+\t+(-|\+)\t+\S+/;
  my $feature_type = $1;
  my $strand       = $2;

  if ($debug) {
    print STDERR "feature type: $feature_type\n";
    print STDERR "index before: $index\n";
  }

  if ($feature_type =~ /5utr/i) {
    if ($debug) {
      print STDERR "5'UTR gene - forward\n";
    }
    ($index, $gene_feature, $cds_feature) = parse5UTRGene ($index, @forward_features);
  }
  elsif ($feature_type =~ /coding|intron|3utr/i) {
    if ($debug) {
      print STDERR "Fuzzy gene - forward\n";
    }
    ($index, $gene_feature, $cds_feature) = parseFuzzyGene ($index, $seq_obj, @forward_features);
  }

  if ($debug) {
    print STDERR "index after: $index\n";
  }

  if (defined $gene_feature) {
    push (@psu_features, $gene_feature);
  }
  if (defined $cds_feature) {
    push (@psu_features, $cds_feature);
  }

  $index++;
}

#####################################################
#
# Processing the reverse features
#
#####################################################

$gene_feature = undef;
$cds_feature  = undef;

$index = 0;
while ($index < @reverse_features) {

  my $line = $reverse_features[$index];

  if ($debug) {
    print STDERR "line: $line\n";
  }

  $line =~ /\w+\t+\S+\t+(\w+)\t+\d+\t+\d+\t+\S+\t+(-|\+)\t+\S+/;
  my $feature_type = $1;
  my $strand = $2;

  if ($debug) {
    print STDERR "feature type: $feature_type\n";
    print STDERR "index before: $index\n";
  }

  if ($feature_type =~ /3utr/i) {
    
    if ($debug) {
      print STDERR "3'UTR gene - reverse\n";
    }
    ($index, $gene_feature, $cds_feature) = parse3UTRGene ($index, $seq_obj, @reverse_features);
  }
  elsif ($feature_type =~ /coding|intron|5utr/i) {
    if ($debug) {
      print STDERR "Fuzzy gene - reverse\n";
    }
    ($index, $gene_feature, $cds_feature) = parseFuzzyGene ($index, $seq_obj, @reverse_features);
  }

  if ($debug) {
    print STDERR "index after: $index\n";
  }
  
  if (defined $gene_feature) {
    push (@psu_features, $gene_feature);
  }
  if (defined $cds_feature) {
    push (@psu_features, $cds_feature);
  }

  $index++;
}

# write new embl file

# if there is features, generate a new sequence entry.

if (@psu_features > 0) {
  my $psu_seq_obj = Bio::PSU::Seq->new (
					-id   => $seq_obj->id,
					-desc => $seq_obj->desc,
					-type => 'dna',
					-features => \@psu_features
				       );  
  $cds_stream->write_seq ($psu_seq_obj);
}

######################
# End
##

#####################################################
#
# Gene which starts with a 5'UTR
#
# @param Hmmgene features
#
# @return a PSU CDS feature
# @return a PSU gene feature
# @return the index for next HMMgene gene processing
# 
#####################################################

sub parse5UTRGene {
  my ($index, @forward_features) = @_;

  my $line = $forward_features[$index];
  
  $line =~ /([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+(.+)/;

  my $seqName     = $1;
  my $nameFeature = $3;
  my $start       = $4;
  my $end         = $5;
  my $score       = $6;
  my $strand      = $7;
  my $frame       = $8;

  if (not defined $start) {
    print STDERR "start not defined!!!\n";
  }

  my $gene_start = $start;
  my $total_scores = 0;
  my $nb_exons = 0;
  my @ranges = ();
  my @scores = ();

  $index++;
  while ($index < @forward_features) {
    $line = $forward_features[$index];
    $line =~ /([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+(.+)/;
    my $seqName     = $1;
    my $nameFeature = $3;
    my $start       = $4;
    my $end         = $5;
    my $score       = $6;
    my $strand      = $7;
    my $frame       = $8;

    if ($debug) {
      print STDERR "nameFeature: $nameFeature\n";
    }

    if ($nameFeature =~ /coding/i) {

      # Exon feature

      $nb_exons++;
      my $range = Bio::PSU::Range->new(
				       -start       => $start,
				       -end         => $end,
				       -strand      => 1
				      );
      push (@ranges, $range);
      push (@scores, $score);
      $total_scores = $total_scores + $score;
    }
    elsif ($nameFeature =~ /intron/i) {
      # Intron feature
      # do nothing !!
    }
    elsif ($nameFeature =~ /3utr/i) {
      
      # case when there is no predicted exons !!!

      if ($nb_exons == 0) {
	return $index, undef, undef;
      }

      # 3' UTR
      # create PSU features objects

      if ($debug) {
	print STDERR "Complete 5' Gene\n";
      }

      my $gene_end = $end;
      my $gene_range = Bio::PSU::Range->new(
					    -start  => $gene_start,
					    -end    => $gene_end,
					    -strand => 1
					   );

      if ($debug) {
	print STDERR "start, end: $gene_start, $gene_end\n";
      }

      my $cds_score = $total_scores / $nb_exons unless $nb_exons == 0;
      $cds_score = round ($cds_score, 2);
      my $note = "HMMGene - exon scores: " . join (', ', @scores);
      
      my $gene_feature = Bio::PSU::Feature->new(
						-key    => 'gene',
						-ranges => $gene_range
					       );
      my $cds_feature = Bio::PSU::Feature->new(
					       -key    => 'CDS',
					       -ranges => \@ranges
					      );
      
      $cds_feature->qadd('note', $note);
      $cds_feature->qadd('method', 'HMMGENE');
      $cds_feature->qadd('score', $cds_score);
      $cds_feature->qadd('colour', '6');
      
      $gene_feature->qadd('method', 'HMMGENE');
      $gene_feature->qadd('note', $note);
      
      return ($index, $gene_feature, $cds_feature);
    }
    
    if (($index+1) >= @forward_features) {
      # this is an unfinished sequence
      
      if ($debug) {
	print STDERR "Uncomplete 5' gene, missing the 3'UTR\n";
      }

      # case when there is no predicted exons !!!
      # not reported

      if ($nb_exons == 0) {
	return ($index, undef, undef);
      }
      
      # create PSU features objects
      
      my $gene_end = $end;
      
      if ($debug) {
	print STDERR "start, end: $gene_start, $gene_end\n";
      }
      
      my $gene_range = Bio::PSU::Range->new(
					    -start => $gene_start,
					    -end   => $gene_end,
					    -strand => 1
					   );
      $gene_range->no_3prime(1);
      
      my $cds_score = $total_scores / $nb_exons unless $nb_exons == 0;
      $cds_score = round ($cds_score, 2);
      my $note = "HMMGene - exon scores: " . join (', ', @scores);
      
      my $gene_feature = Bio::PSU::Feature->new(
						-key    => 'gene',
						-ranges => $gene_range
					       );
      my $cds_feature = Bio::PSU::Feature->new(
					       -key    => 'CDS',
					       -ranges => \@ranges
					      );
      
      $cds_feature->qadd('note', $note);
      $cds_feature->qadd('method', 'HMMGENE');
      $cds_feature->qadd('score', $cds_score);
      $cds_feature->qadd('colour', '6');
      $cds_feature->qadd('is_partial');
      
      $gene_feature->qadd('method', 'HMMGENE');
      $gene_feature->qadd('note', $note);
      
      return ($index, $gene_feature, $cds_feature);
    }

    $index++;
  }

  print STDERR "problem, not supposed to come here !!!!!!!!!!!\n";

  return ($index, undef, undef);
}

#####################################################
#
# Gene which doesn't start with a 5'UTR or doesn't finish with a 3'UTR
#
# @param Hmmgene features
#
# @return a PSU CDS feature
# @return a PSU gene feature
# @return the index for next HMMgene gene processing
# 
#####################################################

sub parseFuzzyGene {
  my ($index, $seq_obj, @hmmgene_features) = @_;

  my $codon_start = undef;
  my $total_scores = 0;
  my $nb_exons = 0;
  my @ranges   = ();
  my @scores   = ();

  my $line = $hmmgene_features[$index];
  
  $line =~ /([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+(.+)/;

  my $seqName     = $1;
  my $nameFeature = $3;
  my $start       = $4;
  my $end         = $5;
  my $score       = $6;
  my $strand      = $7;
  my $frame       = $8;

  my $gene_start = $start;

  if ($debug) {
    print STDERR "gene start: $gene_start\n";
  }

  if ($strand eq "+") {
    $strand = 1;
  }
  else {
    $strand = -1;
  }

  if ($nameFeature =~ /coding/i) {
    
    # The current feature is an exon
    
    $nb_exons++;
    my $range = Bio::PSU::Range->new(
				     -start       => $start,
				     -end         => $end,
				     -strand      => $strand
				    );
    
    push (@ranges, $range);
    push (@scores, $score);
    $total_scores = $total_scores + $score;
  }
  
  $index++;
  while ($index < @hmmgene_features) {
    $line = $hmmgene_features[$index];
    $line =~ /([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+(.+)/;
    
    my $seqName     = $1;
    my $nameFeature = $3;
    my $start       = $4;
    my $end         = $5;
    my $score       = $6;
    my $frame       = $8;
    
    if ($nameFeature =~ /coding/i) {
      
      # Exon feature
      
      $nb_exons++;
      my $range = Bio::PSU::Range->new(
				       -start       => $start,
				       -end         => $end,
				       -strand      => $strand
				      );    
      
      push (@ranges, $range);
      push (@scores, $score);
      $total_scores = $total_scores + $score;
    }
    elsif ($nameFeature =~ /intron/i) {
      # Intron feature
      # do nothing !!
    }
    elsif ($nameFeature =~ /utr/i) {
      
      # 5' or 3' UTR

      # case when there is no predicted exons !!!

      if ($nb_exons == 0) {
	return ($index, undef, undef);
      }
      
      # create PSU features objects
      
      if ($debug) {
	print STDERR "Uncomplete fuzzy gene, one UTR is missing\n";
      }

      # Get Frame

      if (@ranges && ($strand == 1)) {
	$codon_start = getFrame ($ranges[0], $seq_obj);
      }
      
      my $gene_end   = $end;
      
      if ($debug) {
	print STDERR "start, end: $gene_start, $gene_end\n";
      }
      
      my $gene_range = Bio::PSU::Range->new(
					    -start  => $gene_start,
					    -end    => $gene_end,
					    -strand => $strand
					   );
      if ($strand) {
	$gene_range->no_5prime(1);
      }
      else {
	$gene_range->no_3prime(1);
      }
      
      my $cds_score = 0;
      $cds_score = $total_scores / $nb_exons unless $nb_exons == 0;
      $cds_score = round ($cds_score, 2);
      my $note = "HMMGene - exon scores: " . join (', ', @scores);
      
      my $gene_feature = Bio::PSU::Feature->new(
						-key    => 'gene',
						-ranges => $gene_range
					       );
      my $cds_feature = Bio::PSU::Feature->new(
					       -key    => 'CDS',
					       -ranges => \@ranges
					      );
      
      $cds_feature->qadd('note', $note);
      $cds_feature->qadd('method', 'HMMGENE');
      $cds_feature->qadd('score', $cds_score);
      $cds_feature->qadd('colour', '6');

      if (defined $codon_start) {
	$cds_feature->qadd('codon_start', $codon_start);
      }
      
      $gene_feature->qadd('method', 'HMMGENE');
      $gene_feature->qadd('note', $note);
      
      return ($index, $gene_feature, $cds_feature);
    }
    
    if (($index+1) >= @hmmgene_features) {

      # case when there is no predicted exons !!!

      if ($nb_exons == 0) {
	return ($index, undef, undef);
      }

      # this is an unfinished sequence
      
      if ($debug) {
	print STDERR "Uncomplete gene, both UTRs are missing\n";
      }
      
      # create PSU features objects
      
      my $gene_end   = $end;

      if (@ranges) {
	
	# Get Frame
	
	if ($strand == 1) {
	  $codon_start = getFrame ($ranges[0], $seq_obj);
	}
	elsif ($strand == -1) {
	  $codon_start = getFrame ($ranges[@ranges-1], $seq_obj);
	}
      }

      if ($debug) {
	print STDERR "start, end: $gene_start, $gene_end\n";
      }
      
      my $gene_range = Bio::PSU::Range->new(
					    -start  => $gene_start,
					    -end    => $gene_end,
					    -strand => $strand
					   );
      $gene_range->no_5prime(1);
      $gene_range->no_3prime(1);
      
      my $gene_feature = Bio::PSU::Feature->new(
						-key    => 'gene',
						-ranges => $gene_range
					       );
      
      my $cds_feature = Bio::PSU::Feature->new(
					       -key    => 'CDS',
					       -ranges => \@ranges
					      );
      
      my $cds_score = $total_scores / $nb_exons unless $nb_exons == 0;
      $cds_score = round ($cds_score, 2);
      my $note = "HMMGene - exon scores: " . join (', ', @scores);
      
      $cds_feature->qadd('note', $note);
      $cds_feature->qadd('method', 'HMMGENE');
      $cds_feature->qadd('score', $cds_score);
      $cds_feature->qadd('colour', '6');
      $cds_feature->qadd('is_partial');
      
      if (defined $codon_start) {
	$cds_feature->qadd('codon_start', $codon_start);
      }
      
      $gene_feature->qadd('method', 'HMMGENE');
      $gene_feature->qadd('note', $note);
      
      return ($index, $gene_feature, $cds_feature);
    }
    
    $index++;
  }
  
  print STDERR "problem, not supposed to come here !!!!!!!!!!!\n";
  
  return ($index, undef, undef);
}

#####################################################
#
# Gene which starts with a 3'UTR in HMMgene Output
#
# @param Hmmgene features
#
# @return a PSU CDS feature
# @return a PSU gene feature
# @return the index for next HMMgene gene processing
# 
#####################################################

sub parse3UTRGene {
  my ($index, $seq_obj, @reverse_features) = @_;
  
  my $codon_start = undef;
  
  my $line = $reverse_features[$index];
  
  $line =~ /([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+(.+)/;
  
  my $seqName     = $1;
  my $nameFeature = $3;
  my $start       = $4;
  my $end         = $5;
  my $score       = $6;
  my $strand      = $7;
  my $frame       = $8;
  
  my $gene_start = $start;
  
  if ($debug) {
    print STDERR "gene start: $gene_start\n";
  }
  
  my $total_scores = 0;
  my $nb_exons = 0;
  my @ranges = ();
  my @scores = ();
  
  $index++;
  while ($index < @reverse_features) {
    $line = $reverse_features[$index];
    $line =~ /([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+([^\t]+)\t+(.+)/;
    
    my $seqName     = $1;
    my $nameFeature = $3;
    my $start       = $4;
    my $end         = $5;
    my $score       = $6;
    my $strand      = $7;
    my $frame       = $8;
    
    if ($debug) {
      print STDERR "nameFeature: $nameFeature\n";
    }
    
    if ($nameFeature =~ /coding/i) {
      
      # Exon feature
      
      $nb_exons++;
      my $range = Bio::PSU::Range->new(
				       -start       => $start,
				       -end         => $end,
				       -strand      => -1
				      );
      push (@ranges, $range);
      push (@scores, $score);
      $total_scores = $total_scores + $score;
    }
    elsif ($nameFeature =~ /intron/i) {
      # Intron feature
      # do nothing !!
    }
    elsif ($nameFeature =~ /5utr/i) {
      
      # 5' UTR 

      # case when there is no predicted exons !!!

      if ($nb_exons == 0) {
	return ($index, undef, undef);
      }

      # create PSU features objects
      
      if ($debug) {
	print STDERR "Complete 3' gene\n";
      }
      
      my $gene_end = $end;
      
      if ($debug) {
	print STDERR "gene start, end: $gene_start, $gene_end\n";
      }
      
      my $gene_range = Bio::PSU::Range->new(
					    -start  => $gene_start,
					    -end    => $gene_end,
					    -strand => -1
					   );
      my $cds_score = $total_scores / $nb_exons unless $nb_exons == 0;
      $cds_score = round ($cds_score, 2);
      my $note = "HMMGene - exon scores: " . join (', ', @scores);
      
      my $gene_feature = Bio::PSU::Feature->new(
						-key    => 'gene',
						-ranges => $gene_range
					       );
      my $cds_feature = Bio::PSU::Feature->new(
					       -key    => 'CDS',
					       -ranges => \@ranges
					      );
      
      $cds_feature->qadd('note', $note);
      $cds_feature->qadd('method', 'HMMGENE');
      $cds_feature->qadd('score', $cds_score);
      $cds_feature->qadd('colour', '6');
      
      $gene_feature->qadd('method', 'HMMGENE');
      $gene_feature->qadd('note', $note);
      
      return ($index, $gene_feature, $cds_feature);
    }
    
    if (($index+1) >= @reverse_features) {
      # this is an unfinished sequence

      # case when there is no predicted exons !!!

      if ($nb_exons == 0) {
	return ($index, undef, undef);
      }
      
      if ($debug) {
	print STDERR "Uncomplete 3'gene, missing the 5'UTR\n";
      }
      
      # create PSU features objects
      
      if (@ranges) {
	# get the frame
	$codon_start = getFrame ($ranges[@ranges-1], $seq_obj);
      }

      if ($debug) {
	print STDERR "codon start: $codon_start\n";
      }
      
      my $gene_end   = $end;
      
      if ($debug) {
	print STDERR "gene start, end: $gene_start, $gene_end\n";
      }
      
      my $gene_range = Bio::PSU::Range->new(
					    -start  => $gene_start,
					    -end    => $gene_end,
					    -strand => -1
					   );
      $gene_range->no_5prime(1);
      
      my $cds_score = $total_scores / $nb_exons unless $nb_exons == 0;
      $cds_score = round ($cds_score, 2);
      my $note = "HMMGene - exon scores: " . join (', ', @scores);
      
      my $gene_feature = Bio::PSU::Feature->new(
						-key    => 'gene',
						-ranges => $gene_range
					       );
      my $cds_feature = Bio::PSU::Feature->new(
					       -key    => 'CDS',
					       -ranges => \@ranges
					      );
      
      $cds_feature->qadd('note', $note);
      $cds_feature->qadd('method', 'HMMGENE');
      $cds_feature->qadd('score', $cds_score);
      $cds_feature->qadd('colour', '6');
      $cds_feature->qadd('is_partial','');	
      $cds_feature->qadd('codon_start', $codon_start);
      $gene_feature->qadd('method', 'HMMGENE');
      $gene_feature->qadd('note', $note);
      
      return ($index, $gene_feature, $cds_feature);
    }
    
    $index++;
  }
  
  print STDERR "problem, not supposed to come here !!!!!!!!!!!\n";
  
  return ($index, undef, undef);
  
}

sub round {
  my ($number, $nb_decimal) = @_;
  my $length_string = length ($number);
  if ($number =~ /\./) {
    $number =~ m/\./g;
    my $length = pos ($number);
    $length = $length + $nb_decimal;
    my $new_number = substr ($number, 0, $length);
    # print STDERR "number, new_number: $number, $new_number\n";
    return $new_number;
  }
  else {
    return $number;
  }
}

##########################################################################
#
# Resolve the ORF when the first MET codon is missing
# Resolving by returning the frame where there is no stop codon
# Pb : which one to choose when there are frames without stop codons ???
#
# @Param The first exon range object
# @Param The Sequence object
# @Return the codon_start qualifier value for artemis
#
##########################################################################

sub getFrame {
  my ($range, $seq_obj) = @_;
  my $frame = undef;
  
  my $subseq = getSubSequence ($seq_obj, $range);

  if ($debug) {
    print STDERR "range object: " . Dumper ($range) . "\n";
    print STDERR "subseq: $subseq\n";
  }

  my $frame_index = 0;
  my $seq_length = length ($subseq);
  my $found_frame = 0;
  
  while (not ($found_frame) && ($frame_index < 3)) {
    
    if ($debug) {
      print STDERR "working on the frame $frame_index\n";
    }
    
    my $i = $frame_index;
    my $is_stop_codon = 0;
    while (($i<$seq_length-2) && not $is_stop_codon) {
      my $codon = substr ($subseq, $i, 3);
      if ($codon =~ /taa|tag|tga/i) {
	if ($debug) {
	  print STDERR "found a stop codon on frame $frame_index, $codon\n";
	}
	$is_stop_codon = 1;
      }
      $i = $i + 3;
    }
    if (not $is_stop_codon) {
      $found_frame = 1;
      $frame = $frame_index;
    }
    $frame_index++;
  }
  
  if ($debug) {
    print STDERR "frame found (yes/no) ? $found_frame\n";
  }
  
  my $codon_start = undef;
  if ($found_frame) {
    if ($frame =~ /0/) {
      $codon_start = 1;
    }
    elsif ($frame =~ /1/) {
      $codon_start = 2;
    }
    elsif ($frame =~ /2/) {
      $codon_start = 3;
    }
  }
  
  return $codon_start;
}

sub getSubSequence {
  my ($seq_obj, $range) = @_;

  if ($range->strand == -1) {
    my $rev_seq_obj = $seq_obj->revcom;
    my $length = length ($seq_obj->str);
    
    # + 3 to get rid of the last codon which might be a stop codon

    my $end = $length + 1 - $range->start - 3;
    my $start   = $length + 1 - $range->end;
    
    my $sub_seq_obj = $rev_seq_obj->subseq ($start, $end, 0);
    return $sub_seq_obj->str;
  }
  else {
    my $sub_seq_obj = $seq_obj->subseq ($range->start, $range->end - 3, 0);
    return $sub_seq_obj->str;
  }
}
