#!/usr/local/bin/perl --  # -*-Perl-*-

# $Header: /nfs/pathsoft/repository/psu/genexec/perl/src/intergenic_distance.pl,v 1.1 2000/03/27 09:44:54 kmr Exp $

# purpose: This script will either report on the intergenic distances or
# purpose: extract intergenic dna (depending on the switches used)

use strict;

use Embl;
use EMBL::Parser;

sub usage
{
  die <<EOF;
usage: $0 [-<number> [-s]] embl_sequence
   or: $0 embl_sequence

  -s        will wrap the output sequence at 60 characters per line
  -<number> sets the minimum size of the intergenic regions to output.  If this
            flag is -0 or is absent a summary of the intergenic sequence will
            be printed instead.

eg:
  $0 -1000 embl_sequence

will write out a FASTA database of all the intergenic regions that are at least
1000 bases long

EOF
}

open ERRORS, ">>intron_extract_errors";

my $extract_count = 0;

my $split_lines = 0;

while (@ARGV && $ARGV[0] =~ /^-(.+)/) {
  my $letter = $1;

  if ($letter =~ /^\d+$/) {
    $extract_count = $letter;
  } else {
    if ($letter eq 's') {
      $split_lines = 1
    } else {
      usage;
    }
  }
  shift;
}

if (!@ARGV) {
  usage;
}

open IN_FILE, "$ARGV[0]" or die "$0: can't open $ARGV[0]\n";

# read in some embl files from a stream
my $set = Read Embl(\*IN_FILE);

# get an array of all of the EMBL::Entry objects within $set
my @entries = $set->entries();
while (my $entry = pop @entries) {
  my $ft = $entry->FT ();

  my @features = $ft->features ();

  my @cds_features = ();

  for my $f (@features) {
    if ($f->{key} eq "CDS" or $f->{key} =~ /rna/i) {
      push @cds_features, $f;
    }
  }

  my @out_array = ();

  for (my $i = 0 ; $i < scalar (@cds_features) - 1 ; ++$i) {
    my $this_feature = $cds_features[$i];
    my $next_feature = $cds_features[$i + 1];

#    print "this feature: $$this_feature{key}  $$this_feature{location}\n";
#    print "next feature: $$next_feature{key}  $$next_feature{location}\n";

    my $this_location = $this_feature->{location};
    my $next_location = $next_feature->{location};

    if ($extract_count > 0) {
      my $intergene_start = get_end ($this_location) + 1;
      my $intergene_end = get_start ($next_location) - 1;

      my $intergene_length = $intergene_end - $intergene_start + 1;

#      print "$intergene_length $intergene_start $intergene_end\n";

      if ($intergene_length >= $extract_count) {
        my $sequence = $entry->SQ ();
  
        my $sequence_length = $sequence->length ();

        my $sequence_string = $sequence->seq;

        my $intergene_string = 
          substr ($sequence_string, $intergene_start - 1, $intergene_length);

        print ">", $ARGV[0], " ", $entry->ID (),
          " $intergene_length $intergene_start", 
          " $intergene_end\n";

        if ($split_lines) {
          $intergene_string =~ s/(.{60})/\1\n/g;
        }

        print $intergene_string, "\n";
      }
    } else {
      my $out_string = "";

      if ($this_location =~ /complement/) {
        $out_string .= "c ";
      } else { 
        $out_string .= "  ";
      }

      if ($next_location =~ /complement/) {
        $out_string .= "c ";
      } else { 
        $out_string .= "  ";
      }

      $out_string .= (get_start ($next_location) - get_end ($this_location));

      push @out_array, $out_string;
    }
  }

  next if ($extract_count > 0);

  my @together = ();

  my $total = 0;
  my $count = 0;

  for my $k (@out_array) {

    if ($k =~ /^c c (.*)/ || $k =~ /^    (.*)/) { 
      ++$count;
      $total += $1;
    }
  }

  print "tandem:    $count  ", $total/$count, "\n";

  $total = 0;
  $count = 0;

  for my $k (@out_array) {

    if ($k =~ /^c   (.*)/) { 
      ++$count;
      $total += $1;
    }
  }

  print "away:      $count  ", $total/$count, "\n";

  $total = 0;
  $count = 0;

  for my $k (@out_array) {

    if ($k =~ /^  c (.*)/ || $k =~ /^    (.*)/) { 
      ++$count;
      $total += $1;
    }
  }

  print "together:  $count  ", $total/$count, "\n";

  $total = 0;
  $count = 0;

  for my $k (@out_array) {

    if ($k =~ /^. . (.*)/) {
      ++$count;
      $total += $1;
    }
  }

  print "all:       $count  ", $total/$count, "\n";
}

sub get_end
{
  my $location = $_[0];

  my @ranges = get_ranges ($location);

  my $end = 0;

  for my $pos (@ranges) {
    if ($pos > $end) {
      $end = $pos;
    }
  }

  $end;
}

sub get_start
{
  my $location = $_[0];

  my @ranges = get_ranges ($location);

  my $start = 9999999999;

  for my $pos (@ranges) {
    if ($pos < $start) {
      $start = $pos;
    }
  }

  $start;
}


sub get_ranges
{
  return ($_[0] =~ m/(\d+)\.\.>?(\d+)|\d+/g);
}
