#!/usr/local/bin/perl5.6.0 -w

# $Header: /nfs/pathsoft/repository/psu/genexec/perl/src/map_protein_to_contigs.pl,v 1.6 2004/01/29 18:18:54 kmr Exp $

# purpose: This script takes an MSPcrunch file (which the result of the
# purpose: tblastn of a protein database and a FASTA database of contigs)
# purpose: and uses it to pick contigs that match from the contig database.

use strict;

use Getopt::Std;

use Bio::PSU::Seq;
use Bio::PSU::SeqFactory;
use Bio::PSU::IO::BufferFH;

sub usage
{
  die <<EOF;
usage:
  $0 [-k] contig_database mspcrunch_file

The aim of this script is to reorder the contigs in a fasta database of an
assembly so that the contigs appear in same order as the genes in an existing
finished genomes.

Example usage: If you have a finished, contiguous and annotated sequence
from organism A and an unfinished assembly from organism B which is syntenic
with A follow these steps:

1. Create a protein set from A containing all the sequences of all its genes
   in the order they appear in the genome
2. Do a tblastn of the protein database from A against a FASTA database of
   contigs from B then MSPcrunch the output
3. Run this script on the MSPcrunch file which will create a new FASTA
   database that contains only those contigs from B that were hit by the BLAST
   search.  The contigs in the new database will appear in the same order as
   the proteins in the database from A.
4. (optional) Create a BLASTN database of the sequence from genome A, then
   use big_blast.pl to create a crunch file for ACT:
   big_blast.pl -2 -tx [FASTA file from A] [ordered contigs from B]

Options:
  -k keep the contigs that don't match a protein and add them to the end of
     the output file (the default is to discard them)
EOF
}

my $debug = 0;
my $keep_non_matching_contigs = 0;

my %options = ();

getopts ("dk", \%options);

if (exists $options{d}) {
  $debug = 1;
}

if (exists $options{k}) {
  $keep_non_matching_contigs = 1;
}

if (exists $options{h}) {
  usage;
}

if (@ARGV != 2) {
  usage;
}

if ($debug) {
  open DEBUG, ">&STDERR";
} else {
  open DEBUG, ">/dev/null";
}

my %seq_hash = ();
my @seqs = ();

if (!$debug) {
  my $seqfactory =
    Bio::PSU::SeqFactory->make(-file   => $ARGV[0],
                               -format => "FASTA",
                               -type => 'dna');

  while (my $seq = $seqfactory->next_seq) {
    my $id = $seq->id ();

    push @seqs, $seq;
    $seq_hash{$id} = $seq;
  }
}



open MSPCRUNCH, "$ARGV[1]" or die "can't open $ARGV[1]\n";

my @q_names_in_order = ();

my %hits = ();

my %rev_hits = ();

my @hits = ();

while (<MSPCRUNCH>) {

  s/^\s+//;

  my @bits = split /\s+/, $_;

  my ($score,$percent_id,$q_start,$q_end,$q_name,$s_start,$s_end,$s_name) = @bits;

  $q_name =~ s/,$//;

  my $rev_comp = 0;

  if ($q_start > $q_end) {
    $rev_comp = 1;

    ($q_start, $q_end) = ($q_end, $q_start);
  }

  if ($s_start > $s_end) {
    $rev_comp = not $rev_comp;

    ($s_start, $s_end) = ($s_end, $s_start);
  }

  my $new_hit = {
                 q_name => $q_name,
                 s_name => $s_name,
                 q_start => $q_start,
                 q_end => $q_end,
                 rev_comp => $rev_comp,
                 score => $score
                };

  if (!exists $hits{$q_name}) {
    push @q_names_in_order, $q_name;
  }

  push @hits, $new_hit;

  push @{$hits{$q_name}}, $new_hit;

  push @{$rev_hits{$s_name}}, $new_hit;

  print DEBUG "pushed $score, $q_start, $q_end, $q_name, $rev_comp, $s_name, $new_hit, $rev_comp\n";
}

my @extra_hits = ();

sub remove_hit
{
  my $hit = $_[0];

  @hits = grep { $_ ne $hit } @hits;

  for my $s_name (keys %rev_hits) {
    my @this_s_hits = @{$rev_hits{$s_name}};

    @{$rev_hits{$s_name}} = grep {$_ ne $hit} @this_s_hits;
  }

  for my $q_name (keys %hits) {
    my @this_q_hits = @{$hits{$q_name}};

    @{$hits{$q_name}} = grep {$_ ne $hit} @this_q_hits;
  }
}


# if any contig has more than one hit, remove all but the top scoring hit
for my $s_name (keys %rev_hits) {
  print DEBUG "exmining hits to $s_name\n";

  my @this_s_hits = @{$rev_hits{$s_name}};

  my $max_score = -1;
  my $max_hit = undef;

  for (my $i = 0 ; $i < @this_s_hits ; ++$i) {
    my $hit = $this_s_hits[$i];
    my $q_name = $hit->{q_name};
    my $this_score = $hit->{score};

    print DEBUG "hit $hit\n";

    if ($this_score > $max_score) {
      $max_score = $this_score;
      $max_hit = $hit;
      print DEBUG "max: $this_score $i $hit from ",$hit->{q_name}," to ",
      $s_name, "\n";
    } else {
      print DEBUG "not max: $this_score $i $hit ",$hit->{q_name}," to ",
      $s_name, "\n";
    }

    print DEBUG "$i $hit\n";
  }

  for (my $i = @this_s_hits - 1 ; $i >= 0 ; --$i) {
    my $hit = $this_s_hits[$i];

    print DEBUG "\$i: $i $hit\n";

    print DEBUG "$max_hit\n";
    print DEBUG "$hit != $max_hit\n";

    if ($hit != $max_hit) {
      my $hit_q_name = $hit->{q_name};

      print DEBUG "removing $hit_q_name $hit\n";

      #      print DEBUG "new \@this_q_hits: @this_q_hits\n";

      # remove from full list
      print DEBUG "before: ", scalar (@hits), "\n";
      remove_hit ($hit);
      print DEBUG "after: ", scalar (@hits), "\n";
    }
  }
}



# if a protein hits more than one contig remove all but the hit with the
# maximum score (unless one of the other hits is the only hit to it's contig
# in which case save it in @extra_hits)
if (0) {
  for my $q_name (keys %hits) {
    print DEBUG "exmining hits from $q_name\n";

    my @this_q_hits = @{$hits{$q_name}};

    my $max_score = -1;
    my $max_hit = undef;

    print DEBUG "\@this_q_hits: @this_q_hits\n";

    for (my $i = 0 ; $i < @this_q_hits ; ++$i) {
      my $hit = $this_q_hits[$i];
      my $s_name = $hit->{s_name};
      my $this_score = $hit->{score};

      if ($this_score > $max_score) {
        $max_score = $this_score;
        $max_hit = $hit;
        print DEBUG "max: $this_score $i $hit ",$hit->{s_name},"\n";
      } else {
        print DEBUG "not max: $this_score $i $hit ",$hit->{s_name},"\n";
      }

      print DEBUG "$i $hit\n";
    }

    for (my $i = @this_q_hits - 1 ; $i >= 0 ; --$i) {
      my $hit = $this_q_hits[$i];

      print DEBUG "\$i: $i $hit\n";

      print DEBUG "$max_hit\n";
      print DEBUG "$hit != $max_hit\n";

      if ($hit != $max_hit) {
        my $hit_s_name = $hit->{s_name};

        print DEBUG "removing $hit_s_name $hit\n";

        if (@{$rev_hits{$hit_s_name}} <= 1) {
          # save the hit since it is the last hit to the contig
          push @extra_hits, $hit;
        }

        #      print DEBUG "new \@this_q_hits: @this_q_hits\n";

        # remove from full list
        print DEBUG "before: ", scalar (@hits), "\n";
        remove_hit ($hit);
        print DEBUG "after: ", scalar (@hits), "\n";
      }
    }

    @{$hits{$q_name}} = @this_q_hits;
  }
}

my $out_seq = Bio::PSU::SeqFactory->make(-fh     => \*STDOUT,
                                         -format => 'fasta');


my %written_seqs = ();

for my $q_name (@q_names_in_order) {
  my @this_q_hits = @{$hits{$q_name}};

  my $rev_query = 0;

  if ($q_name =~ /\.comp$/) {
    $rev_query = 1;
  }

  # sort by first query base
  @this_q_hits = sort {
    if ($rev_query) {
      $b->{q_end} <=> $a->{q_end}
    } else {
      $a->{q_start} <=> $b->{q_start}
    }
  } @this_q_hits;

  print DEBUG "\@this_q_hits: ";
  print DEBUG join " ", map { $_->{q_name} } @this_q_hits;

  print DEBUG "\n";

  for my $hit (@this_q_hits) {
    my $seq_name = $hit->{s_name};

    print DEBUG "writing: ", $seq_name, " ($hit)\n";

    if (!defined $seq_hash{$seq_name}) {
      warn "$0: can't find ",$seq_name," in contigs file\n";
    }

    if ($hit->{rev_comp} ^ $rev_query) {
      print DEBUG "writing rev_comp ", $hit->{rev_comp}, " ", $rev_query;
    } else {
      print DEBUG "writing ", $hit->{rev_comp}, " ", $rev_query;
    }

    print DEBUG "\n";

    if (!defined $seq_hash{$seq_name}) {
      next;
    }

    my $seq = $seq_hash{$seq_name};

    ++$written_seqs{$seq_name};

# warn $hit->{rev_comp} . "^ $rev_query = " . ($hit->{rev_comp} ^ $rev_query) . "\n";

    if ($hit->{rev_comp} ^ $rev_query) {
      $out_seq->write_seq ($seq->revcom);
    } else {
      $out_seq->write_seq ($seq);
    }
  }
}

if ($keep_non_matching_contigs) {
  for my $seq (@seqs) {
    if (!exists $written_seqs{$seq->id()}) {
      $out_seq->write_seq ($seq);
    }
  }
}
