#!/usr/local/bin/perl5.6.1 -w

=head1 NAME

 mspcrunch_guided_cp - perform copying of EMBL qualifiers from one
 feature to another, guided by an MSPCrunch-format comparison file

=head1 SYNOPSIS

Examples:

 mspcrunch_guided_cp query_embl_file subject_embl_file msp_crunch_file note product

An EMBL file is written to STDOUT

=head1 DESCRIPTION

This script uses the MSPCrunch output generated by the
reciprocal_fasta script. It will identify reciprocally best-matched
CDS features in a pair of EMBL files and transfer selected annotation
from the source (subject) to the destination (query) features.

The command line requires the two EMBL files (source first, then
destination), the MSPCrunch file and then a list of all EMBL
qualifiers whose values are to be transferred.

=head1 METHODS

None

=head1 AUTHOR

Keith James (kdj@sanger.ac.uk)
Ellen Schofield (es2@sanger.ac.uk)

=head1 COPYRIGHT

Copyright (C) 2002 Genome Research Limited. All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind. It may
be used, redistributed and/or modified under the same conditions as
Perl itself.

=cut

#####################################################
#
# Change the script to cope with a query file having more than one entries/sequences.           -DONE
#
# Which means that the unique crunch file will actually point at different quey sequences.
#
#####################################################

use strict;
use Bio::PSU::SeqFactory;
use Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper;
#use Bio::SeqIO;
#use Bio::PSU::Utils::FeatureIdHelper::BioPerlHelper;

use Getopt::Std;

my $_debug = 0;

my %opts;
my $transfert    = 0;
my $_qual_prefix = "";
my $regex;
 
my $usage = "Usage: mspcrunch_guided_cp -t -p -r [regex expression] <query EMBL file> <subject EMBL file> <MSPCrunch file> [<qualifier list>]

\t-t Transfer the qualifers with mapping information, e.g. <<FROM systematic_id=gene1 filename=file1.embl date=20031012>>.
\t-p The mapped qualifiers will be given a prefix, transfered_, e.g. /transfered_product.
";

# Process options
getopts('r:tph', \%opts);

defined $opts{h} and die $usage;
defined $opts{t} and $transfert++;
defined $opts{p} and $_qual_prefix = "transferred_";
defined $opts{r} and $regex = $opts{r};

if ($_debug) {
  print STDERR "transfert: $transfert\n";
}

my $query_file   = shift;
my $subject_file = shift;
my $crunch_file  = shift;

my @qualifiers = @ARGV;

if ((! defined $query_file) || (! defined $subject_file) || (! defined $crunch_file))
{
    die $usage
}

if ($_debug) {
  print STDERR "query_file: $query_file, subject_file: $subject_file, crunch_file: $crunch_file\n";
}

#my $id_helper =
#  Bio::PSU::Utils::FeatureIdHelper::BioPerlHelper->new(-selection_regex => $regex,
#						       -verbose => 1);

my $id_helper =
  Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper->new(-selection_regex => $regex,
						      -verbose => 1);

# BioPSU
my $qi = Bio::PSU::SeqFactory->make(-file   => "<$query_file",
                                    -format => 'embl');
my $si = Bio::PSU::SeqFactory->make(-file   => "<$subject_file",
                                    -format => 'embl');
my $qo = Bio::PSU::SeqFactory->make(-fh     => \*STDOUT,
                                    -format => 'embl');
# Bioperl
#my $qi = Bio::SeqIO->new(-file   => "<$query_file",
#			 -format => 'embl');
#my $si = Bio::SeqIO->new(-file   => "<$subject_file",
#			 -format => 'embl');

my @q_seq;
my %q_hash;
my @s_seq;
my %s_hash;

print STDERR "Analysing Query Sequence .";
while (my $query_seq   = $qi->next_seq){
    print STDERR ".";
    foreach my $feat ($query_seq->features)
	#foreach my $feat ($_->all_SeqFeatures)
    {
	next unless ($feat->key eq 'CDS');
	my $systematic_id = undef;
	eval {
	    $systematic_id = $id_helper->get_systematic_id($feat);
	};
	
	if ($@) {
	    my $loc = join ",", map {
		$_->{start} . ".." . $_->{end}
	    } $feat->ranges ();
	    
	    print STDERR "Error in CDS at: $loc\n";
	    print STDERR "A Systematic Id is missing !!\n";
	    return;
	}
	$q_hash{$systematic_id} = $feat;
    }
}

my @subject_seqs = ();

while (my $subject_seq = $si->next_seq) {
    push @subject_seqs, $subject_seq;
}

print STDERR "\nAnalysing Subject Sequence .";
for my $subject_seq (@subject_seqs){
    print STDERR ".";
    foreach my $feat ($subject_seq->features)
	#foreach my $feat ($_->all_SeqFeatures)
    {
	next unless ($feat->key eq 'CDS');
	my $systematic_id = undef;
	eval {
	    $systematic_id = $id_helper->get_systematic_id($feat);
	};
	
	if ($@) {
	    my $loc = join ",", map {
		$_->{start} . ".." . $_->{end}
	    } $feat->ranges ();
	    
	    print STDERR "Error in CDS at: $loc\n";
	    print STDERR "A Systematic Id is missing !!\n";
	    return;
	}
	$s_hash{$systematic_id} = $feat;
    }
}
print STDERR "\n";

my $date;
{
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime;
    
    $year += 1900;
    $mon++;
    
    $date = sprintf "%s%2.2d%2.2d", $year, $mon, $mday;
}

open(CRUNCH, "<$crunch_file") or die "Failed to open MSPCrunch file $crunch_file: $!\n";

while (<CRUNCH>)
{
    print STDERR "$_";
    my ($score, $percent,
        $q_start, $q_end, $q_name,
        $s_start, $s_end, $s_name) = split;

    my $query_feat = find_feature($q_start, $q_end, $q_name, %q_hash);
    my $subject_feat = find_feature($s_start, $s_end, $s_name, %s_hash);

    if (! defined $query_feat){
        print STDERR "Unable to find destination feature $q_name at $q_start..$q_end\n"
    }
    elsif (! defined $subject_feat){
         print STDERR "Unable to find source feature $s_name at $s_start..$s_end\n"
    }
    else{
        foreach my $qual (@qualifiers){
		
	  if ($transfert){
	    print STDERR "   Transfering qualifiers ($qual)...\n";
	    my $filename      = $query_file;

	    my $systematic_id = undef;
	    my $primary_name  = undef;

	    eval {
	      $systematic_id = $id_helper->get_systematic_id($query_feat);
	    };
	    
	    if ($@) {
	      my $loc = join ",", map {
		$_->{start} . ".." . $_->{end}
	      } $query_feat->ranges ();

	      print STDERR "Error in CDS at: $loc\n";
	      print STDERR "A Systematic Id is missing !!\n";
	      return;
	    }

	    #eval {
	    #  $primary_name = $id_helper->get_primary_name($query_feat);
	    #};
	    
	    if ($@) {
	      my $loc = join ",", map {
		$_->{start} . ".." . $_->{end}
	      } $query_feat->ranges ();

	      if ($_debug) {
		print STDERR "Warning in CDS at: $loc\n";
		print STDERR "The primary name is missing !!\n";
	      }
	    }


	    my $transfert_prefix;

	    if (defined $primary_name) {
	      $transfert_prefix = "<<FROM systematic_id=$systematic_id primary_name=$primary_name filename=$filename date=$date>> ";
	    }
	    else {
	      $transfert_prefix = "<<FROM systematic_id=$systematic_id filename=$filename date=$date>> ";
	    }
	    
	    my @q_values = $query_feat->qvalues($qual);
	    my $q_value_to_add;
	    my $qual_to_add;

	    # my @q_values = $query_feat->each_tag_value($qual);
	    foreach my $q_value (@q_values) {
		$q_value_to_add = $transfert_prefix . $q_value;
		# Add a prefix, "transfered_" if necessary
		$qual_to_add = $_qual_prefix . $qual;
		$subject_feat->qadd($qual_to_add, $q_value_to_add);
		#$subject_feat->add_tag_value($qual, $q_value);
	      
	    }
	  }
	  else {
	    my @q_values = $query_feat->qvalues($qual);
	    $subject_feat->qadd($qual, @q_values);

	    #my @q_values = $query_feat->each_tag_value($qual);
	    #$subject_feat->add_tag_value($qual, @q_values);
	  }
	  #$qo->write_seq($subject_feat)
        }
    }
}

for my $seq (@subject_seqs){
    $qo->write_seq($seq);
}


sub find_feature
{
    my ($start, $end, $name, %seqs) = @_;

    my $feat = $seqs{$name};
    if ($feat->start == $start && $feat->end == $end){
	return $feat;
    }
    return undef;
}



sub find_feature_old
{
    my ($start, $end, $name, @seqs) = @_;

    foreach (@seqs){
	foreach my $feat ($_->features){
	    next unless ($feat->key eq 'CDS');
	    my $systematic_id = undef;
	    eval {
		$systematic_id = $id_helper->get_systematic_id($feat);
	    };
	    
	    if ($@) {
		my $loc = join ",", map {
		    $_->{start} . ".." . $_->{end}
		} $feat->ranges ();
		
		print STDERR "Error in CDS at: $loc\n";
		print STDERR "A Systematic Id is missing !!\n";
		return;
	    }
	    
	    if ($feat->start == $start && $feat->end == $end && $name eq $systematic_id){
		return $feat;
	    }
	}
    }
    return undef;
}
