#!/usr/local/bin/perl

=head1 NAME

pfagss2tab.pl 

=head1 SYNOPSIS

  perl -w pfagss2tab.pl 

  -help a help message
  -i the input
  -m the machine where pfagss will be running

=head1 DESCRIPTION

  This script wraps pfagss gene prediction processing in order to generate a tab file with the gene prediction on the 6 ORFs.
  bear in mind that pfagss only runs on linux

  @author axk

=cut

# External Modules

# Constraint variables declaration before using them
use strict;

use Cwd;

# Bio modules
# PSU
use Bio::PSU::SeqFactory;
# BioPerl
use Bio::SeqIO;

##########################################
#
# Loading Parameters
#
##########################################

my $_help          = "-help";
my $_input         = "-i";
my $_machine       = "-m";

# no input per default
my $input_name     = "-1";
my $machine        = "ics-linux1";

# Check if in the command line there is an input file name (-i option)
# the other options are not compulsory (set of options per default)
my $lgth = @ARGV + 0;
# if no args, display the help message
if ($lgth == 0) {
  print "specify at least an input file with the option \"-i\"\n";
  display_help ();
  exit 1;
}
elsif ($lgth >= 1) {
  my $i = 0;
  foreach my $element (@ARGV) {
    if ($element eq $_help) {
      display_help ();
      exit 1;
    }
    elsif ($element eq $_input) {
      $input_name = $ARGV[$i+1];
    }
    elsif ($element eq $_machine) {
      $machine = $ARGV[$i+1];
    }
    # if at least one of the options is not recognized, display the help message and exit
    elsif ($element =~ /^-/) {
      print STDERR "option not recognized: $element\n";
      display_help ();
      exit 1;
    }
    $i ++;
  }
}
if ($input_name eq "-1") {
  print STDERR "ERROR - specify at least the name of an input file with the -i option\n";
  print STDERR "Syntax : perl -w est_clustering.pl -i [input_file]\n";
  exit 1;
}

# print ("$input_name\n");
my @tmp = split (/\./, $input_name);
my $input_rev_name = $tmp[0] . "_rev.fasta";
my $output_name = $tmp[0] . ".out";
my $output_rev_name = $tmp[0] . "_rev.out";
my $tab_file = $tmp[0] . ".tab";

my $curdir = cwd;
print "current dir : $curdir\n";
my @list = ($curdir);
my @grep_list = grep (/^\/tmp_mnt*/, @list);
if ((@grep_list+0) > 0) {
  my @tmp = split (/\/tmp_mnt/, $curdir);
  $curdir = $tmp[1];
  print ("curdir,tmp : $curdir, @tmp.\n");
}

my $input_name_path = $curdir . "/" . $input_name;
my $input_rev_name_path = $curdir . "/" . $input_rev_name;
my $output_name_path = $curdir . "/" . $output_name;
my $output_rev_name_path = $curdir . "/" . $output_rev_name;

print "pfagss wrapping is being processed...\n\n";

# generation of the reverse genbank file

print "seqret -filter -srev -sequence $input_name -outseq $input_rev_name\n";
my @args = ("seqret -filter -srev -sequence $input_name -outseq $input_rev_name");
system (@args);

# sequence length - using PSU Bio Modules

# PSU
my $s = Bio::PSU::SeqFactory->make(-file   => "<$input_rev_name_path",
                                   -format => 'fasta');
my $lgth_seq = length ($s->next_seq->str);

# BIOPerl
# my $in  = Bio::SeqIO->new('-file' => "$input_rev_name_path",
#                        '-format' => 'Fasta');
# my $seq = $in->next_seq();
# my $lgth_seq = $seq->length();
# my $lgth_seq = 72480;

print "seq length: $lgth_seq\n";

# pfagss running twice

@args = ("rsh $machine \"cd $curdir ; pfagss $input_name\"");
print ("cmd : $args[0]\n");
system (@args);

@args = ("rsh $machine \"cd $curdir ; pfagss $input_rev_name\"");
print ("cmd : $args[0]\n");
system (@args);

# Features list on the 5 prime strand
my @feat_list_5;
my $parse_line = "false";

my $_utr = "misc_feature";
my $_int = "intron";
my $_cdr = "exon";

# parsing the outputs to generate a unique tab file

print "parsing the first output file,$output_name...\n\n";

open (INFILE, "< $output_name_path") or die "can't open this file: $!";
while (<INFILE>) {
  chomp;

  # sometimes there is a ">" (when fasta format actually) at the begining of the line
  # so replace ">" by a couple of space characters
  # in order to have a complient syntax with the pfagss output iwth genbank format file

  substr ($_, 0, 2) =~ s/>/   /g;

  # print ("line: $_\n");

  # split the line to extract from it the information - the pattern matching is several spaces " +"
  my @tmp = split (/ +/, $_);
  # print "tmp:\n@tmp\n";
  my $base = 0;
  my $location = "";
  my $AG_site = "";
  my $region = "";
  my $key = "";
  my $tf = "";
  my $colour = 0;
  my $score = 0;

  if (lc ($parse_line) eq "true") {
    $base    = $tmp[2];
    $AG_site = $tmp[3];
    $region  = $tmp[4];
    $tf      = $tmp[5];
    $score   = $tmp[6];
    
    $location = ($base-2) . "..$base";

    if ($region eq "UTR") {
      $key = $_utr;
    }
    elsif ($region eq "INT") {
      $key = $_int;
    }
    elsif ($region eq "CDR") {
      $key = $_cdr;
    }
    else {
      print "new region, $region, not recognized !!!!\n";
    }

    if (lc($tf) eq "true") {
      $colour = 12;
    }
    else {
      $colour = 13;
    }

    # print ":$base, $AG_site, $region, $tf, $score\n";
    
    my %feat_data = (
		     location => $location,
		     ag       => $AG_site,
		     key      => $key,
		     tf       => $tf,
		     colour   => $colour,
		     score    => $score,
		    );
    
    push (@feat_list_5, \%feat_data);
  }
  
  if ($_ =~ /base/) {
    # print "first line found on $_\n";
    $parse_line = "true";
    # next;
  }
}
close (INFILE);

print "parsing the second output file,$output_rev_name...\n\n";

# Features list on the 3 prime strand
my @feat_list_3;
$parse_line = "false";

open (INFILE_REV, "< $output_rev_name_path") or die "can't open this file: $!";
while (<INFILE_REV>) {
  chomp;

  # sometimes there is a ">" (when fasta format actually) at the begining of the line
  # so replace ">" by a couple of space characters
  # in order to have a compliant syntax with the pfagss output iwth genbank format file

  substr ($_, 0, 2) =~ s/>/   /g;

  # print ("line: $_\n");

  # split the line to extract from it the information - the pattern matching is several spaces " +"
  my @tmp = split (/ +/, $_);
  my $base = 0;
  my $location = "";
  my $AG_site = "";
  my $region = "";
  my $key = "";
  my $tf = "";
  my $colour = 0;
  my $score = 0;

  if (lc ($parse_line) eq "true") {
    $base    = $tmp[2];
    $AG_site = $tmp[3];
    $region  = $tmp[4];
    $tf      = $tmp[5];
    $score   = $tmp[6];

    # the location is based on the forward strand => previous mapping
    # +1 because the sequence starts from 1 and not 0 !!!
    my $complement_base = $lgth_seq - $base + 1;

    $location = "complement($complement_base.." . ($complement_base+2) . ")";

    if ($region eq "UTR") {
      $key = $_utr;
    }
    elsif ($region eq "INT") {
      $key = $_int;
    }
    elsif ($region eq "CDR") {
      $key = $_cdr;
    }
    else {
      print "new region, $region, not recognized !!!!\n";
    }
    
    if (lc($tf) eq "true") {
      $colour = 12;
    }
    else {
      $colour = 13;
    }

    # print ":$base, $AG_site, $region, $tf, $score\n";
    
    my %feat_data = (
		   location => $location,
		   ag       => $AG_site,
		   key      => $key,
		   tf       => $tf,
		   colour   => $colour,
		   score    => $score,
		  );

    push (@feat_list_3, \%feat_data);
  }

  if ($_ =~ /base/) {
     $parse_line = "true";
     # next;
  }
}
close (INFILE_REV);

print "tab file creation...\n";

open (OUTFILE, "> $tab_file") or die "can't create this file: $!";

foreach my $href (@feat_list_5) {
  my %feat_data = %$href;

  my $location   = %feat_data->{location};
  my $key        = %feat_data->{key};
  my $tf         = %feat_data->{tf};
  my $colour     = %feat_data->{colour};
  my $ag         = %feat_data->{ag};
  my $note_ag    = "AG site = $ag";
  my $score      = %feat_data->{score};

  # print STDOUT "location: $location\n";

  printf OUTFILE "FT   %-15.15s        $location\n", $key;
  print  OUTFILE "FT                          /colour=$colour\n";
  print  OUTFILE "FT                          /note=\"$note_ag, score=$score, tf=$tf\"\n";
  # print  OUTFILE "FT                          /score=$score\n";
}

foreach my $href (@feat_list_3) {
  my %feat_data = %$href;

  my $location   = %feat_data->{location};
  my $key        = %feat_data->{key};
  my $tf         = %feat_data->{tf};
  my $colour     = %feat_data->{colour};
  my $ag         = %feat_data->{ag};
  my $note_ag    = "AG site = $ag";
  my $score      = %feat_data->{score};

  printf OUTFILE "FT   %-15.15s        $location\n", $key;
  print  OUTFILE "FT                          /colour=$colour\n";
  print  OUTFILE "FT                          /note=\"$note_ag, score=$score, tf=$tf\"\n";
  # print  OUTFILE "FT                          /score=$score\n";
}

close (OUTFILE);

print "process done\n";

###################################################################
# The end
###################################################################

# display the help message

sub display_help {
  print "perl -w pfagss2tab.pl -help -i [input_file] -m [machine]\n\n";
  print "\t-help a help message\n";
  print "\t-i the input - (no default, always specify it)\n";
  print "\t-m the machine where the process will be done - (default: ics-linux1)\n\n";
}

