#!/usr/local/bin/perl5.6.1 -w

use strict;
use Getopt::Std;
use Data::Dumper;

use Bio::PSU::Range;
use Bio::PSU::SeqFactory;
use Bio::PSU::Seq;
use Bio::PSU::Feature;

my $_debug = 0;

my (%opts);
my ($input, $output) = undef;
getopts('i:o:', \%opts);
defined $opts{i} and $input  = $opts{i};
defined $opts{o} and $output = $opts{o};

if (not defined $input) {
  die "Usage :
phat2tab.pl -i [GFF Phat output file] -o [tab output file]
";
}

open PHAT, "<$input" or die "can't open file, $input\n";

my $sembl;
if (defined $output) {

  print STDERR "Writing output to a new file, $output...\n";

  $sembl = Bio::PSU::SeqFactory->make(
				      -file   => ">$output",
				      -format => 'embl'
				     );
}
else {
  $sembl = Bio::PSU::SeqFactory->make(
				      -fh   => \*STDOUT,
				      -format => 'embl'
				     );
}

my @features = ();
my $seqName  = undef;
my $gene_id  = undef;
my @ranges   = ();
my $score    = 0;
my $what;
my $codon_start = undef;

while (<PHAT>) {
  my $line = $_;
  
  if ($_debug) {
    print STDERR "new line : $line";
  }

  if ($line =~ /[^\t]+\t[^\t]+\t[^\t]+\t[^\t]+\t[^\t]+\t[^\t]+\t[^\t]+\t[^\t]+\t[^\t]+/) {
    $line =~ /([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)/;
    
    $seqName       = $1;
    my $method     = $2;
    my $exon_start = $4;
    my $exon_end   = $5;
    my $proba      = $6;
    my $strand     = $7;
    if ($strand eq "+") {
      $strand = 1;
    }
    elsif ($strand eq "-") {
      $strand = -1;
    }
    else {
      print STDERR "strand unknown!!\n";
      $strand = 0;
    }

    # what's that ????????
    $what       = $8;
    my $rest       = $9;
    
    # Parse $rest to get the gene id
    
    $rest =~ /gene\s\"(\w+)\"\sframe\"(\d)\"/;
    
    my $gene_id_tmp = $1;
    my $frame       = $2;

    if ($_debug) {
      print STDERR "seqName: $seqName\n";#
      print STDERR "method: $method\n";
      print STDERR "exon_start: $exon_start\n";
      print STDERR "exon_end: $exon_end\n";
      print STDERR "proba: $proba\n";
      print STDERR "strand: $strand\n";
      print STDERR "what: $what\n";
      print STDERR "gene_id_tmp: $gene_id_tmp\n";
      print STDERR "frame: $frame\n";
    }
    
    if (not defined $gene_id) {
      $gene_id = $gene_id_tmp;

      $codon_start = getFrame ($what);

      $score += $proba;
      my $range = Bio::PSU::Range->new (
					-start  => $exon_start,
					-end    => $exon_end,
					-strand => $strand
				       );
      push (@ranges, $range);
    }
    elsif ($gene_id_tmp ne $gene_id) {

      if ($_debug) {
	print STDERR "gene, $gene_id\n";
	print STDERR "ranges, " . Dumper (@ranges) . "\n";
      }

      # create a new CDS feature
      my $feature = Bio::PSU::Feature->new (
					    -key    => 'CDS',
					    -ranges => [@ranges],
					   );
      $feature->qadd ('method', 'PHAT');
      my $nbexons = @ranges;

      if ($_debug) {
	print STDERR "nb exons: $nbexons\n";
      }

      $score = $score / $nbexons;
      $feature->qadd ('score', $score);
      $feature->qadd ('gene', $gene_id);

      if (defined $codon_start) {
	$feature->qadd ('codon_start', $codon_start);
      }

      push (@features, $feature);

      if ($_debug) {
	print STDERR "\nparsing a new CDS...\n\n";
      }

      # Initialisation

      $codon_start = undef;
      $gene_id = $gene_id_tmp;
      @ranges = ();
      my $range = Bio::PSU::Range->new (
					-start  => $exon_start,
					-end    => $exon_end,
					-strand => $strand
				       );
      push (@ranges, $range);
      $score = $proba;
    }
    else {
      $score += $proba;
      my $range = Bio::PSU::Range->new (
					-start  => $exon_start,
					-end    => $exon_end,
					-strand => $strand
				       );
      push (@ranges, $range);
    }
  }
}

close PHAT;

if (@ranges > 0) {
  
  # create the last CDS feature
  my $feature = Bio::PSU::Feature->new (
					-key    => 'CDS',
					-ranges => [@ranges],
				       );
  $feature->qadd ('method', 'PHAT');
  my $nbexons = @ranges;
  
  if ($_debug) {
    print STDERR "nb exons: $nbexons\n";
  }
  
  $score = $score / $nbexons;
  $feature->qadd ('score', $score);
  $feature->qadd ('gene', $gene_id);
  
  $codon_start = getFrame ($what);
  $feature->qadd ('codon_start', $codon_start);
  
  push (@features, $feature);
  
  my $seqobj = Bio::PSU::Seq->new (
				   -type => 'dna',
				   -desc => "Phat CDS Predictions",
				   -id   => $seqName,
				  );
  $seqobj->features (@features);
  $sembl->write_seq ($seqobj);
}

# Done !!
  
sub getFrame {
  my ($frame) = @_;
  
  if ($_debug) {
    print STDERR "frame: $frame\n";
  }
  
  if ($frame =~ /0/) {
    $codon_start = 1;
  }
  elsif ($frame =~ /1/) {
    $codon_start = 2;
  }
  elsif ($frame =~ /2/) {
    $codon_start = 3;
  }
  
  if ($_debug) {
    print STDERR "start codon: $codon_start\n";
  }

  return $codon_start;
}    
