#!/usr/local/bin/perl5.6.0 -w

=head1 NAME

prosite_by_perl - Splice-friendly Prosite search and annotation tool
using Perl regular expressions

=head1 SYNOPSIS

 prosite_by_perl filename
 prosite_by_perl -c 8 filename
 prosite_by_perl -c 0 -t 11 filename

=head1 DESCRIPTION

This program searches CDS features for Prosite patterns using Perl
regular expressions. The CDS features should be supplied in the form
of an EMBL file. The features will be translated using the Bacterial
translation table by default. The -t switch may be used to specify
alternative tables using the standard numbering system. Currently
features with /partial or /pseudo qualifiers are skipped because of
the phase problem (at some point I will add support for this by
searching a 3-frame translation in such cases).

The -c switch will allow the colour code of the features produced
to be changed (default colour is 8). The output is an EMBL feature
table sent to STDOUT.

=head1 METHODS

None

=head1 AUTHOR

Keith James (kdj@sanger.ac.uk)

=head1 COPYRIGHT

Copyright (C) 2001 Genome Research Limited. All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind. It may
be used, redistributed and/or modified under the same conditions as
Perl itself.

=cut

use strict;
use Getopt::Std;
use Bio::PSU::SeqFactory;
use Bio::PSU::FeatureFactory;

my $libdir   = "/nfs/disk222/yeastpub/bio-soft/prosite";
my $patterns = "$libdir/perl.regex";

my %patthash;

my $usage = "Usage: prosite_by_perl [-ct] <file name>

\t-c Colour code of features  (default  8)
\t-t Translation table number (default 11)

\tTables available are:

\t 1) Standard
\t 2) Vertebrate Mitochondrial
\t 3) Yeast Mitochondrial
\t11) Bacterial
";

my (%opts, $colour, $table);
getopts('c:ht:', \%opts);

defined $opts{h} and die $usage;
defined $opts{c} and $colour  = $opts{c};
defined $opts{t} and $table   = $opts{t};

# Default colour is green
$colour = 8 unless defined $colour;

# Default translation table is bacterial (11)
$table ||= '11';

my $file = shift;

unless (defined $file) { die $usage }

open(PATT, "<$patterns") or
    die "Unable to open Prosite patterns file $patterns: $!\n";

# File contains tab-delimited: PS accession, Perl regex, Pdoc
# accession, description
while (<PATT>)
{
    chomp;
    my ($acc, $regex, $pdoc, $desc) = split("\t", $_);

    my $compiled = eval { qr/$regex/mi } or
     	warn "Failed to compile regex: $regex\n";

    # Set anon list contents
    $patthash{$acc} = [$compiled, $pdoc, $desc];
}

close(PATT);

# Create EMBL stream and gather one sequence
my $es = Bio::PSU::SeqFactory->make(-file => "<$file");
my $e  = $es->next_seq;

# Handler for printing features as they appear, rather than printing a
# big feature table at the end
my $fth = Bio::PSU::IO::FTHandler->new;

my $anf = Bio::PSU::FeatureFactory->new;

foreach my $f ($e->features)
{
    # Discard non-CDS features
    next unless $f->key eq 'CDS';

    # Discard any CDS with /pseudo or /partial qualifier
    next if $f->qexists('pseudo|partial');

    # Translate with specified translation table
    my $aa = $f->translate($table);

    foreach my $acc (keys %patthash)
    {
	my @hits = search_regex(${$patthash{$acc}}[0], $aa->str);

	foreach my $hit (@hits)
	{
	    # Calculate which amino acid the hits starts on
	    my $aa_start = $hit->[0] - $hit->[1] + 1;

	    # Create a Prosite feature
	    my $psf = $anf->annotate_translation($f, $aa_start, $hit->[1]);

	    # Add some suitable annotation
	    $psf->qadd("note", "$acc " . ${$patthash{$acc}}[2]);
	    $psf->qadd("colour", $colour);

	    # Print features as they are created using the handler
	    print $fth->make_feature_block($psf);
	}
    }
}

=head2 search_regex

 Title   : search_regex
 Usage   : @hits = search_regex($regex, $aa)
 Function: Searches a string with a compiled regular expression and
         : returns a list of hits. Each hit is a list containing the
         : start coordinate and length of the match.
 Example :
 Returns : List of lists
 Args    : Compiled regex, sequence string

=cut

sub search_regex
{
    my ($regex, $aa) = @_;

    my @hits;

    while ($aa =~ /($regex)/g)
    {
	push(@hits, [pos($aa), length($1)])
    }

    return @hits;
}
