#!/usr/local/bin/perl5.6.0 -w

=head1 NAME

reputer_to_embl.pl - turn the output from reputer into a TAB file

=head1 SYNOPSIS

Examples:

reputer_to_embl.pl bob_rep_data > some_file.tab

=head1 DESCRIPTION

This program will make a TAB file from the output of the REPuter
(http://www.genomes.de/) repeat finding program.  The output will
contain one repeat_unit feature per repeat identified.  Each distinct
repeat sequence will get it's own ID, stored in /rpt_family of the
repeat_units.

The REPuter output must contain the sequence of the repeat.

=head1 METHODS

None

=head1 COPYRIGHT

Copyright (C) 2002 Genome Research Limited. All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind. It may
be used, redistributed and/or modified under the same conditions as
Perl itself.

=cut

use strict;
use Getopt::Std;

sub usage
{
  die <<EOF;
usage: $0 [-d <num>] reputer_file > some_file.tab

the -d flag specifies the maximum distance between two repeats.  Only repeat
families that contain at least one pair of repeat units this close or less
will be reported.

type:
  perldoc reputer_to_embl.pl
for more documentation
EOF
}

if (@ARGV && $ARGV[0] eq '-h') {
  usage;
}

my %opts = ();

getopts('d:h', \%opts);

if (defined $opts{h}) {
  usage;
}

my $max_distance = $opts{d};

my %hits = ();

sub save_feature
{
  my $sequence = shift;
  my $match_start = shift;

  ++$hits{$sequence}{$match_start};
}

my ($length, $match1_start, $match2_start, $e_value, $sequence);

while (<>) {
  if (/^\s*(\d+)\s+(\d+)\s+(\d+)\s+[\.\d]+\s+([e\.\+\-\d]+)(?:\s+(\w+))?/) {
    if (defined $sequence) {
      save_feature ($sequence, $match1_start);
      save_feature ($sequence, $match2_start);
      $sequence = undef;
    }

    ($length, $match1_start, $match2_start, $e_value, $sequence) =
      ($1, $2, $3, $4, $5);

    if (defined $sequence) {
      save_feature ($sequence, $match1_start);
      save_feature ($sequence, $match2_start);
      $sequence = undef;
    } else {
      $sequence = "";
    }
  } else {
    if (/^\s*([atgc]+)\s*$/i) {
      $sequence .= $1;
    } else {
      if (!/^\s*$|^REPuter|^Format/) {
        die "$0: exiting - can not parse this line:\n$_\n";
      }
    }
  }
}

if (defined $sequence) {
  save_feature ($sequence, $match1_start);
  save_feature ($sequence, $match2_start);
}

my $family = "rep00000";

for my $seq (sort { length $a <=> length $b } keys %hits) {
  my @starts = sort {$a <=> $b} keys %{$hits{$seq}};

  my $repeat_length = length $seq;

  my $process_this_repeat = 0;

  if (defined $max_distance) {
    for (my $i = 0 ; $i < @starts - 1 ; ++$i) {
      my $this_end = $starts[$i] + $repeat_length - 1;
      my $next_start = $starts[$i + 1];
      if ($next_start - $this_end - 1 <= $max_distance) {
        $process_this_repeat = 1;
      }
    }
  } else {
    $process_this_repeat = 1;
  }

  if (!$process_this_repeat) {
    warn "repeat skipped - repeat units too far apart " .
         "(>$max_distance): @starts - $seq\n";
    next;
  }

  for my $start (@starts) {
    my $location = $start . ".." . ($start + (length $seq) - 1);

    print <<EOF;
FT   repeat_unit     $location
FT                   /rpt_family="$family"
FT                   /note="repeat found by REPuter: $seq"
EOF
  }

  ++$family;
}

