#!/usr/local/bin/perl5.6.1 -w

use strict;
use Getopt::Std;

use Bio::PSU::Feature;
use Bio::PSU::IO::FTHandler;

use Bio::PSU::Utils::RfamParsing;

my %opts;
my $file;
my $cutoff = 12;
my $frags = 0;

my $usage = "
Usage: $0 -h (for help)
   or: $0 -i input file [-o output_file] [-c cutoff] [-f for fragmented RfamScans]
 
This script will parse the output of Rfam scan and create an EMBL feature
table.
 
Required:
   -i <file>    rfam scan results input file.

Options: 
   -o <file>  : will send the output to the given file rather than standard output.
   -c <value> : provides a cutoff score for hits, default = 12.
   -f         : for a fragmented rfam search (default = F)

";

getopts('hi:o:c:f', \%opts);

defined $opts{h} and die $usage;            #help
defined $opts{i} or die $usage;             #input file
if (defined $opts{o}){                      #output file
    close STDOUT;
    open STDOUT, ">$opts{o}" or die "can not open $opts{o}\n";
    #open STDOUT, ">$opts{o}" and die "Output file = $opts{o}\n";
}
defined $opts{c} and $cutoff = $opts{c};   #cutoff score
defined $opts{f} and $frags = 1;

my $rfam = Bio::PSU::Utils::RfamParsing->new(file => $opts{i},
                                             fragments => $frags);

my $fth = Bio::PSU::IO::FTHandler->new;

while (my $domain_obj = $rfam->next_domain()){

    my $sstart = $domain_obj->start;
    my $send = $domain_obj->end;
    my $id = $domain_obj->seqname;
    my $acc = $domain_obj->hmmacc;
    my $name = $domain_obj->hmmname;
    my $dstart = $domain_obj->hstart;
    my $dend = $domain_obj->hend;
    my $score = $domain_obj->bits;
    my $strand = $domain_obj->strand;
    my $key = "misc_RNA";

    if ($name =~ /tRNA/){
	$key = "tRNA";
    }
    elsif ($name =~ /rRNA/){
	$key = "rRNA";
    }

    next unless ($score >= $cutoff);

    my $f = Bio::PSU::Feature->new(-key    => $key,
 				   -start  => $sstart,
 				   -end    => $send,      
 				   -strand => $strand
				   );

    $f->qadd('temporary_systematic_id', $id);
    $f->qadd('domain', "Rfam:$acc\;$name;Score=$score;positions $dstart to $dend");

    print $fth->make_feature_block($f);
} 
