#!/usr/local/bin/perl5.6.0 -w 

#######################################
#######################################
# This script adds Ns within the sequence in an EMBL format, at the coordinates given by the contigs file. Shift the annotations to take into account the insertions.
#######################################
#######################################

use strict;

use Data::Dumper;

# PSU modules for EMBL writing
use Bio::PSU::SeqFactory;
use Bio::PSU::Seq;
use Bio::PSU::Feature;
use Bio::PSU::Range;


my $debug = 1;

my $_help     = "-help";
my $_tab      = "-t";
my $_contigs  = "-c";

# no input per default
my $tab       = "-1";
my $contigs   = "-1";

my $Ns = 100;
my $Ns_seq = "";
for (my $i=0; $i<$Ns; $i++) {
  $Ns_seq = $Ns_seq . "N";
}

# Check if in the command line there is an input file name (-i option)
# the other options are not compulsory (set of options per default)
my $lgth = @ARGV + 0;
my $i = 0;
foreach my $element (@ARGV) {
  if ($element eq $_help) {
    display_help ();
    exit 1;
  }
  elsif ($element eq $_contigs) {
    $contigs = $ARGV[$i+1];
    }
  elsif ($element eq $_tab) {
    $tab = $ARGV[$i+1];
  }
  # if at least one of the options is not recognized, display the help message and exit
  elsif ($element =~ /^-/) {
    print STDERR "option not recognized: $element\n";
    display_help ();
    exit 1;
  }
  $i ++;
}

if ($tab eq "-1" || $contigs eq "-1") {
  display_help ();
  exit 1;
}

my %contigs = getContigsSet ($contigs);

my $output_obj = Bio::PSU::SeqFactory->make(
					    -file   => "<$tab",
					    -format => 'embl'
					   );
my $seq_obj  = $output_obj->next_seq;
my @features = $seq_obj->features;

my $new_tab = "$tab.withNs";
my $new_output_obj = Bio::PSU::SeqFactory->make(
						-file   => ">$new_tab",
						-format => 'embl'
					       );

my $new_seq_str  = setUpdatedSequence ($seq_obj->str, $Ns_seq, %contigs);
my @new_features = setUpdatedFeatures ($Ns, \@features, %contigs);

my $new_seq_obj  = Bio::PSU::Seq->new (
				       -id   => $seq_obj->id,
				       -desc => $seq_obj->desc,
				       -str  => $new_seq_str,
				       -type => 'dna'
				      );
$new_seq_obj->features (@new_features);
$new_output_obj->write_seq ($new_seq_obj);

#####################
# end
##

sub display_help {
  print "shift_annotations.pl -t [embl file] -c [contigs file]\n\n";
  print "the number of inserted Ns is 100\n";
}


sub getContigsSet {
  my ($contigs) = @_;
  my %contigs;
  my $i = 0;

  print STDERR "parsing contigs file, $contigs...\n";

  open FILE, "<$contigs" or die "can't open contigs file, $contigs!\n";

  while (<FILE>) {
    my $line = $_;

    if ($line =~ /(\d+)\.\.(\d+)/) {
      
      my $start = $1;
      my $end   = $2;
      my $range = Bio::PSU::Range->new (
					-start  => $start,
					-end    => $end,
					-strand => 1
				       );
      $contigs{$i} = $range;
      $i++;
    }
    else {
      print STDERR "ERROR, contigs line doesn't match the correct pattern!!\n";
      print STDERR "line: $line\n";
    }
  }

  close FILE;

  print STDERR "parsing done!\n";

  return %contigs;
}

sub setUpdatedSequence {
  my ($seq_str, $Ns_seq, %contigs) = @_;
  my @contig_indexes = keys (%contigs);
  
  my $range = $contigs{0};
  my $start = 0;
  my $end = $range->end - 1;
  my $length = $end - $start + 1;
  my $new_seq_str = substr ($seq_str, 0, $length);

  for (my $i=1; $i<@contig_indexes; $i++) {
    $range = $contigs{$i};
    $start = $range->start - 1;
    $end   = $range->end - 1;
    $length = $end - $start + 1;
    $new_seq_str = $new_seq_str . $Ns_seq . substr ($seq_str, $start, $length);
  }

  # get the remaining !!!

  $start = $start + $length;
  $new_seq_str = $new_seq_str . $Ns_seq . substr ($seq_str, $start);
  
  return $new_seq_str;
}


sub setUpdatedFeatures {
  my ($Ns, $features_ref, %contigs) = @_;
  my @features = @$features_ref;
  my @new_features = ();
  
  foreach my $feature (@features) {
    my @ranges = $feature->ranges;
    my @new_ranges = ();

    foreach my $range (@ranges) {
      my $start     = $range->start;
      my $index_start = getIndex ($start, %contigs);
      my $new_start = $start + $Ns * $index_start;

      my $end       = $range->end;
      my $index_end = getIndex ($end, %contigs);
      my $new_end   = $end + $Ns * $index_end;

      my $new_range = Bio::PSU::Range->new (
					    -start  => $new_start,
					    -end    => $new_end,
					    -strand => $range->strand
					   );
      push (@new_ranges, $new_range);
    }

    print STDERR "Dumping ranges: " . Dumper (@new_ranges) . "\n";

    # create new feature
    my $new_feature = $feature->clone (-ranges => \@new_ranges);
    push (@new_features, $new_feature);
  }
  
  return @new_features;
}


sub getIndex {
  my ($position, %contigs) = @_;
  my @contig_indexes = keys (%contigs);

  for ($i=0; $i<@contig_indexes; $i++) {
    my $range = $contigs{$i};
    my $start = $range->start;
    my $end = $range->end;
    if (($position >= $start) && ($position <= $end)) {
      if ($debug) {
	print STDERR "position, $position, found!\n";
      }
      return $i;
    }
  }

  if ($debug) {
    print STDERR "can't find index for position, $position!!\n";
    print STDERR "returned index, $i\n";
  }
  # exit 1;
  return $i;
}
