#!/usr/local/bin/perl5.6.1

=head1 NAME

signalp2embl - Generate EMBL annotation of signal sequences predicted
by SignalP2.0

=head1 SYNOPSIS

Examples:

 signalp2embl -t 11 -y gram- embl_file
 signalp2embl -t 1 -y euk -c 0 embl_file

The misc_features are written to STDOUT.

=head1 DESCRIPTION

This program requires an EMBL file containing annotated CDS
features. The features will be translated using the specified
translation table (default is table 1; standard), except for those
with /pseudo, /partial or /transl_except qualifiers.

The output is parsed by the script to produce an misc_feature for each
SignalP prediction on STDOUT.

Setting the -y switch (to euk, gram+ or gram- is important, as is
choosing the correct translation table with the -t switch. The -c
switch sets the colour of the features (using the Artemis colour
codes). Set the -p switch to the threshold probability (0-1.0) below
which you do not want to report predictions.

Don't forget to cite the authors if you use this:

"Identification of prokaryotic and eukaryotic signal peptides and
prediction of their cleavage sites", Henrik Nielsen, Jacob
Engelbrecht, Soeren Brunak and Gunnar von Heijne Protein Engineering,
vol. 10, pages 1-6, 1997.

"Prediction of signal peptides and signal anchors by a hidden Markov
model", Henrik Nielsen and Anders Krogh In: Proceedings of the Sixth
International Conference on Intelligent Sys- tems for Molecular
Biology (ISMB 6), AAAI Press, Menlo Park, California, pages 122-130,
1998.

=head1 METHODS

None

=head1 AUTHOR

Keith James (kdj@sanger.ac.uk)

=head1 COPYRIGHT

Copyright (C) 2001 Genome Research Limited. All Rights Reserved.

decodeanhmm, version 1.1e Copyright (C) 1998-1999 by Anders Krogh. All
rights reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind. It may
be used, redistributed and/or modified under the same conditions as
Perl itself.

=cut

use strict;
use File::Temp qw/tempfile/;
use Getopt::Std;

use Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper;
use Bio::PSU::SeqFactory;
use Bio::PSU::FeatureFactory;
use Bio::Tools::Signalp;

use Bio::PSU::Utils::PlasmoAP11;

my $_debug = 0;

# Configuration of SignalP variables
my $bindir       = "/nfs/disk222/yeastpub/bio-soft/signalp-2.0";
my $signalp_bin  = "$bindir/signalp";
my $signalp_args = "-trunc 70 -f summary -m hmm";

my $unknown_id = "Unknown_CDS";

my $filter = 0;

my (%opts, $annot, $apicoAP, $colour, $prob, $regex, $table, $type);
my $usage = "Usage: signalp2embl [-a] [-s] [-c <colour>] [-p prob] [-t <table>] [-f] -y <type> <signalp options>

\t-a annotate original CDS with results
\t-s predict Apicoplast targetting peptide
\t-c Colour code of features  (default  0)
\t-p Cutoff probability       (default 0.6)
\t-r specify regex used to recognise a systematic gene name
\t-t Translation table number (default 1; standard numbering)
\t-y Organism type            (gram+, gram-, euk)
\t-f act as a filter instead of writing just the new features

\tTables available are:

\t 1) Standard
\t 2) Vertebrate Mitochondrial
\t 3) Yeast Mitochondrial
\t11) Bacterial
";

# Process options
getopts('afsc:hp:r:t:y:', \%opts);

defined $opts{h} and die $usage;
defined $opts{f} and $filter++;
defined $opts{a} and $annot++;
defined $opts{s} and $apicoAP++;
defined $opts{c} and $colour  = $opts{c};
defined $opts{p} and $prob    = $opts{p};
defined $opts{r} and $regex   = $opts{r};
defined $opts{t} and $table   = $opts{t};
defined $opts{y} and $type    = $opts{y};

# Default colour is white
$colour = 0 unless defined $colour;

# Default cutoff is 0.6
$prob = 0.6 unless defined $prob;

# Default translation table is universal (1)
$table ||= '1';

unless ($type =~ /^(gram\+$|gram-|euk)/) { die $usage }
$signalp_args .= " -t $type";

my $file;

if (!$filter) {
  $file = shift;
  unless (defined $file) { die $usage }
}

my $internal_id_counter = "000000";

my $id_helper =
  Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper->new(-selection_regex => $regex, -verbose => $_debug);

# Read EMBL stream and get entry
my $es;
my $eo;

if ($filter) {
  $es = Bio::PSU::SeqFactory->make(-fh => \*STDIN, -format => 'embl');
  $eo = Bio::PSU::SeqFactory->make(-fh => \*STDOUT, -format => 'embl');
} else {
  $es = Bio::PSU::SeqFactory->make(-file => "<$file");
}

# Handler for printing features as they appear, rather than printing a
# big feature table at the end
my $fth = Bio::PSU::IO::FTHandler->new;

while (my $entry = $es->next_seq)
{
  my %id_to_cds;

  foreach my $f ($entry->features)
  {
    # Discard non-CDS features
    next unless $f->key eq 'CDS';

    # Discard any CDS with /pseudo or /partial qualifier
    next if $f->qexists('partial');
    next if $f->qexists('pseudo');

    my ($fh, $tempfile) = tempfile("SIGNALP.XXXXXX", UNLINK => 1);
    close($fh);

    # Create Fasta output stream
    my $fo = Bio::PSU::SeqFactory->make(-file   => ">>$tempfile",
                                        -format => 'fasta');

    my $id;

    eval {
      $id = $id_helper->get_systematic_id($f);
    };

    if (!defined $id) {
      $id = ("signalp2embl_unknown_" . $internal_id_counter++ . "_" .
             $f->start . "_" . $f->end);
      warn "feature at " . $f->start . ".." . $f->end .
           " has no gene name\n";
    }

    if ($_debug) {
      print STDERR "Processing CDS, $id...\n";
    }

    $id_to_cds{$id} = $f;

    # Translate with specified translation table
    my $frame = 0;

    if ($f->qexists('codon_start'))
    {
      my @codon_start = $f->qvalues('codon_start');
      $frame = $codon_start[0] - 1;
    }

    my $f_pep = $f->translate($table, $frame);
    my $f_pep_str = $f_pep->str;

    $f_pep_str =~ s/\*$//;

    $f_pep->str ($f_pep_str);

    # Translation exceptions need to be skipped for now
    if (grep /\*/, $f_pep_str)
    {
      if ($f->qexists('transl_except'))
      {
        next;
      }
      else
      {
        warn "Stop codon present in ID $id. Skipping this CDS\n";
        next;
      }
    }

    # Set the id for the Fasta header
    $f_pep->id($id);
    $f_pep->desc(join(':', $f->start, $f->end, $f->strand));

    # Write translation to temp file as Fasta
    $fo->write_seq($f_pep);

    open(SIGP, "$signalp_bin $signalp_args $tempfile |") or
    die "Unable to open pipe from $signalp_bin: $!\n";

    my ($pred, $sprob, $aprob, $cprob, $cpos1, $cpos2, $raw);
    while (<SIGP>)
    {
      $raw .= $_;

      next if /^>/;
      next if /^\*/;
      chomp;

      if (/^Prediction:\s+(.*)/)
      {
        $pred = $1;
        next;
      }

      if (/^Signal peptide probability:\s+(\S+)/)
      {
        $sprob = $1;
        next;
      }

      if (/^Signal anchor probability:\s+(\S+)/)
      {
        $aprob = $1;
        next;
      }

      if (/^Max cleavage site probability:\s+(\S+)\s+between pos\.\s+(\S+)\s+and\s+(\S+)/)
      {
        ($cprob, $cpos1, $cpos2) = ($1, $2, $3);
        next;
      }
    }

    close(SIGP);
    unlink "$tempfile";

    unless (defined $pred  &&
            defined $sprob &&
            defined $cprob &&
            defined $cpos1 &&
            defined $cpos2)
    {
      warn "Unable to parse SignalP result. Offending SignalP output:\n", $raw;
      next;
    }

    next if $pred eq 'Non-secretory protein';

    if ($type eq 'euk' && (! defined $aprob))
    {
      warn "Unable to parse SignalP result (anchor probability). " .
           "Offending SignalP output:\n", $raw;
      next;
    }

    my $note = "";

    if ($type eq 'euk')
    {
      next if (($sprob < $prob) && ($aprob < $prob));

      $note = sprintf("%s predicted for %s by SignalP 2.0 HMM (Signal peptide probabilty %1.3f, signal anchor probability %1.3f) with cleavage site probability %1.3f between residues %d and %d",
                      $pred,
                      $id,
                      $sprob,
                      $aprob,
                      $cprob,
                      $cpos1,
                      $cpos2);
    }
    else
    {
      next if $sprob < $prob;

      $note = sprintf("%s predicted for %s by SignalP 2.0 HMM (Signal peptide probabilty %1.3f) with cleavage site probability %1.3f between residues %d and %d",
                      $pred,
                      $id,
                      $sprob,
                      $cprob,
                      $cpos1,
                      $cpos2);
    }

    my $cds = $id_to_cds{$id};
    my ($cds_start, $cds_end) = ($cds->start, $cds->end);
    my $cds_strand = $cds->strand;

    my ($nuc_start, $nuc_end);

    if ($cds_strand == 1)
    {
      $nuc_start = $cds_start;
      $nuc_end   = $cds_start + (3 * $cpos1) - 1;
      #$nuc_end   = $cds_start + (3 * ($cpos1 - 1));
    }
    else
    {
      #$nuc_start = $cds_end - (3 * ($cpos1 + 1));
      $nuc_start = $cds_end - (3 * $cpos1) + 1;
      $nuc_end   = $cds_end;
    }

    if ($cpos1 < 1 or $cpos2 < 0)
    {
      print STDERR "Prediction was outside of CDS - skipping addition of feature\n"
    }
    else
    {
      my $misc = Bio::PSU::Feature->new(-key => 'misc_feature',
                                        -start => $nuc_start,
                                        -end => $nuc_end,
                                        -strand => $cds_strand);

      $misc->qadd('id', $id);
      $misc->qadd('note', $note);
      $misc->qadd('colour', $colour);

      $misc->qadd('type', "signalp");
      $misc->qadd('algorithm', "SignalP 2.0 HMM");
      $misc->qadd('coord', "$cpos1,$cpos2");
      $misc->qadd('cleavage', $cprob);
      $misc->qadd('anchor', $aprob);
      $misc->qadd('signal', $sprob);

      if ($filter) {
        $entry->features($misc);
      } else {
        my $text = $fth->make_feature_block($misc);
        print $text;
      }
    }

    # Added facility for annotation of original features
    if ($annot)
    {
      $f->qadd('note', $note);
    }

    #########################################
    #
    # PlasmoAP
    #
    #########################################

    if ($apicoAP) {

      # 1/ Get rid of the signal peptide
      # From 1 to $cpos1

      if ($cpos1 <= 0) {
        print STDERR "can't deal with this feature!!\n";
        next;
      }


      my $subpeptide_seqobj = substr $f_pep_str, $cpos2 - 1;

      # 2/ Run ApicoAP


      my $plasmoAP = Bio::PSU::Utils::PlasmoAP11->new (seqobj => $subpeptide_seqobj);
      my $result = $plasmoAP->analysis ();

      # print STDERR "systematic id: $id\n";
      print STDERR "result: $result\n";

      # 3/ If positive, store the result in a note qualifier

      # if + or ++

      if ($result =~ /\+/) {
        $note = "Apicoplast targetting peptide predicted by the PlasmoAP tool";
        $f->qadd('note', $note);
      }
    }
  }

  if ($filter) {
    $eo->write_seq($entry);
  }
}
