#!/usr/local/bin/perl5.6.1 -w

=head1 NAME

sim42tab.pl - Generate EMBL annotation of exons predicted
by sim4

=head1 SYNOPSIS

Examples:

 sim42tab.pl -i sim4.res -o seq.sim4.embl
 sim42tab.pl -c cDNAs.fa -g gDNA.fa

The EMBL entry is written to STDOUT per default.

=head1 DESCRIPTION

This program requires either a sim4 output file or a cDNA database and a genomic sequence. In that case it will run sim4 for you.
Then it is parsing the sim4 output, using Bioperl, and generates a set of exon features. 
Not the all set of sim4 predicted exons are reported, there is a filtering step which can be parametered using -e, -p and -l.
Then there is an extra step which generates a set of mRNA features. When possible, it joins exons one to another. 
The conditions are : 
* First when they are predicted from the same cDNA sequence, 
* Second the alignment coordinates need to be contiguous (e.g. 1..211, 212..456)

The script produces an EMBL entry on STDOUT.
Set -o option to produce an EMBL entry file instead.

=head1 METHODS

None

=head1 AUTHOR

Arnaud Kerhornou (axk@sanger.ac.uk)

=head1 COPYRIGHT

Copyright (C) 2003 Genome Research Limited. All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind. It may
be used, redistributed and/or modified under the same conditions as
Perl itself.

=cut

use strict;
use Data::Dumper;
use Getopt::Std;

use IO::File;

# Bioperl

# Use of a modified version of Bioperl Results.pm module

use Bio::Tools::Sim4::Results;
# some modifications in the code to make the parsing working
# was stopping when parsing a result without any hit !
use Bio::PSU::Utils::Sim4Results;

use Bio::SeqIO;
use Bio::Seq;
use Bio::SeqFeature::Generic;
use Bio::Location::Simple;
use Bio::Location::Split;
use Bio::Location::Fuzzy;

##
# Get the EST description from the sim4 output 
# Bioperl Parsing doesn't get it - Bioperl bug to fix!!
##

my $debug = 0;

# %
my $_EST_coverage   = 40;
# %
my $_Hit_percentage = 90;
# bp
my $_EST_length     = 50;

my $sim4_args = "P=1 A=4";

my (%opts, $cdna, $bac, $sim4_file, $tab_file);
getopt('h:i:c:g:o:e:p:l', \%opts);

if (defined $opts{"h"}) {
  display_help ();
  exit 1;
}

if (not ((defined $opts{"c"} and defined $opts{"g"}) or defined $opts{i})) {
  print STDERR "\nSpecify a sim4 output or the name of an ESTs database and genomic sequence to run sim4\n\n";
  display_help ();
  exit 1;
}

defined $opts{"c"} and $cdna = $opts{"c"};
defined $opts{"g"} and $bac  = $opts{"g"};
defined $opts{"i"} and $sim4_file  = $opts{"i"};
if (not (defined $opts{"o"} and $tab_file = $opts{"o"})) {
  print STDERR "\nno output specified\n\n";
  display_help ();
  exit 1;
}

defined $opts{"e"} and $_EST_coverage   = $opts{"e"};
defined $opts{"p"} and $_Hit_percentage = $opts{"p"};
defined $opts{"l"} and $_EST_length     = $opts{"l"};

if ($debug) {
  print STDERR "Filetring options:\n";
  print STDERR "\tEST coverage percentage: $_EST_coverage\n";
  print STDERR "\tHit percentage identity: $_Hit_percentage\n";
  print STDERR "\tMinimum EST length: $_EST_length\n";
}

my @features  = ();
my $seqId = "";
my $sembl = Bio::SeqIO->new(
			    '-file'   => ">$tab_file",
			    '-format' => 'embl'
			   );

my $sim4_parser;
my $tmp_fh;

if (defined $sim4_file) {

  if ($debug) {
    print STDERR "sim4 output file, $sim4_file\n";
  }

  $sim4_parser = Bio::PSU::Utils::Sim4Results->new(
						   -file => $sim4_file,
						  );
}
else {
  
  if ($debug) {
    print STDERR "Running sim4...\n";
    print STDERR "cDNA file, $cdna\n";
    print STDERR "gDNA file, $bac\n";
  }

  my @sim4 = qx/sim4 $bac $cdna $sim4_args/;

  if ($debug) {
    print STDERR "sim4 run done\n";
  }

  $tmp_fh = IO::File->new_tmpfile() or die "can't create a temporary file, $!\n";
  foreach my $line (@sim4) {
    print $tmp_fh "$line";
  }
  seek ($tmp_fh, 0, 0);

  $sim4_parser = Bio::PSU::Utils::Sim4Results->new(
						   -fh => $tmp_fh,
						  );
}

if ($debug) {
  print STDERR "parsing sim4 output...\n";
}

my %exons_Per_cDNA;

##
# parse the results
##
while (my $exonset = $sim4_parser->next_exonset()) {

  ##
  # gDNA Data
  ##

  $seqId = $exonset->seq_id();
  my $seq_start = $exonset->start();
  my $seq_end   = $exonset->end();

  if ($debug) {
    print STDERR "gDNA Id: $seqId\n";
  }
  
  foreach my $exon ( $exonset->sub_SeqFeature() ) {
    
    ##
    # EST Hit Data
    ##

    my $homol = $exon->est_hit();

    my $EST_seqId      = $homol->seq_id;
    # can't get  it !!!!
    #my $EST_seqDesc    = $homol->seqdesc;
    my $EST_seqLength  = $homol->seqlength;
    my $Hit_start      = $homol->start;
    my $Hit_end        = $homol->end;
    my $Hit_length     = $Hit_end - $Hit_start + 1;

    my $percentage_id = $exon->percentage_id;

    if ($debug) {
      print STDERR "EST Id: $EST_seqId\n";
      # print STDERR "EST Description: $EST_seqDesc\n";
      print STDERR "EST length: $EST_seqLength\n";
      print STDERR "Hit length: $Hit_length\n";
    }

    ################################
    #
    # Filtering
    #
    ################################

    if ($percentage_id < $_Hit_percentage) {
      if ($debug) {
	print STDERR "percentage Id, $percentage_id, too low!!\n";
      }
      next;
    }
    
    if ($EST_seqLength < $_EST_length) {
      if ($debug) {
	print STDERR "filtering hit, EST too small\n";
      }
      next;
    }
    
    my $coverage = 100 * $Hit_length / $EST_seqLength;
    if ($coverage < $_EST_coverage) {
      if ($debug) {
	print STDERR "filtering hit, low coverage, $coverage\n";
      }
      next;
    }

    ################################

    ################################
    # 
    # Set more information in the Bioperl exon object
    #
    ################################
 
    my $note = "exon feature predicted by sim4";
    $exon->add_tag_value ('note', $note);
    $note = "Matched to sequence " . $EST_seqId . " at " . $Hit_start . " to " . $Hit_end;
    $exon->add_tag_value ('note', $note);
    $exon->add_tag_value ('percent_id', $percentage_id);

    ################################

    push (@features, $exon);

    ################################
    #
    # Fill %exons_Per_cDNA
    #
    ################################

    if ($Hit_start > $Hit_end) {
      # commute the coordinates
      my $tmp    = $Hit_start;
      $Hit_start = $Hit_end;
      $Hit_end   = $tmp;
    }
    # my $exon_note = "$Hit_start..$Hit_end";
    my $cDNA_location = Bio::Location::Simple->new (
						    -start  => $Hit_start,
						    -end    => $Hit_end,
						    -strand => 1
						   );
    my $gDNA_location = Bio::Location::Fuzzy->new (
						    -start  => $exon->start(),
						    -end    => $exon->end(),
						    -strand => $exon->strand(),
						   );
    my $exon_hash = {
		     -gDNAlocation => $gDNA_location,
		     -cDNAlocation => $cDNA_location,
		    };
    if (exists ($exons_Per_cDNA{$EST_seqId})) {
      my $exon_hashes = $exons_Per_cDNA{$EST_seqId};
      push (@$exon_hashes, $exon_hash);
      $exons_Per_cDNA{$EST_seqId} = $exon_hashes;
    }
    else {
      $exons_Per_cDNA{$EST_seqId} = [$exon_hash];
    }

  }
  
}

$sim4_parser->close();
undef $tmp_fh;

if ($debug) {
  print STDERR "parsing done.\n";
}

################################
#
# Process %exons_Per_cDNA to join exons together
# When they share the same EST and when the hits coordinates fit
# e.g (1..211),(212..467)
#
################################

my @ESTs = keys (%exons_Per_cDNA);
foreach my $EST (@ESTs) {
  
  if ($debug) {
    print STDERR "processing merging for EST, $EST...\n";
  }

  my $exon_hashes = $exons_Per_cDNA{$EST};
  if (@$exon_hashes > 1) {
    my @joined_ranges = ();
    my @joined_notes  = ();
    my $exon_hash1    = $exon_hashes->[0];
    
    my $i = 1;
    while ($i < @$exon_hashes) {
      my $exon_hash2 = $exon_hashes->[$i];
      
      my $range1 = $exon_hash1->{-cDNAlocation};
      my $range2 = $exon_hash2->{-cDNAlocation};
      
      if ($range2->start < $range1->start) {
	
	if ($debug) {
	  print STDERR "permuting exons...\n";
	}
	
	# permute them !
	my $tmp = $range1;
	$range1 = $range2;
	$range2 = $tmp;
      }
      
      my $start1 = $range1->start;
      my $end1   = $range1->end;
      my $start2 = $range2->start;
      my $end2   = $range2->end;
      
      if ($debug) {
	print STDERR "start1: $start1\n";
	print STDERR "end1: $end1\n";
	print STDERR "start2: $start2\n";
	print STDERR "end2: $end2\n";
      }
      
      if ($end1 == ($start2-1)) {
	# join them
	
	if ($debug) {
	  print STDERR "join the two exons...\n";
	}
	
	if (@joined_ranges == 0) {
	  # Frist range gets fuzzy start
	  my $gDNAlocation1 = $exon_hash1->{-gDNAlocation};
	  $gDNAlocation1->max_start ($gDNAlocation1->start);
	  $gDNAlocation1->min_start (undef);
	  $gDNAlocation1->start (undef);
	  $gDNAlocation1->start_pos_type ("BEFORE");
	  push (@joined_ranges, $gDNAlocation1);
	  my $exon_location1 = $exon_hash1->{-cDNAlocation};
	  my $exon_note1 = $exon_location1->start . ".." . $exon_location1->end;
	  push (@joined_notes, $exon_note1);
	}
	my $gDNAlocation2 = $exon_hash2->{-gDNAlocation};
	if (($i +1) == @$exon_hashes) {
	  # last range gets fuzzy end
	  $gDNAlocation2->min_end ($gDNAlocation2->end);
	  $gDNAlocation2->max_end (undef);
	  $gDNAlocation2->end_pos_type ("AFTER");
	  $gDNAlocation2->end (undef);
	}
	push (@joined_ranges, $gDNAlocation2);
	my $exon_location2 = $exon_hash2->{-cDNAlocation};
	my $exon_note2 = $exon_location2->start . ".." . $exon_location2->end;
	push (@joined_notes, $exon_note2);
      }
      $i++;
    }
    # done !!
    # actually the process should be more complex !!
    # to include gaps !!
    if (@joined_ranges > 1) {
      my $splitlocation = new Bio::Location::Split();
      $splitlocation->add_sub_Location(@joined_ranges);

      if ($debug) {
	print STDERR "Dumping splitlocation: " . Dumper ($splitlocation) . "\n";
      }
      
      my $cds_feature = Bio::SeqFeature::Generic->new (
						       -primary => 'mRNA',
						       -location => $splitlocation
						      );
      my $note = "mRNA feature predicted by sim4";
      $cds_feature->add_tag_value ('note', $note);
      $note = "matching cDNA sequence, $EST at (" . join (', ', @joined_notes) . ")";
      $cds_feature->add_tag_value ('note', $note);
      $cds_feature->add_tag_value ('partial', '');
      push (@features, $cds_feature);
    }
  }
}

if ($debug) {
  print STDERR "Writing sim4 results in tab file, $tab_file...\n";
}

my $seqobj = Bio::Seq->new (
			    -display_id => $seqId,
			    -desc       => "sim4 prediction",
			   );

if (@features > 0) {
  $seqobj->add_SeqFeature (@features);
}
$sembl->write_seq ($seqobj);

if ($debug) {
  print STDERR "file done.\n";
}

##
# end
##

sub display_help {
  my $usage = "Usage: sim42tab.pl [-h] [-c <cDNAs Database>] [-g <genomic DNA sequence>] [-i <sim4 output file>] [-o <tab output file>] [-e <EST coverage percentage id>] [-p <Hit percentage id>] [-l <Minimum EST length>]\n";
  print "$usage\n";
  print "\t-c the cDNA fasta database\n";
  print "\t-g the genomic sequence\n";
  print "\t-i the sim4 output - instead of c and g parameters\n";
  print "\t-o the sim4 tab file\n";
  print "\t-e EST coverage percentage id - Default is 40%\n";
  print "\t-p Hit percentage id - Default is 90%\n";
  print "\t-l Minimum EST length - Default is 50bp\n";
}
