#!/usr/local/bin/perl5.6.1 -w

=head1 NAME

sp2embl.pl - convert the superfamily output into an embl format file

=head1 SYNOPSIS

  sp2embl.pl -help -sp [superfamily output] -embl [embl file wih the CDSs features]

  -help a help message
  -sp the superfamily output file
  -embl the embl annotation file(s) - you can give a file or a directory

=head1 DESCRIPTION

  This script to convert superfamily output into embl.

  It generates a new embl file with superfamily predictions as CDS_DOMAIN feature. It involves a mapping from protein coordinates to DNA coordinates.
  The syntax is contig_name.SPs.tab

  It also generates a new set of embl annotation files. The superfamily predictions are added in there as domain qualifiers within the CDS features.
  The new embl files syntax is contig_name.with_sp_qualifiers.embl

  Involves unique gene names (in gene qualifiers at the moment) and unique contig names (extracted from the embl file names) to do a proper mapping.

  @author Arnaud Kerhornou - axk@sanger.ac.uk

=cut

=head1 ALGORITHM

=cut

##################################################################
##
# There is a bug in the mapping between protein coordinates and the DNA coordinates
# This mapping is done using bioperl but it needs to take into accoung which frame the CDS is => the codon_start qualifier needs to be taken into consideration
##
##################################################################


# External Modules

# Constraint variables declaration before using them
use strict;
use Cwd;
use Data::Dumper;

# Benchmark Module
use Benchmark;
##################################################################

# Bio modules
# Bioperl
use Bio::Location::Simple;
use Bio::SeqIO;
use Bio::Seq;
use Bio::SeqFeature::Generic;
use Bio::Coordinate::GeneMapper;

use Bio::Location::Simple;

# BioPSU
use Bio::PSU::IO::BufferFH;
use Bio::PSU::Utils::FeatureIdHelper::BioPerlHelper;
use IO::File;

##########
# The Superfamily Parser Module
##

 eval "require Bio::PSU::Utils::SuperfamilyParser";
# use Bio::PSU::Utils::SuperfamilyParser;

my $t1 = Benchmark->new ();

##########################################
#
# Verbose
#
##########################################

my $verbose = 0;

##########################################
#
# Loading Parameters
#
##########################################

my $_help        = "-help";
my $_superfamily = "-sp";
my $_embl        = "-embl";
my $_regex       = "-regex";

my $sp   = "-1";
my $embl = "-1";
my $regex = undef;

# Check if in the command line there is an input file name (-i option)
# the other options are not compulsory (set of options per default)
my $lgth = @ARGV + 0;
# if no args, display the help message
if ($lgth < 4) {
  display_help ();
  exit 1;
}
elsif ($lgth >= 4) {
  my $i = 0;
  foreach my $element (@ARGV) {
    if ($element eq $_help) {
      display_help ();
      exit 1;
    }
    elsif ($element eq $_superfamily) {
      $sp = $ARGV[$i+1];
    }
    elsif ($element eq $_embl) {
      $embl = $ARGV[$i+1];
    }
    elsif ($element eq $_regex) {
      $regex = $ARGV[$i+1];
    }
    # if at least one of the options is not recognized, display the help message and exit
    elsif ($element =~ /^-/) {
      print STDERR "option not recognized: $element\n";
      display_help ();
      exit 1;
    }
    $i ++;
  }
}
if (($sp eq "-1") || ($embl eq "-1")) {
  display_help ();
  exit 1;
}

my $id_helper =
  Bio::PSU::Utils::FeatureIdHelper::BioPerlHelper->new(-selection_regex => $regex);

###################################
# Generate an Array of embl files
# ($embl may be a directory)
##

my @embls = ();
my $is_file = 0;

if (-f $embl) {
  push (@embls, $embl);
  $is_file = 1;
}
else {
  
  opendir EMBLDIR, $embl or die "Can not read this directory, $embl!\n";
  @embls = grep {$_ =~ /.+\.embl$/i} readdir EMBLDIR;

  closedir EMBLDIR;
}

if ($verbose) {
  print STDERR "set of embl files: ". join (', ', @embls) . "\n";
}

##########################
# parsing embl file(s)
##

print STDERR "parsing embl file(s)...\n";

my %cds_features_per_gene_name;
my %seq_per_contig;

foreach my $embl_file (@embls) {

  if (not $is_file) {
    # if directory, add the input directory at the path file
    $embl_file = "$embl/$embl_file";
  }
  
  if ($verbose) {
    print STDERR "parsing embl file, $embl_file...\n";
  }
  
  my $sembl    = Bio::SeqIO->new('-file'   => $embl_file,
  				 '-format' => 'embl'
				);
  
  while (my $embl_seq_obj = $sembl->next_seq()) {
    my $embl_contig  = $embl_seq_obj->id();
    
    if ($verbose) {
      print STDERR "contig ID: $embl_contig\n";
      print STDERR "acc: " . $embl_seq_obj->accession_number() . "\n";
      print STDERR "display_id: " . $embl_seq_obj->display_id() . "\n";
      print STDERR "accession: " . $embl_seq_obj->accession() . "\n";
    }
    
    $seq_per_contig{$embl_contig} = $embl_seq_obj;
    
    foreach my $feature ($embl_seq_obj->all_SeqFeatures) {
      if ($feature->primary_tag eq "CDS") {
	my $id = $id_helper->get_systematic_id($feature);
	if (defined $id) {
	  my %feature_hash = (-contig => $embl_contig, -feature => $feature);
	  
	  if (not exists $cds_features_per_gene_name{$id}) {
	    $cds_features_per_gene_name{$id} = \%feature_hash;
	  }
	  else {
	    print STDERR "ERROR - identifier, $id, duplicated!!!\n";
	    print STDERR "Will not be able to map sp hit and related CDS!!\n";
	    if ($verbose) {
	      print STDERR "Dumping feature: " . Dumper ($feature) . "\n";
	    }
	  }
	}
	else {
	  print STDERR "ERROR - CDS feature with no identifier, will not be able to map sp hit and related CDS!!\n";
	  if ($verbose) {
	    print STDERR "Dumping feature: " . Dumper ($feature) . "\n";
	  }
	}
      }
    } # next feature
  } # next sequence entry

} # next embl file

print STDERR "embl file(s) parsing done!\n";

#########################
# get SP Predictions
##

print STDERR "parsing SP assignments, $sp...\n";

my $sp_parser = Bio::PSU::Utils::SuperfamilyParser->new (
							 superfamilyOutput => $sp,
							 verbose => $verbose,
							);
my @hits = $sp_parser->parse ();

print STDERR "sp hits parsing done, found " . @hits . "\n";
# print STDERR "found " . @hits . " hits. " . Dumper (@hits) . "\n";;

########################
# Generating SUPERFAMILY FEATURE OBJECTS AS CDS_DOMAINS
##

print STDERR "Generating superfamily feature objects...\n";

my %sp_fts_per_contig;
my @new_features = ();

foreach my $sp_hit (@hits) {
  
  if ($verbose) {
    print STDERR "Processing superfamily hit for protein, " . $sp_hit->getProtAccession() . "...\n";
  }
  
  # SP hit data

  my $prot_name = $sp_hit->getProtAccession();
  my $start     = $sp_hit->getStart();
  my $end       = $sp_hit->getEnd();
  my $evalue    = $sp_hit->getEvalue();
  my $scop_id   = $sp_hit->getScopID();
  my $scop_name = $sp_hit->getScopName();
  
  if (exists $cds_features_per_gene_name{$prot_name}) {
    
    # corresponding CDS Feature data
    
    my %cds_feature_hash = %{$cds_features_per_gene_name{$prot_name}};
    my $cds_feature = $cds_feature_hash{-feature};
    my $contig_name = $cds_feature_hash{-contig};
    
    if ($verbose) {
      print STDERR "Mapping protein and DNA coordinates of the superfamily hits...\n";
    }
    
    # Mapping Protein coordinates to DNA coordinates of the SP domain
    # Using Bioperl
    
    my $gmap = Bio::Coordinate::GeneMapper->new (
						 -in  => 'peptide',
						 -out => 'chr',
						 -strict => 1,
						 -nozero => 0
						);
    $gmap->cds($cds_feature);
    my $exon_location_object = $cds_feature->location;
    if (ref ($exon_location_object) =~ /Simple/) {
      $gmap->exons ($exon_location_object);
    }
    elsif (ref ($exon_location_object) =~ /Split/) {
      my @exon_locations = $exon_location_object->sub_Location();
      # don't need it !!!
      # my $first_exon_location = $exon_locations[0];
      $gmap->exons (@exon_locations);
    }
    else {
      print STDERR "ERROR - bioperl location class, " . ref ($exon_location_object) . " not recognized !!!\n";
      exit 1;
    }
    
    my $sp_location = Bio::Location::Simple->new (
						  -start => $start,
						  -end   => $end,
						  -strand => 1
						 );
    my $note = "SUPERFAMILY domain match for $prot_name\t(P=$evalue)\tto SCOP entry, $scop_id $scop_name from $start to $end";
    my $protein_feature_protein = create_sp_ft_bioperl ($note, $sp_location);
    my $new_location = $gmap->map($protein_feature_protein);

    $protein_feature_protein->location ($new_location);

    if ($verbose) {
      print STDERR "SP hit coordinates at the protein level : " . Dumper ($sp_location) . "\n";
      print STDERR "mapped SP domain coordinates into the genomic sequence: " . Dumper ($new_location) . "\n";
    }

    # add the new SP feature object to the corresponding contig array

    if (exists $sp_fts_per_contig{$contig_name}) {
      my @sp_fts = @{$sp_fts_per_contig{$contig_name}};
      push (@sp_fts, $protein_feature_protein);
      $sp_fts_per_contig{$contig_name} = \@sp_fts;
    }
    else {
      my @sp_fts = ();
      push (@sp_fts, $protein_feature_protein);
      $sp_fts_per_contig{$contig_name} = \@sp_fts;
    }

    # Generate the domain qualifier

    my $sp_qualifier_value = "SCOP:$scop_id; $scop_name; $evalue; codon $start-$end";
    $cds_feature->add_tag_value ('domain', $sp_qualifier_value);

    $cds_feature_hash{-new_feature} = $cds_feature;
    $cds_features_per_gene_name{$prot_name} = \%cds_feature_hash;

  }
  else {
    if ($verbose) {
      print STDERR "ERROR - can't find the corresponding CDS for SP hit which protein name is, $prot_name!!\n";
    }
  }
} # next SP hit

print STDERR "SP Features creation done.\n";
  
# Create new tab file(s), one per contig with new superfamily feature objects

print STDERR "tab file(s) generation...\n";

my @contig_names_with_sps = keys (%sp_fts_per_contig);

foreach my $contig_name (@contig_names_with_sps) {
  if ($verbose) {
    print STDERR "Adding new CDS_domain features for contig, $contig_name\n";
  }

  my @sp_features = @{$sp_fts_per_contig{$contig_name}};
  my $tab_name = "$contig_name.SPs.tab";

  # generate new tab file
  
  if ($verbose) {
    print STDERR "generating new tab file with the SP assignment features...\n";
  }
  
  my $new_embl_seq_obj = Bio::Seq->new(
				       -display_id => $contig_name,
				       -desc => "Superfamily domain features annotations file for contig, $contig_name",
				      );
  
  $new_embl_seq_obj->add_SeqFeature (@sp_features);
  
  if ($verbose) {
    print STDERR "sequence object done!\n";
  }
  
  my $sembl_new = Bio::SeqIO->new('-file'   => ">$tab_name",
				  '-format' => 'embl');
  $sembl_new->write_seq($new_embl_seq_obj);
  
} # next contig

# Generating a new CDS EMBL file with domain qualifiers

print STDERR "CDS EMBL tab file generation...\n";

my @all_contig_names = keys (%seq_per_contig);

if ($verbose) {
  # print STDERR "all contig names: @all_contig_names\n";
  print STDERR "nb contigs: " . @all_contig_names ."\n";
}

foreach my $contig_name (@all_contig_names) {
  my $seq_obj = $seq_per_contig{$contig_name};
  my $seq_str = $seq_obj->seq;
  my @new_features = ();

  foreach my $feature ($seq_obj->all_SeqFeatures) {
    if ($feature->primary_tag eq "CDS") {

      my $systematic_id = undef;
      eval {
        $systematic_id = $id_helper->get_systematic_id($feature);
      };

      if ($@) {
        my $loc = join ",", map {
          $_->{start} . ".." . $_->{end}
        } $feature->ranges ();
	
        print STDERR "Error in CDS at: $loc -\n", $@;
        next;
      }

      my %cds_feature_hash = %{$cds_features_per_gene_name{$systematic_id}};

      if ($verbose) {
	print STDERR "Dumping cds_feature_hash: " . Dumper (%cds_feature_hash) . "\n";
      }

      if (exists $cds_feature_hash{-new_feature}){
	push (@new_features, $cds_feature_hash{-new_feature});
      }
      else {
	push (@new_features, $feature);
      }
    }
    else {
      push (@new_features, $feature);
    }
  }

  # create new sequence object

  my $new_seq_obj = Bio::Seq->new ( -seq        => $seq_str,
				    -display_id => $contig_name,
				    -alphabet   => 'dna',
				  );
  $new_seq_obj->add_SeqFeature (@new_features);
  
  my $new_seq_factory = Bio::SeqIO->new( -format => 'embl', 
					 -file   => ">$contig_name.with_sp_qualifiers.embl",
				       );
  $new_seq_factory->write_seq ($new_seq_obj);
  
}

print STDERR "process done.\n";

my $t2 = Benchmark->new ();
print  STDERR "\nTotal : ", timestr (timediff ($t2, $t1)), "\n";

###################################################################
# The end
###################################################################

# display the help message

sub display_help {
  print "sp2embl.pl -help -sp [superfamily file] -embl [DNA sequence embl file or directory]\n\n";
  print "\t-help a help message\n";
  print "\t-sp the superfamily prediction file - (no default, always specify it)\n";
  print "\t-embl the DNA sequence input file or directory - (no default, always specify it)\n";
  print "\t-regex a regular expression used to determine the CDS identifier - (optional))\n\n"
}

sub create_sp_ft_psu {
  my ($note, @sp_ranges) = @_;
  
  my $sp_feature_name = "CDS_domain";
  my $colour = 7;
  
  my $feature = Bio::PSU::Feature->new(
				       -key    => $sp_feature_name,
				       -ranges => \@sp_ranges,
				      );
  $feature->qadd('note', $note);
  $feature->qadd('colour', $colour);
  return $feature;
}

sub create_sp_ft_bioperl {
  my ($note, $sp_location) = @_;
  
  my $sp_feature_name = "CDS_domain";
  my $colour = 7;

  my $feature = Bio::SeqFeature::Generic->new (
					       -primary => $sp_feature_name,
					       -tag    => {
							   colour => $colour,
							   note => $note
							  } 
					      );
  $feature->location ($sp_location);

  return $feature;
}
