#!/usr/local/bin/perl5.6.1 -w

=head1 NAME

tmhmm2embl - Generate EMBL annotation of transmembrane helices
predicted by TMHMM2.0

=head1 SYNOPSIS

Examples:

 tmhmm2embl -t 11 embl_file
 tmhmm2embl -c 4 -t 1 embl_file

The EMBL entry is written to STDOUT.

=head1 DESCRIPTION

This program requires an EMBL file containing annotated CDS
features. The features will be translated using the specified
translation table (default is table 1; universal), except for those
with /pseudo, /partial or /transl_except qualifiers. TMHMM2.0 will be
run on these to produce at output file with a '.tmhmm' extension.

The output file is parsed by the script to produce an EMBL table on
STDOUT.

The -c switch sets the colour of the features (using the Artemis
colour codes) and the -v switch will provide verbose annotation of all
the helices.

The -s switch will create a separate feature for each predicted
transmembrane helix.

This script requires the Bio::PSU Perl modules see:

http://www.sanger.ac.uk/Users/kdj/software.html

Don't forget to cite Anders Krogh if you use this:

TMHMM is described in

Erik L.L. Sonnhammer, Gunnar von Heijne, and Anders Krogh:
A hidden Markov model for predicting transmembrane helices in protein
sequences.
In J. Glasgow et al., eds.: Proc. Sixth Int. Conf. on Intelligent
Systems for Molecular Biology, pages 175-182. AAAI Press, 1998.

=head1 METHODS

None

=head1 AUTHOR

Keith James (kdj@sanger.ac.uk)

=head1 COPYRIGHT

Copyright (C) 2000 Genome Research Limited. All Rights Reserved.

decodeanhmm, version 1.1e Copyright (C) 1998-1999 by Anders Krogh. All
rights reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind. It may
be used, redistributed and/or modified under the same conditions as
Perl itself.

=cut

use strict;
use File::Temp qw/tempfile/;
use Getopt::Std;

use Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper;
use Bio::PSU::SeqFactory;
use Bio::Tools::Tmhmm;

# Configuration of TMHMM variables
my $basedir = "/nfs/disk222/yeastpub/bio-soft/TMHMM2.0";
my $bindir = "/nfs/disk222/yeastpub/bio-soft/TMHMM2.0/bin";

my $tmhmm_bin = "$bindir/tmhmm";

my $internal_id_counter = "000000";

my $filter = 0;

my (%opts, $annot, $colour, $load, $regex, $split, $table, $verbose);
my $usage = "Usage:
  tmhmm2embl [-t <table>] [-r <regex>] [-c <colour>] [-av] file
or 
  tmhmm2embl -f [-t <table>] [-r <regex>] [-c <colour>] [-av] < old_file.embl

\t-a annotate CDS feature with new qualifier
\t-t specify translation table (default is 1; standard numbering)
\t-r specify regex used to recognise a systematic gene name
\t-c specify which Artemis colour code to use (default 0)
\t-v verbose annotation (reports coordinates of all helices)
\t-f act as a filter instead of writing just the new features

";

# Process options
getopts('ac:fhr:t:v', \%opts);

defined $opts{h} and die $usage;
defined $opts{a} and $annot++;
defined $opts{c} and $colour  = $opts{c};
defined $opts{f} and $filter++;
defined $opts{r} and $regex   = $opts{r};
defined $opts{t} and $table   = $opts{t};
defined $opts{v} and $verbose = $opts{v};

# Default colour is white
$colour = 0 unless defined $colour;

# Default translation table is universal (1)
$table ||= '1';

my $file;

if (!$filter) {
  $file = shift;
  unless (defined $file) { die $usage }
}

my $id_helper =
  Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper->new(-selection_regex => $regex);

my $es;
my $eo;

if ($filter) {
  $es = Bio::PSU::SeqFactory->make(-fh => \*STDIN, -format => 'embl');
  $eo = Bio::PSU::SeqFactory->make(-fh => \*STDOUT, -format => 'embl');
} else {
  $es = Bio::PSU::SeqFactory->make(-file => "<$file");
}

# Handler for printing features as they appear
my $fth = Bio::PSU::IO::FTHandler->new;

while (my $entry = $es->next_seq)
{
    my %id_to_cds;

    my ($fh, $tempfile) = tempfile("TMHMM.XXXXXX");
    close($fh);

    # Create Fasta output stream
    my $fo = Bio::PSU::SeqFactory->make(-file   => ">>$tempfile",
                                        -format => 'fasta');

    foreach my $f ($entry->features)
    {
        # Discard non-CDS features
        next unless $f->key eq 'CDS';

        # Discard any CDS with /pseudo or /partial qualifier
        next if $f->qexists('partial');
        next if $f->qexists('pseudo');

        my $id;
        eval {
          $id = $id_helper->get_systematic_id($f);
        };

        if (!defined $id) {
          $id = ("tmhmm2embl_unknown_" . $internal_id_counter++ . "_" .
                 $f->start . "_" . $f->end);
          warn "feature at " . $f->start . ".." . $f->end .
            " has no gene name\n";
        }

        $id_to_cds{$id} = $f;

        # Translate with specified translation table
        my $frame = 0;

        if ($f->qexists('codon_start'))
        {
            my @codon_start = $f->qvalues('codon_start');
            $frame = $codon_start[0] - 1;
        }

        my $f_pep = $f->translate($table, $frame);

        my $f_pep_str = $f_pep->str;
        $f_pep_str =~ s/\*$//;

        $f_pep->str ($f_pep_str);

        # Translation exceptions need to be skipped for now
        if (grep /\*/, $f_pep_str)
        {
            if ($f->qexists('transl_except'))
            {
                next;
            }
            else
            {
                warn "Stop codon present in ID $id. Skipping this CDS\n";
                next;
            }
        }

        # Set the id for the Fasta header
        $f_pep->id($id);
        $f_pep->desc(join(':', $f->start, $f->end, $f->strand));

        # Write translation to temp file as Fasta
        $fo->write_seq($f_pep);
    }

    # Run TMHMM
    open(TMHMM, "$tmhmm_bin $tempfile 2> /dev/null |")
        or die "Unable to open pipe from $tmhmm_bin: $!\n";

    my $parser = Bio::Tools::Tmhmm->new(-fh => \*TMHMM);

    my %id_to_coords;

    while (my $tm_helix = $parser->next_result)
    {
        my $id = $tm_helix->seq_id;

        if (! exists $id_to_coords{$id})
        {
            $id_to_coords{$id} = [];
        }

        # Hash of IDs to lists of lists
        push(@{$id_to_coords{$id}}, [$tm_helix->start, $tm_helix->end]);
    }

    foreach my $id (keys %id_to_coords)
    {
        my $cds = $id_to_cds{$id};
        my ($cds_start, $cds_end) = ($cds->start, $cds->end);
        my $cds_strand = $cds->strand;

        my @coord_qualifiers = ();

        my @coords = @{$id_to_coords{$id}};

        my $helixcount = scalar @coords;
        my $helix      = $helixcount > 1 ? "helices" : "helix";

        my $note = "$helixcount probable transmembrane $helix predicted for $id by TMHMM2.0";

        my @locs;

        foreach my $coord (@coords)
        {
            my ($tmh_start, $tmh_end) = @$coord;
            my ($nuc_start, $nuc_end);

            if ($cds_strand == 1)
            {
                $nuc_start = $cds_start + (3 * ($tmh_start - 1));
                $nuc_end   = $cds_start + (3 *  $tmh_end)  - 1;
            }
            else
            {
                $nuc_start = $cds_end - (3 *  $tmh_end)  + 1;
                $nuc_end   = $cds_end - (3 * ($tmh_start - 1));
            }

            push @coord_qualifiers, "$tmh_start,$tmh_end";

            push(@locs, Bio::PSU::Range->new(-start  => $nuc_start,
                                             -end    => $nuc_end,
                                             -strand => $cds_strand));
        }

        my $misc = Bio::PSU::Feature->new(-key    => 'misc_feature',
                                          -ranges => \@locs);

        $misc->qadd ('coord', @coord_qualifiers);

	# Add the aa coordinates of the transmembrane domains to the note
	# Was gone within a test "if ($verbose)" !!!!

	$note .= " at aa ";
	my $text = join(', ', map { $_->[0] . "-" . $_->[1] } @coords);
	$text =~ s/, (\d+-\d+)$/ and $1/;
	
	$note .= $text;

        $misc->qadd('id', $id);
        $misc->qadd('note', $note);
        $misc->qadd('colour', $colour);
        $misc->qadd('algorithm', "TMHMM 2.0");
        $misc->qadd('type', 'TMHMM');

        if ($filter) {
            $entry->features($misc);
        } else {
            my $text = $fth->make_feature_block($misc);
            print $text;
        }

        # Added facility for annotation of original features
        if ($annot)
        {
            $id_to_cds{$id}->qadd('note', $note);
        }
    }

    close(TMHMM);
    # unlink "$tempfile";

    if ($filter) {      
        $eo->write_seq($entry);
    }
}
