#!/usr/local/bin/perl -w
#
# $Header: /nfs/pathsoft/repository/psu/genexec/perl/src/trans_embl.pl,v 1.7 2001/02/14 11:23:22 kdj Exp $
#
# name:   trans_embl
# author: Keith James kdj@sanger.ac.uk
#
# purpose: Given one or more EMBL files this script will create a
# purpose: database of translations in Fasta, EMBL, GCG or raw
# purpose: formats. Each CDS is distinguished from the others by its
# purpose: coding sequence (rather than its name or translation) and
# purpose: any duplicates are set aside. If the -j option is used the
# purpose: script attempts to join partial CDS which are partly in one
# purpose: EMBL file and partly in another (subject to a minimum
# purpose: overlap of 20 aa).
# purpose:
# purpose: Example:
# purpose:
# purpose: trans_embl -f foo -p bar -d bing -m fasta *.embl
# purpose:
# purpose: -f <file>  full CDS are written to this file
# purpose: -p <file>  partial CDS are written to this file
# purpose: -j <file>  joined CDS are written to this file
# purpose: -d <file>  duplicate CDS are written to this file
# purpose:
# purpose: -m <fasta|embl|gcg|raw> output format (required)
# purpose:
# purpose: If one or more or these arguments are omitted, that file is
# purpose: not written.

BEGIN 
{
    unshift(@INC, "/nfs/disk92/PerlSource/Bioperl/Releases/bioperl-0.05.1/");
}

use strict;
use Getopt::Std;
use Embl;
use Bio::Seq;
use Bio::SeqIO;

my (%opts, $printfull, $printjoin, $printpart, $printdups, $format, $printpseu);
my @entries;
my (%to_join, %been_joined, @permute);
my (%uniq_partial, @dup_partial);
my (%uniq_joined, @dup_joined);
my (%uniq_full, @dup_full);
my (%uniq_pseu, @dup_pseu);

# minimum overlap accepted for translations to be joined
my $min_overlap = 20;

my $usage = "Usage: trans_embl [-f <file> -j <file> -p <file> -s <file> -d <bool>] -m <format> files
\t-f write translations of full CDS to file
\t-j write translations of joined CDS to file
\t-p write translations of partial CDS to file
\t-s write nucleotide sequence of pseudogenes to file
\t-d write translations of duplicate CDS to file
\t-m write format (Fasta, EMBL, GCG or Raw)
";

getopts('df:j:m:p:s:', \%opts);

defined $opts{"d"} and $printdups++;
defined $opts{"f"} and $printfull = $opts{"f"};
defined $opts{"j"} and $printjoin = $opts{"j"};
defined $opts{"m"} and $format    = $opts{"m"};
defined $opts{"p"} and $printpart = $opts{"p"};
defined $opts{"s"} and $printpseu = $opts{"s"};

unless (@ARGV) { die $usage }

if (defined $format)
{
    unless ($format =~ /Raw|Fasta|EMBL|GCG/i)
    {
	die "Invalid output format: $format\n"
    }
    $format =~ tr/rAWfSTemblgc/RawFstEMBLGC/;
}
else
{
    warn "No output format specified. Defaulting to Fasta\n";
    $format = "Fasta";
}

###
### open and read the Embl files
###

my @files = @ARGV;

foreach my $file (@files)
{
    open(FH, "$file") or die "Unable to open $file: $!\n";
    my $newset = Embl->Read(\*FH);
    push(@entries, $newset->entries());
    close(FH);
}

###
### parse the feature table of each entry
###

foreach my $entry (@entries)
{
    next unless ($entry->SQ());
    next unless ($entry->FT());

    my $id      = $entry->ID();
    my $f_table = $entry->FT();
    my $bio_seq = Bio::Seq->new( -seq => $entry->SQ()->{ "seq" } );

    my @features = $f_table->features();

    foreach my $feature (@features)
    {
	$feature->key() eq "CDS" and do
	{
	    my (@genes, $product, $partial, $pseudo);
	    my $codon_start = 1;

	    my @qualifiers = $feature->qualifiers();

	    foreach my $qualifier (@qualifiers)
	    {
	    SWITCH:
		{
		    $qualifier->{ "name" } eq "gene" and do
		    {
			my $val = $qualifier->{ "value" };
			$val =~ s/\"//g;
			push(@genes, $val);
			last SWITCH;
		    };

		    $qualifier->{ "name" } eq "product" and do
		    {
			$product = $qualifier->{ "value" };
			$product =~ s/\"//g;
			last SWITCH;
		    };

		    $qualifier->{ "name" } eq "codon_start" and do
		    {
			$codon_start = $qualifier->{ "value" };
			last SWITCH;
		    };

		    $qualifier->{ "name" } eq "partial" and do
		    {
			$partial++;
			last SWITCH;
		    };

		    $qualifier->{ "name" } eq "pseudo" and do
		    {
			$pseudo++;
			last SWITCH;
		    };
		    my $nothing = 1;
		}
	    }

	    unless (@genes) { push(@genes, "undefined") }
	    $product ||= "undefined product";

	    my $location = $feature->location();

	    # make new Bio::Seq object from exons specified by $location
	    # within the Bio::Seq object $seq
	    my $bio_nt = &range_seq($location, $bio_seq);

	    my $offset = substr($bio_nt->str(), $codon_start - 1);
	    $bio_nt->setseq($offset);

	    my $bio_aa = $bio_nt->translate();

	    # set the Bio::Seq object's id and desc
	    my $sep = "";
	    if ($format eq "Fasta") { $sep = ": " };
	    $bio_aa->id(join(',', @genes));
	    $bio_aa->desc("$sep$product");

	    # check for pseudogenes
	    if (defined $pseudo)
	    {
		$bio_nt->id(join(',', @genes));
		$bio_nt->desc("$sep$product (pseudo)");
		if (exists $uniq_pseu{ $bio_nt->str() } )
		{
		    print "Warning: duplicate pseudogene ", $bio_aa->id(), "\n";
		    push(@dup_pseu, $bio_nt);
		}
		else
		{
		    $uniq_pseu{ $bio_nt->str() } = $bio_nt;
		}
	    }
	    # set aside partial sequences for joining
	    elsif ((grep /<|>/, $location) || defined $partial)
	    {
		$bio_aa->desc($bio_aa->desc() . " (partial)");
		if (exists $to_join{ $bio_nt->str() })
		{
		    print "Warning: duplicate partial sequence ", $bio_aa->id(), "\n";
		    push(@dup_partial, $bio_aa);
		}
		else
		{
		    $to_join{ $bio_nt->str() } = $bio_aa;
		}
	    }
	    # otherwise add to list for printing
	    else
	    {
		if (exists $uniq_full{ $bio_nt->str() })
		{
		    print "Warning: duplicate full sequence ", $bio_aa->id(), "\n";
		    push(@dup_full, $bio_aa);
		}
		else
		{
		    $uniq_full{ $bio_nt->str() } = $bio_aa;
		}
	    }
	}
    }
}

###
### try to join each partial sequence to all the others in turn
###

# create a list of anonymous lists
foreach (keys %to_join)
{
    push(@permute, [$_, $to_join{ $_ }]);
}

if (defined $printjoin)
{
    # outer loop stays the same size
 OUTER: foreach my $outer (0..($#permute - 1))
    {
	# inner loop gets smaller with each iteration
	foreach my $inner (($outer + 1)..$#permute)
	{
	    my $consensus = &compare($permute[$outer][1]->str(), $permute[$inner][1]->str(), $min_overlap);	

	    # check for a sequence being joined to more than one other
	    if (defined $consensus)
	    {
		if ($been_joined{ $permute[$outer][0] })
		{
		    print "Warning: ", $permute[$outer][1]->id(), " has been joined more than once\n";
		}
		if ($been_joined{ $permute[$inner][0] })
		{
		    print "Warning: ", $permute[$inner][1]->id(), " has been joined more than once\n";
		}

		# always make the join anyway, for diagnostic purposes
		my $join = Bio::Seq->new
		    (
		     -seq  => $consensus,
		     -id   => $permute[$outer][1]->id()   . "+" . $permute[$inner][1]->id(),
		     -desc => $permute[$outer][1]->desc() . " " . $permute[$inner][1]->desc()
		    );

		# check for duplicate sequences after joining
		if (exists $uniq_joined{ $join->id() })
		{
		    print "Warning: joined CDS ", $join->id(), " present more than once\n";
		    push(@dup_joined, $join)
		}
		else
		{
		    $uniq_joined{ $join->id() } = $join;
		    $been_joined{ $permute[$outer][0] }++;
		    $been_joined{ $permute[$inner][0] }++;
		}
		next OUTER;
	    }
	}
    }
}

# collect the unjoined partials
foreach my $partial (@permute)
{
    if (! exists $been_joined{ $partial->[0] })
    {
	$uniq_partial{ $partial->[0] } = $partial->[1];
    }
}

###
### dump all the sequences to files and print any warnings
###

&dump_file(\%uniq_full,    $printfull, $format) if (defined $printfull);
&dump_file(\%uniq_joined,  $printjoin, $format) if (defined $printjoin);
&dump_file(\%uniq_partial, $printpart, $format) if (defined $printpart);
&dump_file(\%uniq_pseu,    $printpseu, $format) if (defined $printpseu);

if (defined $printdups)
{
    &dump_dup_file(\@dup_full,    "$printfull.dup", $format) if (defined $printfull);
    &dump_dup_file(\@dup_joined,  "$printjoin.dup", $format) if (defined $printjoin);
    &dump_dup_file(\@dup_partial, "$printpart.dup", $format) if (defined $printpart);
    &dump_dup_file(\@dup_pseu,    "$printpseu.dup", $format) if (defined $printpseu);
}
else
{
    &dup_warning(\@dup_full,    "full");
    &dup_warning(\@dup_joined,  "joined");
    &dup_warning(\@dup_partial, "partial");
    &dup_warning(\@dup_pseu,    "pseudo");
}


###
### subroutines
###

sub compare
{
    my ($seq1, $seq2, $min_overlap)  = @_;
    my $consensus;

    # try both ways as we don't know which belongs on the left or right
    $consensus = &slide($seq1, $seq2, $min_overlap);
    $consensus = &slide($seq2, $seq1, $min_overlap) unless (defined $consensus);
    return $consensus;
}

sub slide
{
    my ($leftseq, $rightseq, $min_overlap)  = @_;
    my $consensus;

    my @lengths = sort { $a <=> $b } (length ($leftseq), length($rightseq));

    # slide the end of the left sequence over the start of the right sequence
    foreach my $overlap ($min_overlap..$lengths[0])
    {
	my $leftbit  = substr($leftseq, 0 - $overlap, $lengths[0]);
	my $rightbit = substr($rightseq, 0, $overlap);
	if ($leftbit eq $rightbit)
	{
	    $consensus = $leftseq . substr($rightseq, $overlap);
	    last;
	}
    }
    return $consensus;
}

sub range_seq
{
    my ($location, $bio_seq) = @_;
    my ($comp, @ascending);
    my $join = "";

    grep /complement/, $location and $comp++;

    # clean up the location string
    $location =~ s/complement|join|\(|\)|<|>//g;
    my @subranges = split(',', $location);

    foreach my $subrange (@subranges)
    {
	if ($subrange =~ /(\d+)\.\.(\d+)/)
	{
	    # for a range of 2+ bases
	    my ($coord1, $coord2) = sort { $a <=> $b } ($1, $2);
	    push (@ascending, [$coord1, $coord2]);
	}
	else
	{
	    # for a range of 1 base
	    push (@ascending, [$subrange, $subrange]);
	}
    }

    # sort by the first coordinate
    my @sorted_ascending = sort { @$a[0] <=> @$b[0]; } @ascending;

    foreach my $range (@sorted_ascending)
    {
	$join .= $bio_seq->str(@$range[0], @$range[1]);
    }

    # create new Bio::seq object to return
    my $bio_subseq = Bio::Seq->new( -seq => $join );

    # store original chromosomal boundaries in the object's names hash
    my %names = (
		 "csome_start" => $sorted_ascending[0][0],
		 "csome_end"   => $sorted_ascending[-1][-1]
		);
    $bio_subseq->names(\%names);

    if (defined $comp) { $bio_subseq = $bio_subseq->revcom() }
    return $bio_subseq;
}

sub dump_file
{
    my ($uniq_r, $filename, $format) = @_;

    my $out = Bio::SeqIO->new
	(
	 -file     => ">$filename",
	 '-format' => "$format"
	);
    foreach (keys %$uniq_r)
    {
	$out->write_seq($uniq_r->{ $_ });
    }
}

sub dup_warning
{
    my ($dup_r, $type) = @_;

    foreach (@$dup_r)
    {
	print "Warning: duplicate copy of $type CDS ", $_->id(), " excluded from output\n";
    }
}

sub dump_dup_file
{
    my ($dup_r, $filename, $format) = @_;

    my $out = Bio::SeqIO->new
	(
	 -file     => ">$filename",
	 '-format' => "$format"
	);

    foreach (@$dup_r)
    {
	$out->write_seq($_);
    }
}
