#!/usr/local/bin/perl5.6.1 -w

=head1 NAME

 translate_features - translates EMBL features and writes Fasta format
 output

=head1 SYNOPSIS

 translate_features < <file_1> [.. <file_n>]

=head1 DESCRIPTION

This script takes sequences in EMBL format on STDIN and a regular
expression which is used to match feature keys. All features with
matching keys are translated. The translation will respect codon_start
qualifiers. The translation table must be specified. A feature with a
pseudo qualifier will be skipped. Feature IDs are taken from the
systematic_id qualifier or are generated from the sequence ID if no
gene qualifier is present.

=head1 METHODS

None

=head1 AUTHOR

Keith James (kdj@sanger.ac.uk)

=head1 COPYRIGHT

Copyright (C) 2002 Genome Research Limited. All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind. It may
be used, redistributed and/or modified under the same conditions as
Perl itself.

=cut

use strict;
use Getopt::Std;

use Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper;
use Bio::PSU::SeqFactory;

my $usage = "Usage: translate_features -k <regex> -t <table> [-r <regex> -l <int>] < <file_1> [.. <file_n>]
\t-k regex used to match feature key
\t-l minumum length of coding sequence to translate
\t-r specify regex used to recognise a systematic gene name
\t-t translation table number

\tTables available are:

\t 1) Standard
\t 2) Vertebrate Mitochondrial
\t 3) Yeast Mitochondrial
\t11) Bacterial
";

my (%opts, $kregex, $iregex, $table);
my $len = 3;

getopts('hk:l:r:t:', \%opts);

defined $opts{h} and die $usage;
defined $opts{k} and $kregex = $opts{k};
defined $opts{l} and $len    = $opts{l};
defined $opts{r} and $iregex = $opts{r};
defined $opts{t} and $table  = $opts{t};

unless (defined $kregex) { die $usage }
unless (defined $table) { die $usage }

my $compk = eval { qr/$kregex/m } or
	die "Failed to compile regex: $kregex\n";
my $compi = undef;

if (defined $iregex)
{
    $compi = eval { qr/$iregex/m } or
        die "Failed to compile regex: $iregex\n";
}

my $id_helper =
  Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper->new(-selection_regex => $compi);

my $ei = Bio::PSU::SeqFactory->make(-fh     => \*STDIN,
									-format => 'embl');
my $eo = Bio::PSU::SeqFactory->make(-fh     => \*STDOUT,
                                    -format => 'fasta');

my $scount = 0;
my $fcount = 0;

while (my $seq = $ei->next_seq)
{
	my $seq_id = $seq->id;
	$seq_id ||= "unknown_sequence_" . $scount++;

	foreach my $f ($seq->features)
    {
		next if $f->qexists('pseudo');
        next unless $f->key =~ /^$compk$/;
        next if (length ($f->str) < $len);

        my $id = $id_helper->get_systematic_id($f);
		my $desc = "";

		my @products = $f->product;
		if (@products)
		{
			$desc = shift @products;
		}

		my $aa = $f->translate($table);
		$aa->id($id);
		$aa->desc($desc);
		$eo->write_seq($aa);
	}
}
