#!/usr/local/bin/perl5.6.1 -w

# $Header: /nfs/pathsoft/repository/psu/genexec/perl/src/tribe_to_embl.pl,v 1.5 2003/11/14 13:18:48 kmr Exp $

use strict;

use Bio::PSU::SeqFactory;
use Bio::PSU::IO::BufferFH;

use Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper;

sub usage
{
  die <<EOF;
usage $0 cluster_name embl_file tribe_file > new_embl_file

This script will add a /cluster qualifier to each CDS feature in the EMBL file
indicating which TribeMCL cluster the feature came from.

The cluster qualifier will have this format:
  /cluster="cluster_name:xxxx yy"

where xxxx is the cluster id (padded with zeros so that it's at least
4 characters) and yyy is the number of proteins in that cluster.

If a cluster name of tribe_e-80 is passed on the command line we might
have this qualifier for cluster 45, which has 12 members.
  /cluster="tribe_e-80:0045 12"
EOF
}

if (@ARGV != 3) {
  usage;
}

my $cluster_name = $ARGV[0];
my $embl_file = $ARGV[1];
my $tribe_file = $ARGV[2];

open TRIBE, $tribe_file or die "can't open $tribe_file\n";

# gene id to cluster id mapping
my %tribe_hash = ();

# count of the number of genes in each cluster
my %cluster_counts = ();

while (<TRIBE>) {
  my ($cluster_id, $protein_id, @extra) = split;

  if (@extra > 0) {
    die "garbage at end of line:\n$_";
  }

  # each line is:
  # cluster_id protein_id
  $tribe_hash{$protein_id} = $cluster_id;

  ++$cluster_counts{$cluster_id};
}

my $in = Bio::PSU::SeqFactory->make(-file   => $embl_file,
                                    -format => 'EMBL');
my $out = Bio::PSU::SeqFactory->make(-fh     => \*STDOUT,
                                     -format => 'embl');
my $id_helper =
  Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper->new();

while (my $seq = $in->next_seq) {
  if (defined $seq) {
    for my $f ($seq->features) {

      next unless $f->key () eq "CDS";

      my $gene;

      eval {
        $gene = $id_helper->get_systematic_id($f);
      };

      if ($@) {
        my $loc = join ",", map {
          $_->{start} . ".." . $_->{end}
        } $f->ranges ();

        print STDERR "Error in CDS at: $loc -\n", $@;
        next;
      }

      if (defined $gene && exists $tribe_hash{$gene}) {
        my $tribe_id = $tribe_hash{$gene};
        my $count = $cluster_counts{$tribe_id};
        if ($count > 1) {
          my $new_qualifier =
          sprintf '%s:%04.4s %s', $cluster_name, $tribe_id, $count;
          $f->qadd("cluster", $new_qualifier)
        }
      }
    }

    $out->write_seq($seq);
  } else {
    die "can't open $embl_file\n";
  }
}

