/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.sql;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

/**
 * <code>ConnectionFactory</code> is an interface for obtaining JDBC
 * <code>Connection</code>s and performing operations on them and
 * their <code>Statement</code>s and <code>ResultSet</code>s. An
 * implementation could use <code>java.sql.DriverManager</code>,
 * <code>javax.sql.DataSource</code> or connection pool etc.
 *
 * @author Keith James
 */
public interface ConnectionFactory
{
    /**
     * <code>getConnection</code> obtains a connection. This does not
     * have to be a new connection, but may be one already in use by
     * the current transaction, for example. The behaviour depends on
     * the implementation and should be documented there.
     *
     * @return a <code>Connection</code>.
     * @exception SQLException if an error occurs.
     */
    public Connection getConnection() throws SQLException;

    /**
     * <code>commit</code> commits the current transaction on this
     * connection, dealing with any exceptions appropriately.
     *
     * @param c a <code>Connection</code>.
     */
    public void commit(Connection c);

    /**
     * <code>close</code> closes the connection, dealing with any
     * exceptions appropriately.
     *
     * @param c a <code>Connection</code>.
     */
    public void close(Connection c);

    /**
     * <code>rollback</code> rolls back the current transaction,
     * dealing with any exceptions appropriately.
     *
     * @param c a <code>Connection</code>.
     */
    public void rollback(Connection c);

    /**
     * <code>close</code> closes the statement, dealing with any
     * exceptions appropriately.
     *
     * @param s a <code>Statement</code>.
     */
    public void close(Statement s);

    /**
     * <code>close</code> closes the resultset, dealing with any
     * exceptions appropriately.
     *
     * @param r a <code>ResultSet</code>.
     */
    public void close(ResultSet r);
}
