/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.sql;

import java.io.IOException;
import java.io.InputStream;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ResourceBundle;

import org.apache.log4j.Category;

/**
 * <p><code>ConnectionFactoryDefaultImpl</code> is a simple
 * implementation which uses <code>java.sql.DriverManager</code> to
 * obtain a connection. Once a new connection is obtained using
 * <code>getConnection()</code>, the first thing it does is to call
 * <code>setAutoCommit(false)</code> on it to make operations
 * transactional.</p>
 *
 * <p>The factory loads its configuration from a file named
 * "/ConnectionFactoryDefaultImpl.properties". It looks for a key
 * "jdbc.connection.factory.props" where it expects to find a resource
 * location for the file containing the database settings. This is
 * also a properties file, with the following keys:</p>
 *
 * <ul>
 *  <li>jdbc.driver      &lt;driver name&gt;</li>
 *  <li>jdbc.subprotocol &lt;subprotocol&gt;</li>
 *  <li>jdbc.host        &lt;host name&gt;</li>
 *  <li>jdbc.port        &lt;port&gt;</li>
 *  <li>jdbc.database    &lt;database name&gt;</li>
 * </ul>
 *
 * <p>For example:</p>
 *
 * <ul>
 *  <li>jdbc.driver      &lt;org.postgresql.Driver&gt;</li>
 *  <li>jdbc.subprotocol &lt;postgresql&gt;</li>
 *  <li>jdbc.host        &lt;localhost&gt;</li>
 *  <li>jdbc.port        &lt;5432&gt;</li>
 *  <li>jdbc.database    &lt;foobase&gt;</li>
 * </ul>
 *
 * <p>These values are used to construct the database URL.</p>
 *
 * @author Keith James
 */
public class ConnectionFactoryDefaultImpl implements ConnectionFactory
{
    /**
     * Constant <code>FACTORY_BUNDLE_KEY</code> the
     * <code>ResourceBundle</code> key to specify the
     * <code>ResourceBundle</code> file containing the JDBC settings.
     */
    public static final String FACTORY_BUNDLE_KEY = "jdbc.connection.factory.props";

    /**
     * Constant <code>JDBC_DRIVER_KEY</code> the
     * <code>ResourceBundle</code> key to specify the JDBC driver.
     */
    public static final String JDBC_DRIVER_KEY = "jdbc.driver";

    /**
     * Constant <code>JDBC_SUBPROTOCOL_KEY</code> the
     * <code>ResourceBundle</code> key to specify the subprotocol.
     */
    public static final String JDBC_SUBPROTOCOL_KEY = "jdbc.subprotocol";

    /**
     * Constant <code>JDBC_HOST_KEY</code> the
     * <code>ResourceBundle</code> key to specify the host.
     */
    public static final String JDBC_HOST_KEY = "jdbc.host";

    /**
     * Constant <code>JDBC_PORT_KEY</code> the
     * <code>ResourceBundle</code> key to specify the port.
     */
    public static final String JDBC_PORT_KEY = "jdbc.port";

    /**
     * Constant <code>JDBC_DATABASE_KEY</code> the
     * <code>ResourceBundle</code> key to specify the database.
     */
    public static final String JDBC_DATABASE_KEY = "jdbc.database";

    // Log4j setup
    static Category cat = Category.getInstance(ConnectionFactoryDefaultImpl.class.getName());

    static ResourceBundle bundle = ResourceBundle.getBundle(ConnectionFactoryDefaultImpl.class.getName());
;

    private String configFile;

    private String driverName;
    private String subprotocol;
    private String host;
    private String port;
    private String database;
    private String url;

    private String user;
    private String password;

    /**
     * Creates a new <code>ConnectionFactoryDefaultImpl</code> for the
     * specified user.
     *
     * @param user a <code>String</code>.
     * @param password a <code>String</code>.
     *
     * @exception Exception if an error occurs.
     */
    public ConnectionFactoryDefaultImpl(String user, String password)
        throws Exception
    {
        String confBundleName;

        try
        {
            confBundleName = bundle.getString(FACTORY_BUNDLE_KEY);
        }
        catch (NullPointerException npe)
        {
            throw new Error(FACTORY_BUNDLE_KEY
                            + " is not defined in the ResourceBundle "
                            + ConnectionFactoryDefaultImpl.class.getName()
                            + ". See ConnectionFactoryDefaultImpl documentation for help.");
        }

        cat.debug("Loading factory configuration from " + confBundleName);
        ResourceBundle conf = ResourceBundle.getBundle(confBundleName);

        driverName  = conf.getString(JDBC_DRIVER_KEY);
        subprotocol = conf.getString(JDBC_SUBPROTOCOL_KEY);
        host        = conf.getString(JDBC_HOST_KEY);
        port        = conf.getString(JDBC_PORT_KEY);
        database    = conf.getString(JDBC_DATABASE_KEY);
            
        this.user     = user;
        this.password = password;

        if (driverName == null)
            throw new Error(JDBC_DRIVER_KEY
                            + " is not defined in the ResourceBundle "
                            + ConnectionFactoryDefaultImpl.class.getName()
                            + ". See ConnectionFactoryDefaultImpl documentation for help.");

        // Build database URL PostgreSQL
        if (driverName.equals("org.postgresql.Driver"))
            url = "jdbc:" + subprotocol + "://" + host + ":" + port + "/" + database;
        // Build database URL Oracle
        else if (driverName.equals("oracle.jdbc.driver.OracleDriver"))
            url = "jdbc:" + subprotocol + ":@" + host + ":" + port + ":" + database;
        else
            throw new Exception("Unable to construct database URL for unknown driver '"
                                + driverName
                                + "'");

        cat.debug("Loading database driver " + driverName);
        // Load driver
        Class.forName(driverName);
    }

    /**
     * <code>getConnection</code> gets a new connection.
     *
     * @return a <code>Connection</code>.
     *
     * @exception SQLException if an error occurs.
     */
    public Connection getConnection() throws SQLException
    {
        Connection connection =
            DriverManager.getConnection(url, user, password);

        cat.info("+++ Getting a new connection "
                  + connection
                  + ", " + url
                  + " for user=" + user);

        connection.setAutoCommit(false);
        cat.info(">>> Started transaction with "
                 + connection
                 + ", " + url
                 + " for user=" + user);

        return connection;
    }

    /**
     * <code>commit</code> commits the transaction on the current
     * connection. If the commit fails a rollback will be
     * attempted. Successes in transaction closing or rollback are
     * logged at the log4j INFO level, while failures are logged at
     * the ERROR level. Other operations are logged at the DEBUG
     * level.
     *
     * @param connection a <code>Connection</code>.
     */
    public void commit(Connection connection)
    {
        synchronized(connection)
        {
            try
            {
                if (connection != null)
                {
                    connection.commit();
                    cat.debug("::: Committed to " 
                              + connection
                              + ", " + url
                              + " for user=" + user);
                    cat.info("<<< Ended transaction with "
                             + connection
                             + ", " + url
                             + " for user=" + user);
                }
                else
                {
                    cat.warn("Tried to commit to null connection with " + url
                             + " for user=" + user);
                }
            }
            catch (Exception e1)
            {
                cat.error("Error committing to "
                          + connection
                          + ", " + url
                          + " for user=" + user, e1);

                rollback(connection);
            }
        }
    }

    /**
     * <code>close</code> closes the connection, dealing with any
     * exceptions appropriately. If the close fails a rollback will be
     * attempted.
     *
     * @param connection a <code>Connection</code>.
     */
    public void close(Connection connection)
    {
        synchronized(connection)
        {
            try
            {
                if (connection != null)
                {
                    connection.close();
                    cat.info("--- Closed connection to "
                             + connection
                             + ", " + url
                             + " for user=" + user);
                    connection = null;
                }
                else
                {
                    // Warn if tried to close null connection
                    cat.warn("Tried to close null connection to "
                             + url
                             + " for user=" + user);
                }
            }
            catch (Exception e1)
            {
                cat.error("Error closing Connection "
                          + connection
                          + ", " + url
                          + " for user=" + user, e1);

                // If close failed, try to rollback
                rollback(connection);
            }
        }
    }

    /**
     * <code>close</code> closes the statement, dealing with any
     * exceptions appropriately.
     *
     * @param statement a <code>Statement</code>.
     */
    public void close(Statement statement)
    {
        try
        {
            if (statement != null)
            {
                statement.close();
                cat.debug("  - Closed Statement " + statement);
            }
            else
            {
                cat.warn("Tried to close null Statement");
            }
        }
        catch (Exception e)
        {
            cat.error("Error closing Statement " + statement);

        }
    }

    /**
     * <code>close</code> closes the resultset, dealing with any
     * exceptions appropriately.
     *
     * @param resultSet a <code>ResultSet</code>.
     */
    public void close(ResultSet resultSet)
    {
        try
        {
            if (resultSet != null)
            {
                resultSet.close();
                cat.debug("  - Closed ResultSet " + resultSet);
            }
            else
            {
                cat.warn("Tried to close null ResultSet");
            }
        }
        catch (Exception e)
        {
            cat.error("Error closing ResultSet " + resultSet);
        }
    }

    public void rollback(Connection connection)
    {
        synchronized(connection)
        {
            try
            {
                if (connection != null)
                {
                    connection.rollback();
                    cat.info("@@@ Rollback of transaction with "
                             + connection
                             + ", " + url
                             + " for user=" + user);
                }
                else
                {
                    // Warn if tried to rollback null connection
                    cat.warn("Tried to rollback null connection to "
                             + url
                             + " for user=" + user);
                }
            }
            catch (Exception e2)
            {
                cat.error("Error during rollback from "
                          + connection
                          + ", " + url
                          + " for user=" + user, e2);
            }
        }
    }
}
