/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.sql;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import org.apache.log4j.Category;

/**
 * <code>ConnectionSupport</code> provides logging and exception
 * handling for common connection operations.
 *
 * @author Keith James
 */
public class ConnectionSupport
{
   // Log4j setup
    static Category cat = Category.getInstance(ConnectionSupport.class.getName());

    private String user;

    /**
     * Creates a new <code>ConnectionSupport</code> which will log
     * events under user "<unknown>".
     */
    public ConnectionSupport()
    {
        this("<unknown>");
    }

    /**
     * Creates a new <code>ConnectionSupport</code> which will log
     * events under the specified user name.
     *
     * @param user a <code>String</code>.
     */
    public ConnectionSupport(String user)
    {
        this.user = user;
    }

    /**
     * <code>commit</code> commits the transaction on the
     * connection. If the commit fails a rollback will be
     * attempted. Successes in transaction closing or rollback are
     * logged at the log4j INFO level, while failures are logged at
     * the ERROR level. Other operations are logged at the DEBUG
     * level.
     *
     * @param connection a <code>Connection</code>.
     */
    public void commit(Connection connection)
    {
        synchronized(connection)
        {
            try
            {
                if (connection != null)
                {
                    connection.commit();
                    cat.debug("::: Committed to " 
                              + connection
                              + " for user=" + user);
                    cat.info("<<< Ended transaction with "
                             + connection
                             + " for user=" + user);
                }
                else
                {
                    cat.warn("Tried to commit to null connection for user=" + user);
                }
            }
            catch (Exception e1)
            {
                cat.error("Error committing to "
                          + connection
                          + " for user=" + user, e1);

                rollback(connection);
            }
        }
    }

    /**
     * <code>close</code> closes the connection, dealing with any
     * exceptions appropriately. If the close fails a rollback will be
     * attempted.
     *
     * @param connection a <code>Connection</code>.
     */
    public void close(Connection connection)
    {
        synchronized(connection)
        {
            try
            {
                if (connection != null)
                {
                    connection.close();
                    cat.info("--- Closed connection to "
                             + connection
                             + " for user=" + user);
                    connection = null;
                }
                else
                {
                    // Warn if tried to close null connection
                    cat.warn("Tried to close null connection for user=" + user);
                }
            }
            catch (Exception e1)
            {
                cat.error("Error closing Connection "
                          + connection
                          + " for user=" + user, e1);

                // If close failed, try to rollback
                rollback(connection);
            }
        }
    }

    /**
     * <code>close</code> closes the statement, dealing with any
     * exceptions appropriately.
     *
     * @param statement a <code>Statement</code>.
     */
    public void close(Statement statement)
    {
        try
        {
            if (statement != null)
            {
                statement.close();
                cat.debug("  - Closed Statement " + statement);
            }
            else
            {
                cat.warn("Tried to close null Statement");
            }
        }
        catch (Exception e)
        {
            cat.error("Error closing Statement " + statement);

        }
    }

    /**
     * <code>close</code> closes the resultset, dealing with any
     * exceptions appropriately.
     *
     * @param resultSet a <code>ResultSet</code>.
     */
    public void close(ResultSet resultSet)
    {
        try
        {
            if (resultSet != null)
            {
                resultSet.close();
                cat.debug("  - Closed ResultSet " + resultSet);
            }
            else
            {
                cat.warn("Tried to close null ResultSet");
            }
        }
        catch (Exception e)
        {
            cat.error("Error closing ResultSet " + resultSet);
        }
    }

    /**
     * <code>rollback</code> attempts to rollback a transaction,
     * dealing with any exceptions appropriately.
     *
     * @param connection a <code>Connection</code>.
     */
    public void rollback(Connection connection)
    {
        synchronized(connection)
        {
            try
            {
                if (connection != null)
                {
                    connection.rollback();
                    cat.warn("@@@ Rollback of transaction with "
                             + connection
                             + " for user=" + user);
                }
                else
                {
                    // Warn if tried to rollback null connection
                    cat.warn("Tried to rollback null connection to for user=" + user);
                }
            }
            catch (Exception e2)
            {
                cat.error("Error during rollback from "
                          + connection
                          + " for user=" + user, e2);
            }
        }
    }
}
