/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.sql;

import java.sql.Connection;
import java.sql.SQLException;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;

import oracle.jdbc.pool.OracleDataSource;

import uk.ac.sanger.util.CommandLineDoc;

/**
 * <code>DataSourceBinder</code> creates <code>DataSource</code>s and
 * bind them to contexts. Set the JNDI properties using a standard
 * jndi.properties file.
 *
 * @author Keith James
 */
public class DataSourceBinder
{
    static final int VENDOR_POSTGRESQL = 0;
    static final int      VENDOR_MYSQL = 1;
    static final int     VENDOR_ORACLE = 2;

    private static CommandLineDoc docs;

    public static void main(String [] argv)
    {
        docs = new CommandLineDoc();
        docs.loadResources(DataSourceBinder.class.getName());

        String   dbName = null;
        String     host = null;
        int        port = -1;
        String     user = null;
        String password = null;
        int      vendor = -1;

        LongOpt [] lopts = new LongOpt[6];
        lopts[0] = new LongOpt("database", LongOpt.REQUIRED_ARGUMENT, null, 'd');
        lopts[1] = new LongOpt("host",     LongOpt.REQUIRED_ARGUMENT, null, 'h');
        lopts[2] = new LongOpt("port",     LongOpt.REQUIRED_ARGUMENT, null, 'p');
        lopts[3] = new LongOpt("user",     LongOpt.REQUIRED_ARGUMENT, null, 'u');
        lopts[4] = new LongOpt("vendor",   LongOpt.REQUIRED_ARGUMENT, null, 'v');
        lopts[5] = new LongOpt("password", LongOpt.REQUIRED_ARGUMENT, null, 'w');

        int    c;
        String arg;

        Getopt g = new Getopt("DataSourceBinder", argv, "d:h:p:u:v:w:", lopts);

        while ((c = g.getopt()) != -1)
        {
            switch (c)
            {
                case 'd':
                    // Set database
                    arg = g.getOptarg();
                    if (arg != null)
                        dbName = arg;
                    else
                        exitUsage();
                    break;

                case 'h':
                    // Set host
                    arg = g.getOptarg();
                    if (arg != null)
                        host = arg;
                    else
                        exitUsage();
                    break;

                case 'p':
                    // Set port
                    arg = g.getOptarg();
                    if (arg != null)
                    {
                        try
                        {
                            port = Integer.parseInt(arg);
                        }
                        catch (NumberFormatException nfe)
                        {
                            exitUsage();
                        }
                    }
                    else
                    {
                        exitUsage();
                    }
                    break;

                case 'u':
                    // Set user
                    arg = g.getOptarg();
                    if (arg != null)
                        user = arg;
                    else
                        exitUsage();
                    break;

                case 'v':
                    // Set vendor
                    arg = g.getOptarg();
                    if (arg != null)
                    {
                        if (arg.equals("oracle"))
                            vendor = VENDOR_ORACLE;
                        else
                        {
                            exitUsage();
                        }
                    }
                    else
                    {
                        exitUsage();
                    }
                    break;

                case 'w':
                    arg = g.getOptarg();
                    if (arg != null)
                         password = arg;
                    else
                        exitUsage();
                    break;

                default:
                    exitUsage();
                    break;
            }
        }

        if (dbName   == null ||
            host     == null ||
            user     == null ||
            password == null ||
            port     == -1)
        {
            exitUsage();
        }

        // Remaining argument(s)
        String [] rem = new String [argv.length - g.getOptind()];
        System.arraycopy(argv, g.getOptind(),
                         rem, 0,
                         argv.length - g.getOptind());

        if (rem.length < 1)
            exitUsage();

        try
        {
            bindOracleDataSource(rem[0], host, port,
                                 dbName, user, password);
        }
        catch (Exception e)
        {
            System.err.println("Binding failed. Trace follows:");
            e.printStackTrace();
            System.exit(0);
        }

        try
        {
            testLookup(rem[0]);
        }
        catch (Exception e)
        {
            System.err.println("Test lookup of new datasource failed. Trace follows:");
            e.printStackTrace();
        }
    }

    /**
     * <code>bindOracleDataSource</code> binds an
     * <code>OracleDataSource</code> to a JNDI name.
     *
     * @param dsName a <code>String</code> named subcontext to bind to
     * e.g. 'jdbc/FooDB'.
     * @param host a <code>String</code> host name.
     * @param port an <code>int</code>.
     * @param dbName a <code>String</code> database name.
     * @param user a <code>String</code> user name.
     * @param password a <code>String</code> password.
     *
     * @exception Exception if an error occurs.
     */
    static void bindOracleDataSource(String dsName,
                                     String host, int port, String dbName,
                                     String user, String password)
        throws SQLException, NamingException
    {
        Context context = new InitialContext();
        OracleDataSource datasource = new OracleDataSource();
        datasource.setDriverType("thin");
        datasource.setServerName(host);
        datasource.setPortNumber(port);
        datasource.setDatabaseName(dbName);
        datasource.setUser(user);
        datasource.setPassword(password);

        context.rebind(dsName, datasource);
    }

    static void testLookup(String dsName)
        throws SQLException, NamingException
    {
        Context context = new InitialContext();
        DataSource datasource = (DataSource) context.lookup(dsName);
        Connection con = datasource.getConnection();
        con.close();
    }

    static void exitUsage()
    {
        docs.printLongUsage(System.out);
        System.exit(0);
    }
}
