/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.util;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

/**
 * <code>AbstractPropertyChangeable</code> is a utility class which
 * provides lazy instantiation of <code>PropertyChangeSupport</code>
 * when listeners are added.
 *
 * @author Keith James
 */
public abstract class AbstractPropertyChangeable
    implements PropertyChangeable
{
    private transient PropertyChangeSupport changeSupport;

    /**
     * <code>addPropertyChangeListener</code> adds a
     * <code>PropertyChangeListener</code> to the listener list. The
     * listener is registered for all properties.
     *
     * @param listener a <code>PropertyChangeListener</code>.
     */
    public void addPropertyChangeListener(PropertyChangeListener listener)
    {
        PropertyChangeSupport pcs = getPropertyChangeSupport();

        synchronized (pcs)
        {
            pcs.addPropertyChangeListener(listener);
        }
    }

    /**
     * <code>addPropertyChangeListener</code> adds a
     * <code>PropertyChangeListener</code> for a specific
     * property. The listener will be invoked only when a call on
     * firePropertyChange names that specific property.
     *
     * @param propertyName a <code>String</code>.
     *
     * @param listener a <code>PropertyChangeListener</code>.
     */
    public void addPropertyChangeListener(String                 propertyName,
                                          PropertyChangeListener listener)
    {
        PropertyChangeSupport pcs = getPropertyChangeSupport();

        synchronized (pcs)
        {
            pcs.addPropertyChangeListener(propertyName, listener);
        }
    }

    /**
     * <code>removePropertyChangeListener</code> removes a
     * <code>PropertyChangeListener</code> from the listener
     * list. This removes a <code>PropertyChangeListener</code> that
     * was registered for all properties.
     *
     * @param listener a <code>PropertyChangeListener</code>.
     */
    public void removePropertyChangeListener(PropertyChangeListener listener)
    {
        PropertyChangeSupport pcs = getPropertyChangeSupport();

        synchronized (pcs)
        {
            pcs.removePropertyChangeListener(listener);
        }
    }

    /**
     * <code>removePropertyChangeListener</code> removes a
     * <code>PropertyChangeListener</code> for a specific property.
     *
     * @param propertyName a <code>String</code>.
     *
     * @param listener a <code>PropertyChangeListener</code>.
     */
    public void removePropertyChangeListener(String                 propertyName,
                                             PropertyChangeListener listener)
    {
        PropertyChangeSupport pcs = getPropertyChangeSupport();

        synchronized (pcs)
        {
            pcs.removePropertyChangeListener(propertyName, listener);
        }
    }

    /**
     * <code>hasListeners</code> checks if there are any listeners.
     *
     * @return a <code>boolean</code>.
     */
    public boolean hasListeners()
    {
        return changeSupport != null;
    }

    /**
     * <code>hasListeners</code> checks if there are any listeners for
     * a specific property.
     *
     * @param propertyName a <code>String</code>.
     *
     * @return a <code>boolean</code>.
     */
    public boolean hasListeners(String propertyName)
    {
        return (changeSupport != null &&
                getPropertyChangeSupport().hasListeners(propertyName));
    }

    /**
     * <code>getPropertyChangeSupport</code>.
     *
     * @return a <code>PropertyChangeSupport</code> value.
     */
    protected PropertyChangeSupport getPropertyChangeSupport()
    {
        if (changeSupport == null)
            changeSupport = new PropertyChangeSupport(this);

        return changeSupport;
    }
}
