/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.util;

import java.io.PrintStream;
import java.text.BreakIterator;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * <code>CommandLineDoc</code> manages command line option
 * documentation. It is designed to be used in printing usage
 * summaries and help for any available command line options. Text may
 * be associated with short (-h, -v) or long (--help, --version)
 * options. Internationalized <code>ResourceBundle</code>s loaded from
 * <code>Properties</code> files allow the documentation to be
 * maintained separately from the code. The properties files should be
 * named <code>&lt;classname&gt;Doc.properties</code>.
 *
 * @author Keith James
 */
public class CommandLineDoc
{
    private static String  UNAVAILABLE = "<No documentation available>";
    private static String SHORT_PREFIX = "-";
    private static String  LONG_PREFIX = "--";
    private static String  INIT_INDENT = "    ";
    private static int  MAX_LINELENGTH = 70;

    private String classDoc;
    private ResourceBundle resources;
    private SortedMap shortDocs;
    private SortedMap longDocs;

    /**
     * Creates a new <code>CommandLineDoc</code>.
     */
    public CommandLineDoc()
    {
        shortDocs = new TreeMap();
        longDocs  = new TreeMap();
    }

    /**
     * <code>loadResources</code> loads the documentation
     * <code>ResourceBundle</code>. The documentation should
     * be in a <code>Properties</code> file named according the Java
     * <code>ResourceBundle</code> naming convention.
     *
     * @param className a <code>String</code> class name of the
     * program being documented.
     */
    public void loadResources(String className)
    {
        resources = ResourceBundle.getBundle(className,
                                             Locale.getDefault());

        Enumeration e = resources.getKeys();
        while (e.hasMoreElements())
        {
            String key = (String) e.nextElement();
            String val = resources.getString(key);

            if (val == null)
                continue;

            if (key.equals(className))
            {
                setDoc(val);
            }
            else if (key.length() == 1)
            {
                setShortDoc(key.charAt(0), val);
            }
            else
            {
                setLongDoc(key, val);
            }
        }
    }

    /**
     * <code>getDoc</code> returns a documentation summary
     * <code>String</code> which describes the program. For example,
     * "FooMangler [-f] <file name>".
     *
     * @return a <code>String</code>.
     */
    public String getDoc()
    {
        return classDoc;
    }

    /**
     * <code>setDoc</code> sets a documentation <code>String</code>
     * which describes the program. For example, "FooMangler [-f]
     * <file name>".
     *
     * @param doc a <code>String</code>.
     */
    public void setDoc(String doc)
    {
        if (doc == null)
            doc = UNAVAILABLE;
        classDoc = doc;
    }

    /**
     * <code>getShortDoc</code> returns a documentation
     * <code>String</code> which describes a short option.
     *
     * @param opt a <code>char</code> option.
     *
     * @return a <code>String</code>.
     */
    public String getShortDoc(char opt)
    {
        return (String) shortDocs.get(new Character(opt));
    }

    /**
     * <code>setShortDoc</code> sets a documentation
     * <code>String</code> which describes a short option.
     *
     * @param opt a <code>char</code> option.
     *
     * @return a <code>String</code>.
     */
    public void setShortDoc(char opt, String doc)
    {
        if (doc == null)
            doc = UNAVAILABLE;
        shortDocs.put(new Character(opt), doc);
    }

    /**
     * <code>getLongDoc</code> returns a documentation
     * <code>String</code> which describes a long option.
     *
     * @param opt a <code>String</code> option.
     *
     * @return a <code>String</code>.
     */
    public String getLongDoc(String opt)
    {
        return (String) longDocs.get(opt);
    }

    /**
     * <code>setLongDoc</code> sets a documentation
     * <code>String</code> which describes a long option.
     *
     * @param opt a <code>String</code> option.
     *
     * @return a <code>String</code>.
     */
    public void setLongDoc(String opt, String doc)
    {
        if (doc == null)
            doc = UNAVAILABLE;
        longDocs.put(opt, doc);
    }

    /**
     * <code>printShortUsage</code> prints a usage statement using
     * short options only;
     *
     * @param ps a <code>PrintStream</code>.
     */
    public void printShortUsage(PrintStream ps)
    {
        StringBuffer sb = new StringBuffer();
        sb.append("Usage:\n\n");
        sb.append(getDoc());
        sb.append("\n\n");

        for (Iterator oi = shortDocs.keySet().iterator(); oi.hasNext();)
        {
            Object o = oi.next();

            if (o instanceof Character)
            {
                Character c = (Character) o;
                sb.append("Option ");
                sb.append(SHORT_PREFIX);
                sb.append(c);
                sb.append(" :\n");

                String doc = (String) shortDocs.get(c);
                formatDoc(doc, sb);
            }
        }

        ps.println(sb.substring(0));
    }

    /**
     * <code>printLongUsage</code> prints a usage statement using
     * long options only;
     *
     * @param ps a <code>PrintStream</code>.
     */
    public void printLongUsage(PrintStream ps)
    {
        StringBuffer sb = new StringBuffer();
        sb.append("Usage:\n\n");
        sb.append(getDoc());
        sb.append("\n\n");

        for (Iterator oi = longDocs.keySet().iterator(); oi.hasNext();)
        {
            Object o = oi.next();

            if (o instanceof String)
            {
                String s = (String) o;
                sb.append("Option ");
                sb.append(LONG_PREFIX);
                sb.append(s);
                sb.append(" :\n");

                String doc = (String) longDocs.get(s);
                formatDoc(doc, sb);
            }
        }

        ps.print(sb.substring(0));
    }

    /**
     * <code>formatDoc</code> performs first-line indentation and
     * subsequent line-wrapping.
     *
     * @param doc a <code>String</code> to format.
     * @param sb a <code>StringBuffer</code> into which to append the
     * result.
     */
    private void formatDoc(String doc, StringBuffer sb)
    {
        BreakIterator bit = BreakIterator.getLineInstance();
        bit.setText(doc);
        int start = bit.first();
        int   end = bit.next();
        int lineLength = 0;

        sb.append(INIT_INDENT);

        while (BreakIterator.DONE != end)
        {
            String word = doc.substring(start, end);
            lineLength = lineLength + word.length();

            if (lineLength >= MAX_LINELENGTH)
            {
                sb.append("\n");
                lineLength = word.length();
            }

            sb.append(word);
            start = end;
            end = bit.next();
        }

        sb.append("\n\n");
    }
}
