/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.util;

import java.beans.PropertyChangeListener;

/**
 * <code>PropertyChangeable</code> is an interface which specifies
 * that an object has bound properties to which a
 * <code>PropertyChangeListener</code> may listen. These methods are
 * specified in an interface so that it may be inherited by other
 * interfaces which need to be sure that all their implementations are
 * able to accept listeners.
 *
 * @author Keith James
 */
public interface PropertyChangeable
{
    /**
     * <code>addPropertyChangeListener</code> adds a
     * <code>PropertyChangeListener</code> to the listener list. The
     * listener is registered for all properties.
     *
     * @param listener a <code>PropertyChangeListener</code>.
     */
    public void addPropertyChangeListener(PropertyChangeListener listener);

    /**
     * <code>addPropertyChangeListener</code> adds a
     * <code>PropertyChangeListener</code> for a specific
     * property. The listener will be invoked only when a call on
     * firePropertyChange names that specific property.
     *
     * @param propertyName a <code>String</code>.
     *
     * @param listener a <code>PropertyChangeListener</code>.
     */
    public void addPropertyChangeListener(String                 propertyName,
                                          PropertyChangeListener listener);

    /**
     * <code>removePropertyChangeListener</code> removes a
     * <code>PropertyChangeListener</code> from the listener
     * list. This removes a <code>PropertyChangeListener</code> that
     * was registered for all properties.
     *
     * @param listener a <code>PropertyChangeListener</code>.
     */
    public void removePropertyChangeListener(PropertyChangeListener listener);

    /**
     * <code>removePropertyChangeListener</code> removes a
     * <code>PropertyChangeListener</code> for a specific property.
     *
     * @param propertyName a <code>String</code>.
     *
     * @param listener a <code>PropertyChangeListener</code>.
     */
    public void removePropertyChangeListener(String                 propertyName,
                                             PropertyChangeListener listener);

    /**
     * <code>hasListeners</code> checks if there are any listeners.
     *
     * @return a <code>boolean</code>.
     */
    public boolean hasListeners();

    /**
     * <code>hasListeners</code> checks if there are any listeners for
     * a specific property.
     *
     * @param propertyName a <code>String</code>.
     *
     * @return a <code>boolean</code>.
     */
    public boolean hasListeners(String propertyName);
}
