/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.util;

import java.util.HashMap;
import java.util.Map;

/**
 * <code>PrototypeFactory</code> is an abstract base class for
 * factories which create new instances of objects implementing the
 * <code>Prototype</code> interface. Classes implementing
 * <code>Prototype</code> should use the public clone method of the
 * interface as the default means of creating new instances.
 *
 * @author Keith James
 */
public abstract class PrototypeFactory
{
    // Central registry of factories
    private static Map factories = new HashMap();

    /**
     * <code>addFactory</code> adds a new
     * <code>PrototypeFactory</code> to the central registry of
     * factories.
     *
     * @param className a <code>String</code>.
     * @param factory a <code>PrototypeFactory</code>.
     */
    public static void addFactory(final String           className,
                                  final PrototypeFactory factory)
    {
        factories.put(className, factory);
    }

    /**
     * <code>makePrototype</code> creates a new instance of a
     * <code>Prototype</code> from the correct factory housed in the
     * registry.
     *
     * @param className a <code>String</code> which is the name of the
     * concrete class implementing <code>Prototype</code> to create a
     * new instance of.
     *
     * @return a <code>Prototype</code>.
     *
     * @exception NoSuchPrototypeException if there is no such
     * implementation of <code>Prototype</code>.
     * @exception CloneNotSupportedException if the
     * <code>Prototype</code> cloning fails.
     */
    public static final Prototype makePrototype(final String className)
        throws NoSuchPrototypeException, CloneNotSupportedException
    {
        // If there is no factory for this class try to load the class
        // now. A static block within the class should register its
        // factory in the static Map
        if (! factories.containsKey(className))
        {
            try
            {
                Class.forName(className);
            }
            catch (ClassNotFoundException cnfe)
            {
                throw new NoSuchPrototypeException("Failed to find class "
                                                   + className);
            }
        }

        return ((PrototypeFactory) factories.get(className)).make();
    }

    /**
     * <code>make</code> should be implemented by concrete
     * factories. As the class implementing <code>Prototype</code>
     * will probably not have a public constructor (so that cloning
     * becomes the only available means of creating new instances),
     * the factory will probably be an inner class.
     *
     * @return a <code>Prototype</code>.
     *
     * @exception CloneNotSupportedException if cloning fails.
     */
    protected abstract Prototype make() throws CloneNotSupportedException;
}
