/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.util;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * <code>PrototypeRegistry</code> provides a registry of available
 * <code>Prototype</code>s and manages cloning of new
 * <code>Prototype</code> instances.
 *
 * @author Keith James
 */
public class PrototypeRegistry
{
    private Map registry;

    /**
     * Creates a new, empty <code>PrototypeRegistry</code>.
     */
    public PrototypeRegistry()
    {
        registry = new HashMap();
    }

    /**
     * <code>bindPrototype</code> adds a copy of a
     * <code>Prototype</code> to the registry.
     *
     * @param key an <code>Object</code>.
     * @param prototype a corresponding <code>Prototype</code>.
     *
     * @exception CloneNotSupportedException if an error occurs.
     */
    public void bindPrototype(Object key, Prototype prototype)
        throws CloneNotSupportedException
    {
        Prototype p = (Prototype) prototype.clone();
        registry.put(key, p);
    }

    /**
     * <code>unbindPrototype</code> removes a <code>Prototype</code>
     * from the registry.
     *
     * @param key an <code>Object</code>.
     *
     * @exception NoSuchPrototypeException if the key does not
     * exist in the registry.
     */
    public void unbindPrototype(Object key)
        throws NoSuchPrototypeException
    {
        if (registry.isEmpty())
            throw new NoSuchPrototypeException("Failed to unbind Prototype for key '"
                                               + key
                                               + "' as the registry is empty");
        if (! registry.containsKey(key))
            throw new NoSuchPrototypeException("Failed to unbind Prototype for key '"
                                               + key
                                               + "' as the registry does not contain this key");
        registry.remove(key);
    }

    /**
     * <code>copyPrototype</code> retrieves a copied
     * <code>Prototype</code> for a given registry key.
     *
     * @param key an <code>Object</code>.
     *
     * @return a <code>Prototype</code>.
     *
     * @exception CloneNotSupportedException if an error occurs.
     * @exception NoSuchPrototypeException if the key
     * does not exist in the registry.
     */
    public Prototype lookup(Object key) throws
        CloneNotSupportedException, NoSuchPrototypeException
    {
        if (registry.isEmpty())
            throw new NoSuchPrototypeException("Failed to lookup Prototype for key '"
                                               + key
                                               + "' as the registry is empty");
        if (! registry.containsKey(key))
            throw new NoSuchPrototypeException("Failed to lookup Prototype for key '"
                                               + key
                                               + "' as the registry does not contain this key");

        return (Prototype) ((Prototype) registry.get(key)).clone();
    }

    /**
     * <code>getAllPrototypes</code> returns all the
     * <code>Prototype</code>s in the registry.
     *
     * @return a <code>Set</code> of <code>Prototype</code>s.
     */
    public Set getAllPrototypes()
    {
        return  Collections.unmodifiableSet(registry.entrySet());
    }

    /**
     * <code>getRegistryKeySet</code> returns all the keys in the
     * registry.
     *
     * @return a <code>Set</code> of <code>Object</code>s.
     */
    public Set getRegistryKeySet()
    {
        return Collections.unmodifiableSet(registry.keySet());
    }
}
