/*
 * This file is part of the Pathogen Sequencing Unit Java library.
 * Copyright (c) 2001 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.util.gui;

import java.awt.Image;
import java.awt.Toolkit;
import java.io.File;
import java.net.URL;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.NoSuchElementException;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.Set;

import org.apache.log4j.Category;

/**
 * <code>ImageFactoryDefaultImpl</code> is the default implementation
 * of <code>ImageFactory</code>. It loads images from files specified
 * in the properties file loaded as a <code>ResourceBundle</code>.
 *
 * @author Keith James
 */
public class ImageFactoryDefaultImpl implements ImageFactory
{
    // Log4j setup
    static Category cat = Category.getInstance(ImageFactoryDefaultImpl.class.getName());

    static ResourceBundle bundle;

    // Singleton instance of this class
    private static final ImageFactory INSTANCE = new ImageFactoryDefaultImpl();

    // Map of keys to images
    private static Map images;

    /**
     * Creates a new <code>ImageFactoryDefaultImpl</code>. Private
     * constructor; use getInstance to get the singleton.
     *
     * @exception Exception if an error occurs.
     */
    private ImageFactoryDefaultImpl()
    {
        images = new HashMap();

        Toolkit kit = Toolkit.getDefaultToolkit();

        bundle = ResourceBundle.getBundle(ImageFactoryDefaultImpl.class.getName());

        Enumeration e = bundle.getKeys();
        
        String key, val;

        while (e.hasMoreElements())
        {
            key = (String) e.nextElement();

            try
            {
                val = bundle.getString(key);
            }
            catch (NullPointerException npe)
            {
                throw new Error("Malformed ResourceBundle "
                                + ImageFactoryDefaultImpl.class.getName()
                                + " contains key '"
                                + key
                                + "' missing a corresponding value");
            }

            URL fileURL = ImageFactoryDefaultImpl.class.getResource(val);
            try
            {
                cat.debug("Loading '" + fileURL + "'");
                images.put(key, kit.getImage(fileURL));
            }
            catch (NullPointerException npe)
            {
                cat.error("Failed to load image '"
                          + fileURL
                          + "'");
            }
        }
    }

    /**
     * <code>getInstance</code> returns the singleton instance of the
     * class.
     *
     * @return an <code>ImageFactory</code>.
     */
    public static ImageFactory getInstance()
    {
        return INSTANCE;
    }

    public Image createImage(String identifer) throws NoSuchImageException
    {
        if (! (images.keySet().contains(identifer)))
            throw new NoSuchImageException("Failed to retrieve image for identifier '"
                                           + identifer
                                           + "'");

        return (Image) images.get(identifer);
    }

    public Collection createImages()
    {
        return Collections.unmodifiableCollection(images.values());
    }
}
