/*
 * This file is part of the Pathogen Sequencing Unit Java library.
 * Copyright (c) 2001 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.util.seq;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

import org.biojava.bio.BioError;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.io.SequenceFormat;
import org.biojava.bio.symbol.Alphabet;

/**
 * <code>SequenceFormatTesterDefaultImpl</code> tests ASCII flat files
 * containing sequence data to determine their format.
 *
 * @author Keith James
 */
public class SequenceFormatTesterDefaultImpl implements SequenceFormatTester
{
    // ResourceBundle describing known formats
    private static ResourceBundle bundle;

    // Singleton instance of this class
    private static final SequenceFormatTester INSTANCE =
        new SequenceFormatTesterDefaultImpl();

    // Map of line formats to sequence formats.
    private static Map formats;

    /**
     * Creates a new
     * <code>SequenceFormatTesterDefaultImpl</code>. Private
     * constructor; use getInstance to get the singleton.
     */
    private SequenceFormatTesterDefaultImpl()
    {
        formats = new HashMap();

        String key = null;
        String val = null;

        try
        {
            bundle = ResourceBundle.getBundle(SequenceFormatTesterDefaultImpl.class.getName());
            Enumeration e = bundle.getKeys();

            while (e.hasMoreElements())
            {
                key = (String) e.nextElement();

                try
                {
                    val = bundle.getString(key);
                }
                catch (NullPointerException npe)
                {
                    throw new Error("Malformed ResourceBundle "
                                    + SequenceFormatTesterDefaultImpl.class.getName()
                                    + " contains key '"
                                    + key
                                    + "' missing a corresponding value");
                }

                formats.put(key, Class.forName(val).newInstance());
            }
        }
        catch (MissingResourceException mre)
        {
            throw new Error("Failed to find ResourceBundle "
                            + SequenceFormatTesterDefaultImpl.class.getName());
        }
        catch (ClassNotFoundException cnfe)
        {
            throw new Error("Malformed ResourceBundle "
                            + SequenceFormatTesterDefaultImpl.class.getName()
                            + " requested unavailable class name '"
                            + val
                            + "'");
        }
        catch (InstantiationException ie)
        {
            throw new Error("Unable to instantiate '"
                            + val
                            + "' defined in ResourceBundle "
                            + SequenceFormatTesterDefaultImpl.class.getName());
        }
        catch (IllegalAccessException iae)
        {
            throw new Error("Unable to access constructor of '"
                            + val
                            + "' defined in ResourceBundle "
                            + SequenceFormatTesterDefaultImpl.class.getName());
        }
    }

    public static SequenceFormatTester getInstance()
    {
        return INSTANCE;
    }

    public SequenceFormat getSequenceFormat(File seqFile) throws BioException
    {
        Object matching = null;

        try
        {
            BufferedReader br = new BufferedReader(new 
                InputStreamReader(new FileInputStream(seqFile)));

            String line = br.readLine();

            br.close();
            br = null;

            for (Iterator fi = formats.keySet().iterator(); fi.hasNext();)
            {
                String diagnostic = (String) fi.next();

                if (line.startsWith(diagnostic))
                {
                    matching = formats.get(diagnostic);
                    break;
                }
            }
        }
        catch (IOException ioe)
        {
            throw new BioException(ioe, "An error ocurred while trying to determine the format of file "
                                   + seqFile.getName());
        }

        if (matching == null)
            throw new BioException("File "
                                   + seqFile.getName()
                                   + " is in an unknown format");

        return (SequenceFormat) matching;
    }
}
