package Bio::PSU::Utils::GenericPSU_to_GO;

use strict;
use Data::Dumper;
use Carp;

use Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper;

my $debug = 1;

sub new {
  my $invocant = shift;
  my $class    = ref ($invocant) || $invocant;
  if(@_ % 2) {
    croak "Default options must be name => value pairs (odd number supplied)";
  }

  my $self = {
	      @_,
	     };
  
  bless ($self, $class);
  return $self;
}

sub embl_parsing {
  my $self = shift;
  my ($seq, $tigr_format, $db_name, $taxon_id, $assigned_by, $regex) = @_;
  
  my $id_helper =
    Bio::PSU::Utils::FeatureIdHelper::BioPSUHelper->new(-selection_regex => $regex,
							-verbose => 0);
  
  my $time;
  
  {
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime;
    
    $year += 1900;
    
    $time = sprintf "%s%2.2d%2.2d", $year, $mon, $mday;
  }
  
  if (defined $seq) {
    my @features = ();
    
    for my $feature ($seq->features) {
      next unless (($feature->key () eq "CDS") || ($feature->key () eq "mRNA"));
      
      if ((not $tigr_format) && $feature->qexists ('pseudo')) {
	if ($debug) {
	  print STDERR "pseudogene!\n";
	}
	next;
      }

      # processing systematic id

      my $systematic_id = undef;
      
      eval {
	$systematic_id = $id_helper->get_systematic_id($feature);
      };
      
      if ($@) {
	my $loc = join ",", map {
	  $_->{start} . ".." . $_->{end}
	} $feature->ranges ();
	
	print STDERR "Error in CDS at: $loc\n";
	print STDERR "A Systematic Id is missing !!\n";
	next;
      }

      if (defined $systematic_id) {
	if ($debug) {
	  print STDERR "\nprocessing gene, $systematic_id...\n";
	}
      }
      
      # processing standard name

      my $primary_name = undef;

      eval {
	$primary_name = $id_helper->get_primary_name($feature);
      };

      if (not defined $primary_name) {
	if ($debug) {
	  print STDERR "no gene primary name at: " . get_range_string($feature) . "\n";
	  print STDERR "the primary name will be given as the systematic name\n";
	}
	$primary_name = $systematic_id;
      }

      # processing product

      my $product = getProduct ($feature);

      my ($GO_components, $GO_functions, $GO_processes) = getGOTerms ($feature);

      # processing synonyms

      my $synonym = getSynonyms ($feature);
	
      if ($debug) {
	print STDERR "\nprocessing GO qualifiers...\n\n";
      }

      # processing GO qualifiers ...

      my @all_go_qualifiers = ();
      
      push @all_go_qualifiers, map { "component; $_" } @$GO_components;
      push @all_go_qualifiers, map { "function; $_" } @$GO_functions;
      push @all_go_qualifiers, map { "process; $_" } @$GO_processes;
      
      if ($debug) {
	print STDERR "go qualifiers: " . @all_go_qualifiers . "\n";
      }

      for my $go_qualifier (@all_go_qualifiers) {

	if ($debug) {
	  print STDERR "\ngo qualifier: $go_qualifier\n";
	}
	
	# $go_qualifier =~ s/^\s*(.*)\s*$/$1/;
	
	my ($ontology, $go, $evidence_code, $ev_code_db_ids, $the_rest) =
	  split /\s*;\s*/, $go_qualifier, 5;
	
	if ($debug) {
	  print STDERR "go: $go\n";
	}

	my $go_id = undef;
	my $go_name = undef;
	if ($go =~ /^(GO\:\d+)\s*\((.*)/) {
	  $go =~ /^(GO\:\d+)\s*\((.*)/;
	  
	  $go_id = $1;
	  $go_name = undef;
	  if (defined $2) {
	    $go_name = $2;
	    if ($go_name =~ /\)$/) {
	      chop ($go_name);
	    }
	  }
	}
	else {
	  $go_id = $go;
	}
	
	if (!defined $go_id) {
	  if ($debug) {
	    warn "no go id in /GO_$ontology qualifier for $systematic_id\n";
	  }
	  next;
	}
	else {
	  if ($debug) {
	    print STDERR "got go id, $go_id, for ontology, $ontology\n";
	  }
	}
	
	if (!defined $evidence_code) {
	  if ($debug) {
	    warn "no evidence_code in /GO_$ontology qualifier for $systematic_id\n";
	  }
	  next;
	}
	else {
	  if ($debug) {
	    print STDERR "got evidence code, $evidence_code.\n";
	  }
	}
	
	my $db_ref_id = "";
	if (!defined $ev_code_db_ids) {
	  if ($debug) {
	    warn "no evidence_code db ids in /GO_$ontology qualifier for $systematic_id\n";
	  }
	  next;
	}
	else {
	  if ($debug) {
	    print STDERR "got evidence code db ids, $ev_code_db_ids\n";
	  }

	  ############################################
	  #
	  # Cleaning the Reference column
	  #
	  ############################################
	  
	  

	  if ($ev_code_db_ids =~ /date/) {
	    # date parsing
	    $ev_code_db_ids =~ /date\:(\d+)/;
	    $time = $1;
	  }
	  
	  # ND Evidence code, make sure it has GO_REF:nd in the reference column
	  
	  if ($evidence_code eq "ND") {
	    if ($debug) {
	      print STDERR "cleaned ND Reference for gene, $systematic_id!\n";
	    }
	    $ev_code_db_ids = "GO_REF:nd";
	    $db_ref_id = $ev_code_db_ids;
	  }
	  else {
	    $db_ref_id = $ev_code_db_ids;
	    
	    # a Reference column must have a datasource:id syntax
	    # if not, get rid of it !!
	    if (not $db_ref_id =~ /\:/) {
	      $db_ref_id = "";
	    }
	    
	    # Substitute GOC for PMID
	    #if ($db_ref_id =~ /^GOC\:/) {
	    #  print STDERR "Replacing GOC Reference by PMID!!\n";
	    #  $db_ref_id =~ s/^GOC/PMID/;
	    #}

	    ########################################
	    # cleaning SWALL DB Reference
	    #
	    ########################################
	    
	    if ($db_ref_id =~ /^SWALL\:\D+\|.+/) {

	      if ($debug) {
		print STDERR "Ref. is not clean at all!!!\n";
		print STDERR "Ref: $db_ref_id\n";
	      }

	      # get rid of SWALL:
	      $db_ref_id =~ s/^SWALL\://;
	      # replace th fist '|' occurrence by a ':'
	      $db_ref_id =~ s/\|/\:/;

	      if ($debug) {
		print STDERR "Ref. should be a bit cleaner now !!!\n";
		print STDERR "Ref: $db_ref_id\n";
	      }

	    }
	    
	    if ($db_ref_id =~ /\s\(/) {
	      # get rid of the '()'
	      $db_ref_id =~ s/\s\(/\|/g;
	      $db_ref_id =~ s/\)//g;

	      $db_ref_id =~ s/\(//;

	      if ($debug) {
		print STDERR "Ref. should be so much cleaner now !!!\n";
		print STDERR "Ref: $db_ref_id\n";
	      }

	    }

	    if ($db_ref_id =~ /TAIR\:TAIR\:gene\:/) {
	      # just keep TAIR once !!
	      $db_ref_id =~ s/TAIR\:TAIR\:gene\:/TAIR\:/g;

	      if ($debug) {
		print STDERR "ohhhh and imagine now !!!\n";
		print STDERR "Ref: $db_ref_id\n";
	      }

	    }

	    if ($db_ref_id =~ /MGI\:MGI\:/) {
	      # just keep MGI once !!
	      $db_ref_id =~ s/MGI\:MGI\:/MGI\:/g;

	      if ($debug) {
		print STDERR "wasn't fully done yet !!!\n";
		print STDERR "Ref: $db_ref_id\n";
	      }

	    }

	    # Clean FB in tryp Tb chr. 1
	    # get rid of the ()

	    $db_ref_id =~ s/\s\(/\|/g;
	    $db_ref_id =~ s/\)//g;
	    $db_ref_id =~ s/\(//;

	  }
	}

	my $ontology_id;
	
	if ($ontology =~ /^([pcf])/i) {
	  $ontology_id = uc $1;
	} else {
	  if ($debug) {
	    die "internal error - can not figure out the ontology!!\n";
	  }
	}
	
	# parsing the rest !

	########################################
	#
	## source information is deprecated
	#
	########################################

	my $source_ev_code;
	my $source_db_id;

	if (defined $the_rest && $the_rest =~ /source \((\S+)\s*;\s*(\S+)\)/) {
	  $source_ev_code = $1;
	  $source_db_id = $2;
	}
	
	########################################
	
	my $with_or_from = "";
	
	# a with_or_from column must have a datatype:id syntax
	# if not, get rid of it !!
	if (not $with_or_from =~ /\:/) {
	  $with_or_from = "";
	}
	
	if (defined $the_rest && $the_rest =~ /date/) {
	  # date parsing
	  $the_rest =~ /date\:(\d+)/;
	  $time = $1;
	}
	elsif (defined $the_rest && $the_rest =~ /GO\:/) {
	  # GO From field parsing
	  $the_rest =~ /(GO\:\d+)/;
	  $with_or_from = $1;
	}
	elsif (defined $the_rest && $the_rest =~ /^\w+\:/) {
	  # With field parsing

	  # keep it, can have more than one reference !!

	  # $the_rest =~ /(\w+\:\d+)/;
	  # $with_or_from = $1;
	}
	
	##################
	# NOT information
	##################

	my $not_flag = 0;
	
	if ($go_id =~ s/^NOT //) {
	  $not_flag = 1;
	}
	
	my $db_object_id     = $systematic_id;
	my $db_object_symbol = undef;
	if (defined $primary_name) {
	  $db_object_symbol = $primary_name;
	}
	else {
	  $db_object_symbol = $systematic_id;
	}
	
	print "$db_name\t$db_object_id\t$db_object_symbol\t\t$go_id\t$db_ref_id\t$evidence_code\t$with_or_from\t$ontology_id\t$product\t$synonym\tgene\ttaxon:$taxon_id\t$time\t$assigned_by\n";
      }
    }
  } else {
    if ($debug) {
      die "End of file while reading tab file\n";
    }
  }

}

#########
# END
##

sub get_range_string
{
  my ($feature) = shift;
  my @ranges = map {
    $_->{start} . ".." . $_->{end}
  } $feature->ranges ();

  return join ",", @ranges;
}


sub getSystematicId {
  my ($feature) = @_;
  
  my $systematic_id = undef;
  if ($feature->systematic_id ()) {
    $systematic_id = $feature->systematic_id ();
  }
  elsif ($feature->temporary_systematic_id ()) {
    $systematic_id = $feature->temporary_systematic_id ();
  }
  else {
    my @genes = $feature->gene ();
    $systematic_id = $genes[0];
  }

  return $systematic_id;
}

sub getPrimaryName {
  my ($feature) = @_;

  my $primary_name = undef;
  
  if ($feature->primary_name) {
    $primary_name = $feature->primary_name;
  }
  else {
    my @genes = $feature->gene ();
    if (@genes>1) {
      $primary_name = $genes[1];
    }
  }

  return $primary_name;
}


sub getProduct {
  my ($feature) = @_;

  my $product = "";
  
  if ($feature->qexists ('product')) {
    my @products = $feature->qvalues ('product');
    
    # concatenate all the product qualifiers
    # $product = join "|", @products;
    
    # NOW just pick the first one !!!
    $product = $products[0];
  }
  else {
    if ($debug) {
      print STDERR "no /product at: " . get_range_string ($feature);
    }
  }
  
  return $product;
  
}

sub getSynonyms {
  my ($feature) = @_;
  my $synonym = "";

  if ($feature->qexists ('synonym')) {
    my @synonyms = $feature->qvalues('synonym');
    $synonym = join "|", @synonyms;
  }
  else {
    if ($debug) {
      print STDERR "no /synonym at: " . get_range_string ($feature);
    }
  }

  return $synonym;

}

sub getGOTerms {
  my ($feature) = @_;
  
  my @GO_component = ();
  
  if ($feature->qexists ("GO_component")) {
    @GO_component = $feature->qvalues ("GO_component");
  }
  
  my @GO_function = ();
  
  if ($feature->qexists ("GO_function")) {
    @GO_function = $feature->qvalues ("GO_function")
  }
  
  my @GO_process = ();
  
  if ($feature->qexists ("GO_process")) {
    @GO_process = $feature->qvalues ("GO_process");
  }

  return (\@GO_component, \@GO_function, \@GO_process);
  
}

1;
