package uk.ac.sanger.psu.gfmerge.analysis;

import java.util.*;
import java.io.*;
import java.util.regex.*;

import org.biojava.bio.*;
import org.biojava.bio.seq.*;
import org.biojava.bio.symbol.*;
import org.biojava.bio.seq.impl.*;
import org.biojava.bio.seq.io.*;

import uk.ac.sanger.psu.gfmerge.*;
import uk.ac.sanger.psu.gfmerge.datacomp.*;
import uk.ac.sanger.psu.gfmerge.util.io.*;
import uk.ac.sanger.psu.gfmerge.util.getopt.*;
import uk.ac.sanger.psu.gfmerge.util.biotools.*;
import uk.ac.sanger.psu.gfmerge.util.algotools.*;
import uk.ac.sanger.psu.gfmerge.util.debug.*;
import uk.ac.sanger.psu.gfmerge.analysis.*;
import uk.ac.sanger.psu.gfmerge.analysis.tools.*;
import uk.ac.sanger.psu.gfmerge.analysis.simfeature_analysis.*;

/** class which wraps analysis of gene predictions
 * <p>
 * Gene predicter, prediction, genemodel region, cDNA region, Blast region objects are created within the class.
 * <br>
 * An region array of picked gene models is returned.
 *
 * @see uk.ac.sanger.psu.gfmerge.datacomp.GFMergeRegion
 * @see uk.ac.sanger.psu.gfmerge.datacomp.GeneModel
 * @see uk.ac.sanger.psu.gfmerge.datacomp.Prediction
 * @author Sebastian R. Spiegler
 * @version 1.0
 */
public class GFMerge_Analysis {

  // attributes
  /** ArrayList of gene predicter objects
   */
  private ArrayList genePredicters;
  /** ArrayList of prediction objects
   */
  private ArrayList predictions;
  /** ArrayList of genemodel regions which contains all of the original gene models
   */
  private ArrayList arrListOfGMRegions_org;
  /** ArrayList of genemodels from which low scoring genemodels are removed during analysis
   */
  private ArrayList arrListOfGMRegions;
  /** ArrayList of cDNA regions
   */
  private ArrayList arrListOfCDnaRegions;
  /** ArrayList of Blast regions
   */
  private ArrayList arrListOfBlastRegions;
  /** ArrayList which contains analysis objects
   * (<code>CDnaSpliceSiteAnalysis</code>, <code>CDnaOverlapAnalysis</code>, <code>BlastOverlapAnalysis</code>, <code>TotalExonLengthAnalysis</code>, <code>GeneModelLengthAnalysis</code>, <code>AvgCondProbAnalysis</code>)
   */
  private ArrayList analysisObjArr;
  /** return value of getMergedFeatureTable,
   * contains regions of picked genemodels
   */
  private ArrayList mergedFeatureTableArr;
  /** object which contains all command line parameters
   */
  private SimpleArg allArguments;

  private static int _aminoacidsLength = 40;

  // constructor
  /** constructor of GFMerge_Analysis objects
   * <p>
   * It takes a SimpleArg object as parameter which contains itself all
   * command line arguments.
   *
   * @param _allArguments object
   * @see uk.ac.sanger.psu.gfmerge.util.getopt.SimpleArg
   */

  public GFMerge_Analysis(SimpleArg _allArguments) {

    // set command line parameters
    setAllArguments(_allArguments);

    // Create GenePredicter objects and calculate their accuracy
    setGenePredicters(
        GFMerge_Analysis.createGeneFinderObjects(allArguments));
    // debug 102_1 - GenePredicter
    Debug.debug_102_1(allArguments.getDebug(), genePredicters);

    // Create Prediction objects
    setPredictions(
        GFMerge_Analysis.createPredictionObjects(allArguments, genePredicters));
    // debug 102_2 - prediction objects
    // within createPredictionObjects()
    // debug 102_3 - predictions
    Debug.debug_102_3(allArguments.getDebug(), predictions);

    // Create arraylist of gm regions
    setArrListOfGMRegions(
        GFMerge_Analysis.calcArrListOfGMRegions(predictions, allArguments));

    // Create arraylist of cDNA feature regions
    setArrListOfCDnaRegions(
        GFMerge_Analysis.calcArrListOfCDnaRegions(allArguments));

    // Create arraylist of Blast feature regions
    setArrListOfBlastRegions(
        GFMerge_Analysis.calcArrListOfBlastRegions(allArguments));

    // backup arrListOfGMRegions
    setArrListOfGMRegions_org(
        GFMerge_Analysis.copyArrOfFeatRegions(arrListOfGMRegions));

    // save array of all analysis objects
    setAnalysisObjArr(
        // create analysis objects and put them in array
        this.createAnalysisObjArr(this.getArrListOfGMRegions(),
                                  this.getArrListOfCDnaRegions(),
                                  this.getArrListOfBlastRegions(),
                                  this.getAllArguments()));
  }

  /** method which returns an array list of regions only containing the
   * genemodels which have been picked during the process of scoring
   * <p>
   * There should be no overlapping gene models anymore.
   *
   * @return <code>ArrayList</code> of regions
   * @see uk.ac.sanger.psu.gfmerge.datacomp.GFMergeRegion
   * @see uk.ac.sanger.psu.gfmerge.datacomp.GeneModel
   */
  public ArrayList getMergedFeatureTable() {

    System.out.println("--------------------------------------------------");
    System.out.println("Number of Blast hits");
    countFeat(arrListOfBlastRegions);

    System.out.println("--------------------------------------------------");
    System.out.println("Number of cDNA hits");
    countFeat(arrListOfCDnaRegions);

    System.out.println("--------------------------------------------------");
    System.out.println("Number of Gene models");
    countFeat(arrListOfGMRegions_org);

    if (mergedFeatureTableArr == null) {
      this.setMergedFeatureTableArr(this.calcMergedFeatureTable());

      System.out.println("--------------------------------------------------");
      System.out.println("Number of Merged Features");
      countFeat(mergedFeatureTableArr);

      return mergedFeatureTableArr;
    }
    else {

      System.out.println("--------------------------------------------------");
      System.out.println("Number of Merged Features");
      countFeat(mergedFeatureTableArr);

      return mergedFeatureTableArr;
    }


  }

  private void countFeat(ArrayList ArrListOfRegions)
  {
    long countedFeat  = 0;
    double avgNoOfExons  = 0;
    double avgFeatLength = 0;
    double avgFeatTotalLength = 0;

    for(Iterator itReg = ArrListOfRegions.iterator(); itReg.hasNext();)
    {
      GFMergeRegion region = (GFMergeRegion) itReg.next();
      for(Iterator itFeat = region.getLocateable().iterator(); itFeat.hasNext();)
      {
        GFMergeFeature feature = (GFMergeFeature) itFeat.next();
        countedFeat++;
        avgNoOfExons += feature.getExons().size();
        avgFeatLength += feature.getLength();
        avgFeatTotalLength += feature.getTotalExonLength();
      }
    }

    avgNoOfExons        = avgNoOfExons / countedFeat;
    avgFeatLength       = avgFeatLength / countedFeat;
    avgFeatTotalLength  = avgFeatTotalLength / countedFeat;

    System.out.println("Number of features : " + countedFeat);
    System.out.println("Avg Number of exons: " + avgNoOfExons);
    System.out.println("Avg feature length : " + avgFeatLength);
    System.out.println("Avg Total feat len : " + avgFeatTotalLength);
    System.out.println("--------------------------------------------------");
  }

  /** method which is called from getMergedFeatureTable if variable <code>mergedFeatureTableArr </code> is <code>null</code>. It calculates the high scoring genemodels which are picked during analysis.
   *
   * @return ArrayList of regions which contain picked genemodels
   */
  private ArrayList calcMergedFeatureTable() {
    ArrayList _arrListOfGMRegions = this.getArrListOfGMRegions();
    ArrayList _analysisObjArr = this.getAnalysisObjArr();
    ArrayList analysedGms = new ArrayList();
    int debug = this.getAllArguments().getDebug();

    // return array which is going to be modified
    analysedGms = _arrListOfGMRegions;
    //debug 701 - array of analyzed genemodels
    Debug.debug_701(debug, "original array", analysedGms);

    // loop over all analysis objects and do analysis
    for (Iterator itAnalysisObj = _analysisObjArr.iterator();
         itAnalysisObj.hasNext(); ) {
      // backup of region array
      ArrayList analysedGms_loopBak =
          GFMerge_Analysis.copyArrOfFeatRegions(analysedGms);

      // analysis object
      BasicFeatureAnalysisAble analysisObj =
          (BasicFeatureAnalysisAble) itAnalysisObj.next();

      // do analysis, get high scoring gm region array list
      analysedGms = analysisObj.getHighScoringGmRegArr(analysedGms);

      //debug 701 - array of analyzed genemodels
      Debug.debug_701(debug, (analysisObj.getClass()).toString(),
                      analysedGms);

      // SAVE removed genemodels (analysis level) in output file
      if (this.getAllArguments().doFileOfRemovedGms()) {
        GFMergeOutput.saveRemovedFeat2File(
            this.getAllArguments().getOutputFilename() + ".removed" + "."
            + analysisObj.getClassName() + ".tab",
            analysedGms_loopBak,
            analysedGms,
            (Prediction)this.getPredictions().get(0));
      }

    } // END FOR

    return analysedGms;
  }

  /** method which generated an ArrayList of analysis objects (<code>CDnaSpliceSiteAnalysis</code>, <code>CDnaOverlapAnalysis</code>, <code>BlastOverlapAnalysis</code>, <code>TotalExonLengthAnalysis</code>, <code>GeneModelLengthAnalysis</code>, <code>AvgCondProbAnalysis</code>)
   *
   * @param _arrListOfGMRegions ArrayList which contains region of genemodels
   * @param _arrListOfCDnaRegions ArrayList which contains region of cDNA
   *      objects
   * @param _arrListOfBlastRegions ArrayList which contains region of
   *      Blast objects
   * @param _allArguments object which contains all command line
   *     arguments
   * @return ArrayList of analysis objects (<code>BasicFeatureAnalysis</code>)
   */
  private ArrayList createAnalysisObjArr(ArrayList _arrListOfGMRegions,
                                         ArrayList _arrListOfCDnaRegions,
                                         ArrayList _arrListOfBlastRegions,
                                         SimpleArg _allArguments) {
    ArrayList analysisObjArr = new ArrayList();

    analysisObjArr.add(new CDnaSpliceSiteAnalysis(_arrListOfGMRegions,
                                                  _arrListOfCDnaRegions,
                                                  _allArguments));

    analysisObjArr.add(new CDnaOverlapAnalysis(_arrListOfGMRegions,
                                               _arrListOfCDnaRegions,
                                               _allArguments));

    analysisObjArr.add(new BlastOverlapAnalysis(_arrListOfGMRegions,
                                                _arrListOfBlastRegions,
                                                _allArguments));

    analysisObjArr.add(new TotalExonLengthAnalysis(_arrListOfGMRegions,
        _allArguments));

    analysisObjArr.add(new GeneModelLengthAnalysis(_arrListOfGMRegions,
        _allArguments));

    analysisObjArr.add(new AvgCondProbAnalysis(_arrListOfGMRegions,
                                               _allArguments));

    return analysisObjArr;
  }

  /**
   * accessor method which returns an array list of regions which correspond
   * to the original array used in the constructor
   *
   * @return original <code>ArrayList</code> of regions
   * @see uk.ac.sanger.psu.gfmerge.datacomp.GFMergeRegion
   * @see uk.ac.sanger.psu.gfmerge.datacomp.GeneModel
   */
  public ArrayList getArrListOfGMRegions_org() {
    return arrListOfGMRegions_org;
  }

  /** accessor method which returns an array list of prediction objects
   *
   * @return <code>ArrayList</code> of predictions
   * @see uk.ac.sanger.psu.gfmerge.datacomp.Prediction
   */
  public ArrayList getPredictions() {
    return predictions;
  }

  // copy ArrayList of regions, creating new regions objects pointing to the
  // same gm objects like in the original region object
  /** method which does a deep copy of an region array, it keeps the references to the genemodel objects
   *
   * @param arrListOfFeatRegions ArrayList of feature regions (cDNA, Blast)
   * @return ArrayList of GFMergeRegion objects
   */
  private static ArrayList copyArrOfFeatRegions(ArrayList arrListOfFeatRegions) {
    ArrayList arrListOfFeatRegions_copy = new ArrayList();

    for (Iterator itReg = arrListOfFeatRegions.iterator(); itReg.hasNext(); ) {
      GFMergeRegion region = (GFMergeRegion) itReg.next();
      Location regLocation = region.getLocation();
      ArrayList featArr = new ArrayList(region.getLocateable());
      GFMergeRegion region_copy = new GFMergeRegion(regLocation, featArr);
      arrListOfFeatRegions_copy.add(region_copy);
    }

    return arrListOfFeatRegions_copy;
  }

  ////////////////////////////////////////////////////////////////////////
  // create genefinder objects and calculate their accuracy
  //
  /** methode which creates genefinder objects and calculates their accuracy
   *
   * @param _allArguments object which contains all command line arguments
   * @return ArrayList of GenePredicter objects
   */
  private static ArrayList createGeneFinderObjects(SimpleArg _allArguments) {
    String filenameAnnotated = _allArguments.getAnnotatedTrainSet();
    ArrayList PredictionTrainSet = _allArguments.getPredictedTrainSet();

    ArrayList thisGenePredicters = new ArrayList();

    // get annotated sequence
    SequenceIterator sequencesAnnotated = GFMergeInput.getSeq(filenameAnnotated);

    Sequence seqAnnotated = null;
    try
    {
      seqAnnotated = BioTools.seqs2Seq(sequencesAnnotated);
    }
    catch (BioException ex)
    {
      //not in EMBL format
      System.err.println("Error in file " + filenameAnnotated);
      ex.printStackTrace();
    }

    // create genefinder objects
    for (Iterator it = PredictionTrainSet.iterator(); it.hasNext(); ) {
      String filename = (String) it.next();
      SequenceIterator sequences = GFMergeInput.getSeq(filename);


      Sequence seqGenePredicter = null;
      try
      {
        seqGenePredicter = BioTools.seqs2Seq(sequences);
      }
      catch (BioException ex)
      {
        //not in EMBL format
        System.err.println("Error in file " + filename);
        ex.printStackTrace();
      }

      GenePredicter genePredicter = new GenePredicter(seqAnnotated,
          seqGenePredicter,
          filename);
      thisGenePredicters.add(genePredicter);
    }
    return thisGenePredicters;
  }

  ////////////////////////////////////////////////////////////////////////
  // create prediction objects
  //
  /** method which creates <code>Prediction</code> objects
   *
   * @param _allArguments object which contains all command line arguments
   * @param _genePredicters ArrayList which contains all genepredicters
   * @return ArrayList of Prediction objects
   */
  private static ArrayList createPredictionObjects(SimpleArg _allArguments,
      ArrayList _genePredicters) {
    ArrayList PredictionFiles = _allArguments.getPredictions();
    ArrayList thisPredictions = new ArrayList();

    for (Iterator it = PredictionFiles.iterator(); it.hasNext(); ) {
      String filename = (String) it.next();
      SequenceIterator sequences = GFMergeInput.getSeq(filename);


      Sequence seq = null;
      try
      {
        seq = BioTools.seqs2Seq(sequences);
      }
      catch (BioException ex)
      {
        //not in EMBL format
        System.err.println("Error in file " + filename);
        ex.printStackTrace();
      }

      // filter sequence, take only CDS features
      // Also check for stop codons sanity within the ORF
      // Also filter on aminoacids length
      Sequence onlyCdsSeq = GFMerge_Analysis.removeNonCdsFeatures(seq, _aminoacidsLength);

      Prediction pred = new Prediction(onlyCdsSeq, filename, _genePredicters);
      thisPredictions.add(pred);

      //debug 102_2 - prediction objects
      Debug.debug_102_2(_allArguments.getDebug(), pred.getGenePredicterName(),
                        filename);
    }
    return thisPredictions;
  }

  ////////////////////////////////////////////////////////////////////////
  // filter sequence, take only CDS features
  // This methods also check for protein sequence consistency, 
  // if the protein sequence contains a stop codon or if it is below an aminoacids length threshold, it will be removed.
  //
  /** method which removes non-CDS features from sequence
   *
   * @param sequence BioJava sequence object
   * @return cleaned sequence with only CDS features
   */
  private static Sequence removeNonCdsFeatures(Sequence sequence, int _aminoacidsLength)
  {
    // get CDS filter
    FeatureFilter cdsFf  = new FeatureFilter.ByType("CDS");

    FeatureHolder cdsFh = sequence.filter(cdsFf);

    for(Iterator itFeat = cdsFh.features(); itFeat.hasNext();)
    {
      Feature feature = (Feature) itFeat.next();
      
      ///////////////////////////////////////////////
      // Check also for stop codons !!!
      
      SymbolList CDSSequence = feature.getSymbols();

      // Transcription
      
      SymbolList rnaForm = null;
      try {
	  rnaForm = RNATools.transcribe (CDSSequence);
      } catch (IllegalAlphabetException exp) {
	  System.err.println ("Cannot transcribe the given sequence, wrong alphabet !");
      }
      
      // Translation
      
      SymbolList proteinForm = null;
      try {
	  proteinForm = RNATools.translate (rnaForm);
	  
	  // Checking
	  
	  String proteinSequence = proteinForm.seqString();
	  
	  // Check the protein length

	  if (proteinSequence.length () < _aminoacidsLength) {
	      try
		  {
		      sequence.removeFeature(feature);
		  }
	      catch (Exception ex)
		  {
		      System.err.println("Error while filtering CDS features");
		      System.exit( -1);
		  }
	  }
	  else {
	      // System.err.println ("protein sequence fine!");
	  }

	  StringTokenizer token = new StringTokenizer (proteinSequence, "*");
	  if (token.countTokens() > 1) {
	      Location location = feature.getLocation();
	      System.out.println ("has a stop codon!!");
	      System.out.println ("This CDS (" + location.getMin() + "," + location.getMax() + " will not be processed!");
	      // remove it
	      try
		  {
		      sequence.removeFeature(feature);
		  }
	      catch (Exception ex)
	      {
		  System.err.println("Error while filtering CDS features");
		  System.exit( -1);
	      }
	  }
	  else {
	      // System.err.println ("protein sequence fine!");
	  }
      } catch (IllegalAlphabetException exp) {
	  System.err.println ("Cannot transcribe the given sequence, wrong alphabet !");
	  System.err.println ("IllegalAlphabetException - RNA sequence: " + rnaForm.seqString() + ".");
      } catch (IllegalArgumentException exp) {
	  //System.err.println ("IllegalArgumentException - Cannot transcribe the given sequence, sequence length not divided by 3 ?!");
	  //String rnaSequence = rnaForm.seqString();
	  //int length = rnaSequence.length();
	  //System.err.println ("RNA sequence: " + rnaSequence);
	  //System.err.println ("RNA sequence length: " + length + ".");
	  //Location location = feature.getLocation();
	  //System.err.println ("min/max: " + location.getMin() + ", " + location.getMax());

	  // The gene is truncated, that's fine (could be because of unfinished sequence contig), keep this CDS.

      }

      //////////////////////////////////////////////

    }

    // get non CDS filter
    FeatureFilter nonCdsFf = new FeatureFilter.Not(cdsFf);

    // get non CDS FeatureHolder
    FeatureHolder nonCdsFh = sequence.filter(nonCdsFf);

    for(Iterator itFeat = nonCdsFh.features(); itFeat.hasNext();)
    {
      Feature feature = (Feature) itFeat.next();

      try
      {
        sequence.removeFeature(feature);
      }
      catch (Exception ex)
      {
        System.err.println("Error while filtering CDS features");
        System.exit( -1);
      }
    }
    return sequence;
  }





  ////////////////////////////////////////////////////////////////////////
  // get ArrayList of gm regions
  //
  /** method which calulates prediction objects
   *
   * @param _predictions ArrayList which contains all prediction objects
   * @param _allArguments object which contains all command line
   *     arguments
   * @return ArrayList of Prediction objects
   */
  private static ArrayList calcArrListOfGMRegions(ArrayList _predictions,
                                                  SimpleArg _allArguments) {
    ArrayList thisArrListOfGMRegions = new ArrayList();

    // Create tree structure of gene models
    // (gms with same min location in array)
    MyTreeMap allGMTreeMap = new MyTreeMap();
    allGMTreeMap = GFMerge_Analysis.createTreeOfGMs(_predictions, allGMTreeMap);
    // debug - d 103 => gm treemap
    Debug.debug_103(_allArguments.getDebug(), "GM TreeMap", allGMTreeMap);

    thisArrListOfGMRegions = BioTools.search(allGMTreeMap);
    // debug - d 104 => regions of gene models
    Debug.debug_104(_allArguments.getDebug(), thisArrListOfGMRegions,
                    allGMTreeMap);

    return thisArrListOfGMRegions;
  }

  ////////////////////////////////////////////////////////////////////////
  // Create tree structure of gene models
  // (gms with same min location in array)
  //
  /** method which creates a tree structure of genemodels which are sorted by their minimum location
   *
   * @param _predictions ArrayList which contains all prediction objects
   * @param _allGmTreeMap empty MyTreeMap which will contain the return
   *      value
   * @return MyTreeMap object
   */
  private static MyTreeMap createTreeOfGMs(ArrayList _predictions,
                                           MyTreeMap _allGmTreeMap) {

    for (Iterator itPred = _predictions.iterator(); itPred.hasNext(); ) {
      Prediction pred = (Prediction) itPred.next();
      BioTools.addToTreeMap(_allGmTreeMap, pred.getGeneModels());
    }

    return _allGmTreeMap;
  }

  ////////////////////////////////////////////////////////////////////////
  // get ArrayList of cDNA feature regions
  //
  /** method which creates ArrayList of cDNA regions
   *
   * @param _allArguments object which contains all command line arguments
   * @return ArrayList of cDNA regions
   */
  private static ArrayList calcArrListOfCDnaRegions(SimpleArg _allArguments) {
    ArrayList _ArrListOfCDnaRegions = new ArrayList();

    // Create cDNA objects
    MyTreeMap allCDnaTreeMap = new MyTreeMap();
    allCDnaTreeMap = GFMerge_Analysis.createSimFeatObjects(_allArguments.
        getCDNAFile(),
        allCDnaTreeMap);
    // debug - d 105 => cDNA treemap
    Debug.debug_105(_allArguments.getDebug(), "cDNA TreeMap", allCDnaTreeMap);

    // Create array list of cDNA regions
    _ArrListOfCDnaRegions = BioTools.search(allCDnaTreeMap);
    // debug - d 106 => regions of cDNA hits
    Debug.debug_106(_allArguments.getDebug(), "cDNA", _ArrListOfCDnaRegions,
                    allCDnaTreeMap);

    return _ArrListOfCDnaRegions;
  }

  ////////////////////////////////////////////////////////////////////////
  // get ArrayList of Blast feature regions
  //
  /** method which creates ArrayList of Blast regions
   *
   * @param _allArguments object which contains all command line arguments
   * @return ArrayList of Blast regions
   */
  private static ArrayList calcArrListOfBlastRegions(SimpleArg _allArguments) {
    ArrayList _ArrListOfBlastRegions = new ArrayList();

    // Create Blast objects
    MyTreeMap allBlastTreeMap = new MyTreeMap();
    allBlastTreeMap = GFMerge_Analysis.createSimFeatObjects
        (_allArguments.getBlastFile(),
         allBlastTreeMap);
    // debug - d 105 => Blast treemap
    Debug.debug_105(_allArguments.getDebug(), "Blast TreeMap", allBlastTreeMap);

    // Create array list of blast regions
    _ArrListOfBlastRegions = BioTools.search(allBlastTreeMap);
    // debug - d 106 => regions of blast hits
    Debug.debug_106(_allArguments.getDebug(), "Blast", _ArrListOfBlastRegions,
                    allBlastTreeMap);

    return _ArrListOfBlastRegions;
  }

  ////////////////////////////////////////////////////////////////////////
  // Creates sim feature objects
  //
  /** method which creates similarity feature objects (cDNA, Blast), puts them in a MyTreeMap object and returns this tree map
   *
   * @param featFilename filename of the embl file which contains features
   * @param _allFeatTreeMap empty MyTreeMap which will contain the return
   *      value
   * @return MyTreeMap of similarity features (cDNA, Blast)
   */
  private static MyTreeMap createSimFeatObjects(String featFilename,
                                                MyTreeMap _allFeatTreeMap) {
    SequenceIterator sequences = GFMergeInput.getSeq(featFilename);


    Sequence sequence = null;
    try
    {
      sequence = BioTools.seqs2Seq(sequences);
    }
    catch (BioException ex)
    {
      //not in EMBL format
      System.err.println("Error in file " + featFilename);
      ex.printStackTrace();
    }

    ArrayList featArrList = new ArrayList();

    for (Iterator it = sequence.features(); it.hasNext(); ) {
      StrandedFeature feature = (StrandedFeature) it.next();
      SimFeature simFeat = new SimFeature(feature);
      featArrList.add(simFeat);
    }
    BioTools.addToTreeMap(_allFeatTreeMap, featArrList);

    return _allFeatTreeMap;
  }

  // accessor methods
  // get methods
  /** accessor method which returns ArrayList of genemodel regions
   *
   * @return ArrayList of genemodel regions
   */
  public ArrayList getArrListOfGMRegions() {
    return this.arrListOfGMRegions;
  }

  /** accessor method which returns ArrayList of cDNA regions
   *
   * @return ArrayList of cDNA regions
   */
  private ArrayList getArrListOfCDnaRegions() {
    return this.arrListOfCDnaRegions;
  }

  /** accessor method which return ArrayList of Blast regions
   *
   * @return ArrayList of Blast regions
   */
  private ArrayList getArrListOfBlastRegions() {
    return this.arrListOfBlastRegions;
  }

  /** accessor method which returns ArrayList of analysis objects generated in calcAnalysisObjArr
   * (<code>CDnaSpliceSiteAnalysis</code>, <code>CDnaOverlapAnalysis</code>, <code>BlastOverlapAnalysis</code>, <code>TotalExonLengthAnalysis</code>, <code>GeneModelLengthAnalysis</code>, <code>AvgCondProbAnalysis</code>)
   *
   * @return ArrayList of analysis objects
   */
  private ArrayList getAnalysisObjArr() {
    return this.analysisObjArr;
  }

  /** accessor method which returns the object containg all command line parameters
   *
   * @return SimpleArg object
   */
  private SimpleArg getAllArguments() {
    return allArguments;
  }

  // set methods
  /** accessor method which sets ArrayList of genemodel regions
   *
   * @param _arrListOfGMRegions ArrayList which contains region of  genemodel
   *     objects
   */
  private void setArrListOfGMRegions(ArrayList _arrListOfGMRegions) {
    this.arrListOfGMRegions = _arrListOfGMRegions;
  }

  /** accessor method which sets ArrayList of cDNA regions
   *
   * @param _arrListOfCDnaRegions ArrayList of feature cDNA regions
   */
  private void setArrListOfCDnaRegions(ArrayList _arrListOfCDnaRegions) {
    this.arrListOfCDnaRegions = _arrListOfCDnaRegions;
  }

  /** accessor method which sets ArrayList of Blast regions
   *
   * @param _arrListOfBlastRegions ArrayList of feature Blast regions
   */
  private void setArrListOfBlastRegions(ArrayList _arrListOfBlastRegions) {
    this.arrListOfBlastRegions = _arrListOfBlastRegions;
  }

  /** accessor method which sets SimpleArg object containing all command line parameters
   *
   * @param _allArguments object which contains all command line  arguments
   */
  private void setAllArguments(SimpleArg _allArguments) {
    this.allArguments = _allArguments;
  }

  /** accessor method which sets ArrayList of prediction objects
   *
   * @param _predictions ArrayList of prediction objects
   */
  private void setPredictions(ArrayList _predictions) {
    this.predictions = _predictions;
  }

  /** accessor method which sets ArrayList of genepredicter objects
   *
   * @param _genePredicters ArrayList of genepredicter objects
   */
  private void setGenePredicters(ArrayList _genePredicters) {
    this.genePredicters = _genePredicters;
  }

  /** accessor method which sets ArrayList of original genemodel regions
   *
   * @param _arrListOfGMRegions_org ArrayList of original genemodel regions
   */
  private void setArrListOfGMRegions_org(ArrayList _arrListOfGMRegions_org) {
    this.arrListOfGMRegions_org = _arrListOfGMRegions_org;
  }

  /** accessor method which sets ArrayList of analysis objects generated in calcAnalysisObjArr
   * (<code>CDnaSpliceSiteAnalysis</code>, <code>CDnaOverlapAnalysis</code>, <code>BlastOverlapAnalysis</code>, <code>TotalExonLengthAnalysis</code>, <code>GeneModelLengthAnalysis</code>, <code>AvgCondProbAnalysis</code>)
   *
   * @param _analysisObjArr ArrayList of analysis objects
   */
  private void setAnalysisObjArr(ArrayList _analysisObjArr) {
    this.analysisObjArr = _analysisObjArr;
  }

  /** accessor method which sets mergedFeatureTableArr with ArrayList of regions which contain high scoring genemodels
   *
   * @param _mergedFeatureTableArr ArrayList of regions which contain high
   *     scoring genemodels
   */
  private void setMergedFeatureTableArr(ArrayList _mergedFeatureTableArr) {
    mergedFeatureTableArr = _mergedFeatureTableArr;
  }

}
