package uk.ac.sanger.psu.gfmerge.analysis.simfeature_analysis.tools;

import java.util.*;

import org.biojava.bio.symbol.*;
import org.biojava.bio.seq.*;

import uk.ac.sanger.psu.gfmerge.datacomp.*;
import uk.ac.sanger.psu.gfmerge.util.algotools.*;
import uk.ac.sanger.psu.gfmerge.util.debug.*;

/** class which contains computing tools for CDnaSpliceSiteAnalysis.
 *
 * @author Sebastian R. Spiegler
 * @version 1.0
 */
public class CDnaSpliceSiteAnalysisTools extends BasicFeatureAnalysisTools {

  /** method which returns HashMap of genemodels pointing to their score using a cDNA splice site scoring system.
   *
   * @param debug debugging mode
   * @param fuzzyCDnaBoundaries contains allowed difference of cDNA and
   *      genemodel splice site
   * @param considerStrand switch for strand consideration
   * @param _arrListOfGMRegions ArrayList of genemodel regions
   * @param _arrListOfCDnaRegions ArrayList of cDNA regions
   * @return HashMap of genemodels pointing to their cDNA splice site score
   */
  public static HashMap getCDnaSpliceSiteScoredGmHash(int debug,
      int fuzzyCDnaBoundaries,
      boolean considerStrand,
      ArrayList _arrListOfGMRegions,
      ArrayList _arrListOfCDnaRegions) {
    HashMap cDnaSpliceSiteScoredGmHash = new HashMap();


    // getCDnaScoredSpliceSiteHash (get hashCDnaOverlappedGms)
    HashMap hashCDnaOverlappedGms =
        FeatureOverlapAnalysisTools.getGmHashOfOverlappedFeat(debug,
        considerStrand,
        _arrListOfGMRegions,
        _arrListOfCDnaRegions);
    //debug 201 - hash of by cDNAs overlapped gms
    Debug.debug_201(debug, hashCDnaOverlappedGms);


    // get scores for gm splice sites
    cDnaSpliceSiteScoredGmHash =
        CDnaSpliceSiteAnalysisTools.getCDnaScoredSpliceSiteHash(debug,
        fuzzyCDnaBoundaries, hashCDnaOverlappedGms);

    //debug 203 - hash of cDNA scored gms
    Debug.debug_203(debug, "CDnaSpliceSite - cDNA scored gms",
                    cDnaSpliceSiteScoredGmHash);

    return cDnaSpliceSiteScoredGmHash;
  }


  /** method which gets scored splice sites of genemodels using cDNA evidence.
   *
   * @param debug debugging mode
   * @param fuzzyCDnaBoundaries contains allowed difference of cDNA and
   *      genemodel splice site
   * @param hashCDnaOverlappedGms HashMap of genemodels pointing to
   *      ArrayList of overlapping cDNAs
   * @return HashMap of genemodels pointing their splice site score
   */
  private static HashMap getCDnaScoredSpliceSiteHash(
      int debug, int fuzzyCDnaBoundaries, HashMap hashCDnaOverlappedGms) {
    HashMap cDnaScoredSpliceSiteHash = new HashMap();

    for (Iterator itGm = hashCDnaOverlappedGms.keySet().iterator();
         itGm.hasNext(); ) {
      GeneModel gm = (GeneModel) itGm.next();
      ArrayList featArr = (ArrayList) hashCDnaOverlappedGms.get(gm);

      Double score =
          new Double(CDnaSpliceSiteAnalysisTools.getCDnaScoredSpliceSiteForGm(
          debug, fuzzyCDnaBoundaries, gm, featArr));
      cDnaScoredSpliceSiteHash.put(gm, score);
    }
    return cDnaScoredSpliceSiteHash;
  }



  /** methode which calculates the score for one gm (either intron boundaries or splice sites).
   *
   * @param debug debugging mode
   * @param fuzzyCDnaBoundaries contains allowed difference of cDNA and
   *      genemodel splice site
   * @param gm genemodel
   * @param featureArr ArrayList of overlapping SimFeature objects
   * @return score
   */
  private static double getCDnaScoredSpliceSiteForGm(int debug,
      int fuzzyCDnaBoundaries, GeneModel gm, ArrayList featureArr) {
    double gmScore = 0;
    MyTreeMap minPosTree = new MyTreeMap();
    MyTreeMap maxPosTree = new MyTreeMap();

    // get cDNA scored splice sites
    boolean isMinPosTree = true;
    minPosTree =
        CDnaSpliceSiteAnalysisTools.getGmSpliceSiteTree(debug,
        fuzzyCDnaBoundaries, isMinPosTree, gm,featureArr);

    isMinPosTree = false; // MaxPosTree
    maxPosTree =
        CDnaSpliceSiteAnalysisTools.getGmSpliceSiteTree(debug,
        fuzzyCDnaBoundaries, isMinPosTree, gm,featureArr);

    // compare content of record objects and calculate score for gm
    gmScore =
        CDnaSpliceSiteAnalysisTools.calcCDnaSpliceSiteGmScore(minPosTree,
        maxPosTree,featureArr);

    return gmScore;
  }


  /** method which creates a tree of either max or min splice sites of the genemodel and overlapping features.
   * The tree contains RecordGmCDna objects which are used for finding overlapping splice sites of genemodels and cDNA features
   *
   * @param debug debugging mode
   * @param fuzzyCDnaBoundaries contains allowed difference of cDNA and
   *      genemodel splice site
   * @param isMinPosTree switch for consideration of either start or end
   *      position of the feature
   * @param gm genemodel
   * @param featureArr ArrayList of overlapping SimFeature objects
   * @return tree of RecordGmCDna objects
   */
  private static MyTreeMap getGmSpliceSiteTree(int debug,
                                               int fuzzyCDnaBoundaries,
                                               boolean isMinPosTree,
                                               GeneModel gm,
                                               ArrayList featureArr) {
    MyTreeMap positionTree = new MyTreeMap();

    // add GENEMODEL exon boundaries to treemap
    positionTree =
        CDnaSpliceSiteAnalysisTools.putGmInPosTree(isMinPosTree,
        fuzzyCDnaBoundaries, gm.getExons(),featureArr);

    // add FEATURE exon boundaries to treemap
    for (Iterator itFeatArr = featureArr.iterator(); itFeatArr.hasNext(); ) {
      SimFeature feature = (SimFeature) itFeatArr.next();

    // INTRON BOUNDARIES !!!
      positionTree =
          CDnaSpliceSiteAnalysisTools.putFeatPosInGmTree(positionTree,
          fuzzyCDnaBoundaries, isMinPosTree,feature.getIntrons());

//    // EXON BOUNDARIES !!!
//      positionTree =
//          CDnaSpliceSiteAnalysisTools.putFeatPosInGmTree(positionTree,
//          fuzzyCDnaBoundaries, isMinPosTree,feature.getExons());

    }
    //debug 202 - Debugging CDnaSpliceSite - PostionTree
    Debug.debug_202(debug, gm, isMinPosTree, positionTree);
    return positionTree;
  }


  /** method which returns a tree of featureComponent positions (Intron, Exon).
   *
   * @param isMinPosTree switch for consideration of either start or end
   *     position of the feature
   * @param fuzzyCDnaBoundaries contains allowed difference of cDNA and
   *      genemodel splice site
   * @param gmFeatCompArr ArrayList of genemodel components (genemodel
   *      component positions)
   * @param featureArr ArrayList of overlapping features
   * @return tree of RecordGmCDna objects (featureComponent positions)
   */
  private static MyTreeMap putGmInPosTree(boolean isMinPosTree,
                                          int fuzzyCDnaBoundaries,
                                          ArrayList gmFeatCompArr,
                                          ArrayList featureArr) {
    MyTreeMap gmFeatPosTree = new MyTreeMap();

    // add featureComponent boundaries to TreeMap
    for (Iterator itGmFeat = gmFeatCompArr.iterator(); itGmFeat.hasNext(); )
    {
      FeatureComponentAble featComp = (FeatureComponentAble) itGmFeat.next();

      int pos =
          CDnaSpliceSiteAnalysisTools.getPositionForPosTree(isMinPosTree,
          featComp);


      // if boundary overlaps cDna feature
      if(CDnaSpliceSiteAnalysisTools.doesPosOverlapSimFeatures(pos,
          fuzzyCDnaBoundaries, featureArr))
      {
        // create record object (add gm pos)
        RecordGmCDna record = new RecordGmCDna();
        record.gmBoundary = new Integer(pos);

        // add record object to treemap, key = featureComponent boundaries pos
        gmFeatPosTree.put(new Integer(pos), record);
      }
    }
    return gmFeatPosTree;
  }




  /** method which returns true if given position overlaps a object in SimFeature ArrayList.
   *
   * @param pos position of a genemodel (either start or end position)
   * @param fuzzyCDnaBoundaries contains allowed difference of cDNA and
   *      genemodel splice site
   * @param featureArr ArrayList of overlapping SimFeature objects
   * @return true for overlapping / false for not overlapping
   */
  private static boolean doesPosOverlapSimFeatures(int pos,
                                                   int fuzzyCDnaBoundaries,
                                                   ArrayList featureArr)
  {
    boolean doesPosOverlapSimFeatures = false;


    int minLoc = pos - fuzzyCDnaBoundaries;
    int maxLoc = pos + fuzzyCDnaBoundaries;

    Location fuzzyBoundLocation = LocationTools.makeLocation(minLoc, maxLoc);

    for(Iterator itFeat = featureArr.iterator(); itFeat.hasNext();)
    {
      SimFeature feature       = (SimFeature) itFeat.next();
      ArrayList contigLocArr  = new ArrayList();

      // put all intron locations in array
      for(Iterator itInt = feature.getIntrons().iterator();
          itInt.hasNext();)
      {
        FeatureComponent intron = (FeatureComponent) itInt.next();
        Location intLoc         = intron.getLocation();
//        System.err.println("intLoc " + intLoc);
        contigLocArr.add(intLoc);
      }

      // create continous location
      Location featLoc       = feature.getLocation();
//      System.err.println("featLoc " + featLoc);
      contigLocArr.add(featLoc);
      Location featContigLoc = LocationTools.union(contigLocArr);
//      System.err.println("featContigLoc " + featContigLoc);
//      System.err.println();

      // if boundary of gm overlaps continous cDNA feature location
      // and cDNA feature has more than 1 exons
      if(LocationTools.overlaps(fuzzyBoundLocation,featContigLoc) &&
         (feature.getExons().size() > 1))
      {
        doesPosOverlapSimFeatures = true;
      }
    }
    return doesPosOverlapSimFeatures;
  }




  /** method which returns position of exon boundary for position tree.
   *
   * @param isMinPosTree switch for consideration of either start or end
   *     position of the feature
   * @param exon Exon object
   * @return position of FeatureComponent boundary
   */
  private static int getPositionForPosTree(boolean isMinPosTree,
                                           FeatureComponentAble exon)
  {
    int pos = 0;

    // take MinPos
    if (isMinPosTree) {
      pos = exon.getLocation().getMin();
    }

    // take MaxPos
    else {
      pos = exon.getLocation().getMax();
    }

    return pos;
  }


  /** method which adds SimFeature boundaries to gm boundary tree.
   * If same boundary => add to gm record,
   * If diff boundary => create new record object
   *
   * @param positionTree tree containing RecordGmCDna objects
   * @param fuzzyCDnaBoundaries contains allowed difference of cDNA and
   *      genemodel splice site
   * @param isMinPosTree switch for consideration of either start or end
   *      position of the feature
   * @param featureCompArr ArrayList of genemodel components
   * @return tree of RecordGmCDna objects
   */
  private static MyTreeMap putFeatPosInGmTree(MyTreeMap positionTree,
                                              int fuzzyCDnaBoundaries,
                                              boolean isMinPosTree,
                                              ArrayList featureCompArr) {
    for (Iterator itFeatInt = featureCompArr.iterator(); itFeatInt.hasNext(); )
    {
      FeatureComponentAble featureComp = (FeatureComponentAble) itFeatInt.next();
      int posFeat = 0;

      // take MinPos
      if (isMinPosTree) {
        // minPos of exon = maxPos of intron + 1
        posFeat = featureComp.getLocation().getMax() + 1;
      }

      // take MaxPos
      else
      {
        // maxPos of exon = minPos of intron - 1
        posFeat = featureComp.getLocation().getMin() - 1;
      }

      // check if feature position is in treemap (fuzzy bound distance)
      // ALLOWED DISTANCE FOR FUZZY BOUNDARIES COULD BE COMMAND LINE ARGUMENT !!!
      RecordGmCDna nodeRecord = checkFeatPosInTree(posFeat,
          fuzzyCDnaBoundaries,
          positionTree);

      // feature position is not in tree
      if (nodeRecord == null) {
        // create record object (add feature pos)
        RecordGmCDna record = new RecordGmCDna();
        record.cDnaBoundaries.add(new Integer(posFeat));

        // add record object to treemap, key = pos
        positionTree.put(new Integer(posFeat), record);
      }

      // feature position is in tree
      else {
        nodeRecord.cDnaBoundaries.add(new Integer(posFeat));

        // add record object to treemap, key = pos
//        positionTree.put(nodeRecord.gmBoundary, nodeRecord);
      }
    }
    return positionTree;
  }


  /** method which checks whether position is in tree or not.
   * If position is in tree, record object is returned,
   * else null is returned
   *
   * @param featPos position of a genemodel (either start or end position)
   * @param maxDistance contains allowed difference of cDNA and
   *     genemodel splice site
   * @param posTree position tree containing all splice sites of
   *     genemodel and overlapping SimFeature objects (either start or end
   *      position)
   * @return either RecordGmCDna object or null
   */
  private static RecordGmCDna checkFeatPosInTree(int featPos,
                                                 int maxDistance,
                                                 MyTreeMap posTree) {
    RecordGmCDna returnRecord = null;

    for (Iterator itTree = posTree.keySet().iterator(); itTree.hasNext(); ) {
      int featPosComp = ( (Integer) itTree.next()).intValue();

      // use modulus to check feature is less away than maxDistance from
      // compared feature
      if (Math.abs(featPos - featPosComp) <= maxDistance) {

        // get record object
        returnRecord = (RecordGmCDna) posTree.get(new Integer(featPosComp));
      }
    }
    return returnRecord;
  }


  /** method which calculates the score of a gm.
   *
   * @param minPosTree tree containing RecordGmCDna objects (only start
   *     positions)
   * @param maxPosTree tree containing RecordGmCDna objects (only end
   *     positions)
   * @param featureArr ArrayList of genemodel overlapping features
   * @return absolute score
   */
  private static double calcCDnaSpliceSiteGmScore(MyTreeMap minPosTree,
                                                  MyTreeMap maxPosTree,
                                                  ArrayList featureArr) {
    double gmScore = 0;

    // minPosTree
    gmScore += CDnaSpliceSiteAnalysisTools.getTreeScore(minPosTree, featureArr);

    // maxPosTree
    gmScore += CDnaSpliceSiteAnalysisTools.getTreeScore(maxPosTree, featureArr);

    return gmScore;
  }


  /** method which calculates the score of a tree.
   * Either start or end positions => partial score of genemodel
   *
   * @param tree tree containing RecordGmCDna objects
   * @param featureArr ArrayList of genemodel overlapping features
   * @return absolute score
   */
  private static double getTreeScore(MyTreeMap tree, ArrayList featureArr) {
    double treeScore = 0;
    for (Iterator itTree = tree.keySet().iterator(); itTree.hasNext(); ) {
      Object key = (Object) itTree.next();
      RecordGmCDna record = (RecordGmCDna) tree.get(key);
      Integer gmBound = record.gmBoundary;
      ArrayList cDnaArr = (ArrayList) record.cDnaBoundaries;

      // gm boundary and cDNA boundary
      if ( (gmBound != null) && (cDnaArr.size() > 0)) {
        treeScore += cDnaArr.size();
      }

      // no gm boundary and cDNA boundary
      else if ( (gmBound == null) && (cDnaArr.size() > 0)) {
        treeScore -= cDnaArr.size();
      }

//      // gm boundary and no cDNA boundary
//      // if location overlaps a cDNA intron
//      else if    ((gmBound != null) &&
//                 (cDnaArr.size() == 0) &&
//                 (BioTools.doesBoundaryOverlapCDnaIntron(gmBound.intValue(),
//                                                         featureArr) == true))
//        {
//          treeScore -= 1;
//        }

      // gm boundary and no cDNA boundary
      else if ( (gmBound != null) && (cDnaArr.size() == 0)) {
        treeScore -= 1;
      }
    }
    return treeScore;
  }


  /** method which checks if genemodel boundary is overlapping a cDNA intron.
   *
   * @param gmBound boundary of a genemodel
   * @param featureArr ArrayList of overlapping cDNA features
   * @return true for overlap
   */
  private static boolean doesBoundaryOverlapCDnaIntron(int gmBound,
      ArrayList featureArr) {
    boolean doesBoundaryOverlapCDnaIntron = false;
    for (Iterator itFeat = featureArr.iterator(); itFeat.hasNext(); ) {
      SimFeature feature = (SimFeature) itFeat.next();
      for (Iterator itIntron = feature.getIntrons().iterator();
           itIntron.hasNext(); ) {
        Intron intron = (Intron) itIntron.next();
        int minLoc = intron.getLocation().getMin();
        int maxLoc = intron.getLocation().getMax();

        // if gm boundary overlaps intron (lies inbetween min and max loc)
        if ( (minLoc < gmBound) && (gmBound < maxLoc)) {
          doesBoundaryOverlapCDnaIntron = true;
        }
      }
    }
    return doesBoundaryOverlapCDnaIntron;
  }

}
