package uk.ac.sanger.psu.gfmerge.analysis.simfeature_analysis.tools;

import java.io.*;
import java.util.*;
import java.util.regex.*;
import org.biojava.bio.*;
import org.biojava.bio.seq.*;
import org.biojava.bio.seq.io.*;
import org.biojava.bio.seq.impl.*;
import org.biojava.bio.symbol.*;

import uk.ac.sanger.psu.gfmerge.util.algotools.*;
import uk.ac.sanger.psu.gfmerge.datacomp.*;
import uk.ac.sanger.psu.gfmerge.util.debug.*;
import uk.ac.sanger.psu.gfmerge.analysis.*;
import uk.ac.sanger.psu.gfmerge.analysis.simfeature_analysis.*;
import uk.ac.sanger.psu.gfmerge.analysis.simfeature_analysis.
    BasicFeatureAnalysis;
import uk.ac.sanger.psu.gfmerge.*;

import java.lang.Math;

/** class which contains basic overlap computing tools which are used by derived overlap analysis classes (cDNA, Blast)
 *
 * @author Sebastian R. Spiegler
 * @version 1.0
 */
public class FeatureOverlapAnalysisTools
    extends BasicFeatureAnalysisTools {

  /** method which returns hash of gms pointing to their longest absolute overlap (BLAST).
   * Different hits of a single protein to a genemodel are considered and accumulated. Longest protein overlap is taken for scoring the genemodel.
   *
   * @param debug debugging mode
   * @param considerStrand switch for strand consideration
   * @param arrListOfGMRegions ArrayList of genemodel regions
   * @param arrListOfFeatRegions ArrayList of SimFeature regions (Blast)
   * @return hash of genemodels pointing to their blast overlap score
   */
  public static HashMap getGmHashOfLongestAbsoluteFeatOverlap(
      int debug,
      boolean considerStrand,
      boolean isPercentage,
      ArrayList arrListOfGMRegions,
      ArrayList arrListOfFeatRegions,
      boolean considerIntrons) {
    HashMap gmHashOfLongestPercentFeatOverlap = new HashMap();

    // HashMap of gms pointing to their ArrayList of overlapping features
    HashMap hashFeatOverlappedGms
        = BasicFeatureAnalysisTools.getGmHashOfOverlappedFeat(debug,
        considerStrand,
        arrListOfGMRegions,
        arrListOfFeatRegions);

    // HashMap of gms pointing to their LONGEST feature overlap (longest protein hit)
    // key=> gm, value=> overlap length at exon level
    HashMap gmHashOfLongestFeatOverlap =
        FeatureOverlapAnalysisTools.calcHashOfLongestFeatOverlap(debug,
        considerStrand, hashFeatOverlappedGms, considerIntrons);

    // HashMap of gms pointing to their absolute overlap
    gmHashOfLongestPercentFeatOverlap =
        FeatureOverlapAnalysisTools.calcHashOfScoredOverlap(
        gmHashOfLongestFeatOverlap, isPercentage);

    return gmHashOfLongestPercentFeatOverlap;
  }

  /** method which calculates hash of longest feature overlap.
   * Considers multiple protein hits to a single genemodel and takes total
   *     length of the longest protein hit.
   *     (key=> gm, value=> longest overlap [long])
   *
   * @param debug debugging mode
   * @param considerStrand switch for strand consideration
   * @param hashFeatOverlappedGms hash of genemodels pointing to an
   *      ArrayList of overlapping features
   * @return HashMap of genemodels pointing to their longest protein overlap
   */
  private static HashMap calcHashOfLongestFeatOverlap(
      int debug,
      boolean considerStrand,
      HashMap hashFeatOverlappedGms,
      boolean considerIntrons) {
    HashMap hashOfLongestFeatOverlap = new HashMap();

    for (Iterator itGm = hashFeatOverlappedGms.keySet().iterator();
         itGm.hasNext(); ) {

      GeneModel gm = (GeneModel) itGm.next();
      ArrayList featArr = (ArrayList) hashFeatOverlappedGms.get(gm);

      long overlap = FeatureOverlapAnalysisTools.calcGmLongestOverlap(
          debug, considerStrand, gm, featArr, considerIntrons);

      hashOfLongestFeatOverlap.put(gm, new Long(overlap));
    }
    return hashOfLongestFeatOverlap;
  }

  /** method which calculates longest overlap for a single genemodel.
   *
   * @param debug debugging mode
   * @param considerStrand switch for strand consideration
   * @param gm genemodel
   * @param overlappingFeatArr ArrayList of overlapping features
   * @return overlap length (longest)
   */
  private static long calcGmLongestOverlap(int debug,
                                           boolean considerStrand,
                                           GeneModel gm,
                                           ArrayList overlappingFeatArr,
                                           boolean considerIntrons) {
    long gmLongestOverlap = 0;

    // get hashmap proteins pointing to array of hits from the same protein
    // - 1 -
    HashMap proteinHash =
        FeatureOverlapAnalysisTools.getProteinHash(debug,
        gm, overlappingFeatArr);

    // calculate for each protein the length of overlap at exon level
    // - 2 -
    HashMap proteinOverlapHash =
        FeatureOverlapAnalysisTools.getProteinOverlapHash(debug,
        considerStrand,
        gm,
        proteinHash,
        considerIntrons);

    // find longest overlap and return hash (key=>proteinID,value=>length[Long])
    // - 3 -
    HashMap gmLongestOverlapHash =
        FeatureOverlapAnalysisTools.findLongestProteinOverlap(
        proteinOverlapHash);

    String proteinID = (String) gmLongestOverlapHash.keySet().iterator().next();
    gmLongestOverlap = ( (Long) gmLongestOverlapHash.get(proteinID)).longValue();

    // debug 405 - length of longest protein overlap of single gm
    Debug.debug_405(debug, gm, proteinID, gmLongestOverlap);

    return gmLongestOverlap;
  }

  // - 3 -
  /** method which finds longest protein overlap.
   * Returns a hash where protein ID is the key and the length is the value.
   *
   * @param proteinOverlapLength hash of proteins (protein ID) pointing to
   *     their overlap length
   * @return hash, longest protein (protein ID) pointing to its overlap length
   */
  private static HashMap findLongestProteinOverlap(HashMap proteinOverlapLength) {
    long maxProteinOverlap = 0;
    String maxProteinID = "";

    for (Iterator itProt = proteinOverlapLength.keySet().iterator();
         itProt.hasNext(); ) {
      String proteinID = (String) itProt.next();
      long overlapLength =
          ( (Long) proteinOverlapLength.get(proteinID)).longValue();

      if (overlapLength > maxProteinOverlap) {
        maxProteinOverlap = overlapLength;
        maxProteinID = proteinID;
      }
    }

    HashMap returnHash = new HashMap();
    returnHash.put(maxProteinID, new Long(maxProteinOverlap));

    return returnHash;
  }

  // - 2 -
  /** method which returns hash map of proteins pointing to overlap length.
   * (all proteins are overlapping a single gm)
   *
   * @param debug debugging mode
   * @param considerStrand switch for strand consideration
   * @param gm genemodel
   * @param proteinHash hash of proteins (protein ID) pointing to their
   *      overlapping feature hits
   * @return hash map of proteins pointing to overlap length
   */
  private static HashMap getProteinOverlapHash(int debug,
                                               boolean considerStrand,
                                               GeneModel gm,
                                               HashMap proteinHash,
                                               boolean considerIntrons) {
    HashMap proteinOverlapLength = new HashMap();

    for (Iterator itProt = proteinHash.keySet().iterator(); itProt.hasNext(); ) {
      String proteinID = (String) itProt.next();
      ArrayList arrFeat = (ArrayList) proteinHash.get(proteinID);
      long singleProteinOverlap = 0;
      singleProteinOverlap +=
          FeatureOverlapAnalysisTools.getSingleProteinOverlap(debug,
          considerStrand,
          gm, arrFeat, considerIntrons);
      // put protein with overlap in hash
      proteinOverlapLength.put(proteinID, new Long(singleProteinOverlap));

    }

    // debug 404 - HashMap of proteins pointing to overlap length
    Debug.debug_404(debug, gm, proteinOverlapLength);

    return proteinOverlapLength;
  }

  /** method which returns single protein overlap for a genemodel.
   *
   * @param debug debugging mode
   * @param considerStrand switch for strand consideration
   * @param gm genemodel
   * @param arrFeat ArrayList of different hits from the same protein
   * @return overlap length
   */
  private static long getSingleProteinOverlap(int debug,
                                              boolean considerStrand,
                                              GeneModel gm,
                                              ArrayList arrFeat,
                                              boolean considerIntrons) {
    long singleProteinOverlap = 0;

    // get TOTAL feature overlap of single gm
//    singleProteinOverlap =
//        FeatureOverlapAnalysisTools.calcGmTotalOverlap(debug,
//        considerStrand,
//        gm,
//        arrFeat);
    singleProteinOverlap =
        FeatureOverlapAnalysisTools.calcGmTotalExIntOverlap(debug,
        considerStrand,
        gm,
        arrFeat,
        considerIntrons);

    return singleProteinOverlap;
  }

  // - 1 -
  /** method which returns protein hash.
   * Group feature hits from same protein.
       * key=> protein ID, value=>ArrayList of overlapping features of this protein.
   *
   * @param debug debugging mode
   * @param gm genemodel
   * @param overlappingFeatArr ArrayList of genemodel overlapping
   *      features
   * @return hash of proteins pointing to their overlapping hits
   */
  private static HashMap getProteinHash(int debug,
                                        GeneModel gm,
                                        ArrayList overlappingFeatArr) {
    HashMap proteinHash = new HashMap();

    // for each feature which overlaps gm
    for (Iterator itFeat = overlappingFeatArr.iterator(); itFeat.hasNext(); ) {
      SimFeature feature = (SimFeature) itFeat.next();

      // get proteinID
      String proteinID = FeatureOverlapAnalysisTools.getProteinID(feature);

      // put feature in proteinHash
      proteinHash =
          FeatureOverlapAnalysisTools.putFeatureInProteinHash(proteinID,
          feature,
          proteinHash);
    }

    // debug 403 - HashMap protein hash for a single gene model
    Debug.debug_403(debug, gm, proteinHash);

    return proteinHash;
  }

  /** method which returns the protein ID of a feature.
   *
   * @param feature overlapping feature
   * @return protein ID
   */
  private static String getProteinID(GFMergeFeature feature) {
    String proteinID = "";

    String annotation = (feature.getFeature().getAnnotation()).toString();

//    Pattern p            = Pattern.compile(".*\\)\\s+to\\s+(\\w*\\s+\\w*).*",Pattern.CASE_INSENSITIVE);
//    Pattern p = Pattern.compile(".*\\)\\s+to\\s+(\\w*\\s+\\w*).*");
//    Matcher m = p.matcher(annotation);
//
//    if (m.matches()) {
//      proteinID = m.group(1);
//    }

    // BlastCDS feature from Dicty Perl script
    Pattern pattern1 = Pattern.compile(".*\\)\\s+to\\s+(\\w*\\s+\\w*).*");

    // BlastCDS feature generated by exblx_to_tab.pl script
    //"FT                   /note="max score = 497  AAP37962 AAP37962 Heat shock protein 70."
    Pattern pattern2 = Pattern.compile(
        "^.*max\\s+score\\s+=\\s+\\d+\\s+(\\w+\\s+\\w+).*$");

    Matcher matcher1 = pattern1.matcher(annotation);
    Matcher matcher2 = pattern2.matcher(annotation);

    if (matcher1.matches()) {
      proteinID = matcher1.group(1);
    }

    else if (matcher2.matches()) {
      proteinID = matcher2.group(1);
    }

    else { // error
      System.err.println("Error - Feature without protein information");
      System.err.println("Feature   : " + feature.getLocation());
      System.err.println("Annotation: " + feature.getFeature().getAnnotation());
      System.exit(1);
    }
    return proteinID;
  }

  /** method which stores feature in protein hash.
   *
   * @param proteinID ID of the protein
   * @param feature overlapping SimFeature
   * @param proteinHash protein hash
   * @return protein hash (key=> protein ID, value=> ArrayList of hits from
   *      the this protein)
   */
  private static HashMap putFeatureInProteinHash(String proteinID,
                                                 SimFeature feature,
                                                 HashMap proteinHash) {
    // proteinID is already in hash
    if (proteinHash.containsKey(proteinID)) {
      ArrayList tempArr = (ArrayList) proteinHash.get(proteinID);
      tempArr.add(feature);
      proteinHash.put(proteinID, tempArr);
    }

    // proteinID does not exist
    else {
      ArrayList tempArr = new ArrayList();
      tempArr.add(feature);
      proteinHash.put(proteinID, tempArr);
    }

    return proteinHash;
  }

  /** method which returns hash of genemodels pointing to their ABSOLUTE overlap (CDNA).
   *
   * @param debug debugging mode
   * @param considerStrand switch for strand consideration
   * @param arrListOfGMRegions ArrayList of genemodel regions
   * @param arrListOfFeatRegions ArrayList of feature regions (cDNA)
   * @return hash of genemodels pointing to cDNA overlap
   */
  public static HashMap getGmHashOfAbsoluteFeatOverlap(
      int debug,
      boolean considerStrand,
      boolean isPercentage,
      ArrayList arrListOfGMRegions,
      ArrayList arrListOfFeatRegions,
      boolean considerIntrons) {
    HashMap gmHashOfAbsoluteFeatOverlap = new HashMap();

    // HashMap of gms pointing to their ArrayList of overlapping features
    HashMap gmHashOfOverlappedFeat
        = FeatureOverlapAnalysisTools.getGmHashOfOverlappedFeat(debug,
        considerStrand,
        arrListOfGMRegions,
        arrListOfFeatRegions);

    // HashMap of gms pointing to their TOTAL feature overlap (cDNA)
    HashMap gmHashOfTotalFeatOverlap =
        FeatureOverlapAnalysisTools.calcHashOfTotalOverlap(debug,
        considerStrand,
        gmHashOfOverlappedFeat,
        considerIntrons);

    // debug 301 - HashMap of genemodels pointing to total feature overlap
    Debug.debug_301(debug,
                    "FeatureOverlapAnalysisTools - cDNA total feature overlap",
                    gmHashOfTotalFeatOverlap);

    // HashMap of gms pointing to their absolute overlap
    gmHashOfAbsoluteFeatOverlap =
        FeatureOverlapAnalysisTools.calcHashOfScoredOverlap(
        gmHashOfTotalFeatOverlap, isPercentage);

    // debug 302 - HashMap of genemodels pointing to their absolute overlap
    Debug.debug_302(debug,
        "FeatureOverlapAnalysisTools - cDNA absolute scored overlap",
        gmHashOfAbsoluteFeatOverlap);

    return gmHashOfAbsoluteFeatOverlap;
  }

  /** method which calculates hash of total cDNA overlap for genemodels.
   *
   * @param debug debugging mode
   * @param considerStrand switch for strand consideration
   * @param gmHashOfOverlappedFeat hash of genemodels pointing to
   *     ArrayList of overlapping cDNA features
   * @return hash of genemodels pointing to their cDNA overlap
   */
  private static HashMap calcHashOfTotalOverlap(
      int debug,
      boolean considerStrand,
      HashMap gmHashOfOverlappedFeat,
      boolean considerIntrons) {
    HashMap hashOfTotalFeatOverlap = new HashMap();

    for (Iterator itGm = gmHashOfOverlappedFeat.keySet().iterator();
         itGm.hasNext(); ) {
      GeneModel gm = (GeneModel) itGm.next();
      ArrayList featArr = (ArrayList) gmHashOfOverlappedFeat.get(gm);
      long overlap =
//          FeatureOverlapAnalysisTools.calcGmTotalOverlap(debug, considerStrand,
//          gm, featArr);
          FeatureOverlapAnalysisTools.calcGmTotalExIntOverlap(debug,
          considerStrand,
          gm, featArr, considerIntrons);

      hashOfTotalFeatOverlap.put(gm, new Long(overlap));
    }
    return hashOfTotalFeatOverlap;
  }

  /** method which calculates total cDNA/single Blast protein overlap for a single genemodel.
   *
   * @param debug debugging mode
   * @param considerStrand switch for strand consideration
   * @param gm genemodel
   * @param featArr ArrayList of genemodel overlapping cDNA features
   * @return total cDNA overlap for a single genemodel
   */
  private static long calcGmTotalOverlap(int debug,
                                         boolean considerStrand,
                                         GeneModel gm,
                                         ArrayList featArr) {
    long overlap = 0;

    // HashMap of exons pointing to their ArrayList of overlapping features
    HashMap hashFeatOverlappedExons =
        FeatureOverlapAnalysisTools.getExonHashOfOverlappedFeat(debug,
        considerStrand,
        gm,
        featArr);

    for (Iterator itEx = gm.getExons().iterator(); itEx.hasNext(); ) {
      Exon exon = (Exon) itEx.next();
      overlap +=
          FeatureOverlapAnalysisTools.calcExonTotalOverlap(exon,
          hashFeatOverlappedExons);
    }
//    System.err.println("calcGmTotalOverlap " + gm.getLocation());
//    System.err.println("overlap " + overlap);
    return overlap;
  }

  private static long calcGmTotalExIntOverlap(int debug,
                                              boolean considerStrand,
                                              GeneModel gm,
                                              ArrayList featArr,
                                              boolean considerIntrons) {
    long overlap = 0;

    // EXONS
    // HashMap of exons pointing to their ArrayList of overlapping features
    HashMap hashFeatOverlappedExons =
        FeatureOverlapAnalysisTools.getGmCompHashOfOverlappedFeat(debug,
        considerStrand,
        gm.getExons(),
        featArr);

    boolean isExon = true;
    for (Iterator itEx = gm.getExons().iterator(); itEx.hasNext(); ) {
      Exon exon = (Exon) itEx.next();
      overlap +=
          FeatureOverlapAnalysisTools.calcGmCompTotalOverlap(exon,
          hashFeatOverlappedExons, isExon);
    }

    // INTRONS
    if (considerIntrons == true) {
//      System.out.println("gm: " + gm.getLocation());

      HashMap hashFeatOverlappedIntrons =
          FeatureOverlapAnalysisTools.getGmCompHashOfOverlappedFeat(debug,
          considerStrand,
          gm.getIntrons(),
          featArr);

//      System.out.println("before intron: " + overlap);

      isExon = false;
      for (Iterator itInt = gm.getIntrons().iterator(); itInt.hasNext(); ) {
        Intron intron = (Intron) itInt.next();
        overlap +=
            FeatureOverlapAnalysisTools.calcGmCompTotalOverlap(intron,
            hashFeatOverlappedIntrons, isExon);
      }

//      System.out.println("after  intron: " + overlap);
//      System.out.println();
    }

    return overlap;
  }

  /** method which calculates total cDNA overlap on genemodel exon level.
   *
   * @param exon Exon object
   * @param hashFeatOverlappedExons hash of exons pointing to their
   *     overlapping cDNA / Blast features
   * @return total cDNA / Blast overlap at genemodel nucleotide level
   */
  private static long calcExonTotalOverlap(Exon exon,
                                           HashMap hashFeatOverlappedExons) {
    long totalExonFeatOverlap = 0;

    Location exLoc = exon.getLocation();
    ArrayList subLocsArr = new ArrayList();

    if (hashFeatOverlappedExons.containsKey(exon)) {
      ArrayList featArr = (ArrayList) hashFeatOverlappedExons.get(exon);
      for (Iterator itFeat = featArr.iterator(); itFeat.hasNext(); ) {
        SimFeature feature = (SimFeature) itFeat.next();
        Location featLoc = feature.getLocation();

        // fill subLocsArr with feat overlaps on exon level
        Location subLoc = LocationTools.intersection(exLoc, featLoc);
        subLocsArr.add(subLoc);
      }
      Location mergedSubLocs = LocationTools.union(subLocsArr);
      totalExonFeatOverlap =
          FeatureOverlapAnalysisTools.calcTotalLocationLength(mergedSubLocs);
    }
    return totalExonFeatOverlap;
  }

  private static long calcGmCompTotalOverlap(FeatureComponentAble gmComp,
                                             HashMap hashFeatOverlappedGmComps,
                                             boolean isExon) {
    long totalGmCompFeatOverlap = 0;

    Location gmCompLoc = gmComp.getLocation();
    ArrayList subLocsArr = new ArrayList();

    if (hashFeatOverlappedGmComps.containsKey(gmComp)) {
      ArrayList featArr = (ArrayList) hashFeatOverlappedGmComps.get(gmComp);

      // exon comparison
      if (isExon == true) {
        // for all overlapping features
        for (Iterator itFeat = featArr.iterator(); itFeat.hasNext(); ) {
          SimFeature feature = (SimFeature) itFeat.next();
          ArrayList exonArr = feature.getExons();

          for (Iterator itEx = exonArr.iterator(); itEx.hasNext(); ) {
            Exon exon = (Exon) itEx.next();
            Location exLoc = exon.getLocation();

            // fill subLocsArr with feat overlaps on exon level
            Location subLoc = LocationTools.intersection(gmCompLoc, exLoc);
            subLocsArr.add(subLoc);
          }
        }
      }
      // intron comparison
      else {
        // for all overlapping features
        for (Iterator itFeat = featArr.iterator(); itFeat.hasNext(); ) {
          SimFeature feature = (SimFeature) itFeat.next();
          ArrayList intronArr = feature.getIntrons();

          for (Iterator itInt = intronArr.iterator(); itInt.hasNext(); ) {
            Intron intron = (Intron) itInt.next();
            Location intLoc = intron.getLocation();

            // fill subLocsArr with feat overlaps on exon level
            Location subLoc = LocationTools.intersection(gmCompLoc, intLoc);
            subLocsArr.add(subLoc);
          }
        }

      }
      Location mergedSubLocs = LocationTools.union(subLocsArr);
      totalGmCompFeatOverlap =
          FeatureOverlapAnalysisTools.calcTotalLocationLength(mergedSubLocs);
    }
    return totalGmCompFeatOverlap;
  }

  /** method which accumulates total overlap length of exons/introns by
   * aggregating sublocations of a compound overlap.
   *
   * @param mergedSubLocs merged Locations of all htis at exon level
   * @return total overlap lenght of genemodel on exon level
   */
  private static long calcTotalLocationLength(Location mergedSubLocs) {
    long totalLocationLength = 0;
    for (Iterator itLoc = mergedSubLocs.blockIterator(); itLoc.hasNext(); ) {
      Location loc = (Location) itLoc.next();
      totalLocationLength += (loc.getMax() - loc.getMin() + 1);
    }
    return totalLocationLength;
  }

  /** method which calculates scored overlap of genemodels.
   * (returns hash of gms pointing to their [Double] scored overlap).
   *
   * @param gmHashOfTotalFeatOverlap hash genemodels pointing to their
   *     absolute overlap [Long]
   * @param isPercentage switch for percentage overlap otherwise absolute
   *     overlap [Double]
   * @return returns hash genemodels pointint to their percentage feature
   *     overlap
   */
  private static HashMap calcHashOfScoredOverlap(
      HashMap gmHashOfTotalFeatOverlap, boolean isPercentage) {
    HashMap hashOfScoredOverlap = new HashMap();

    for (Iterator itGm = gmHashOfTotalFeatOverlap.keySet().iterator();
         itGm.hasNext(); ) {
      GeneModel gm = (GeneModel) itGm.next();
      long gmTotalLength = gm.getTotalExonLength();
      long featOverlap = ( (Long) gmHashOfTotalFeatOverlap.get(gm)).longValue();
      double scoredOverlap = 0;

      // percentage overlap
      if (isPercentage == true) {
        scoredOverlap = (double) featOverlap / (double) gmTotalLength;
      }
      // absolute overlap
      else {
        scoredOverlap = (double) featOverlap;
      }

      hashOfScoredOverlap.put(gm, new Double(scoredOverlap));
    }

    return hashOfScoredOverlap;
  }

  /** method which returns hash of exons pointing to ArrayList of overlapping features.
   *
   * @param debug debugging mode
   * @param considerStrand switch for strand consideration
   * @param gm genemodel
   * @param overlappingFeatArr ArrayList of genemodel overlapping
   *     features
   * @return hash of exons pointing to ArrayList of overlapping features
   */
  private static HashMap getExonHashOfOverlappedFeat
      (int debug,
       boolean considerStrand,
       GeneModel gm,
       ArrayList overlappingFeatArr) {
    HashMap exonHashOfOverlappedFeat = new HashMap();

    for (Iterator itEx = gm.getExons().iterator(); itEx.hasNext(); ) {
      Exon exon = (Exon) itEx.next();
      ArrayList tempArr = FeatureOverlapAnalysisTools.getFeatOverlapArrForExon(
          exon,
          overlappingFeatArr);
      exonHashOfOverlappedFeat.put(exon, tempArr);
    }

    // debug 402 - HashMap of exons pointing to array of overlapping features
    Debug.debug_402(debug,
                    "FeatureOverlapTools - HashMap of exons pointing"
                    + " to array of overlapping features",
                    exonHashOfOverlappedFeat);

    return exonHashOfOverlappedFeat;
  }

  /** method which returns hash of gm features (exons or introns) pointing to
   * ArrayList of overlapping features (Blast or cDNA).
   *
   * @param debug debugging mode
   * @param considerStrand switch for strand consideration
   * @param gmCompArr ArrayList of genemodel components (exons or introns)
   * @param overlappingFeatArr ArrayList of genemodel overlapping
   *     features
   * @return hash of exons pointing to ArrayList of overlapping features
   */
  private static HashMap getGmCompHashOfOverlappedFeat
      (int debug,
       boolean considerStrand,
       ArrayList gmCompArr,
       ArrayList overlappingFeatArr) {

    HashMap gmCompHashOfOverlappedFeat = new HashMap();

    for (Iterator itGmComp = gmCompArr.iterator(); itGmComp.hasNext(); ) {
      FeatureComponentAble gmComp = (FeatureComponentAble) itGmComp.next();
      ArrayList tempArr = FeatureOverlapAnalysisTools.
          getFeatOverlapArrForGmComp(
          gmComp,
          overlappingFeatArr);
      gmCompHashOfOverlappedFeat.put(gmComp, tempArr);
    }

    // debug 402 - HashMap of exons pointing to array of overlapping features
    Debug.debug_402(debug,
        "FeatureOverlapTools - HashMap of gmComp (exons, introns) pointing"
        + " to array of overlapping features",
        gmCompHashOfOverlappedFeat);

    return gmCompHashOfOverlappedFeat;
  }

  /** method which returns overlapping features for a single exon.
   *
   * @param exon Exon object
   * @param featArr ArrayList of genemodel overlapping features
   * @return ArrayList of overlapping features for a single exon
   */
  private static ArrayList getFeatOverlapArrForExon(Exon exon,
      ArrayList featArr) {
    ArrayList featOverlapArr = new ArrayList();

    Location exonLoc = exon.getLocation();

    for (Iterator itFeat = featArr.iterator(); itFeat.hasNext(); ) {
      SimFeature feature = (SimFeature) itFeat.next();
      Location featLoc = feature.getLocation();

      if (LocationTools.overlaps(exonLoc, featLoc)) {
        featOverlapArr.add(feature);
      }
    }

    return featOverlapArr;
  }

  private static ArrayList getFeatOverlapArrForGmComp(FeatureComponentAble
      gmComp,
      ArrayList featArr) {

    ArrayList featOverlapArr = new ArrayList();

    Location gmCompLoc = gmComp.getLocation();

    for (Iterator itFeat = featArr.iterator(); itFeat.hasNext(); ) {
      SimFeature feature = (SimFeature) itFeat.next();
      Location featLoc = feature.getLocation();

      if (LocationTools.overlaps(gmCompLoc, featLoc)) {
        featOverlapArr.add(feature);
      }
    }

    return featOverlapArr;
  }

}