package uk.ac.sanger.psu.gfmerge.analysis.simfeature_analysis.tools;

import java.util.*;

import org.biojava.bio.symbol.*;

import uk.ac.sanger.psu.gfmerge.datacomp.*;
import uk.ac.sanger.psu.gfmerge.util.algotools.*;
import uk.ac.sanger.psu.gfmerge.util.debug.*;

/** class which contains computing tools for TotalExonLengthAnalysis.
 *
 * @author Sebastian R. Spiegler
 * @version 1.0
 */
public class TotalExonLengthAnalysisTools /*extends BasicFeatureAnalysisTools*/ {

  /** method which returns total exon length of a genemodel.
   * Total exon length means accumulated length of all exons within a genemodel excluding introns.
   *
   *
   * @param debug debugging mode
   * @param _arrListOfGMRegions ArrayList of genemodel regions
   * @return hash of genemodels pointing to their total exon length
   */
  public static HashMap getTlExLengthScoredGmHash(int debug,
                                                  ArrayList _arrListOfGMRegions) {
    HashMap tlExLengthScoredGmHash = new HashMap();

    // get HashMap of genemodels pointing to the number of equal gms
    HashMap gmHashOfEqualGms =
        TotalExonLengthAnalysisTools.getGmHashOfEqualGms(_arrListOfGMRegions);

    for (Iterator itReg = _arrListOfGMRegions.iterator(); itReg.hasNext(); ) {
      GFMergeRegion region = (GFMergeRegion) itReg.next();
      ArrayList gmArr = (ArrayList) region.getLocateable();

      for (Iterator itGm = gmArr.iterator(); itGm.hasNext(); ) {
        GeneModel gm = (GeneModel) itGm.next();
        double score =
            TotalExonLengthAnalysisTools.getTotalExonLengthScore(gm,
            gmHashOfEqualGms);
        tlExLengthScoredGmHash.put(gm, new Double(score));
      }
    }
    return tlExLengthScoredGmHash;
  }

  /** method which calculates score for a genemodel using the total exon length as scoring system.
   * Genemodel which have equal companions get a higher score.
   * This score is calculated by using the length of the sequence and multiplying it by the number of equal genemodels. The reason for this implementation is that in the case of overlapping genemodels rather equal genemodels should be chosen.
   *
   * @param gm genemodel
   * @param gmHashOfEqualGms HashMap of genemodels pointing to the number
   *      of equal genemodels (including the genemodel itself)
   * @return genemodel score using the total exon length scoring system
   */
  private static double getTotalExonLengthScore(GeneModel gm,
                                                HashMap gmHashOfEqualGms) {
    double totalExonLengthScore = 0;
    long numberOfEqualGms = 0;

    // if gm has equal companions
    if (gmHashOfEqualGms.containsKey(gm)) {
      long sequenceLength = gm.getFeature().getSequence().length();
      numberOfEqualGms = ( (Integer) gmHashOfEqualGms.get(gm)).intValue();
      totalExonLengthScore = sequenceLength * numberOfEqualGms;
    }
    // if not
    else {
      totalExonLengthScore = gm.getTotalExonLength();
    }

    // debug
//      System.out.println("gm                   : " + gm.getLocation());
//      System.out.println("gm length            : " + gm.getLength());
//      System.out.println("equal gms            : " + numberOfEqualGms);
//      System.out.println("totalExonLengthScore : " + totalExonLengthScore);
//      System.out.println();

    return totalExonLengthScore;
  }

  /** method which calculates hash of gene models pointing to number of equal gms
   * (including the genemodel itself)
   *
   * @param _arrListOfGmRegions ArrayList of genemodel regions
   * @return HashMap of genemodel pointing to the number of equal genemodels
   */
  private static HashMap getGmHashOfEqualGms(ArrayList _arrListOfGmRegions) {
    HashMap gmHashOfEqualGms = new HashMap();

    HashMap gmHashOfLocs =
        TotalExonLengthAnalysisTools.getGmHashOfLocations(_arrListOfGmRegions);

    for (Iterator itLoc = gmHashOfLocs.keySet().iterator(); itLoc.hasNext(); ) {
      Location location = (Location) itLoc.next();
      ArrayList gmArr = (ArrayList) gmHashOfLocs.get(location);

      // if more gm with the same location (equal gms)
      if (gmArr.size() > 1) {
        for (Iterator itGm = gmArr.iterator(); itGm.hasNext(); ) {
          GeneModel gm = (GeneModel) itGm.next();
          int numberOfEqualGms = gmArr.size();
          gmHashOfEqualGms.put(gm, new Integer(numberOfEqualGms));
        }
      }
    }

    // debug
//      for(Iterator itGm = gmHashOfEqualGms.keySet().iterator();
//          itGm.hasNext();)
//      {
//        GeneModel gm  = (GeneModel) itGm.next();
//        System.out.println("gm    : " + gm.getLocation());
//        Integer score = (Integer) gmHashOfEqualGms.get(gm);
//        System.out.println("score : " + score);
//        System.out.println();
//      }
//      System.out.println();

    return gmHashOfEqualGms;
  }

  /** method which calculates hash of locations pointing to their array of
   * gms with the same location (equal genemodels).
   *
   * @param _arrListOfGmRegions ArrayList of genemodel regions
   * @return HashMap of Locations pointing to ArrayList of genemodels with
   *     this particular Location
   */
  private static HashMap getGmHashOfLocations(ArrayList _arrListOfGmRegions) {
    HashMap gmHashOfLocs = new HashMap();

    // fill HashMap with locations
    for (Iterator itReg = _arrListOfGmRegions.iterator(); itReg.hasNext(); ) {
      GFMergeRegion region = (GFMergeRegion) itReg.next();
      ArrayList gmArr = (ArrayList) region.getLocateable();

      for (Iterator itGm = gmArr.iterator(); itGm.hasNext(); ) {
        GeneModel gm = (GeneModel) itGm.next();
        Location gmLoc = (Location) gm.getLocation();

        // if hash contains location already
        if (gmHashOfLocs.containsKey(gmLoc)) {
          ArrayList arrListTemp
              = (ArrayList) gmHashOfLocs.get(gm.getLocation());
          arrListTemp.add(gm);
          gmHashOfLocs.put(gmLoc, arrListTemp);
        }
        // if hash does not contain start location
        else {
          ArrayList arrListTemp = new ArrayList();
          arrListTemp.add(gm);
          gmHashOfLocs.put(gmLoc, arrListTemp);
        }
      }
    }

    // debug
//    for(Iterator itLoc = gmHashOfLocs.keySet().iterator(); itLoc.hasNext();)
//    {
//      Location location = (Location) itLoc.next();
//      ArrayList gmArr   = (ArrayList) gmHashOfLocs.get(location);
//      System.out.println("location : " + location);
//
//      for(Iterator itGm = gmArr.iterator(); itGm.hasNext();)
//      {
//        GeneModel gm = (GeneModel) itGm.next();
//        System.out.println("       gm : " + gm.getLocation());
//      }
//      System.out.println();
//    }

    return gmHashOfLocs;
  }

}