package uk.ac.sanger.psu.gfmerge.analysis.tools;

import java.util.*;
import org.biojava.bio.*;
import org.biojava.bio.seq.*;
import org.biojava.bio.symbol.*;

import uk.ac.sanger.psu.gfmerge.datacomp.*;
import uk.ac.sanger.psu.gfmerge.util.debug.*;
import uk.ac.sanger.psu.gfmerge.util.biotools.*;

public class HighScoringPath {

////////////////////////////////////////////////////////////////////////
// Recursive genemodel picking using highest scoring path in a region
//

  // (1)
  // gets highscoring path using the whole sequence
  // hash must be clustered before
  //
  public static HashMap getHighScoringPath(HashMap nonClusteredObjScoreHash,
                                           ArrayList arrListOfGmRegions,
                                           boolean isAverage) {
    HashMap objsOnHighScoringPath = new HashMap();

    // cluster hash into regions
    for(Iterator itReg = arrListOfGmRegions.iterator(); itReg.hasNext();)
    {
      GFMergeRegion region = (GFMergeRegion) itReg.next();

      HashMap clusteredObjScoreHash      = new HashMap();
      HashMap highScoringClusterPathHash = new HashMap();

      // put all gms of region into clusteredObjScoreHash
      for(Iterator itObj = region.getLocateable().iterator(); itObj.hasNext();)
      {
        Object obj                    = (Object) itObj.next();

        if(nonClusteredObjScoreHash.containsKey(obj))
        {
          Object score = (Object) nonClusteredObjScoreHash.get(obj);
          clusteredObjScoreHash.put(obj,score);
        }
        // !!!! OBJECTS WHICH ARE NOT IN HASH GET SCORE 0 !!!!
        else
        {
          Double score = new Double(0);
          clusteredObjScoreHash.put(obj,score);
        }
      }
      // get cluster highscoring path
      highScoringClusterPathHash
          = HighScoringPath.calcHighScoringPath(clusteredObjScoreHash, isAverage);

      // add cluster highscoring path to sequence highscoring path hash
      objsOnHighScoringPath.putAll(highScoringClusterPathHash);
    }

    return objsOnHighScoringPath;
  }


  // (2)
  // gets highscoring path using a cluster (hash contains only gm cluster)
  //
  // compares scores of all pathes and picks the path with the highest score
  // all objects on this path will be added to a hashmap and returned
  //
  private static HashMap calcHighScoringPath(HashMap clusteredObjScoreHash,
                                             boolean isAverage) {

    // initialize allPathArr, singlePathArr, objsOnHighScoringPath, maxScoreSeen
    TreeMap allPathList           = new TreeMap();
    ArrayList singlePathArr       = new ArrayList();
    HashMap objsOnHighScoringPath = new HashMap();

    // get all pathes by calling getAllPaths
    allPathList = HighScoringPath.getAllPaths(allPathList,
                                              singlePathArr,
                                              clusteredObjScoreHash,
                                              isAverage);

    // get high score of all paths
    Double maxScoreSeen = (Double) allPathList.lastKey();

    // find high score path(s) and add objects
    ArrayList maxScoreArr = (ArrayList) allPathList.get(maxScoreSeen);

    // get high score path hash,
    // key=> gm, value=> score
    for (Iterator itMaxPath = maxScoreArr.iterator(); itMaxPath.hasNext(); )
    {
      ArrayList pathArr     = (ArrayList) itMaxPath.next();

      for (Iterator itObj = pathArr.iterator(); itObj.hasNext(); )
      {
        LocateAble obj    = (LocateAble) itObj.next();
        Double objScore   = (Double) clusteredObjScoreHash.get(obj);
        objsOnHighScoringPath.put(obj, objScore);
      }
    }

    return objsOnHighScoringPath;
  }



  //
  // retrieve recursively all paths in a set of objects
  // parameters are allPathArr, singlePathArr, objScoreHash
  //
  private static TreeMap getAllPaths(TreeMap allPathList,
                                     ArrayList singlePathArr,
                                     HashMap objScoreHash,
                                     boolean isAverage) {

    // get all non overlapping objects by calling getNonOverlappingObjs
    ArrayList arrOfNonOverlappingObjs = HighScoringPath.getNonOverlappingObjs(
        singlePathArr, objScoreHash);

    // OPTIMIZATION 1 (only pathes allowed which start at the "left corner" of a cluster)
    if (HighScoringPath.isNonOverlappingObjUpstream(singlePathArr,
        arrOfNonOverlappingObjs) == true) {
      // exit condition
    }

    // OPTIMIZATION 3
    // (only pahtes are allowed which do not have a predecessor which should be further downstream)
    else if (HighScoringPath.isNonOverlappingDownstreamObjFurtherUpstream(
        singlePathArr) == true) {
      // exit condition
    }

    // if there are no nonoverlapping objects left, do nothing / exit condition
    else if (arrOfNonOverlappingObjs.size() == 0) {
      // allPathArr.add(singlePathArr);

      // OPTIMIZATION 2 (list of pathes sorted by their score which saves further iterations)
      Double pathScore = new Double(getPathScore(singlePathArr, objScoreHash, isAverage));

      if (allPathList.containsKey(pathScore)) {
        ArrayList tempArr = (ArrayList) allPathList.get(pathScore);
        tempArr.add(singlePathArr);
        allPathList.put(pathScore, tempArr);
      }
      else {
        ArrayList tempArr = new ArrayList();
        tempArr.add(singlePathArr);
        allPathList.put(pathScore, tempArr);
      }

      // exit condition
    }

    // if there are nonoverlapping objects left, call getAllPaths recursivly
    else {
      for (Iterator itObj = arrOfNonOverlappingObjs.iterator(); itObj.hasNext(); ) {
        LocateAble obj = (LocateAble) itObj.next();
        ArrayList tempArr = new ArrayList(singlePathArr); // create new array
        tempArr.add(obj);
        getAllPaths(allPathList, tempArr, objScoreHash, isAverage);
      }
    }
    return allPathList;
  }



  //
  // get score of path
  // sum the scores of all the objects in path
  //
  private static double getPathScore(ArrayList pathArr,
                                     HashMap objScoreHash,
                                     boolean isAverage) {
    double pathScore = 0;
    for (Iterator itObj = pathArr.iterator(); itObj.hasNext(); ) {
      LocateAble obj = (LocateAble) itObj.next();
      double score = ( (Double) objScoreHash.get(obj)).doubleValue();
      pathScore += score;
    }

    // if average score, divide score by number of objects (array size)
    if(isAverage == true)
    {
      return pathScore / pathArr.size();
    }
    // else return score
    else
    {
      return pathScore;
    }
  }



  //
  // get non overlapping genemodels
  // parameters are singlePathArr, objScoreHash
  //
  private static ArrayList getNonOverlappingObjs(ArrayList singlePathArr,
                                                 HashMap objScoreHash) {
    ArrayList arrOfNonOverlappingObjs = new ArrayList();
    HashMap overlappingObjHash = new HashMap();

    // singlePathArr is EMPTY => add all objects of hash to return array
    if (singlePathArr.size() == 0) {
      // iterate over all genemodels in hash and add them to return array
      for (Iterator itObj = objScoreHash.keySet().iterator(); itObj.hasNext(); ) {
        LocateAble obj = (LocateAble) itObj.next();
        arrOfNonOverlappingObjs.add(obj);
      }
    }

    // singlePathArr is NOT EMPTY => add none overlapping objects of hash to return array
    else {
      // iterate over genemodels in path
      for (Iterator itObj = singlePathArr.iterator(); itObj.hasNext(); ) {
        LocateAble obj = (LocateAble) itObj.next();
        Location objLoc = obj.getLocation();

        // iterate over all genemodels in hash
        for (Iterator itObjComp = objScoreHash.keySet().iterator();
             itObjComp.hasNext(); ) {
          LocateAble objComp = (LocateAble) itObjComp.next();
          Location objCompLoc = objComp.getLocation();

          // all the genemodels which overlap the original one, add to temporary hash
//          System.out.println("Loc 1: " + objLoc);
//          System.out.println("Loc 2: " + objCompLoc);
//          System.out.println("Overlaps: " + LocationTools.overlaps(objLoc, objCompLoc));
//          System.out.println("---");
          if(BioTools.overlaps(objLoc, objCompLoc))
          //if (LocationTools.overlaps(objLoc, objCompLoc))
          // changed to eliminate the bug, where a single exon gene is
          // within an intron location of another gene and the method
          // LocationTools.overlaps() returns false
          {
            overlappingObjHash.put(objComp, "overlapping");
          }
        }
      }

      // compare hash of overlapping objects to objScoreHash,
      // take all objects which are not in the hash
      for (Iterator itHash = objScoreHash.keySet().iterator(); itHash.hasNext(); ) {
        LocateAble obj = (LocateAble) itHash.next();
        if (! (overlappingObjHash.containsKey(obj))) {
          arrOfNonOverlappingObjs.add(obj);
        }
      }

    }

    return arrOfNonOverlappingObjs;
  }


  //
  // find out if there are non overlapping objects further upstream
  //
  private static boolean isNonOverlappingObjUpstream(ArrayList singlePathArr,
      ArrayList arrOfNonOverlappingObjs) {
    boolean isNonOverlappingObjUpstream = false;

    // do it only for first object
    if (singlePathArr.size() == 1) {
      // take first element
      LocateAble obj = (LocateAble) singlePathArr.get(0);
      Location locObj = obj.getLocation();
      for (Iterator it = arrOfNonOverlappingObjs.iterator(); it.hasNext(); ) {
        LocateAble compObj = (LocateAble) it.next();
        Location locCompObj = compObj.getLocation();
        if (locCompObj.getMin() < locObj.getMin() &&
            locCompObj.getMax() < locObj.getMax()) {
          isNonOverlappingObjUpstream = true;
        }
      }
    }

    return isNonOverlappingObjUpstream;
  }



  //
  // find out if path has wrong order (downstream object before upstream object)
  //
  private static boolean isNonOverlappingDownstreamObjFurtherUpstream(ArrayList
      singlePathArr) {
    boolean isNonOverlappingDownstreamObjFurtherUpstream = false;

    // do not do it for empty array
    if (singlePathArr.size() > 0) {
      // compare all elements in array to each other
      // find out if upstream elements come after downstream elements
      // p.e. [0,10] [30,35] [15,20] => [15,20] comes after downstream element
      for (int i = 1; i < singlePathArr.size(); i++) {
        LocateAble obj = (LocateAble) singlePathArr.get(i);
        Location locObj = obj.getLocation();

        //predecessor
        LocateAble predObj = (LocateAble) singlePathArr.get(i - 1); ;
        Location locPredObj = predObj.getLocation();

        if (locPredObj.getMin() > locObj.getMin() &&
            locPredObj.getMax() > locObj.getMax()) {
          isNonOverlappingDownstreamObjFurtherUpstream = true;
        }
      }
    }
    return isNonOverlappingDownstreamObjFurtherUpstream;
  }



  ////////////////////////////////////////////////////////////////////////
  // remove low scoring gms from array of regions
  //
  public static ArrayList removeLowScoreGmsFromRegArr(ArrayList
      _arrListOfGMRegions,
      HashMap highScorePathGms) {
    for (Iterator itReg = _arrListOfGMRegions.iterator(); itReg.hasNext(); ) {
      GFMergeRegion gmReg = (GFMergeRegion) itReg.next();
      for (Iterator itGm = gmReg.getLocateable().iterator(); itGm.hasNext(); ) {
        GeneModel gm = (GeneModel) itGm.next();
        // if gm is not on high scoring path, remove from region
        if (!highScorePathGms.containsKey(gm)) {
          itGm.remove();
        }
      }
    }

    return _arrListOfGMRegions;
  }

}