package uk.ac.sanger.psu.gfmerge.datacomp;

import java.util.*;
import org.biojava.bio.seq.*;
import org.biojava.bio.symbol.*;
import uk.ac.sanger.psu.gfmerge.util.algotools.MyTreeMap;

/** class Exon is representing an biological exon which is a sublocation of a feature.
 * <p>
 * For example:
 * <pre>
 * Exon exon = new Exon(geneModel, location);
 * </pre>
 *
 * @author Sebastian R. Spiegler
 * @see uk.ac.sanger.psu.gfmerge.datacomp.Exon
 * @see uk.ac.sanger.psu.gfmerge.datacomp.FeatureComponent
 * @see uk.ac.sanger.psu.gfmerge.datacomp.ExonAble
 * @version 1.0
 */

public class Exon extends FeatureComponent implements ExonAble
{

  ///////////////////////////////////////
  // attributes

    /** typesafe enumeration of the Exon type
     */
    private ExonType exonType;

  ///////////////////////////////////////
  // constructor

    /** constructor for Exon object
     *
     * @param _gfmFeature reference to genemodel exon belongs to
     * @param _location Location of exon
     */
    public  Exon(GFMergeFeature _gfmFeature, Location _location)
    {
      super(_gfmFeature, _location);
      setExonType();
    } // end Exon


    ///////////////////////////////////////
    // operations

    /** accessor method which return exon type
     *
     * @return type of exon
     */
    public ExonType getExonType()
    {
        return exonType;
    } // end getExonType

    /** method which returns frame of the exon.
     * Not implemented !!! Returns 0.
     *
     * @return returns 0. Not implemented
     */
    public int getFrame()
    {
      // to be coded
      return 0;
    } // end getFrame


    /** accessor method which sets exon type.
     */
    private void setExonType()
    {
      int minLocGeneModel = this.getGfmFeature().getLocation().getMin();
      int maxLocGeneModel = this.getGfmFeature().getLocation().getMax();
      int minLocExon = this.getLocation().getMin();
      int maxLocExon = this.getLocation().getMax();
      StrandedFeature.Strand strand = this.getGfmFeature().getStrand();

      // Strand has to be considered
      if ((minLocGeneModel == minLocExon) && (maxLocGeneModel == maxLocExon))
      {
        exonType = ExonType.TYPE_SINGLE_EXON_GENE;
      }
      else if ((minLocGeneModel == minLocExon) && (maxLocGeneModel != maxLocExon)
               && (strand == StrandedFeature.POSITIVE))
      {
        exonType = ExonType.TYPE_5_END;
      }
      else if ((minLocGeneModel == minLocExon) && (maxLocGeneModel != maxLocExon)
               && (strand == StrandedFeature.NEGATIVE))
      {
        exonType = ExonType.TYPE_3_END;
      }
      else if ((minLocGeneModel != minLocExon) && (maxLocGeneModel == maxLocExon)
               && (strand == StrandedFeature.POSITIVE))
      {
        exonType = ExonType.TYPE_3_END;
      }
      else if ((minLocGeneModel != minLocExon) && (maxLocGeneModel == maxLocExon)
               && (strand == StrandedFeature.NEGATIVE))
      {
        exonType = ExonType.TYPE_5_END;
      }
      else
      {
        exonType = ExonType.TYPE_INTERNAL;
      }
      //debug
//        System.out.println("GeneModel Loc: " + this.getGeneModel().getLocation() + " " + this.getGeneModel().getStrand());
//        System.out.println("Exon      loc: " + this.getLocation()                + " " + this.getExonType());
//        System.out.println("");
    } // end setExonType


} // end Exon
