package uk.ac.sanger.psu.gfmerge.datacomp;

import java.util.*;
import org.biojava.bio.seq.*;
import org.biojava.bio.symbol.*;
import uk.ac.sanger.psu.gfmerge.util.biotools.*;
import uk.ac.sanger.psu.gfmerge.util.algotools.*;


/** abstract class which contains basic implementation of genemodels, blast and cDNA features.
 *
 * @see uk.ac.sanger.psu.gfmerge.datacomp.GeneModel
 * @see uk.ac.sanger.psu.gfmerge.datacomp.SimFeature
 * @version 1.0
 */
public abstract class GFMergeFeature implements GFMergeFeatureAble {

  ///////////////////////////////////////
  // attributes


    /** BioJava feature object
     */
    private StrandedFeature feature;
    /** ArrayList of exon within a genemodel
     */
    private ArrayList exons   = new ArrayList();
    /** ArrayList of introns within a genemodel
     */
    private ArrayList introns = new ArrayList();

  ///////////////////////////////////////
  // constructor

    /** constructor which contains basic implementation for GeneModel and SimFeature objects
     *
     * @param feature BioJava feature
     */
    public  GFMergeFeature(StrandedFeature feature)
    {
      setFeature(feature);
      setExons(feature);
      setIntrons(feature);
    } // end BasicFeature

    ///////////////////////////////////////
    // operations

    /** accessor method which returns the appertaining StrandedFeature retrieved from file
     *
     * @return BioJava StrandedFeature
     */
    public StrandedFeature getFeature()  //done
    {
        return feature;
    } // end getFeature

    /** accessor method which returns ArrayList of exons
     *
     * @return ArrayList of exons
     */
    public ArrayList getExons()  //done
    {
        return exons;
    } // end getExons

    /** accessor method which returns ArrayList of introns
     *
     * @return ArrayList of introns
     */
    public ArrayList getIntrons()  //done
    {
        return introns;
    } // end getIntrons

    /** accessor method which returns Location of the BioJava feature
     *
     * @return Location of the BioJava feature
     */
    public Location getLocation()  //done
    {
      return this.getFeature().getLocation();
    } // end getLocation

    /** accessor method which returns Strand of the BioJava feature
     *
     * @return Strand of the BioJava feature
     */
    public StrandedFeature.Strand getStrand()
    {
      return this.getFeature().getStrand();
    } // end getStrand


    /** accessor method which returns the length of the feature
     *
     * @return length of the feature
     */
    public long getLength()
    {
      long length;
      length = this.getLocation().getMax() - this.getLocation().getMin() + 1;
      return length;
    }

    /** accessor method which returns total exon length of the feature
     *
     * @return total exon length of the feature
     * @see uk.ac.sanger.psu.gfmerge.analysis.simfeature_analysis.TotalExonLengthAnalysis
     */
    public long getTotalExonLength()
    {
      long totalExonLength = 0;
      for(Iterator itEx = this.getExons().iterator(); itEx.hasNext();)
      {
        Exon       exon  = (Exon) itEx.next();
        totalExonLength += exon.getLength();
      }
      return totalExonLength;
    }



    /** accessor method which sets BioJava feature
     *
     * @param _feature BioJava feature
     */


    private void setFeature(StrandedFeature _feature)  //done
    {
        feature = _feature;
    } // end setFeature


    /** accessor method which calculates exons of feature.
     * Exon means sub Locations of BioJava feature.
     *
     * @param _feature BioJava feature
     */
    private void setExons(Feature _feature)  //done
    {
      for (Iterator it = _feature.getLocation().blockIterator(); it.hasNext(); )
      {
        Location l = (Location) it.next();
        exons.add(new Exon(this,l));
      }
    } // end setExons


    /** accessor method which calculates introns of feature.
     * Intron means gaps inbetween sub Locations of BioJava feature.
     *
     * @param _feature BioJava feature
     */
    private void setIntrons(Feature _feature)
    {
      int minLocGeneModel = this.getLocation().getMin();
      int maxLocGeneModel = this.getLocation().getMax();
      ArrayList exonLocs  = new ArrayList();

      for (Iterator it = _feature.getLocation().blockIterator(); it.hasNext(); )
      {
        Location l = (Location) it.next();
        exonLocs.add(l);
      }

      for (int i = 0; i < exonLocs.size() - 1 ; i++)
      {
        int intronStart = ((Location) exonLocs.get(i)).getMax() + 1;
        int intronEnd   = ((Location) exonLocs.get(i + 1)).getMin() - 1;
        Location l      = LocationTools.makeLocation(intronStart, intronEnd);
        introns.add(new Intron(this,l));
      }
    } // end setIntrons
} // end BasicFeature
