package uk.ac.sanger.psu.gfmerge.util;

import java.util.*;
import org.biojava.bio.seq.*;
import org.biojava.bio.symbol.*;

/** class PredAccuracy wraps the calculation of the average conditional probability for a genepredicter program.
 */
public class PredAccuracy
{

  ///////////////////////////////////////
  // attributes

  /** attribute records the number of true positive bases (bases which are predicted to be coding and are coding in the annotation).
   */
  private int tp;
  /** attribute records the number of true negative bases (bases which are predicted to be non-coding and are non-coding in the annotation).
   */
  private int tn;
  /** attribute records the number of false positive bases (bases which are predicted to be coding but are non-coding in the annotation).
   */
  private int fp;
  /** attribute records the number of false negative bases (bases which are predicted to be non-coding but are coding in the annotation).
   */
  private int fn;
  /** Average Conditional Probability (ACP) is suggested as an appropriate measure of global prediction accuracy (Burset and Guigo, 1996).
   * <p>
   * It is calculated by comparing the prediction to the annotation on the base level.
   * Formula used: ACP = [TP/(TP+FN)+TP/(TP+FP)+TN/(TN+FP)+TN/(TN+FN)]/4
   */
  private double acp = 0;

  ///////////////////////////////////////
  // constructor

  /** constructor which creates PredAccuracy object.
   * Calls calcACP method which will calculate the average conditional probability.
   * 
   * @param seq_org BioJava sequence object containing the annotation
   * @param seq_comp BioJava sequence object containing the prediction
   */
  public PredAccuracy(Sequence seq_org, Sequence seq_comp)
  {
    this.setTP(0);
    this.setTN(0);
    this.setFP(0);
    this.setFN(0);
    this.setACP( calcACP(seq_org,seq_comp) );
  }

  ///////////////////////////////////////
  // operations

  /** accessor method which returns number of true positive bases.
   * 
   * @return number of true positive bases
   */
  public int getTP()  {
    return this.tp;
  }

  /** accessor method which setsnumber of true positive bases.
   * 
   * @param _tp number of true positive bases
   */
  private void setTP(int _tp)  {
    this.tp = _tp;
  }

  /** accessor method which returns number of true negative bases.
   * 
   * @return number of true negative bases
   */
  public int getTN()  {
    return this.tn;
  }

  /** accessor method which sets number of true negative bases.
   * 
   * @param _tn number of true negative bases
   */
  private void setTN(int _tn)  {
    this.tn = _tn;
  }

  /** accessor method which gets number of false positive bases.
   * 
   * @return number of false positive bases
   */
  public int getFP()  {
    return this.fp;
  }

  /** accessor method which sets number of false positive bases.
   * 
   * @param _fp number of false positive bases
   */
  private void setFP(int _fp)  {
    this.fp = _fp;
  }

  /** accessor method which returns number of false negative bases.
   * 
   * @return number of false negative bases
   */
  public int getFN()  {
    return this.fn;
  }

  /** accessor method which sets number of false negative bases.
   * 
   * @param _fn accessor method which returns number of false negative bases.
   */
  private void setFN(int _fn)  {
    this.fn = _fn;
  }

  /** accessor method which returns average conditional probability value.
   * 
   * @return average conditional probability
   */
  public double getACP()   {
    return this.acp;
  }

  /** accessor method which sets average conditional probability value.
   * 
   * @param _acp average conditional probability
   */
  private void setACP(double _acp)   {
    this.acp = _acp;
  }


  /** method which calculates ACP (average conditional probability).
   * called from constructor.
   * 
   * @param seq_org annotation sequence
   * @param seq_comp annotation sequence
   * @return average conditional probability
   */
  private double calcACP(Sequence seq_org, Sequence seq_comp)
  {
    double thisAcp;
    int    thisSeqLength_org          = seq_org.length();

    // get features (in featureholder) from strand
    FeatureHolder thisFh_org_fw   = getFeaturesFromStrand(true ,seq_org);
    FeatureHolder thisFh_org_rev  = getFeaturesFromStrand(false,seq_org);
    FeatureHolder thisFh_comp_fw  = getFeaturesFromStrand(true ,seq_comp);
    FeatureHolder thisFh_comp_rev = getFeaturesFromStrand(false,seq_comp);

    // call method which gets lists of locations
    ArrayList thisArrList_org_fw    = getAListOfLocs(thisFh_org_fw);
    ArrayList thisArrList_org_rev   = getAListOfLocs(thisFh_org_rev);
    ArrayList thisArrList_comp_fw   = getAListOfLocs(thisFh_comp_fw);
    ArrayList thisArrList_comp_rev  = getAListOfLocs(thisFh_comp_rev);

    // call method to compare both strands of orignial and predicted seq
    // arguments = length, lists of locations
    compSeq(thisSeqLength_org,
            thisArrList_org_fw,
            thisArrList_org_rev,
            thisArrList_comp_fw,
            thisArrList_comp_rev);

    double tpC = (double) this.getTP(); // casting to double
    double tnC = (double) this.getTN(); // casting to double
    double fpC = (double) this.getFP(); // casting to double
    double fnC = (double) this.getFN(); // casting to double

    thisAcp = 0.25 * ( (tpC/(tpC+fnC)) + (tpC/(tpC+fpC)) + (tnC/(tnC+fpC)) + (tnC/(tnC+fnC)) );
    return thisAcp;
  }

  /** method which returns ArrayList of locations when sequence is given.
   * Called from calcACP
   * 
   * @param fh BioJava FeatureHolder
   * @return ArrayList of Locations
   */
  private static ArrayList getAListOfLocs(FeatureHolder fh)
  {
    ArrayList arrList = new ArrayList();

    // for all features
    for (Iterator itFeat = fh.features(); itFeat.hasNext(); )
    {
      Feature feature = (Feature) itFeat.next();

      // for sublocations in feature
      for (Iterator itLoc = feature.getLocation().blockIterator(); itLoc.hasNext(); )
      {
        Location loc = (Location) itLoc.next();
        arrList.add(loc);
      }
    }
    return arrList;
  }

  /** method which compares two sequences for matches and overlaps.
   * Considers forward and reverse strand seperately.
   * 
   * @param seqLength_org length of the sequence
   * @param arrList_org_fw ArrayList of bases (either c for coding or n  
   *      for non-coding) of annotaion, forward strand
   * @param arrList_org_rev ArrayList of bases (either c for coding or n 
   *      for non-coding) of annotaion, reverse  strand
   * @param arrList_comp_fw ArrayList of bases (either c for coding or n 
   *      for non-coding) of prediction, forward strand
   * @param arrList_comp_rev ArrayList of bases (either c for coding or n
   *      for non-coding) of prediction, reverse strand
   */
  private void compSeq(int seqLength_org,
                       ArrayList arrList_org_fw,
                       ArrayList arrList_org_rev,
                       ArrayList arrList_comp_fw,
                       ArrayList arrList_comp_rev)
  {
    StringBuffer seq_org_fw   = new StringBuffer();
    StringBuffer seq_org_rev  = new StringBuffer();
    StringBuffer seq_comp_fw  = new StringBuffer();
    StringBuffer seq_comp_rev = new StringBuffer();

    // initiate sequences to non-codingg
    for (int ind = 0; ind < seqLength_org; ind++)
    {
      seq_org_fw.append("N");
      seq_org_rev.append("N");
      seq_comp_fw.append("N");
      seq_comp_rev.append("N");
    }

    // mark coding nukleotides
    seq_org_fw   = markCoding(arrList_org_fw,seq_org_fw);     // org fw strand
    seq_org_rev  = markCoding(arrList_org_rev,seq_org_rev);   // org rev strand
    seq_comp_fw  = markCoding(arrList_comp_fw,seq_comp_fw);   // comp fw strand
    seq_comp_rev = markCoding(arrList_comp_rev,seq_comp_rev); // comp rev strand

    // compare orignal seq to seq_comp
    HashMap countMap = new HashMap();
    countMap = countNucleotides(countMap, seqLength_org, seq_org_fw, seq_comp_fw);    //fw strand
    countMap = countNucleotides(countMap, seqLength_org, seq_org_rev, seq_comp_rev);  //rev strand

    // set attributes
    setAttributes(countMap);

    //debug
//         System.out.println("TP:  " + tp);
//         System.out.println("TN:  " + tn);
//         System.out.println("FP:  " + fp);
//         System.out.println("FN:  " + fn);
  }

  /** method which marks coding exons as 'C' on a sequence.
   * Called from compSeq
   * 
   * @param arrListLoc ArrayList of coding Locations
   * @param seq String which contains binary variable for each position, 
   *      at the beginning all postions are non-coding, during processing     
   *     coding bases will be marked "C"
   * @return String containg coding/non-coding information for each base
   */
  private static StringBuffer markCoding(ArrayList arrListLoc,
                                         StringBuffer seq)
  {
    // for each location in array
    for (Iterator itLocArr = arrListLoc.iterator(); itLocArr.hasNext();)
    {
      Location loc = (Location) itLocArr.next();

      //mark position on sequence
      for (int indPos = loc.getMin() - 1; indPos < loc.getMax(); indPos++)
      //index starts with 0, sequence starts with 1
      {
        seq.setCharAt(indPos, 'C');
      }
    }
    return seq;
  }

  /** method which counts TP,TN,FP,FN bases.
   * Called from compSeq
   * 
   * @param thisCountMap HashMap containing tp, tn, fp, fn as keys pointing to
   *      their value, number of bases of the other strand will be added
   * @param seqLength length of sequence
   * @param seq_org annotation
   * @param seq_comp prediction
   * @return HashMap containing tp, tn, fp, fn pointing to their values
   */
  private static HashMap countNucleotides(HashMap thisCountMap,
                                          int seqLength,
                                          StringBuffer seq_org,
                                          StringBuffer seq_comp )
  {
    int myTp = 0, myTn = 0, myFp = 0,  myFn = 0;

    for (int ind = 0; ind < seqLength; ind++)
    {
      // nucleotid is predicted coding and is coding
      char orgBase  = seq_org.charAt(ind);
      char compBase = seq_comp.charAt(ind);

      // comparison
      if (( orgBase == compBase ) && ( compBase == 'C'))
      {
        myTp++;
      }

      // nucleotid is predicted non-coding and is non-coding
      else if (( orgBase == compBase ) && ( compBase == 'N'))
      {
        myTn++;
      }

      // nucleotid is predicted coding and is non-coding
      else if (( orgBase != compBase ) && ( compBase == 'C'))
      {
        myFp++;
      }

      // nucleotid is predicted non-coding and is coding
      else if (( orgBase != compBase ) && ( compBase == 'N'))
      {
        myFn++;
      }
    }

    // add counted bases to hash which will be returned
    if(thisCountMap.containsKey("tp"))
    {
      myTp += ((Integer) thisCountMap.get("tp")).intValue();
      thisCountMap.put("tp",new Integer(myTp));
    }
    else
    {
      thisCountMap.put("tp",new Integer(myTp));
    }


    if(thisCountMap.containsKey("tn"))
    {
      myTn += ((Integer) thisCountMap.get("tn")).intValue();
      thisCountMap.put("tn",new Integer(myTn));
    }
    else
    {
      thisCountMap.put("tn",new Integer(myTn));
    }


    if(thisCountMap.containsKey("fp"))
    {
      myFp += ((Integer) thisCountMap.get("fp")).intValue();
      thisCountMap.put("fp",new Integer(myFp));
    }
    else
    {
      thisCountMap.put("fp",new Integer(myFp));
    }


    if(thisCountMap.containsKey("fn"))
    {
      myFn += ((Integer) thisCountMap.get("fn")).intValue();
      thisCountMap.put("fn",new Integer(myFn));
    }
    else
    {
      thisCountMap.put("fn",new Integer(myFn));
    }

    return thisCountMap;
  }

  /** method which returns featureholder of all features on either forward or
   * reverse strand
   * 
   * @param pos_strand switch for strand
   * @param fh BioJava FeatuerHolder
   * @return featureholder of all features on either forward or reverse strand
   */
  private static FeatureHolder getFeaturesFromStrand(boolean pos_strand, FeatureHolder fh)
  {
    FeatureHolder new_fh;
    FeatureFilter ff;

    if(pos_strand)
    {
      ff = new FeatureFilter.StrandFilter(StrandedFeature.POSITIVE);
    }
    else
    {
      ff = new FeatureFilter.StrandFilter(StrandedFeature.NEGATIVE);
    }
    new_fh = fh.filter(ff);
    return new_fh;
  }

  /** method which saves HashMap containing the tp, tn, fp, fn values to their appropriate attributes.
   * 
   * @param _countMap HashMap containing the tp, tn, fp, fn values
   */
  private void setAttributes(HashMap _countMap)
  {
    setTP(   (  (Integer) _countMap.get("tp")   ).intValue()  );
    setTN(   (  (Integer) _countMap.get("tn")   ).intValue()  );
    setFP(   (  (Integer) _countMap.get("fp")   ).intValue()  );
    setFN(   (  (Integer) _countMap.get("fn")   ).intValue()  );
  }

}
