package uk.ac.sanger.psu.gfmerge.util.biotools;

import java.io.*;
import java.util.*;
import org.biojava.bio.*;
import org.biojava.bio.seq.*;
import org.biojava.bio.seq.io.*;
import org.biojava.bio.seq.impl.*;
import org.biojava.bio.symbol.*;

import uk.ac.sanger.psu.gfmerge.util.algotools.*;
import uk.ac.sanger.psu.gfmerge.datacomp.*;
import uk.ac.sanger.psu.gfmerge.util.debug.*;
import uk.ac.sanger.psu.gfmerge.analysis.*;
import uk.ac.sanger.psu.gfmerge.*;

import java.lang.Math;

/** class BioTools contains basic processing tools within GFMerge.
 *
 * @author Sebastian R. Spiegler
 * @version 1.0
 */
public class BioTools {

  /** Seq2Seq method returns a sequence object when sequence iterator is given.
   * <p>
   * Called by the main method in GFMerge.
   *
   * @param sequences SequenceIterator object which may contain more than one
   *     more than one sequence, but only the first sequence is returned
   * @return Sequence  Sequence object
   * @see org.biojava.bio.seq.SequenceIterator
   * @see org.biojava.bio.seq.Sequence
   */
  public static Sequence seqs2Seq(SequenceIterator sequences) throws BioException
  {
    Sequence seq = null;

    //iterate through the sequences
    while (sequences.hasNext()) {
      try {
        seq = sequences.nextSequence();
      }

//      catch (BioException ex) {
//        //not in EMBL format
//        ex.printStackTrace();
//      }

      catch (NoSuchElementException ex) {
        //request for more sequence when there isn't any
        ex.printStackTrace();
      }
    }
    return seq;
  }

  /** method returns genefinder name of geneprediction if all features have the same 'method' value in the prediction file otherwise it throughs an exception.
   * <p>
   * Called by the main method in GFMerge.
   *
   * @exception NoSuchElementException If no method entry is in file
   * @exception "Exception" If different method values are in file
   * @param _sequence Sequence object
   * @param _fname filename of the prediction file which is needed for
   *     the exception method
   * @return String                    method value of the prediction file
   * @see org.biojava.bio.seq.Sequence
   */
  public static String getGenePredicterFromSequence(Sequence _sequence,
      String _fname) {
    String method = null;
    ArrayList genePredArr = new ArrayList();
    boolean errorFlag = false;

    // for each feature get 'method'
    for (Iterator it = _sequence.features(); it.hasNext(); ) {
      Feature f = (Feature) it.next();
      try {
        genePredArr.add(f.getAnnotation().getProperty("method"));
      }
      // if there are no "method" entries, exit
      catch (NoSuchElementException ex) {
        System.err.println("Error in " + _fname);
        System.err.println("Features have no method entries");
        System.err.println(
            "class uk.ac.sanger.psu.gfmerge.util.biotools.BioTools");
        System.exit(1);
      }
    }

    // check "method" of each feature
    if (genePredArr.size() > 1) {
      for (int i = 1; i < genePredArr.size(); i++) {
        String pred = (String) genePredArr.get(i - 1);
        String succ = (String) genePredArr.get(i);

        //debug
//            System.out.print(pred + " " + succ + " ");
//            System.out.println(pred.compareTo(succ));

        if (pred.compareTo(succ) != 0) {
          errorFlag = true;
        }
      }
    }

    // if not all "method" entries are the same, exit
    if (errorFlag == true) {
      System.err.println("Error in file " + _fname);
      System.err.println("Features have different method entries");
      System.err.println(
          "class uk.ac.sanger.psu.gfmerge.util.biotools.BioTools");
      System.exit(1);
    }

    try
    {
      method = (String) genePredArr.get(0);
    }
    catch(Exception ex)
    {

      System.out.println("Warning while retrieving sequence from prediction file");
      System.out.println("Sequence might not have features...");

      // System.err.println(ex.getMessage());
      // System.exit(-1);
      return "Undefined";
    }
    return method;
  }





////////////////////////////////////////////////////////////////////////
// Creates sequence object of a sequence and picked genemodels
//
/** method which creates sequence object of a sequence and picked genemodels.
 *
 * @param thisPredSeq BioJava sequence object
 * @param thisRegionsOfPickedGmsArr ArrayList of genemodel regions
 *     containing highscoring genemodels
 * @return new BioJava sequence object containing picked genemodels
 */
public static Sequence createNewSeqOfPickedGms(Sequence thisPredSeq,
                                               ArrayList thisRegionsOfPickedGmsArr)
{
  SymbolList thisSymList = thisPredSeq.subList(1, thisPredSeq.length());
  SimpleSequence newSeq = new SimpleSequence(thisSymList, thisPredSeq.getURN(),
                                             thisPredSeq.getName(),
                                             Annotation.EMPTY_ANNOTATION);

  for (Iterator it = thisRegionsOfPickedGmsArr.iterator(); it.hasNext(); )
  {
    GFMergeRegion gmsArr  = (GFMergeRegion) it.next();

    for (Iterator itGms = gmsArr.getLocateable().iterator(); itGms.hasNext(); )
    {
      GeneModel gm = (GeneModel) itGms.next();
      StrandedFeature.Template ft = new StrandedFeature.Template();
      try
      {
        ft.populate(gm.getFeature());
      }
      catch (Exception ex) {}

      try
      {
        newSeq.createFeature(ft);
      }
      catch (Exception ex) {}
    }
  }
  return newSeq;
}






////////////////////////////////////////////////////////////////////////
// Recursive computing of by start position sorted features
// to retrieve regions of features
//

  /** Search method calculates regions of overlapping features.
   * <p>
   * Recursive computing of features sorted by start position in order to retrieve regions of continuously overlapping features (genemodel, cDNA, Blast)
   *
   * @param tMap binary tree instanced from {@link MyTreeMap} each
   *     key-value-pair contains the start location of features as key and an
   *     array of {@link LocateAble} as value
   * @return <code>ArrayList</code> of Regions which contain themselves
   *     LocateAble objects
   * @see uk.ac.sanger.psu.gfmerge.util.algotools.MyTreeMap
   * @see uk.ac.sanger.psu.gfmerge.datacomp.LocateAble
   */
  public static ArrayList search(MyTreeMap tMap) {
    ArrayList allRegionsArrList = new ArrayList(); // ArrayList which contains regions
    HashMap doneObjects = new HashMap();

    for (Iterator itTM = tMap.keySet().iterator(); itTM.hasNext(); ) { //for each array of LocateAble objects
      ArrayList objArrList = (ArrayList) tMap.get(itTM.next());

      for (Iterator itObj = objArrList.iterator(); itObj.hasNext(); ) { //for each object within an array
        LocateAble obj = (LocateAble) itObj.next();
        ArrayList regionArrList = new ArrayList();

        regionArrList = find(obj, doneObjects, tMap);

        // if region is not empty then add it to allRegionsArray
        if (regionArrList != null) { // region must not be null (empty)!!!!
          // get location of region
          Location regionLoc;
          int minLoc = Integer.MAX_VALUE;
          int maxLoc = Integer.MIN_VALUE;

          for (Iterator it = regionArrList.iterator(); it.hasNext(); ) {
            LocateAble thisObj = (LocateAble) it.next();
            int thisMinLoc = thisObj.getLocation().getMin();
            int thisMaxLoc = thisObj.getLocation().getMax();
            if (thisMinLoc < minLoc) {
              minLoc = thisMinLoc;
            }
            if (thisMaxLoc > maxLoc) {
              maxLoc = thisMaxLoc;
            }
          }

          regionLoc = LocationTools.makeLocation(minLoc, maxLoc);

          // add region(location, arraylist of genemodels) to array list of regions
          allRegionsArrList.add(new GFMergeRegion(regionLoc, regionArrList));

        } // end if
      }
    }
    return allRegionsArrList;
  }

  /** Find method contains recursive search for regions.
   * <p>
   * Called by the search method in BioTools.
   *
   * @param obj feature object with Location (genemodel, cDNA, Blast)
   * @param doneObjects HashMap of seen objects
   * @param tMap binary tree instanced from {@link MyTreeMap} each
   *     key-value-pair contains the start location of features as key and an
   *     array of {@link LocateAble} as value
   * @param gm objects of {@link LocateAble} for which overlaps are
   *     searched for
   * @param doneObjs tree of {@link HashMap} which contains processed
   *     objects
   * @return <code>ArrayList</code> of regions which are list of overlapping
   *     LocateAble objects
   */
  private static ArrayList find(LocateAble obj, HashMap doneObjects,
                                MyTreeMap tMap) {
    // conditions for recursion
    // case 1: if key is already done, return null
    if (doneObjects.containsKey(obj)) {
      return null;
    }

    // case 2: if key is new, start processing
    // search for overlaps for obj and search for overlaps of each overlapping
    // objs
    else {
      ArrayList regionTemp = new ArrayList();

      doneObjects.put(obj, "done"); // mark object processed
      ArrayList overlappingObjs = new ArrayList();
      overlappingObjs = findOverlaps(obj, tMap);
      regionTemp.add(obj);

      for (Iterator itObj = overlappingObjs.iterator(); itObj.hasNext(); ) {
        LocateAble overlappingObj = (LocateAble) itObj.next();
        ArrayList temp = find(overlappingObj, doneObjects, tMap);
        if (temp != null) {
          regionTemp.addAll(temp);
        }
      }
      return regionTemp;
    }
  }

  /** FindOverlaps method returns all overlapping objects in tree of objects.
   * <p>
   * Called by the find method in BioTools.
   *
   * @param obj objects of {@link LocateAble} for which overlaps are searched
   *     for
   * @param tMap binary tree instanced from {@link MyTreeMap} each
   *     key-value-pair contains the start location of features as key and an
   *     array of {@link LocateAble} as value
   * @return <code>ArrayList</code> of LocateAble objects which overlap a
   *     certain LocateAble object
   */
  private static ArrayList findOverlaps(LocateAble obj, MyTreeMap tMap) {
    ArrayList overlapingObjsTemp = new ArrayList();

    for (Iterator itTM = tMap.keySet().iterator(); itTM.hasNext(); ) { //for each array of objects

      ArrayList objArrList = (ArrayList) tMap.get(itTM.next());

      for (Iterator itObj = objArrList.iterator(); itObj.hasNext(); ) { //for each object within an array
        LocateAble objComp = (LocateAble) itObj.next();
        Location locOrg = obj.getLocation();
        Location locComp = objComp.getLocation();

//        System.out.println("Region overlap loc org : "+locOrg);
//        System.out.println("Region overlap loc comp: "+locComp);
//        System.out.println("Region overlap: "+LocationTools.overlaps(locOrg, locComp));
//        System.out.println("---");

        if(BioTools.overlaps(locOrg, locComp))
        //if(LocationTools.intersection(locOrg, locComp) != Location.empty)
        //if (LocationTools.overlaps(locOrg, locComp))
        // changed to eliminate the bug, where a single exon gene is
        // within an intron location of another gene and the method
        // LocationTools.overlaps() returns false
        {
          overlapingObjsTemp.add(objComp);
        }
      }
    }
    return overlapingObjsTemp;
  }





  /** Overlaps method contains the BioJava overlap functionality and checks
   * whether one location is contained by the other
   * <p>
   * It was introduced because BioJava just compares sub locations and does not
   * consider overlaps of introns
   *
   * @param locOrg original location
   * @param locComp location to compare
   * @return boolean overlaps
   */
  public static boolean overlaps(Location locOrg, Location locComp)
  {
    boolean overlaps = false;

    if(LocationTools.overlaps(locOrg, locComp))
    {
      return overlaps = true;
    }
    // check whether one gene model is within the other (within an intron)
    else if ( (locOrg.getMin()  < locComp.getMin() && locOrg.getMax()  > locComp.getMax()) ||
              (locComp.getMin() < locOrg.getMin()  && locComp.getMax() > locOrg.getMax()))
    {
      return overlaps = true;
    }
    else
    {
      return overlaps = false;
    }
  }




////////////////////////////////////////////////////////////////////////
// Tools for creating a tree structure of geneModels
//
//





/** AddToTreeMap method creates a binary tree of locateable objects sorted by
 * their start location.
 * <p>
 * Key is the start location of a LocateAble object,
 * Value is an array of LocateAble objects with the same start location
 *
 * @param tMap binary tree instanced from {@link MyTreeMap} each
 *     key-value-pair contains the start location of features as key and an
 *     array of {@link LocateAble} as value. tree is empty when called first,
 *     new leaves will be added.
 * @param locatableArrList <code>ArrayList</code> of LocateAble objects
 * @return MyTreeMap        binary tree of LocateAble objects
 * @see uk.ac.sanger.psu.gfmerge.util.algotools.MyTreeMap
 * @see uk.ac.sanger.psu.gfmerge.datacomp.LocateAble
 */

public static MyTreeMap addToTreeMap(MyTreeMap tMap,
                                     ArrayList locatableArrList)
{
  for (Iterator it = locatableArrList.iterator(); it.hasNext(); )
  {
    // array list contains locatable objects
    LocateAble locatable = (LocateAble) it.next();

    // get location to calculate start pos. + length of gene model
    Location loc = locatable.getLocation();
    Long minLoc = new Long(loc.getMin());
    long length = loc.getMax() - loc.getMin() + 1;

    // max length of locateable objects stored in class MyTreeMap
    long maxObjectFound = tMap.getMaxObjectFound();

    if (tMap.containsKey(minLoc) == false)
    { // key does not exist so far
      ArrayList arrList_new = new ArrayList();
      arrList_new.add(locatable);

      tMap.put(minLoc, arrList_new);
    }
    else
    { // key exists already
      ArrayList arrList_existed = (ArrayList) tMap.get(minLoc);
      arrList_existed.add(locatable);
      tMap.put(minLoc, arrList_existed);
    }

    if (maxObjectFound < length)
    {
      tMap.setMaxObjectFound(length);
    }
  }
  return tMap;
}
}
