package uk.ac.sanger.psu.gfmerge.util.getopt;

import java.util.*;
import uk.ac.sanger.psu.gfmerge.util.getopt.*;
import org.biojava.bio.seq.*;
import org.biojava.bio.BioException;

import uk.ac.sanger.psu.gfmerge.util.biotools.*;
import uk.ac.sanger.psu.gfmerge.util.io.*;


/** class for objects containing the command line arguments.
 * Input files used have to be in EMBL format.
 *
 * @author Sebastian R. Spiegler
 * @version 1.0
 */
public class SimpleArg {

  ///////////////////////////////////////
  // attributes

  /** annotation of training set (for acp calculation)
   */
  private String    annotatedTrainSet = null;                 //contains filename of annotated training set
  /** predictions of training set (for acp calculation).
   */
  private ArrayList predictedTrainSet = new ArrayList();      //contains filename array of predictions on training set
  /** predictions which will be collapsed into one (EMBL format).
   */
  private ArrayList predictions       = new ArrayList();      //contains filename array of predictions
  /** blast file which is used for evidence of correct predictions (EMBL format).
   */
  private String    blastFile         = null;                 //contains filename of blast file
  /** cDNA file which is used for evidence of correct predictions (EMBL format).
   */
  private String    cDNAFile          = null;                 //contains filename of cDNA file
  /** debugging mode.
   */
  private int debug                   = 0;

  /** filename of output file.
   */
  private String filename             = null;                 //contains filename of output file
  /** switch for output of a file containt all removed genemodels as well as a file of each analysis level with the removed genemodels.
   */
  private boolean fileOfRemovedGms    = false;                //contains filename of output file
  /** switch for strand consideration of cDNA.
   * Default is false.
   */
  private boolean considerCDnaStrand  = false;                //contains flag for cDNA strand consideration
  /** allowed difference between genemodel and cDNA splice site to be assigned as correct splice site.
   */
  private int fuzzyBoundaries         = 15;                    //contains fuzzy boundary value (cDNA overlap)


  ///////////////////////////////////////
  // constructor

  /** constructor of SimpleArg objects.
   *
   * @param args command line arguments from main method
   */
  public SimpleArg(String[] args)
  {
    setArguments(args);
    checkArguments();
  }

  ///////////////////////////////////////
  // operations

  /** method which retrieves command line arguments and saves them.
   *
   * @param _args
   */
  private void setArguments(String[] _args)
  {
    // initialization
    GetOpt go = new GetOpt("a:t:p:b:c:f:rx:yd:");
    String   optA = null;        // option a contains annotated training set
    String[] optT = null;        // option t contains predicted training set
    String[] optP = null;        // option p contains predictions
    String   optB = null;        // option b contains blast file
    String   optC = null;        // option c contains cDNA file
    String   optF = null;        // option f contains filename of output file
    String   optR = null;        // option r switch for output of removed gms
    String   optX = null;        // option x contains fuzzy boundaries length
    String   optY = null;        // option x consider cDNA strand
    String   optD = null;        // debug mode
    char c;

    // checking arguments
    while ((c = go.getopt(_args)) != GetOpt.DONE)
    {
      switch(c)
      {
        case 'h':                                         // help
          doHelp(0);
          break;
        case 'a':                                         // 1 annotated training set
          optA = go.optarg(  );
          this.annotatedTrainSet = optA;
          break;
        case 't':                                         // x predicted training set
          optT = (go.optarg(  )).split("\\s+");
          predictedTrainSet = strArrToArrList(optT);
          break;
        case 'p':                                         // x prediction files
          optP = (go.optarg(  )).split("\\s+");
          predictions = strArrToArrList(optP);
          break;
        case 'b':                                         // 1 blast file
          optB = go.optarg(  );
          this.blastFile = optB;
          break;
        case 'c':                                         // 1 cDNA file
          optC = go.optarg(  );
          this.cDNAFile = optC;
          break;

        case 'f':                                         // 1 output filename
          optF = go.optarg(  );
          this.filename = optF;
          break;

        case 'r':                                         // boolean switch for
          optR = go.optarg(  );                           // output of removed gms
          this.fileOfRemovedGms = true;
          break;

        case 'x':                                         // length of fuzzy bound.
          optX = go.optarg(  );
          this.fuzzyBoundaries = new Integer(optX).intValue();
          break;

        case 'y':                                         // boolean switch for
          optY = go.optarg(  );                           // cDNA strand consideration
          this.considerCDnaStrand = true;
          break;

        case 'd':                                         // debug
          optD     = go.optarg(  );
          this.debug = new Integer(optD).intValue();
          break;

        default:
          System.err.println("Unknown option character " + c);
          doHelp(1);
      }
    }
  }

  /** Stub for providing help on usage
   *
   * @param returnValue
   */
  private void doHelp(int returnValue)
  {
    System.err.println("Usage: GFMerge ");
    System.err.println("[-h]                            help");
    System.err.println("[-a]                            annotated training set");
    System.err.println("[-t \"file1 file2 .. fileN\"]     predictions on training set");
    System.err.println("[-p \"file1 file2 .. fileN\"]     predictions of gene predicters");
    System.err.println("[-b]                            blast file");
    System.err.println("[-c]                            cDNA file");
    System.err.println("[-f]                            filename for output");
    System.err.println("[-r]                            switch for output of removed features");
    System.err.println("[-x]                            fuzzy cDNA boundaries");
    System.err.println("[-y]                            consider cDNA strand");
    System.exit(returnValue);
  }

  /** method which converts String[] into an ArrayList
   *
   * @param strArr String[]
   * @return ArrayList
   */
  private ArrayList strArrToArrList (String[] strArr)
  {
    ArrayList arrList = new ArrayList();
    for(int i = 0 ; i < strArr.length; i++)
    {
      arrList.add(strArr[i].toString());
    }
    return arrList;
  }
  /** method which checks arguments (briefly)
   */
  private void checkArguments()
  {
    if( (annotatedTrainSet == null)                ||
        (predictedTrainSet.size() == -1)           ||
        (predictedTrainSet.size() == 0)            ||
        (predictions.size() == -1)                 ||
        (predictions.size() == 0)                  ||
        (filename == null)                         ||
        (predictedTrainSet.size() != predictions.size()))
    {
      System.err.println("insufficient arguments");
      doHelp(1);
    }


  }

  ///////////////////////////////////////
  // accessor operations


  /** accessor method which returns the filename of the annotated training set file
   * stored in attribute 'annotatedTrainSet',
   * retrieved from command-line argument (-a)
   *
   * @return String
   */
  public String getAnnotatedTrainSet()
  {
    return this.annotatedTrainSet;
  }


  /** accessor method which returns a sequence object of the annotated training set,
   * used filename is stored in attribute 'annotatedTrainSet' ,
   * retrieved from command-line argument (-a)
   *
   * @return Sequence
   */
  public Sequence getAnnotatedTrainSet_Sequence()
  {
    // get filename
    String filename = (String) this.getAnnotatedTrainSet();

    // get single sequence out of sequence iterators
    SequenceIterator sequences = GFMergeInput.getSeq(filename);

    Sequence seq = null;

    try
    {
      seq = BioTools.seqs2Seq(sequences);
    }
    catch (BioException ex)
    {
      //not in EMBL format
      System.err.println("Error in file " + filename);
      ex.printStackTrace();
    }


    return seq;
  }


  /** accessor method which returns an array list of filenames of predictions from
   * different gene prediction programs on the training set,
   * retrieved from command-line argument (-t)
   *
   * @return ArrayList
   */
  public ArrayList getPredictedTrainSet()
  {
    return this.predictedTrainSet;
  }


  /** accessor method which returns an array list of sequence objects containing
   * predictions on the training set of different gene predicters,
   * used filename are stored in attribute 'predictedTrainSet' ,
   * retrieved from command-line argument (-t)
   *
   * @return ArrayList of Sequences
   */
  public ArrayList getPredictedTrainSet_Sequences()
  {
    ArrayList seqArrList = new ArrayList();

    for (Iterator it = this.getPredictedTrainSet().iterator(); it.hasNext();)
    {
      // get filename
      String filename = (String) it.next();

      // get single sequence out of sequence iterators
      SequenceIterator sequences = GFMergeInput.getSeq(filename);

      Sequence seq = null;
      try
      {
        seq = BioTools.seqs2Seq(sequences);
      }
      catch (BioException ex)
      {
        //not in EMBL format
        System.err.println("Error in file " + filename);
        ex.printStackTrace();
      }

      seqArrList.add(seq);
    }
    return seqArrList;
  }

  /** accessor method which returns an array list of filenames of predictions,
   * stored in attribute 'predictions',
   * retrieved from command-line argument (-p)
   *
   * @return ArrayList of Strings
   */
  public ArrayList getPredictions()
  {
    return this.predictions;
  }


   /** accessor method which returns an array list of sequence objects containing
    * predictions of different gene predicters,
    * used filename are stored in attribute 'predictions' ,
    * retrieved from command-line argument (-p)
    *
    * @return ArrayList of Sequences
    */
   public ArrayList getPredictions_Sequences()
   {
     ArrayList seqArrList = new ArrayList();

     for (Iterator it = this.getPredictions().iterator(); it.hasNext();)
     {
       // get filename
       String filename = (String) it.next();

       // get single sequence out of sequence iterators
       SequenceIterator sequences = GFMergeInput.getSeq(filename);

       Sequence seq = null;
       try
       {
         seq = BioTools.seqs2Seq(sequences);
       }
       catch (BioException ex)
       {
         //not in EMBL format
         System.err.println("Error in file " + filename);
         ex.printStackTrace();
       }

       seqArrList.add(seq);
     }
     return seqArrList;
   }


//   /**
//    * get-method which returns an array list of filenames of weightmatrices,
//    * stored in attribute 'weightMat',
//    * retrieved from command-line argument (-q)
//    * @return ArrayList of Strings
//    */
//  public ArrayList getWeightMat()
//  {
//    return weightMat;
//  }

  /** accessor method which returns debug mode (int variable).
   *
   * @return int
   * @see uk.ac.sanger.psu.gfmerge.util.debug.Debug
   */
  public int getDebug()
  {
    return this.debug;
  }

  /** accessor method which returns an blast filename,
   * retrieved from command-line argument (-b)
   *
   * @return blast filename
   */
  public String getBlastFile()
  {
    return this.blastFile;
  }

  /** accessor method which returns an cDNA filename,
   * retrieved from command-line argument (-c)
   *
   * @return cDNA filename
   */
  public String getCDNAFile()
  {
    return this.cDNAFile;
  }

  /** accessor method which returns an output filename,
   * retrieved from command-line argument (-f)
   *
   * @return output filename
   */
  public String getOutputFilename()
  {
    return this.filename;
  }

  /** accessor method returns switch whether removed genemodels should be saved to file or not. retrieved from command-line argument (-b). Default is no output.
   *
   * @return boolean switch for ouput of removed files
   */
  public boolean doFileOfRemovedGms()
  {
    return this.fileOfRemovedGms;
  }

  /** accessor method which returns the value of tolerance inbetween cDNA and genemodel splice sites.
   *
   * @return tolerance inbetween cDNA and genemodel splice sites
   */
  public int getFuzzyBoundariesValue()
  {
    return this.fuzzyBoundaries;
  }

  /** accessor method which returns switch for cDNA strand consideration. Default is false/no strand consideration.
   *
   * @return switch for cDNA strand consideration
   */
  public boolean doConsiderCDnaStrand()
  {
    return this.considerCDnaStrand;
  }



}
