/*
 * Copyright (c) 2003 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <lsf/lsbatch.h>

/**********************************************************************
 *
 * bjobs_plus
 *
 * The standard LSF bjobs program does not fully report program state
 * or exit codes and does not produce a format suitable for machine
 * processing. This program addresses these issues.
 *
 * It takes the following command line arguments:
 *
 *  -h               -- prints usage and exits
 *  -q <queue name>  -- sets the queue to query (default all queues)
 *  -u <user name>   -- sets the user to query (default all users)
 *
 * Prints job info in CSV format to STDOUT, one line per job. The
 * format is job id, user id, status, from host, exec hosts, exit
 * value. Where there is more than one exec host, their names are
 * separated by spaces.
 *
 * Author: Keith James <kdj@sanger.ac.uk>
 *
 **********************************************************************/

void print_job_info(struct jobInfoEnt *job);

void usage();

int main(int argc, char **argv)
{
    int c;

    int options = ALL_JOB;
    char *username_ptr = "all";
    char *qname_ptr = NULL;

    struct jobInfoEnt *job;
    int more;

    while ((c = getopt (argc, argv, "hq:u:")) != -1)
    {
        switch (c)
        {
            case 'h':
                usage();
                break;

            case 'q':
                qname_ptr = optarg;
                break;

            case 'u':
                username_ptr = optarg;
                break;

            default:
                usage();
                break;
        }
    }

    if (lsb_init(NULL))
    {
        fprintf(stderr, "Call to lsb_init failed: %s\n",
                lsb_sysmsg());
        exit(-1);
    }

    if (lsb_openjobinfo(0, NULL, username_ptr, qname_ptr, NULL, options) < 0)
    {
        /* This condition occurs if no matching job is found, which
         * isn't really an error */
        exit(0);
    }

    for (;;)
    {
        job = lsb_readjobinfo(&more);
        
        if (job == NULL)
        {
            fprintf(stderr, "Call to lsb_readjobinfo failed: %s\n",
                    lsb_sysmsg());
            exit(-1);
        }

        print_job_info(job);

        if (! more) 
            break;
    }

    lsb_closejobinfo();
    exit(0);
}

/**********************************************************************
 *      Name:  print_job_info
 *   Purpose:  Prints job info in CSV format to STDOUT, one line per
 *             job. The format is job id, user id, status, from host,
 *             exec hosts, exit value. Where there is more than one
 *             exec host, their names are separated by spaces.
 *
 * Arguments:  job -- a jobInfoEnt struct
 *
 *   Returns:  void
 **********************************************************************/
void print_job_info(struct jobInfoEnt *job)
{
    int i;
    char sep;

    printf("%d,%s,%d,%s",
           job->jobId, job->user, job->status, job->fromHost);

    if (job->numExHosts == 0)
    {
        printf(",null");
    }
    else if (job->numExHosts == 1)
    {
        printf(",%s", (job->exHosts)[0]);
    }
    else
    {
        sep = ',';
        for (i = 0; i < job->numExHosts; i++)
        {
            if (i > 0)
            {
                sep = ' ';
            }

            printf("%c%s", sep, (job->exHosts)[i]);
        }
    }

    printf(",%d\n", job->exitStatus);
}

void usage()
{
    fprintf(stderr, "Usage: bjobs_plus [-q <queue>] [-u <user>] \n");
    exit(8);
}
