/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

#include <stdlib.h>
#include <string.h>
#include <lsf/lsbatch.h>
#include "lsf_util.h"
#include "uk_ac_sanger_jcon_lsf_LSFBatchInfo.h"

/* For ArrayList */
int initial_cap;

/* java.util.ArrayList */
jclass    alClazz;
jmethodID alCon;
jmethodID alAdd;

/* java.lang.String */
jclass    stClazz;
jmethodID stCon;
jmethodID stLength;

/* uk.ac.sanger.jcon.lsf.LSFJobInfo */
jclass    jiClazz;
jmethodID jiCon;

/* Functions local */
void get_classes_and_methods(JNIEnv *env);

jobject create_job_info(JNIEnv *env, struct jobInfoEnt *job);

/*
 * Returns an uk.ac.sanger.psu.jobcontrol.lsf.LSFJobInfo object for a
 * job with the specified LSF job id.
 */
JNIEXPORT jobject JNICALL Java_uk_ac_sanger_jcon_lsf_LSFBatchInfo_getJobInfo__I
(JNIEnv *env, jobject this, jint id)
{
    /* For LSF call */
    struct jobInfoEnt *job;
    int options = ALL_JOB;

    jobject jobInfo;
    int more;

    char message [256];
    char *message_ptr = message;

    get_classes_and_methods(env);

    /* Start LSF calls */
    if (lsb_init(NULL))
    {
        throw_lsf_exception(env, "Call to lsb_init failed",
                            lsb_sysmsg());
        return NULL;
    }

    if (lsb_openjobinfo(id, NULL, NULL, NULL, NULL, options) < 0)
    {
        snprintf(message_ptr, 256,
                 "Call to lsb_openjobinfo failed for job %d", id);
        throw_lsf_exception(env, message_ptr, lsb_sysmsg());
        return NULL;
    }

    for (;;)
    {
        job = lsb_readjobinfo(&more);

        if (job == NULL)
        {
            snprintf(message_ptr, 256,
                 "Call to lsb_readjobinfo failed for job %d", id);
            throw_lsf_exception(env, message_ptr, lsb_sysmsg());
            return NULL;
        }

        if (job->jobId == id)
        {
            jobInfo = create_job_info(env, job);
            break;
        }

        if (! more)
            break;
    }

    lsb_closejobinfo();
    /* End LSF calls */

    if (jobInfo != NULL)
        return jobInfo;

    throw_lsf_exception(env, "Call to lsb_openjobinfo failed",
                        lsb_sysmsg());
    return NULL;
}

/*
 * Returns java.util.ArrayList of
 * uk.ac.sanger.jcon.lsf.LSFJobInfo objects describing jobs
 * specified by the array of jint LSF job ids.
 */
JNIEXPORT jobject JNICALL Java_uk_ac_sanger_jcon_lsf_LSFBatchInfo_getJobInfo___3I
(JNIEnv *env, jobject this, jintArray queryIds)
{
    /* For LSF call */
    struct jobInfoEnt *job;
    int options= ALL_JOB;

    jint jobIdCount;
    jboolean isCopy;
    jint *queryId_ptr;
    int queryId;

    jobject jobInfo;
    jobject jobInfoList;
    int more;
    int i;

    char message [256];
    char * message_ptr = message;

   /* Initial capacity of ArrayList */
    initial_cap = 20;

    get_classes_and_methods(env);
    jobInfoList = (*env)->NewObject(env, alClazz, alCon, initial_cap);

    /* Start LSF calls */
    if (lsb_init(NULL))
    {
        throw_lsf_exception(env, "Call to lsb_init failed",
                            lsb_sysmsg());
        return jobInfoList;
    }

    jobIdCount = (*env)->GetArrayLength(env, queryIds);
    queryId_ptr = (*env)->GetIntArrayElements(env, queryIds, &isCopy);

    for (i = 0; i < jobIdCount; i++)
    {
        queryId = queryId_ptr[i];

        if (lsb_openjobinfo(queryId, NULL, NULL, NULL, NULL, options) < 0)
        {
            snprintf(message_ptr, 256,
                 "Call to lsb_openjobinfo failed for job %d", queryId);
            throw_lsf_exception(env, message_ptr, lsb_sysmsg());
            return jobInfoList;
        }

        for (;;)
        {
            job = lsb_readjobinfo(&more);

            if (job == NULL)
            {
                snprintf(message_ptr, 256,
                         "Call to lsb_readjobinfo failed for job %d", queryId);
                throw_lsf_exception(env, message_ptr, lsb_sysmsg());
                return jobInfoList;
            }

            if (job->jobId == queryId)
            {
                jobInfo = create_job_info(env, job);
                (*env)->CallBooleanMethod(env, jobInfoList, alAdd, jobInfo);
            }

            if (! more)
                break;
        }

        lsb_closejobinfo();
    }
    /* End LSF calls */

    if (isCopy == JNI_TRUE)
    {
        (*env)->ReleaseIntArrayElements(env, queryIds, queryId_ptr, 0);
    }

    return jobInfoList;
}

/*
 * Returns java.util.ArrayList of
 * uk.ac.sanger.jcon.lsf.LSFJobInfo objects describing jobs
 * on queue specified by the name jstring queue.
 */
JNIEXPORT jobject JNICALL
Java_uk_ac_sanger_jcon_lsf_LSFBatchInfo_getJobInfo__Ljava_lang_String_2
(JNIEnv *env, jobject this, jstring queue)
{
    /* For LSF call */
    struct jobInfoEnt *job;
    int options = ALL_JOB;

    char *username_ptr = "all";
    char *qname_ptr = NULL;

    jobject jobInfo;
    jobject jobInfoList;
    int more;

    /* Initial capacity of ArrayList */
    initial_cap = 20;
    
    get_classes_and_methods(env);
    qname_ptr = dynamic_jstring_chars(env, stLength,
                                      queue, qname_ptr);

    jobInfoList = (*env)->NewObject(env, alClazz, alCon, initial_cap);

    /* Start LSF calls */
    if (lsb_init(NULL))
    {
        free(qname_ptr);
        throw_lsf_exception(env, "Call to lsb_init failed",
                            lsb_sysmsg());
        return jobInfoList;
    }

    if (lsb_openjobinfo(0, NULL, username_ptr, qname_ptr, NULL, options) < 0)
    {
        free(qname_ptr);
        throw_lsf_exception(env, "Call to lsb_openjobinfo failed",
                            lsb_sysmsg());
        return jobInfoList;
    }

    for (;;)
    {
        job = lsb_readjobinfo(&more);

        if (job == NULL)
        {
            free(qname_ptr);
            throw_lsf_exception(env, "Call to lsb_readjobinfo failed",
                                lsb_sysmsg());
            return jobInfoList;
        }

        jobInfo = create_job_info(env, job);
        (*env)->CallBooleanMethod(env, jobInfoList, alAdd, jobInfo);

        if (! more)
            break;
    }

    lsb_closejobinfo();
    /* End LSF calls */

    free(qname_ptr);

    return jobInfoList;
}

/*
 * Returns java.util.ArrayList of
 * uk.ac.sanger.jcon.lsf.LSFJobInfo objects describing jobs
 * on queue specified by the name jstring queue belonging to the user
 * named by jstring user.
 */
JNIEXPORT jobject JNICALL
Java_uk_ac_sanger_jcon_lsf_LSFBatchInfo_getJobInfo__Ljava_lang_String_2Ljava_lang_String_2
(JNIEnv *env, jobject this, jstring username, jstring queue)
{
    /* For LSF call */
    struct jobInfoEnt *job;
    int options = ALL_JOB;

    char *username_ptr = NULL;
    char *qname_ptr = NULL;

    jobject jobInfo;
    jobject jobInfoList;
    int more;

    /* Initial capacity of ArrayList */
    initial_cap = 20;

    get_classes_and_methods(env);

    username_ptr = dynamic_jstring_chars(env, stLength,
                                         username, username_ptr);
    qname_ptr = dynamic_jstring_chars(env, stLength,
                                      queue, qname_ptr);

    jobInfoList = (*env)->NewObject(env, alClazz, alCon, initial_cap);

    /* Start LSF calls */
    if (lsb_init(NULL))
    {
        free(username_ptr);
        free(qname_ptr);
        throw_lsf_exception(env, "Call to lsb_init failed",
                            lsb_sysmsg());
        return jobInfoList;
    }

    if (lsb_openjobinfo(0, NULL, username_ptr, qname_ptr, NULL, options) < 0)
    {
        free(username_ptr);
        free(qname_ptr);
        throw_lsf_exception(env, "Call to lsb_openjobinfo failed",
                            lsb_sysmsg());
        return jobInfoList;
    }

    for (;;)
    {
        job = lsb_readjobinfo(&more);

        if (job == NULL)
        {
            free(username_ptr);
            free(qname_ptr);
            throw_lsf_exception(env, "Call to lsb_readjobinfo failed",
                                lsb_sysmsg());
            return jobInfoList;
        }

        jobInfo = create_job_info(env, job);
        (*env)->CallBooleanMethod(env, jobInfoList, alAdd, jobInfo);

        if (! more)
            break;
    }

    lsb_closejobinfo();
    /* End LSF calls */

    free(username_ptr);
    free(qname_ptr);

    return jobInfoList;
}

/*
 * Returns java.util.ArrayList of
 * uk.ac.sanger.jcon.lsf.LSFJobInfo objects describing jobs
 * specified by the by the array of jstring job names.
 */
JNIEXPORT jobject JNICALL
Java_uk_ac_sanger_jcon_lsf_LSFBatchInfo_getNamedJobInfo
(JNIEnv *env, jobject this, jobjectArray jobNames)
{
    /* For LSF call */
    struct jobInfoEnt *job;
    int options = ALL_JOB;

    char *username_ptr = "all";
    char *jname_ptr = NULL;

    jint jobNameCount;
    jstring jobName;
    jobject jobInfo;
    jobject jobInfoList;
    int more;
    int i;

    /* LSF variables */
    options = ALL_JOB;

   /* Initial capacity of ArrayList */
    initial_cap = 20;

    get_classes_and_methods(env);
    jobInfoList = (*env)->NewObject(env, alClazz, alCon, initial_cap);

    /* Start LSF calls */
    if (lsb_init(NULL))
    {
        throw_lsf_exception(env, "Call to lsb_init failed",
                            lsb_sysmsg());
        return jobInfoList;
    }

    jobNameCount = (*env)->GetArrayLength(env, jobNames);

    for (i = 0; i < jobNameCount; i++)
    {
        jobName = (jstring) (*env)->GetObjectArrayElement(env, jobNames, i);

        jname_ptr = dynamic_jstring_chars(env, stLength,
                                          jobName, jname_ptr);
        (*env)->DeleteLocalRef(env, jobName);

        if (lsb_openjobinfo(0, jname_ptr, username_ptr,
                            NULL, NULL, options) < 0)
        {
            free(jname_ptr);
            throw_lsf_exception(env, "Call to lsb_openjobinfo failed",
                                lsb_sysmsg());
            return jobInfoList;
        }
        
        for (;;)
        {
            job = lsb_readjobinfo(&more);

            if (job == NULL)
            {
                free(jname_ptr);
                throw_lsf_exception(env, "Call to lsb_readjobinfo failed",
                                    lsb_sysmsg());
                return jobInfoList;
            }

            jobInfo = create_job_info(env, job);
            (*env)->CallBooleanMethod(env, jobInfoList, alAdd, jobInfo);

            if (! more)
                break;
        }

        lsb_closejobinfo();
        free(jname_ptr);
    }
    /* End LSF calls */

    return jobInfoList;
}

/*
 * Sets up the required classes and methods.
 */
void get_classes_and_methods(JNIEnv *env)
{
    alClazz = (*env)->FindClass(env, "java.util.ArrayList");
    alCon   = (*env)->GetMethodID(env, alClazz, "<init>", "(I)V");
    alAdd   = (*env)->GetMethodID(env, alClazz, "add",
                                  "(Ljava/lang/Object;)Z");

    stClazz = (*env)->FindClass(env, "java.lang.String");
    stCon   = (*env)->GetMethodID(env, stClazz, "<init>", "()V");
    stLength = (*env)->GetMethodID(env, stClazz, "length",
                                   "()I");

    jiClazz = (*env)->FindClass(env, "uk.ac.sanger.jcon.lsf.LSFJobInfo");
    jiCon   = (*env)->GetMethodID(env, jiClazz, "<init>",
                                  "(ILjava/lang/String;IILjava/lang/String;[Ljava/lang/String;)V");
}

/*
 * Adds a new LSFJobInfo object, created from the job struct, to the
 * ArrayList.
 */
jobject create_job_info(JNIEnv *env, struct jobInfoEnt *job)
{
    jint    jobId;
    jstring user;
    jstring fromHost;
    jobjectArray exHostsArr;
    jstring exHost;

    jint status  = -1;
    jint exitVal = -1;

    int numExHosts;
    int i;

    jobId    = job->jobId;
    user     = (*env)->NewStringUTF(env, job->user);
    status   = job->status;
    exitVal  = job->exitStatus;
    fromHost = (*env)->NewStringUTF(env, job->fromHost);

    numExHosts = job->numExHosts;
    exHostsArr = (*env)->NewObjectArray(env, numExHosts, stClazz, NULL);

    for (i = 0; i < numExHosts; i++)
    {
        exHost = (*env)->NewStringUTF(env, (job->exHosts)[i]);
        (*env)->SetObjectArrayElement(env, exHostsArr, i, exHost);
    }

    return (*env)->NewObject(env, jiClazz, jiCon,
                             jobId, user, status, exitVal,
                             fromHost, exHostsArr);
}
