/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <lsf/lsbatch.h>
#include "lsf_util.h"
#include "uk_ac_sanger_jcon_lsf_LSFBatchSubmit.h"

/* Pointers for jstring chars */
struct myPointers
{
    char *jname_ptr;
    char *qname_ptr;
    char *command_ptr;
    char *infile_ptr;
    char *outfile_ptr;
    char *errfile_ptr;
};

/* Functions local */
void freePointers(struct myPointers *mptrs);

/*
 * Class:     uk_ac_sanger_jcon_lsf_LSFBatchSubmit
 * Method:    submit
 * Signature: (Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;
 *             Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL
Java_uk_ac_sanger_jcon_lsf_LSFBatchSubmit_submit
(JNIEnv *env, jobject this, jstring jobName, jstring queue, jstring inFile,
 jstring outFile, jstring errFile, jstring command)
{
    struct submit      request;
    struct submitReply reply;
    struct myPointers  mptrs = { NULL, NULL, NULL, NULL, NULL, NULL };

    /* java.lang.String */
    jclass    stClazz;
    jmethodID stLength;

    int jobId;
    int options; /* Options for job request */
    int dup_fd;  /* Holder for saved stdout file descriptor */
    int i;

    memset(&request, 0, sizeof(struct submit));
    memset(&reply,   0, sizeof(struct submitReply));

    stClazz = (*env)->FindClass(env, "java.lang.String");
    stLength = (*env)->GetMethodID(env, stClazz, "length",
                                   "()I");

    jobId = -1;

    if (jobName == NULL)
    {
        throw_lsf_exception(env, "Job name was null", "");
        return -1;
    }
    if (queue == NULL)
    {
        throw_lsf_exception(env, "Queue was null", "");
        return -1;
    }
    if (command == NULL)
    {
        throw_lsf_exception(env, "Command was null", "");
        return -1;
    }

    request.jobName = dynamic_jstring_chars(env, stLength,
                                            jobName, mptrs.jname_ptr);
    request.queue = dynamic_jstring_chars(env, stLength,
                                          queue, mptrs.qname_ptr);
    request.command = dynamic_jstring_chars(env, stLength,
                                            command, mptrs.command_ptr);

    options = 0 | SUB_JOB_NAME | SUB_QUEUE;

    if (inFile != NULL)
    {
        options |= SUB_IN_FILE;
        request.inFile = dynamic_jstring_chars(env, stLength,
                                               inFile, mptrs.infile_ptr);
    }

    if (outFile != NULL)
    {
        options |= SUB_OUT_FILE;
        request.outFile = dynamic_jstring_chars(env, stLength,
                                                outFile, mptrs.outfile_ptr);
    }

    if (errFile != NULL)
    {
        options |= SUB_ERR_FILE;
        request.errFile = dynamic_jstring_chars(env, stLength,
                                                errFile, mptrs.errfile_ptr);
    }

    if (lsb_init(NULL))
    {
        freePointers(&mptrs);
        throw_lsf_exception(env, "Call to lsb_init failed",
                            lsb_sysmsg());
        return -1;
    }

    request.options  = options;
    request.options2 = 0;
    request.resReq   = NULL;

    for (i = 0; i < LSF_RLIM_NLIMITS; i++)
        request.rLimits[i] = DEFAULT_RLIMIT;
    
    request.hostSpec         = NULL;
    request.numProcessors    = 1;
    request.maxNumProcessors = 1;
    request.beginTime        = 0;
    request.termTime         = 0;
    request.nxf              = 0;
    request.delOptions       = 0;

    /*
     * It is necessary to dup and close STDOUT to protect the call to
     * lsb_submit. The lsb_submit function always writes to STDOUT and
     * this crashes the Java VM unless action is taken to prevent
     * it.
     */
    if ((dup_fd = dup(STDOUT_FILENO)) < 0)
    {
        freePointers(&mptrs);
        throw_jni_exception(env, "Failed to dup STDOUT");
        return dup_fd;
    }

    if (close(STDOUT_FILENO) < 0)
    {
        freePointers(&mptrs);
        throw_jni_exception(env, "Failed to close STDOUT");
        return -1;
    }

    /* Call protected by dup */
    jobId = lsb_submit(&request, &reply);

    if (dup2(dup_fd, STDOUT_FILENO) < 0)
    {
        freePointers(&mptrs);
        throw_jni_exception(env, "Failed to retrieve saved STDOUT");
        return -1;
    }

    if (close(dup_fd) < 0)
    {
        freePointers(&mptrs);
        throw_jni_exception(env, "Failed to close saved STDOUT");
        return -1;
    }

    if (jobId < 0)
    {
        freePointers(&mptrs);
        throw_lsf_exception(env, "Failed to submit job", lsb_sysmsg());
        return -1;
    }

    freePointers(&mptrs);

    return jobId;
}

void freePointers(struct myPointers *mptrs)
{
    if (mptrs->jname_ptr)   free(mptrs->jname_ptr);
    if (mptrs->qname_ptr)   free(mptrs->qname_ptr);
    if (mptrs->command_ptr) free(mptrs->command_ptr);
    if (mptrs->infile_ptr)  free(mptrs->infile_ptr);
    if (mptrs->outfile_ptr) free(mptrs->outfile_ptr);
    if (mptrs->errfile_ptr) free(mptrs->errfile_ptr);
}




