/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

#include <jni.h>
#include <stdlib.h>
#include <string.h>

/**********************************************************************
 *      Name:  copy_jstring_chars
 *   Purpose:  Copies const char array in the jstring to pointer
 *
 * Arguments:  env -- the JNI struct
 *             string -- the jstring to copy from
 *             char_ptr -- pointer to array to copy into
 *
 *   Returns:  void
 **********************************************************************/
void copy_jstring_chars(JNIEnv *env, jstring string, char *char_ptr)
{
    jboolean isCopy;
    const char *jschar_ptr;

    jschar_ptr = (*env)->GetStringUTFChars(env, string, &isCopy);
    strcpy(char_ptr, jschar_ptr);
    if (isCopy == JNI_TRUE)
    {
        (*env)->ReleaseStringUTFChars(env, string, jschar_ptr);
    }
}

/**********************************************************************
 *      Name:  throw_lsf_exception
 *   Purpose:  Throws an exception containing a message and the LSF
 *             error message
 *
 * Arguments:  env -- the JNI struct
 *             message_ptr -- pointer to the main exception message
 *             lsf_message_ptr -- pointer to the LSF error message
 *
 *   Returns:  void
 **********************************************************************/
void throw_lsf_exception(JNIEnv *env, char *message_ptr, char *lsf_message_ptr)
{
    jstring message;
    jstring lsf_message;

    jclass    leClazz;
    jmethodID leCon;
    jthrowable lsfException;

    leClazz = (*env)->FindClass(env, "uk.ac.sanger.jcon.lsf.LSFBatchException");
    leCon   = (*env)->GetMethodID(env, leClazz, "<init>",
                                  "(Ljava/lang/String;Ljava/lang/String;)V");

    message = (*env)->NewStringUTF(env, message_ptr);
    lsf_message = (*env)->NewStringUTF(env, lsf_message_ptr);

    lsfException = (jthrowable) (*env)->NewObject(env, leClazz, leCon,
                                                  message,lsf_message);

    (*env)->Throw(env, lsfException);
}

/**********************************************************************
 *      Name:  throw_jni_exception
 *   Purpose:  Throws an exception containing a message indicating a 
 *             serious error in the JNI layer
 *
 * Arguments:  env -- the JNI struct
 *             message_ptr -- pointer to the exception message
 *
 *   Returns:  void
 **********************************************************************/
void throw_jni_exception(JNIEnv *env, char *message_ptr)
{
    jstring message;

    jclass    jeClazz;
    jmethodID jeCon;
    jthrowable jniException;

    jeClazz = (*env)->FindClass(env, "uk.ac.sanger.psu.jobcontrol.lsf.JNIException");
    jeCon   = (*env)->GetMethodID(env, jeClazz, "<init>",
                                  "(Ljava/lang/String;)V");

    message = (*env)->NewStringUTF(env, message_ptr);
    jniException = (jthrowable) (*env)->NewObject(env, jeClazz, jeCon,
                                                  message);

    (*env)->Throw(env, jniException);
}

/**********************************************************************
 *      Name:  dynamic_jstring_chars
 *   Purpose:  Will allocate sufficient memory to hold a copy of the
 *             chars in jstring and then makes the copy to the
 *             supplied pointer
 *
 * Arguments:  env -- the JNI struct
 *             stLength -- method ID for length() method of String
 *             string -- the jstring
 *             char_ptr -- pointer to allocate
 *
 *   Returns:  pointer to copy of jstring chars
 **********************************************************************/
char * dynamic_jstring_chars(JNIEnv *env, jmethodID stLength,
                             jstring string, char *char_ptr)
{
    int length = (*env)->CallIntMethod(env, string, stLength);

    char_ptr = (char *) malloc((length + 1) * sizeof(char));

    if (char_ptr == NULL)
    {
        throw_jni_exception(env, "Malloc for jstring chars failed");
        return char_ptr;
    }

    copy_jstring_chars(env, string, char_ptr);

    return char_ptr;
}
