
-- EXECUTABLE represents a specific program or script e.g. blastall

DROP SEQUENCE exec_id_seq;
CREATE SEQUENCE exec_id_seq;

DROP TABLE executable CASCADE CONSTRAINTS;
CREATE TABLE executable
(
        exec_id         integer         CONSTRAINT executable_pkey PRIMARY KEY,
        name            varchar(64)     NOT NULL,
        version         varchar(32)     ,
        description     varchar(128)    NOT NULL
);

-- STATUS represents the state of a Job. Values can be WAITING, READY,
-- STARTED, COMPLETED, FAILED, SKIPPED or CANCELLED

DROP SEQUENCE status_id_seq;
CREATE SEQUENCE status_id_seq;

DROP TABLE status CASCADE CONSTRAINTS;
CREATE TABLE status
(
        status_id       integer         CONSTRAINT status_pkey PRIMARY KEY,
        description     varchar(12)     NOT NULL
);

-- OWNER represents the owner and point of contact for a Task

DROP SEQUENCE owner_id_seq;
CREATE SEQUENCE owner_id_seq;

DROP TABLE owner CASCADE CONSTRAINTS;
CREATE TABLE owner
(
        owner_id        integer         CONSTRAINT owner_pkey PRIMARY KEY,
        user_name       varchar(8)      NOT NULL,
        first_name      varchar(32)     NOT NULL,
        last_name       varchar(32)     NOT NULL,
        email           varchar(32)     NOT NULL,
        telephone       varchar(32)     NOT NULL
);

-- TASK represents a collection of Jobs

DROP SEQUENCE task_id_seq;
CREATE SEQUENCE task_id_seq;

DROP TABLE task CASCADE CONSTRAINTS;
CREATE TABLE task
(
        task_id         integer         CONSTRAINT task_pkey PRIMARY KEY,
        name            varchar(32)     NOT NULL,
        description     varchar(255)	NOT NULL,
        owner_id        integer         NOT NULL
                                        CONSTRAINT task_owner_fkey
                                        REFERENCES owner,
        priority        integer         NOT NULL
                                        CONSTRAINT priority_chk
                                        CHECK (priority BETWEEN 1 AND 10),
        submitted       integer         NOT NULL
                                        CONSTRAINT submitted_chk
                                        CHECK (submitted = 0 OR
                                               submitted = 1),
        submission_time date
);

DROP INDEX task_submitted_idx;
CREATE BITMAP INDEX task_submitted_idx ON task(submitted);

-- JOB represents an atomic stage in an analysis. For example, running
-- a search of one sequence against a database. Parsing the search output
-- would be a second Job dependent on the first

DROP SEQUENCE job_id_seq;
CREATE SEQUENCE job_id_seq;

DROP TABLE job CASCADE CONSTRAINTS;
CREATE TABLE job
(
        job_id          integer         CONSTRAINT job_pkey PRIMARY KEY,
        task_id         integer         NOT NULL,
                                        --
                                        -- Put this constraint back in when
                                        -- we move to Oracle 9. Temporarily
                                        -- removed for createJob speed hack
                                        --
                                        -- CONSTRAINT job_task_fkey
                                        -- REFERENCES task,
                                        --
        parent_id       integer         NOT NULL
                                        CONSTRAINT job_parent_chk
                                        CHECK (parent_id >= 0),
        low             integer         NOT NULL
                                        CONSTRAINT low_chk
                                        CHECK (low >= 0),
        high            integer         NOT NULL
                                        CONSTRAINT high_chk
                                        CHECK (high >= 1),
        operator        varchar(3)      NOT NULL
                                        CONSTRAINT operator_chk
                                        CHECK (operator IN ('AND', 'OR')),
        exec_id         integer         NOT NULL
                                        CONSTRAINT job_exec_fkey
                                        REFERENCES executable,
        status_id       integer         NOT NULL
                                        CONSTRAINT job_status_fkey
                                        REFERENCES status,
        batch_id        integer         NOT NULL,
        queue           varchar(12)     ,
        start_time      date            ,
        completion_time date            ,
        input_name      varchar(255)    ,
        output_name     varchar(255)    ,
        error_name      varchar(255)    ,
        cmd_templ       varchar(255)    NOT NULL,
        work_dir_name   varchar(255)    ,
        environment     varchar(255)    ,
        max_retry_count integer         NOT NULL
                                        CONSTRAINT max_retry_count_chk
                                        CHECK (max_retry_count >= 0),
        retry_count     integer         NOT NULL
                                        CONSTRAINT retry_count_chk
                                        CHECK (retry_count >= 0),
        exit_value      integer
);

-- status_id has low cardinality, but is subject to some change
DROP INDEX job_status_idx;
CREATE BITMAP INDEX job_status_idx ON job(status_id);

DROP INDEX job_task_idx;
CREATE INDEX job_task_idx ON job(task_id);

DROP INDEX job_parent_idx;
CREATE INDEX job_parent_idx ON job(parent_id);

DROP INDEX job_low_idx;
CREATE INDEX job_low_idx ON job(low);


-- JOB_HISTORY is a log of changes in Job state (not used)

DROP SEQUENCE history_id_seq;
CREATE SEQUENCE history_id_seq;

DROP TABLE job_history CASCADE CONSTRAINTS;
CREATE TABLE job_history
(
        history_id      integer         CONSTRAINT history_pkey PRIMARY KEY,
        job_id          integer         NOT NULL
                                        CONSTRAINT history_job_fkey
                                        REFERENCES job,
        status_id       integer         NOT NULL
                                        CONSTRAINT history_status_fkey
                                        REFERENCES status,
        event_time      date            NOT NULL,
        description     varchar(255)
);

-- HOST is a compute host where Jobs are executed. Some systems, such as
-- LSF allow a single Job to be run on > 1 host, so an additional job_host
-- table is used

DROP SEQUENCE host_id_seq;
CREATE SEQUENCE host_id_seq;

DROP TABLE host CASCADE CONSTRAINTS;
CREATE TABLE host
(
        host_id         integer         CONSTRAINT host_pkey PRIMARY KEY,
        name            varchar(12)     NOT NULL
);

-- JOB_HOST relates Jobs and hosts

DROP TABLE job_host CASCADE CONSTRAINTS;
CREATE TABLE job_host
(
        job_id          integer         NOT NULL
                                        CONSTRAINT job_host_fkey
                                        REFERENCES job,
        host_id         integer         NOT NULL
                                        CONSTRAINT host_job_fkey
                                        REFERENCES host
);

DROP INDEX job_host_idx;
CREATE INDEX job_host_idx ON job_host(job_id, host_id);

-- JOB_HOST_VW is a view which allows direct association of a Job to a
-- host name

DROP VIEW job_host_vw;
CREATE VIEW job_host_vw AS
    SELECT job.job_id, host.name
    FROM job, host, job_host
    WHERE
        job_host.job_id = job.job_id
        AND job_host.host_id = host.host_id;


-- TASK_SUBMITTED_VW is a view which lists Tasks which have been
-- submitted, but are not completed

DROP VIEW task_submitted_vw;
-- CREATE VIEW task_submitted_vw AS
--     SELECT DISTINCT task.task_id, task.owner_id, task.submission_time
--     FROM job, task
--     WHERE
--         task.task_id = job.task_id
--         AND task.submitted = 1
--         AND job.parent_id = 0
--         AND job.status_id = 1
--          OR job.status_id = 2
--          OR job.status_id = 3;

-- The following has roughly half the cost of the above because it
-- allows a nested loops join instead of sort merge join with full
-- table scans of job and task

CREATE VIEW task_submitted_vw AS
    SELECT DISTINCT task.task_id, task.owner_id, task.submission_time
    FROM job, task
    WHERE
        task.task_id = job.task_id
        AND task.submitted = 1
        AND job.parent_id = 0
        AND job.status_id = 1
    UNION ALL
    SELECT DISTINCT task.task_id, task.owner_id, task.submission_time
    FROM job, task
    WHERE task.task_id = job.task_id
        AND task.submitted = 1
        AND job.parent_id = 0
        AND job.status_id = 2
    UNION ALL
    SELECT DISTINCT task.task_id, task.owner_id, task.submission_time
    FROM job, task
    WHERE task.task_id = job.task_id
        AND task.submitted = 1
        AND job.parent_id = 0
        AND job.status_id = 3;

-- TASK_BREAKDOWN_VW is a view which gives a breakdown of total
-- numbers of Jobs with each status_id for each Task

DROP VIEW task_breakdown_vw;
CREATE VIEW task_breakdown_vw AS
    SELECT job.task_id, job.status_id AS status_id,
           count(job.status_id) AS num_jobs
    FROM job
    GROUP BY job.task_id, job.status_id;

-- JOB_PARENT_STATUS_VW is a view which extracts parent-child
-- associations and status from the nested set data in the job
-- table. Creates a parent list. For experimentation only. Do not use
-- because it is too slow.

DROP VIEW job_parent_status_vw;
CREATE VIEW job_parent_status_vw AS
    SELECT child.job_id AS job_id,
           child.task_id AS task_id,
           parent.job_id AS parent_id,
           child.status_id AS status_id
    FROM job child, job parent
    WHERE
        child.task_id = parent.task_id
        AND child.low BETWEEN parent.low AND parent.high
        AND child.low != parent.low
        AND NOT EXISTS
            (SELECT * FROM job inter
                WHERE
                    inter.task_id = child.task_id
                    AND inter.low BETWEEN parent.low + 1 AND parent.high - 1
                    AND child.low BETWEEN inter.low + 1 AND inter.high - 1);
