
-- EXECUTABLE represents a specific program or script e.g. blastall

CREATE SEQUENCE exec_id_seq;

CREATE TABLE executable
(
        exec_id         integer         CONSTRAINT executable_pkey UNIQUE PRIMARY KEY,
        name            varchar(64)     NOT NULL,
        version         varchar(32)     ,
        description     varchar(128)    NOT NULL
);

-- STATUS represents the state of a Job. Values can be WAITING, READY,
-- STARTED, COMPLETED, FAILED, SKIPPED or CANCELLED

CREATE SEQUENCE status_id_seq;

CREATE TABLE status
(
        status_id       integer         CONSTRAINT status_pkey UNIQUE PRIMARY KEY,
        description     varchar(12)     NOT NULL
);

-- OWNER represents the owner and point of contact for a Task

CREATE SEQUENCE owner_id_seq;

CREATE TABLE owner
(
        owner_id        integer         CONSTRAINT owner_pkey UNIQUE PRIMARY KEY,
        user_name       varchar(12)     NOT NULL,
        first_name      varchar(32)     NOT NULL,
        last_name       varchar(32)     NOT NULL,
        email           varchar(32)     NOT NULL,
        telephone       varchar(32)     NOT NULL
);

-- TASK represents a collection of Jobs

CREATE SEQUENCE task_id_seq;

CREATE TABLE task
(
        task_id         integer         CONSTRAINT task_pkey UNIQUE PRIMARY KEY,
        name            varchar(32)     NOT NULL,
        description     varchar(255)	NOT NULL,
        owner_id        integer         NOT NULL
                                        CONSTRAINT task_owner_fkey
                                        REFERENCES owner,
        priority        integer         NOT NULL,
        submission_time timestamp
);

-- JOB represents an atomic stage in an analysis. For example, running
-- a search of one sequence against a database. Parsing the search output
-- would be a second Job dependent on the first

CREATE SEQUENCE job_id_seq;

CREATE TABLE job
(
        job_id          integer         CONSTRAINT job_pkey UNIQUE PRIMARY KEY,
        task_id         integer         NOT NULL
                                        CONSTRAINT job_task_fkey
                                        REFERENCES task,
        parent_id       integer         CONSTRAINT job_parent_fkey
                                        REFERENCES job,
        low             integer         NOT NULL,
        high            integer         NOT NULL,
        operator        varchar(3)      NOT NULL
                                        CONSTRAINT operator_chk
                                        CHECK (operator IN ('AND', 'OR')),
        exec_id         integer         NOT NULL
                                        CONSTRAINT job_exec_fkey
                                        REFERENCES executable,
        status_id       integer         NOT NULL
                                        CONSTRAINT job_status_fkey
                                        REFERENCES status,
        batch_id        integer         NOT NULL,
        queue           varchar(12)     ,
        start_time      timestamp       ,
        completion_time timestamp       ,
        input_name      varchar(64)     ,
        output_name     varchar(64)     ,
        error_name      varchar(64)     ,
        cmd_templ       varchar(255)    NOT NULL,
        work_dir_name   varchar(255)    NOT NULL,
        environment     varchar(255)    ,
        max_retry_count integer         NOT NULL
                                        CONSTRAINT max_retry_count_chk
                                        CHECK (max_retry_count >= 0),
        retry_count     integer         NOT NULL
                                        CONSTRAINT retry_count_chk
                                        CHECK (retry_count >= 0),
        exit_value      integer
);

-- DROP INDEX job_task_idx;
CREATE INDEX job_task_idx ON job(task_id);
-- DROP INDEX job_low_idx;
CREATE INDEX job_low_idx ON job(low);

-- JOB_HISTORY is a log of changes in Job state

CREATE SEQUENCE history_id_seq;

CREATE TABLE job_history
(
        history_id      integer         CONSTRAINT history_pkey UNIQUE PRIMARY KEY,
        job_id          integer         NOT NULL
                                        CONSTRAINT history_job_fkey
                                        REFERENCES job,
        status_id       integer         NOT NULL
                                        CONSTRAINT history_status_fkey
                                        REFERENCES status,
        event_time      timestamp       NOT NULL,
        description     varchar(255)
);

-- HOST is a compute host where Jobs are executed. Some systems, such as
-- LSF allow a single Job to be run on > 1 host, so an additional job_host
-- table is used

CREATE SEQUENCE host_id_seq;

CREATE TABLE host
(
        host_id         integer         CONSTRAINT host_pkey UNIQUE PRIMARY KEY,
        name            varchar(12)     NOT NULL
);

-- JOB_HOST relates Jobs and hosts

CREATE TABLE job_host
(
        job_id          integer         NOT NULL
                                        CONSTRAINT job_host_fkey
                                        REFERENCES job,
        host_id         integer         NOT NULL
                                        CONSTRAINT host_job_fkey
                                        REFERENCES host
);

CREATE INDEX job_host_idx ON job_host(job_id, host_id);

-- JOB_HOST_VW is a view which allows direct association of a Job to a
-- host name

DROP VIEW job_host_vw;
CREATE VIEW job_host_vw AS
    SELECT job.job_id, host.name
    FROM job, host, job_host
    WHERE
        job_host.job_id = job.job_id
        AND job_host.host_id = host.host_id;

-- TASK_SUBMITTED_VW is a view which lists Tasks which have been
-- submitted, but are not completed

DROP VIEW task_submitted_vw;
CREATE VIEW task_submitted_vw AS
    SELECT DISTINCT task.task_id, task.owner_id, task.submission_time
    FROM task, job
    WHERE
        task.task_id = job.task_id
        AND task.submission_time IS NOT NULL
        AND job.parent_id IS NULL 
        AND job.status_id < 4;

-- TASK_BREAKDOWN_VW is a view which gives a breakdown of total
-- numbers of Jobs with each status_id for each Task

DROP VIEW task_breakdown_vw;
CREATE VIEW task_breakdown_vw AS
    SELECT job.task_id, job.status_id AS status_id,
           count(job.status_id) AS num_jobs
    FROM job
    GROUP BY job.task_id, job.status_id;
