/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon;

import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;

import uk.ac.sanger.jcon.protocol.*;
import uk.ac.sanger.util.CommandLineDoc;

/**
 * <p><code>JSubmit</code> is the front end for launching all task
 * related <code>Protocol</code>s. A <code>Protocol</code> is a way of
 * defining how to create one or more <code>Task</code>s in a way
 * which is more formal then writing a BeanShell script.</p>
 *
 * <p><code>JSubmit</code> is configured by means of a properties file
 * which contains keys which are <code>Protocol</code> aliases and
 * corresponding values which are fully qualified class names of
 * implementations of the <code>Protocol</code> interface. The alias
 * is specified by the --protocol command line argument and the
 * relevant <code>Protocol</code> implementation will be
 * instantiated. After processing with GNU getopt, remaining arguments
 * are passed to the <code>Protocol</code>'s <code>apply</code>
 * method.</p>
 *
 * @author Keith James
 */
public class JSubmit
{
    /**
     * <code>PROTOCOL_PROPS</code> the properties file containing the
     * protocol configuration
     */
    public static final String PROTOCOL_PROPS = "/protocol.properties";

    public static final String     HELP_KEY = "help";
    public static final String    OWNER_KEY = "owner";
    public static final String PROTOCOL_KEY = "protocol";
    public static final String   CONFIG_KEY = "config";
    public static final String    QUEUE_KEY = "queue";

	public static final int        GETOPT_CODE = 1;
	public static final int      LOCATION_CODE = 10;
	public static final int INSTANTIATION_CODE = 20;
	public static final int     EXECUTION_CODE = 30;

	private static Properties protocolProps = loadProperties();

	public static void main(String [] argv)
	{
        // Load documentation. This implementation should probably be
        // phased out (along with GNU getopt) and replaced with
        // Jakarta Commons CLI to fall in line with biojava
        // dependencies
		CommandLineDoc doc = new CommandLineDoc();
		doc.loadResources(JSubmit.class.getName());

        // Parse options
		LongOpt [] longopts = new LongOpt[5];
		longopts[0] = new LongOpt(CONFIG_KEY,
                                  LongOpt.REQUIRED_ARGUMENT, null, 'c');
		longopts[1] = new LongOpt(HELP_KEY,
                                  LongOpt.NO_ARGUMENT, null, 'h');
		longopts[2] = new LongOpt(OWNER_KEY,
                                  LongOpt.REQUIRED_ARGUMENT, null, 'o');
		longopts[3] = new LongOpt(PROTOCOL_KEY,
                                  LongOpt.REQUIRED_ARGUMENT, null, 'p');
        longopts[4] = new LongOpt(QUEUE_KEY,
                                  LongOpt.REQUIRED_ARGUMENT, null, 'q');

		Getopt g = new Getopt("JSubmit", argv, "c:ho:p:q:", longopts);

		String    owner = null;
		String protocol = null;
        String    queue = null;
        String fileName = null;
        Map config = new HashMap();

		int c = 0;
		while ((c = g.getopt()) != -1)
		{
			switch (c)
			{
                case 'c':
                    fileName = g.getOptarg();
                    protocol = "UserProtocol";
                    break;

				case 'h':
					doc.printShortUsage(System.err);
				    System.exit(0);
					break;

				case 'o':
					owner = g.getOptarg();
					break;

				case 'p':
					protocol = g.getOptarg();
                    config.put(PROTOCOL_KEY, protocol);
					break;

                case 'q':
                    queue = g.getOptarg();
                    config.put(QUEUE_KEY, queue);
                    break;

				default:
					System.err.println("Failed to recognise option '"
									   + c
									   + "'");
					doc.printLongUsage(System.err);
					System.exit(GETOPT_CODE);
			}
		}

        // Sanity checks
		if (owner == null)
		{
			System.err.println("Owner was not defined");
			doc.printLongUsage(System.err);
			System.exit(GETOPT_CODE);
		}

		if ((fileName == null) && (protocol == null))
		{
			System.err.println("Protocol was not defined");
			doc.printLongUsage(System.err);
			System.exit(GETOPT_CODE);
		}

		if ((fileName != null) && (! protocol.equals("UserProtocol")))
		{
			System.err.println("Both a user configuration and a Protocol were defined");
			doc.printLongUsage(System.err);
			System.exit(GETOPT_CODE);
		}

		if (queue == null)
		{
			System.err.println("Batch queue was not defined");
			doc.printLongUsage(System.err);
			System.exit(GETOPT_CODE);
		}

        // Instantiate the required Protocol impl
		String className = protocolProps.getProperty(protocol);

		if (className == null)
		{
			System.err.println("Unknown protocol. Known protocols are:");

			Enumeration enum = protocolProps.propertyNames();
			while (enum.hasMoreElements())
			{
				System.err.println(enum.nextElement());
			}

			System.exit(INSTANTIATION_CODE);
		}

		Protocol p = null;
		try
		{
			p = (Protocol) Class.forName(className).newInstance();
		}
		catch (ClassNotFoundException cnfe)
		{
			System.err.println("Failed to find protocol");
			cnfe.printStackTrace();
			System.exit(LOCATION_CODE);
		}
		catch (Exception e)
		{
			System.err.println("Failed to create protocol");
			e.printStackTrace();
			System.exit(INSTANTIATION_CODE);
		}

        // Remaining args are file names
		int protocolArgCount = argv.length - g.getOptind();
		String [] protocolArgs = new String [protocolArgCount];
		System.arraycopy(argv, g.getOptind(),
						 protocolArgs, 0, protocolArgCount);

        // Populate the config in case the Protocol wants these values
        config.put(CONFIG_KEY, fileName);
        config.put(QUEUE_KEY, queue);

        // Submit the Protocol
		try
		{
			p.apply(owner, config, protocolArgs);
		}
		catch (Exception e)
		{
			System.err.println("Failed to execute protocol");
			e.printStackTrace();
			System.exit(EXECUTION_CODE);
		}

		System.exit(0);
	}

	private static Properties loadProperties()
	{
		Properties props = new Properties();

        InputStream is = null;

        try
        {
            is = JSubmit.class.getResourceAsStream(PROTOCOL_PROPS);

            if (is == null)
                throw new IOException("Failed to find resource '"
                                      + PROTOCOL_PROPS
                                      + "'");

            props.load(is);
        }
        catch (IOException ioe1)
        {
            System.err.println("Failed to load resource '"
                               + PROTOCOL_PROPS
                               + "'");
            try
            {
                is.close();
                throw ioe1;
            }
            catch (IOException ioe2)
            {
                System.err.println("Failed to close '"
                                   + PROTOCOL_PROPS
                                   + "'");
            }
        }

		return props;
	}
}
