/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon;

import java.sql.SQLException;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.dao.*;
import uk.ac.sanger.jcon.manager.Administrator;
import uk.ac.sanger.jcon.manager.AdministratorDefaultImpl;
import uk.ac.sanger.jcon.manager.TaskManager;
import uk.ac.sanger.jcon.manager.TaskManagerDefaultImpl;

/**
 * <code>JobControl</code> is the bootstrap object for the system. It
 * contains public methods which return the utility objects
 * <code>Administrator</code> which performs system management and
 * <code>TaskManager</code>, which is used to create and submit new
 * <code>Task</code>s.
 *
 * @author Keith James
 */
public final class JobControl
{
    // Log4j setup
    static Category cat = Category.getInstance(JobControl.class.getName());

    static ResourceBundle bundle =
        ResourceBundle.getBundle(JobControl.class.getName());

    static final String JNDI_DATASOURCE_KEY = "jobcontrol.jndi.datasource";

	private DataSource datasource;

    /**
     * Creates a new <code>JobControl</code> instance.
     */
    public JobControl() throws NamingException
    {
        String name = null;
        try
        {
            name = bundle.getString(JNDI_DATASOURCE_KEY);
        }
        catch (NullPointerException npe)
        {
            throw new Error("Malformed configuration ResourceBundle "
                            + "contains key '"
                            + JNDI_DATASOURCE_KEY
                            + "' missing a corresponding value");
        }

		Context context = new InitialContext();
		datasource = (DataSource) context.lookup(name);
    }

    /**
     * <code>getConfiguration</code> returns the current system
     * configuration.
     *
     * @return a <code>ResourceBundle</code>.
     */
    public static ResourceBundle getConfiguration()
    {
        return bundle;
    }

    /**
     * <code>createAdministrator</code> returns a new
     * <code>Administrator</code>.
     *
     * @return an <code>Administrator</code>.
     */
    public Administrator createAdministrator()
    {
        return new AdministratorDefaultImpl(new OwnerSQLDAO(datasource),
											new ExecutableSQLDAO(datasource),
											new StatusSQLDAO(datasource),
											new TaskSQLDAO(datasource));
    }

    /**
     * <code>createTaskManager</code> returns a new
     * <code>TaskManager</code>.
     *
     * @return a <code>TaskManager</code>.
     */
    public TaskManager createTaskManager()
    {
        return new TaskManagerDefaultImpl(new TaskSQLDAO(datasource),
										  new JobSQLDAO(datasource));
    }

    public ExecutableDAO createExecutableDAO()
    {
        return new ExecutableSQLDAO(datasource);
    }

    public JobDAO createJobDAO()
    {
		return new JobSQLDAO(datasource);
    }

    public OwnerDAO createOwnerDAO()
    {
        return new OwnerSQLDAO(datasource);
    }

    public StatusDAO createStatusDAO()
    {
        return new StatusSQLDAO(datasource);
    }

    public TaskDAO createTaskDAO()
    {
        return new TaskSQLDAO(datasource);
    }

	public HostNameDAO createHostNameDAO()
	{
		return new HostNameSQLDAO(datasource);
	}

    public TaskStatisticDAO createTaskStatisticDAO()
    {
        return new TaskStatisticSQLDAO(datasource);
    }
}

