/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon;

import java.io.File;
import java.util.Collection;
import java.util.Iterator;
import javax.naming.NamingException;

import gnu.getopt.Getopt;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.dao.JobDAO;
import uk.ac.sanger.jcon.dao.StatusDAO;
import uk.ac.sanger.jcon.dao.TaskDAO;
import uk.ac.sanger.jcon.dao.TaskStatisticDAO;
import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.run.*;
import uk.ac.sanger.util.CommandLineDoc;

/**
 * <p><code>JobControlApp</code> is a command line application which
 * starts up a scheduler to run tasks on the system. There should only
 * be one <code>JobControlApp</code> running.</p>
 *
 * <p>On starting this program checks the database for jobs which are
 * marked as running. These are considered to have been left by a
 * previous system failure and are set back to being ready to run so
 * that they may be restarted.</p>
 *
 * <p>This program launches a daemon which checks for the presence of
 * a specific file (usually a PID file) at regular intervals. If this
 * file is deleted a safe shutdown will be attempted. This involves
 * launching no new tasks and waiting for any currently running jobs
 * to complete before halting.</p>
 *
 * @author Keith James
 */
public class JobControlApp
{
    // Log4j setup
    static Category cat =
        Category.getInstance(JobControlApp.class.getName());

    private static CommandLineDoc docs;

    private JobControl jobControl;
    private File pidFile;

    private Coordinator coordinator;
    private TaskScheduler scheduler;

    // FIXME - command line options and help

    public static void main(String [] argv) throws Exception
    {
        cat.info ("Starting");
        docs = new CommandLineDoc();
        // docs.loadResources(JobControlApp.class.getName());

        JobControlApp jcApp = new JobControlApp("jobcontrol.pid");
        jcApp.startUpLSFSQL();
    }

    JobControlApp(String pidFileName) throws NamingException
    {
        pidFile = new File(pidFileName);
        jobControl = new JobControl();
    }

    private void startUpLSFSQL()
        throws Exception
    {
        JobDAO       jobDAO = jobControl.createJobDAO();
        StatusDAO statusDAO = jobControl.createStatusDAO();
        TaskDAO     taskDAO = jobControl.createTaskDAO();
        TaskStatisticDAO taskStatDAO = jobControl.createTaskStatisticDAO();

        coordinator =
            new CoordinatorLSFImpl(ProcessorLSFImpl.FACTORY,
                                   jobDAO,
                                   statusDAO);

        scheduler = new TaskSchedulerDefaultImpl(taskDAO, coordinator);

        scheduler.setSendMail (false);

        scheduler.setPolicy(new PolicyRoundRobinImpl(taskStatDAO));
        cleanUp(taskStatDAO, statusDAO, jobDAO);

        PIDFileWatcher watcher = new PIDFileWatcher(pidFile);
        watcher.start();

        scheduler.startScheduling();
        coordinator.startCoordinating();
    }

   /**
     * <code>cleanUp</code> is run before starting the
     * <code>Coordinator</code> to pick up any <code>Job</code>s
     * marked as STARTED. These will not have any listeners and result
     * from a system failure. These are not known to have completed
     * and are updated as READY before any other action takes
     * place. They will then be restarted. The maximum number which
     * can be in this state is equal to the maximum number which the
     * <code>Coordinator</code> can handle concurrently.
     *
     * @param taskStatDAO a <code>TaskStatisticDAO</code>.
     * @param statusDAO a <code>StatusDAO</code>.
     * @param jobDAO a <code>JobDAO</code>.
     */
    private void cleanUp(TaskStatisticDAO taskStatDAO,
                         StatusDAO        statusDAO,
                         JobDAO           jobDAO)
    {
        try
        {
            Status started = statusDAO.readStatusById(Status.STARTED);

            int [] submittedTaskIds = taskStatDAO.readSubmittedTaskId();

            cat.info("Found "
                     + submittedTaskIds.length
                     + " submitted Tasks on startup");

            for (int i = 0; i < submittedTaskIds.length; i++)
            {
                cat.info("Checking state of Task ["
                         + submittedTaskIds[i]
                         + "]");

                Collection startedJobs =
                    jobDAO.readJobsByStatusAndTaskId(started,
                                                     submittedTaskIds[i]);
                cat.info("Found "
                         + startedJobs.size()
                         + " started Jobs whose current state is unknown");

                for (Iterator ji = startedJobs.iterator(); ji.hasNext();)
                {
                    Job j = (Job) ji.next();
                    cat.warn("Cleaning up " + j);
                    jobDAO.updateJobReady(j);
                }
            }
        }
        catch (Exception e)
        {
            cat.error("Failed to clean up Jobs on restart", e);
        }
    }

    /**
     * <code>PIDFileWatcher</code> is a file watcher daemon. It will
     * halt both <code>TaskScheduler</code> and
     * <code>Coordinator</code> if the file is not found.
     */
    private class PIDFileWatcher implements Runnable
    {
        private Thread thread;
        private File pidFile;

        PIDFileWatcher(File pidFile)
        {
            this.pidFile = pidFile;

            thread = new Thread(this);
            thread.setDaemon(true);
        }

        public void start()
        {
            thread.start();
        }

        public void run()
        {
            while (checkPIDFile())
            {
                waitFor(60000);
            }
        }

        /**
         * <code>checkPIDFile</code> checks for the presence of the
         * <code>File</code>. If it is absent, the
         * <code>TaskScheduler</code> and <code>Coordinator</code> are
         * stopped.
         */
        private boolean checkPIDFile()
        {
            if (! (pidFile.exists() && pidFile.isFile()))
            {
                cat.info("PID file " + pidFile + " indicates halt required.");
                scheduler.stopScheduling();
                coordinator.stopCoordinating();
                return false;
            }

            return true;
        }

        private synchronized void waitFor(int milliSeconds)
        {
            try
            {
                cat.info("Waiting for " + milliSeconds + " ms");
                wait(milliSeconds);
            }
            catch (InterruptedException ie)
            {
                cat.error("Caught an Exception", ie);
            }
        }
    }
}
