/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.dao;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.ResourceBundle;
import javax.sql.DataSource;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.job.Executable;
import uk.ac.sanger.jcon.job.ExecutableDefaultImpl;

/**
 * <code>ExecutableSQLDAO</code> implements <code>ExecutableDAO</code>
 * for persistence tasks using a relational database. Canonical
 * instances of <code>Executable</code> are cached after the first
 * retrieval as the executable table is static.
 *
 * @author Keith James
 */
public class ExecutableSQLDAO extends SQLDAOBaseImpl implements ExecutableDAO
{
    // SQL statement resource
    static ResourceBundle sqlStatements =
            ResourceBundle.getBundle(ExecutableSQLDAO.class.getName());

    // Map to cache instances, keyed by ID
    private HashMap canonical = new HashMap(50);

    /**
     * Creates a new <code>ExecutableSQLDAO</code> with no
     * <code>DataSource</code>.
     */
    public ExecutableSQLDAO()
    {
        super();
    }

    /**
     * Creates a new <code>ExecutableSQLDAO</code> with the specified
     * <code>DataSource</code>.
     *
     * @param datasource a <code>DataSource</code>.
     */
    public ExecutableSQLDAO(DataSource datasource)
    {
        super(datasource);
    }

    public synchronized void createExecutable(Executable executable)
        throws Exception
    {
        String sql = sqlStatements.getString("createExecutable");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        boolean rollback = false;

        try
        {
            int newId = createNewExecutableId();

            setObjectIntField(executable, "id", newId);

            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, executable.getId());
            ps.setString(2, executable.getName());

            String version = executable.getVersion();
            if (version != null)
                ps.setString(3, version);
            else
                ps.setNull(3, Types.VARCHAR);

            ps.setString(4, executable.getDescription());
            cat.debug("Executing SQL: " + ps);
            ps.executeUpdate();
        }
        catch (Exception e)
        {
            cat.error("Caught an Exception; rolling back transaction", e);
            rollback = true;
            throw e;
        }
        finally
        {
            cSupport.close(ps);

            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    public synchronized Collection readAllExecutables()
        throws SQLException
    {
        String sql = sqlStatements.getString("readAllExecutables");

        Connection c = getManagedConnection();
        Statement s = null;
        ResultSet r = null;
        List executables = new ArrayList();

        try
        {
            cat.debug("Preparing SQL: " + sql);
            s = c.createStatement();
            cat.debug("Executing SQL: " + s);
            r = s.executeQuery(sql);

            while (r.next())
            {
                executables.add(populateExecutable(r));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(s);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return executables;
    }

    public synchronized Executable readExecutableById(int execId)
        throws SQLException, JobControlException
    {
        // Test canonical instances first
        Integer canonKey = new Integer(execId);
        if (canonical.containsKey(canonKey))
        {
            cat.debug("Fetching canonical Executable [" + canonKey + "] from cache");
            return (Executable) canonical.get(canonKey);
        }

        String sql = sqlStatements.getString("readExecutableById");

        Connection c = getManagedConnection();
        PreparedStatement  ps = null;
        ResultSet           r = null;
        Executable executable = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, execId);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            if (r.next())
            {
                executable = populateExecutable(r);
                // Add to canonical instances
                canonical.put(canonKey, executable);
            }
            else
            {
                throw new JobControlException("Requested a non-existent Executable with ID '"
                                              + execId
                                              + "'");
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return executable;
    }

    public synchronized Collection readExecutablesByName(String name)
        throws SQLException
    {
        String sql = sqlStatements.getString("readExecutablesByName");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        List executables = new ArrayList();

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setString(1, name);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            while (r.next())
            {
                executables.add(populateExecutable(r));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return executables;
    }

    public synchronized void deleteExecutable(Executable executable)
        throws SQLException
    {
        String sql = sqlStatements.getString("deleteExecutable");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        boolean rollback = false;

        int execId = executable.getId();

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, execId);
            cat.debug("Executing SQL: " + ps);
            ps.executeUpdate();
        }
        catch (SQLException se)
        {
            cat.error("Caught an Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        finally
        {
            cSupport.close(ps);

            if (rollback)
            {
                cSupport.rollback(c);
            }
            else
            {
                cSupport.commit(c);
                canonical.remove(new Integer(execId));
            }

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    /**
     * <code>populateExecutable</code> creates a new instance of
     * <code>Executable</code>.
     *
     * @param r a <code>ResultSet</code>.
     *
     * @return an <code>Executable</code>.
     *
     * @exception SQLException if an error occurs.
     * @exception NoSuchFieldException if an error occurs.
     * @exception IllegalAccessException if an error occurs.
     */
    private Executable populateExecutable(ResultSet r)
        throws SQLException
    {
        Executable executable = new ExecutableDefaultImpl(r.getString("name"),
                                                          r.getString("version"),
                                                          r.getString("description"));
        setObjectIntField(executable, "id", r.getInt("exec_id"));

        return executable;
    }

    /**
     * <code>createNewExecutableId</code> gets a new unique ID from a
     * sequence.
     *
     * @return an <code>int</code> ID.
     *
     * @exception Exception if an error occurs.
     */
    private int createNewExecutableId() throws Exception
    {
        String sql = sqlStatements.getString("createNewExecutableId");

        Connection c = getConnection();
        Statement s = null;
        ResultSet r = null;
        int id = 0;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            s = c.createStatement();
            cat.debug("Executing SQL: " + s);
            r = s.executeQuery(sql);
            if (r.next())
                id = r.getInt("exec_id");
            else
                throw new Exception("Failed to retrieve a new Executable ID");
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(s);
        }

        return id;
    }
}
