/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.dao;

import java.util.Collection;

/**
 * <code>HostNameDAO</code> describes data access objects for
 * persistence of valid host names. Possible stores are relational
 * databases or XML files. It's slightly overkill for these simple
 * <code>String</code>s, but we may use a full <code>Host</code> class
 * in future.
 *
 * @author Keith James
 */
public interface HostNameDAO
{
    /**
     * <code>createHostName</code> creates a new host name in the
     * store.
     *
     * @param name a <code>String</code> host name.
     *
     * @exception Exception if an error occurs.
     */
    public void createHostName(String name) throws Exception;

    /**
     * <code>readAllHostNames</code> reads all host names from the
     * store.
     *
     * @return a <code>Collection</code> of <code>String</code>s.
     *
     * @exception Exception if an error occurs.
     */
    public Collection readAllHostNames() throws Exception;

    /**
     * <code>readHostNameById</code> reads a host name from the store
     * by ID.
     *
     * @param hostId an <code>int</code> host ID.
     *
     * @return a <code>String</code> host name.
     *
     * @exception Exception if an error occurs.
     */
    public String readHostNameById(int hostId) throws Exception;

    /**
     * <code>readHostIdByName</code> reads a host ID from the store by
     * name.
     *
     * @param name a <code>String</code>.
     *
     * @return an <code>int</code> host ID.
     *
     * @exception Exception if an error occurs.
     */
    public int readHostIdByName(String name) throws Exception;

    /**
     * <code>readHostNamesByJobId</code> reads the name(s) of the
     * host(s) on which a job ran.
     *
     * @param jobId an <code>int</code> ID.
     *
     * @return a <code>String []</code> array of host names.
     *
     * @exception Exception if an error occurs.
     */
    public String [] readHostNamesByJobId(int jobId) throws Exception;

    /**
     * <code>deleteHostName</code> deletes a host name from the store.
     *
     * @param name a <code>String</code> host name.
     *
     * @exception Exception if an error occurs.
     */
    public void deleteHostName(String name) throws Exception;
}
