/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.dao;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.ResourceBundle;
import javax.sql.DataSource;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControlException;

/**
 * <code>HostNameSQLDAO</code> implements <code>HostNameDAO</code> for
 * persistence tasks using a relational database.
 *
 * @author Keith James
 */
public class HostNameSQLDAO extends SQLDAOBaseImpl implements HostNameDAO
{
    // Log4j setup
    static Category cat = Category.getInstance(HostNameSQLDAO.class.getName());

    // SQL statement resource
    static ResourceBundle sqlStatements =
        ResourceBundle.getBundle(HostNameSQLDAO.class.getName());

    // Map to cache instances, keyed by ID
    private HashMap canonical = new HashMap(50);

    private String [] stringProto = new String [0];

    /**
     * Creates a new <code>HostNameSQLDAO</code> with no
     * <code>DataSource</code>.
     */
    public HostNameSQLDAO()
    {
        super();
    }

    /**
     * Creates a new <code>HostNameSQLDAO</code> with the specified
     * <code>DataSource</code>.
     *
     * @param datasource a <code>DataSource</code>.
     */
    public HostNameSQLDAO(DataSource datasource)
    {
        super(datasource);
    }

    public synchronized void createHostName(String name) throws Exception
    {
        String sql = sqlStatements.getString("createHostName");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        boolean rollback = false;

        try
        {
            int newId = createNewHostId();

            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, newId);
            ps.setString(2, name);
            cat.debug("Executing SQL: " + ps);
            ps.executeUpdate();
        }
        catch (Exception e)
        {
            cat.error("Caught an Exception; rolling back transaction", e);
            rollback = true;
            throw e;
        }
        finally
        {
            cSupport.close(ps);

            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    public synchronized Collection readAllHostNames() throws SQLException
    {
        String sql = sqlStatements.getString("readAllHostNames");

        Connection c = getManagedConnection();
        Statement s = null;
        ResultSet r = null;
        List names = new ArrayList();

        try
        {
            cat.debug("Preparing SQL: " + sql);
            s = c.createStatement();
            cat.debug("Executing SQL: " + s);
            r = s.executeQuery(sql);

            while (r.next())
            {
                names.add(r.getString("name"));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(s);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return names;
    }

    public synchronized String readHostNameById(int hostId)
        throws SQLException, JobControlException
    {
        // Test canonical instances first
        Integer canonKey = new Integer(hostId);
        if (canonical.containsKey(canonKey))
        {
            cat.debug("Fetching canonical host name [" + canonKey + "] from cache");
            return (String) canonical.get(canonKey);
        }

        String sql = sqlStatements.getString("readHostNameById");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        String          name = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, hostId);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            if (r.next())
            {
                name = r.getString("name");;
                // Add to canonical instances
                canonical.put(canonKey, name);
            }
            else
            {
                throw new JobControlException("Requested a non-existent host name with ID '"
                                              + hostId
                                              + "'");
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return name;
    }

    public synchronized int readHostIdByName(String name)
        throws SQLException, JobControlException
    {
        int hostId = 0;

        if (canonical.containsValue(name))
        {
            Iterator ei = canonical.entrySet().iterator();

            while (ei.hasNext())
            {
                Map.Entry e = (Map.Entry) ei.next();
                if (((String) e.getValue()).equals(name))
                    hostId = ((Integer) e.getKey()).intValue();
            }
        }
        else
        {
            String sql = sqlStatements.getString("readHostIdByName");

            Connection c = getManagedConnection();
            PreparedStatement ps = null;
            ResultSet          r = null;

            try
            {
                cat.debug("Preparing SQL: " + sql);
                ps = c.prepareStatement(sql);
                ps.setString(1, name);
                cat.debug("Executing SQL: " + ps);
                r = ps.executeQuery();

                if (r.next())
                {
                    hostId = r.getInt("host_id");
                    // Add to canonical instances
                    canonical.put(new Integer(hostId), name);
                }
                else
                {
                     throw new JobControlException("Requested a non-existent host name with ID '"
                                                   + hostId
                                                   + "'");
                }
            }
            finally
            {
                cSupport.close(r);
                cSupport.close(ps);

                if (isManagingConnections())
                    cSupport.close(c);
            }
        }

        return hostId;
    }

    public synchronized String [] readHostNamesByJobId(int jobId)
        throws SQLException
    {
        String sql = sqlStatements.getString("readHostNamesByJobId");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        List hosts = new ArrayList();

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, jobId);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();
            
            while (r.next())
            {
                hosts.add(r.getString("name"));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return (String []) hosts.toArray(stringProto);
    }

    public synchronized void deleteHostName(String name)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("deleteHostName");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        boolean rollback = false;

        int hostId = readHostIdByName(name);

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, hostId);
            cat.debug("Executing SQL: " + ps);
            ps.executeUpdate();
        }
        catch (SQLException e)
        {
            cat.error("Caught an Exception; rolling back transaction", e);
            rollback = true;
            throw e;
        }
        finally
        {
            cSupport.close(ps);

            if (rollback)
            {
                cSupport.rollback(c);
            }
            else
            {
                cSupport.commit(c);
                canonical.remove(new Integer(hostId));
            }
        }
        }

    /**
     * <code>createNewHostId</code> gets a new unique ID from a
     * sequence.
     *
     * @return an <code>int</code> host ID.
     *
     * @exception Exception if an error occurs.
     */
    private int createNewHostId() throws Exception
    {
        String sql = sqlStatements.getString("createNewHostId");

        Connection c = getConnection();
        Statement s = null;
        ResultSet r = null;
        int id = 0;

        try
        {
            s = c.createStatement();
            cat.debug("Executing SQL: " + s);
            r = s.executeQuery(sql);

            if (r.next())
                id = r.getInt("host_id");
            else
                throw new Exception("Failed to retrieve a new host ID");
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(s);
        }

        return id;
    }
}
