/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.dao;

import java.util.Collection;
import java.util.List;

import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.Status;

/**
 * <code>JobDAO</code> describes data access objects for persistence
 * of <code>Job</code>s. Possible stores are relational databases
 * or XML files.
 *
 * @author Keith James
 */
public interface JobDAO
{
    /**
     * <code>readJobById</code> reads a single <code>Job</code> from
     * the store.
     *
     * @param jobId an <code>int</code> ID.
     *
     * @return a <code>Job</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Job readJobById(int jobId) throws Exception;

    /**
     * <code>readJobsByTaskId</code> reads all <code>Job</code>s
     * belonging to the <code>Task</code> with the specified ID.
     *
     * @param taskId an <code>int</code> iD.
     *
     * @return a <code>Collection</code> of <code>Job</code>s.
     *
     * @exception Exception if an error occurs.
     */
    public Collection readJobsByTaskId(int taskId) throws Exception;

    /**
     * <code>readTaskIdByJob</code> returns the ID of the
     * <code>Task</code> to which this <code>Job</code> belongs.
     *
     * @param job a <code>Job</code>.
     *
     * @return an <code>int</code> <code>Task</code> ID, or -1 if the
     * <code>Task</code> can not be identified.
     *
     * @exception Exception if an error occurs.
     */
    public int readTaskIdByJob(Job job) throws Exception;

    /**
     * <code>readJobPath</code> returns a <code>List</code> of
     * <code>Job</code>s which form the path between a
     * <code>Job</code> and the root <code>Job</code> (exclusive of
     * the start <code>Job</code>, but inclusive of the root).
     *
     * @param job a <code>Job</code>.
     *
     * @return a <code>List</code> of <code>Job</code>s.
     *
     * @exception Exception if an error occurs.
     */
    public List readJobPath(Job job) throws Exception;

    /**
     * <code>readJobsByStatusAndTaskId</code> reads all
     * <code>Job</code>s for a particular <code>Task</code> from the
     * store which have the requested <code>Status</code>. The
     * <code>Task</code> is specified by its ID.
     *
     * @param status a <code>Status</code>.
     * @param taskId an <code>int</code> ID.
     *
     * @return a <code>Collection</code> of <code>Job</code>s.
     *
     * @exception Exception if an error occurs.
     */
    public Collection readJobsByStatusAndTaskId(Status status,
                                                int    taskId)
        throws Exception;


    /**
     * <code>readNJobsByStatusAndTaskId</code> reads n
     * <code>Job</code>s for a particular <code>Task</code> from the
     * store which have the requested <code>Status</code>. The
     * <code>Task</code> is specified by its ID.
     *
     * @param n an <code>int</code> maximum number of
     * <code>Job</code>s to return.
     * @param status a <code>Status</code>.
     * @param taskId an <code>int</code> <code>Task</code> ID.
     *
     * @return a <code>Collection</code> of <code>Job</code>s.
     *
     * @exception Exception if an error occurs.
     */
    public Collection readNJobsByStatusAndTaskId(int    n,
                                                 Status status,
                                                 int    taskId)
        throws Exception;

    /**
     * <code>updateJob</code> updates a <code>Job</code> with new
     * values.
     *
     * @param job a <code>Job</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void updateJob(Job job) throws Exception;

    /**
     * <code>updateJobReady</code> updates a <code>Job</code> as
     * ready.
     *
     * @param job a <code>Job</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void updateJobReady(Job job) throws Exception;

    /**
     * <code>updateJobStarted</code> updates a <code>Job</code> as
     * started. It can be expected to set start time and status.
     *
     * @param job a <code>Job</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void updateJobStarted(Job job) throws Exception;

    /**
     * <code>updateJobCompleted</code> updates a <code>Job</code> as
     * reaching completion.
     *
     * @param job a <code>Job</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void updateJobCompleted(Job job) throws Exception;

    /**
     * <code>updateJobRetry</code> updates a <code>Job</code> as being
     * retried.
     *
     * @param job a <code>Job</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void updateJobRetry(Job job) throws Exception;

    /**
     * <code>updateJobFailed</code> updates a <code>Job</code> as
     * failed.
     *
     * @param job a <code>Job</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void updateJobFailed(Job job) throws Exception;

    /**
     * <code>updateJobFailedCascade</code> updates a path of
     * <code>Job</code>s towards the root such that any viable
     * <code>Job</code> which can not be made READY is set to
     * FAILED. This is defined by it being WAITING or READY, having
     * the OR operator and not having all its children FAILED. Such a
     * case breaks the failure path at that point.
     *
     * @param path a <code>Job</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void updateJobFailedCascade(Job job) throws Exception;

    public void updateJobRestore(Job job) throws Exception;

    /**
     * <code>updateJobCancelled</code> updates a <code>Job</code> as
     * cancelled.
     *
     * @param job a <code>Job</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void updateJobCancelled(Job job) throws Exception;
}
