/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.dao;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Stack;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.job.Job;

/**
 * <p><code>JobNestedSetHelper</code> traverses a tree in preorder and
 * produces two integers for each node such that the integers for each
 * member of the subtree below that node fall between that node's
 * values. This means that the subtrees are resolved into nested
 * sets. See chapter 29 of "Joe Celko's SQL for Smarties" (ISBN
 * 1-55860-576-2).</p>
 *
 * <p>The values are returned as a preorder <code>List</code> of
 * <code>Job</code>s.</p>
 *
 * @author Keith James
 */
public class JobNestedSetHelper
{
    // Log4j setup
    static Category cat = Category.getInstance(JobNestedSetHelper.class.getName());

    /**
     * <code>calculateNestedSets</code> calculates the left (low) and
     * right (high) values for representing a tree as nested sets.
     *
     * @param job a <code>Job</code>.
     * @param low an <code>int</code> to be the (left) low value of
     * the root.
     *
     * @return a <code>List</code> of nodes in preorder.
     */
    public List calculateNestedSets(Job job, int low)
    {
        Stack enumStack = new Stack();
        Stack  jobStack = new Stack();
        List    jobList = new ArrayList();

        job.setLow(low);
        cat.debug("Set root Job low to " + low);
        jobList.add(job);

        // Set counter to low
        int count = low;
        enumStack.push(job.children());

        // While the node we are dealing with has more children
        while (! enumStack.isEmpty() &&
               ((Enumeration) enumStack.peek()).hasMoreElements())
        {
            // Get the remaining children
            Enumeration childEnum = (Enumeration) enumStack.peek();

            // Increment ready to give low number
            count++;

            // Get the first child and push it onto the job stack
            // while giving it a low number
            Job i = (Job) childEnum.nextElement();
            i.setLow(count);
            cat.debug("Set Job low to " + count);
            jobStack.push(i);
            cat.debug("Pushed Job onto stack");

            // Also add it to the List
            jobList.add(i);

            // If this child has children push them onto the
            // enumeration stack (preorder enumeration)
            Enumeration grandChildEnum = i.children();

            if (grandChildEnum.hasMoreElements())
            {
                enumStack.push(grandChildEnum);
                // Start work on the first grandchild as new child
                continue;
            }
            else
            {
                // Increment ready to give high number
                count++;

                // We ran out of children so we get this child back
                // off the job stack and give it a high number,
                // closing the set
                Job j = (Job) jobStack.pop();
                cat.debug("Popped Job from stack");
                j.setHigh(count);
                cat.debug("Set Job high to " + count);
            }

            // We are out of children if we reach here. Now empty the
            // stack. Each time there is an empty enumeration left on
            // the enumeration stack there should be a job on the job
            // stack which requires its set to be closed
            while ((! enumStack.isEmpty() &&
                    (! ((Enumeration) enumStack.peek()).hasMoreElements())))
            {
                enumStack.pop();

                if (! jobStack.isEmpty())
                {
                    // Increment ready to give high number
                    count++;

                    // Close the set encompassing the children
                    Job j = (Job) jobStack.pop();
                    cat.debug("Popped Job from stack");
                    j.setHigh(count);
                    cat.debug("Set Job high to " + count);
                }
            }
        }

        // Finally increment ready to give high number for the root
        count++;
        job.setHigh(count);
        cat.debug("Set root Job high to " + count);

        return jobList;
    }
}
