/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.dao;

import java.sql.*;
import java.util.*;
import javax.sql.DataSource;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.job.*;

/**
 * <p><code>JobSQLDAO</code> implements <code>JobDAO</code> for
 * persistence tasks using a relational database. It covers
 * manipulation of <code>Job</code>s outside the context of a
 * <code>Task</code>. <code>Job</code>s should only be created within
 * the context of a <code>Task</code>, therefore to create new
 * <code>Job</code>s in the store use a <code>TaskDAO</code> which
 * will store it a <code>Task</code>'s <code>Job</code>s
 * implicitly.</p>
 *
 * @author Keith James
 */
public class JobSQLDAO extends SQLDAOBaseImpl implements JobDAO
{
    static final int CASCADE_EXITVALUE = 999;

    // Log4j setup
    static Category cat = Category.getInstance(JobSQLDAO.class.getName());

    // SQL statement resource
    static ResourceBundle sqlStatements =
            ResourceBundle.getBundle(JobSQLDAO.class.getName());

    private ExecutableSQLDAO executableSQLDAO;
    private StatusSQLDAO     statusSQLDAO;
    private HostNameSQLDAO   hostNameSQLDAO;

    private JobNestedSetHelper setHelper = new JobNestedSetHelper();

    /**
     * Creates a new <code>JobSQLDAO</code> with no
     * <code>DataSource</code>.
     */
    public JobSQLDAO()
    {
        super();

        executableSQLDAO = new ExecutableSQLDAO();
        statusSQLDAO     = new StatusSQLDAO();
        hostNameSQLDAO   = new HostNameSQLDAO();
    }

    /**
     * Creates a new <code>JobSQLDAO</code> with the specified
     * <code>DataSource</code>.
     *
     * @param datasource a <code>DataSource</code>.
     */
    public JobSQLDAO(DataSource datasource)
    {
        super(datasource);

        executableSQLDAO = new ExecutableSQLDAO();
        statusSQLDAO     = new StatusSQLDAO();
        hostNameSQLDAO   = new HostNameSQLDAO();
    }

    public synchronized Job readJobById(int jobId)
        throws Exception, JobControlException
    {
        Connection c = getManagedConnection();
        Job job = null;

        try
        {
            job = _readJobById(jobId);
        }
        finally
        {
            if (isManagingConnections())
                cSupport.close(c);
        }

        return job;
    }

    public synchronized Collection readJobsByTaskId(int taskId)
        throws SQLException, JobControlException
    {
         String sql = sqlStatements.getString("readJobsByTaskId");

         Connection c = getManagedConnection();
         PreparedStatement ps = null;
         ResultSet          r = null;

         List jobs = new ArrayList();

         try
         {
             cat.debug("Preparing SQL: " + sql);
             ps = c.prepareStatement(sql);
             ps.setInt(1, taskId);
             cat.debug("Executing SQL: " + ps);
             r = ps.executeQuery();

            while (r.next())
            {
                jobs.add(populateJob(r));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return jobs;
    }

    public synchronized int readTaskIdByJob(Job job)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("readTaskIdByJob");

        Connection c = getManagedConnection();
        int taskId = 0;

        try
        {
            taskId = _readTaskIdByJob(job);
        }
        finally
        {
            if (isManagingConnections())
                cSupport.close(c);
        }

        return taskId;
    }

    public synchronized Collection readJobsByStatusAndTaskId(Status status,
                                                             int    taskId)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("readJobsByStatusAndTaskId");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        List jobs = new ArrayList();

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            // Select all Jobs for this Task which have required Status
            ps.setInt(1, taskId);
            ps.setInt(2, status.getId());
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            while (r.next())
            {
                jobs.add(populateJob(r));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return jobs;
    }                                 

    public synchronized Collection readNJobsByStatusAndTaskId(int    n,
                                                              Status status,
                                                              int    taskId)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("readNJobsByStatusAndTaskId");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        List jobs = new ArrayList();

        int rowCount = 0;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, taskId);
            ps.setInt(2, status.getId());
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            while (r.next() && (rowCount < n))
            {
                jobs.add(populateJob(r));
                ++rowCount;
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return jobs;
    }

    public synchronized List readJobPath(Job job)
        throws SQLException, JobControlException
    {
        Connection c = getManagedConnection();
        List path = null;

        try
        {
            path = _readJobPath(job);
        }
        finally
        {
            if (isManagingConnections())
                cSupport.close(c);
        }

        cat.debug("Path to root from Job ["
                  + job.getId()
                  + "] is "
                  + path);

        return path;
    }

    /**
     * <code>updateJob</code> updates a <code>Job</code> with new
     * values. The values updated are start time, completion time,
     * exit value, retry count and <code>Status</code>.
     *
     * @param job a <code>Job</code>.
     *
     * @exception SQLException if an error occurs.
     * @exception JobControlException if an error occurs.
     */
    public synchronized void updateJob(Job job)
        throws SQLException, JobControlException
    {
        int                jobId = job.getId();
        Timestamp      startTime = job.getStartTime();
        Timestamp completionTime = job.getCompletionTime();
        int            exitValue = job.getExitValue();
        int           retryCount = job.getRetryCount();
        int             statusId = job.getStatus().getId();

        String sql = sqlStatements.getString("updateJobById");

        Connection c = getManagedConnection();
        PreparedStatement us = null;
        boolean rollback = false;

        try
        {
            lockJob(jobId);

            // Update statement
            cat.debug("Preparing SQL: " + sql);
            us = c.prepareStatement(sql);
            us.setInt(1, statusId);

            if (startTime != null)
                us.setTimestamp(2, startTime);
            else
                us.setNull(2, Types.TIMESTAMP);

            if (completionTime != null)
                us.setTimestamp(3, completionTime);
            else
                us.setNull(3, Types.TIMESTAMP);

            us.setInt(4, retryCount);
            us.setInt(5, exitValue);
            us.setInt(6, jobId);
            cat.debug("Executing SQL: " + us);
            us.executeUpdate();
        }
        catch (SQLException se)
        {
            cat.error("Caught Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        catch (JobControlException je)
        {
            cat.error("Caught Exception; rolling back transaction", je);
            rollback = true;
            throw je;
        }
        finally
        {
            cSupport.close(us);

            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    public synchronized void updateJobReady(Job job)
        throws SQLException, JobControlException
    {
        Connection c = getManagedConnection();
        boolean rollback = false;

        try
        {
            _updateJobReady(job);
        }
        catch (SQLException se)
        {
            cat.error("Caught Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        catch (JobControlException je)
        {
            cat.error("Caught Exception; rolling back transaction", je);
            rollback = true;
            throw je;
        }
        finally
        {
            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    public synchronized void updateJobStarted(Job job)
        throws SQLException, JobControlException
    {
        int      jobId = job.getId();
        int   statusId = job.getStatus().getId();
        int retryCount = job.getRetryCount();

        int    batchId = 0;
        String   queue = null;

        if (job instanceof BatchJob)
        {
            BatchJob batchJob = (BatchJob) job;
            batchId = batchJob.getBatchId();
            queue = batchJob.getQueue();
        }

        // Assert READY
        if (statusId != Status.READY)
            throw new JobControlException("Failed to start Job ["
                                          + jobId
                                          + "] as it has status '"
                                          + job.getStatus().getDescription()
                                          + "'");

        String sql = sqlStatements.getString("updateJobStartedById");

        Connection c = getManagedConnection();
        PreparedStatement us = null;
        boolean rollback = false;

        try
        {
            lockJob(jobId);

            // Update statement
            cat.debug("Preparing SQL: " + sql);
            us = c.prepareStatement(sql);
            us.setInt(1, Status.STARTED);
            us.setInt(2, batchId);

            if (queue != null)
                us.setString(3, queue);
            else
                us.setNull(3, Types.VARCHAR);

            us.setInt(4, retryCount);
            us.setInt(5, jobId);
            cat.debug("Executing SQL: " + us);
            us.executeUpdate();
        }
        catch (SQLException se)
        {
            cat.error("Caught Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        catch (JobControlException je)
        {
            cat.error("Caught Exception; rolling back transaction", je);
            rollback = true;
            throw je;
        }
        finally
        {
            cSupport.close(us);

            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    public synchronized void updateJobCompleted(Job job)
        throws SQLException, JobControlException
    {
        int     jobId = job.getId();
        int  statusId = job.getStatus().getId();
        int exitValue = job.getExitValue();

        // Assert STARTED.
        if (! (statusId == Status.STARTED))
            throw new JobControlException("Failed to complete Job ["
                                          + jobId
                                          + "] as it has status '"
                                          + job.getStatus().getDescription()
                                          + "'");

        String sql = sqlStatements.getString("updateJobCompletedById");

        Connection c = getManagedConnection();
        PreparedStatement us = null;
        boolean rollback = false;

        try
        {
            lockJob(jobId);

            // Update statement
            cat.debug("Preparing SQL: " + sql);
            us = c.prepareStatement(sql);
            us.setInt(1, Status.COMPLETED);
            us.setInt(2, exitValue);
            us.setInt(3, jobId);
            cat.debug("Executing SQL: " + us);
            us.executeUpdate();

            if (job instanceof BatchJob)
            {
                updateJobHosts(job);
            }
        }
        catch (SQLException se)
        {
            cat.error("Caught Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        catch (JobControlException je)
        {
            cat.error("Caught Exception; rolling back transaction", je);
            rollback = true;
            throw je;
        }
        finally
        {
            cSupport.close(us);

            if (rollback)
            {
                cSupport.rollback(c);
                cSupport.close(c);
            }
        }

        int parentId = _readJobParentId(job);
        int   taskId = _readTaskIdByJob(job);;

        // If parent_id is 0 this Job has no parent and is the root
        if (parentId == 0)
        {
            // The goal Job has low == 0, so if we are there we are done
            if (job.getLow() == 0)
            {
                cSupport.commit(c);
                if (isManagingConnections())
                    cSupport.close(c);

                return;
            }
            else
            {
                try
                {
                    // Here we kick off the leaf Jobs belonging to the
                    // next root Job in line
                    updateNextJobTreeReady(job, taskId);
                }
                catch (SQLException se)
                {
                    cat.error("Caught Exception; rolling back transaction", se);
                    rollback = true;
                    throw se;
                }
                finally
                {
                    if (rollback)
                        cSupport.rollback(c);
                    else
                        cSupport.commit(c);

                    if (isManagingConnections())
                        cSupport.close(c);
                }

                return;
            }
        }

        try
        {
            // Read parent
            Job parent = _readJobById(parentId);
            String parentOperator = parent.getOperator();
            int parentStatusId = parent.getStatus().getId();

            if (parentOperator.equals(Job.OPERATOR_AND))
            {
                // Case where all siblings must complete before parent can
                // move from WAITING to READY
                switch (parentStatusId)
                {
                    // Major case. Parent is WAITING and may be set to
                    // READY if all children are COMPLETED
                    case (Status.WAITING):
                        sql = sqlStatements.getString("readChildStatusSummaryById");

                        PreparedStatement ps = null;
                        ResultSet   childRes = null;
                        boolean allComplete = commonChildStatus(parent,
                                                                Status.COMPLETED,
                                                                taskId);

                        // If all siblings are complete we can update the
                        // parent as READY
                        if (allComplete)
                        {
                            cat.info("All children of Job ["
                                     + parentId
                                     + "] are complete");
                            _updateJobReady(parent);
                        }
                        break;

                    case (Status.FAILED):
                        cat.info("Parent of Job ["
                                 + jobId
                                 + "] could not be set READY as it had already FAILED");
                        break;

                    case (Status.SKIPPED):
                        cat.info("Parent could not be set READY as it had already SKIPPED");
                        break;

                    case (Status.CANCELLED):
                        cat.info("Parent could not be set READY as it had already CANCELLED");
                        break;

                    default:
                        cat.warn("Unexpected Job parent state '"
                                 + parentStatusId
                                 + "' ("
                                 + parent.getStatus().getDescription()
                                 + ")");
                }
            }
            else if (parentOperator.equals(Job.OPERATOR_OR))
            {
                switch (parentStatusId)
                {
                    case (Status.WAITING):
                        // Should get all child Jobs which are WAITING
                        // or READY below this parent and set them to
                        // SKIPPED
                        updateRedundantChildJobs(parentId,
                                                 Status.SKIPPED,
                                                 taskId);
                        _updateJobReady(parent);
                        break;

                    case (Status.READY):
                        // Parent has already been updated as READY so the
                        // thread which did that should be handling
                        // skipping of redundant child Jobs
                        break;

                    case (Status.STARTED):
                        // Parent has already been updated as READY,
                        // then STARTED so the threads which did that
                        // should be handling skipping of redundant
                        // child Jobs
                        break;

                    case (Status.COMPLETED):
                        cat.warn("Parent could not be set READY as it had already COMPLETED");
                        break;

                    case (Status.SKIPPED):
                        cat.warn("Parent could not be set READY as it had already been SKIPPED");
                        break;

                    case (Status.FAILED):
                        cat.warn("Parent could not be set READY as it had already FAILED");
                        break;

                    default:
                        cat.warn("Unexpected Job parent state '"
                                 + parentStatusId
                                 + "' ("
                                 + parent.getStatus().getDescription()
                                 + ")");
                }
            }
            else
            {
                cat.warn("Unexpected Job parent operator '"
                         + parent.getOperator()
                         + "'");
            }
        }
        catch (SQLException se)
        {
            cat.error("Caught Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        catch (JobControlException je)
        {
            cat.error("Caught Exception; rolling back transaction", je);
            rollback = true;
            throw je;
        }
        finally
        {
            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    public void updateJobRetry(Job job)
        throws SQLException, JobControlException
    {
        int         jobId = job.getId();
        int      statusId = job.getStatus().getId();
        int     exitValue = job.getExitValue();
        int    retryCount = job.getRetryCount();
        int maxRetryCount = job.getMaxRetryCount();

        // Assert READY or STARTED
        if (! (statusId == Status.READY || statusId == Status.STARTED))
            throw new JobControlException("Failed to retry Job ["
                                          + jobId
                                          + "] as it has status '"
                                          + job.getStatus().getDescription()
                                          + "'");

        String sql = sqlStatements.getString("updateJobRetryById");

        retryCount++;

        cat.info("Job ["
                 + jobId
                 + "] failed with exit value "
                 + exitValue
                 + ". Now running retry "
                 + retryCount
                 + " of maximum "
                 + maxRetryCount);

        Connection c = getManagedConnection();
        PreparedStatement us = null;
        boolean rollback = false;

        try
        {
            lockJob(jobId);

            // Update statement
            cat.debug("Preparing SQL: " + sql);
            us = c.prepareStatement(sql);
            us.setInt(1, Status.READY);
            us.setInt(2, retryCount);
            us.setInt(3, jobId);
            cat.debug("Executing SQL: " + us);
            us.executeUpdate();
        }
        catch (SQLException se)
        {
            cat.error("Caught Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        finally
        {
            cSupport.close(us);

            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    public synchronized void updateJobFailed(Job job)
        throws SQLException, JobControlException
    {
        int         jobId = job.getId();
        int      statusId = job.getStatus().getId();
        int     exitValue = job.getExitValue();
        int    retryCount = job.getRetryCount();
        int maxRetryCount = job.getMaxRetryCount();

        // Assert READY or STARTED
        if (! (statusId == Status.READY || statusId == Status.STARTED))
            throw new JobControlException("Failed to fail Job ["
                                          + jobId
                                          + "] as it has status '"
                                          + job.getStatus().getDescription()
                                          + "'");

        if (retryCount < maxRetryCount)
        {
            updateJobRetry(job);
        }
        else
        {
            cat.info("Job ["
                     + jobId
                     + "] failed with exit value "
                     + exitValue);
            cat.debug("Updating single Job failure");

            String sql = sqlStatements.getString("updateJobFailedById");

            Connection c = getManagedConnection();
            PreparedStatement us = null;
            boolean rollback = false;

            try
            {
                lockJob(jobId);

                // Update statement
                cat.debug("Preparing SQL: " + sql);
                us = c.prepareStatement(sql);
                us.setInt(1, Status.FAILED);
                us.setInt(2, exitValue);
                us.setInt(3, jobId);
                cat.debug("Executing SQL: " + us);
                us.executeUpdate();

                if (job instanceof BatchJob)
                    updateJobHosts(job);
            }
            catch (SQLException se)
            {
                cat.error("Caught Exception; rolling back transaction", se);
                rollback = true;
                throw se;
            }
            catch (JobControlException je)
            {
                cat.error("Caught Exception; rolling back transaction", je);
                rollback = true;
                throw je;
            }
            finally
            {
                cSupport.close(us);

                if (rollback)
                    cSupport.rollback(c);
                else
                    cSupport.commit(c);

                if (isManagingConnections())
                    cSupport.close(c);
            }
        }
    }

    public synchronized void updateJobFailedCascade(Job job)
        throws SQLException, JobControlException
    {
        int         jobId = job.getId();
        int      statusId = job.getStatus().getId();
        int     exitValue = job.getExitValue();
        int    retryCount = job.getRetryCount();
        int maxRetryCount = job.getMaxRetryCount();

        // Assert READY or STARTED
        if (! (statusId == Status.READY || statusId == Status.STARTED))
            throw new JobControlException("Failed to fail Job ["
                                          + jobId
                                          + "] as it has status '"
                                          + job.getStatus().getDescription()
                                          + "'");

        if (retryCount < maxRetryCount)
        {
            updateJobRetry(job);
        }
        else
        {
            cat.info("Job ["
                     + jobId
                     + "] failed with exit value "
                     + exitValue);
            cat.debug("Updating Job failure cascade");

            String sql = sqlStatements.getString("updateJobFailedById");

            Connection c = getManagedConnection();
            PreparedStatement us = null;
            boolean rollback = false;

            try
            {
                int taskId = _readTaskIdByJob(job);

                List  jobPath = _readJobPath(job);
                List failPath = filterJobFailurePath(jobPath, taskId);

                if (job instanceof BatchJob)
                    updateJobHosts(job);

                cat.debug("Preparing SQL: " + sql);
                us = c.prepareStatement(sql);

                // Mark the Job which had a runtime failure as FAILED
                lockJob(jobId);
                us.setInt(1, Status.FAILED);
                us.setInt(2, exitValue);
                us.setInt(3, jobId);
                us.addBatch();

                // Mark dependent Jobs as CANCELLED so that we don't
                // try to run them
                for (Iterator fpi = failPath.iterator(); fpi.hasNext();)
                {
                    Job   fail = (Job) fpi.next();
                    int failId = fail.getId();
                    lockJob(failId);

                    us.setInt(1, Status.CANCELLED);
                    us.setInt(2, CASCADE_EXITVALUE);
                    us.setInt(3, failId);
                    us.addBatch();
                }

                cat.debug("Executing SQL: " + us);
                us.executeBatch();
            }
            catch (SQLException se)
            {
                cat.error("Caught Exception; rolling back transaction", se);
                rollback = true;
                throw se;
            }
            catch (JobControlException je)
            {
                cat.error("Caught Exception; rolling back transaction", je);
                rollback = true;
                throw je;
            }
            finally
            {
                cSupport.close(us);

                if (rollback)
                    cSupport.rollback(c);
                else
                    cSupport.commit(c);

                if (isManagingConnections())
                    cSupport.close(c);
            }
        }
    }

    public void updateJobRestore(Job job)
        throws SQLException, JobControlException
    {
        int    jobId = job.getId();
        int statusId = job.getStatus().getId();

        // Case where someone tries to continue a Job which has not
        // failed
        if (statusId != Status.FAILED)
            throw new JobControlException("Failed to continue Job ["
                                          + jobId
                                          + "] as it has status '"
                                          + job.getStatus().getDescription()
                                          + ". Unable to continue a Job which has not failed");

        String sql = sqlStatements.getString("updateSkippedJobRestoreById");

        Connection c = getManagedConnection();
        PreparedStatement us = null;
        boolean rollback = false;

        try
        {
            lockJob(jobId);
            int taskId = _readTaskIdByJob(job);

            List skipped = lockRestorableJobs(jobId, Status.SKIPPED, taskId);
            List failed = lockRestorableJobs(jobId, Status.FAILED, taskId);

            if (skipped.size() > 0)
            {
                cat.debug("Preparing SQL: " + sql);
                us = c.prepareStatement(sql);

                // Update cascaded SKIPPED Jobs to WAITING and null their
                // dummy exit_value
                for (Iterator si = skipped.iterator(); si.hasNext();)
                {
                    Integer id = (Integer) si.next();
                    cat.info("Marking SKIPPED Job ["
                             + id
                             + "] as restorable");

                    us.setInt(1, Status.WAITING);
                    us.setNull(2, Types.INTEGER);
                    us.setInt(3, id.intValue());
                    us.addBatch();
                }

                cat.debug("Executing SQL: " + us);
                us.executeBatch();
                cSupport.close(us);
            }

            if (failed.size() > 0)
            {
                sql = sqlStatements.getString("updateFailedJobRestoreById");

                cat.debug("Preparing SQL: " + sql);
                us = c.prepareStatement(sql);

                // Update FAILED Jobs to READY, null their start_time,
                // completion_time and exit_value
                for (Iterator fi = failed.iterator(); fi.hasNext();)
                {
                    Integer id = (Integer) fi.next();
                    cat.info("Marking FAILED Job ["
                             + id
                             + "] as restorable");

                    us.setInt(1, Status.READY);
                    us.setNull(2, Types.TIMESTAMP);
                    us.setNull(3, Types.TIMESTAMP);
                    us.setInt(4, 0);
                    us.setInt(5, 0);
                    us.setNull(6, Types.INTEGER);
                    us.setInt(7, id.intValue());
                    us.addBatch();
                }

                cat.debug("Executing SQL: " + us);
                us.executeBatch();
                cSupport.close(us);

                deleteJobHosts(failed);
            }
        }
        catch (SQLException se)
        {
            cat.error("Caught Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        finally
        {
            cSupport.close(us);

            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    public synchronized void updateJobCancelled(Job job)
        throws SQLException, JobControlException
    {
        int         jobId = job.getId();
        int      statusId = job.getStatus().getId();
        int    retryCount = job.getRetryCount();
        int maxRetryCount = job.getMaxRetryCount();

        // Case where someone tries to cancel a Job which would not
        // run anyway because it failed too many times
        if (statusId == Status.FAILED && retryCount < maxRetryCount)
            throw new JobControlException("Failed to cancel Job ["
                                          + jobId
                                          + "] as it has status '"
                                          + job.getStatus().getDescription()
                                          + "' and "
                                          + "retryCount "
                                          + retryCount
                                          + "(maximum "
                                          + maxRetryCount
                                          + "). Unable to cancel a Job which can not be started");

        // Assert READY or STARTED
        if (! (statusId == Status.READY || statusId == Status.STARTED))
            throw new JobControlException("Failed to fail Job ["
                                          + jobId
                                          + "] as it has status '"
                                          + job.getStatus().getDescription()
                                          + "'. Unable to cancel a Job which can not be started");

        String sql = sqlStatements.getString("updateJobCancelledById");

        Connection c = getManagedConnection();
        PreparedStatement us = null;
        boolean rollback = false;

        try
        {
            lockJob(jobId);
            int taskId = _readTaskIdByJob(job);

            // Update statement
            cat.debug("Preparing SQL: " + sql);
            us = c.prepareStatement(sql);
            us.setInt(1, Status.CANCELLED);
            us.setInt(2, jobId);
            cat.debug("Executing SQL: " + us);
            us.executeUpdate();

            // If a Job is CANCELLED all its child Jobs which are WAITING
            // or READY should be CANCELLED too.
            updateRedundantChildJobs(jobId, Status.CANCELLED, taskId);
        }
        catch (SQLException se)
        {
            cat.error("Caught Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        finally
        {
            cSupport.close(us);

            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    /**
     * <code>createJob</code> creates a new <code>Job</code> in the
     * database. This method is package private because a
     * <code>Job</code> should never be created in the database
     * outside the context of a <code>Task</code> (in fact, an attempt
     * to do this will most likely fail because of foreign key
     * constraints).
     *
     * @param job a <code>Job</code>.
     * @param taskId an <code>int</code> ID.
     * @param low an <code>int</code> to be the low value for the
     * root.
     *
     * @return high an <code>int</code> which is the high value
     * assigned for newly created root.
     *
     * @exception Exception if an error occurs.
     */
    synchronized int createJob(Job job, int taskId, int low) throws Exception
    {
        String sql = sqlStatements.getString("createJob");

        List preorderList = setHelper.calculateNestedSets(job, low);

        Connection c = getConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;

        // The following is a hack because our Oracle version is 8.1.7
        // which does not have JDBC 3 and consequently does not have
        // the method getGeneratedKeys on Statement.

        // WARNINING! This required temporarily disabling the foreign
        // key constraint between job and task. Remember to put this
        // back when the Oracle upgrade is done.

        // Temporary value so that we can query the generated keys
        // where task_id = 0
        int   tempTaskId = 0;
        int tempParentId = 0;
        int     keyCount = 0;

        // Insert jobs with task_id = 0 and parent_id = 0
        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);

            for (Iterator ji = preorderList.iterator(); ji.hasNext();)
            {
                Job j = (Job) ji.next();

                ps.setInt(1, tempTaskId);
                ps.setInt(2, tempParentId);
                ps.setInt(3, j.getLow());
                ps.setInt(4, j.getHigh());
                ps.setString(5, j.getOperator());

                Executable executable = j.getExecutable();
                if (executable == null)
                    throw new JobControlException("Failed to create Job "
                                                  + "as it has a null "
                                                  + "Executable");
                else
                    ps.setInt(6, executable.getId());

                Status status = j.getStatus();
                if (j == null)
                    throw new JobControlException("Failed to create Job "
                                                  + "as it has a null Status");
                else
                    ps.setInt(7, status.getId());

                ps.setNull(8, Types.TIMESTAMP);
                ps.setNull(9, Types.TIMESTAMP);

                String inputName = j.getInputName();
                if (inputName != null)
                    ps.setString(10, inputName);
                else
                    ps.setNull(10, Types.VARCHAR);

                String outputName = j.getOutputName();
                if (outputName != null)
                    ps.setString(11, outputName);
                else
                    ps.setNull(11, Types.VARCHAR);

                String errorName = j.getErrorName();
                if (errorName != null)
                    ps.setString(12, errorName);
                else
                    ps.setNull(12, Types.VARCHAR);

                String commandTemplate = j.getCommandTemplate();
                if (commandTemplate != null)
                    ps.setString(13, commandTemplate);
                else
                    ps.setNull(13, Types.VARCHAR);

                String workDirName = j.getWorkDirectoryName();
                if (workDirName != null)
                    ps.setString(14, workDirName);
                else
                    ps.setNull(14, Types.VARCHAR);

                String environment = j.getEnvironment();
                if (environment != null)
                    ps.setString(15, environment);
                else
                    ps.setNull(15, Types.VARCHAR);

                if (j instanceof BatchJob)
                {
                    ps.setString(16, ((BatchJob) j).getQueue());
                    ps.setInt(17, ((BatchJob) j).getBatchId());
                }
                else
                {
                    ps.setNull(16, Types.VARCHAR);
                    ps.setInt(17, 0);
                }

                ps.setInt(18, j.getMaxRetryCount());
                ps.setInt(19, j.getRetryCount());
                ps.setNull(20, Types.INTEGER);

                ps.addBatch();
            }

            cat.debug("Executing SQL: " + ps);
            ps.executeBatch();
        }
        finally
        {
            cSupport.close(ps);
        }

        // Query back the generated keys and set the job IDs on the
        // Job objects
        try
        {
            sql = sqlStatements.getString("readJobIdsByTaskId");
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, tempTaskId);
            cat.debug("Executing SQL: " + ps);

            r = ps.executeQuery();
            Iterator ji = preorderList.iterator();
            while (r.next() && ji.hasNext())
            {
                int jobId = r.getInt("job_id");
                keyCount++;

                Job j = (Job) ji.next();
                setObjectIntField(j, "id", jobId);                    
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);
        }

        // Sanity check - did we get back the same number of IDs as we
        // have Jobs?
        if (preorderList.size() != keyCount)
            throw new JobControlException("Created "
                                          + preorderList.size()
                                          + " Jobs but retrieved "
                                          + keyCount
                                          + " autoincremented IDs.");

        // Finally update the job table with the real task_id and
        // parent_id
        try
        {
            sql = sqlStatements.getString("updateJobTaskParentById");
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);

            for (Iterator ji = preorderList.iterator(); ji.hasNext();)
            {
                Job j = (Job) ji.next();

                ps.setInt(1, taskId);

                Job parent = (Job) j.getParent();
                if (parent == null)
                {
                    ps.setInt(2, 0);
                }
                else
                {
                    ps.setInt(2, parent.getId());
                }

                ps.setInt(3, j.getId());

                ps.addBatch();
            }

            cat.debug("Executing SQL: " + ps);
            ps.executeBatch();
        }
        finally
        {
            cSupport.close(ps);
        }

        return job.getHigh();
    }

    /**
     * <code>readRootJobsByTaskId</code> reads from the store the root
     * <code>Job</code>s for the <code>Task</code> with the specified
     * ID.
     *
     * @param taskId an <code>int</code> ID.
     *
     * @return a <code>Job []</code> array.
     *
     * @exception SQLException if an error occurs.
     * @exception JobControlException if an error occurs.
     */
    synchronized Job [] readRootJobsByTaskId(int taskId)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("readRootJobsByTaskId");

        Connection c = getConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        List jobs = new ArrayList();

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            // Select the Jobs for this Task which have a null parent
            ps.setInt(1, taskId);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            while (r.next())
            {
                jobs.add(populateJob(r));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);
        }

        return (Job []) jobs.toArray(new Job [0]);
    }

    synchronized void updateNextJobTreeReady(Job rootJob, int taskId)
        throws SQLException, JobControlException
    {
        int jobId = rootJob.getId();
        int low = rootJob.getLow();

        if (low == 0)
            throw new JobControlException("Job ["
                                          + jobId
                                          + "] is the goal for Task ["
                                          + taskId
                                          + "]");

        String sql = sqlStatements.getString("readNextRootJob");

        Connection c = getConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        Job              job = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, taskId);
            ps.setInt(2, jobId);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            if (r.next())
            {
                job = populateJob(r);
            }
            else
            {
                throw new JobControlException("Unable to update next tree after Job ["
                                              + jobId
                                              + "] in Task ["
                                              + taskId
                                              + "]");
            }

            updateLeafJobsReady(job, taskId);
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);
        }
    }

    /**
     * <code>updateLeafJobsReady</code> sets the leaf
     * <code>Job</code>s of a specific root <code>Job</code> to ready.
     *
     * @param rootJob a <code>Job</code>.
     * @param taskId an <code>int</code> ID of the <code>Task</code>.
     *
     * @exception JobControlException if an error occurs.
     * @exception SQLException if an error occurs.
     */
    synchronized void updateLeafJobsReady(Job rootJob, int taskId)
        throws SQLException, JobControlException
    {
         String sql = sqlStatements.getString("updateLeafJobsReady");

         Connection c = getConnection();
         PreparedStatement ps = null;

         try
         {
             cat.debug("Preparing SQL: " + sql);
             ps = c.prepareStatement(sql);
             ps.setInt(1, Status.READY);
             ps.setInt(2, taskId);
             ps.setInt(3, rootJob.getLow());
             ps.setInt(4, rootJob.getHigh());
             ps.setInt(5, Status.WAITING);
             cat.debug("Executing SQL: " + ps);
             ps.executeUpdate();
         }
         finally
         {
             cSupport.close(ps);
         }
    }

    synchronized void deleteJob(int jobId, int taskId)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("deleteJob");

        Connection c = getConnection();
        PreparedStatement us = null;

        try
        {
            List selected = lockJobTree(jobId, taskId);

            if (selected.size() > 0)
            {
                cat.debug("Preparing SQL: " + sql);
                us = c.prepareStatement(sql);

                for (Iterator si = selected.iterator(); si.hasNext();)
                {
                    Integer id = (Integer) si.next();
                    cat.info("Marking Job ["
                             + id
                             + "] for deletion");
                    us.setInt(1, id.intValue());
                    us.addBatch();
                }

                cat.debug("Executing SQL: " + us);
                us.executeBatch();

                deleteJobHosts(selected);
            }
        }
        finally
        {
            cSupport.close(us);
        }
    }

    /**
     * <code>lockJob</code> obtains a write lock on a
     * <code>Job</code>.
     *
     * @param jobId an <code>int</code> ID.
     *
     * @exception SQLException if an error occurs.
     * @exception JobControlException if an error occurs.
     */
    private void lockJob(int jobId) throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("selectJobForUpdateById");

        Connection c = getConnection();
        PreparedStatement ls = null;
        ResultSet          r = null;

        try
        {
            // Lock statement
            cat.debug("Preparing SQL: " + sql);
            ls = c.prepareStatement(sql);
            ls.setInt(1, jobId);
            cat.debug("Executing SQL: " + ls);
            r = ls.executeQuery();

            if (! r.next())
                throw new JobControlException("Failed to find Job ["
                                              + jobId
                                              + "] for update");
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ls);
        }
    }

    private List lockJobTree(int jobId, int taskId)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("selectTreeForUpdate");

        Connection c = getConnection();
        PreparedStatement ls = null;
        ResultSet          r = null;
        List selected = new ArrayList();

        try
        {
            // Lock statement
            cat.debug("Preparing SQL: " + sql);
            ls = c.prepareStatement(sql);
            ls.setInt(1, taskId);
            ls.setInt(2, taskId);
            ls.setInt(3, jobId);
            cat.debug("Executing SQL: " + ls);
            r = ls.executeQuery();

            while (r.next())
            {
                int id = r.getInt("job_id");
                selected.add(new Integer(id));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ls);
        }

        return selected;
    }

    private List lockChildJobsByStatus(int parentId, int taskId,
                                       int statusIdX, int statusIdY)
        throws SQLException
    {
        String sql = sqlStatements.getString("selectChildTreeForUpdateByStatus");

        Connection c = getConnection();
        PreparedStatement ls = null;
        ResultSet          r = null;
        List selected = new ArrayList();

        try
        {
            // Lock statement
            cat.debug("Preparing SQL: " + sql);
            ls = c.prepareStatement(sql);
            ls.setInt(1, taskId);
            ls.setInt(2, taskId);
            ls.setInt(3, parentId);
            ls.setInt(4, statusIdX);
            ls.setInt(5, statusIdY);
            cat.debug("Executing SQL: " + ls);
            r = ls.executeQuery();

            while (r.next())
            {
                int jobId = r.getInt("job_id");
                selected.add(new Integer(jobId));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ls);
        }

        return selected;
    }

    private List lockRestorableJobs(int jobId, int statusId, int taskId)
        throws SQLException
    {
        String sql = sqlStatements.getString("selectRestorableJobsForUpdate");

        Connection c = getConnection();
        PreparedStatement ls = null;
        ResultSet          r = null;
        List selected = new ArrayList();

        try
        {
            // Lock statement
            cat.debug("Preparing SQL: " + sql);
            ls = c.prepareStatement(sql);
            ls.setInt(1, taskId);
            ls.setInt(2, taskId);
            ls.setInt(3, jobId);
            ls.setInt(4, statusId);
            ls.setInt(5, 0);
            cat.debug("Executing SQL: " + ls);
            r = ls.executeQuery();

            while (r.next())
            {
                int id = r.getInt("job_id");
                selected.add(new Integer(id));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ls);
        }

        return selected;
    }

    /**
     * <code>updateRedundantChildJobs</code> finds all the children of
     * a <code>Job</code> which has completed and checks to see if
     * they are WAITING or READY and sets their status_id. Used for
     * SKIPPING or CANCELLING child Jobs.
     *
     * @param parentId an <code>int</code> ID.
     * @param newStatusId an <code>int</code> ID.
     * @param taskId an <code>int</code> ID.
     *
     * @exception SQLException if an error occurs.
     */
    private void updateRedundantChildJobs(int parentId,
                                          int newStatusId,
                                          int taskId)
        throws SQLException
    {
        if (! (newStatusId == Status.SKIPPED ||
               newStatusId == Status.CANCELLED))
            throw new IllegalArgumentException("New Status ID was not one of "
                                               + Status.SKIPPED
                                               + " or "
                                               + Status.CANCELLED);

        String sql = sqlStatements.getString("updateJobStatus");
  
        Connection c = getConnection();
        PreparedStatement us = null;

        try
        {
            List redundant = lockChildJobsByStatus(parentId, taskId,
                                                   Status.WAITING,
                                                   Status.READY);

            if (redundant.size() > 0)
            {
                cat.debug("Preparing SQL: " + sql);
                us = c.prepareStatement(sql);

                for (Iterator ri = redundant.iterator(); ri.hasNext();)
                {
                    Integer id = (Integer) ri.next();
                    cat.info("Marking Job ["
                             + id
                             + "] as redundant");

                    us.setInt(1, newStatusId);
                    us.setInt(2, id.intValue());
                    us.addBatch();
                }
                cat.debug("Executing SQL: " + us);
                us.executeBatch();
            }
        }
        finally
        {
            cSupport.close(us);
        }
    }

    /**
     * <code>populateJob</code> creates a new instance of
     * <code>Job</code>.
     *
     * @param r a <code>ResultSet</code> 
     *
     * @return a <code>Job</code>.
     *
     * @exception SQLException if an error occurs.
     * @exception JobControlException if an error occurs.
     */
    private Job populateJob(ResultSet r)
        throws SQLException, JobControlException
    {
        Connection c = getConnection();
        executableSQLDAO.setConnection(c);
        statusSQLDAO.setConnection(c);
        hostNameSQLDAO.setConnection(c);

        int    jobId = r.getInt("job_id");
        int parentId = r.getInt("parent_id");
        int   execId = r.getInt("exec_id");
        int statusId = r.getInt("status_id");

        int  low = r.getInt("low");
        int high = r.getInt("high");

        Timestamp      startTime = r.getTimestamp("start_time");
        Timestamp completionTime = r.getTimestamp("completion_time");

        String          operator = r.getString("operator");
        String         inputName = r.getString("input_name");
        String        outputName = r.getString("output_name");
        String         errorName = r.getString("error_name");
        String   commandTemplate = r.getString("cmd_templ");
        String workDirectoryName = r.getString("work_dir_name");
        String       environment = r.getString("environment");

        int maxRetryCount = r.getInt("max_retry_count");
        int    retryCount = r.getInt("retry_count");
        int     exitValue = r.getInt("exit_value");

        // Batch Jobs only
        int  batchId = r.getInt("batch_id");
        String queue = r.getString("queue");

        Executable executable =
            executableSQLDAO.readExecutableById(execId);
        Status status = statusSQLDAO.readStatusById(statusId);
        String [] hosts = hostNameSQLDAO.readHostNamesByJobId(jobId);

        Job job = null;

        if (queue != null)
        {
            job = new JobBatchImpl(operator, executable, status,
                                   startTime, completionTime,
                                   inputName, outputName, errorName,
                                   commandTemplate, workDirectoryName, environment,
                                   maxRetryCount, retryCount, exitValue,
                                   queue);
            ((BatchJob) job).setBatchId(batchId);
            ((BatchJob) job).setHosts(hosts);
        }
        else
        {
            job = new JobDefaultImpl(operator, executable, status,
                                     startTime, completionTime,
                                     inputName, outputName, errorName,
                                     commandTemplate, workDirectoryName, environment,
                                     maxRetryCount, retryCount, exitValue);
        }

        setObjectIntField(job, "id", jobId);

        job.setLow(low);
        job.setHigh(high);

        return job;
    }

    /**
     * <code>populateJobTrees</code> recreates trees of
     * <code>Job</code>s under a single <code>Job</code>. This method
     * is package private because client code should only be able to
     * create <code>Job</code> trees in the context of a
     * <code>Task</code> i.e. via a <code>TaskSQLDAO</code>.
     *
     * @param taskId an <code>int</code>.
     *
     * @return a <code>Job []</code> array.
     *
     * @exception SQLException if an error occurs.
     * @exception JobControlException if an error occurs.
     */
    synchronized Job [] populateJobTrees(int taskId)
        throws SQLException, JobControlException
    {
        // Map containing all possible parents keyed by their ID
        Map parentIdToParent = new HashMap();
        // Map containing all lists of children keyed by their parent
        // ID
        Map parentIdToChildren  = new HashMap();

        String sql = sqlStatements.getString("readJobsByTaskId");

        Connection c = getConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, taskId);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            while (r.next())
            {
                Job job = populateJob(r);
                parentIdToParent.put(new Integer(job.getId()), job);

                int parentId = r.getInt("parent_id");

                Integer pi = new Integer(parentId);
                if (parentIdToChildren.containsKey(pi))
                {
                    // If there is already a List, add to it
                    ((List) parentIdToChildren.get(pi)).add(job);
                }
                else
                {
                    // Otherwise make new List
                    List children = new ArrayList();
                    children.add(job);
                    parentIdToChildren.put(pi, children);
                }
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);
        }

        // This map simply collects references to root Jobs as we come
        // across them and keeps them in the correct order (by
        // ascending ID)
        Map rootJobs = new TreeMap();

        // Iterate through children by their parent ID, setting their
        // parent to the corresponding Job obtained from the parent
        // map
        for (Iterator ji = parentIdToChildren.keySet().iterator();
             ji.hasNext();)
        {
            Integer pi = (Integer) ji.next();

            Job parent = (Job) parentIdToParent.get(pi);
            if (parent == null)
            {
                // We found a root Job
                rootJobs.put(pi, ((List) parentIdToChildren.get(pi)).get(0));
            }
            else
            {
                List children = (List) parentIdToChildren.get(pi);
                for (Iterator ci = children.iterator(); ci.hasNext();)
                {
                    Job child = (Job) ci.next();
                    parent.add(child);
                }
            }
        }

        return (Job []) rootJobs.values().toArray(new Job[0]);
    }

    /**
     * <code>createNewJobId</code> gets a new unique ID from a
     * sequence.
     *
     * @return an <code>int</code> ID.
     *
     * @exception Exception if an error occurs.
     */
    private int createNewJobId() throws Exception
    {
        String sql = sqlStatements.getString("createNewJobId");

        Connection c = getConnection();
        Statement s = null;
        ResultSet r = null;
        int id = 0;

        try
        {
            s = c.createStatement();
            cat.debug("Executing SQL: " + s);
            r = s.executeQuery(sql);

            if (r.next())
                id = r.getInt("job_id");
            else
                throw new Exception("Failed to retrieve a new Job ID");
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(s);
        }

        return id;
    }

    /**
     * <code>commonChildStatus</code> returns true if all children of
     * a <code>Job</code> in the specified <code>Task</code> share the
     * same status_id.
     *
     * @param job a <code>Job</code> to test.
     * @param statusId an <code>int</code> ID.
     * @param taskId an <code>int</code> ID of the <code>Task</code>
     * to which the <code>Job</code> belongs.
     *
     * @return a <code>boolean</code>.
     *
     * @exception SQLException if an error occurs.
     */
    private boolean commonChildStatus(Job job, int statusId, int taskId)
        throws SQLException
    {
        String sql = sqlStatements.getString("readChildStatusSummaryById");

        Connection c = getConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;

        boolean allMatch = true;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, taskId);
            ps.setInt(2, job.getId());
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            while (r.next())
            {
                int childStatusId = r.getInt("status_id");
                if (childStatusId == statusId)
                {
                    continue;
                }
                else
                {
                    allMatch = false;
                    break;
                }
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);
        }

        return allMatch;
    }

    /**
     * <code>filterJobFailurePath</code> filters the <code>Job</code>
     * failure path by checking each node towards the root from the
     * failed <code>Job</code> to see whether it could possibly
     * succeed. This is defined by it being WAITING or READY, having
     * the OR operator and not having all its children FAILED. Such a
     * case breaks the failure path at that point. The path is also
     * broken when a STARTED Job is reached.
     *
     * @param path a <code>List</code> of <code>Job</code>s.
     * @param taskId an <code>int</code> ID.
     *
     * @return a <code>List</code> of <code>Job</code>s.
     *
     * @exception SQLException if an error occurs.
     * @exception JobControlException if an error occurs.
     */
    private List filterJobFailurePath(List path, int taskId)
        throws SQLException, JobControlException
    {
        List filtered = new ArrayList();
        boolean continguous = true;

    PATH:
        for (int i = 0; i < path.size(); i++)
        {
            Job         job = (Job) path.get(i);
            int       jobId = job.getId();
            String operator = job.getOperator();
            int    statusId = job.getStatus().getId();

        STATUS:
            switch (statusId)
            {
                case (Status.WAITING):
                case (Status.READY):
                    if (operator.equals(Job.OPERATOR_AND))
                    {
                        if (continguous)
                            filtered.add(job);
                    }
                    else if (operator.equals(Job.OPERATOR_OR))
                    {
                        if (commonChildStatus(job, Status.FAILED, taskId))
                            filtered.add(job);
                    }
                    break STATUS;

                case (Status.STARTED):
                    break PATH;

                case (Status.FAILED):
                case (Status.SKIPPED):
                case (Status.CANCELLED):
                    continguous = false;
                    break STATUS;

                default:
                    throw new JobControlException("Unknown Status ID '"
                                                  + statusId
                                                  + "'");
            }
        }

        cat.debug("Failure path to root from Job is now "
                  + filtered);

        return filtered;
    }

    /**
     * <code>updateJobHosts</code> updates the job_host link table
     * with the host(s) on which a <code>Job</code> was executed.
     *
     * @param job a <code>Job</code>.
     *
     * @exception SQLException if an error occurs.
     * @exception JobControlException if an error occurs.
     */
    private void updateJobHosts(Job job)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("updateJobHosts");

        Connection c = getConnection();
        hostNameSQLDAO.setConnection(c);

        PreparedStatement us = null;

        int jobId = job.getId();

        String [] hosts = ((JobBatchImpl) job).getHosts();

        try
        {
            cat.debug("Preparing SQL: " + sql);
            us = c.prepareStatement(sql);

            for (int i = 0; i < hosts.length; i++)
            {
                int hostId = hostNameSQLDAO.readHostIdByName(hosts[i]);

                us.setInt(1, jobId);
                us.setInt(2, hostId);
                us.addBatch();
            }

            cat.debug("Executing SQL: " + us);
            us.executeBatch();
        }
        finally
        {
            cSupport.close(us);
        }
    }

    /**
     * <code>deleteJobHosts</code> deletes from the job_host link
     * table when <code>Job</code>a are deleted.
     *
     * @param jobIds a <code>List</code> of IDs.
     *
     * @exception SQLException if an error occurs.
     */
    private void deleteJobHosts(List jobIds) throws SQLException
    {
        String sql = sqlStatements.getString("deleteJobHosts");

        Connection c = getConnection();
        
        PreparedStatement us = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            us = c.prepareStatement(sql);

            for (Iterator ji = jobIds.iterator(); ji.hasNext();)
            {
                Integer id = (Integer) ji.next();

                us.setInt(1, id.intValue());
                us.addBatch();
            }

            cat.debug("Executing SQL: " + us);
            us.executeBatch();
        }
        finally
        {
            cSupport.close(us);
        }
    }

    private Job _readJobById(int jobId)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("readJobById");

        Connection c = getConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        Job              job = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, jobId);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            if (r.next())
                job = populateJob(r);
            else
                throw new JobControlException("Requested a non-existent Job with ID '"
                                              + jobId
                                              + "'");
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);
        }

        return job;
    }

    private int _readTaskIdByJob(Job job)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("readTaskIdByJob");

        Connection c = getConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;

        int taskId = -1;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, job.getId());
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            if (r.next())
                taskId = r.getInt("task_id");
            else
                throw new JobControlException("Failed to find Job ["
                                              + job.getId()
                                              + "]");
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);
        }

        return taskId;
    }

    /**
     * <code>_readJobParentId</code> reads the job_id of a
     * <code>Job</code>'s parent. This uses a more efficient query
     * than the more elegant adjacency list view of the nested set
     * data. Requires the job table to be denormalized.
     *
     * @param job a <code>Job</code>
     *
     * @return an <code>int</code> ID or 0 if the <code>Job</code> has
     * no parent.
     *
     * @exception SQLException if an error occurs.
     */
    private int _readJobParentId(Job job)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("readJobParentId");

        Connection c = getConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;

        int parentId = 0;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, job.getId());
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            if (r.next())
                parentId = r.getInt("parent_id");
            else
                throw new JobControlException("Failed to find Job ["
                                              + job.getId()
                                              + "]");
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);
        }

        return parentId;
    }

    public List _readJobPath(Job job) throws SQLException, JobControlException
    {
        int jobId = job.getId();
        int   low = job.getLow();
        int  high = job.getHigh();

        int taskId = _readTaskIdByJob(job);

        String sql = sqlStatements.getString("readJobPath");

        Connection c = getConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;

        List path = new ArrayList();

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, taskId);
            ps.setInt(2, low);
            ps.setInt(3, high);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            int childId = jobId;

            while (r.next())
            {
                Job j = populateJob(r);
                path.add(j);
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);
        }

        cat.debug("Path to root from Job ["
                  + jobId
                  + "] is "
                  + path);

        return path;
    }

    private void _updateJobReady(Job job)
        throws SQLException, JobControlException
    {
        int jobId = job.getId();

        String sql = sqlStatements.getString("updateJobReadyById");

        Connection c = getConnection();
        PreparedStatement us = null;

        try
        {
            lockJob(jobId);

            // Update statement
            cat.debug("Preparing SQL: " + sql);
            us = c.prepareStatement(sql);
            us.setInt(1, Status.READY);
            us.setInt(2, jobId);
            cat.debug("Executing SQL: " + us);
            us.executeUpdate();
        }
        finally
        {
            cSupport.close(us);
        }
    }
}
