/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.dao;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ResourceBundle;
import javax.sql.DataSource;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.job.Owner;
import uk.ac.sanger.jcon.job.OwnerDefaultImpl;

/**
 * <code>OwnerSQLDAO</code> implements <code>OwnerDAO</code> for
 * persistence tasks using a relational database.
 *
 * @author Keith James
 */
public class OwnerSQLDAO extends SQLDAOBaseImpl implements OwnerDAO
{
    // Log4j setup
    static Category cat = Category.getInstance(OwnerSQLDAO.class.getName());

    // SQL statement resource
    static ResourceBundle sqlStatements =
        ResourceBundle.getBundle(OwnerSQLDAO.class.getName());

    // Map to cache instances, keyed by ID
    private HashMap canonical = new HashMap(50);

    /**
     * Creates a new <code>OwnerSQLDAO</code> with no
     * <code>DataSource</code>.
     */
    public OwnerSQLDAO()
    {
        super();
    }

    /**
     * Creates a new <code>OwnerSQLDAO</code> with the specified
     * <code>DataSource</code>.
     *
     * @param datasource a <code>DataSource</code>.
     */
    public OwnerSQLDAO(DataSource datasource)
    {
        super(datasource);
    }

    public synchronized void createOwner(Owner owner) throws Exception
    {
        String sql = sqlStatements.getString("createOwner");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        boolean rollback = false;

        try
        {
            int newId = createNewOwnerId();
            setObjectIntField(owner, "id", newId);

            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, owner.getId());
            ps.setString(2, owner.getUserName());
            ps.setString(3, owner.getFirstName());
            ps.setString(4, owner.getLastName());
            ps.setString(5, owner.getEMail());
            ps.setString(6, owner.getTelephone());
            cat.debug("Executing SQL: " + ps);
            ps.executeUpdate();
        }
        catch (Exception e)
        {
            cat.error("Caught an Exception; rolling back transaction", e);
            rollback = true;
            throw e;
        }
        finally
        {
            cSupport.close(ps);

            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    public synchronized Collection readAllOwners() throws SQLException
    {
        String sql = sqlStatements.getString("readAllOwners");

        Connection c = getManagedConnection();
        Statement s = null;
        ResultSet r = null;
        List owners = new ArrayList();
        
        try
        {
            cat.debug("Preparing SQL: " + sql);
            s = c.createStatement();
            cat.debug("Executing SQL: " + s);
            r = s.executeQuery(sql);

            while (r.next())
            {
                owners.add(populateOwner(r));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(s);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return owners;
    }

    public synchronized Owner readOwnerById(int ownerId)
        throws SQLException, JobControlException
    {
        // Test canonical instances first
        Integer canonKey = new Integer(ownerId);
        if (canonical.containsKey(canonKey))
        {
            cat.debug("Fetching canonical Owner [" + canonKey + "] from cache");
            return (Owner) canonical.get(canonKey);
        }

        String sql = sqlStatements.getString("readOwnerById");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        Owner          owner = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, ownerId);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            if (r.next())
            {
                owner = populateOwner(r);
                // Add to canonical instances
                canonical.put(canonKey, owner);
            }
            else
            {
                throw new JobControlException("Requested a non-existent Owner with ID '"
                                              + ownerId
                                              + "'");
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return owner;
    }

    public synchronized Owner readOwnerByUserName(String userName)
        throws SQLException, JobControlException
    {
        // Test canonical instances first
        for (Iterator oi = canonical.values().iterator(); oi.hasNext();)
        {
            Owner owner = (Owner) oi.next();
            if (userName.equals(owner.getUserName()))
            {
                cat.debug("Fetching canonical Owner ["
                          + owner.getId()
                          + "] from cache");
                return owner;
            }
        }

        String sql = sqlStatements.getString("readOwnerByUserName");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        Owner          owner = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setString(1, userName);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            if (r.next())
            {
                owner = populateOwner(r);
                // Add to canonical instances
                canonical.put(new Integer(owner.getId()), owner);
            }
            else
            {
                throw new JobControlException("Requested a non-existent Owner with user name '"
                                              + userName +
                                              "'");
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return owner;
    }

    public synchronized void deleteOwner(Owner owner) throws SQLException
    {
        String sql = sqlStatements.getString("deleteOwner");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        boolean rollback = false;

        int ownerId = owner.getId();

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, ownerId);
            cat.debug("Executing SQL: " + ps);
            ps.executeUpdate();
        }
        catch (SQLException se)
        {
            cat.error("Caught an Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        finally
        {
            cSupport.close(ps);

            if (rollback)
            {
                cSupport.rollback(c);
            }
            else
            {
                cSupport.commit(c);
                canonical.remove(new Integer(ownerId));
            }

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    /**
     * <code>populateOwner</code> creates a new instance of
     * <code>Owner</code>.
     *
     * @param r a <code>ResultSet</code>.
     *
     * @return an <code>Owner</code>.
     *
     * @exception SQLException if an error occurs.
     */
    private Owner populateOwner(ResultSet r) throws SQLException
    {
        Owner owner = new OwnerDefaultImpl(r.getString("first_name"),
                                           r.getString("last_name"),
                                           r.getString("user_name"),
                                           r.getString("email"),
                                           r.getString("telephone"));
        setObjectIntField(owner, "id", r.getInt("owner_id"));

        return owner;
    }

    /**
     * <code>createNewOwnerId</code> gets a new unique ID from a
     * sequence.
     *
     * @return an <code>int</code> ID.
     *
     * @exception Exception if an error occurs.
     */
    private int createNewOwnerId() throws Exception
    {
        String sql = sqlStatements.getString("createNewOwnerId");
 
        Connection c = getConnection();
        Statement s = null;
        ResultSet r = null;
        int id = 0;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            s = c.createStatement();
            cat.debug("Executing SQL: " + s);
            r = s.executeQuery(sql);

            if (r.next())
                id = r.getInt("owner_id");
            else
                throw new Exception("Failed to retrieve a new Owner ID");
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(s);
        }

        return id;
    }
}


