/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.dao;

import java.lang.reflect.Field;
import java.sql.Connection;
import java.sql.SQLException;
import javax.sql.DataSource;

import org.apache.log4j.Category;

import uk.ac.sanger.sql.ConnectionSupport;

/**
 * <p><code>SQLDAOBaseImpl</code> is the base class for data access
 * objects in this package.</p>
 *
 * @author Keith James
 */
public class SQLDAOBaseImpl
{
    // Log4j setup
    static Category cat = Category.getInstance(SQLDAOBaseImpl.class.getName());

    protected DataSource datasource;
    protected Connection connection;
    protected ConnectionSupport cSupport;

    /**
     * Creates a new <code>SQLDAOBaseImpl</code> with no
     * <code>DataSource</code>. Use of this constructor creates a DAO
     * which cannot create new <code>Connection</code>s and must rely
     * on another DAO to provide them. This is the constructor to use
     * when creating delegate DAOs whose <code>Connection</code>s must
     * be closely controlled as in the case where a transaction
     * involving a chain of several DAOs is being undertaken.
     */
    public SQLDAOBaseImpl()
    {
        this(null);
    }

    /**
     * Creates a new <code>SQLDAOBaseImpl</code> using the specified
     * <code>DataSource</code>. Use of this constructor creates a DAO
     * which will create new <code>Connection</code>s on demand and
     * will manage them for any delegate DAOs. This is the constructor
     * to use when creating a DAO which will undertake transactions
     * involving a chain of delegate DAOs.
     *
     * @param datasource a <code>DataSource</code>.
     */
    public SQLDAOBaseImpl(DataSource datasource)
    {
        this.datasource = datasource;
        cSupport = new ConnectionSupport();
    }

    /**
     * <code>getUser</code> returns the user to be mentioned in the
     * logs and error messages.
     *
     * @return a <code>String</code>.
     */
//     public String getUser()
//     {
//         return cSupport.getUser();
//     }

    /**
     * <code>setUser</code> sets the user to be mentioned in the logs
     * and error messages.
     *
     * @param user a <code>String</code>.
     */
//     public void setUser(String user)
//     {
//         cSupport.setUser(user);
//     }

    /**
     * <code>getConnection</code> returns the <code>Connection</code>
     * currently being used.
     *
     * @return a <code>Connection</code>.
     *
     * @exception SQLException if the <code>Connection</code> is null
     * or has been closed.
     */
    public synchronized Connection getConnection() throws SQLException
    {
        if (connection == null)
            throw new SQLException("Failed to get current connection as it was null");
        else if (connection.isClosed())
            throw new SQLException("Failed to get current connection as it was closed");

        cat.debug("... Getting current connection " + connection);
        return connection;
    }

    /**
     * <code>setConnection</code> sets the <code>Connection</code>
     * currently being used. This is the only way in which DAOs with
     * no <code>DataSource</code> can obtain a <code>Connection</code>.
     *
     * @param connection a <code>Connection</code>.
     *
     * @exception SQLException if the <code>Connection</code> is null
     * or has been closed.
     */
    public synchronized void setConnection(Connection connection)
        throws SQLException
    {
        if (connection == null)
            throw new SQLException("Failed to set current connection as it was null");
        else if (connection.isClosed())
            throw new SQLException("Failed to set current connection as it was closed");

        cat.debug("=== Setting current connection to " + connection);
        this.connection = connection;
    }

    /**
     * <code>isManagingConnections</code> returns true if the DAO is
     * managing its own <code>Connection</code> lifecycles. This is
     * true only for those which have a <code>DataSource</code>.
     *
     * @return a <code>boolean</code>.
     */
    public boolean isManagingConnections()
    {
        return datasource != null;
    }

    /**
     * <code>getManagedConnection</code> returns a
     * <code>Connection</code> which is being managed by a DAO.
     *
     * @return a <code>Connection</code>.
     *
     * @exception SQLException if an error occurs.
     */
    protected Connection getManagedConnection() throws SQLException
    {
        Connection c = null;
        if (isManagingConnections())
        {
            c = datasource.getConnection();
            cat.debug("+++ Got a new connection " + c);
            c.setAutoCommit(false);
            cat.info(">>> Started transaction with " + c);
            setConnection(c);
        }
        else
        {
            c = getConnection();
        }

        return c;
    }

    protected void setObjectIntField(Object object, String fieldName, int value)
    {
        try
        {
            Field field = object.getClass().getDeclaredField(fieldName);
            field.setAccessible(true);
            field.set(object, new Integer(value));
            field.setAccessible(false);
        }
        catch (NoSuchFieldException nsfe)
        {
            throw new Error("Failed to set object field value: "
                            + nsfe.getMessage());
        }
        catch (IllegalAccessException iae)
        {
            throw new Error("Failed to set object field value: "
                            + iae.getMessage());
        }
    }
}
