/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.dao;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.ResourceBundle;
import javax.sql.DataSource;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.job.StatusDefaultImpl;

/**
 * <code>StatusSQLDAO</code> implements <code>StatusDAO</code> for
 * persistence tasks using a relational database.
 *
 * @author Keith James
 */
public class StatusSQLDAO extends SQLDAOBaseImpl implements StatusDAO
{
    // Log4j setup
    static Category cat = Category.getInstance(StatusSQLDAO.class.getName());

    // SQL statement resource
    static ResourceBundle sqlStatements =
            ResourceBundle.getBundle(StatusSQLDAO.class.getName());

    // Map to cache instances, keyed by ID
    private HashMap canonical = new HashMap(10);

    /**
     * Creates a new <code>StatusSQLDAO</code> with no
     * <code>DataSource</code>.
     */
    public StatusSQLDAO()
    {
        super();
    }

    /**
     * Creates a new <code>StatusSQLDAO</code> with the specified
     * <code>DataSource</code>.
     *
     * @param datasource a <code>DataSource</code>.
     */
    public StatusSQLDAO(DataSource datasource)
    {
        super(datasource);
    }

    public synchronized Collection readAllStatus() throws SQLException
    {
        String sql = sqlStatements.getString("readAllStatus");

        Connection c = getManagedConnection();
        Statement s = null;
        ResultSet r = null;
        List status = new ArrayList();

        try
        {
            cat.debug("Preparing SQL: " + sql);
            s = c.createStatement();
            cat.debug("Executing SQL: " + s);
            r = s.executeQuery(sql);

            while (r.next())
            {
                status.add(populateStatus(r));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(s);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return status;
    }

    public synchronized Status readStatusById(int statusId)
        throws SQLException, JobControlException
    {
        // Test canonical instances first
        Integer canonKey = new Integer(statusId);
        if (canonical.containsKey(canonKey))
        {
            cat.debug("Fetching canonical Status ["
                      + canonKey
                      + "] from cache");
            return (Status) canonical.get(canonKey);
        }

        String sql = sqlStatements.getString("readStatusById");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        Status        status = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, statusId);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            if (r.next())
            {
                status = populateStatus(r);
                // Add to canonical instances
                canonical.put(canonKey, status);
            }
            else
            {
                throw new JobControlException("Requested a non-existent Status with ID '"
                                              + statusId
                                              + "'");
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return status;
    }

    /**
     * <code>populateStatus</code> creates a new instance of
     * <code>Status</code>.
     *
     * @param r a <code>ResultSet</code>.
     *
     * @return a <code>Status</code>.
     *
     * @exception SQLException if an error occurs.
     */
    private Status populateStatus(ResultSet r) throws SQLException
    {
        Status status = new StatusDefaultImpl(r.getString("description"));
        setObjectIntField(status, "id", r.getInt("status_id"));

        return status;
    }
}
