/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.dao;

import java.sql.Timestamp;
import java.util.Collection;

import uk.ac.sanger.jcon.job.Owner;
import uk.ac.sanger.jcon.job.Task;

/**
 * <code>TaskDAO</code> describes data access objects for persistence
 * of <code>Task</code>s. Possible stores are relational databases or
 * XML files.
 *
 * @author Keith James
 */
public interface TaskDAO
{
    /**
     * <code>INCLUDE_JOB_TREE</code> is a constant used to indicate
     * that an operation on a <code>Task</code> should include the
     * full job tree.
     */
    public static final int INCLUDE_JOB_TREE = 0;

    /**
     * <code>EXCLUDE_JOB_TREE</code> is a constant used to indicate
     * that an operation on a <code>Task</code> should not include the
     * full job tree, but just the root job.
     */
    public static final int EXCLUDE_JOB_TREE = 1;

    /**
     * <code>createTask</code> creates a new <code>Task</code> in the
     * store.
     *
     * @param task a <code>Task</code>.
     *
     * @return an <code>int</code> the ID of the new
     * <code>Task</code>.
     *
     * @exception Exception if an error occurs.
     */
    public int createTask(Task task) throws Exception;

   /**
     * <code>readTaskById</code> reads a single <code>Task</code> from
     * the store. The object instantiated will contain either a full
     * <code>Job</code> tree, or just the root <code>Job</code>, as
     * specified by the <code>treeOperation</code> argument.
     *
     * @param taskId an <code>int</code> ID.
     * @param treeOperation an <code>int</code> indicating whether the
     * full <code>Job</code> tree should be instantiated. Acceptable
     * values are <code>INCLUDE_JOB_TREE</code> or
     * <code>EXCLUDE_JOB_TREE</code>.
     *
     * @return a <code>Task</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Task readTaskById(int taskId, int treeOperation) throws Exception;

    /**
     * <code>readAllTasks</code> reads all <code>Task</code>s from the
     * store.
     *
     * @param treeOperation an <code>int</code> indicating whether the
     * full <code>Job</code> tree should be instantiated. Acceptable
     * values are <code>INCLUDE_JOB_TREE</code> or
     * <code>EXCLUDE_JOB_TREE</code>.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection readAllTasks(int treeOperation) throws Exception;

    /**
     * <code>readSubmittedTasks</code> reads all <code>Task</code>s
     * from the store which have been submitted, but which have not
     * yet completed.
     *
     * @param treeOperation an <code>int</code> an <code>int</code>
     * indicating whether the full <code>Job</code> tree should be
     * instantiated. Acceptable values are
     * <code>INCLUDE_JOB_TREE</code> or <code>EXCLUDE_JOB_TREE</code>.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs
     */
    public Collection readSubmittedTasks(int treeOperation) throws Exception;

    /**
     * <code>readTasksByOwner</code> reads all <code>Task</code>s from
     * the store belonging to the specified <code>Owner</code>.
     *
     * @param owner an <code>Owner</code>.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection readTasksByOwner(Owner owner) throws Exception;

    /**
     * <code>readTasksByName</code> reads all <code>Task</code>s from
     * the store with the specified <code>name</code>. This uses the
     * SQL 'LIKE' pattern matching.
     *
     * @param name a <code>String</code>.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection readTasksByName(String name) throws Exception;

    /**
     * <code>readTasksByPriority</code> reads all <code>Task</code>s from
     * the store with the specified priority.
     *
     * @param priority an <code>int</code>.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection readTasksByPriority(int priority) throws Exception;

    /**
     * <code>readTasksSubmittedBefore</code> reads all
     * <code>Task</code>s from the store which were submitted before
     * the specified time.
     *
     * @param time a <code>Timestamp</code>.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection readTasksSubmittedBefore(Timestamp time)
        throws Exception;

    /**
     * <code>readTasksSubmittedAfter</code> reads all
     * <code>Task</code>s from the store which were submitted after
     * the specified time.
     *
     * @param time a <code>Timestamp</code>.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection readTasksSubmittedAfter(Timestamp time)
        throws Exception;

    /**
     * <code>readTasksSubmittedBetween</code> reads all
     * <code>Task</code>s from the store which were submitted between
     * the specified times.
     *
     * @param earliest a <code>Timestamp</code>.
     * @param latest a <code>Timestamp</code>.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection readTasksSubmittedBetween(Timestamp earliest,
                                                Timestamp latest)
        throws Exception;

    /**
     * <code>updateTask</code> updates a <code>Task</code>, but does
     * not traverse any <code>Job</code> tree. The submission time
     * cannot be changed by this method.
     *
     * @param task a <code>Task</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void updateTask(Task task) throws Exception;

    /**
     * <code>updateTaskSubmitted</code> updates a <code>Task</code> as
     * having been submitted for running. A <code>Task</code> must be
     * created in the persistent store before this method is called.
     *
     * @param taskId an <code>int</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void updateTaskSubmitted(int taskId) throws Exception;

    public void updateTaskRestored(int taskId) throws Exception;

    /**
     * <code>updateTaskCancelled</code> cancels an existing
     * <code>Task</code>.
     *
     * @param taskId an <code>int</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void updateTaskCancelled(int taskId) throws Exception;

    /**
     * <code>deleteTask</code> deletes a <code>Task</code> from the
     * store.
     *
     * @param taskId an <code>int</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void deleteTask(int taskId) throws Exception;
}
