/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.dao;

import java.sql.*;
import java.util.*;
import javax.sql.DataSource;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.job.*;

/**
 * <p><code>TaskSQLDAO</code> implements <code>TaskDAO</code> for
 * persistence tasks using a relational database. This class also
 * creates its own <code>Job</code> instances in the store.</p>
 *
 * <p>This is a naive implementation which always hits the
 * database. If we find this a problem, then caching will be
 * added.</p>
 *
 * @author Keith James
 */
public class TaskSQLDAO extends SQLDAOBaseImpl implements TaskDAO
{
    // Log4j setup
    static Category cat = Category.getInstance(TaskSQLDAO.class.getName());

    // SQL statement resource
    static ResourceBundle sqlStatements =
            ResourceBundle.getBundle(TaskSQLDAO.class.getName());

    private OwnerSQLDAO           ownerSQLDAO;
    private ExecutableSQLDAO executableSQLDAO;
    private JobSQLDAO               jobSQLDAO;

    /**
     * Creates a new <code>TaskSQLDAO</code> with the specified
     * <code>DataSource</code>.
     */
    public TaskSQLDAO()
    {
        super();
        ownerSQLDAO = new OwnerSQLDAO();
        jobSQLDAO   = new JobSQLDAO();
    }

    /**
     * Creates a new <code>TaskSQLDAO</code> with the specified
     * <code>DataSource</code>.
     *
     * @param datasource a <code>DataSource</code>.
     */
    public TaskSQLDAO(DataSource datasource)
    {
        super(datasource);
        ownerSQLDAO = new OwnerSQLDAO();
        jobSQLDAO   = new JobSQLDAO();
    }

    public synchronized int createTask(Task task) throws Exception
    {
        String sql = sqlStatements.getString("createTask");

        Connection c = getManagedConnection();
        jobSQLDAO.setConnection(c);

        PreparedStatement ps = null;
        int newId = 0;
        boolean rollback = false;

        try
        {
            newId = createNewTaskId();
            setObjectIntField(task, "id", newId);

            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, task.getId());
            ps.setString(2, task.getName());
            ps.setString(3, task.getDescription());
            ps.setInt(4, task.getOwner().getId());
            ps.setInt(5, task.getPriority());
            ps.setInt(6, 0);
            ps.setNull(7, Types.TIMESTAMP);
            cat.debug("Executing SQL: " + ps);
            ps.executeUpdate();

            // The root jobs
            Job [] rootJobs = task.getJobs();

            // Ordering preserved by low/high values across root
            // jobs. Each root job has a low value one greater than
            // the previous one. The first root job has a low of 0.
            int  low = 0;
            int high = 0;

            for (int i = 0; i < rootJobs.length; i++)
            {
                high = jobSQLDAO.createJob(rootJobs[i], task.getId(), low);
                low = ++high;
            }
        }
        catch (Exception e)
        {
            cat.error("Caught Exception; rolling back transaction", e);
            rollback = true;
            throw e;
        }
        finally
        {
            cSupport.close(ps);

            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return newId;
    }

    public synchronized Task readTaskById(int taskId, int treeOperation)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("readTaskById");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        Task            task = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, taskId);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            if (r.next())
            {
                switch (treeOperation)
                {
                    case (TaskDAO.INCLUDE_JOB_TREE):
                        task = populateTask(r);
                        break;

                    case (TaskDAO.EXCLUDE_JOB_TREE):
                        task = populateTaskLite(r);
                        break;

                    default:
                        throw new IllegalArgumentException("Invalid job treeOperation flag "
                                                           + treeOperation);
                }
            }
            else
            {
                throw new JobControlException("Requested a non-existent Task "
                                              + "with ID '"
                                              + taskId
                                              + "'");
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return task;
    }

    public synchronized Collection readAllTasks(int treeOperation)
        throws Exception
    {
        String sql = sqlStatements.getString("readAllTasks");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        Collection     tasks = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);

            tasks = processCollectionQuery(ps, treeOperation);
        }
        finally
        {
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return tasks;
    }

    public synchronized Collection readSubmittedTasks(int treeOperation)
        throws Exception
    {
        String sql = sqlStatements.getString("readSubmittedTasks");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        Collection     tasks = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);

            tasks = processCollectionQuery(ps, treeOperation);
        }
        finally
        {
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return tasks;
    }

    public synchronized Collection readTasksByOwner(Owner owner)
        throws Exception
    {
        String sql = sqlStatements.getString("readTasksByOwner");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        Collection     tasks = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, owner.getId());
            tasks = processCollectionQuery(ps, TaskDAO.INCLUDE_JOB_TREE);
        }
        finally
        {
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return tasks;
    }

    public synchronized Collection readTasksByName(String name)
        throws Exception
    {
        String sql = sqlStatements.getString("readTasksByName");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        Collection     tasks = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setString(1, name);
            tasks = processCollectionQuery(ps, TaskDAO.INCLUDE_JOB_TREE);
        }
        finally
        {
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return tasks;
    }

    public synchronized Collection readTasksByPriority(int priority)
        throws Exception
    {
        String sql = sqlStatements.getString("readTasksByPriority");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        Collection     tasks = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, priority);
            tasks = processCollectionQuery(ps, TaskDAO.INCLUDE_JOB_TREE);
        }
        finally
        {
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return tasks;
    }

    public synchronized Collection readTasksSubmittedBefore(Timestamp time)
        throws Exception
    {
        String sql = sqlStatements.getString("readTasksSubmittedBefore");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        Collection     tasks = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setTimestamp(1, time);
            tasks = processCollectionQuery(ps, TaskDAO.INCLUDE_JOB_TREE);
        }
        finally
        {
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return tasks;
    }

    public synchronized Collection readTasksSubmittedAfter(Timestamp time)
        throws Exception
    {
        String sql = sqlStatements.getString("readTasksSubmittedAfter");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        Collection     tasks = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setTimestamp(1, time);
            tasks = processCollectionQuery(ps, TaskDAO.INCLUDE_JOB_TREE);
        }
        finally
        {
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return tasks;
    }

    public synchronized Collection readTasksSubmittedBetween(Timestamp earliest,
                                                             Timestamp latest)
        throws Exception
    {
        String sql = sqlStatements.getString("readTasksSubmittedBetween");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        Collection     tasks = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setTimestamp(1, earliest);
            ps.setTimestamp(2, latest);
            tasks = processCollectionQuery(ps, TaskDAO.INCLUDE_JOB_TREE);
        }
        finally
        {
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return tasks;
    }

    /**
     * <code>updateTask</code> updates a <code>Task</code> with new
     * values. The values updated are description and priority.
     *
     * @param task a <code>Task</code>.
     *
     * @exception Exception if an error occurs.
     */
    public synchronized void updateTask(Task task) throws Exception
    {
        int         taskId = task.getId();
        String description = task.getDescription();
        int       priority = task.getPriority();

        String sql = sqlStatements.getString("updateTaskById");

        Connection c = getManagedConnection();
        PreparedStatement us = null;
        boolean rollback = false;

        try
        {
            lockTask(taskId);

            // Update statement
            cat.debug("Preparing SQL: " + sql);
            us = c.prepareStatement(sql);
            us.setString(1, description);
            us.setInt(2, priority);
            us.setInt(3, taskId);
            cat.debug("Executing SQL: " + us);
            us.executeUpdate();
        }
        catch (Exception e)
        {
            cat.error("Caught Exception; rolling back transaction", e);
            rollback = true;
            throw e;
        }
        finally
        {
            cSupport.close(us);

            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    public synchronized void updateTaskSubmitted(int taskId)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("updateTaskSubmittedById");

        Connection c = getManagedConnection();
        jobSQLDAO.setConnection(c);

        PreparedStatement ps = null;
        boolean rollback = false;

        try
        {
            int submitted = lockTask(taskId);

            // Assert not previously submitted
            if (submitted == -1)
                throw new JobControlException("Failed to submit task ["
                                              + taskId
                                              + "] as its current submission "
                                              + "status could not be determined");

            if (submitted == 1)
                throw new JobControlException("Failed to submit task ["
                                              + taskId
                                              + "] as it has already been submitted");

            Task task = _readTaskById(taskId, TaskDAO.EXCLUDE_JOB_TREE);

            // Update statement
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, taskId);
            cat.debug("Executing SQL: " + ps);
            ps.executeUpdate();
            cSupport.close(ps);

            // Update leaf Jobs READY
            Job [] rootJobs = task.getJobs();
            int lastIndex = rootJobs.length - 1;

            jobSQLDAO.updateLeafJobsReady(rootJobs[lastIndex], taskId);
        }
        catch (SQLException se)
        {
            cat.error("Caught Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        finally
        {
            cSupport.close(ps);

            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    public synchronized void updateTaskRestored(int taskId) throws Exception
    {
        Connection c = getManagedConnection();
        jobSQLDAO.setConnection(c);

        PreparedStatement ps = null;
        boolean rollback = false;

        try
        {
            lockTask(taskId);
            Task task = _readTaskById(taskId, TaskDAO.EXCLUDE_JOB_TREE);

            Job [] jobs = task.getJobs();

            for (int i = jobs.length; --i >= 0;)
                jobSQLDAO.updateJobRestore(jobs[i]);
        }
        catch (SQLException se)
        {
            cat.error("Caught Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        finally
        {
            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    public synchronized void updateTaskCancelled(int taskId) throws Exception
    {
        Connection c = getManagedConnection();
        jobSQLDAO.setConnection(c);

        PreparedStatement ps = null;
        boolean rollback = false;

        try
        {
            lockTask(taskId);
            Task task = _readTaskById(taskId, TaskDAO.EXCLUDE_JOB_TREE);
            Job [] jobs = task.getJobs();

            for (int i = jobs.length; --i >= 0;)
                jobSQLDAO.updateJobCancelled(jobs[i]);
         }
        catch (SQLException se)
        {
            cat.error("Caught Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        finally
        {
            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    public synchronized void deleteTask(int taskId) throws Exception
    {
        String sql = sqlStatements.getString("deleteTask");

        Connection c = getManagedConnection();
        jobSQLDAO.setConnection(c);

        PreparedStatement ps = null;
        boolean rollback = false;

        try
        {
            lockTask(taskId);

            Task task = _readTaskById(taskId, TaskDAO.EXCLUDE_JOB_TREE);

            Job [] jobs = task.getJobs();
            for (int i = jobs.length; --i >= 0;)
                jobSQLDAO.deleteJob(jobs[i].getId(), taskId);

            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, task.getId());
            ps.executeUpdate();
        }
        catch (SQLException se)
        {
            cat.error("Caught Exception; rolling back transaction", se);
            rollback = true;
            throw se;
        }
        finally
        {
            cSupport.close(ps);

            if (rollback)
                cSupport.rollback(c);
            else
                cSupport.commit(c);

            if (isManagingConnections())
                cSupport.close(c);
        }
    }

    private List processCollectionQuery(PreparedStatement ps, int treeOperation)
        throws Exception
    {
        ResultSet r = null;
        List tasks = new ArrayList();

        try
        {
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            switch (treeOperation)
            {
                case (TaskDAO.INCLUDE_JOB_TREE):
                    while (r.next())
                    {
                        tasks.add(populateTask(r));
                    }
                    break;

                case (TaskDAO.EXCLUDE_JOB_TREE):
                    while (r.next())
                    {
                        tasks.add(populateTaskLite(r));
                    }
                    break;

                default:
                    throw new IllegalArgumentException("Invalid job treeOperation flag " + treeOperation);
            }
        }
        finally
        {
            cSupport.close(r);
        }

        return tasks;
    }

    private int lockTask(int taskId)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("selectTaskForUpdateById");

        Connection c = getConnection();
        PreparedStatement ls = null;
        ResultSet          r = null;
        int        submitted = -1;

        try
        {
            // Lock statement
            ls = c.prepareStatement(sql);
            ls.setInt(1, taskId);
            cat.debug("Executing SQL: " + ls);
            r = ls.executeQuery();

            if (r.next())
            {
                submitted = r.getInt("submitted");
            }
            else
            {
                throw new JobControlException("Failed to find Task ["
                                              + taskId
                                              + "] for update");
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ls);
        }

        return submitted;
    }

    /**
     * <code>populateTask</code> creates a new instance of
     * <code>Task</code>.
     *
     * @param r a <code>ResultSet</code>.
     *
     * @return a <code>Task</code>.
     *
     * @exception SQLException if an error occurs.
     * @exception JobControlException if an error occurs.
     */
    private Task populateTask(ResultSet r)
        throws SQLException, JobControlException
    {
        int taskId =  r.getInt("task_id");
        String name = r.getString("name");
        String description = r.getString("description");
        int ownerId = r.getInt("owner_id");
        int priority = r.getInt("priority");
        Timestamp submissionTime = r.getTimestamp("submission_time");

        Connection c = getConnection();
        ownerSQLDAO.setConnection(c);
        jobSQLDAO.setConnection(c);

        Owner owner = ownerSQLDAO.readOwnerById(ownerId);
        Job [] jobs = jobSQLDAO.populateJobTrees(taskId);

        cat.debug("Creating Task object with ID:"
                  + taskId
                  + " name:"
                  + name
                  + " desc: "
                  + description
                  + " owner: "
                  + owner
                  + " # root jobs: "
                  + jobs.length);

        Task task = new TaskDefaultImpl(name, description, owner, jobs);
        task.setPriority(priority);
        task.setSubmissionTime(submissionTime);
        setObjectIntField(task, "id", taskId);

        return task;
    }

    /**
     * <code>populateTaskLite</code> creates a new instance of
     * <code>Task</code> with its root <code>Job</code>.
     *
     * @param r a <code>ResultSet</code>.
     *
     * @return a <code>Task</code> with its root <code>Job</code>s.
     *
     * @exception SQLException if an error occurs.
     * @exception JobControlException if an error occurs.
     */
    private Task populateTaskLite(ResultSet r)
        throws SQLException, JobControlException
    {
        int taskId =  r.getInt("task_id");
        String name = r.getString("name");
        String description = r.getString("description");
        int ownerId = r.getInt("owner_id");
        int priority = r.getInt("priority");
        Timestamp submissionTime = r.getTimestamp("submission_time");

        Connection c = getConnection();
        ownerSQLDAO.setConnection(c);
        jobSQLDAO.setConnection(c);

        Owner owner = ownerSQLDAO.readOwnerById(ownerId);

        // These are already ordered by ID
        Job [] jobs = jobSQLDAO.readRootJobsByTaskId(taskId);

        cat.debug("Creating Task with ID: "
                  + taskId
                  + " name:"
                  + name
                  + " desc: "
                  + description
                  + " owner: "
                  + owner
                  + " # root jobs: "
                  + jobs.length);

        Task task = new TaskDefaultImpl(name, description, owner, jobs);
        task.setPriority(priority);
        task.setSubmissionTime(submissionTime);
        setObjectIntField(task, "id", taskId);

        return task;
    }

    /**
     * <code>createNewTaskId</code> gets a new unique ID from a
     * sequence.
     *
     * @return an <code>int</code> ID.
     *
     * @exception Exception if an error occurs.
     */
    private int createNewTaskId() throws Exception
    {
        String sql = sqlStatements.getString("createNewTaskId");

        Connection c = getConnection();
        Statement s = null;
        ResultSet r = null;
        int id = 0;

        try
        {
            s = c.createStatement();
            cat.debug("Executing SQL: " + s);
            r = s.executeQuery(sql);

            if (r.next())
                id = r.getInt("task_id");
            else
                throw new Exception("Failed to retrieve a new Task ID");
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(s);
        }

        return id;
    }

	private Task _readTaskById(int taskId, int treeOperation)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("readTaskById");

        Connection c = getConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        Task            task = null;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, taskId);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            if (r.next())
            {
                switch (treeOperation)
                {
                    case (TaskDAO.INCLUDE_JOB_TREE):
                        task = populateTask(r);
                        break;

                    case (TaskDAO.EXCLUDE_JOB_TREE):
                        task = populateTaskLite(r);
                        break;

                    default:
                        throw new IllegalArgumentException("Invalid job treeOperation flag " + treeOperation);
                }
            }
            else
            {
                cat.warn("Requested a non-existent Task with ID '" + taskId + "'");
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);
        }

        return task;
    }
}
