/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.dao;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.ResourceBundle;
import javax.sql.DataSource;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.job.Owner;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.job.TaskState;

/**
 * <code>TaskStatisticSQLDAO</code> implements
 * <code>TaskStatisticDAO</code> for persistence tasks using a
 * relational database.
 *
 * @author Keith James
 */
public class TaskStatisticSQLDAO extends SQLDAOBaseImpl
    implements TaskStatisticDAO
{
    // Log4j setup
    static Category cat =
        Category.getInstance(TaskStatisticSQLDAO.class.getName());

    // SQL statement resource
    static ResourceBundle sqlStatements =
            ResourceBundle.getBundle(TaskStatisticSQLDAO.class.getName());

    private StatusSQLDAO statusSQLDAO;

    /**
     * Creates a new <code>TaskStatisticSQLDAO</code> with no
     * <code>DataSource</code>.
     */
    public TaskStatisticSQLDAO()
    {
        super();
        statusSQLDAO = new StatusSQLDAO();
    }

    /**
     * Creates a new <code>TaskStatisticSQLDAO</code> with the
     * specified <code>DataSource</code>.
     *
     * @param datasource a <code>DataSource</code>.
     */
    public TaskStatisticSQLDAO(DataSource datasource)
    {
        super(datasource);
        statusSQLDAO = new StatusSQLDAO();
    }

    public synchronized int [] readSubmittedTaskId() throws SQLException
    {
        String sql = sqlStatements.getString("readSubmittedTaskId");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;

        List taskIds = new ArrayList();
        int taskId = 0;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            cat.debug("Executing SQL: " + sql);
            r = ps.executeQuery();

            while (r.next())
            {
                taskId = r.getInt("task_id");
                cat.debug("Submitted Task ID is " + taskId);
                taskIds.add(new Integer(taskId));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        int [] submittedTaskIds = new int [taskIds.size()];

        for (int i = 0;  i < submittedTaskIds.length; i++)
        {
            submittedTaskIds[i] = ((Integer) taskIds.get(i)).intValue();
        }

        return submittedTaskIds;
    }

    public synchronized int [] readNSubmittedTaskId(int n)
        throws SQLException
    {
        String sql = sqlStatements.getString("readNSubmittedTaskId");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;

        int [] taskIds = new int[n];
        int rowCount = 0;

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            cat.debug("Executing SQL: " + sql);
            r = ps.executeQuery();

            while (r.next() && (rowCount < n))
            {
                int taskId = r.getInt("task_id");
                taskIds[rowCount] = taskId;
                ++rowCount;
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        // If less than n Task IDs were found, shrink the array to fit
        if (rowCount < n)
        {
            int [] actualIds = new int [rowCount];
            System.arraycopy(taskIds, 0, actualIds, 0, rowCount);
            taskIds = actualIds;
        }

        return taskIds;
    }

    public synchronized int [] readSubmittedTaskIdByOwner(int ownerId)
        throws SQLException
    {
        String sql = sqlStatements.getString("readSubmittedTaskIdByOwner");

        Connection c = getManagedConnection();
        PreparedStatement ps = null;
        ResultSet          r = null;
        List taskIds = new ArrayList();

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, ownerId);
            cat.debug("Executing SQL: " + sql);
            r = ps.executeQuery();

            while (r.next())
            {
                taskIds.add(new Integer(r.getInt("task_id")));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        int [] submittedTaskIds = new int [taskIds.size()];

        for (int i = 0; i < submittedTaskIds.length; i++)
        {
            submittedTaskIds[i] = ((Integer) taskIds.get(i)).intValue();
        }

        return submittedTaskIds;
    }

    public synchronized Collection readSubmittedTaskState()
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("readSubmittedTaskState");

        Connection c = getManagedConnection();
        statusSQLDAO.setConnection(c);

        PreparedStatement ps = null;
        ResultSet          r = null;
        Status        status = null;
        Map tStats = new HashMap();

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            while (r.next())
            {
                Integer taskId = new Integer(r.getInt("task_id"));

                cat.debug("Got statistic for Task ["
                          + taskId.toString()
                          + "]");

                TaskState tStat = null;
                if (tStats.containsKey(taskId))
                {
                    tStat = (TaskState) tStats.get(taskId);
                }
                else
                {
                    tStat = new TaskState(taskId.intValue());
                    
                    cat.debug("Created TaskState ["
                              + taskId.intValue()
                              + "]");

                    tStats.put(taskId, tStat);
                }

                status = statusSQLDAO.readStatusById(r.getInt("status_id"));
                tStat.setJobCount(status, r.getInt("num_jobs"));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return tStats.values();
    }

    public synchronized TaskState readTaskState(int taskId)
        throws SQLException, JobControlException
    {
        String sql = sqlStatements.getString("readTaskState");

        Connection c = getManagedConnection();
        statusSQLDAO.setConnection(c);

        PreparedStatement ps = null;
        ResultSet          r = null;
        Status status        = null;
        TaskState tStat = new TaskState(taskId);

        try
        {
            cat.debug("Preparing SQL: " + sql);
            ps = c.prepareStatement(sql);
            ps.setInt(1, taskId);
            cat.debug("Executing SQL: " + ps);
            r = ps.executeQuery();

            while (r.next())
            {
                status = statusSQLDAO.readStatusById(r.getInt("status_id"));
                tStat.setJobCount(status, r.getInt("num_jobs"));
            }
        }
        finally
        {
            cSupport.close(r);
            cSupport.close(ps);

            if (isManagingConnections())
                cSupport.close(c);
        }

        return tStat;
    }
}
