/* TaskViewer.java
 *
 * created: Tue Aug 12 2003
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 2003  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/pathsoft/repository/psu/jobcontrol/src/uk/ac/sanger/jcon/gui/TaskViewer.java,v 1.1 2004/06/17 14:14:24 kmr Exp $
 */

package uk.ac.sanger.jcon.gui;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.sql.DataSource;

import uk.ac.sanger.jcon.dao.*;
import uk.ac.sanger.jcon.job.Owner;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.job.Task;
import uk.ac.sanger.jcon.job.TaskState;

import java.awt.*;
import javax.swing.*;
import java.util.*;

/**
 *  A component for viewing the status of jobs and tasks from the PSU
 *  jobcontrol system.
 *
 *  @author Kim Rutherford <kmr@sanger.ac.uk>
 *  @version $Id: TaskViewer.java,v 1.1 2004/06/17 14:14:24 kmr Exp $
 **/

public class TaskViewer extends JPanel implements Runnable {
  /**
   *  Create a component to monitor the status of the Tasks with the given IDs
   **/
  public TaskViewer (final int [] ids)
      throws uk.ac.sanger.jcon.JobControlException {
    ResourceBundle bundle = uk.ac.sanger.jcon.JobControl.getConfiguration();

    final String val =
      bundle.getString ("jobcontrol.gui.task_viewer.update_time");

    try
      {
        SLEEP_TIME = 1000 * Integer.parseInt (val);
      }
    catch (NumberFormatException nfe)
      {
        throw new Error("Number format error (" + val + ") in "
                        + "configuration ResourceBundle: " +
                        nfe.getMessage ());
      }

    try {
      Context context = new InitialContext();

      datasource = (DataSource) context.lookup("jdbc/JobControl");
    
      makeComponents (ids, datasource);
    } catch (javax.naming.NamingException e) {
      throw new Error ("NamingException: " + e.getMessage ());
    }

    new Thread (this).start ();
  }

  public void run () {
    try {
      StatusDAO statusDAO = new StatusSQLDAO(datasource);
      TaskSQLDAO taskSQLDAO = new TaskSQLDAO(datasource);
      TaskStatisticDAO taskStatDAO = new TaskStatisticSQLDAO(datasource);

      Status completed = statusDAO.readStatusById(Status.COMPLETED);
      Status    failed = statusDAO.readStatusById(Status.FAILED);

      while (true) {
        if (finishFlag) {
          return;
        }

        try {
          Collection tStats = taskStatDAO.readSubmittedTaskState();
          Map statMap = createTaskStatMap(tStats);

          for (Iterator swi = progressbars.keySet ().iterator();
               swi.hasNext(); ) {
            final Integer id = (Integer) swi.next ();
          
            final TaskState tStat = (TaskState) statMap.get(id);

            final JProgressBar progressBar =
              (JProgressBar) progressbars.get (id);
            final JLabel label =
              (JLabel) labels.get (id);

            if (tStat == null) {
              SwingUtilities.invokeLater(new Runnable () {
                public void run () {
                  progressBar.setValue (progressBar.getMaximum());
                  label.setText (getLabelText (id.intValue (),
                                               progressBar.getMaximum(),
                                               progressBar.getMaximum()));
                }
              });
            } else {
              final int compCount = tStat.getJobCount(completed);
              final int failCount = tStat.getJobCount(failed);
              SwingUtilities.invokeLater(new Runnable () {
                public void run () {
                  progressBar.setValue (compCount + failCount);
                  label.setText (getLabelText (id.intValue (),
                                               tStat.getTotalJobCount (),
                                               compCount + failCount));
                }
              });
            }
          }
          Thread.sleep (SLEEP_TIME);
        } catch (IllegalStateException e) {
          throw new Error ("IllegalStateException: " + e.getMessage ()); 
        } catch (InterruptedException _) {
          // ignore interrupt during a sleep()
        }
      }
    } catch (Exception e) {
      e.printStackTrace ();
      throw new Error ("Exception: " + e.getMessage ());
    }
  }

  public void makeComponents (final int [] ids, DataSource datasource) {
    try {
      TaskSQLDAO taskSQLDAO = new TaskSQLDAO(datasource);
      TaskStatisticDAO taskStatDAO = new TaskStatisticSQLDAO(datasource);

      BoxLayout layout = new BoxLayout (this, BoxLayout.Y_AXIS);
      this.setLayout (layout);

      Collection tStats = taskStatDAO.readSubmittedTaskState();
      Map statMap = createTaskStatMap(tStats);

      for (int i = 0 ; i < ids.length ; ++i) {
        final Integer id = new Integer(ids[i]);
        Task task = taskSQLDAO.readTaskById(ids[i], TaskDAO.EXCLUDE_JOB_TREE);
        TaskState tStat = (TaskState) statMap.get(id);

        if (tStat != null) {
          JLabel label1 = new JLabel (task.getDescription ());
          JLabel label2 =
            new JLabel (getLabelText (id.intValue (),
                                      tStat.getTotalJobCount (), 0));

          JProgressBar progressBar =
            new JProgressBar(0, tStat.getTotalJobCount ());

          progressBar.setStringPainted (true);

          add (label1);
          add (label2);
          add (progressBar);

          progressbars.put (id, progressBar);
          labels.put (id, label2);
        }
      }

    } catch (javax.naming.NamingException e) {
      throw new Error ("TaskViewer initialisation failed because of a " +
                       "NamingException: " + e.getMessage ());
    } catch (Exception e) {
      e.printStackTrace ();
      throw new Error ("Exception: " + e);
    }
  }

  private Map createTaskStatMap(Collection tStats)
  {
    Map statMap = new HashMap();

    for (Iterator i = tStats.iterator(); i.hasNext();)
      {
        TaskState tStat = (TaskState) i.next();
        statMap.put(new Integer(tStat.getTaskId()), tStat);
      }

    return statMap;
  }

  /**
   *  
   **/
  private String getLabelText (int id, int totalJobs, int completedJobs) {
    if (totalJobs == completedJobs) {
      return "(task #" + id + ") completed " + totalJobs;
    } else {
      return "(task #" +
        id + ") completed " + completedJobs + "/" + totalJobs;
    }
  }

  /**
   *  The length of time (in seconds) to sleep between updates.
   **/
  private final int SLEEP_TIME;

  private HashMap progressbars = new HashMap ();
  private HashMap labels = new HashMap ();

  private DataSource datasource = null;

  private boolean finishFlag = false;
}
