/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.job;

/**
 * <p><code>ExecutableDefaultImpl</code> represents a task to be
 * executed, including a description of its purpose, a name indicating
 * the resource (binary executable etc.), a name indicating the
 * working directory, any command arguments and environment
 * variables.</p>
 *
 * @author Keith James
 */
public class ExecutableDefaultImpl implements Executable
{
    /**
     * The unique ID of a <code>ExecutableDefaultImpl</code> instance.
     */ 
    private int id;

    /**
     * <code>name</code> is the name used to resolve the binary
     * executable or script which is to be launched.
     */
    protected String name;

    /**
     * <code>version</code> indicates the version of the binary
     * executable or script which is to be launched
     */
    protected String version;

    /**
     * <code>description</code> is a freetext description of the
     * executable's purpose.
     */
    protected String description;

    /**
     * Creates a new, empty <code>ExecutableDefaultImpl</code>.
     */
    public ExecutableDefaultImpl() { }

    /**
     * Creates a new <code>ExecutableDefaultImpl</code> with the
     * supplied name, version and description.
     *
     * @param name a <code>String</code> name used to resolve the
     * binary executable or script which is to be launched.
     * @param version a <code>String</code>.
     * @param description a <code>String</code> freetext description.
     */
    public ExecutableDefaultImpl(String name, String version,
                                 String description)
    {
        if (name == null)
            throw new IllegalArgumentException("Failed to set name as it was null");
        if (description == null)
            throw new IllegalArgumentException("Failed to set description as it was null");

        checkFieldLength(name, Executable.MAX_NAME_LEN);
        checkFieldLength(description, Executable.MAX_DESC_LEN);

        if (version != null)
            checkFieldLength(version, Executable.MAX_VER_LEN);

        this.name        = name;
        this.version     = version;
        this.description = description;
    }

    /**
     * <code>getId</code> returns the ID.
     *
     * @return an <code>int</code>.
     */
    public int getId()
    {
        return id;
    }

    /**
     * <code>getName</code> returns the name used to locate the
     * resource to execute. Currently it is expected that this is
     * simply a file name in the local filesystem. Later this could be
     * applied to other systems such as using a URL or JNDI.
     *
     * @return a <code>String</code>.
     */
    public String getName()
    {
        return name;
    }

    /**
     * <code>setName</code> sets the name used to locate the resource
     * to execute.
     *
     * @param name a <code>String</code>.
     */
    public void setName(String name)
    {
        if (name == null)
            throw new IllegalArgumentException("Failed to set name as it was null");
        checkFieldLength(name, Executable.MAX_NAME_LEN);
        this.name = name;
    }

    /**
     * <code>getVersion</code> returns the version of the resource to
     * be executed.
     *
     * @return a <code>String</code>.
     */
    public String getVersion()
    {
        return version;
    }

    /**
     * <code>setVersion</code> sets the version of the resource to be
     * executed. May be null.
     */
    public void setVersion(String version)
    {
        if (version != null)
            checkFieldLength(version, Executable.MAX_VER_LEN);
        this.version = version;
    }

    /**
     * <code>getDescription</code> returns the free-text description
     * of the purpose of the executable.
     *
     * @return a <code>String</code>.
     */
    public String getDescription()
    {
        return description;
    }

    /**
     * <code>setDescription</code> sets the free-text description of
     * the purpose of the executable.
     *
     * @param description a <code>String</code>.
     */
    public void setDescription(String description)
    {
        if (description == null)
            throw new IllegalArgumentException("Failed to set description as it was null");
        checkFieldLength(description, Executable.MAX_DESC_LEN);
        this.description = description;
    }

    public String toString()
    {
        StringBuffer sb = new StringBuffer("Executable [");
        sb.append(id);
        sb.append("]");
        sb.append(" Name: ");
        sb.append(name);
        sb.append(" Version: ");
        sb.append(version);

        return sb.substring(0);
    }

    protected void checkFieldLength(String field, int maxLength)
    {
        if (field.length() > maxLength)
            throw new IllegalArgumentException("Field "
                                               + field
                                               + " is longer than the permitted maximum of "
                                               + maxLength
                                               + " characters");
    }
}
