/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.job;

import java.io.Serializable;
import java.sql.Timestamp;
import java.util.Enumeration;
import javax.swing.tree.MutableTreeNode;

import uk.ac.sanger.sql.MutableNestedSetTreeNode;
import uk.ac.sanger.util.PropertyChangeable;

/**
 * <p><code>Job</code>s represent computational tasks. These may be
 * single operations or may consist of a tree of operations with
 * completion of the parent being dependent on completion of its child
 * operations.</p>
 *
 * @author Keith James
 */
public interface Job extends Serializable, MutableTreeNode,
                             MutableNestedSetTreeNode, PropertyChangeable
{
    /**
     * <code>OPERATOR_AND</code> is the operator used to resolve job
     * completion by combining with "AND".
     */
    public static final String OPERATOR_AND = "AND";

    /**
     * <code>OPERATOR_OR</code> is the operator used to resolve job
     * completion by combining with "OR".
     */
    public static final String OPERATOR_OR = "OR";

    /**
     * <code>MAX_MAX_RETRY_COUNT</code> is the maximum number of
     * retries permitted.
     */
    public static final int MAX_MAX_RETRY_COUNT = 4;

    /**
     * <code>DEFAULT_MAX_RETRY_COUNT</code> is the default number of
     * retries.
     */
    public static final int DEFAULT_MAX_RETRY_COUNT = 1;

    /**
     * <code>MAX_IO_NAME_LEN</code> is the maximum length permitted
     * for an input, output or error name.
     */
    public static final int MAX_IO_NAME_LEN = 255;

    /**
     * <code>MAX_CMD_TMPL_LEN</code> is the maximum length permitted
     * for a command template.
     */
    public static final int MAX_CMD_TMPL_LEN = 255;

    /**
     * <code>MAX_WORK_DIR_LEN</code> is the maximum length permitted
     * for a working directory path.
     */
    public static final int MAX_WORK_DIR_LEN = 255;

    /**
     * <code>MAX_ENV_LEN</code> is the maximum length permitted
     * for an environment string.
     */
    public static final int MAX_ENV_LEN = 255;

    /**
     * <code>AMD_PREFIX_KEY</code> the <code>ResourceBundle</code> key to
     * specify the prefixes that need to be removed from working directories
     * names before passing to LSF
     */
    public static final String AMD_PREFIX_KEY =
        "jobcontrol.job.amd_prefix";
    
    /**
     * <code>getId</code> returns the ID.
     *
     * @return an <code>int</code>.
     */
    public int getId();

    /**
     * <code>getOperator</code> returns the operator used to combine
     * the success of jobs on which the job depends. May be set to
     * 'AND' or 'OR'.
     *
     * @return a <code>String</code>.
     */
    public String getOperator();

    /**
     * <code>setOperator</code> sets the operator used to combine the
     * success of jobs on which the job depends. May be set to 'AND'
     * or 'OR'.
     *
     * @param operator a <code>String</code>.
     */
    public void setOperator(String operator);

    /**
     * <code>getExecutable</code> returns the executable launched by
     * this job.
     *
     * @return an <code>Executable</code>.
     */
    public Executable getExecutable();

    /**
     * <code>setExecutable</code> sets the executable launched by
     * the job.
     *
     * @param executable an <code>Executable</code>.
     */
    public void setExecutable(Executable executable);

    /**
     * <code>getStatus</code> returns the current status of the job.
     *
     * @return a <code>Status</code>.
     */
    public Status getStatus();

    /**
     * <code>setStatus</code> sets the current status of the job.
     *
     * @param status a <code>Status</code>.
     */
    public void setStatus(Status status);

    /**
     * <code>getStartTime</code> returns the time at which the job
     * started.
     *
     * @return a <code>Timestamp</code>.
     */
    public Timestamp getStartTime();

    /**
     * <code>setStartTime</code> sets the time at which the job
     * started.
     *
     * @param startTime a <code>Timestamp</code>.
     */
    public void setStartTime(Timestamp startTime);

    /**
     * <code>getCompletionTime</code> returns the time at which the
     * job successfully completed.
     *
     * @return a <code>Timestamp</code>.
     */
    public Timestamp getCompletionTime();

    /**
     * <code>setCompletionTime</code> sets the time at which the
     * job successfully completed.
     *
     * @param completionTime a <code>Timestamp</code>.
     */
    public void setCompletionTime(Timestamp completionTime);

    /**
     * <code>getInputName</code> returns the name used to locate the
     * resource which will contain input for the resource being
     * executed. This will be the STDIN of the program being run.
     *
     * @return a <code>String</code>.
     */
    public String getInputName();

    /**
     * <code>setInputName</code> sets the name used to locate the
     * resource which will contain input for the resource being
     * executed. This will be the STDIN of the program being run.
     *
     * @param name a <code>String</code>.
     */
    public void setInputName(String name);

    /**
     * <code>getOutputName</code> returns the name used to locate the
     * resource which will contain output from the resource being
     * executed. This will be the STDOUT of the program being run.
     *
     * @return a <code>String</code>.
     */
    public String getOutputName();

    /**
     * <code>setOutputName</code> sets the name used to locate the
     * resource which will contain output from the resource being
     * executed. This will be the STDOUT of the program being run.
     *
     * @param name a <code>String</code>.
     */
    public void setOutputName(String name);

    /**
     * <code>getErrorName</code> returns the name used to locate the
     * resource which will contain errors from the resource being
     * executed. This will be the STDERR of the program being run.
     *
     * @return a <code>String</code>.
     */
    public String getErrorName();

    /**
     * <code>setErrorName</code> sets the name used to locate the
     * resource which will contain errors from the resource being
     * executed. This will be the STDERR of the program being run.
     *
     * @param name a <code>String</code>.
     */
    public void setErrorName(String name);

    /**
     * <code>getCommandTemplate</code> returns the command line
     * template which will be used to run the <code>Executable</code>.
     *
     * @return a <code>String</code>.
     */
    public String getCommandTemplate();

    /**
     * <code>setCommandTemplate</code> sets the command line template
     * which will be used to run the <code>Executable</code>.
     *
     * @param commandTemplate <code>String</code>.
     */
    public void setCommandTemplate(String commandTemplate);

    /**
     * <code>getWorkDirectoryName</code> returns the name used to
     * locate the working directory. Currently it is expected that
     * this is simply a directory name in the local filesystem. Later
     * this could be applied to other systems such as using a URL or
     * JNDI.
     *
     * @return a <code>String</code>.
     */
    public String getWorkDirectoryName();

    /**
     * <code>setWorkDirectoryName</code>sets the name used to locate
     * the working directory.
     *
     * @param workDirectoryName a <code>String</code>.
     */
    public void setWorkDirectoryName(String workDirectoryName);

    /**
     * <code>getEnvironment</code> returns the evironment variables
     * which will be supplied to the resource to be executed. These
     * are expected to be whitespace-delimited alternating tags and
     * values.
     *
     * @return a <code>String</code>.
     */
    public String getEnvironment();

    /**
     * <code>setEnvironment</code> sets the evironment variables
     * which will be supplied to the resource to be executed. These
     * are expected to be whitespace-delimited alternating tags and
     * values.
     *
     * @param environment a <code>String</code>.
     */
    public void setEnvironment(String environment);

    /**
     * <code>getMaxRetryCount</code> returns the maximum number of
     * times the job can be retried after a failure before it is
     * suspended and the owner contacted.
     *
     * @return an <code>int</code>.
     */
    public int getMaxRetryCount();

    /**
     * <code>setMaxRetryCount</code> sets the maximum number of
     * times the job can be retried after a failure before it is
     * suspended and the owner contacted.
     *
     * @param maxRetryCount an <code>int</code>.
     */
    public void setMaxRetryCount(int maxRetryCount);

    /**
     * <code>getRetryCount</code> returns the number of times the job
     * has been retried after failures.
     *
     * @return an <code>int</code>.
     */
    public int getRetryCount();

    /**
     * <code>setRetryCount</code> sets the number of times the job
     * has been retried after failures.
     *
     * @param retryCount an <code>int</code>.
     */
    public void setRetryCount(int retryCount);

    /**
     * <code>getExitValue</code> returns the exit value of the
     * executable launched by the job after it has completed.
     *
     * @return an <code>int</code>.
     */
    public int getExitValue();

    /**
     * <code>setExitValue</code> sets the exit value of the
     * executable launched by the job after it has completed.
     *
     * @param exitValue an <code>int</code>.
     */
    public void setExitValue(int exitValue);

    /**
     * <code>preorderEnumeration</code> returns an
     * <code>Enumeration</code> in preorder of the tree below this
     * <code>Job</code>.
     *
     * @return an <code>Enumeration</code>.
     */
    public Enumeration preorderEnumeration();

    /**
     * <code>postorderEnumeration</code> returns an
     * <code>Enumeration</code> in postorder of the tree below this
     * <code>Job</code>.
     *
     * @return an <code>Enumeration</code>.
     */
    public Enumeration postorderEnumeration();

    /**
     * <code>add</code> adds a <code>Job</code> as a child of this
     * <code>Job</code>.
     *
     * @param job a <code>Job</code> to be set as a child.
     */
    public void add(Job job);
}
