/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.job;

import java.sql.Timestamp;

/**
 * <code>JobBatchImpl</code> is a <code>Job</code> suitable for
 * assignment to a batch queue.
 *
 * @author Keith James
 */
public class JobBatchImpl extends JobDefaultImpl implements BatchJob
{
    /**
     * The unique ID of a <code>JobBatchImpl</code> instance.
     */
    private int id;

    /**
     * The unique ID of a <code>JobBatchImpl</code> assigned by a
     * batch system when it is submitted.
     */
    private int batchId;

    /**
     * The batch queue.
     */
    protected String queue;

    /**
     * <code>hosts</code> where the job ran (usually just one).
     */
    protected String [] hosts;

    /**
     * Creates a new, empty <code>JobBatchImpl</code> with null parent,
     * default values for priority and maximum number of retries to
     * attempt. 
     */
    public JobBatchImpl()
    {
        super();
    }

    /**
     * Creates a new <code>JobBatchImpl</code> with null parent,
     * default values for priority and maximum number of retries to
     * attempt.
     *
     * @param executable an <code>Executable</code> to run.
     */
    public JobBatchImpl(Executable executable)
    {
        super(executable);
    }

    /**
     * Creates a new <code>JobBatchImpl</code> with null parent and
     * all values explicitly set by the constructor.
     *
     * @param operator a <code>String</code>.
     * @param executable an <code>Executable</code> to run.
     * @param status a <code>Status</code>.
     * @param startTime a <code>Timestamp</code>.
     * @param completionTime a <code>Timestamp</code>.
     * @param inputName a <code>String</code>.
     * @param outputName a <code>String</code>.
     * @param errorName a <code>String</code>.
     * @param commandTemplate a <code>String</code>.
     * @param workDirectoryName a <code>String</code>.
     * @param environment a <code>String</code>.
     * @param maxRetryCount an <code>int</code>.
     * @param retryCount an <code>int</code>.
     * @param exitValue an <code>int</code>.
     * @param queue a <code>String</code> batch queue name.
     */
    public JobBatchImpl(String operator, Executable executable, Status status,
                        Timestamp startTime, Timestamp completionTime,
                        String inputName, String outputName, String errorName,
                        String commandTemplate, String workDirectoryName, String environment,
                        int maxRetryCount, int retryCount, int exitValue,
                        String queue)
    {
        super(operator, executable, status,
              startTime, completionTime,
              inputName, outputName, errorName,
              commandTemplate, workDirectoryName, environment,
              maxRetryCount, retryCount, exitValue);

        if (queue == null)
            throw new IllegalArgumentException("Failed to set queue as it was null");

        if (workDirectoryName == null)
            throw new IllegalArgumentException("Failed to set workDirectoryName as it was null");

        checkFieldLength(queue, BatchJob.MAX_QUEUE_LEN);
        this.queue = queue;

        // Hosts not known until after Job is run
        hosts = new String [0];
    }

    /**
     * <code>getId</code> returns the ID.
     *
     * @return an <code>int</code>.
     */
    public int getId()
    {
        return id;
    }

    public void setWorkDirectoryName(String workDirectoryName)
    {
        if (workDirectoryName == null)
            throw new IllegalArgumentException("Failed to set workDirectoryName as it was null");

        super.setWorkDirectoryName(workDirectoryName);
    }

    /**
     * <code>getBatchId</code> returns the job ID which was assigned
     * by a batch system when the job was submitted, or 0 if the job
     * has not yet been submitted.
     *
     * @return an <code>int</code>.
     */
    public int getBatchId()
    {
        return batchId;
    }

    /**
     * <code>setBatchId</code> sets the job ID which was assigned by a
     * batch system when the job was submitted.
     *
     * @param batchId an <code>int</code>.
     */
    public void setBatchId(int batchId)
    {
        this.batchId = batchId;
    }

    /**
     * <code>getQueue</code> returns the batch queue.
     *
     * @return a <code>String</code>.
     */
    public String getQueue()
    {
        return queue;
    }

    /**
     * <code>setQueue</code> sets the batch queue.
     *
     * @param queue a <code>String</code>.
     */
    public void setQueue(String queue)
    {
       if (queue == null)
            throw new IllegalArgumentException("Failed to set queue as it was null");
        checkFieldLength(queue, BatchJob.MAX_QUEUE_LEN);
        this.queue = queue;
    }

    /**
     * <code>getHosts</code> returns the host(s) on which the batch
     * system ran the job.
     *
     * @return a <code>String []</code> array of host names, or an
     * empty array if the job has not yet been assigned to a host.
     */
    public String [] getHosts()
    {
        return hosts;
    }

    /**
     * <code>setHosts</code> sets the host(s) on which the batch
     * system ran the job.
     *
     * @param hosts a <code>String []</code> array of host names.
     */
    public void setHosts(String [] hosts)
    {
        this.hosts = hosts;
    }

    public String toString()
    {
        StringBuffer sb = new StringBuffer("Batch Job [");
        sb.append(id);
        sb.append("] '");
        sb.append(executable);
        sb.append("', queue ");
        sb.append(queue);
        sb.append(", ");
        sb.append(status);

        return sb.substring(0);
    }
}
