/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.job;

import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.sql.Timestamp;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Vector;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;

import java.util.*;

import uk.ac.sanger.jcon.JobControl;
import uk.ac.sanger.jcon.util.PostorderEnumeration;
import uk.ac.sanger.jcon.util.PreorderEnumeration;
import uk.ac.sanger.util.AbstractPropertyChangeable;

/**
 * <p><code>JobDefaultImpl</code>s represent computational
 * tasks. These may be single operations or may consist of a tree of
 * operations with completion of the parent being dependent on
 * completion of its child operations.</p>
 *
 * @author Keith James
 */
public class JobDefaultImpl extends AbstractPropertyChangeable implements Job
{
    /**
     * The unique ID of a <code>JobDefaultImpl</code> instance.
     */
    private int id;

    /**
     * <code>low</code> is the nested set low value.
     */
    protected int low;

    /**
     * <code>high</code> is the nested set high value.
     */
    protected int high;

    /**
     * <code>parent</code> is the parent of the job.
     */
    protected MutableTreeNode parent;

    /**
     * <code>children</code> is the list of jobs on which the job is
     * dependent.
     */
    protected Vector children;

    /**
     * <code>operator</code> used to resolve completion of the child
     * jobs.
     */
    protected String operator;

    /**
     * The <code>Executable</code> which this job will launch.
     */
    protected Executable executable;

    /**
     * The current <code>Status</code> of the job instance.
     */
    protected Status status;

    /**
     * The <code>Timestamp</code> this job was started.
     */
    protected Timestamp startTime;

    /**
     * The <code>Timestamp</code> this job was successfully completed.
     */
    protected Timestamp completionTime;

   /**
     * <code>inputName</code> is the name used to resolve STDIN for
     * the binary executable or script which is to be launched.
     */
    protected String inputName;

    /**
     * <code>outputName</code> is the name used to resolve STDOUT from
     * the binary executable or script which is to be launched.
     */
    protected String outputName;

    /**
     * <code>inputName</code> is the name used to resolve STDERR from
     * the binary executable or script which is to be launched.
     */
    protected String errorName;

    /**
     * <code>workDirectoryName</code> is the name to be used to
     * resolve the working directory.
     */
    protected String workDirectoryName;

    /**
     * <code>commandTemplate</code> used to run the executable.
     */
    protected String commandTemplate;

    /**
     * <code>environment</code> to be supplied to the executable.
     */
    protected String environment;

    /**
     * The maximum number of retries which may be attempted before
     * failure is reported to the owner.
     */
    protected int maxRetryCount;

    /**
     * The number of retries so far.
     */
    protected int retryCount;

    /**
     * The exit value of the executable launched by the
     * <code>Executable</code>.
     */
    protected int exitValue;

    /**
     * <code>userObject</code> associated user object.
     */
    protected transient Object userObject;

    protected PropertyChangeSupport changeSupport;

    /**
     * Creates a new, empty <code>JobDefaultImpl</code> with null
     * parent, default values for priority and maximum number of
     * retries to attempt.
     */
    public JobDefaultImpl()
    {
        super();
        operator      = Job.OPERATOR_AND;
        maxRetryCount = Job.DEFAULT_MAX_RETRY_COUNT;
        exitValue     = -1;
    }

    /**
     * Creates a new <code>JobDefaultImpl</code> with null parent,
     * default values for priority and maximum number of retries to
     * attempt.
     *
     * @param executable an <code>Executable</code> to run.
     */
    public JobDefaultImpl(Executable executable)
    {
        super();
        if (executable == null)
            throw new IllegalArgumentException("Failed to set Executable as it was null");

        operator      = Job.OPERATOR_AND;
        maxRetryCount = Job.DEFAULT_MAX_RETRY_COUNT;
        exitValue     = -1;

        this.executable = executable;
    }

    /**
     * Creates a new <code>JobDefaultImpl</code> with null parent and
     * all values explicitly set by the constructor.
     *
     * @param operator a <code>String</code>.
     * @param executable an <code>Executable</code> to run.
     * @param status a <code>Status</code>.
     * @param startTime a <code>Timestamp</code>.
     * @param completionTime a <code>Timestamp</code>.
     * @param inputName a <code>String</code>.
     * @param outputName a <code>String</code>.
     * @param errorName a <code>String</code>.
     * @param commandTemplate a <code>String</code>.
     * @param workDirectoryName a <code>String</code>.
     * @param environment a <code>String</code>.
     * @param maxRetryCount an <code>int</code>.
     * @param retryCount an <code>int</code>.
     * @param exitValue an <code>int</code>.
     */
    public JobDefaultImpl(String operator, Executable executable, Status status,
                          Timestamp startTime, Timestamp completionTime,
                          String inputName, String outputName, String errorName,
                          String commandTemplate, String workDirectoryName, String environment,
                          int maxRetryCount, int retryCount, int exitValue)
    {
        super();
        if (maxRetryCount < 0 || maxRetryCount > Job.MAX_MAX_RETRY_COUNT)
            throw new IllegalArgumentException("Invalid maxRetryCount '"
                                               + maxRetryCount
                                               + "' passed to constructor. Valid range is "
                                               + "0 >= id <= "
                                               + Job.MAX_MAX_RETRY_COUNT);

        if (! (operator.equals(Job.OPERATOR_AND) ||
               operator.equals(Job.OPERATOR_OR)))
            throw new IllegalArgumentException("Invalid operator '"
                                               + operator
                                               + "' passed to constructor. Valid options are: '"
                                               + Job.OPERATOR_AND
                                               + "', '"
                                               + Job.OPERATOR_OR
                                               + "'");

        if (executable == null)
            throw new IllegalArgumentException("Failed to set Executable as it was null");
        if (status == null)
            throw new IllegalArgumentException("Failed to set Status as it was null");
        if (commandTemplate == null)
            throw new IllegalArgumentException("Failed to set commandTemplate as it was null");

        checkFieldLength(inputName, Job.MAX_IO_NAME_LEN);
        checkFieldLength(outputName, Job.MAX_IO_NAME_LEN);
        checkFieldLength(errorName, Job.MAX_IO_NAME_LEN);
        checkFieldLength(commandTemplate, Job.MAX_CMD_TMPL_LEN);
        checkFieldLength(workDirectoryName, Job.MAX_WORK_DIR_LEN);

        this.operator          = operator;
        this.executable        = executable;
        this.status            = status;
        this.startTime         = startTime;
        this.completionTime    = completionTime;
        this.inputName         = inputName;
        this.outputName        = outputName;
        this.errorName         = errorName;
        this.commandTemplate   = commandTemplate;
        this.workDirectoryName = workDirectoryName;
        this.environment       = environment;
        this.maxRetryCount     = maxRetryCount;
        this.retryCount        = retryCount;
        this.exitValue         = exitValue;
    }

    /**
     * <code>getId</code> returns the ID.
     *
     * @return an <code>int</code>.
     */
    public int getId()
    {
        return id;
    }

    /**
     * <code>getOperator</code> returns the operator used to combine
     * the success of jobs on which the job depends. May be set to
     * 'AND' or 'OR'.
     *
     * @return a <code>String</code>.
     */
    public String getOperator()
    {
        return operator;
    }

    /**
     * <code>setOperator</code> sets the operator used to combine the
     * success of jobs on which the job depends. May be set to 'AND'
     * or 'OR'.
     *
     * @param operator a <code>String</code>.
     */
    public void setOperator(String operator)
    {
        if (! (operator.equals(Job.OPERATOR_AND) ||
               operator.equals(Job.OPERATOR_OR)))
            throw new IllegalArgumentException("Invalid operator '"
                                               + operator
                                               + "' passed to method. Valid options are: '"
                                               + Job.OPERATOR_AND
                                               + "', '"
                                               + Job.OPERATOR_OR
                                               + "'");
        this.operator = operator;
    }

    /**
     * <code>getExecutable</code> returns the executable launched by
     * this job.
     *
     * @return an <code>Executable</code>.
     */
    public Executable getExecutable()
    {
        return executable;
    }

    /**
     * <code>setExecutable</code> sets the executable launched by
     * the job.
     *
     * @param executable an <code>Executable</code>.
     */
    public void setExecutable(Executable executable)
    {
        if (executable == null)
            throw new IllegalArgumentException("Failed to set Executable as it was null");

        this.executable = executable;
    }

    /**
     * <code>getStatus</code> returns the current status of the job.
     *
     * @return a <code>Status</code>.
     */
    public Status getStatus()
    {
        return status;
    }

    /**
     * <code>setStatus</code> sets the current status of the job.
     *
     * @param status a <code>Status</code>.
     */
    public void setStatus(Status status)
    {
        if (status == null)
            throw new IllegalArgumentException("Failed to set Status as it was null");

        if (hasListeners())
        {
            PropertyChangeSupport pcs = getPropertyChangeSupport();
            pcs.firePropertyChange("status", this.status, status);
        }

        this.status = status;
    }

    /**
     * <code>getStartTime</code> returns the time at which the job
     * started.
     *
     * @return a <code>Timestamp</code>.
     */
    public Timestamp getStartTime()
    {
        return startTime;
    }

    /**
     * <code>setStartTime</code> sets the time at which the job
     * started.
     *
     * @param startTime a <code>Timestamp</code>.
     */
    public void setStartTime(Timestamp startTime)
    {
        this.startTime = startTime;
    }

    /**
     * <code>getCompletionTime</code> returns the time at which the
     * job successfully completed.
     *
     * @return a <code>Timestamp</code>.
     */
    public Timestamp getCompletionTime()
    {
        return completionTime;
    }

    /**
     * <code>setCompletionTime</code> sets the time at which the
     * job successfully completed.
     *
     * @param completionTime a <code>Timestamp</code>.
     */
    public void setCompletionTime(Timestamp completionTime)
    {
        this.completionTime = completionTime;
    }

    /**
     *  Remove or add NFS prefixes to directory names
     **/
    public String fixName(String name) {
        if (name.startsWith ("/yeastpub4")) {
            System.err.println ("setting name to: " +
                                "/nfs/disk222/" + name);
            return "/nfs/disk222" + name;
        } else {
            ResourceBundle bundle = JobControl.getConfiguration();

            Enumeration e = bundle.getKeys();

            while (e.hasMoreElements())
            {
                final String key = (String) e.nextElement();

                if (key.startsWith (AMD_PREFIX_KEY)) {
                    try
                    {
                        final String val = bundle.getString(key);

                        if (name.startsWith (val)) {
                            return name.substring (val.length());
                        }
                    }
                    catch (NullPointerException npe)
                    {
                        throw new Error("Malformed configuration " +
                                        "ResourceBundle " +
                                        "contains key '" + key +
                                        "' missing a corresponding value");
                    }
                }

            }
        }

        return name;
    }

    /**
     * <code>getInputName</code> returns the name used to locate the
     * resource which will contain input from the resource being
     * executed. This will be the STDOUT of the program being run.
     *
     * @return a <code>String</code>.
     */
    public String getInputName()
    {
        return inputName;
    }

    /**
     * <code>setInputName</code> sets the name used to locate the
     * resource which will contain input from the resource being
     * executed. This will be the STDOUT of the program being run.
     *
     * @param inputName a <code>String</code>.
     */
    public void setInputName(String inputName)
    {
        inputName = fixName(inputName);
        checkFieldLength(inputName, Job.MAX_IO_NAME_LEN);
        this.inputName = inputName;
   }

    /**
     * <code>getOutputName</code> returns the name used to locate the
     * resource which will contain input for the resource being
     * executed. This will be the STDIN of the program being run.
     *
     * @return a <code>String</code>.
     */
    public String getOutputName()
    {
        return outputName;
    }

    /**
     * <code>setOutputName</code> sets the name used to locate the
     * resource which will contain input for the resource being
     * executed. This will be the STDIN of the program being run.
     *
     * @param outputName a <code>String</code>.
     */
    public void setOutputName(String outputName)
    {
        outputName = fixName(outputName);
        checkFieldLength(outputName, Job.MAX_IO_NAME_LEN);
        this.outputName = outputName;
    }

    /**
     * <code>getErrorName</code> returns the name used to locate the
     * resource which will contain errors from the resource being
     * executed. This will be the STDERR of the program being run.
     *
     * @return a <code>String</code>.
     */
    public String getErrorName()
    {
        return errorName;
    }

    /**
     * <code>setErrorName</code> sets the name used to locate the
     * resource which will contain errors from the resource being
     * executed. This will be the STDERR of the program being run.
     *
     * @param errorName a <code>String</code>.
     */
    public void setErrorName(String errorName)
    {
        errorName = fixName(errorName);
        checkFieldLength(errorName, Job.MAX_IO_NAME_LEN);
        this.errorName = errorName;
    }

    /**
     * <code>getCommandTemplateName</code> returns the command line
     * template which will be used to run the <code>Executable</code>.
     *
     * @param commandTemplate <code>String</code>.
     */
    public String getCommandTemplate()
    {
        return commandTemplate;
    }

    /**
     * <code>setCommandTemplate</code> sets the command line template
     * string which will be used to run the <code>Executable</code>.
     *
     * @param commandTemplate <code>String</code>.
     */
    public void setCommandTemplate(String commandTemplate)
    {
        if (commandTemplate == null)
            throw new IllegalArgumentException("Failed to set commandTemplate as it was null");

        checkFieldLength(commandTemplate, Job.MAX_CMD_TMPL_LEN);
        this.commandTemplate = commandTemplate;
    }

    /**
     * <code>getWorkDirectoryName</code> returns the name used to
     * locate the working directory. Currently it is expected that
     * this is simply a directory name in the local filesystem. Later
     * this could be applied to other systems such as using a URL or
     * JNDI.
     *
     * @return a <code>String</code>.
     */
    public String getWorkDirectoryName()
    {
        return workDirectoryName;
    }

    /**
     * <code>setWorkDirectoryName</code> sets the name used to locate
     * the working directory.
     *
     * @param workDirectoryName a <code>String</code>.
     */
    public void setWorkDirectoryName(String workDirectoryName)
    {
        workDirectoryName = fixName(workDirectoryName);
        checkFieldLength(workDirectoryName, Job.MAX_WORK_DIR_LEN);

        this.workDirectoryName = workDirectoryName;
    }

    /**
     * <code>getEnvironment</code> returns the evironment variables
     * which will be supplied to the resource to be executed. These
     * are expected to be whitespace-delimited alternating tags and
     * values.
     *
     * @return a <code>String</code>.
     */
    public String getEnvironment()
    {
        return environment;
    }

    /**
     * <code>setEnvironment</code> sets the evironment variables
     * which will be supplied to the resource to be executed. These
     * are expected to be whitespace-delimited alternating tags and
     * values.
     *
     * @param environment a <code>String</code>.
     */
    public void setEnvironment(String environment)
    {
        checkFieldLength(environment, Job.MAX_ENV_LEN);
        this.environment = environment;
    }

    /**
     * <code>getMaxRetryCount</code> returns the maximum number of
     * times the job can be retried after a failure before it is
     * suspended and the owner contacted.
     *
     * @return an <code>int</code>.
     */
    public int getMaxRetryCount()
    {
        return maxRetryCount;
    }

    /**
     * <code>setMaxRetryCount</code> sets the maximum number of
     * times thiejob can be retried after a failure before it is
     * suspended and the owner contacted.
     *
     * @param maxRetryCount an <code>int</code>.
     */
    public void setMaxRetryCount(int maxRetryCount)
        throws IllegalArgumentException
    {
        if (maxRetryCount < 0 || maxRetryCount > Job.MAX_MAX_RETRY_COUNT)
            throw new IllegalArgumentException("Invalid maxRetryCount '"
                                               + maxRetryCount
                                               + "' passed to method. Valid range is "
                                               + "0 >= priority <= "
                                               + Job.MAX_MAX_RETRY_COUNT);
        this.maxRetryCount = maxRetryCount;
    }

    /**
     * <code>getRetryCount</code> returns the number of times the job
     * has been retried after failures.
     *
     * @return an <code>int</code>.
     */
    public int getRetryCount()
    {
        return retryCount;
    }

    /**
     * <code>setRetryCount</code> sets the number of times the job
     * has been retried after failures.
     *
     * @param retryCount an <code>int</code>.
     */
    public void setRetryCount(int retryCount)
    {
        if (hasListeners())
        {
            PropertyChangeSupport pcs = getPropertyChangeSupport();
            pcs.firePropertyChange("retryCount", this.retryCount, retryCount);
        }

        this.retryCount = retryCount;
    }

    /**
     * <code>getExitValue</code> returns the exit value of the
     * executable launched by the job after it has completed.
     *
     * @return an <code>int</code>.
     */
    public int getExitValue()
    {
        return exitValue;
    }

    /**
     * <code>setExitValue</code> sets the exit value of the
     * executable launched by the job after it has completed.
     *
     * @param exitValue an <code>int</code>.
     */
    public void setExitValue(int exitValue)
    {
        if (hasListeners())
        {
            PropertyChangeSupport pcs = getPropertyChangeSupport();
            pcs.firePropertyChange("exitValue", this.exitValue, exitValue);
        }

        this.exitValue = exitValue;
    }

    public Enumeration children()
    {
        if (children == null)
            return DefaultMutableTreeNode.EMPTY_ENUMERATION;
        else
            return children.elements();
    }

    public boolean getAllowsChildren()
    {
        return true;
    }

    public TreeNode getChildAt(int childIndex)
    {
        if (children == null)
            throw new ArrayIndexOutOfBoundsException("Job has no children");
        else
            return (MutableTreeNode) children.get(childIndex);
    }

    public int getChildCount()
    {
        if (children == null)
            return 0;
        else
            return children.size();
    }

    public int getIndex(TreeNode job)
    {
        if (job == null)
            throw new IllegalArgumentException("Job argument was null");
        else if (! isOurChild((Job) job))
            return -1;
        else
            return children.indexOf(job);
    }

    public TreeNode getParent()
    {
        return parent;
    }

    public void setParent(MutableTreeNode parent)
    {
        this.parent = parent;
    }

    public boolean isLeaf()
    {
        return (getChildCount() == 0);
    }

    public void add(Job newChild)
    {
        if (newChild.getParent() == this)
            throw new IllegalArgumentException("Failed to added child Job as it is already a child of this Job");
        else
            insert(newChild, getChildCount());
    }

    public void insert(MutableTreeNode newChild, int childIndex)
    {
        if (! (newChild instanceof Job))
            throw new IllegalArgumentException("Failed to add child Job as it was not an instance of uk.ac.sanger.jcon.job.Job");
        else if (newChild == null)
            throw new IllegalArgumentException("Failed to add child Job as it was null");
        else if (isOurAncestor(newChild))
            throw new IllegalArgumentException("Failed to add child Job as it is an ancestor");

        MutableTreeNode oldParent = (MutableTreeNode) newChild.getParent();

        if (oldParent != null)
            oldParent.remove(newChild);

        newChild.setParent(this);

        if (children == null)
            children = new Vector();

        children.insertElementAt(newChild, childIndex);
    }

    public void remove(int childIndex)
    {
        MutableTreeNode child = (MutableTreeNode) getChildAt(childIndex);
        children.remove(childIndex);
        child.setParent(null);
    }

    public void remove(MutableTreeNode child)
    {
        if (child == null)
            throw new IllegalArgumentException("Failed to remove child Job as it was null");
        else if (! isOurChild(child))
            throw new IllegalArgumentException("Failed to remove child as it the Job is not its parent");
        else
            remove(getIndex(child));
    }

    public void removeFromParent()
    {
        MutableTreeNode parent = (MutableTreeNode) getParent();
        if (parent != null)
            parent.remove(this);
    }

    public Object getUserObject()
    {
        return userObject;
    }

    public void setUserObject(Object userObject)
    {
        this.userObject = userObject;
    }

    public Enumeration preorderEnumeration()
    {
        return new PreorderEnumeration(this);
    }

    public Enumeration postorderEnumeration()
    {
        return new PostorderEnumeration(this);
    }

    public int getLow()
    {
        return low;
    }

    public void setLow(int low)
    {
        this.low = low;
    }

    public int getHigh()
    {
        return high;
    }

    public void setHigh(int high)
    {
        this.high = high;
    }

    public String toString()
    {
        StringBuffer sb = new StringBuffer("Job [");
        sb.append(id);
        sb.append("] '");
        sb.append(executable);
        sb.append("', ");
        sb.append(status);

        return sb.substring(0);
    }

    protected void checkFieldLength(String field, int maxLength)
    {
        if (field != null && field.length() > maxLength)
            throw new IllegalArgumentException("Field "
                                               + field
                                               + " is longer than the permitted maximum of "
                                               + maxLength
                                               + " characters");
    }

    protected boolean isOurChild(TreeNode child)
    {
        if (child == null)
            return false;
        else if (getChildCount() == 0)
            return false;
        else
            return (child.getParent() == this);
    }

    protected boolean isOurAncestor(TreeNode job)
    {
        if (job == null)
            return false;

        TreeNode ancestor = this;

        while (ancestor != null)
        {
            ancestor = ancestor.getParent();
            if (ancestor == job)
                return true;
        }

        return false;
    }

    protected void writeObject(ObjectOutputStream stream) throws IOException
    {
        stream.defaultWriteObject();

        Object [] objects;

        if (userObject != null && userObject instanceof Serializable)
            objects = new Object [] { "userObject" , userObject };
        else
            objects = new Object[0];

        stream.writeObject(objects);
    }

    protected void readObject(ObjectInputStream stream)
        throws IOException, ClassNotFoundException
    {
        stream.defaultReadObject();

        Object [] objects;

        objects = (Object []) stream.readObject();

        if (objects.length > 0 && objects[0].equals("userObject"))
            userObject = objects[1];
    }
}
