/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.job;

import java.io.Serializable;
import java.sql.Timestamp;

/**
 * <code>Task</code> represents the top level of a set of computes. An
 * instance contains a freetext description of the overall purpose of
 * the computes, an indication of the status of the compute as a
 * whole, the details of its owner and a <code>Job</code> to
 * execute. Also stored is a timestamp which is set to the time when
 * the computes were submitted to the system.
 *
 * @author Keith James
 */
public interface Task extends Serializable
{
    /**
     * <code>MAX_PRIORITY</code> is the maximum (highest) priority a
     * task may recieve.
     */
    public static final int MAX_PRIORITY = 10;

    /**
     * <code>MIN_PRIORITY</code> is the minimum (lowest) priority a
     * task may recieve.
     */
    public static final int MIN_PRIORITY = 1;

    /**
     * <code>DEFAULT_PRIORITY</code> is the default priority for a
     * task.
     */
    public static final int DEFAULT_PRIORITY = 4;

    /**
     * <code>MAX_NAME_LEN</code> is the maximum length permitted for a
     * name.
     */
    public static final int MAX_NAME_LEN = 32;

    /**
     * <code>MAX_DESC_LEN</code> is the maximum length permitted for a
     * description.
     */
    public static final int MAX_DESC_LEN = 255;

    /**
     * <code>getId</code> returns the ID.
     *
     * @return an <code>int</code>.
     */
    public int getId();

    /**
     * <code>getName</code> returns the name of this type of task.
     *
     * @return a <code>String</code>.
     */
    public String getName();

    /**
     * <code>setName</code> sets the name of this type of task.
     *
     * @param name a <code>String</code>.
     */
    public void setName(String name);

    /**
     * <code>getDescription</code> returns the free text description of
     * the task.
     *
     * @return a <code>String</code>.
     */
    public String getDescription();

    /**
     * <code>setDescription</code> sets the free text description of
     * the task.
     *
     * @param description a <code>String</code>.
     */
    public void setDescription(String description);

    /**
     * <code>getOwner</code> returns the owner of the job group.
     *
     * @return an <code>Owner</code>.
     */
    public Owner getOwner();

    /**
     * <code>setOwner</code> sets the owner of the job group.
     *
     * @param owner an <code>Owner</code>.
     */
    public void setOwner(Owner owner);

    /**
     * <code>getStatus</code> returns the status of the task
     * calculated from the status of its jobs. This will be WAITING
     * before task is submitted for running, then STARTED once
     * submission occurs and will finally be given the same status as
     * the final job, once that has been run. The final status may be
     * any legal status.
     *
     * @return a <code>Status</code>.

     * @exception IllegalStateException if the <code>Task</code>
     * contains no <code>Job</code>s.
     */
    public Status getStatus();

    /**
     * <code>getJobs</code> returns the jobs to be executed by this
     * task. The final job is at index 0.
     *
     * @return a <code>Job []</code> array.
     */
    public Job [] getJobs();

    /**
     * <code>setJob</code> sets the jobs to be executed by this
     * task. The final job is at index 0.
     *
     * @param jobs a <code>Job []</code> array.
     */
    public void setJobs(Job [] jobs);

    /**
     * <code>getPriority</code> returns the priority of the task.
     *
     * @return an <code>int</code>.
     */
    public int getPriority();

    /**
     * <code>setPriority</code> sets the priority of the task.
     *
     * @param priority an <code>int</code>.
     */
    public void setPriority(int priority);

    /**
     * <code>getSubmissionTime</code> returns the time at which the
     * task was submitted.
     *
     * @return a <code>Timestamp</code>.
     */
    public Timestamp getSubmissionTime();

    /**
     * <code>setSubmissionTime</code> sets the time at which the task
     * was submitted.
     *
     * @param submissionTime a <code>Timestamp</code>.
     */
    public void setSubmissionTime(Timestamp submissionTime);
}
