/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.job;

import javax.sql.DataSource;
import javax.naming.Context;
import javax.naming.InitialContext;

import uk.ac.sanger.jcon.dao.JobDAO;
import uk.ac.sanger.jcon.dao.JobSQLDAO;

import java.sql.Timestamp;

/**
 * <code>TaskDefaultImpl</code> represents the top level of a set of
 * computes. An instance contains a freetext description of the
 * overall purpose of the computes, an indication of the status of the
 * compute as a whole, the details of its owner and a <code>Job</code>
 * to execute. Also stored is a timestamp which is set to the time
 * when the computes were submitted to the system.
 *
 * @author Keith James
 */
public class TaskDefaultImpl implements Task
{
    /**
     * The unique ID of a <code>Task</code> instance.
     */
    private int id;

    /**
     * <code>name</code> is a name unique to this type of task.
     */
    protected String name;

    /**
     * <code>description</code> is a freetext description of the
     * task's purpose.
     */
    protected String description;

    /**
     * <code>owner</code> is the owner of the task.
     */
    protected Owner owner;

   /**
     * <code>jobs</code> is the job array to be executed by this
     * task.
     */
    protected Job [] jobs;

    /**
     * The priority of the instance.
     */
    protected int priority;

    /**
     * The <code>Timestamp</code> this task was submitted.
     */
    protected Timestamp submissionTime;

    /**
     * Creates a new <code>TaskDefaultImpl</code>.
     *
     * @param name a <code>String</code> identifier for this type of
     * task.
     * @param description a <code>String</code> freetext description
     * of the task purpose.
     * @param owner an <code>Owner</code>.
     */
    public TaskDefaultImpl(String name, String description, Owner owner)
    {
        if (name == null)
            throw new IllegalArgumentException("Failed to set name "
                                               + "as it was null");
        if (description == null)
            throw new IllegalArgumentException("Failed to set description "
                                               + "as it was null");
        if (owner == null)
            throw new IllegalArgumentException("Failed to set Owner "
                                               + "as it was null");

        checkFieldLength(name, Task.MAX_NAME_LEN);
        checkFieldLength(description, Task.MAX_DESC_LEN);

        this.name        = name;
        this.description = description;
        this.owner       = owner;

        priority = Task.DEFAULT_PRIORITY;
    }

    /**
     * Creates a new <code>TaskDefaultImpl</code>.
     *
     * @param name a <code>String</code> identifier for this type of
     * task.
     * @param description a <code>String</code> freetext description
     * of the taks purpose.
     * @param owner an <code>Owner</code>.
     * @param jobs a <code>Job []</code> array to be executed.
     */
    public TaskDefaultImpl(String name, String description,
                           Owner owner, Job [] jobs)
    {
        if (name == null)
            throw new IllegalArgumentException("Failed to set name "
                                               + "as it was null");
        if (description == null)
            throw new IllegalArgumentException("Failed to set description "
                                               + "as it was null");
        if (owner == null)
            throw new IllegalArgumentException("Failed to set Owner "
                                               + "as it was null");

       if (jobs == null || jobs.length == 0)
            throw new IllegalArgumentException("Jobs array must contain "
                                               + "at least one Job");

        checkFieldLength(name, Task.MAX_NAME_LEN);
        checkFieldLength(description, Task.MAX_DESC_LEN);

        this.name        = name;
        this.description = description;
        this.owner       = owner;
        this.jobs        = jobs;

        priority = Task.DEFAULT_PRIORITY;
    }

    /**
     * <code>getId</code> returns the ID.
     *
     * @return an <code>int</code>.
     */
    public int getId()
    {
        return id;
    }

    /**
     * <code>getName</code> returns the name of this type of task.
     *
     * @return a <code>String</code>.
     */
    public String getName()
    {
        return name;
    }

    /**
     * <code>setName</code> sets the name of this type of task.
     *
     * @param name a <code>String</code>.
     */
    public void setName(String name)
    {
        if (name == null)
            throw new IllegalArgumentException("Failed to set name "
                                               + "as it was null");
        checkFieldLength(name, Task.MAX_NAME_LEN);
        this.name = name;
    }

    /**
     * <code>getDescription</code> returns the free text description of
     * the job group.
     *
     * @return a <code>String</code>.
     */
    public String getDescription()
    {
        return description;
    }

    /**
     * <code>setDescription</code> sets the free text description of
     * the task.
     *
     * @param description a <code>String</code>.
     */
    public void setDescription(String description)
    {
        if (description == null)
            throw new IllegalArgumentException("Failed to set description "
                                               + "as it was null");
        checkFieldLength(description, Task.MAX_DESC_LEN);
        this.description = description;
    }

    /**
     * <code>getOwner</code> returns the owner of the task.
     *
     * @return an <code>Owner</code>.
     */
    public Owner getOwner()
    {
        return owner;
    }

    /**
     * <code>setOwner</code> sets the owner of the task.
     *
     * @param owner an <code>Owner</code>.
     */
    public void setOwner(Owner owner)
    {
        if (owner == null)
            throw new IllegalArgumentException("Failed to set Owner "
                                               + "as it was null");
        this.owner = owner;
    }

    /**
     * <code>getStatus</code> returns the status of the task which is
     * defined as the status of the final goal job.
     *
     * @return a <code>Status</code> or null if an error occurs while reading
     * from the database.
     */
    public Status getStatus()
    {
        try {
            // This is probably wrong-headed, but we don't want to call
            // createTask().getStatus() everytime we need the status because
            // that would be slow
            Context context = new InitialContext();
            DataSource datasource =
                (DataSource) context.lookup("jdbc/JobControl");
            JobSQLDAO jobSQLDAO = new JobSQLDAO(datasource);
            Job job_zero = jobSQLDAO.readJobById(jobs[0].getId());
            return job_zero.getStatus();
        } catch (java.lang.Exception e) {
            return null;
        }
    }

    /**
     * <code>getJobs</code> returns the jobs to be executed by this
     * task.
     *
     * @return a <code>Job []</code> array.
     */
    public Job [] getJobs()
    {
        return jobs;
    }

    /**
     * <code>setJob</code> sets the job to be executed by this
     * task.
     *
     * @param jobs a <code>Job []</code> array.
     */
    public void setJobs(Job [] jobs)
    {
        if (jobs == null || jobs.length == 0)
            throw new IllegalArgumentException("Jobs array must contain "
                                               + "at least one Job");
        this.jobs = jobs;
    }

    /**
     * <code>getPriority</code> returns the priority of the task.
     *
     * @return an <code>int</code>.
     */
    public int getPriority()
    {
        return priority;
    }

    /**
     * <code>setPriority</code> sets the priority of the task.
     *
     * @param priority an <code>int</code>.
     */
    public void setPriority(int priority) throws IllegalArgumentException
    {
        if (priority < Task.MIN_PRIORITY || priority > Task.MAX_PRIORITY)
            throw new IllegalArgumentException("Invalid priority '"
                                               + priority
                                               + "' passed to method. Valid range is "
                                               + Task.MIN_PRIORITY
                                               + " >= priority <= "
                                               + Task.MAX_PRIORITY);
        this.priority = priority;
    }

    /**
     * <code>getSubmissionTime</code> returns the time at which the
     * task was submitted.
     *
     * @return a <code>Timestamp</code>.
     */
    public Timestamp getSubmissionTime()
    {
        return submissionTime;
    }

    /**
     * <code>setSubmissionTime</code> sets the time at which the task
     * was submitted.
     *
     * @param submissionTime a <code>Timestamp</code>.
     */
    public void setSubmissionTime(Timestamp submissionTime)
    {
        this.submissionTime = submissionTime;
    }

    public String toString()
    {
        StringBuffer sb = new StringBuffer("Task [");
        sb.append(id);
        sb.append("] '");
        sb.append(name);
        sb.append("'\n");
        sb.append(owner);

        if (submissionTime == null)
        {
            sb.append("\nNot submitted");
        }
        else
        {
            sb.append("\nSubmitted: ");
            sb.append(submissionTime);
        }

        for (int i = 0; i < jobs.length; i++)
            sb.append("\n" + jobs[i]);

        return sb.substring(0);
    }

    protected void checkFieldLength(String field, int maxLength)
    {
        if (field.length() > maxLength)
            throw new IllegalArgumentException("Field "
                                               + field
                                               + " is longer than the permitted maximum of "
                                               + maxLength
                                               + " characters");
    }
}
