/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.job;

/**
 * <code>TaskState</code> is a value object which contains information
 * about the state of a <code>Task</code>. Its purpose is to contain
 * information which has been obtained by querying a store (e.g. an
 * SQL database) without instantiating <code>Task</code> objects.
 *
 * @author Keith James
 */
public class TaskState
{
    private int taskId;
    private int [] jobCounts;

    /**
     * Creates a new <code>TaskState</code>.
     *
     * @param taskId an <code>int</code> ID.
     */
    public TaskState(int taskId)
    {
        this.taskId = taskId;
        jobCounts = new int [7];
    }

    /**
     * <code>getTaskId</code> returns the ID of the <code>Task</code>
     * which the object describes.
     *
     * @return an <code>int</code> ID.
     */
    public int getTaskId()
    {
        return taskId;
    }

    /**
     * <code>setJobCount</code> sets the number of <code>Job</code>s
     * which have the specified <code>Status</code>.
     *
     * @param status a <code>Status</code>.
     * @param jobCount an <code>int</code> number of <code>Job</code>s
     * with that <code>Status</code>.
     */
    public void setJobCount(Status status, int jobCount)
    {
        jobCounts[status.getId() - 1] = jobCount;
    }

    /**
     * <code>getJobCount</code> returns the number of
     * <code>Job</code>s which have the specified <code>Status</code>.
     *
     * @param status a <code>Status</code>.
     */
    public int getJobCount(Status status)
    {
        return jobCounts[status.getId() - 1];
    }

    /**
     * <code>getTotalJobCount</code> returns the total number of
     * <code>Job</code>s in the <code>Task</code>.
     *
     * @return an <code>int</code>.
     */
    public int getTotalJobCount()
    {
        int total = 0;

        for (int i = 0; i < jobCounts.length; i++)
        {
            total += jobCounts[i];
        }

        return total;
    }

    public String toString()
    {
        return "TaskState [" + taskId + "]";
    }
}
