/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.lsf;

import java.util.List;
import java.util.Iterator;

/**
 * <code>LSFBatchInfo</code> is a class which provides access to the
 * LSF batch queue information C API. The native methods are
 * synchronized, which effectively serialises access to the C library
 * as there is only one instance of the class to obtain a lock on.
 *
 * @author Keith James
 */
public final class LSFBatchInfo
{
    static
    {
        System.loadLibrary("LSFBatchInfo");
    }

    // Singleton instance of this class
    private static final LSFBatchInfo INSTANCE = new LSFBatchInfo();

    /**
     * Creates a new <code>LSFBatchSubmit</code>. Private constructor;
     * use getInstance to get the singleton.
     */
    private LSFBatchInfo() { }

    /**
     * <code>getInstance</code> returns the singleton instance of this
     * class.
     *
     * @return an <code>LSFBatchInfo</code>.
     */
    public static LSFBatchInfo getInstance()
    {
        return INSTANCE;
    }

    /**
     * <code>getJobInfo</code> returns an <code>LSFJobInfo</code>
     * object representing a job with the specified
     * <strong>LSF</strong> job id.
     *
     * @param lsfJobId an <code>int</code>.
     *
     * @return a <code>LSFJobInfo</code>.
     *
     * @exception LSFBatchException if an error occurs or the
     * specified job does not exist.
     * @exception JNIException if a serious C API level error occurs.
     */
    public native LSFJobInfo getJobInfo(int lsfJobId)
        throws LSFBatchException, JNIException;

    /**
     * <code>getJobInfo</code> returns a <code>List</code> of
     * <code>LSFJobInfo</code> objects representing jobs with the
     * specified <strong>LSF</strong> job ids.
     *
     * @param lsfJobIds an <code>int []</code> array.
     *     
     * @return a <code>List</code> of <code>LSFJobInfo</code>
     * objects.
     *
     * @exception LSFBatchException if an error occurs or one of the
     * specified jobs does not exist.
     * @exception JNIException if a serious C API level error occurs.
     */
    public native List getJobInfo(int [] lsfJobIds)
        throws LSFBatchException, JNIException;

    /**
     * <code>getJobInfo</code> returns a <code>List</code> of
     * <code>LSFJobInfo</code> objects representing all jobs on a
     * particular queue. In this context, all jobs means jobs of any
     * status (including pending, suspended, current, done), belonging
     * to any user and executing on any host(s).
     *
     * @param queue a <code>String</code> queue name.
     *
     * @return a <code>List</code> of <code>LSFJobInfo</code>
     * objects.
     *
     * @exception LSFBatchException if an error occurs or the
     * specified queue does not exist.
     * @exception JNIException if a serious C API level error occurs.
     */
    public native List getJobInfo(String queue)
        throws LSFBatchException, JNIException;

    /**
     * <code>getJobInfo</code> returns a <code>List</code> of
     * <code>LSFJobInfo</code> objects representing all jobs on a
     * particular queue belonging to a specific user.
     *
     * @param user a <code>String</code> user name.
     * @param queue a <code>String</code> queue name.
     *
     * @return a <code>List</code> of <code>LSFJobInfo</code>
     * objects.
     *
     * @exception LSFBatchException if an error occurs or if the
     * specified user or queue do not exist.
     * @exception JNIException if a serious C API level error occurs.
     */
    public native List getJobInfo(String user, String queue)
        throws LSFBatchException, JNIException;

    /**
     * <code>getJobInfo</code> returns a <code>List</code> of
     * <code>LSFJobInfo</code> objects representing all jobs specified
     * by the array of job names.
     *
     * @param jobNames a <code>String []</code> array of job names.
     *
     * @return a <code>List</code> of <code>LSFJobInfo</code>
     * objects.
     *
     * @exception LSFBatchException if an error occurs or one or more
     * of the specified job names do not exist.
     * @exception JNIException if a serious C API level error occurs.
     */
    public native List getNamedJobInfo(String [] jobNames)
        throws LSFBatchException, JNIException;
}
