/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.lsf;

/**
 * <code>LSFBatchSubmit</code> is a class which provides access to the
 * LSF batch submission C API. The native method is synchronized,
 * which effectively serialises access to the C library as there is
 * only one instance of the class to obtain a lock on.
 *
 * @author Keith James
 */
public final class LSFBatchSubmit
{
    static
    {
        System.loadLibrary("LSFBatchSubmit");
    }

    // Singleton instance of this class
    private static final LSFBatchSubmit INSTANCE = new LSFBatchSubmit();

    /**
     * Creates a new <code>LSFBatchSubmit</code>. Private constructor;
     * use getInstance to get the singleton.
     */
    private LSFBatchSubmit() { }

    /**
     * <code>getInstance</code> returns the singleton instance of this
     * class.
     *
     * @return an <code>LSFBatchSubmit</code>.
     */
    public static LSFBatchSubmit getInstance()
    {
        return INSTANCE;
    }

    /**
     * <code>submit</code> submits a job to an LSF batch queue. This
     * calls the C function <code>lsb_submit</code> via
     * JNI. Consequently it has the same caveats. Notably, if a null
     * output file is specified, the submitter will recieve mail
     * notification from LSF (the job report). The implementation
     * endeavours to avoid sending mail, but the LSF API gives little
     * choice.
     *
     * @param jobName a <code>String</code> to be used as the job
     * name.
     * @param queue a <code>String</code> queue name to submit to.
     * @param infile a <code>String</code> file name for input (or
     * null if none is required).
     * @param outFile a <code>String</code> file name for output (or
     * null if none is required).
     * @param errFile a <code>String</code> file name for errors (or
     * null if none is required).
     * @param command a <code>String</code>.
     *
     * @return an <code>int</code> equal to the LSF job ID assigned to
     * the job.
     *
     * @exception LSFBatchException if an error occurs.
     * @exception JNIException if a serious C API level error occurs.
     */
    public synchronized native int submit(String jobName, String queue,
                                          String inFile, String outFile, String errFile,
                                          String command)
        throws LSFBatchException, JNIException;
}
