/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.lsf;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.run.BufferedStringSink;

/**
 * <code>LSFBjobsExec</code> is an alternative to
 * <code>LSFBatchInfo</code> which does not use JNI, but instead runs
 * the C program "bjobs_plus", a replacement for "bjobs" which
 * includes information not reported by "bjobs" (such as full job
 * status, job exit code) and writes only one line per job.
 *
 * @author Keith James
 */
public final class LSFBjobsExec
{
    // Log4j setup
    static Category cat =
        Category.getInstance(LSFBjobsExec.class.getName());

    // Singleton instance of this class
    private static final LSFBjobsExec INSTANCE = new LSFBjobsExec();

    /**
     * Creates a new <code>LSFBjobsExec</code>. Private constructor;
     * use getInstance to get the singleton.
     */
    private LSFBjobsExec() { }

    /**
     * <code>getInstance</code> returns the singleton instance of this
     * class.
     *
     * @return an <code>LSFBjobsExec</code>.
     */
    public static LSFBjobsExec getInstance()
    {
        return INSTANCE;
    }

    /**
     * <code>getJobInfo</code> returns a <code>List</code> of
     * <code>LSFJobInfo</code> objects representing jobs whose LSF job
     * IDs are in <code>lsfJobIds</code>.
     *
     * @param lsfJobIds a <code>Set</code> of <code>Integer<code> LSF
     * job IDs.
     *
     * @return a <code>List</code> of <codeLSFJobInfo</code> objects.
     *
     * @exception LSFBatchException if an error occurs.
     */
    public List getJobInfo(Set lsfJobIds)
        throws LSFBatchException
    {
        Process process;

        List bjobsCommand = new ArrayList();
        bjobsCommand.add("bjobs_plus");

        List lsfJobInfo = new ArrayList();

        try
        {
            Runtime runtime = Runtime.getRuntime();

            String [] tokens = (String []) bjobsCommand.toArray(new String [0]);
            process = runtime.exec(tokens);         

            BufferedStringSink stderrSink =
                new BufferedStringSink(process.getErrorStream());
            stderrSink.start();

            Runnable parser = new Parser(process.getInputStream(),
                                         lsfJobIds,
                                         lsfJobInfo);

            Thread thread = new Thread(parser);
            thread.start();

            int exitValue = process.waitFor();

            stderrSink.join();
            thread.join();

            String stderr = stderrSink.getBuffer().substring(0);

            if (exitValue != 0)
                throw new LSFBatchException("Recieved exit code "
                                            + exitValue
                                            + "from bjobs with stderr '"
                                            + stderr
                                            + "'");
        }
        catch (InterruptedException ie)
        {
            cat.error("Caught Exception while getting Job info", ie);
            throw new LSFBatchException("Failed to get Job info from LSF");
        }
        catch (IOException ioe)
        {
            cat.error("Caught Exception while getting Job info", ioe);
            throw new LSFBatchException("Failed to get Job info from LSF");
        }
        finally
        {
            process = null;
        }

        return lsfJobInfo;
    }

    private class Parser implements Runnable
    {
        BufferedReader br;
        Set lsfJobIds;
        List target;

        Parser(InputStream is, Set lsfJobIds, List target)
        {
            br = new BufferedReader(new InputStreamReader(is));
            this.lsfJobIds = lsfJobIds;
            this.target = target;
        }

        public void run()
        {
            String line = null;

            try
            {
                while ((line = br.readLine()) != null)
                {
                    try
                    {
                        LSFJobInfo info = parseLine(line);
                        Integer lsfJobId = new Integer(info.getJobId());

                        if (lsfJobIds.contains(lsfJobId))
                            target.add(info);
                    }
                    catch (NumberFormatException nfe)
                    {
                        cat.error("Caught Exception while parsing line '"
                                  + line
                                  + "'", nfe);
                    }
                }
            }
            catch (IOException ioe)
            {
                cat.error("Caught Exception", ioe);
            }
        }

        private LSFJobInfo parseLine(String line)
        {
            String [] tokens = tokenize(line, ",");

            int jobId = Integer.parseInt(tokens[0]);
            String user = tokens[1];
            int status = Integer.parseInt(tokens[2]);
            String fromHost = tokens[3];
            String [] exHosts = tokenize(tokens[4], " ");
            int exitValue = Integer.parseInt(tokens[5]);

            return new LSFJobInfo(jobId, user, status, exitValue,
                                  fromHost, exHosts);
        }

        private String [] tokenize(String line, String delim)
        {
            if (line.equals("null"))
                return new String [0];

            if (line.indexOf(delim) == -1)
                return new String [] { line };

            StringTokenizer st = new StringTokenizer(line, delim);

            int tokenCount = st.countTokens();
            String [] tokens = new String [tokenCount];

            for (int i = 0; i < tokenCount; i++)
            {
                tokens[i] = st.nextToken();
            }

            return tokens;
        }
    }
}
