/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.lsf;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.run.BufferedStringSink;

/**
 * <code>LSFBsubExec</code> is an alternative to
 * <code>LSFBatchSubmit</code> which does not use JNI, but instead
 * runs the LSF "bsub" program once for each submitted job. 
 *
 * @author Keith James
 */
public final class LSFBsubExec
{
    // Log4j setup
    static Category cat =
        Category.getInstance(LSFBsubExec.class.getName());

    // Singleton instance of this class
    private static final LSFBsubExec INSTANCE = new LSFBsubExec();

    /**
     * Creates a new <code>LSFBsubExec</code>. Private constructor;
     * use getInstance to get the singleton.
     */
    private LSFBsubExec() { }

    /**
     * <code>getInstance</code> returns the singleton instance of this
     * class.
     *
     * @return an <code>LSFBsubExec</code>.
     */
    public static LSFBsubExec getInstance()
    {
        return INSTANCE;
    }

    /**
     * <code>submit</code> submits a job to an LSF batch queue using
     * "bsub".
     *
     * @param jobName a <code>String</code> to be used as the job
     * name.
     * @param queue a <code>String</code> queue name to submit to.
     * @param infile a <code>String</code> file name for input (or
     * null if none is required).
     * @param outFile a <code>String</code> file name for output (or
     * null if none is required).
     * @param errFile a <code>String</code> file name for errors (or
     * null if none is required).
     * @param command a <code>String</code>.
     * @param env a <code>String []</code> of key-value pairs to be
     * used by the <code>Runtime.exec</code> call which runs "bsub".
     * @param workDir a <code>String</code> working directory used by
     * the <code>Runtime.exec</code> call which runs "bsub".
     *
     * @return an <code>int</code> equal to the LSF job ID assigned to
     * the job.
     *
     * @exception Exception if an error occurs
     */
    public synchronized int submit(String jobName, String queue,
                                   String inFile, String outFile, String errFile,
                                   String command, String [] env, String workDir)
        throws Exception
    {
        if (workDir == null)
            throw new IllegalArgumentException("The working directory may not be null");

        Process process;

        List bsubCommand = new ArrayList();
        bsubCommand.add("bsub");
        bsubCommand.add("-J");
        bsubCommand.add(jobName);
        bsubCommand.add("-q");
        bsubCommand.add(queue);

        if (inFile != null)
        {
            bsubCommand.add("-i");
            bsubCommand.add(inFile);
        }

        if (outFile != null)
        {
            bsubCommand.add("-o");
            bsubCommand.add(outFile);
        }

        if (errFile != null)
        {
            bsubCommand.add("-e");
            bsubCommand.add(errFile);
        }

        bsubCommand.add(command);

        cat.debug("Running bsub command '"
                  + bsubCommand
                  + "' in "
                  + workDir);

        File dir = new File(workDir);

        if (!(dir.exists () && dir.isDirectory () && dir.canWrite ())) {
          // Runtime.exec() hangs on Tru64 if the directory doesn't exist
          cat.debug("Directory doesn't exist: " + workDir);
          throw new LSFBatchException("Not writable directory: " + workDir);
        }

        int lsfJobId = -1;

        try
        {
            Runtime runtime = Runtime.getRuntime();

            String [] tokens = (String []) bsubCommand.toArray(new String [0]);
            process = runtime.exec(tokens, env, dir);

            BufferedStringSink stdoutSink =
                new BufferedStringSink(process.getInputStream());
            BufferedStringSink stderrSink =
                new BufferedStringSink(process.getErrorStream());

            stdoutSink.start();
            stderrSink.start();

            int exitValue = process.waitFor();

            stdoutSink.join();
            stderrSink.join();

            String stdout = stdoutSink.getBuffer().substring(0);
            String stderr = stderrSink.getBuffer().substring(0);

            if (exitValue != 0)
                throw new LSFBatchException("Recieved exit code "
                                            + exitValue
                                            + "from bsub with stderr '"
                                            + stderr
                                            + "'");

            try
            {
                lsfJobId = parseBsubStdout(stdout);
            }
            catch (Exception e)
            {
                cat.error("Caught Exception ", e);
                throw new LSFBatchException("Failed to parse new Job ID from '"
                                            + stdout
                                            + "'");
            }

        }
        finally
        {
            process = null;
        }

        return lsfJobId;
    }

    private int parseBsubStdout(String stdout)
    {
        // Job <n> is submitted to queue <name>.

        int lAngle = stdout.indexOf('<');
        int rAngle = stdout.indexOf('>');

        return Integer.parseInt(stdout.substring(lAngle + 1, rAngle));
    }
}
