/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.lsf;

import java.io.Serializable;

/**
 * <code>LSFJobInfo</code> describes an LSF job status using some of
 * the data obtained from the C struct jobInfoEnt returned by the
 * LSBLIB function <code>lsb_readjobinfo</code>.
 *
 * Note that the static job status fields used in this class are
 * currently values. A native method to obtain the actual constants
 * from the C headers should be written and used instead.
 *
 * @author Keith James
 */
public final class LSFJobInfo implements Serializable
{
    /**
     * <code>NULL</code> is the null status value.
     */
    public static final int NULL = 0x00;

    /**
     * <code>PEND</code> is the status value if the job is pending.
     */
    public static final int PEND = 0x01;

    /**
     * <code>PSUSP</code> is the status value if the job is held.
     */
    public static final int PSUSP = 0x02;

    /**
     * <code>RUN</code> is the status value if the job is running.
     */
    public static final int RUN = 0x04;

    /**
     * <code>SSUSP</code> is the status value if the job is suspended
     * by the LSF Batch system.
     */
    public static final int SSUSP = 0x08;

    /**
     * <code>USUSP</code> is the status value if the job is suspended
     * by the user.
     */
    public static final int USUSP = 0x10;

    /**
     * <code>EXIT</code> is the status value if the job has exited.
     */
    public static final int EXIT = 0x20;

    /**
     * <code>DONE</code> is the status value if the job has completed
     * successfully.
     */
    public static final int DONE = 0x40;

    /**
     * <code>PDONE</code> is the status value if the post job process
     * has completed successfully.
     */
    public static final int PDONE = 0x80;

    /**
     * <code>PERR</code> is the status value if the post job process
     * has an error.
     */
    public static final int PERR = 0x100;

    /**
     * <code>WAIT</code> is the status value of a chunk job waiting
     * its turn to exec.
     */
    public static final int WAIT = 0x200;

    /**
     * <code>UNKWN</code> is the status value if the job status is
     * unknown.
     */
    public static final int UNKWN = 0x10000;

    private int jobId;
    private String user;
    private int status;
    private int exitValue;
    private String fromHost;
    private String [] exHosts;

    /**
     * Creates a new <code>LSFJobInfo</code>.
     *
     * @param jobId an <code>int</code> equal to the LSF job ID.
     * @param user a <code>String</code>.
     * @param status an <code>int</code> equal to one of the status
     * constants.
     * @param exitValue an <code>int</code> equal to the exit value of
     * the job, or -1 if it has not yet exited.
     * @param fromHost a <code>String</code>.
     * @param exHosts a <code>String []</code> of host names on which
     * this job is running/ran.
     */
    public LSFJobInfo(int jobId, String user, int status, int exitValue,
                      String fromHost, String [] exHosts)
    {
        this.jobId     = jobId;
        this.user      = user;
        this.status    = status;
        this.exitValue = exitValue;
        this.fromHost  = fromHost;
        this.exHosts   = exHosts;
    }

    /**
     * <code>getJobId</code> returns the LSF job ID.
     *
     * @return an <code>int</code>.
     */
    public int getJobId()
    {
        return jobId;
    }

    /**
     * <code>getUser</code> returns the LSF user name.
     *
     * @return a <code>String</code>.
     */
    public String getUser()
    {
        return user;
    }

    /**
     * <code>getStatus</code> returns the LSF status number. This hex
     * number is equal to one of the public static final fields
     * contained in this class.
     *
     * @return an <code>int</code>.
     */
    public int getStatus()
    {
        return status;
    }

    /**
     * <code>getExitValue</code> returns the exit code of the LSF job.
     *
     * @return an <code>int</code>.
     */
    public int getExitValue()
    {
        return exitValue;
    }

    /**
     * <code>getFromHost</code> returns the host from which the LSF
     * job was submitted.
     *
     * @return a <code>String</code> host name.
     */
    public String getFromHost()
    {
        return fromHost;
    }

    /**
     * <code>getExHosts</code> returns the name (names for a parallel
     * job) of hosts on which the LSF job ran.
     *
     * @return a <code>String []</code> array of host names.
     */
    public String [] getExHosts()
    {
        return exHosts;
    }

    public String toString()
    {
        StringBuffer sb = new StringBuffer();
        sb.append("Job [");
        sb.append(jobId);
        sb.append("] status '");
        sb.append(status);
        sb.append("' for user <");
        sb.append(user);
        sb.append("> from host <");
        sb.append(fromHost);
        sb.append("> exec host <");

        int last = exHosts.length - 1;

        for (int i = 0; i <= last; i++)
        {
            sb.append(exHosts[i]);

            if (i != last)
                sb.append(" ");
        }
        sb.append(">");

        return sb.substring(0);
    }
}
