/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.lsf;

import java.util.Map;
import java.util.List;
import java.util.Iterator;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.dao.StatusDAO;
import uk.ac.sanger.jcon.job.BatchJob;
import uk.ac.sanger.jcon.job.Status;

/**
 * <code>LSFJobUpdater</code> updates <code>Job</code>
 * <code>Status</code> by calling the LSF batch system.
 *
 * @author Keith James
 */
public final class LSFJobUpdater
{
    // Log4j setup
    static Category cat =
        Category.getInstance(LSFJobUpdater.class.getName());

    // Masks all except the Job run states
    static final int runMask =
        ~ (LSFJobInfo.PDONE | LSFJobInfo.PERR | LSFJobInfo.WAIT |
           LSFJobInfo.UNKWN);

    // Masks all except the Job post process states
    static final int postMask =
        ~ (LSFJobInfo.PEND | LSFJobInfo.PSUSP | LSFJobInfo.RUN |
           LSFJobInfo.SSUSP | LSFJobInfo.USUSP | LSFJobInfo.EXIT |
           LSFJobInfo.DONE | LSFJobInfo.WAIT | LSFJobInfo.UNKWN);

    private StatusDAO statusDAO;

    /**
     * Creates a new <code>LSFJobInformer</code> instance.
     *
     * @param statusDAO a <code>StatusDAO</code> to use.
     */
    public LSFJobUpdater(StatusDAO statusDAO)
    {
        this.statusDAO = statusDAO;
    }

    /**
     * <code>updateJobs</code> updates each <code>Job</code> if a
     * significant change has occurred in its state (it has started
     * since last checked or it has completed successfully or exited).
     *
     * @param queue a <code>String</code> queue name.
     * @param jobs a <code>Map</code> of LSF job IDs to the
     * corresponding <code>Job</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void updateJobs(String queue, Map jobs) throws Exception
    {
        if (jobs.isEmpty())
            return;

        // Alternative to using the JNI method in
        // LSFBatchSubmit. Since we upgraded to LSF 4.2 the native
        // methods have been crashing the VM at the point where
        // getMethodID is called.
        LSFBjobsExec lsfBatchInfo = LSFBjobsExec.getInstance();

        // Currently 'queue' not used
        List jobInfo = lsfBatchInfo.getJobInfo(jobs.keySet());

        for (Iterator ii = jobInfo.iterator(); ii.hasNext();)
        {
            LSFJobInfo info = (LSFJobInfo) ii.next();
            Integer lsfJobId = new Integer(info.getJobId());

            BatchJob job = (BatchJob) jobs.get(lsfJobId);

            int lsfStatus = info.getStatus();
            int runStatus = lsfStatus & runMask;
            int postStatus = lsfStatus & postMask;

            int statusId = job.getStatus().getId();

            switch (postStatus)
            {
                case (LSFJobInfo.NULL):
                case (LSFJobInfo.PDONE):
                    break;

                case (LSFJobInfo.PERR):
                    job.setHosts(info.getExHosts());
                    job.setExitValue(info.getExitValue());
                    job.setStatus(statusDAO.readStatusById(Status.FAILED));
                    jobs.remove(lsfJobId);
                    cat.warn("Post processing failed for Job ["
                             + job.getId()
                             + "]");
                    return;

                default:
                    throw new LSFBatchException("Unknown job post process status '"
                                                + lsfStatus
                                                + "' received for Job ["
                                                + job.getId()
                                                + "] from LSF Batch system");
            }

            switch (runStatus)
            {
                case (LSFJobInfo.NULL):
                case (LSFJobInfo.PEND):
                case (LSFJobInfo.PSUSP):
                case (LSFJobInfo.SSUSP):
                case (LSFJobInfo.USUSP):
                case (LSFJobInfo.UNKWN):
                    break;

                case (LSFJobInfo.RUN):
                    if (statusId == Status.READY)
                        job.setStatus(statusDAO.readStatusById(Status.STARTED));
                    break;

                case (LSFJobInfo.EXIT):
                    job.setHosts(info.getExHosts());
                    job.setExitValue(info.getExitValue());
                    job.setStatus(statusDAO.readStatusById(Status.FAILED));
                    jobs.remove(lsfJobId);
                    break;

                case (LSFJobInfo.DONE):
                    job.setHosts(info.getExHosts());
                    job.setExitValue(info.getExitValue());
                    job.setStatus(statusDAO.readStatusById(Status.COMPLETED));
                    jobs.remove(lsfJobId);
                    break;

                default:
                    throw new LSFBatchException("Unknown job run status '"
                                                + lsfStatus
                                                + "' received for Job ["
                                                + job.getId()
                                                + "] from LSF Batch system");
            }
        }
    }
}
