/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.lsf;

import java.util.*;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControl;
import uk.ac.sanger.jcon.dao.StatusDAO;
import uk.ac.sanger.jcon.job.BatchJob;

/**
 * <code>LSFQueueWatcher</code> watches LSF queues for changes in job
 * state.
 *
 * @author Keith James
 */
public class LSFQueueWatcher
{
    /**
     * <code>BATCH_QUEUE_KEY</code> is the <code>Properties</code> key
     * which indicates the name of an LSF batch queue to
     * monitor. Multiple entries of this key may be used to specify
     * several batch queues.
     */
    public static final String BATCH_QUEUE_KEY = "jobcontrol.queue.name"; 

    // Log4j setup
    static Category cat =
        Category.getInstance(LSFQueueWatcher.class.getName());
    // Map of queues to Sets of Jobs in that queue
    private Map queues;
    // Job update interface
    private LSFJobUpdater jobUpdater;
    // Status object source
    private StatusDAO statusDAO;

    /**
     * Creates a new <code>LSFQueueWatcher</code>.
     */
    public LSFQueueWatcher(StatusDAO statusDAO)
    {
        this.statusDAO = statusDAO;
        queues = new HashMap();
        jobUpdater = new LSFJobUpdater(statusDAO);

        ResourceBundle bundle = JobControl.getConfiguration();

        String key;
        String val;

        try
        {
            Enumeration e = bundle.getKeys();

            while (e.hasMoreElements())
            {
                key = (String) e.nextElement();

                try
                {
                    val = bundle.getString(key);
                }
                catch (NullPointerException npe)
                {
                    throw new Error("Malformed configuration ResourceBundle "
                                    + "contains key '"
                                    + key
                                    + "' missing a corresponding value");
                }

                if (key.equals(BATCH_QUEUE_KEY))
                {
                    StringTokenizer st = new StringTokenizer(val);
                    while (st.hasMoreTokens())
                    {
                        String token = st.nextToken();

                        cat.info("Adding queue '"
                                 + token
                                 + "' to known queues");
                        queues.put(token, new HashMap());
                    }
                }
            }
        }
        catch (NumberFormatException nfe)
        {
            throw new Error("Number format error in "
                            + "configuration ResourceBundle");
        }
    }

    /**
     * <code>watchFor</code> adds a <code>batchJob</code> to be watched for
     * on the designated queues.
     *
     * @param job a <code>BatchJob</code>.
     *
     * @exception LSFBatchException if the <code>BatchJob</code>'s
     * queue is unknown or the <code>BatchJob</code> is already being
     * watched.
     */
    public void watchFor(BatchJob job) throws LSFBatchException
    {
        String queue = job.getQueue();

        cat.debug("Starting to watch Job ["
                  + job.getId()
                  + "] on queue "
                  + queue);

        if (! queues.containsKey(queue))
        {
            throw new LSFBatchException("Unknown batch queue '"
                                        + queue
                                        + "'");
        }
        else
        {
            Map jobs = (Map) queues.get(queue);

            synchronized(jobs)
            {
                if (jobs.containsValue(job))
                {
                    throw new LSFBatchException("Batch Job ["
                                                + job.getId()
                                                + "] batch job ID "
                                                + job.getBatchId()
                                                + " is already being watched");
                }
                else
                {
                    // Key Job to an Integer wrapping its LSF job ID
                    jobs.put(new Integer(job.getBatchId()), job);
                }
            }
        }
    }

    public void stopWatching(BatchJob job) throws LSFBatchException
    {
        String queue = job.getQueue();

        cat.debug("Stopping watch over Job ["
                  + job.getId()
                  + "] on queue "
                  + queue);

        if (! queues.containsKey(queue))
        {
            throw new LSFBatchException("Unknown batch queue '"
                                        + queue
                                        + "'");
        }
        else
        {
            Map jobs = (Map) queues.get(queue);

            synchronized(jobs)
            {
                if (! jobs.containsValue(job))
                {
                    throw new LSFBatchException("Batch Job ["
                                                + job.getId()
                                                + "] batch job ID "
                                                + job.getBatchId()
                                                + " is not being watched");
                }
                else
                {
                    jobs.remove(new Integer(job.getBatchId()));
                }
            }
        }
    }

    public void update()
    {
        try
        {
            for (Iterator ki = queues.keySet().iterator(); ki.hasNext();)
            {
                String queue = (String) ki.next();
                Map jobs = (Map) queues.get(queue);

                if (jobs != null)
                {
                    synchronized(jobs)
                    {
                        jobUpdater.updateJobs(queue, jobs);
                    }
                }
            }
        }
        catch (Exception e)
        {
            cat.error("Caught Exception while checking batch queues", e);
        }
    }
}
