/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.manager;

import java.sql.Timestamp;
import java.util.Collection;
import java.util.List;

import uk.ac.sanger.jcon.job.Executable;
import uk.ac.sanger.jcon.job.Owner;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.job.Task;

/**
 * <code>Administrator</code> is an interface for operational
 * management of the system.
 *
 * @author Keith James
 */
public interface Administrator
{
    /**
     * <code>createExecutable</code> creates a new
     * <code>Executable</code> in the store.
     *
     * @param owner an <code>Executable</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void createExecutable(Executable executable) throws Exception;

    /**
     * <code>getExecutableById</code> returns the
     * <code>Executable</code> with the specified ID or null if none
     * exists.
     *
     * @param execId an <code>int</code>.
     *
     * @return an <code>Executable</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Executable getExecutableById(int execId) throws Exception;

    /**
     * <code>getExecutables</code> returns all
     * <code>Executable</code>s.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection getExecutables() throws Exception;

    /**
     * <code>getExecutablesByName</code> returns the
     * <code>Executable</code>s matching the specified name pattern.
     *
     * @param name a <code>String</code> pattern.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs
     */
    public Collection getExecutablesByName(String name) throws Exception;

    /**
     * <code>createOwner</code> creates a new <code>Owner</code> in
     * the store.
     *
     * @param owner an <code>Owner</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void createOwner(Owner owner) throws Exception;

    /**
     * <code>getOwnerById</code> returns the <code>Owner</code> with
     * the specified ID or null if none exists.
     *
     * @param ownerId an <code>int</code>.
     *
     * @return an <code>Owner</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Owner getOwnerById(int ownerId) throws Exception;

   /**
     * <code>getOwners</code> returns all <code>Owner</code>s.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection getOwners() throws Exception;

    /**
     * <code>getOwnerByUserName</code> returns the <code>Owner</code> with
     * the specified user name or null if none exists.
     *
     * @param userName a <code>String</code>.
     *
     * @return an <code>Owner</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Owner getOwnerByUserName(String userName) throws Exception;

    /**
     * <code>getStatusById</code> returns the <code>Status</code> with
     * the specified ID or null if none exists.
     *
     * @param statusId an <code>int</code>.
     *
     * @return a <code>Status</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Status getStatusById(int statusId) throws Exception;

    /**
     * <code>getStatus</code> returns all <code>Status</code>s.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection getStatus() throws Exception;

    /**
     * <code>getTaskById</code> returns the <code>Task</code> with the
     * specified ID or null if none exists.
     *
     * @param taskId an <code>int</code>.
     *
     * @return a <code>Task</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Task getTaskById(int taskId) throws Exception;

    /**
     * <code>getTasksByOwner</code> returns all <code>Task</code>s
     * belonging to the specified <code>Owner</code>.
     *
     * @param owner an <code>Owner</code>.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection getTasksByOwner(Owner owner) throws Exception;

    /**
     * <code>getTasksByName</code> returns all <code>Task</code>s
     * matching the specified name pattern.
     *
     * @param name a <code>String</code> pattern.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection getTasksByName(String name) throws Exception;

    /**
     * <code>getTasksByPriority</code> returns all <code>Task</code>s
     * with the specified priority.
     *
     * @param priority an <code>int</code>
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection getTasksByPriority(int priority) throws Exception;

    /**
     * <code>getTasksSubmittedBefore</code> returns all
     * <code>Task</code> submitted before the specified time.
     *
     * @param time a <code>Timestamp</code>.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection getTasksSubmittedBefore(Timestamp time) throws Exception;

    /**
     * <code>getTasksSubmittedAfter</code> returns all
     * <code>Task</code> submitted after the specified time.
     *
     * @param time a <code>Timestamp</code>.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection getTasksSubmittedAfter(Timestamp time) throws Exception;

    /**
     * <code>getTasksSubmittedBetween</code> returns all
     * <code>Task</code> submitted between the specified times.
     *
     * @param earliest a <code>Timestamp</code>.
     * @param latest a <code>Timestamp</code>.
     *
     * @return a <code>Collection</code>.
     *
     * @exception Exception if an error occurs.
     */
    public Collection getTasksSubmittedBetween(Timestamp earliest,
                                               Timestamp latest)
        throws Exception;
}
