/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 *
 */

package uk.ac.sanger.jcon.protocol;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Map;
import javax.naming.NamingException;

import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.EmblLikeFormat;
import org.biojava.bio.seq.io.FastaFormat;
import org.biojava.bio.seq.io.SequenceFormat;

import uk.ac.sanger.jcon.JobControl;
import uk.ac.sanger.jcon.job.BatchJob;
import uk.ac.sanger.jcon.job.JobBatchImpl;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.manager.Administrator;
import uk.ac.sanger.jcon.manager.TaskManager;

/**
 * <code>AbstractProtocol</code> implements some utility methods which
 * will be common to many <code>Protocol</code>s.
 *
 * @author Keith James
 */
abstract class AbstractProtocol implements Protocol
{
	protected JobControl jobControl;
	protected Administrator admin;
	protected TaskManager manager;

	/**
	 * Creates a new <code>AbstractProtocol</code> instance.
	 *
	 * @exception Exception if an error occurs when resolving the
	 * required resources.
	 */
	public AbstractProtocol() throws Exception
	{
		jobControl = new JobControl();
		admin   = jobControl.createAdministrator();
		manager = jobControl.createTaskManager();
	}

	/**
	 * <code>createEmptyJob</code> creates a <code>BatchJob</code>
	 * which does nothing apart from run the Unix 'true'
	 * command. Output and error are sent to /dev/null.
	 *
	 * @param queue a <code>String</code> batch queue name.
	 *
	 * @return a <code>Job</code>.
	 *
	 * @exception IOException if an error occurs.
	 */
	protected BatchJob createEmptyBatchJob(String queue) throws Exception
	{
		BatchJob job = new JobBatchImpl(admin.getExecutableById(1));
		job.setStatus(admin.getStatusById(Status.WAITING));
		job.setOutputName("/dev/null");
		job.setCommandTemplate("$executable");
		job.setQueue(queue);

		return job;
	}

	/**
	 * <code>writeFasta</code> writes a sequence to a file, in Fasta
	 * format.
	 *
	 * @param seq a <code>Sequence</code> to write.
	 * @param file a <code>File</code>.
	 *
	 * @exception IOException if an error occurs.
	 */
	protected void writeFasta(Sequence seq, File file)
		throws IOException
	{
		PrintStream ps = new PrintStream(new FileOutputStream(file));

		SequenceFormat format = new FastaFormat();
		format.writeSequence(seq, ps);
	}

	/**
	 * <code>writeEMBL</code> writes a sequence to a file, in EMBL
	 * format.
	 *
	 * @param seq a <code>Sequence</code> to write.
	 * @param file a <code>File</code>.
	 *
	 * @exception IOException if an error occurs.
	 */
	protected void writeEMBL(Sequence seq, File file)
		throws IOException
	{
		PrintStream ps = new PrintStream(new FileOutputStream(file));

		SequenceFormat format = new EmblLikeFormat();
		format.writeSequence(seq, ps);
	}

    protected File createAbsoluteFile(String dir, String fileName)
    {
        return new File(dir, fileName).getAbsoluteFile();
    }

	public abstract void apply(String ownerName, Map config, String [] data)
		throws Exception;
}
