/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.protocol;

import java.io.File;
import java.io.StringWriter;
import java.util.Map;
import java.util.Properties;
import java.util.ResourceBundle;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.sql.DataSource;

import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.VelocityEngine;

import uk.ac.sanger.jcon.dao.ExecutableDAO;
import uk.ac.sanger.jcon.dao.ExecutableSQLDAO;
import uk.ac.sanger.jcon.dao.StatusDAO;
import uk.ac.sanger.jcon.dao.StatusSQLDAO;
import uk.ac.sanger.jcon.job.*;
import uk.ac.sanger.jcon.util.Configuration;

/**
 * <code>ConfigurableProtocol</code> is configurable with the correct
 * properties file. Subclasses define actual jobs and submission.
 *
 * @author Keith James
 *
 * @see uk.ac.sanger.jcon.util.Configuration
 */
abstract class ConfigurableProtocol extends AbstractProtocol
{
    /**
     * <code>VELOCITY_PROPS</code> the properties file containing the
     * Velocity configuration used to initialise the Velocity engine.
     */
    public static final String VELOCITY_PROPS = "/velocity.properties";

    /**
     * <code>LOG_TAG</code> the log-tag used in log statements issued
     * by the Velocity engine.
     */
    public static final String LOG_TAG = "CMD_TMPL_GEN";

	static ResourceBundle resource =
		ResourceBundle.getBundle(ConfigurableProtocol.class.getName());

	protected Configuration config;
	private ExecutableDAO execDAO;
	private StatusDAO statusDAO;

	private VelocityEngine vEngine;

    ConfigurableProtocol() throws Exception
	{
        execDAO   = jobControl.createExecutableDAO();
		statusDAO = jobControl.createStatusDAO();

		config = new Configuration(resource);

		Properties velocityProps = new Properties();
		velocityProps.load(getClass().getResourceAsStream(VELOCITY_PROPS));
		vEngine = new VelocityEngine();
		vEngine.init(velocityProps);
	}

	/**
	 * <p><code>createJob</code> creates a new <code>Job</code> using
	 * the supplied configuration. The <code>Job</code> will have
	 * default field values with the following exceptions:
	 * input/output/errorName will be set as configured,
	 * <code>Executable</code> will be set as configured, batch queue
	 * name will be set as configured, command template will be set as
	 * configured and <code>Status</code> will be set to WAITING.</p>
	 *
	 * @param baseNames a <code>String []</code> array consisting of
	 * the base filenames for the input, output and error files. These
	 * should be in that order, starting at index 0. Nulls should be
	 * used where no filename is appropriate.
	 * @param organismAlias a <code>String</code> alias specified in
	 * the configuration file.
	 * @param execAlias a <code>String</code> alias specified in
	 * the configuration file.
	 *
	 * @return a <code>Job</code>.
	 *
	 * @exception Exception if an error occurs.
	 */
	Job createJob(String [] baseNames,
                  String organismAlias, String execAlias)
		throws Exception
	{
		Job job = _createJob(baseNames, organismAlias, execAlias);
		job.setCommandTemplate(config.getTemplate(organismAlias, execAlias));

		return job;
	}

	/**
	 * <p><code>createJob</code> creates a new <code>Job</code> using
	 * the supplied configuration. The <code>Job</code> will have
	 * default field values with the following exceptions:
	 * input/output/errorName will be set as configured,
	 * <code>Executable</code> will be set as configured, batch queue
	 * name will be set as configured, command template will be set as
	 * configured and <code>Status</code> will be set to WAITING.</p>
	 *
	 * <p>The <code>VelocityContext</code> is used to preprocess the
	 * command template prior to assignment to the new
	 * <code>Job</code>.
	 *
	 * @param baseNames a <code>String []</code> array consisting of
	 * the base filenames for the input, output and error files. These
	 * should be in that order, starting at index 0. Nulls should be
	 * used where no filename is appropriate.
	 * @param context a <code>VelocityContext</code>.
	 * @param organismAlias a <code>String</code> alias specified in
	 * the configuration file.
	 * @param execAlias a <code>String</code> alias specified in
	 * the configuration file.
	 *
	 * @return a <code>Job</code>.
	 *
	 * @exception Exception if an error occurs.
	 */
	Job createJob(String [] baseNames, VelocityContext context,
                  String organismAlias, String execAlias)
		throws Exception
	{
		Job job = _createJob(baseNames, organismAlias, execAlias);

		String template = config.getTemplate(organismAlias, execAlias);

		StringWriter writer = new StringWriter(256);
		vEngine.evaluate(context, writer, LOG_TAG, template);
		job.setCommandTemplate(writer.toString());

		return job;
	}

	private Job _createJob(String [] baseNames,
						   String organismAlias, String execAlias)
		throws Exception
	{
		int  taxId = config.getTaxonomyId(organismAlias);
		int execId = config.getExecutableId(execAlias);

		Executable exec = execDAO.readExecutableById(execId);
		Status waiting = statusDAO.readStatusById(Status.WAITING);
		String queueName = config.getString(Configuration.EXEC_QUEUE_KEY);

		BatchJob job = new JobBatchImpl(exec);
		job.setStatus(waiting);
		job.setQueue(queueName);

		if (baseNames[0] != null)
			job.setInputName(baseNames[0]);

		if (baseNames[1] != null)
			job.setOutputName(baseNames[1]);

		if (baseNames[2] != null)
			job.setErrorName(baseNames[2]);

		return job;
	}
}
