/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.protocol;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.Map;

import org.apache.log4j.Category;
import org.apache.velocity.VelocityContext;

import org.biojava.bio.BioException;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.SeqIOTools;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.job.*;
import uk.ac.sanger.jcon.util.Configuration;

/**
 * <code>DdiscoideumDNABlast</code> is a protocol for running Blast
 * searches on D discoideum Fasta format DNA sequences.
 *
 * @author Keith James
 */
public class DdiscoideumDNABlast extends ConfigurableProtocol
{
	// Log4j setup
    static Category cat =
		Category.getInstance(DdiscoideumDNABlast.class.getName());

	/**
	 * Creates a new <code>DdiscoideumDNABlast</code> instance.
	 *
	 * @exception Exception if an error occurs when resolving the
	 * required resources.
	 */
	public DdiscoideumDNABlast() throws Exception
	{
		super();
	}

	/**
	 * <code>apply</code> applies the protocol to a the sequences in
	 * files named by the argument strings. Each file may contain many
	 * DNA sequences. There is no limit on the number of arguments.
	 *
	 * @param ownerName a <code>String</code> username.
     * @param config a <code>Map</code> of configuration parameters
     * and corresponding values.
	 * @param data a <code>String []</code> array of filenames.
	 *
	 * @exception Exception if an error occurs.
	 */
	public void apply(String ownerName, Map config, String [] data)
        throws Exception
	{
		if (ownerName == null)
			throw new IllegalArgumentException("ownerName was null");

		Owner owner = admin.getOwnerByUserName(ownerName);

		for (int i = 0; i < data.length; i++)
		{
			try
			{
				applySingle(owner, data[i]);
			}
			catch (Exception e)
			{
				System.err.println("Exception while processing '"
								   + data[i]
								   + "'. Skipping this element");
				e.printStackTrace();
			}
		}
	}

	private void applySingle(Owner owner, String arg) throws Exception
	{
		BufferedReader br = new BufferedReader(new FileReader(arg));
		SequenceIterator seqI = SeqIOTools.readFastaDNA(br);

		String ntSuffix = config.getString(Configuration.FASTA_NT_EXTENSION_KEY);
		String tabSuffix = config.getString(Configuration.FTABLE_EXTENSION_KEY);
		String inpSuffix = config.getString(Configuration.INPUT_EXTENSION_KEY);
		String outSuffix = config.getString(Configuration.OUTPUT_EXTENSION_KEY);
		String errSuffix = config.getString(Configuration.ERROR_EXTENSION_KEY);
		String queue = config.getString(Configuration.EXEC_QUEUE_KEY);

        String workDirectory = System.getProperty("user.dir");

		String repeatsLogName = "blastn_repeats.log";
		String rdnaLogName = "blastn_rdna.log";
		String estsLogName = "blastn_ests.log";
		String happymapLogName = "blastn_happymap.log";
		String swallLogName = "blastx_swall.log";
		String blast2tabLogName = "blast2tab.log";

		Job rootJob = createEmptyBatchJob(queue);
        rootJob.setWorkDirectoryName(workDirectory);

		while (seqI.hasNext())
		{
			Sequence seq = seqI.nextSequence();
			String name = seq.getName();

			if (name.equals(""))
				throw new JobControlException("Sequence used empty string as identifier");

			String ntFileName = name + "." + ntSuffix;

            File ntFile = new File(workDirectory, ntFileName);
			if (! ntFile.exists())
				writeFasta(seq, ntFile);

			String repeatsOutName = name + ".repeats.bln." + outSuffix;
			String rdnaOutName = name + ".rdna.bln." + outSuffix;
			String estsOutName = name + ".ests.bln." + outSuffix;
			String happymapOutName = name + ".happymap.bln." + outSuffix;
			String swallOutName = name + ".swall.blx." + outSuffix;

			VelocityContext context = new VelocityContext();
			// Make this a valid reference to keep paranoid logger
			// quiet
			context.put("executable",
						"<Error: this should not have been expanded>");

			// repeats
			context.put("output_name", repeatsOutName);
			Job repeatsRun =
				createJob(new String [] { ntFileName,
                                          repeatsLogName,
                                          null },
                          context, "d_discoideum", "blastn_v_repeats");
            repeatsRun.setWorkDirectoryName(workDirectory);

			context.put("output_name", repeatsOutName + "." + tabSuffix);
			context.put("ident_filter", "0");
			context.put("score_filter", "100");
            context.put("eval_filter", "1");
			context.put("feature_key", "repeat_region");
			Job repeatsParse =
				createJob(new String [] { repeatsOutName,
                                          blast2tabLogName,
                                          null },
                          context, "d_discoideum", "blast2tab");
            repeatsParse.setWorkDirectoryName(workDirectory);
			repeatsParse.add(repeatsRun);
			rootJob.add(repeatsParse);

			// rDNA
			context.put("output_name", rdnaOutName);
			Job rdnaRun =
				createJob(new String [] { ntFileName,
                                          rdnaLogName,
                                          null },
                          context, "d_discoideum", "blastn_v_rdna");

            rdnaRun.setWorkDirectoryName(workDirectory);

			context.put("output_name", rdnaOutName + "." + tabSuffix);
			context.put("ident_filter", "90");
			context.put("feature_key", "rRNA");
			Job rdnaParse =
				createJob(new String [] { rdnaOutName,
                                          blast2tabLogName,
                                          null },
                          context, "d_discoideum", "blast2tab");

            rdnaParse.setWorkDirectoryName(workDirectory);
			rdnaParse.add(rdnaRun);
			rootJob.add(rdnaParse);

			// ESTs
			context.put("output_name", estsOutName);
			Job estsRun =
				createJob(new String [] { ntFileName,
                                          estsLogName,
                                          null },
                          context, "d_discoideum", "blastn_v_ests");

            estsRun.setWorkDirectoryName(workDirectory);

			context.put("output_name", estsOutName + "." + tabSuffix);
			context.put("ident_filter", "0");
			context.put("feature_key", "misc_feature");
			Job estsParse =
				createJob(new String [] { estsOutName,
                                          blast2tabLogName,
                                          null },
                          context, "d_discoideum", "blast2tab");

            estsParse.setWorkDirectoryName(workDirectory);
			estsParse.add(estsRun);
			rootJob.add(estsParse);

			// happy map
			context.put("output_name", happymapOutName);
			Job happymapRun =
				createJob(new String [] { ntFileName,
                                          happymapLogName,
                                          null },
                          context, "d_discoideum", "blastn_v_happymap");

            happymapRun.setWorkDirectoryName(workDirectory);

			context.put("output_name", happymapOutName + "." + tabSuffix);
			context.put("ident_filter", "90");
			context.put("feature_key", "misc_feature");
			Job happymapParse =
				createJob(new String [] { happymapOutName,
                                          blast2tabLogName,
                                          null },
                          context, "d_discoideum", "blast2tab");

            happymapParse.setWorkDirectoryName(workDirectory);
			happymapParse.add(happymapRun);
			rootJob.add(happymapParse);

			// blastx SWALL
			context.put("output_name", swallOutName);
			Job swallRun =
				createJob(new String [] { ntFileName,
                                          swallLogName,
                                          null },
                          context, "d_discoideum", "blastx_v_swall");

            swallRun.setWorkDirectoryName(workDirectory);

			context.put("output_name", swallOutName + "." + tabSuffix);
			context.put("ident_filter", "0");
			context.put("feature_key", "BLASTCDS");
			Job swallParse =
				createJob(new String [] { swallOutName,
                                          blast2tabLogName,
                                          null },
                          context, "d_discoideum", "blast2tab");

            swallParse.setWorkDirectoryName(workDirectory);
			swallParse.add(swallRun);
			rootJob.add(swallParse);
		}

		Task task = new TaskDefaultImpl("d_discoideum_dnablast",
										"D. discoideum DNA contig Blast searches",
										owner);

		task.setJobs(new Job [] { rootJob });

		int taskId = manager.createTask(task);
		manager.submitTask(taskId);
	}
}
