/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.protocol;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.Map;

import org.apache.log4j.Category;
import org.apache.velocity.VelocityContext;

import org.biojava.bio.BioException;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.SeqIOTools;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.Owner;
import uk.ac.sanger.jcon.job.Task;
import uk.ac.sanger.jcon.job.TaskDefaultImpl;
import uk.ac.sanger.jcon.util.Configuration;

/**
 * <code>DdiscoideumDeNovo</code> is a protocol for running de novo
 * gene predictions on D discoideum Fasta format DNA sequences.
 *
 * @author Keith James
 */
public class DdiscoideumDeNovo extends ConfigurableProtocol
{
	// Log4j setup
    static Category cat =
		Category.getInstance(DdiscoideumDeNovo.class.getName());

	/**
	 * Creates a new <code>DdiscoideumDeNovo</code> instance.
	 *
	 * @exception Exception if an error occurs when resolving the
	 * required resources.
	 */
	public DdiscoideumDeNovo() throws Exception
	{
		super();
	}

	/**
	 * <code>apply</code> applies the protocol to a the sequences in
	 * files named by the argument strings. Each file may contain many
	 * DNA sequences. There is no limit on the number of arguments.
	 *
	 * @param ownerName a <code>String</code> username.
	 * @param args a <code>String []</code> array of filenames.
	 * @param config a <code>Map</code> of configuration parameters
     * and corresponding values.
     *
	 * @exception Exception if an error occurs.
	 */
	public void apply(String ownerName, Map config, String [] data)
        throws Exception
	{
		if (ownerName == null)
			throw new IllegalArgumentException("ownerName was null");

		Owner owner = admin.getOwnerByUserName(ownerName);

		for (int i = 0; i < data.length; i++)
		{
			try
			{
				applySingle(owner, data[i]);
			}
			catch (Exception e)
			{
				System.err.println("Exception while processing '"
								   + data[i]
								   + "'. Skipping this element");
				e.printStackTrace();
			}
		}
	}

	private void applySingle(Owner owner, String arg) throws Exception
	{
		BufferedReader br = new BufferedReader(new FileReader(arg));
		SequenceIterator seqI = SeqIOTools.readFastaDNA(br);

		String ntSuffix = config.getString(Configuration.FASTA_NT_EXTENSION_KEY);
		String aaSuffix = config.getString(Configuration.FASTA_AA_EXTENSION_KEY);
		String emSuffix = config.getString(Configuration.EMBL_EXTENSION_KEY);
        String tabSuffix = config.getString(Configuration.FTABLE_EXTENSION_KEY);
		String inpSuffix = config.getString(Configuration.INPUT_EXTENSION_KEY);
		String outSuffix = config.getString(Configuration.OUTPUT_EXTENSION_KEY);
		String errSuffix = config.getString(Configuration.ERROR_EXTENSION_KEY);
		String queue = config.getString(Configuration.EXEC_QUEUE_KEY);

        String workDirectory = System.getProperty("user.dir");

// 		String gfLogName =
//             createAbsoluteFile(dir, "genefinder.log").toString();
// 		String gf2tabLogName =
//             createAbsoluteFile(dir, "gf2tab.log").toString();
		String geneidLogName = "geneid.log";
		String geneid2tabLogName = "geneid2tab.log";
// 		String hmmgeneLogName =
//             createAbsoluteFile(dir, "hmmgene.log").toString();
// 		String hmmgene2tabLogName =
//             createAbsoluteFile(dir, "hmmgene2tab.log").toString();
		String merge2emblLogName = "merge2embl.log";
        String createidsLogName = "createids.log";
		String translateLogName = "translate.log";
        String catLogName = "cat.log";
		String trnascanLogName = "trnascan.log";

        VelocityContext context = new VelocityContext();
        // Make 'executable' a valid reference to keep paranoid
        // logger quiet
        context.put("executable",
                    "<Error: this should not have been expanded>");

        // genefinders
//         context.put("input_name1", dir);
//         context.put("input_name2", "*." + ntSuffix);
//         Job genefinderRun =
//             jobFactory.createJob(new String [] { null,
//                                                  gfLogName,
//                                                  null },
//                                  context, "d_discoideum", "genefinder");
		Job rootJob = createEmptyBatchJob(queue);
        rootJob.setWorkDirectoryName(workDirectory);

		while (seqI.hasNext())
		{
			Sequence seq = seqI.nextSequence();
			String name = seq.getName();

			if (name.equals(""))
				throw new JobControlException("Sequence used empty string as identifier");

            String ntFileName = name + "." + ntSuffix;

            File ntFile = new File(workDirectory, ntFileName);

			if (! ntFile.exists())
				writeFasta(seq, ntFile);

// 			String genefinderOutName =
//                 createAbsoluteFile(dir, name + ".genefinder." + outSuffix).toString();
//             String gf2tabOutName =
//                 createAbsoluteFile(dir, name + ".genefinder.cds." + tabSuffix).toString();
			String geneidOutName = name + ".geneid." + outSuffix;
            String geneid2tabOutName = name + ".geneid.cds." + tabSuffix;
// 			String hmmgeneOutName =
//                 createAbsoluteFile(dir, name + ".hmmgene." + outSuffix).toString();
//             String hmmgene2tabOutName =
//                 createAbsoluteFile(dir, name + ".hmmgene.cds." + tabSuffix).toString();
			String trnascanOutName = name + ".trnscan." + outSuffix;
			String merge2emblOutName = name + "." + emSuffix;
            String createidsOutName = name + ".ids." + emSuffix;
			String translateOutName = name + "." + aaSuffix;
            String concatTranslateOutName = arg + "." + aaSuffix;
            String trnascanParseOutName = name + ".trnas";

			context = new VelocityContext();
			// Make 'executable' a valid reference to keep paranoid
			// logger quiet
			context.put("executable",
						"<Error: this should not have been expanded>");

// 			context.put("input_name1", genefinderOutName);
// 			context.put("input_name2", ntFileName);
//             context.put("output_name1", gf2tabOutName);
//             context.put("output_name2", "/dev/null");
// 			Job genefinderParse =
// 				createJob(new String [] { null,
// 										  gf2tabLogName,
// 										  null },
// 						  context, "d_discoideum", "gf2tab");

			// geneid
			context.put("input_name", ntFileName);
			context.put("output_name", geneidOutName);
			Job geneidRun =
				createJob(new String [] { null,
                                          geneidLogName,
                                          null },
                          context, "d_discoideum", "geneid");

            geneidRun.setWorkDirectoryName(workDirectory);

			context.put("input_name", geneidOutName);
			context.put("output_name1", geneid2tabOutName);
            context.put("output_name2", "/dev/null");
			Job geneidParse =
				createJob(new String [] { null,
                                          geneid2tabLogName,
                                          null },
                          context, "d_discoideum", "geneid2tab");

            geneidParse.setWorkDirectoryName(workDirectory);
			geneidParse.add(geneidRun);

			// hmmgene
// 			context.put("input_name", ntFileName);
// 			context.put("output_name", hmmgeneOutName);
// 			Job hmmgeneRun =
// 				createJob(new String [] { ntFileName,
// 										  hmmgeneLogName,
// 										  null },
// 						  context, "d_discoideum", "hmmgene");

// 			context.put("input_name1", hmmgeneOutName);
// 			context.put("input_name2", ntFileName);
// 			context.put("output_name1", hmmgene2tabOutName);
//             context.put("output_name2", "/dev/null");
// 			Job hmmgeneParse =
// 				createJob(new String [] { null,
// 										  hmmgene2tabLogName,
// 									   	  null },
// 						  context, "d_discoideum", "hmmgene2tab");

// 			hmmgeneParse.add(hmmgeneRun);

			// Merge once parsing is done
// 			context.put("input_name1", gf2tabOutName);
 			context.put("input_name2", geneid2tabOutName);
// 			context.put("input_name2", hmmgene2tabOutName);
// 			context.put("output_name", merge2emblOutName);
			Job merge2emblRun =
				createJob(new String [] { ntFileName,
                                          merge2emblLogName,
                                          null },
                          context, "d_discoideum", "merge2embl");

            merge2emblRun.setWorkDirectoryName(workDirectory);
// 			merge2emblRun.add(genefinderParse);
 			merge2emblRun.add(geneidParse);
// 			merge2emblRun.add(hmmgeneParse);

            // Create IDs
            context.put("input_name", name);
            context.put("output_name", createidsOutName);
            Job createidsRun =
                createJob(new String [] { merge2emblOutName,
                                          createidsLogName,
                                          null },
                          context, "d_discoideum", "create_ids");

            createidsRun.setWorkDirectoryName(workDirectory);
            createidsRun.add(merge2emblRun);

			// Translate features in merged EMBL file
			context.put("output_name", translateOutName);
			Job translateRun =
				createJob(new String [] { createidsOutName,
                                          translateLogName,
                                          null },
                          context, "d_discoideum", "translate_features");

            translateRun.setWorkDirectoryName(workDirectory);
			translateRun.add(createidsRun);
			rootJob.add(translateRun);

			// trnscan
			context.put("input_name", ntFileName);
			context.put("output_name", trnascanOutName);
			Job trnascanRun =
				createJob(new String [] { null,
                                          trnascanLogName,
                                          null },
                          context, "d_discoideum", "trnascanse");

            trnascanRun.setWorkDirectoryName(workDirectory);

			context.put("output_name", trnascanParseOutName);
			Job trnascanParse =
				createJob(new String [] { trnascanOutName,
                                          trnascanLogName,
                                          null },
                          context, "d_discoideum", "trnascanse2tab");

            trnascanParse.setWorkDirectoryName(workDirectory);
			trnascanParse.add(trnascanRun);
			rootJob.add(trnascanParse);
		}

		Task task = new TaskDefaultImpl("d_discoideum_denovo",
										"D. discoideum de novo gene prediction",
										owner);

//         task.setJobs(new Job [] { rootJob, genefinderRun });
        task.setJobs(new Job [] { rootJob });

		int taskId = manager.createTask(task);
		manager.submitTask(taskId);
	}
}
