/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.protocol;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.Map;

import org.apache.log4j.Category;
import org.apache.velocity.VelocityContext;

import org.biojava.bio.BioException;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.SeqIOTools;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.Owner;
import uk.ac.sanger.jcon.job.Task;
import uk.ac.sanger.jcon.job.TaskDefaultImpl;
import uk.ac.sanger.jcon.util.Configuration;

/**
 * <code>DdiscoideumPeptideAnalysis</code> is a protocol for running
 * Pfam and Blast searches on D discoideum Fasta format peptide
 * sequences.
 *
 * @author Keith James
 */
public class DdiscoideumPeptideAnalysis extends ConfigurableProtocol
{
	// Log4j setup
    static Category cat =
		Category.getInstance(DdiscoideumPeptideAnalysis.class.getName());

	/**
	 * Creates a new <code>DdiscoideumPeptideAnalysis</code> instance.
	 *
	 * @exception Exception if an error occurs when resolving the
	 * required resources.
	 */
	public DdiscoideumPeptideAnalysis() throws Exception
	{
		super();
	}

	/**
	 * <code>apply</code> applies the protocol to a the sequences in
	 * files named by the argument strings. Each file may contain many
	 * DNA sequences. There is no limit on the number of arguments.
	 *
	 * @param ownerName a <code>String</code> username.
     * @param config a <code>Map</code> of configuration parameters
     * and corresponding values.
	 * @param data a <code>String []</code> array of filenames.
	 *
	 * @exception Exception if an error occurs.
	 */
	public void apply(String ownerName, Map config, String [] data)
        throws Exception
	{
		if (ownerName == null)
			throw new IllegalArgumentException("ownerName was null");

		Owner owner = admin.getOwnerByUserName(ownerName);

		for (int i = 0; i < data.length; i++)
		{
			try
			{
				applySingle(owner, data[i]);
			}
			catch (Exception e)
			{
				System.err.println("Exception while processing '"
								   + data[i]
								   + "'. Skipping this element");
				e.printStackTrace();
			}
		}
	}

	private void applySingle(Owner owner, String arg) throws Exception
	{
		BufferedReader br = new BufferedReader(new FileReader(arg));
		SequenceIterator seqI = SeqIOTools.readFastaProtein(br);

		String aaSuffix = config.getString(Configuration.FASTA_AA_EXTENSION_KEY);
        String tabSuffix = config.getString(Configuration.FTABLE_EXTENSION_KEY);
		String inpSuffix = config.getString(Configuration.INPUT_EXTENSION_KEY);
		String outSuffix = config.getString(Configuration.OUTPUT_EXTENSION_KEY);
		String errSuffix = config.getString(Configuration.ERROR_EXTENSION_KEY);
		String queue = config.getString(Configuration.EXEC_QUEUE_KEY);

        String workDirectory = System.getProperty("user.dir");

//         String pfamscanLogName = "pfamscan.log";
//         String pfamscan2tabLogName = "pfamscan2tab.log";
		String swallLogName = "blastp_swall.log";
		String blast2tabLogName = "blast2tab.log";
        String pepfeaturemapLogName = "pepfeaturemap.log";

		Job rootJob = createEmptyBatchJob(queue);
        rootJob.setWorkDirectoryName(workDirectory);

		while (seqI.hasNext())
		{
            Sequence seq = seqI.nextSequence();
            String name = seq.getName();

            if (name.equals(""))
				throw new JobControlException("Sequence used empty string as identifier");

            String aaFileName = name + "." + aaSuffix;

            File aaFile = new File(workDirectory, aaFileName);
            if (! aaFile.exists())
				writeFasta(seq, aaFile);

            VelocityContext context = new VelocityContext();
			// Make this a valid reference to keep paranoid logger
			// quiet
			context.put("executable",
						"<Error: this should not have been expanded>");

            // String pfamscanOutName = name + ".pfam." + outSuffix;
            String swallOutName = name + ".swall.blp." + outSuffix;

            // pfamscan
//             context.put("output_name", pfamscanOutName);
//             Job pfamscanRun =
//                 createJob(new String [] { aaFileName,
// 										   	 pfamscanLogName,
// 											 null },
// 							 context, "d_discoideum", "pfamscan");

//             pfamscanRun.setWorkDirectoryName(workDirectory);

//             context.put("output_name", pfamscanOutName + "." + tabSuffix);
//             Job pfamscanParse =
//                 createJob(new String [] { pfamscanOutName,
// 											 pfamscan2tabLogName,
// 											 null },
// 							 context, "d_discoideum", "pfamscan2tab");

//             pfamscanParse.setWorkDirectoryName(workDirectory);
//             pfamscanParse.add(pfamscanRun);
//             rootJob.add(pfamscanParse);

            // blastp SWALL
            context.put("output_name", swallOutName);
            Job swallRun =
                createJob(new String [] { aaFileName,
                                          swallLogName,
                                          null },
                          context, "d_discoideum", "blastp_v_swall");

            swallRun.setWorkDirectoryName(workDirectory);

            context.put("output_name", swallOutName + "." + tabSuffix);
            context.put("ident_filter", "0");
			context.put("score_filter", "0");
			context.put("eval_filter", "10e-5");
			context.put("feature_key", "BLASTCDS");
            Job swallParse =
                createJob(new String [] { swallOutName,
                                          blast2tabLogName,
                                          null },
                          context, "d_discoideum", "blast2tab");

            swallParse.setWorkDirectoryName(workDirectory);
            swallParse.add(swallRun);
            rootJob.add(swallParse);
		}

		Task task = new TaskDefaultImpl("d_discoideum_pepanalysis",
										"D. discoideum peptide analysis",
										owner);

		task.setJobs(new Job [] { rootJob });

		int taskId = manager.createTask(task);
		manager.submitTask(taskId);
	}
}
