/*
 * Copyright (c) 2003 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.protocol;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.util.List;
import java.util.Map;
import java.util.PropertyResourceBundle;
import java.util.ResourceBundle;

import org.apache.log4j.Category;

import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.io.SeqIOTools;
import org.biojava.bio.seq.io.SequenceBuilderFactory;
import org.biojava.bio.seq.io.SequenceFormat;
import org.biojava.bio.seq.io.StreamReader;
import org.biojava.bio.seq.io.SymbolTokenization;
import org.biojava.bio.symbol.Alphabet;

import uk.ac.sanger.jcon.JSubmit;
import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.dao.*;
import uk.ac.sanger.jcon.job.*;
import uk.ac.sanger.jcon.util.UserConfiguration;

/**
 * <code>ProtocolUserImpl</code> is configurable by means of a
 * user-supplied properties file whose format is described in the user
 * guide. It allows simple chaining of arbitrary jobs where the
 * executables involved communicate data only via STDIN/STDOUT.
 *
 * @author Keith James
 */
public class ProtocolUserImpl extends AbstractProtocol
{
    public static final String  INPUT_SUFFIX = "in";
    public static final String OUTPUT_SUFFIX = "out";
    public static final String  ERROR_SUFFIX = "err";
    public static final String    LOG_SUFFIX = "log";

	// Log4j setup
    static Category cat =
		Category.getInstance(ProtocolUserImpl.class.getName());

	private ExecutableDAO execDAO;
	private StatusDAO statusDAO;

    /**
     * Creates a new <code>ProtocolUserImpl</code> instance.
     *
     * @exception Exception if an error occurs.
     */
    public ProtocolUserImpl() throws Exception
    {
		execDAO   = jobControl.createExecutableDAO();
		statusDAO = jobControl.createStatusDAO();
    }

	public void apply(String ownerName, Map config, String [] data)
        throws Exception
	{
		if (ownerName == null)
			throw new IllegalArgumentException("ownerName was null");

		Owner owner = admin.getOwnerByUserName(ownerName);

        if (! config.containsKey(JSubmit.CONFIG_KEY))
            throw new IllegalArgumentException("Config map did not "
                                               + "contain the mandatory key '"
                                               + JSubmit.CONFIG_KEY
                                               + "'");

        String configFileName = (String) config.get(JSubmit.CONFIG_KEY);
        File configFile = new File(configFileName);

        if ((! configFile.exists()) || (! configFile.canRead()))
            throw new JobControlException("Configuration file '"
                                          + configFile.getAbsolutePath()
                                          + "' is missing or unreadable");

        if (! config.containsKey(JSubmit.QUEUE_KEY))
            throw new IllegalArgumentException("Config map did not "
                                               + "contain the mandatory key '"
                                               + JSubmit.QUEUE_KEY
                                               + "'");

        String queueName = (String) config.get(JSubmit.QUEUE_KEY);

        ResourceBundle bundle =
            new PropertyResourceBundle(new FileInputStream(configFile));
        UserConfiguration taskConfig = new UserConfiguration(bundle);

		for (int i = 0; i < data.length; i++)
		{
			try
			{
				applySingle(owner, taskConfig, queueName, data[i]);
			}
			catch (Exception e)
			{
				System.err.println("Exception while processing '"
								   + data[i]
								   + "'. Skipping this element");
				e.printStackTrace();
			}
		}
    }

    private void applySingle(Owner owner,
                             UserConfiguration taskConfig,
                             String queueName,
                             String fileName)
        throws Exception
    {
        String taskName = taskConfig.getProtocolName();
        String taskDesc = taskConfig.getProtocolDescription();
        List prededence = taskConfig.getJobPrecedenceList();

        if (prededence.isEmpty())
            throw new JobControlException("No jobs were defined");

        BufferedReader br = new BufferedReader(new FileReader(fileName));

        String workDirectory = System.getProperty("user.dir");
        Status waiting = statusDAO.readStatusById(Status.WAITING);

        Job [] jobs = new Job [prededence.size()];

        String prevOutputName = null;

        for (int i = 0; i < prededence.size(); i++)
        {
            int jobDefIndex = ((Integer) prededence.get(i)).intValue();

            int execId = taskConfig.getExecutableId(jobDefIndex);
            int maxRetry = taskConfig.getMaxRetryCount(jobDefIndex);
            String template = taskConfig.getTemplate(jobDefIndex);

            Executable exec = execDAO.readExecutableById(execId);
            BatchJob job = new JobBatchImpl(exec);
            job.setStatus(waiting);
            job.setQueue(queueName);
            job.setMaxRetryCount(maxRetry);
            job.setWorkDirectoryName(workDirectory);

            String baseName =
                createBaseName(owner.getUserName(), taskName, jobDefIndex);
            job.setOutputName(createName(baseName, LOG_SUFFIX));

            if (i == 0)
            {
                // Initial input goes into leaf
                job.setInputName(fileName);
            }
            else
            {
                // Otherwise chain previous output
                job.setInputName(prevOutputName);
                job.add(jobs[i - 1]);
            }

            prevOutputName = createName(baseName, OUTPUT_SUFFIX);
            jobs[i] = job;

            // We redirect the real output to avoid LSF polluting it
            // with its job report
            job.setCommandTemplate(createRedirectTemplate(template,
                                                          createName(baseName,
                                                                     OUTPUT_SUFFIX)));
        }

        Job rootJob = jobs[jobs.length - 1];

		Task task = new TaskDefaultImpl(taskName, taskDesc, owner);
		task.setJobs(new Job [] { rootJob });

		int taskId = manager.createTask(task);
		manager.submitTask(taskId);
	}

    private String createBaseName(String ownerName,
                                  String taskName,
                                  int jobDefIndex)
    {
        StringBuffer sb = new StringBuffer(ownerName);
        sb.append(".");
        sb.append(taskName);
        sb.append(".");
        sb.append("job.");
        sb.append(jobDefIndex);

        return sb.substring(0);
    }

    private String createName(String baseName, String suffix)
    {
        StringBuffer sb = new StringBuffer(baseName);
        sb.append(".");
        sb.append(suffix);

        return sb.substring(0);
    }

    private String createRedirectTemplate(String template,
                                          String redirectName)
    {
        StringBuffer sb = new StringBuffer(template);
        sb.append(" > ");
        sb.append(redirectName);

        return sb.substring(0);
    }
}
