/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.run;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.text.DateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.SortedMap;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.dao.JobDAO;
import uk.ac.sanger.jcon.dao.StatusDAO;
import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.job.Task;
import uk.ac.sanger.jcon.util.Mailer;

/**
 * <code>AbstractCoordinator</code> is an abstract base class for
 * making <code>Coordinator</code>s. It contains the basic data access
 * objects, the <code>ProcessorFactory</code>, an inner class listener
 * for changes in <code>Job</code> <code>Status</code> and some
 * utility methods for managing the array of <code>Processor</code>s.
 *
 * @author Keith James
 */
public abstract class AbstractCoordinator implements Coordinator
{
    // Log4j setup
    static Category cat =
        Category.getInstance(AbstractCoordinator.class.getName());

    /**
     * <code>Job</code> data access object.
     */
    protected JobDAO jobDAO;

    /**
     * <code>Status</code> data access object.
     */
    protected StatusDAO statusDAO;

    /**
     * <code>Processor</code> factory.
     */
    protected ProcessorFactory pFactory;

    /**
     * <code>CommandLineFactory</code> factory.
     */
    protected CommandLineFactory clFactory;

    /**
     * <code>Task</code> we are currently running <code>Job</code>s
     * from.
     */
    protected Task taskLite;

    /**
     * Number of <code>Job</code>s we must have free slots for before
     * grabbing more from database. Also the number of
     * <code>Job</code>s we grab on one go.
     */
    protected int jobGrabSize;

    /**
     * The interval between checks for new READY <code>Job</code>s.
     */
    protected int jobPollInterval;

    /**
     * Maximum number of <code>Job</code>s we can have running at
     * once.
     */
    protected int maxConcurrentJobs;

    /**
     * The array of <code>Processor</code>s used.
     */
    protected Processor [] processors;

    /**
     * Coordinator thread start/stop switch.
     */
    protected boolean isCoordinating;

    /**
     * <code>stopRequested</code> indicates whether the
     * <code>Coordinator</code> has been asked to stop.
     */
    protected boolean stopRequested;

    protected Status started;
    protected Status ready;

    /**
     * Creates a new <code>AbstractCoordinator</code>.
     *
     * @param pFactory a <code>ProcessorFactory</code>.
     * @param jobDAO a <code>JobDAO</code>.
     * @param statusDAO a <code>StatusDAO</code>.
     */
    public AbstractCoordinator(ProcessorFactory pFactory,
                               JobDAO           jobDAO,
                               StatusDAO        statusDAO)
    {
       this.pFactory  = pFactory;
       this.jobDAO    = jobDAO;
       this.statusDAO = statusDAO;

       clFactory = CommandLineFactoryVelocityImpl.getInstance();

       isCoordinating = false;
       stopRequested  = false;
    }

    /**
     * <code>getTask</code> returns the <code>Task</code> from which
     * further <code>Job</code>s will be processed.
     *
     * @return a <code>Task</code>.
     */
    public Task getTask()
    {
        return taskLite;
    }

    /**
     * <code>setTask</code> sets the <code>Task</code> from which
     * further <code>Job</code>s will be processed. This should be a
     * "lite" implementation (one which only references a root
     * <code>Job</code>, rather than a full <code>Job</code> tree). A
     * full <code>Task</code> implementation is acceptable, but simply
     * uses more memory for no extra functionality.
     *
     * @param taskLite a <code>Task</code>.
     */
    public synchronized void setTask(Task taskLite)
    {
        cat.info("Changed current Task to " + taskLite);
        this.taskLite = taskLite;
    }

    /**
     * <code>getMaxConcurrentJobs</code> returns the maximum number of
     * <code>Job</code>s which will be run at once.
     *
     * @return an <code>int</code>.
     */
    public int getMaxConcurrentJobs()
    {
        return maxConcurrentJobs;
    }

    /**
     * <code>getJobGrabSize</code> returns the number of slots which
     * must be free before more <code>Job</code>s will be read from
     * the store to fill them.
     *
     * @return an <code>int</code>.
     */
    public int getJobGrabSize()
    {
        return jobGrabSize;
    }

    /**
     * <code>setJobGrabSize</code> sets the number of slots which must
     * be free before more <code>Job</code>s will be read from the
     * store to fill them.
     *
     * @param jobGrabSize an <code>int</code>.
     *
     * @exception JobControlException if the number is less than 1 or
     * more than the maximum number of concurrent jobs.
     */
    public void setJobGrabSize(int jobGrabSize) throws JobControlException
    {
        if (jobGrabSize < 1 || jobGrabSize > maxConcurrentJobs)
            throw new JobControlException("Failed to set Job grab size to a "
                                          + "value exceeding the maximum "
                                          + "concurrent number of Jobs");

        cat.debug("Now fetching groups of "
                  + jobGrabSize
                  + " jobs from storage");
        this.jobGrabSize = jobGrabSize;
    }

   /**
     * <code>startCoordinating</code> starts the
     * <code>Coordinator</code>.
     *
     * @exception IllegalStateException if already started.
     */
    public abstract void startCoordinating() throws IllegalStateException;

    /**
     * <code>stopCoordinating</code> stops the
     * <code>Coordinator</code>.
     *
     * @exception IllegalStateException if already stopped.
     */
    public abstract void stopCoordinating() throws IllegalStateException;

    /**
     * <code>coordinate</code> executes the main coordination loop.
     */
    protected void coordinate()
    {
        try
        {
            started = statusDAO.readStatusById(Status.STARTED);
            ready   = statusDAO.readStatusById(Status.READY);
        }
        catch (Exception e)
        {
            cat.error("Caught Exception while reading startup values.", e);
        }

        while (isCoordinating)
        {
            int freeSlots = countFreeProcessors();
            cat.info("Currently running: "
                     + (maxConcurrentJobs - freeSlots)
                     + " jobs. Max concurrent Jobs allowed: "
                     + maxConcurrentJobs
                     + " ("
                     + freeSlots
                     + " free slots; grab size "
                     + jobGrabSize
                     + ")");

            // Only stop if there are no Jobs running
            if (stopRequested)
            {
                if (freeSlots == processors.length)
                {
                    cat.info("No Jobs running. Coordinator stopping now.");
                    isCoordinating = false;
                    return;
                }
                else
                {
                    cat.info("Jobs still running. Awaiting their completion "
                             + "before Coordinator is stopped.");
                    waitFor(jobPollInterval);
                    checkRunningJobs();
                    continue;
                }
            }

            if (taskLite == null)
            {
                cat.info("Not Task available");
                waitFor(jobPollInterval);
                continue;
            }
            else if (freeSlots >= jobGrabSize)
            {
                // We have enough free processors to make a database hit
                // worthwhile
                try
                {
                    synchronized(this)
                    {
                        Collection readyJobs =
                            jobDAO.readNJobsByStatusAndTaskId(freeSlots,
                                                              ready,
                                                              taskLite.getId());
                        assignReadyJobs(readyJobs);
                    }
                }
                catch (Exception e)
                {
                    cat.error("Caught Exception while fetching ready Jobs.", e);
                }
            }

            waitFor(jobPollInterval);
            checkRunningJobs();
        }
    }

    protected void checkRunningJobs() { }

    /**
     * <code>waitFor</code> causes the thread to sleep for a number of
     * millis and posts an info message to this effect.
     *
     * @param milliSeconds an <code>int</code>.
     */
    protected synchronized void waitFor(int milliSeconds)
    {
        try
        {
            cat.info("Waiting for " + milliSeconds + " ms");
            wait(milliSeconds);
        }
        catch (InterruptedException ie)
        {
            cat.error("Caught Exception", ie);
        }
    }

    protected void createProcessors()
    {
        processors = new Processor [getMaxConcurrentJobs()];

        for (int i = 0; i < processors.length; i++)
            processors[i] = pFactory.createProcessor(clFactory, statusDAO);
    }

    /**
     * <code>countFreeProcessors</code> counts the number of free
     * <code>Processor</code>s.
     *
     * @return an <code>int</code>.
     */
    protected int countFreeProcessors()
    {
        int freeProcs = 0;

        for (int i = 0; i < processors.length; i++)
        {
            if (processors[i].getFree())
                freeProcs++;
        }

        return freeProcs;
    }

    /**
     * <code>getFreeProcessor</code> returns the first free
     * <code>Processor</code>.
     *
     * @return a <code>Processor</code>, or null if none are free.
     */
    protected Processor getFreeProcessor()
    {
        for (int i = 0; i < processors.length; i++)
        {
            cat.debug("Testing whether processor "
                      + i
                      + " is free");
            if (processors[i].getFree())
            {
                cat.debug("Processor "
                          + i
                          + " is free");
                return processors[i];
            }
        }

        return null;
    }

    /**
     * <code>setFreeProcessor</code> sets a <code>Processor</code> as
     * freed from a specific <code>Job</code> by finding the one with
     * a reference to that <code>Job</code> in the processors array.
     *
     * @param job a <code>Job</code>.
     */
    protected void setFreeProcessor(Job job)
    {
        int index = findProcessorIndex(job.getId());

        if (index == -1)
        {
            cat.error("Failed to find which Processor was processing Job ["
                      + job.getId()
                      + "]");
        }
        else
        {
            cat.debug("Set processor "
                      + index
                      + " free on completion of Job ["
                      + job.getId()
                      + "]");
            processors[index].setFree(true);
        }
    }

    /**
     * <code>findProcessorIndex</code> returns the index of the
     * <code>Processor</code> which is handling the specified
     * <code>Job</code> or -1 if the <code>Job</code> is not being
     * processed.
     *
     * @param jobId an <code>int</code>.
     *
     * @return an <code>int</code>.
     */
    protected int findProcessorIndex(int jobId)
    {
        for (int i = 0; i < processors.length; i++)
        {
            if (processors[i].getFree())
                continue;

            Job procJob =  processors[i].getJob();
            if ((procJob != null) && (jobId == procJob.getId()))
                return i;
        }

        return -1;
    }

    /**
     * <code>assignReadyJobs</code> assigns ready <code>Job</code>s
     * to available <code>Processor</code>s.
     *
     * @param ready a <code>Collection</code> of ready
     * <code>Job</code>s.
     */
    protected void assignReadyJobs(Collection readyJobs)
    {
        for (Iterator ri = readyJobs.iterator(); ri.hasNext();)
        {
            Job job = (Job) ri.next();            

            Processor proc = getFreeProcessor();

            if (proc != null)
            {
                cat.debug("About to start " + job);

                // Add listener for changes in job status made by the
                // processor
                PropertyChangeListener listener = createUpdateListener();
                job.addPropertyChangeListener("status", listener);

                try
                {
                    proc.process(job);
                }
                catch (Exception e)
                {
                    cat.error("Caught Exception while starting Job ["
                              + job.getId()
                              + "]", e);

                    cat.debug("Removing status update listener from Job ["
                              + job.getId()
                              + "]");

                    job.removePropertyChangeListener(listener); 
                    setFreeProcessor(job);
                }
            }
            else
            {
                cat.error("Failed to find a free Processor");
            }
        }
    }

    protected PropertyChangeListener createUpdateListener()
    {
        return new UpdateListener();
    }

    /**
     * <code>UpdateListener</code> is a listener for
     * <code>Status</code> changes in a <code>Job</code> which will
     * update the <code>Job</code> in the store and then remove itself
     * from the <code>Job</code>.
     */
    class UpdateListener implements PropertyChangeListener
    {
        /**
         * <code>propertyChange</code> updates the <code>Job</code>
         * via the enclosing class' <code>JobDAO</code>. If the status
         * change is to COMPLETED or FAILED the listener removes
         * itself from the <code>Job</code> and frees the relevant
         * <code>Processor</code> in the enclosing class.
         *
         * @param evt a <code>PropertyChangeEvent</code>.
         */
        public void propertyChange(PropertyChangeEvent evt)
        {
            Job job = (Job) evt.getSource();
            int oldState = ((Status) evt.getOldValue()).getId();
            int newState = ((Status) evt.getNewValue()).getId();

            boolean terminated = false;

            try
            {
                switch (newState)
                {
                    case (Status.STARTED):
                        jobDAO.updateJobStarted(job);
                        break;

                    case(Status.COMPLETED):
                        jobDAO.updateJobCompleted(job);
                        terminated = true;
                        break;

                    case (Status.FAILED):
                        jobDAO.updateJobFailedCascade(job);
                        terminated = true;
                        break;

                    case (Status.CANCELLED):
                        terminated = true;
                        break;

                    case (Status.SKIPPED):
                        terminated = true;
                        break;

                    default:
                        cat.warn("Coordinator was asked to apply unexpected update of Job ["
                                 + job.getId()
                                 + "] from status "
                                 + oldState
                                 + " to "
                                 + newState); 
                }
            }
            catch (Exception e)
            {
                cat.error("Coordinator failed to update Job ["
                          + job.getId()
                          + "]", e);
            }
            finally
            {
                if (terminated)
                {
                    setFreeProcessor(job);
                    job.removePropertyChangeListener(this);
                }
            }
        }
    }
}
