/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.run;

import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.dao.StatusDAO;
import uk.ac.sanger.jcon.job.Job;

/**
 * <code>AbstractProcessor</code> is an abstract base class for
 * <code>Processor</code> implementations. These are the objects which
 * actually launch job processes, either locally or on remote
 * queues. This class contains the methods for creating the running
 * thread, handling checks to see if an instance is busy and for
 * passing a <code>Job</code> object for processing.
 *
 * @author Keith James
 */
public abstract class AbstractProcessor implements Processor, Runnable
{
    /**
     * <code>clFactory</code> is the <code>CommandLineFactory</code>.
     */
    protected CommandLineFactory clFactory;

    /**
     * <code>statusDAO</code> is the source of <code>Status</code>
     * objects.
     */
    protected StatusDAO statusDAO;

    /**
     * <code>thread</code> is the <code>Thread</code> object for this
     * <code>Runnable</code>.
     */
    protected Thread thread;

    /**
     * <code>isFree</code> is a flag indicating whether the
     * <code>Processor</code> is free.
     */
    protected boolean isFree;

    /**
     * <code>job</code> is the <code>Job</code> being run.
     */
    protected Job job;

    /**
     * Creates a new <code>AbstractProcessor</code>.
     *
     * @param clFactory a <code>CommandLineFactory</code>.
     * @param statusDAO a <code>StatusDAO</code>.
     */
    public AbstractProcessor(CommandLineFactory clFactory,
                             StatusDAO          statusDAO)
    {
        this.clFactory = clFactory;
        this.statusDAO = statusDAO;
        isFree = true;
    }

    /**
     * <code>getJob</code> returns the <code>Job</code> currently
     * being processed.
     *
     * @return a <code>Job</code>.
     */
    public Job getJob()
    {
        return job;
    }

    /**
     * <code>getFree</code> returns true if the processor is free to
     * accept a new <code>Job</code>.
     *
     * @return a <code>boolean</code>.
     */
    public boolean getFree()
    {
        return isFree;
    }

    public void setFree(boolean isFree)
    {
        this.isFree = isFree;

        if (isFree)
            job = null;
    }

    /**
     * <code>process</code> starts the Job running thread.
     *
     * @param job a <code>Job</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void process(Job job) throws Exception
    {
        if (! getFree())
            throw new JobControlException("Failed to process Job as Processor is not free");

        setFree(false);
        this.job = job;

        // FIXME - use thread pool to prevent too many threads
        thread = new Thread(this);
        thread.start();
    }

    /**
     * Implement <code>run</code> to provide a means for launching
     * the <code>Job</code>.
     */
    public abstract void run();

    /**
     * <code>stringTokenise</code> creates an array of strings by
     * splitting on whitespace with the default
     * <code>StringTokenizer</code>.
     *
     * @param string a <code>String</code>.
     * @return a <code>String []</code>.
     */
    protected String [] stringTokenise(String string)
    {
        StringTokenizer st = new StringTokenizer(string);

        List tokens = new ArrayList();

        while (st.hasMoreTokens())
        {
            tokens.add(st.nextToken());
        }

        return (String []) tokens.toArray(new String [0]);
    }
}
