/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.run;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.text.DateFormat;
import java.util.Date;
import java.util.Map;
import java.util.HashMap;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.job.Task;
import uk.ac.sanger.jcon.dao.TaskDAO;
import uk.ac.sanger.jcon.util.Mailer;

/**
 * <code>AbstractTaskScheduler</code> is an abstract base class for
 * making <code>TaskScheduler</code>s. It contains the basic data
 * access object, the scheduler <code>Policy</code> and an inner class
 * listener for changes in <code>Task</code> root <code>Job</code>
 * <code>Status</code> which will send email to the <code>Task</code>
 * <code>Owner</code>.
 *
 * @author Keith James
 */
public abstract class AbstractTaskScheduler implements TaskScheduler
{
   // Log4j setup
    static Category cat =
        Category.getInstance(AbstractTaskScheduler.class.getName());

    /**
     * <code>taskDAO</code> the <code>Task</code> data access object.
     */
    protected TaskDAO taskDAO;

    /**
     * <code>isScheduling</code> flag indicating whether the scheduler
     * is running.
     */
    protected boolean isScheduling;

    /**
     * <code>policy</code> the <code>Task</code> selection policy.
     */
    protected Policy policy;

    /**
     * <code>coordinator</code> the <code>Task</code>
     * <code>Coordinator</code>.
     */
    protected Coordinator coordinator;

    /**
     * <code>policyPollInterval</code> the interval between polls to
     * the <code>Policy</code> object for <code>Task</code>
     * selection.
     */
    protected int policyPollInterval;

    /**
     * <code>taskCache</code> caches "lite" <code>Task</code>s so that
     * changes in their state may be relayed to their
     * <code>Owner</code> by email.
     */
    protected Map taskCache;

    /**
     * <code>sendMail</code> determines whether <code>Task</code>
     * <code>Owner</code>s will be informed of state changes by email.
     */
    protected boolean sendMail;

    /**
     * <code>mailFrom</code> is the email address used in the "From"
     * fields of messages sent to the <code>Owner</code>.
     */
    protected String mailFrom;

    /**
     * <code>df</code> is the <code>DateFormat</code> used in creating
     * the email content.
     */
    protected DateFormat df;

    /**
     * Creates a new <code>AbstractTaskScheduler</code> which has
     * email notification enabled.
     *
     * @param taskDAO a <code>TaskDAO</code>.
     * @param coordinator a <code>Coordinator</code>.
     */
    public AbstractTaskScheduler(TaskDAO     taskDAO,
                                 Coordinator coordinator)
    {
        this.taskDAO     = taskDAO;
        this.coordinator = coordinator;

        taskCache = new HashMap();
        df = DateFormat.getDateTimeInstance();
        sendMail = true;
    }

    public Policy getPolicy()
    {
        return policy;
    }

    public boolean getSendMail()
    {
        return sendMail;
    }

    public void setSendMail(boolean sendMail)
    {
        this.sendMail = sendMail;
    }

    /**
     * <code>setPolicy</code> sets the scheduling <code>Policy</code>.
     *
     * @param policy a <code>Policy</code>.
     *
     * @exception IllegalStateException if the method is called while
     * the <code>TaskScheduler</code> is running.
     */
    public void setPolicy(Policy policy) throws IllegalStateException
    {
        if (isScheduling == true)
            throw new IllegalStateException("Failed to set Policy as "
                                            + " the TaskScheduler is running");
        this.policy = policy;
    }

    public abstract void startScheduling();

    public abstract void stopScheduling();

    /**
     * <code>schedule</code> executes the main scheduling loop.
     */
    protected void schedule()
    {
        int currentTaskId = -1;

        while (isScheduling)
        {
            try
            {
                int newTaskId = policy.selectTaskId();

                cat.debug("Current Task ID: "
                          + currentTaskId
                          + ", new Task ID: "
                          + newTaskId);

                if (newTaskId == -1)
                {
                    cat.debug("No Tasks available");

                    // Update for current Task
                    if (currentTaskId > 0)
                    {
                        taskUpdate(currentTaskId);
                        currentTaskId = -1;
                    }
                }
                else if (newTaskId != currentTaskId)
                {
                    // Update for current Task if there is one
                    if (currentTaskId > 0)
                        taskUpdate(currentTaskId);

                    // Update for new Task
                    taskUpdate(newTaskId);

                    currentTaskId = newTaskId;
                    cat.debug("Scheduler selected Task ["
                              + currentTaskId
                              + "]");

                    Task taskLite = taskDAO.readTaskById(currentTaskId,
                                                         TaskDAO.EXCLUDE_JOB_TREE);

                    coordinator.setTask(taskLite);
                }

                // Current Task re-selected
                waitFor(policyPollInterval);
            }
            catch (Exception e)
            {
                cat.error("Caught Exception while scheduling.", e);
            }
        }
    }

    /**
     * <code>taskUpdate</code> compares a <code>Task</code>'s current
     * state with its last known state. If a significant state change
     * has occurred, email is sent to the <code>Task</code>'s
     * <code>Owner</code>.
     *
     * @param taskId an <code>int</code>.
     *
     * @exception Exception if an error occurs.
     */
    protected void taskUpdate(int taskId) throws Exception
    {
        Integer key = new Integer(taskId);

        Task current = taskDAO.readTaskById(taskId, TaskDAO.EXCLUDE_JOB_TREE);
        Job [] currJobs = current.getJobs();
        String description = currJobs[0].getStatus().getDescription();

        if (taskCache.containsKey(key))
        {
            cat.debug("Task cache contains: " + current);

            Task previous = (Task) taskCache.get(key);
            Job [] prevJobs  = previous.getJobs();

            if (prevJobs.length != currJobs.length)
                throw new JobControlException("The number of root Jobs in Task ["
                                              + taskId
                                              + "] has changed while running");

            int currStatId, prevStatId;

            for (int i = currJobs.length; --i >= 0;)
            {
                currStatId = currJobs[i].getStatus().getId();
                prevStatId = prevJobs[i].getStatus().getId();

                if (currStatId == prevStatId)
                    continue;

                boolean mailableEventFlag = false;

                switch (currStatId)
                {
                    case (Status.WAITING):
                    case (Status.READY):
                    case (Status.STARTED):
                        break;

                    case (Status.COMPLETED):
                        if (i == 0)
                        {
                            mailableEventFlag = true;
                            taskCache.remove(key);
                        }
                        break;

                    case (Status.FAILED):
                        mailableEventFlag = true;
                        taskCache.remove(key);
                        break;

                    case (Status.CANCELLED):
                        mailableEventFlag = true;                        
                        taskCache.remove(key);
                        break;

                    case (Status.SKIPPED):
                        // This is caused where a Job failure cascades
                        // up to the root
                        mailableEventFlag = true;                        
                        taskCache.remove(key);
                        break;

                    default:
                        taskCache.remove(key);
                        throw new JobControlException("Task ["
                                                      + taskId
                                                      + "] was updated to an unexpected Status ["
                                                      + currStatId
                                                      + "]");
                }

                if (mailableEventFlag && sendMail) {
                    sendUpdateMail(current, description);
                }
            }
        }
        else
        {
            cat.debug("Adding to Task cache: " + current);
            taskCache.put(key, current);

            if (sendMail)
                sendUpdateMail(current, "STARTED");
        }
    }

    /**
     * <code>sendUpdateMail</code> informs the <code>Task</code>'s
     * <code>Owner</code> of a state change.
     *
     * @param task a <code>Task</code> which changed state.
     * @param state a <code>String</code> describing the
     * <code>Task</code>'s state.
     *
     * @exception Exception if an error occurs.
     */
    protected void sendUpdateMail(Task task, String label)
        throws Exception
    {
        String address = task.getOwner().getEMail();

        StringBuffer sb = new StringBuffer();
        sb.append("Task [");
        sb.append(task.getId());
        sb.append("] ");
        sb.append(task.getName());
        sb.append(" updated to ");
        sb.append(label);

        String subject = sb.substring(0);

        sb.setLength(0);
        sb.append(task.getDescription());
        sb.append("\n\nSubmitted at ");
        sb.append(df.format(task.getSubmissionTime()));
        sb.append(" was updated to ");
        sb.append(label);

        Mailer.sendMessage(new String [] { address },
                           mailFrom,
                           subject,
                           sb.substring(0));
    }

    /**
     * <code>waitFor</code> causes the thread to sleep for a number of
     * millis and posts an info message to this effect.
     *
     * @param milliSeconds an <code>int</code>.
     */
    protected synchronized void waitFor(int milliSeconds)
    {
        try
        {
            cat.info("Waiting for " + milliSeconds + " ms");
            wait(milliSeconds);
        }
        catch (InterruptedException ie)
        {
            cat.error("Caught Exception", ie);
        }
    }
}
