/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.run;

import java.io.StringWriter;
import java.util.Properties;

import org.apache.log4j.Category;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.VelocityEngine;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.job.Executable;
import uk.ac.sanger.jcon.job.Job;

/**
 * <p><code>CommandLineFactoryVelocityImpl</code> creates command line
 * strings from <code>Job</code>s using the Velocity template
 * engine. The template is expanded using references to the
 * <code>Job</code>'s inputName, outputName, errorName and the
 * <code>Executable</code>'s name (as ${input_name}, ${output_name},
 * ${error_name} and ${executable}, respectively).</p>
 *
 * <p>The default references are:
 *
 * <ul>
 *   <li>${executable} is replaced by the name of Job's Executable</li>
 *   <li>${input_name} is replaced by the Job inputName</li>
 *   <li>${output_name} is replaced by the Job outputName</li>
 *   <li>${error_name} is replaced by the Job errorName</li>
 *   <li>${job_id} is replaced by the Job's unique ID</li>
 * </ul>
 *
 * <p>Other references may be expanded by using the alternative
 * createCommandLine method which accepts a
 * <code>VelocityContext</code> as an argument. This allows arbitrary
 * references to be defined and expanded.</p>
 *
 * @author Keith James
 */
public class CommandLineFactoryVelocityImpl implements CommandLineFactory
{
    /**
     * <code>VELOCITY_PROPS</code> the properties file containing the
     * Velocity configuration used to initialise the Velocity engine.
     */
    public static final String VELOCITY_PROPS = "/velocity.properties";

    /**
     * <code>LOG_TAG</code> the log-tag used in log statements issued
     * by the Velocity engine.
     */
    public static final String LOG_TAG = "CMD_TMPL";

    // Log4j setup
    static Category cat =
        Category.getInstance(CommandLineFactoryVelocityImpl.class.getName());

    // Singleton instance of this class
    private static CommandLineFactory clFactory =
        new CommandLineFactoryVelocityImpl();

    private VelocityEngine vEngine;

    private static String EXEC_REF = "$" + CommandLineFactory.EXECUTABLE;
    private static String EXEC_FORMAL_REF = "${" + CommandLineFactory.EXECUTABLE + "}";

    /**
     * Creates a new
     * <code>CommandLineFactoryVelocityImpl</code>. Private
     * constructor; use getInstance to get the singleton.
     */
    private CommandLineFactoryVelocityImpl()
    {
        try
        {
            Properties velocityProps = new Properties();
            velocityProps.load(getClass().getResourceAsStream(VELOCITY_PROPS));

            vEngine = new VelocityEngine();

            // Initialise with loaded Properties
            vEngine.init(velocityProps);
        }
        catch (Exception e)
        {
            cat.fatal("Failed to create Velocity CommandLineFactory", e);
            throw new Error("Failed to create Velocity CommandLineFactory");
        }
    }

    public static CommandLineFactory getInstance()
    {
        return clFactory;
    }

    public synchronized String createCommandLine(Job job) throws Exception
    {
        return expandTemplate(job, new VelocityContext());
    }

    public synchronized String createCommandLine(Job             job,
                                                 VelocityContext context)
        throws Exception
    {
        // Chain contexts
        VelocityContext context2 = new VelocityContext(context);
        return expandTemplate(job, context2);
    }

    private String expandTemplate(Job             job,
                                  VelocityContext context)
        throws Exception
    {
        String       inputName = job.getInputName();
        String      outputName = job.getOutputName();
        String       errorName = job.getErrorName();
        String commandTemplate = job.getCommandTemplate();

        if (commandTemplate.indexOf(EXEC_REF)        == -1 &&
            commandTemplate.indexOf(EXEC_FORMAL_REF) == -1)
            throw new JobControlException("Job ["
                                          + job.getId()
                                          + "] command line template '"
                                          + commandTemplate
                                          + "' did not contain token for executable");

        Executable executable = job.getExecutable();

        context.put(CommandLineFactory.JOB_ID, new Integer(job.getId()));

        if (inputName != null)
            context.put(CommandLineFactory.INPUT_NAME, inputName);
        if (outputName != null)
            context.put(CommandLineFactory.OUTPUT_NAME, outputName);
        if (errorName != null)
            context.put(CommandLineFactory.ERROR_NAME, errorName);

        if (executable == null)
            throw new IllegalStateException("Failed to create command line as the Executable was null");
        else
            context.put(CommandLineFactory.EXECUTABLE, executable.getName());

        if (commandTemplate == null)
            return executable.getName();

        StringWriter writer = new StringWriter(256);
        vEngine.evaluate(context, writer, LOG_TAG, commandTemplate);

        return writer.toString();
    }
}
