/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.run;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControl;
import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.dao.JobDAO;
import uk.ac.sanger.jcon.dao.StatusDAO;
import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.job.Task;

/**
 * <p><code>Coordinator</code>s are responsible for reading groups of
 * <code>Job</code>s from a persistent store, attaching a listener for
 * changes in their status and then assigning each of them to a free
 * <code>Processor</code> to be executed. If no <code>Job</code>s are
 * ready to be run the coordinator thread will sleep for a
 * configurable period before polling the store again.</p>
 *
 * <p>The listener attached to each job updates the persistent copy of
 * that <code>Job</code> when its status changes in recognised way and
 * then removes itself from the <code>Job</code>.</p>
 *
 * <p>The <code>Coordinator</code> starts <code>Job</code>s from one
 * <code>Task</code> at a time. However, it is safe to switch
 * <code>Task</code>s at any point. The next set of <code>Job</code>s
 * will then be drawn from the new <code>Task</code>.</p>
 *
 * @author Keith James
 */
public interface Coordinator
{
    /**
     * <code>MAX_CONCURRENT_KEY</code> the <code>ResourceBundle</code> key
     * to specify the maximum number of <code>Job</code>s to run
     * concurrently.
     */
    public static final String MAX_CONCURRENT_KEY =
        "jobcontrol.coordinator.max_concurrent";

    /**
     * <code>GRAB_SIZE_KEY</code> the <code>ResourceBundle</code> key to
     * specify the number of <code>Job</code> slots which must be free
     * before a database hit is made.
     */
    public static final String GRAB_SIZE_KEY =
        "jobcontrol.coordinator.grab_size";

    /**
     * <code>POLL_INTERVAL_KEY</code> the <code>ResourceBundle</code> key
     * to specify the polling interval in millseconds at which a
     * completely idle <code>Coordinator</code> will check the
     * database for new ready <code>Job</code>s.
     */
    public static final String POLL_INTERVAL_KEY =
        "jobcontrol.coordinator.poll_interval";

    /**
     * <code>getTask</code> returns the <code>Task</code> from which
     * further <code>Job</code>s will be processed.
     *
     * @return a <code>Task</code>.
     */
    public Task getTask();

    /**
     * <code>setTask</code> sets the <code>Task</code> from which
     * further <code>Job</code>s will be processed.
     *
     * @param task a <code>Task</code>.
     */
    public void setTask(Task task) throws Exception;

    /**
     * <code>getMaxConcurrentJobs</code> returns the maximum number of
     * <code>Job</code>s which will be run at once.
     *
     * @return an <code>int</code>.
     */
    public int getMaxConcurrentJobs();

    /**
     * <code>getJobGrabSize</code> returns the number of slots which
     * must be free before more <code>Job</code>s will be read from
     * the store to fill them.
     *
     * @return an <code>int</code>.
     */
    public int getJobGrabSize();

    /**
     * <code>setJobGrabSize</code> sets the number of slots which must
     * be free before more <code>Job</code>s will be read from the
     * store to fill them.
     *
     * @param jobGrabSize an <code>int</code>.
     *
     * @exception JobControlException if the number is less than 1 or
     * more than the maximum number of concurrent jobs.
     */
    public void setJobGrabSize(int jobGrabSize) throws JobControlException;

    /**
     * <code>startCoordinating</code> starts the coordinator
     * thread.
     */
    public void startCoordinating();

    /**
     * <code>stopCoordinating</code> stops the coordinator thread. Any
     * programs currently running will be unaffected. No new programs
     * will be started.
     */
    public void stopCoordinating();
}
