/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.run;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.text.DateFormat;
import java.util.*;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControl;
import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.dao.JobDAO;
import uk.ac.sanger.jcon.dao.StatusDAO;
import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.Owner;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.job.Task;
import uk.ac.sanger.jcon.util.Mailer;

/**
 * <p><code>CoordinatorDefaultImpl</code> is an implementation which
 * holds a list of <code>Processor</code>s, which each of which can
 * run one <code>Job</code>. It assigns <code>Job</code>s from a store
 * to free <code>Processor</code>s as they become available and
 * updates the store as they are processed. It runs in a separate
 * thread.</p>
 *
 * <p>By passing the relevant <code>ProcessorFactory</code> to the
 * constructor it is possible to use this coordinator with either the
 * the default (Java-based stream IO) or LSF systems. Be aware that
 * LSF version 4.2 on Compaq Tru64 appears to have problems supporting
 * C API requests via JNI which come from any thread other than
 * <code>main</code>.</p>
 *
 * @author Keith James
 */
public class CoordinatorDefaultImpl extends AbstractCoordinator
    implements Runnable
{
    /**
     * <code>DEFAULT_MAX_CONCURRENT</code> is the default maximum
     * number of <code>Job</code>s to run concurrently.
     */
    public static final int DEFAULT_MAX_CONCURRENT = 20;

    /**
     * <code>DEFAULT_GRAB_SIZE</code> is the default number of
     * <code>Job</code> slots which must be free before a database hit
     * is made.
     */
    public static final int DEFAULT_GRAB_SIZE = 10;

    /**
     * <code>DEFAULT_POLL_INTERVAL</code> is the default polling
     * interval in millseconds at which a <code>Coordinator</code>
     * will check the database for new ready <code>Job</code>s.
     */
    public static final int DEFAULT_POLL_INTERVAL = 60000;

    // Log4j setup
    static Category cat =
        Category.getInstance(CoordinatorDefaultImpl.class.getName());

    // Coordinator thread
    private Thread thread;

    /**
     * Creates a new <code>CoordinatorDefaultImpl</code> using the
     * specified data access objects. By varying the
     * <code>ProcessorFactory</code>, jobs can be launched on
     * different systems. There are separate factories for local runs,
     * batch queue runs etc.
     *
     * @param pFactory a <code>ProcessorFactory</code>.
     * @param jobDAO a <code>JobDAO</code>.
     * @param statusDAO a <code>StatusDAO</code>.
     */
    public CoordinatorDefaultImpl(ProcessorFactory pFactory,
                                  JobDAO           jobDAO,
                                  StatusDAO        statusDAO)
    {
        super(pFactory, jobDAO, statusDAO);

        maxConcurrentJobs = DEFAULT_MAX_CONCURRENT;
        jobGrabSize       = DEFAULT_GRAB_SIZE;
        jobPollInterval   = DEFAULT_POLL_INTERVAL;

        ResourceBundle bundle = JobControl.getConfiguration();

        String key;
        String val;

        try
        {
            Enumeration e = bundle.getKeys();

            while (e.hasMoreElements())
            {
                key = (String) e.nextElement();

                try
                {
                    val = bundle.getString(key);
                }
                catch (NullPointerException npe)
                {
                    throw new Error("Malformed configuration ResourceBundle "
                                    + "contains key '"
                                    + key
                                    + "' missing a corresponding value");
                }

                if (key.equals(Coordinator.MAX_CONCURRENT_KEY))
                    maxConcurrentJobs = Integer.parseInt(val);
                else if (key.equals(Coordinator.GRAB_SIZE_KEY))
                    jobGrabSize = Integer.parseInt(val);
                else if (key.equals(Coordinator.POLL_INTERVAL_KEY))
                    jobPollInterval = Integer.parseInt(val);
            }
        }
        catch (NumberFormatException nfe)
        {
            throw new Error("Number format error in "
                            + "configuration ResourceBundle");
        }

        thread = new Thread(this);
    }

    public void startCoordinating() throws IllegalStateException
    {
        if (isCoordinating == true)
            throw new IllegalStateException("Coordinator is already running");

        isCoordinating = true;
        stopRequested = false;

        createProcessors();
        cat.info("Job execution Coordinator started");
        thread.start();
    }

    public void stopCoordinating() throws IllegalStateException
    {
        if (isCoordinating == false)
            throw new IllegalStateException("Coordinator is already stopped");

        stopRequested = true;
        cat.info("Job execution Coordinator stop requested");
    }

    /**
     * <code>run</code> is the main coordinator.
     */
    public void run()
    {
        coordinate();
    }
}
