/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.run;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.util.*;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControl;
import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.dao.JobDAO;
import uk.ac.sanger.jcon.dao.StatusDAO;
import uk.ac.sanger.jcon.job.BatchJob;
import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.job.Task;
import uk.ac.sanger.jcon.lsf.LSFBatchException;
import uk.ac.sanger.jcon.lsf.LSFJobInfo;
import uk.ac.sanger.jcon.lsf.LSFQueueWatcher;

/**
 * <p><code>CoordinatorLSFImpl</code> is an implementation for
 * LSF.</p>
 *
 * @author Keith James
 */
public class CoordinatorLSFImpl implements Coordinator, Runnable
{
    /**
     * <code>DEFAULT_MAX_CONCURRENT</code> is the default maximum
     * number of <code>Job</code>s to run concurrently.
     */
    public static final int DEFAULT_MAX_CONCURRENT = 20;

    /**
     * <code>DEFAULT_GRAB_SIZE</code> is the default number of
     * <code>Job</code> slots which must be free before a database hit
     * is made.
     */
    public static final int DEFAULT_GRAB_SIZE = 10;

    /**
     * <code>DEFAULT_POLL_INTERVAL</code> is the default polling
     * interval in millseconds at which a completely idle
     * <code>Coordinator</code> will check the database for new ready
     * <code>Job</code>s.
     */
    public static final int DEFAULT_POLL_INTERVAL = 60000;

    /**
     * <code>BATCH_QUEUE_KEY</code> is the <code>Properties</code> key
     * which indicates the name of an LSF batch queue to
     * monitor. Multiple entries of this key may be used to specify
     * several batch queues.
     */
    public static final String BATCH_QUEUE_KEY = "jobcontrol.queue.name";

    // Log4j setup
    static Category cat =
        Category.getInstance(CoordinatorLSFImpl.class.getName());

    /**
     * <code>Job</code> data access object.
     */
    protected JobDAO jobDAO;

    /**
     * <code>Status</code> data access object.
     */
    protected StatusDAO statusDAO;

    /**
     * <code>Processor</code> factory.
     */
    protected ProcessorFactory pFactory;

    /**
     * <code>CommandLineFactory</code> factory.
     */
    protected CommandLineFactory clFactory;

    /**
     * <code>Task</code> we are currently running <code>Job</code>s
     * from.
     */
    protected Task taskLite;

    /**
     * Number of <code>Job</code>s we must have free slots for before
     * grabbing more from database. Also the number of
     * <code>Job</code>s we grab on one go.
     */
    protected int jobGrabSize;

    /**
     * The interval between checks for new READY <code>Job</code>s.
     */
    protected int jobPollInterval;

    /**
     * Maximum number of <code>Job</code>s we can have running at
     * once, per queue.
     */
    protected int maxConcurrentJobs;

    /**
     * Coordinator thread start/stop switch.
     */
    protected boolean isCoordinating;

    /**
     * <code>stopRequested</code> indicates whether the
     * <code>Coordinator</code> has been asked to stop.
     */
    protected boolean stopRequested;

    protected Status started;
    protected Status ready;


    // Coordinator thread
    private Thread thread;

    // LSF query interface
    private LSFQueueWatcher queueWatcher;
    // Map of queue names to arrays of Processors; one array per queue
    private Map queueProcessors;

    /**
     * Creates a new <code>CoordinatorLSFImpl</code> using the
     * specified data access objects.
     *
     * @param pFactory a <code>ProcessorFactory</code>. This should be
     * the <code>ProcessorLSFImpl.FACTORY</code> in order for the
     * <code>Job</code>s to be submitted to an LSF batch queue.
     * @param jobDAO a <code>JobDAO</code>.
     * @param statusDAO a <code>StatusDAO</code>.
     */
    public CoordinatorLSFImpl(ProcessorFactory pFactory,
                              JobDAO           jobDAO,
                              StatusDAO        statusDAO)
    {
        this.pFactory  = pFactory;
        this.jobDAO    = jobDAO;
        this.statusDAO = statusDAO;

        clFactory = CommandLineFactoryVelocityImpl.getInstance();

        isCoordinating = false;
        stopRequested  = false;

        queueProcessors = new HashMap();

        maxConcurrentJobs = DEFAULT_MAX_CONCURRENT;
        jobGrabSize       = DEFAULT_GRAB_SIZE;
        jobPollInterval   = DEFAULT_POLL_INTERVAL;

        ResourceBundle bundle = JobControl.getConfiguration();

        String key;
        String val;

        try
        {
            Enumeration e = bundle.getKeys();

            while (e.hasMoreElements())
            {
                key = (String) e.nextElement();

                try
                {
                    val = bundle.getString(key);
                }
                catch (NullPointerException npe)
                {
                    throw new Error("Malformed configuration ResourceBundle "
                                    + "contains key '"
                                    + key
                                    + "' missing a corresponding value");
                }

                if (key.equals(BATCH_QUEUE_KEY))
                {
                    StringTokenizer st = new StringTokenizer(val);
                    while (st.hasMoreTokens())
                    {
                        String token = st.nextToken();

                        cat.info("Adding queue '"
                                 + token
                                 + "' to known queues");
                        queueProcessors.put(token, null);
                    }
                }
                else if (key.equals(Coordinator.MAX_CONCURRENT_KEY))
                {
                    maxConcurrentJobs = Integer.parseInt(val);
                }
                else if (key.equals(Coordinator.GRAB_SIZE_KEY))
                {
                    jobGrabSize = Integer.parseInt(val);
                }
                else if (key.equals(Coordinator.POLL_INTERVAL_KEY))
                {
                    jobPollInterval = Integer.parseInt(val);
                }
            }
        }
        catch (NumberFormatException nfe)
        {
            throw new Error("Number format error in "
                            + "configuration ResourceBundle");
        }

        queueWatcher = new LSFQueueWatcher(statusDAO);
        thread = new Thread(this);
    }

    /**
     * <code>getTask</code> returns the <code>Task</code> from which
     * further <code>Job</code>s will be processed.
     *
     * @return a <code>Task</code>.
     */
    public Task getTask()
    {
        return taskLite;
    }

    /**
     * <code>setTask</code> sets the <code>Task</code> from which
     * further <code>Job</code>s will be processed. This should be a
     * "lite" implementation (one which only references a root
     * <code>Job</code>, rather than a full <code>Job</code> tree). A
     * full <code>Task</code> implementation is acceptable, but simply
     * uses more memory for no extra functionality.
     *
     * @param taskLite a <code>Task</code>.
     */
    public synchronized void setTask(Task taskLite)
    {
        cat.info("Changed current Task to " + taskLite);
        this.taskLite = taskLite;
    }

    /**
     * <code>getMaxConcurrentJobs</code> returns the maximum number of
     * <code>Job</code>s which will be run at once.
     *
     * @return an <code>int</code>.
     */
    public int getMaxConcurrentJobs()
    {
        return maxConcurrentJobs;
    }

    /**
     * <code>getJobGrabSize</code> returns the number of slots which
     * must be free before more <code>Job</code>s will be read from
     * the store to fill them.
     *
     * @return an <code>int</code>.
     */
    public int getJobGrabSize()
    {
        return jobGrabSize;
    }

    /**
     * <code>setJobGrabSize</code> sets the number of slots which must
     * be free before more <code>Job</code>s will be read from the
     * store to fill them.
     *
     * @param jobGrabSize an <code>int</code>.
     *
     * @exception JobControlException if the number is less than 1 or
     * more than the maximum number of concurrent jobs.
     */
    public void setJobGrabSize(int jobGrabSize) throws JobControlException
    {
        if (jobGrabSize < 1 || jobGrabSize > maxConcurrentJobs)
            throw new JobControlException("Failed to set Job grab size to a "
                                          + "value exceeding the maximum "
                                          + "concurrent number of Jobs");

        cat.debug("Now fetching groups of "
                  + jobGrabSize
                  + " jobs from storage");
        this.jobGrabSize = jobGrabSize;
    }

    /**
     * <code>startCoordinating</code> starts the
     * <code>Coordinator</code>.
     *
     * @exception IllegalStateException if already started.
     */
    public void startCoordinating() throws IllegalStateException
    {
        if (isCoordinating == true)
            throw new IllegalStateException("Coordinator is already running");

        isCoordinating = true;
        stopRequested = false;

        createProcessors();
        cat.info("Job execution Coordinator started");
        thread.start();
    }

    /**
     * <code>stopCoordinating</code> stops the
     * <code>Coordinator</code>.
     *
     * @exception IllegalStateException if already stopped.
     */
    public void stopCoordinating() throws IllegalStateException
    {
         if (isCoordinating == false)
            throw new IllegalStateException("Coordinator is already stopped");

         stopRequested = true;
         cat.info("Job execution Coordinator stop requested");
    }

    /**
     * <code>run</code> is the main coordinator.
     */
    public void run()
    {
        coordinate();
    }

    /**
     * <code>coordinate</code> executes the main coordination loop.
     */
    protected void coordinate()
    {
        try
        {
            started = statusDAO.readStatusById(Status.STARTED);
            ready   = statusDAO.readStatusById(Status.READY);
        }
        catch (Exception e)
        {
            cat.error("Caught Exception while reading startup values.", e);
        }

        while (isCoordinating)
        {
            int freeSlots = countFreeProcessors();
            int queues = queueProcessors.keySet().size();

            cat.info("Currently running: "
                     + ((queues * maxConcurrentJobs) - freeSlots)
                     + " jobs. Max concurrent Jobs per queue allowed: "
                     + maxConcurrentJobs
                     + " ("
                     + freeSlots
                     + " combined free slots; grab size "
                     + jobGrabSize
                     + ")");

            // Only stop if there are no Jobs running
            if (stopRequested)
            {
                if (freeSlots == (queues * maxConcurrentJobs))
                {
                    cat.info("No Jobs running. Coordinator stopping now.");
                    isCoordinating = false;
                    return;
                }
                else
                {
                    cat.info("Jobs still running. Awaiting their completion "
                             + "before Coordinator is stopped.");
                    waitFor(jobPollInterval);
                    checkRunningJobs();
                    continue;
                }
            }

            if (taskLite == null)
            {
                waitFor(jobPollInterval);
                continue;
            }
            else if (freeSlots >= jobGrabSize)
            {
                // We have enough free processors to make a database hit
                // worthwhile
                try
                {
                    synchronized(this)
                    {
                        Collection readyJobs =
                            jobDAO.readNJobsByStatusAndTaskId(freeSlots,
                                                              ready,
                                                              taskLite.getId());
                        cat.debug("Found ready Jobs: "
                                  + readyJobs);

                        assignReadyJobs(readyJobs);
                    }
                }
                catch (Exception e)
                {
                    cat.error("Caught Exception while fetching ready Jobs.", e);
                }
            }

            waitFor(jobPollInterval);
            checkRunningJobs();
        }
    }

    protected void checkRunningJobs()
    {
        queueWatcher.update();
    }

    /**
     * <code>waitFor</code> causes the thread to sleep for a number of
     * millis and posts an info message to this effect.
     *
     * @param milliSeconds an <code>int</code>.
     */
    protected synchronized void waitFor(int milliSeconds)
    {
        try
        {
            cat.info("Waiting for " + milliSeconds + " ms");
            wait(milliSeconds);
        }
        catch (InterruptedException ie)
        {
            cat.error("Caught Exception", ie);
        }
    }

    protected void createProcessors()
    {
        for (Iterator qi = queueProcessors.keySet().iterator(); qi.hasNext();)
        {
            String queue = (String) qi.next();

            cat.debug("Creating "
                      + getMaxConcurrentJobs()
                      + " processors for queue '"
                      + queue
                      + "'");

            Processor [] processors = new Processor [getMaxConcurrentJobs()];

            for (int i = 0; i < processors.length; i++)
                processors[i] = pFactory.createProcessor(clFactory, statusDAO);

            queueProcessors.put(queue, processors);
        }
    }

    /**
     * <code>countFreeProcessors</code> counts the number of free
     * <code>Processor</code>s across all queues.
     *
     * @return an <code>int</code>.
     */
    protected int countFreeProcessors()
    {
        int totalFreeProcs = 0;

        for (Iterator qi = queueProcessors.keySet().iterator(); qi.hasNext();)
        {
            String queue = (String) qi.next();

            Processor [] processors =
                (Processor []) queueProcessors.get(queue);

            int freeProcs = 0;

            for (int i = 0; i < processors.length; i++)
            {
                if (processors[i].getFree())
                    freeProcs++;
            }

            totalFreeProcs += freeProcs;

            cat.debug("Found "
                      + freeProcs
                      + " for queue '"
                      + queue
                      + "'");
        }

        return totalFreeProcs;
    }

    /**
     * <code>getFreeProcessor</code> returns the first free
     * <code>Processor</code> for a given queue.
     *
     * @param queue a <code>String</code> queue name.
     *
     * @return a <code>Processor</code>, or null if none are free.
     */
    protected Processor getFreeProcessor(String queue)
    {
        Processor [] processors =
                (Processor []) queueProcessors.get(queue);

        cat.debug("Checking for free slots in " + processors);

        for (int i = 0; i < processors.length; i++)
        {
            cat.debug("Testing whether processor "
                      + i
                      + " is free");
            if (processors[i].getFree())
            {
                cat.debug("Processor "
                          + i
                          + " is free");
                return processors[i];
            }
        }

        return null;
    }

    protected void setFreeProcessor(BatchJob job)
    {
        String queue = job.getQueue();
        int index = findProcessorIndex(job.getId(), queue);

        if (index == -1)
        {
            cat.error("Failed to find which Processor was processing Job ["
                      + job.getId()
                      + "]");
        }
        else
        {
            cat.debug("Set processor "
                      + index
                      + " free on completion of Job ["
                      + job.getId()
                      + "]");

            Processor [] processors =
                (Processor []) queueProcessors.get(queue);
            processors[index].setFree(true);
        }

        try
        {
            queueWatcher.stopWatching(job);
        }
        catch (LSFBatchException lbe)
        {
            cat.error("Caught Exception while freeing Job ["
                      + job.getId()
                      + "] from queue watcher", lbe);
        }
    }

    /**
     * <code>findProcessorIndex</code> returns the index of the
     * <code>Processor</code> which is handling the specified
     * <code>Job</code> or -1 if the <code>Job</code> is not being
     * processed.
     *
     * @param jobId an <code>int</code>.
     * @param queue a <code>String</code> queue name.
     *
     * @return an <code>int</code>.
     */
    protected int findProcessorIndex(int jobId, String queue)
    {
        Processor [] processors =
            (Processor []) queueProcessors.get(queue);

        for (int i = 0; i < processors.length; i++)
        {
            if (processors[i].getFree())
                continue;

            Job procJob =  processors[i].getJob();

            if ((procJob != null) && (jobId == procJob.getId()))
                return i;
        }

        return -1;
    }

    protected void assignReadyJobs(Collection readyJobs)
    {
        // For recording cases where there are no empty slots
        Set fullProcessors = new HashSet();

        for (Iterator ri = readyJobs.iterator(); ri.hasNext();)
        {
            Job j = (Job) ri.next();

            if (! (j instanceof BatchJob))
            {
                cat.error("Job ["
                          + j.getId()
                          + "] is not a batch job. Skipping...");
                continue;
            }

            BatchJob job = (BatchJob) j;
            String queue = job.getQueue();

            // A little shortcut to save iterating through the
            // processor array if we already know that there are no
            // empty slots
            if (fullProcessors.contains(queue))
            {
                cat.debug("Not looking for free processor for queue '"
                          + queue
                          + "' because previous attempt showed none available");
                continue;
            }

            cat.debug("Looking for free processor for queue '"
                      + queue
                      + "'");

            Processor proc = getFreeProcessor(queue);

            if (proc != null)
            {
                cat.debug("About to start " + job);

                // Add listener for changes in job status made by the
                // processor
                PropertyChangeListener listener = createUpdateListener();
                job.addPropertyChangeListener("status", listener);

                try
                {
                    proc.process(job);
                    queueWatcher.watchFor(job);
                }
                catch (Exception e)
                {
                    cat.error("Caught Exception while starting Job ["
                              + job.getId()
                              + "]", e);

                    cat.debug("Removing status update listener from Job ["
                              + job.getId()
                              + "]");

                    job.removePropertyChangeListener(listener); 
                    setFreeProcessor(job);
                }
            }
            else
            {
                cat.debug("Failed to find a free Processor for queue '"
                          + queue
                          + "'. Will not check again");
                fullProcessors.add(queue);
            }
        }
    }

    protected PropertyChangeListener createUpdateListener()
    {
        return new UpdateListener();
    }

    /**
     * <code>UpdateListener</code> is a listener for
     * <code>Status</code> changes in a <code>Job</code> which will
     * update the <code>Job</code> in the store and then remove itself
     * from the <code>Job</code>.
     */
    class UpdateListener implements PropertyChangeListener
    {
        /**
         * <code>propertyChange</code> updates the <code>Job</code>
         * via the enclosing class' <code>JobDAO</code>. If the status
         * change is to COMPLETED or FAILED the listener removes
         * itself from the <code>Job</code> and frees the relevant
         * <code>Processor</code> in the enclosing class.
         *
         * @param evt a <code>PropertyChangeEvent</code>.
         */
        public void propertyChange(PropertyChangeEvent evt)
        {
            BatchJob job = (BatchJob) evt.getSource();
            int oldState = ((Status) evt.getOldValue()).getId();
            int newState = ((Status) evt.getNewValue()).getId();

            boolean terminated = false;

            try
            {
                switch (newState)
                {
                    case (Status.STARTED):
                        jobDAO.updateJobStarted(job);
                        break;

                    case(Status.COMPLETED):
                        jobDAO.updateJobCompleted(job);
                        terminated = true;
                        break;

                    case (Status.FAILED):
                        jobDAO.updateJobFailedCascade(job);
                        terminated = true;
                        break;

                    case (Status.CANCELLED):
                        terminated = true;
                        break;

                    case (Status.SKIPPED):
                        terminated = true;
                        break;

                    default:
                        cat.warn("Coordinator was asked to apply unexpected update of Job ["
                                 + job.getId()
                                 + "] from status "
                                 + oldState
                                 + " to "
                                 + newState); 
                }
            }
            catch (Exception e)
            {
                cat.error("Coordinator failed to update Job ["
                          + job.getId()
                          + "]", e);
            }
            finally
            {
                if (terminated)
                {
                    setFreeProcessor(job);
                    job.removePropertyChangeListener(this);
                }
            }
        }
    }
}
