/*
 * Copyright (c) 2003 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.run;

import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.dao.TaskStatisticDAO;

/**
 * <code>PolicyRoundRobinImpl</code> implements a round-robin
 * selection policy where the unfinished <code>Task</code>s selected
 * in turn.
 *
 * @author Keith James
 */
public final class PolicyRoundRobinImpl implements Policy
{
    // Log4j setup
    static Category cat =
        Category.getInstance(PolicyRoundRobinImpl.class.getName());

    private TaskStatisticDAO taskStatisticDAO;
    private LinkedList idQueue;
    private Set queuedIds;

    /**
     * Creates a new <code>PolicyFIFOImpl</code> using the specified
     * data access object.
     *
     * @param taskStatisticDAO a <code>TaskStatisticDAO</code>.
     */
    public PolicyRoundRobinImpl(TaskStatisticDAO taskStatisticDAO)
    {
        this.taskStatisticDAO = taskStatisticDAO;
        idQueue = new LinkedList();
        queuedIds = new HashSet();
    }

    public int selectTaskId()
    {
        int [] taskIds;
        int taskId = -1;

        // All queued IDs should be accounted for in the set
        if (queuedIds.size() != idQueue.size())
        {
            cat.warn("Task ID queue '"
                     + idQueue
                     + " does not contain the "
                     + "same elements as Task ID set '"
                     + queuedIds);
        }

        // First add any IDs to the queue (and set) if they are not
        // already known
        try
        {
            cat.debug("Reading Task statistics");

            taskIds = taskStatisticDAO.readSubmittedTaskId();

            cat.debug("Found "
                      + taskIds.length
                      + " submitted Task(s)");

            Arrays.sort(taskIds);

            for (int i = 0; i < taskIds.length; i++)
            {
                Integer id = new Integer(taskIds[i]);

                if (! queuedIds.contains(id))
                {
                    cat.debug("Found new submitted Task with ID '"
                              + id
                              + "'");
                    queuedIds.add(id);
                    idQueue.add(id);
                }
            }
        }
        catch (Exception e)
        {
            cat.error("Caught Exception while fetching Task IDs", e);
            return -1;
        }

        // Now dequeue IDs (and remove from set) until we find one
        // which was present when we last queried the database
        while (idQueue.size() > 0)
        {
            Integer id = (Integer) idQueue.removeFirst();
            queuedIds.remove(id);

            cat.debug("Examining Task ["
                      + id
                      + "] ...");

            if (Arrays.binarySearch(taskIds, id.intValue()) < 0)
            {
                cat.debug("Task ["
                          + id
                          + "] is no longer running");

                continue;
            }
            else
            {
                cat.debug("Task ["
                          + id
                          + "] has been selected");

                taskId = id.intValue();
                break;
            }
        }

        return taskId;
    }
}
