/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.run;

import uk.ac.sanger.jcon.job.Job;

/**
 * <code>Processor</code>s processes one <code>Job</code> at a time by
 * launching the specified program. In general a
 * <code>Processor</code> may not know when its <code>Job</code> has
 * completed because it may delegate to another system, such as a
 * load-sharing system or batch queue. Having each
 * <code>Processor</code> thread wait or poll an object rather than
 * simply exiting could prevent the system scaling to large numbers of
 * concurrent <code>Job</code>s. The role of a <code>Processor</code>
 * is to do the minimum work necessary to ensure that a
 * <code>Job</code> runs and to change the <code>Job</code>'s
 * <code>Status</code> accordingly. The <code>Coordinator</code>
 * should have assigned listeners for relevant changes in
 * <code>Status</code> so that it (rather than the
 * <code>Processor</code>) knows when it succeeds or fails.
 *
 * @author Keith James
 */
public interface Processor
{
    /**
     * <code>getJob</code> returns the <code>Job</code> currently
     * being processed.
     *
     * @return a <code>Job</code>.
     */
    public Job getJob();

    /**
     * <code>getFree</code> returns true if the processor is free to
     * accept a new <code>Job</code>.
     *
     * @return a <code>boolean</code>.
     */
    public boolean getFree();

    /**
     * <code>setFree</code> sets the processor as free to accept a new
     * <code>Job</code>.
     *
     * @param setFree a <code>boolean</code>.
     */
    public void setFree(boolean isFree);

    /**
     * <code>process</code> starts the Job running thread.
     *
     * @param job a <code>Job</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void process(Job job) throws Exception;
}
