/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.run;

import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.dao.StatusDAO;
import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.Status;

/**
 * <code>ProcessorDefaultImpl</code> processes one <code>Job</code> at
 * a time by launching the specified program. This is for running
 * local jobs. For using batch queues/LSF there is a separate
 * implementation. See the <code>Processor</code> documentation for
 * further information on the expected behaviour of
 * <code>Processor</code>s.
 *
 * @author Keith James
 */
public class ProcessorDefaultImpl extends AbstractProcessor
{
    // Log4j setup
    static Category cat =
        Category.getInstance(ProcessorDefaultImpl.class.getName());

    /**
     * <code>FACTORY</code> for creating new
     * <code>ProcessorDefaultImpl</code>.
     */
    public static final ProcessorFactory FACTORY =
        new ProcessorFactory()
        {
            public Processor createProcessor(CommandLineFactory clFactory,
                                             StatusDAO          statusDAO)
            {
                return new ProcessorDefaultImpl(clFactory, statusDAO);
            }
        };

    private int    jobId;
    private String inputName;
    private String outputName;
    private String errorName;
    private String workDirectory;
    private String environment;
    private String command;

    /**
     * Creates a new <code>ProcessorDefaultImpl</code> using the
     * specified <code>CommandLineFactory</code> and data access
     * object.
     *
     * @param statusDAO a <code>StatusDAO</code>.
     */
    public ProcessorDefaultImpl(CommandLineFactory clFactory,
                                StatusDAO          statusDAO)
    {
        super(clFactory, statusDAO);
    }

    /**
     * <code>run</code> launches the program specified by the
     * <code>Job</code> and waits for it to exit. The input, output
     * and error streams of the program are handled. The
     * <code>Job</code>'s exit value and <code>Status</code> are
     * updated.
     */
    public void run()
    {
        Process process;

        try
        {
            jobId         = job.getId();
            inputName     = job.getInputName();
            outputName    = job.getOutputName();
            errorName     = job.getErrorName();
            workDirectory = job.getWorkDirectoryName();
            environment   = job.getEnvironment();

            File dir = null;
            if (workDirectory != null)
                dir = new File(workDirectory);

            String [] env = null;
            if (environment != null)
                env = stringTokenise(environment);

            // Create the command and arguments
            command = clFactory.createCommandLine(job);

            Runtime runtime = Runtime.getRuntime();
            process = runtime.exec(command, env, dir);

            job.setStatus(statusDAO.readStatusById(Status.STARTED));

            cat.info("Running Job ["
                     + jobId
                     + "] command '"
                     + command + "'");

            // FIXME - use a thread pool to prevent too many threads

            // Threads to perform IO asynchronously
            BufferedStreamEmptier  inEmptier = null;
            BufferedStreamEmptier outEmptier = null;
            BufferedStreamEmptier errEmptier = null;

            // If we need to get STDIN from a file
            if (inputName != null)
            {
                File stdinFile = null;
                if  (dir != null)
                    stdinFile = new File(dir, inputName);
                else
                    stdinFile = new File(inputName);

                InputStream stdinFileStream = new FileInputStream(stdinFile);
                // Read from file on separate thread
                inEmptier = new BufferedStreamEmptier(stdinFileStream,
                                                      process.getOutputStream());
                inEmptier.start();
            }

            // If STDOUT needs to go to a file
            if (outputName != null)
            {
                File stdoutFile = null;
                if (dir != null)
                    stdoutFile = new File(dir, outputName);
                else
                    stdoutFile = new File(outputName);

                OutputStream stdoutFileStream = new FileOutputStream(stdoutFile);
                // Read from program's STDOUT on separate thread
                outEmptier = new BufferedStreamEmptier(process.getInputStream(),
                                                       stdoutFileStream);
                outEmptier.start();
            }

            // If we need to save STDERR to a file
            if (errorName != null)
            {
                File stderrFile = null;
                if (dir != null)
                    stderrFile = new File(dir, errorName);
                else
                    stderrFile = new File(errorName);

                OutputStream stderrFileStream = new FileOutputStream(stderrFile);
                // Read from program's STDERR on separate thread
                errEmptier = new BufferedStreamEmptier(process.getErrorStream(),
                                                       stderrFileStream);
                errEmptier.start();
            }

            // FIXME - need to think about timing out programs which hang
            int exitValue = process.waitFor();

            // If we started any of these threads wait here for them to
            // flush their buffers, otherwise IO may be truncated
            if (inEmptier != null)
                inEmptier.join();
            if (outEmptier != null)
                outEmptier.join();
            if (errEmptier != null)
                errEmptier.join();

            job.setExitValue(exitValue);

            // Set the Status, triggering listeners
            if (exitValue == 0)
                job.setStatus(statusDAO.readStatusById(Status.COMPLETED));
            else
                job.setStatus(statusDAO.readStatusById(Status.FAILED));

            cat.info("Completed Job ["
                     + jobId
                     + "] command '"
                     + command
                     + "' with exit value "
                     + exitValue);
        }
        catch (Exception e1)
        {
            cat.error("Caught Exception while running", e1);
            try
            {
                if (job != null)
                    job.setStatus(statusDAO.readStatusById(Status.FAILED));
            }
            catch (Exception e2)
            {
                cat.error("Caught Exception while failing Job", e2);
            }
        }
        finally
        {
            process = null;
        }
    }
}
