/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.run;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControlException;
import uk.ac.sanger.jcon.dao.StatusDAO;
import uk.ac.sanger.jcon.job.BatchJob;
import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.lsf.LSFBatchException;
import uk.ac.sanger.jcon.lsf.LSFBsubExec;

/**
 * <code>ProcessorLSFImpl</code> processes one <code>Job</code> at a
 * time by launching the specified program. This is for running jobs
 * using LSF batch queues. There is a separate implementation for
 * running local jobs. See the <code>Processor</code> documentation for
 * further information on the expected behaviour of
 * <code>Processor</code>s.
 *
 * @author Keith James
 */
public final class ProcessorLSFImpl implements Processor
{
    // Log4j setup
    static Category cat =
        Category.getInstance(ProcessorLSFImpl.class.getName());

    /**
     * <code>FACTORY</code> for creating new
     * <code>ProcessorLSFImpl</code>.
     */
    public static final ProcessorFactory FACTORY =
        new ProcessorFactory()
        {
            public Processor createProcessor(CommandLineFactory clFactory,
                                             StatusDAO          statusDAO)
            {
                return new ProcessorLSFImpl(clFactory, statusDAO);
            }
        };

    private CommandLineFactory clFactory;
    private StatusDAO          statusDAO;

    private boolean isFree;
    private Job     job;

    private String inputName;
    private String outputName;
    private String errorName;
    private String workDirectory;
    private String environment;
    private String command;

    /**
     * <code>getJob</code> returns the <code>Job</code> currently
     * being processed.
     *
     * @return a <code>Job</code>.
     */
    public Job getJob()
    {
        return job;
    }

    /**
     * <code>getFree</code> returns true if the processor is free to
     * accept a new <code>Job</code>.
     *
     * @return a <code>boolean</code>.
     */
    public boolean getFree()
    {
        return isFree;
    }

    public void setFree(boolean isFree)
    {
        this.isFree = isFree;

        if (isFree)
            job = null;
    }

    /**
     * Creates a new <code>ProcessorLSFImpl</code> using the
     * specified <code>CommandLineFactory</code> and data access
     * object.
     *
     * @param statusDAO a <code>StatusDAO</code>.
     */
    public ProcessorLSFImpl(CommandLineFactory clFactory,
                            StatusDAO          statusDAO)
    {
        this.clFactory = clFactory;
        this.statusDAO = statusDAO;
        isFree = true;
    }

    /**
     * <code>process</code> starts the Job running thread. At the
     * moment this only accepts <code>Job</code>s which are instances
     * of <code>BatchJob</code>. In future this may change to that a
     * <code>Job</code> passed here is handled by a default queue
     * instead of being rejected.
     *
     * @param job a <code>Job</code>.
     *
     * @exception Exception if an error occurs.
     */
    public void process(Job job) throws Exception
    {
        if (! (job instanceof BatchJob))
            throw new JobControlException("Failed to process Job ["
                                          + job.getId()
                                          + "] as it was not a batch job");
        if (! getFree())
            throw new JobControlException("Failed to process Job as "
                                          + "Processor is not free");

        setFree(false);
        this.job = job;

        BatchJob lsfJob = (BatchJob) job;
 
        String [] env = null;
        if (environment != null)
            env = stringTokenize(environment);

        // Create the command and arguments
        command = clFactory.createCommandLine(job);

        String jobIdStr = Integer.toString(lsfJob.getId());

        cat.info("Running batch Job ["
                 + jobIdStr
                 + "] command '"
                 + command
                 + "'");

        LSFBsubExec lbs = LSFBsubExec.getInstance();

        int lsfJobId = lbs.submit(jobIdStr,
                                  lsfJob.getQueue(),
                                  lsfJob.getInputName(),
                                  lsfJob.getOutputName(),
                                  lsfJob.getErrorName(),
                                  command,
                                  env,
                                  lsfJob.getWorkDirectoryName());

        cat.debug("Obtained new LSF Job ID "
                  + lsfJobId
                  + " for Job ["
                  + jobIdStr
                  + "]");

        // If we get this far we can set batch ID and new Status. The
        // Coordinator handles all the cleanup if an Exception is
        // thrown
        lsfJob.setBatchId(lsfJobId);
        lsfJob.setStatus(statusDAO.readStatusById(Status.STARTED));
    }

    /**
     * <code>stringTokenize</code> creates an array of strings by
     * splitting on whitespace with the default
     * <code>StringTokenizer</code>.
     *
     * @param string a <code>String</code>.
     * @return a <code>String []</code>.
     */
    private String [] stringTokenize(String string)
    {
        StringTokenizer st = new StringTokenizer(string);
		int count = st.countTokens();

        String [] tokens = new String [count];

		for (int i = 0; i < count; i++)
        {
            tokens[i] = st.nextToken();
        }

        return tokens;
    }
}
