/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.run;

/**
 * <p>A <code>TaskScheduler</code> is a scheduler which uses a
 * <code>Policy</code> to set the unfinished <code>Task</code> its
 * <code>Coordinator</code> should process. At intervals it will use
 * the <code>Policy</code> to reasses which <code>Task</code> it
 * should assign to the <code>Coordinator</code>.</p>
 *
 * <p>Custom <code>Policy</code> implementations may be used to change
 * the behaviour of the <code>TaskScheduler</code>.</p>
 *
 * @author Keith James
 */
public interface TaskScheduler
{
    /**
     * <code>POLL_INTERVAL_KEY</code> the <code>Properties</code> key
     * to specify the polling interval in millseconds at which a
     * <code>TaskScheduler</code> will check its scheduling
     * <code>Policy</code> and potentially change the
     * <code>Task</code> being processed.
     */
    public static final String POLL_INTERVAL_KEY = "jobcontrol.scheduler.poll_interval";

    /**
     * <code>MAIL_FROM_KEY</code> the <code>Properties</code> key to
     * specify the email address of the "From" field for messages sent
     * to <code>Task</code> <code>Owner</code>s.
     */
    public static final String MAIL_FROM_KEY = "jobcontrol.scheduler.mail_from";

    /**
     * <code>getPolicy</code> returns the scheduling
     * <code>Policy</code> currently being used.
     *
     * @return a <code>Policy</code>.
     */
    public Policy getPolicy();

    /**
     * <code>setPolicy</code> sets the scheduling <code>Policy</code>.
     *
     * @param policy a <code>Policy</code>.
     */
    public void setPolicy(Policy policy);

    /**
     * <code>startScheduling</code> starts the
     * <code>TaskScheduler</code>.
     */
    public void startScheduling();

    /**
     * <code>stopScheduling</code> stops the
     * <code>TaskScheduler</code>.
     */
    public void stopScheduling();

    /**
     * <code>getSendMail</code> returns whether state changes are
     * being sent out by email.
     *
     * @return a <code>boolean</code>.
     */
    public boolean getSendMail();

    /**
     * <code>setSendMail</code> sets whether state changes are to be
     * sent out by email.
     *
     * @param sendMail a <code>boolean</code>.
     */
    public void setSendMail(boolean sendMail);
}
