/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.run;

import java.util.Enumeration;
import java.util.ResourceBundle;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.JobControl;
import uk.ac.sanger.jcon.job.Task;
import uk.ac.sanger.jcon.dao.TaskDAO;

/**
 * <p><code>TaskSchedulerDefaultImpl</code> is a scheduler which uses
 * a <code>Policy</code> to set the unfinished <code>Task</code> its
 * <code>Coordinator</code> should process. It runs in a separate
 * thread.</p>
 *
 * @author Keith James
 */
public class TaskSchedulerDefaultImpl extends AbstractTaskScheduler
    implements Runnable
{
    /**
     * <code>DEFAULT_POLL_INTERVAL</code> is the default polling
     * interval in millseconds at which
     * <code>TaskSchedulerDefaultImpl</code> will check the database
     * for new ready <code>Job</code>s.
     */
    public static final int DEFAULT_POLL_INTERVAL = 30000;

    /**
     * <code>DEFAULT_MAIL_FROM</code> is the default email address of
     * the "From" field for messages sent to <code>Task</code>
     * <code>Owner</code>s.
     */
    public static final String DEFAULT_MAIL_FROM = "jobcontrol@sanger.ac.uk";

    // Log4j setup
    static Category cat =
        Category.getInstance(TaskSchedulerDefaultImpl.class.getName());

    // Scheduler thread
    private Thread thread;

    /**
     * Creates a new <code>TaskSchedulerDefaultImpl</code>.
     *
     * @param pidFileName a <code>String</code> file name.
     * @param taskDAO a <code>TaskDAO</code>.
     * @param coordinator a <code>Coordinator</code>.
     */
    public TaskSchedulerDefaultImpl(TaskDAO     taskDAO,
                                    Coordinator coordinator)
    {
        super(taskDAO, coordinator);

        policyPollInterval = DEFAULT_POLL_INTERVAL;
        mailFrom           = DEFAULT_MAIL_FROM;

        ResourceBundle bundle = JobControl.getConfiguration();

        String key;
        String val;

        try
        {
            Enumeration e = bundle.getKeys();
            while (e.hasMoreElements())
            {
                key = (String) e.nextElement();

                try
                {
                    val = bundle.getString(key);
                }
                catch (NullPointerException npe)
                {
                    throw new Error("Malformed configuration ResourceBundle "
                                    + "contains key '"
                                    + key
                                    + "' missing a corresponding value");
                }

                if (key.equals(TaskScheduler.POLL_INTERVAL_KEY))
                    policyPollInterval = Integer.parseInt(val);

                if (key.equals(TaskScheduler.MAIL_FROM_KEY))
                    mailFrom = val;
            }
        }
        catch (NumberFormatException nfe)
        {
            throw new Error("Number format error in "
                            + "configuration ResourceBundle");
        }

        thread = new Thread(this);
        isScheduling = false;
    }

    /**
     * <code>startScheduling</code> starts the
     * <code>TaskScheduler</code>.
     *
     * @exception IllegalStateException if the
     * <code>TaskScheduler</code> is running or the
     * <code>Policy</code> is null.
     */
    public void startScheduling() throws IllegalStateException
    {
        if (isScheduling == true)
            throw new IllegalStateException("TaskScheduler is already running");
        else if (policy == null)
            throw new IllegalStateException("Failed to start TaskScheduler as "
                                            + " its scheduling Policy is null");

        isScheduling = true;
        thread.start();
        cat.info("TaskScheduler started");
    }

    /**
     * <code>stopScheduling</code> stops the
     * <code>TaskScheduler</code>.
     *
     * @exception IllegalStateException if the
     * <code>TaskScheduler</code> is not running.
     */
    public void stopScheduling() throws IllegalStateException
    {
        if (isScheduling == false)
             throw new IllegalStateException("TaskScheduler is already stopped");

        isScheduling = false;
        cat.info("TaskScheduler stopped");
    }

    /**
     * <code>run</code> is the main scheduler.
     */
    public void run()
    {
        schedule();
    }
}
