/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.servlets;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * <p><code>Table</code> represents a simple table consisting of rows
 * of cells. Rows may have different numbers of cells from each
 * other. Tables and rows automatically expand or contract when
 * elements (rows or cells, repectively) are added or removed. As cell
 * width is undefined, there is no API for columns (vertical stacks of
 * cells).</p>
 *
 * @author Keith James
 */
public class Table
{
    private List rows;

    /**
     * Creates a new empty <code>Table</code> instance with a capacity
     * of 10 rows.
     */
    public Table()
    {
        this(10);
    }

    /**
     * Creates a new <code>Table</code> instance with the specified
     * initial capacity.
     *
     * @param rowCount an <code>int</code> initial capacity.
     */
    public Table(int rowCount)
    {
        rows = new ArrayList(rowCount);
    }

    /**
     * <code>addRow</code> adds a row to the bottom of the table.
     *
     * @param row a <code>Row</code>.
     */
    public void addRow(Row row)
    {
        rows.add(row);
    }

    /**
     * <code>getRowCount</code> returns the number of rows in the
     * table.
     *
     * @return an <code>int</code>.
     */
    public int getRowCount()
    {
        return rows.size();
    }

    /**
     * <code>getRowAt</code> returns the row at the specified index.
     *
     * @param rowIndex an <code>int</code> index.
     *
     * @return a <code>Row</code>.
     */
    public Row getRowAt(int rowIndex)
    {
        return (Row) rows.get(rowIndex);
    }

    /**
     * <code>getRowList</code> returns an unmodifiable list of all
     * rows.
     *
     * @return a <code>List</code> of <code>Row</code>s.
     */
    public List getRowList()
    {
        return Collections.unmodifiableList(rows);
    }

    /**
     * <code>rowIterator</code> returns an iterator over all rows.
     *
     * @return an <code>Iterator</code>.
     */
    public Iterator rowIterator()
    {
        return rows.iterator();
    }

    public String toString()
    {
        StringBuffer sb = new StringBuffer();

        for (Iterator ri = rowIterator(); ri.hasNext();)
        {
            sb.append(((Row) ri.next()).toString());
            sb.append("\n");
        }

        return sb.substring(0);
    }

    /**
     * <code>Row</code> represents a row of <code>Cell</code>s in
     * the table.
     */
    public class Row
    {
        private List cells;
        private Object meta;

        /**
         * Creates a new <code>Row</code> instance with the specified
         * initial capacity.
         *
         * @param cellCount an <code>int</code> capacity.
         */
        public Row(int cellCount)
        {
            cells = new ArrayList(cellCount);
        }

        /**
         * <code>addCell</code> adds cell to the right hand end of
         * the row.
         *
         * @param cell a <code>Cell</code>.
         */
        public void addCell(Cell cell)
        {
            cells.add(cell);
        }

        /**
         * <code>getCellCount</code> returns the number of cells in
         * the row.
         *
         * @return an <code>int</code>.
         */
        public int getCellCount()
        {
            return cells.size();
        }

        /**
         * <code>getCellAt</code> returns the cell at the specified
         * index.
         *
         * @param cellIndex an <code>int</code> index.
         *
         * @return a <code>Cell</code>.
         */
        public Cell getCellAt(int cellIndex)
        {
            return (Cell) cells.get(cellIndex);
        }

        /**
         * <code>getCellList</code> returns an unmodifiable list of
         * all cells in the row.
         *
         * @return a <code>List</code> of <code>Cell</code>s.
         */
        public List getCellList()
        {
            return Collections.unmodifiableList(cells);
        }

        /**
         * <code>cellIterator</code> returns an iterator over all
         * cells in the row.
         *
         * @return an <code>Iterator</code>.
         */
        public Iterator cellIterator()
        {
            return cells.iterator();
        }

        /**
         * <code>getMetaData</code> returns metadata for the row.
         *
         * @return an <code>Object</code>.
         */
        public Object getMetaData()
        {
            return meta;
        }

        /**
         * <code>setMetaData</code> sets metadata for the row.
         *
         * @param meta an <code>Object</code>.
         */
        public void setMetaData(Object meta)
        {
            this.meta = meta;
        }

        public String toString()
        {
            StringBuffer sb = new StringBuffer();

            for (Iterator ci = cellIterator(); ci.hasNext();)
            {
                sb.append(((Cell) ci.next()).toString());
                sb.append("\t");
            }

            return sb.substring(0, sb.length() - 1);
        }
    }

    /**
     * <code>Cell</code> represents a cell of a single
     * <code>Row</code>. It does not represent a vertical slice
     * through the whole table.
     */
    public class Cell
    {
        private Object data;
        private Object meta;

        /**
         * Creates a new <code>Cell</code> instance containing the
         * specified data.
         *
         * @param data an <code>Object</code>.
         */
        public Cell(Object data)
        {
            this.data = data;
        }

        /**
         * <code>getData</code> returns the data in the cell.
         *
         * @return an <code>Object</code>.
         */
        public Object getData()
        {
            return data;
        }

        /**
         * <code>getMetaData</code> returns metadata for the cell.
         *
         * @return an <code>Object</code>.
         */
        public Object getMetaData()
        {
            return meta;
        }

        /**
         * <code>setMetaData</code> sets metadata for the cell.
         *
         * @param meta an <code>Object</code>.
         */
        public void setMetaData(Object meta)
        {
            this.meta = meta;
        }

        public String toString()
        {
            return data.toString();
        }
    }
}
