/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.servlets;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import javax.sql.DataSource;

import org.apache.log4j.Category;
import org.apache.velocity.Template;
import org.apache.velocity.context.Context;
import org.apache.velocity.exception.ParseErrorException;
import org.apache.velocity.exception.ResourceNotFoundException;

import uk.ac.sanger.jcon.dao.*;
import uk.ac.sanger.jcon.job.Owner;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.job.Task;
import uk.ac.sanger.jcon.job.TaskState;

/**
 * <code>TemplateFactoryTaskStateImpl</code> creates a template
 * populated with <code>Task</code> status data.
 *
 * @author Keith James
 */
class TemplateFactoryTaskStateImpl implements TemplateFactory
{
    // Log4j setup
    static Category cat =
        Category.getInstance(TemplateFactoryTaskStateImpl.class.getName());

    private JobControlServlet servlet;

    TemplateFactoryTaskStateImpl(JobControlServlet servlet)
    {
        this.servlet = servlet;
    }

    public Template createTemplate(Context context) throws Exception
    {
        Template template = null;

        try
        {
            DataSource datasource = servlet.getDataSource();

            TaskDAO              taskDAO = new TaskSQLDAO(datasource);
            StatusDAO          statusDAO = new StatusSQLDAO(datasource);
            TaskStatisticDAO taskStatDAO = new TaskStatisticSQLDAO(datasource);

            Status   waiting = statusDAO.readStatusById(Status.WAITING);
            Status     ready = statusDAO.readStatusById(Status.READY);
            Status   started = statusDAO.readStatusById(Status.STARTED);
            Status completed = statusDAO.readStatusById(Status.COMPLETED);
            Status    failed = statusDAO.readStatusById(Status.FAILED);
            int waitCount, readyCount, startCount,
                compCount, failCount, totalCount;

            // Two transactions - may need to address this
            Collection tasks =
                taskDAO.readSubmittedTasks(TaskDAO.EXCLUDE_JOB_TREE);
            Collection tStats = taskStatDAO.readSubmittedTaskState();
            Map statMap = createTaskStatMap(tStats);

            Table taskTable = new Table();

            for (Iterator ti = tasks.iterator(); ti.hasNext();)
            {
                Task t = (Task) ti.next();
                Integer id = new Integer(t.getId());

                TaskState tStat = (TaskState) statMap.get(id);

                // With the current implementation it's possible that
                // Task could complete between reading submittedTasks
                // and reading the submittedTaskStats
                if (tStat == null)
                {
                    cat.warn("Task ["
                             + id
                             + "] completed before stats could be read");
                    continue;
                }

                waitCount  = tStat.getJobCount(waiting);
                readyCount = tStat.getJobCount(ready);
                startCount = tStat.getJobCount(started);
                compCount  = tStat.getJobCount(completed);
                failCount  = tStat.getJobCount(failed);
                totalCount = tStat.getTotalJobCount();

                Table.Row row = taskTable.new Row(5);

                Table.Cell cell = taskTable.new Cell(id.toString());
                cell.setMetaData(ServletConstants.CSS_ALIGN_R);

                row.addCell(cell);
                row.addCell(taskTable.new Cell(t.getName()));
                row.addCell(taskTable.new Cell(t.getOwner().getUserName()));
                row.addCell(taskTable.new Cell(t.getDescription()));

                String [] vals =
                    new String [] { Integer.toString(t.getPriority()),
                                    Integer.toString(waitCount),
                                    Integer.toString(readyCount),
                                    Integer.toString(startCount),
                                    Integer.toString(compCount),
                                    Integer.toString(failCount),
                                    Integer.toString(totalCount) };

                for (int i = 0; i < vals.length; i++)
                {
                    cell = taskTable.new Cell(vals[i]);
                    cell.setMetaData(ServletConstants.CSS_ALIGN_R);
                    row.addCell(cell);
                }

                if (id.intValue() % 2 == 0)
                    row.setMetaData(ServletConstants.CSS_DARK_BG);
                else
                    row.setMetaData(ServletConstants.CSS_PALE_BG);

                taskTable.addRow(row);
            }

            cat.debug("Result table is:\n" + taskTable);

            context.put("taskTable", taskTable);
            template = servlet.getTemplate("taskstate.vm");
        }
        catch (ResourceNotFoundException rnfe)
        {
            System.err.println("JobControlServlet: template not found");
            throw rnfe;
        }
        catch (ParseErrorException pee)
        {
            System.err.println("JobControlServlet: parse error for template");
            throw pee;
        }

        return template;
    }

    private Map createTaskStatMap(Collection tStats)
    {
        Map statMap = new HashMap();

        for (Iterator i = tStats.iterator(); i.hasNext();)
        {
            TaskState tStat = (TaskState) i.next();
            statMap.put(new Integer(tStat.getTaskId()), tStat);
        }

        return statMap;
    }
}
