/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.util;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * <p><code>Configuration</code> provides a view onto a simply
 * structured <code>ResourceBundle</code> such that software
 * configuration may be carried out using terms (aliases) defined
 * externally to the compiled code.</p>
 *
 * <p>This is a temporary means of configuring a
 * <code>JobFactory</code> until a more structured approach is
 * developed.</p>
 *
 * <p>For example, although organisms are identified by their NCBI
 * taxonomy ID, users may prefer the more familiar genus-species
 * terms. The configuration would define a genus-species alias and a
 * mapping from that alias to the relevant ID. Software using the
 * configuration would then accept instructions using the alias and
 * convert to an ID internally.</p>
 *
 * <p>The configuration structure consists of nested fields and is as
 * follows:</p>
 *
 * <ul>
 *  <li><code>executable</code> may contain</li>
 *    <ul>
 *      <li><code>alias</code></li>
 *      <li><code>id</code></li>
 *      <li><code>run</code></li>
 *    </ul>
 *
 *  <li><code>extension</code> may contain</li>
 *    <ul>
 *      <li><code>embl</code></li>
 *      <li><code>error</code></li>
 *      <li><code>fasta</code></li>
 *      <li><code>featuretable</code></li>
 *      <li><code>input</code></li>
 *      <li><code>output</code></li>
 *    </ul>
 *
 *  <li><code>fasta</code> may contain</li>
 *    <ul>
 *      <li><code>aminoacid</code></li>
 *      <li><code>nucleotide</code></li>
 *    </ul>
 *
 *  <li><code>file</code> may contain</li>
 *    <ul>
 *      <li><code>extension</code></li>
 *    </ul>
 *
 *  <li><code>organism</code> may contain</li>
 *    <ul>
 *      <li><code>alias</code></li>
 *      <li><code>taxid</code></li>
 *    </ul>
 *
 *  <li><code>run</code> may contain</li>
 *    <ul>
 *      <li><code>file</code></li>
 *      <li><code>queue</code></li>
 *    </ul>
 *
 *  <li><code>sequence</code> may contain</li>
 *    <ul>
 *      <li><code>file</code></li>
 *    </ul>
 *
 *  <li><code>template</code> may contain</li>
 *    <ul>
 *      <li><code>alias</code></li>
 *    </ul>
 *  </ul>
 *
 * <p>One or more aliases may be appended to the terminal field to
 * specify a user-configurable mapping.</p>
 *
 * <p>Example:</p>
 *
 * <pre>
 * sequence.file.extension.fasta.nucleotide   nt
 * sequence.file.extension.fasta.aminoacid    aa
 * sequence.file.extension.embl               embl
 * sequence.file.extension.featuretable       tab
 *
 * executable.run.file.extension.input        inp
 * executable.run.file.extension.output       out
 * executable.run.file.extension.error        err
 * executable.run.queue                       babelq2
 *
 * # Define aliases
 * executable.alias                           genefinder geneid hmmgene
 * # Use aliases to map executable.id to value
 * executable.id.genefinder                   22
 * executable.id.geneid                       23
 * executable.id.hmmgene                      25
 *
 * # Define alias
 * organism.alias                             d_discoideum
 * # Use alias to map organism.taxid to value
 * organism.taxid.d_discoideum                44689
 * </pre>
 *
 * @author Keith James
 */
public class Configuration
{
	public static final String         FILE = "file";
	public static final String    EXTENSION = "extension";
	public static final String           ID = "id";
	public static final String        ALIAS = "alias";
	public static final String     TEMPLATE = "template";

	public static final String     SEQUENCE = "sequence";
	public static final String      NUCLEIC = "nucleotide";
	public static final String        AMINO = "aminoacid";
	public static final String        FASTA = "fasta";
	public static final String         EMBL = "embl";
	public static final String FEATURETABLE = "featuretable";

	public static final String   EXECUTABLE = "executable";
	public static final String          RUN = "run";
	public static final String        INPUT = "input";
	public static final String       OUTPUT = "output";
	public static final String        ERROR = "error";
	public static final String        QUEUE = "queue";

	public static final String     ORGANISM = "organism";
	public static final String        TAXID = "taxid";

	public static final String FASTA_NT_EXTENSION_KEY =
		joinComponents(new String [] { SEQUENCE, FILE, EXTENSION,
									   FASTA, NUCLEIC });
	
	public static final String FASTA_AA_EXTENSION_KEY =
		joinComponents(new String [] { SEQUENCE, FILE, EXTENSION,
									   FASTA, AMINO });

	public static final String EMBL_EXTENSION_KEY =
		joinComponents(new String [] { SEQUENCE, FILE, EXTENSION,
									   EMBL });

	public static final String FTABLE_EXTENSION_KEY =
		joinComponents(new String [] { SEQUENCE, FILE, EXTENSION,
									   FEATURETABLE });

	public static final String INPUT_EXTENSION_KEY =
		joinComponents(new String [] { EXECUTABLE, RUN, FILE, EXTENSION,
									   INPUT });

	public static final String OUTPUT_EXTENSION_KEY =
		joinComponents(new String [] { EXECUTABLE, RUN, FILE, EXTENSION,
									   OUTPUT });

	public static final String ERROR_EXTENSION_KEY =
		joinComponents(new String [] { EXECUTABLE, RUN, FILE, EXTENSION,
									   ERROR });

	public static final String EXEC_QUEUE_KEY =
		joinComponents(new String [] { EXECUTABLE, RUN, QUEUE });

	public static final String EXEC_ALIAS_KEY =
		joinComponents(new String [] { EXECUTABLE, ALIAS });

	public static final String EXEC_ID_KEY =
		joinComponents(new String [] { EXECUTABLE, ID });

	public static final String ORG_ALIAS_KEY =
		joinComponents(new String [] { ORGANISM, ALIAS });

	public static final String ORG_TAXID_KEY =
		joinComponents(new String [] { ORGANISM, TAXID });

	private ResourceBundle resource;
	private Set execAliases;
	private Set organismAliases;

	private Map execIds;
	private Map taxIds;

	/**
	 * Creates a new <code>Configuration</code> instance.
	 *
	 * @param resource a <code>ResourceBundle</code>.
	 */
	public Configuration(ResourceBundle resource)
	{
		this.resource = resource;

		initConfiguration();
	}

	/**
	 * <code>getExecutableAliases</code> returns all aliases of
	 * executables.
	 *
	 * @return a <code>Set</code> of <code>String</code>s.
	 */
	public Set getExecutableAliases()
	{
		return Collections.unmodifiableSet(execAliases);
	}

	/**
	 * <code>getOrganismAliases</code> returns all aliases of
	 * organisms.
	 *
	 * @return a <code>Set</code> of <code>String</code>s.
	 */
	public Set getOrganismAliases()
	{
		return Collections.unmodifiableSet(organismAliases);
	}

	/**
	 * <code>getExecutableId</code> returns the ID bound to an
	 * executable alias.
	 *
	 * @param execAlias a <code>String</code>.
	 *
	 * @return an <code>int</code> ID.
	 */
	public int getExecutableId(String execAlias)
	{
		if (! execAliases.contains(execAlias))
			throw new IllegalArgumentException("Unknown alias '"
											   + execAlias
											   + "'");
		if (! execIds.containsKey(execAlias))
			throw new IllegalArgumentException("No ID bound to alias '"
											   + execAlias
											   + "'");

		return ((Integer) execIds.get(execAlias)).intValue();
	}

	/**
	 * <code>getOrganismId</code> returns the ID bound to an organism
	 * alias.
	 *
	 * @param organismAlias a <code>String</code>.
	 *
	 * @return an <code>int</code> ID.
	 */
	public int getTaxonomyId(String organismAlias)
	{
		if (! organismAliases.contains(organismAlias))
			throw new IllegalArgumentException("Unknown alias '"
											   + organismAlias
											   + "'");

		if (! taxIds.containsKey(organismAlias))
			throw new IllegalArgumentException("No Taxonomy ID bound to alias '"
											   + organismAlias
											   + "'");

		return ((Integer) taxIds.get(organismAlias)).intValue();
	}

	/**
	 * <code>getTemplate</code> returns a command line template for
	 * the specified organism and executable aliases.
	 *
	 * @param organismAlias a <code>String</code>.
	 * @param execAlias a <code>String</code>.
	 *
	 * @return a <code>String</code>.
	 */
	public String getTemplate(String organismAlias, String execAlias)
	{
		if (! organismAliases.contains(organismAlias))
			throw new IllegalArgumentException("Unknown alias '"
											   + organismAlias
											   + "'");
		if (! execAliases.contains(execAlias))
			throw new IllegalArgumentException("Unknown alias '"
											   + execAlias
											   + "'");

		return resource.getString(TEMPLATE
								  + "."
								  + organismAlias
								  + "."
								  + execAlias);
	}



	/**
	 * <code>getString</code> gets an object from the underlying
	 * <code>ResourceBundle</code>.
	 *
	 * @param key a <code>String</code>.
	 *
	 * @return a <code>String</code>.
	 */
	public String getString(String key)
	{
		return resource.getString(key);
	}

	private void initConfiguration()
	{
		execAliases     = initAliases(EXECUTABLE);
		organismAliases = initAliases(ORGANISM);

		execIds = initIdMap(EXEC_ID_KEY, execAliases);
		taxIds  = initIdMap(ORG_TAXID_KEY, organismAliases);
	}

	private Set initAliases(String baseKey)
	{
		Set aliases = new HashSet();

		String key = baseKey + "." + ALIAS;
		String [] val = stringTokenize(resource.getString(key));

		for (int i = 0; i < val.length; i++)
		{
			aliases.add(val[i]);
		}

		return aliases;
	}

	private Map initIdMap(String baseKey, Set aliases)
	{
		Map ids = new HashMap();

		String key, val;

		for (Iterator i = aliases.iterator(); i.hasNext();)
		{
			String alias = (String) i.next();

			key = baseKey + "." + alias;
			val = resource.getString(key);

			ids.put(alias, Integer.decode(val));
		}

		return ids;
	}

	private String [] stringTokenize(String val)
	{
		StringTokenizer st = new StringTokenizer(val);

		int count = st.countTokens();

		String [] tokens = new String [count];
		for (int i = 0; i < count; i++)
		{
			tokens[i] = st.nextToken();
		}

		return tokens;
	}

	private static String joinComponents(String [] components)
	{
		StringBuffer sb = new StringBuffer();

		for (int i = 0; i < components.length; i++)
		{
			sb.append(components[i]);
			sb.append(".");
		}

		return sb.substring(0, sb.length() - 1);
	}
}
