/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.util;

import java.util.Enumeration;
import java.util.MissingResourceException;
import java.util.Properties;
import java.util.ResourceBundle;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;

import org.apache.log4j.Category;

/**
 * <code>Mailer</code> is a simple class for sending email messages.
 *
 * @author Keith James
 */
public class Mailer
{
    // Log4j setup
    static Category cat = Category.getInstance(Mailer.class.getName());

    static ResourceBundle bundle = ResourceBundle.getBundle(Mailer.class.getName());

    private static Session session;

    /**
     * Creates a new <code>Mailer</code>.
     */
    public Mailer() throws MissingResourceException
    {
        Enumeration e = bundle.getKeys();
        Properties props = new Properties();

        String key;
        String val;

        while (e.hasMoreElements())
        {
             key = (String) e.nextElement();

             try
             {
                 val = bundle.getString(key);
             }
             catch (NullPointerException npe)
             {
                 throw new Error("Malformed ResourceBundle "
                                 + Mailer.class.getName()
                                 + " contains key '"
                                 + key
                                 + "' missing a corresponding value");
             }

             props.setProperty(key, val);
        }

        session = Session.getDefaultInstance(props);
    }

    /**
     * <code>sendMessage</code> sends a message with the specified
     * subject and content to the recipients.
     *
     * @param recipients a <code>String []</code> array of recipients'
     * email addresses.
     * @param from a <code>String</code> sender address.
     * @param subject a <code>String</code> message subject.
     * @param content a <code>String</code> message body.
     *
     * @exception MessagingException if an error occurs.
     */
    public static void sendMessage(String [] recipients, String from,
                                   String subject, String content)
        throws MessagingException
    {
        if (recipients.length == 0)
            throw new IllegalArgumentException("No recipients defined");

        final Message message = new MimeMessage(session);

        InternetAddress fromAddr = new InternetAddress(from);
        InternetAddress [] recipientAddr =
            new InternetAddress[recipients.length];
        for (int i = 0; i < recipientAddr.length; i++)
        {
            recipientAddr[i] = new InternetAddress(recipients[i]);
        }

        message.setFrom(fromAddr);
        message.setRecipients(Message.RecipientType.TO, recipientAddr);
        message.setSubject(subject);
        message.setContent(content, "text/plain");

        StringBuffer logDetail = new StringBuffer("Mailer sending to ");
        logDetail.append(recipients[0]);

        if (recipients.length > 1)
        {
            logDetail.append(" and ");
            logDetail.append(recipients.length - 1);
            logDetail.append(" others");
        }

        logDetail.append(", from ");
        logDetail.append(fromAddr);
        logDetail.append(". Subject: ");
        logDetail.append(subject);

        cat.info(logDetail);

        Thread mailThread = new Thread()
        {
            public void run()
            {
                try
                {
                    Transport.send(message);
                }
                catch (MessagingException me)
                {
                    cat.error("Mailer failed to send message", me);
                }
            }
        };

        mailThread.start();
    }
}
