/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.util;

import java.util.Enumeration;
import javax.swing.tree.DefaultMutableTreeNode;

import uk.ac.sanger.jcon.job.Job;

/**
 * <code>PostorderEnumeration</code> implements postorder traversal of
 * a <code>Job</code> tree. Uses the <code>Enumeration</code>
 * interface instead of <code>Iterator</code> because the
 * <code>TreeNode</code> interface specifies the former in some
 * methods. Based on code from <code>DefaultMutableTreeNode</code>
 * from Sun JDK by Rob Davis.
 *
 * @author Keith James
 */
public final class PostorderEnumeration implements Enumeration
{
    private Job rootNode;
    private Enumeration children;
    private Enumeration subTree;

    /**
     * Creates a new <code>PostorderEnumeration</code>.
     *
     * @param rootNode a <code>Job</code>.
     */
    public PostorderEnumeration(Job rootNode)
    {
        this.rootNode = rootNode;
        children = rootNode.children();
        subTree = DefaultMutableTreeNode.EMPTY_ENUMERATION;
    }

    public boolean hasMoreElements()
    {
        return rootNode != null;
    }

    public Object nextElement()
    {
        Object node = null;

        if (subTree.hasMoreElements())
        {
            node = subTree.nextElement();
        }
        else if (children.hasMoreElements())
        {
            // Extend chain recursively
            subTree = new PostorderEnumeration((Job) children.nextElement());
            // Absent sub-tree has been replaced with EMPTY_ENUMERATION
            node = subTree.nextElement();
        }
        else
        {
            node = rootNode;
            rootNode = null;
        }

        return node;
    }
}
