/*
 * Copyright (c) 2003 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.util;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.StringTokenizer;
import java.util.TreeMap;

import uk.ac.sanger.jcon.JobControlException;

/**
 * <code>UserConfiguration</code> is a representation of configuration
 * data supplied as a properties file whose format is described in the
 * user guide. It allows simple chaining of arbitrary jobs where the
 * executables involved communicate data only via STDIN/STDOUT.
 *
 * @author Keith James
 */
public class UserConfiguration
{
    /**
     * <code>PROTOCOL_NAME_KEY</code> is the key for protocol name in
     * the user defined configuration file
     */
    public static final String PROTOCOL_NAME_KEY = "protocol.name";

    /**
     * <code>PROTOCOL_DESC_KEY</code> is the key for protocol
     * description in the user defined configuration file.
     */
    public static final String PROTOCOL_DESC_KEY = "protocol.description";

    /**
     * <code>JOB_DEF_KEY</code> is the key for job definitions in the
     * user defined configuration file.
     */
    public static final String JOB_DEF_KEY = "job";

    /**
     * <code>MAX_PROTOCOL_NAME_LEN</code> is the maximum length
     * permitted for a protocol name.
     */
    public static final int MAX_PROTOCOL_NAME_LEN = 32;

    /**
     * <code>MAX_PROTOCOL_NAME_LEN</code> is the maximum length
     * permitted for a protocol description.
     */
    public static final int MAX_PROTOCOL_DESC_LEN = 255;

    private ResourceBundle bundle;

    // Map of jobdefs sorted by precedence
    private Map jobDefs;

    /**
     * Creates a new <code>UserConfiguration</code> instance.
     *
     * @param bundle a <code>ResourceBundle</code> derived from the
     * relevant properties file.
     *
     * @exception JobControlException if the file is badly formatted.
     */
    public UserConfiguration(ResourceBundle bundle) throws JobControlException
    {
        this.bundle = bundle;
        jobDefs = initConfiguration();
    }

    /**
     * <code>getProtocolName</code> returns a name which make be used
     * as a <code>Task</code> name.
     *
     * @return a <code>String</code>.
     *
     * @exception JobControlException if an error occurs
     */
    public String getProtocolName() throws JobControlException
    {
        String name = null;

        try
        {
            name = bundle.getString(PROTOCOL_NAME_KEY);
            if (name.length() > MAX_PROTOCOL_NAME_LEN)
            {
                throw new JobControlException("Protocol name '"
                                              + name
                                              + "' exceeds maximum permitted length of "
                                              + MAX_PROTOCOL_NAME_LEN
                                              + " characters");
            }

            for (int i = 0; i < name.length(); i++)
            {
                if ((name.charAt(i) != '_') &&
                    (name.charAt(i) != '-') &&
                    (! Character.isLetterOrDigit(name.charAt(i))))
                {
                    throw new JobControlException("Protocol name '"
                                                  + name
                                                  + "' contained an illegal character '"
                                                  + name.charAt(i)
                                                  + "'");
                }
            }
        }
        catch (MissingResourceException mre)
        {
            throw new JobControlException("Protocol name under key '"
                                          + PROTOCOL_NAME_KEY
                                          + "' missing from user configuration");
        }

        return name;
    }

    /**
     * <code>getProtocolDescription</code> returns a name which make
     * be used as a <code>Task</code> description.
     *
     * @return a <code>String</code>.
     *
     * @exception JobControlException if an error occurs
     */
    public String getProtocolDescription() throws JobControlException
    {
        String desc = null;

        try
        {
            desc = bundle.getString(PROTOCOL_DESC_KEY);
            if (desc.length() > MAX_PROTOCOL_DESC_LEN)
            {
                throw new JobControlException("Protocol description '"
                                              + desc
                                              + "' exceeds maximum permitted length of "
                                              + MAX_PROTOCOL_DESC_LEN
                                              + " characters");
            }
        }
        catch (MissingResourceException mre)
        {
            throw new JobControlException("Protocol description under key '"
                                          + PROTOCOL_DESC_KEY
                                          + "' missing from user configuration");
        }

        return desc;
    }

    /**
     * <code>getJobPrecedenceList</code> returns a <code>List</code>
     * of <code>Integer</code> which indicate the order in which the
     * jobs are to be executed. The root job (last to be executed) has
     * the lowest value. These values are also the indices of the job
     * definitions contained within the configuration object.
     *
     * @return a <code>List</code> of <code>Integer</code>s.
     */
    public List getJobPrecedenceList()
    {
        List precedence = new ArrayList(jobDefs.size());
        precedence.addAll(jobDefs.keySet());

        return precedence;
    }

    /**
     * <code>getExecutableId</code> returns the executable ID for the
     * specified job definition.
     *
     * @param jobDefIndex an <code>int</code> job definition index.
     *
     * @return an <code>int</code> ID.
     */
    public int getExecutableId(int jobDefIndex)
    {
        int id = 0;

        Integer i = new Integer(jobDefIndex);

        if (! jobDefs.containsKey(i))
            throw new IllegalArgumentException("There is no job definition '"
                                               + jobDefIndex
                                               + "' in this user configuration.");

        return ((JobDef) jobDefs.get(i)).getExecutableId();
    }

    /**
     * <code>getMaxRetryCount</code> returns the maximum number of job
     * retries for the specified job definition.
     *
     * @param jobDefIndex an <code>int</code> job definition index.
     *
     * @return an <code>int</code>.
     */
    public int getMaxRetryCount(int jobDefIndex)
    {
        int count = 0;

        Integer i = new Integer(jobDefIndex);

        if (! jobDefs.containsKey(i))
            throw new IllegalArgumentException("There is no job definition '"
                                               + jobDefIndex
                                               + "' in this user configuration.");

        return ((JobDef) jobDefs.get(i)).getMaxRetryCount();
    }

    /**
     * <code>getTemplate</code> returns the job command line template
     * for the specified job definition.
     *
     * @param jobDefIndex an <code>int</code> job definition index.
     *
     * @return a <code>String</code>.
     */
    public String getTemplate(int jobDefIndex)
    {
        String template = "";

        Integer i = new Integer(jobDefIndex);

        if (! jobDefs.containsKey(i))
            throw new IllegalArgumentException("There is no job definition '"
                                               + jobDefIndex
                                               + "' in this user configuration.");

        return ((JobDef) jobDefs.get(i)).getTemplate();
    }

    private Map initConfiguration() throws JobControlException
    {
        Map jobDefs = new TreeMap();

        for (Enumeration keys = bundle.getKeys(); keys.hasMoreElements();)
        {
            String key = (String) keys.nextElement();

            if (key.startsWith(JOB_DEF_KEY))
            {
                String val = null;

                try
                {
                    val = bundle.getString(key);
                }
                catch (MissingResourceException mre)
                {
                    throw new JobControlException("Definition of job '"
                                                  + key
                                                  + "' missing from user configuration");
                }

                int keyDelim = key.indexOf(".");

                if (keyDelim == -1)
                    throw new JobControlException("Definition of job '"
                                                  + key
                                                  + "' contains a syntax error");

                String precedenceKey = key.substring(keyDelim + 1);

                StringTokenizer st = new StringTokenizer(val, ":");
                int tokenCount = st.countTokens();

                if ((tokenCount < 2) || (tokenCount > 3))
                    throw new JobControlException("Definition of job '"
                                                  + key
                                                  + "' contains a syntax error");
                int jobPrecedence = 0;
                int executableId = 0;
                int maxRetryCount = 0;
                StringBuffer template = new StringBuffer("$executable");

                String [] tokens = new String [tokenCount];
                for (int i = 0; i < tokenCount; i++)
                {
                    tokens[i] = st.nextToken().trim();
                }

                try
                {
                    jobPrecedence = Integer.parseInt(precedenceKey);
                    executableId = Integer.parseInt(tokens[0]);
                    maxRetryCount = Integer.parseInt(tokens[1]);
                }
                catch (NumberFormatException nfe)
                {
                    nfe.printStackTrace();

                    throw new JobControlException("Definition of job '"
                                                  + key
                                                  + "' contains a syntax error");
                }

                if (tokenCount == 3)
                {
                    template.append(" ");
                    template.append(tokens[2]);
                }

                jobDefs.put(new Integer(jobPrecedence),
                            new JobDef(executableId, maxRetryCount,
                                       template.substring(0)));
            }
        }

        return jobDefs;
    }

    private static class JobDef
    {
        int execId;
        int maxRetryCount;
        String template;

        JobDef(int execId, int maxRetryCount, String template)
        {
            this.execId = execId;
            this.maxRetryCount = maxRetryCount;
            this.template = template;
        }

        int getExecutableId()
        {
            return execId;
        }

        int getMaxRetryCount()
        {
            return maxRetryCount;
        }

        String getTemplate()
        {
            return template;
        }

        public String toString()
        {
            StringBuffer sb = new StringBuffer("<Job def: ");
            sb.append("execId ");
            sb.append(execId);
            sb.append("; maxRetryCount ");
            sb.append(maxRetryCount);
            sb.append("; template ");
            sb.append(template);
            sb.append(">");

            return sb.substring(0);
        }
    }
}
