/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.xml;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.sql.Timestamp;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;

import uk.ac.sanger.jcon.job.BatchJob;
import uk.ac.sanger.jcon.job.Executable;
import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.Owner;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.util.PostorderEnumeration;
import uk.ac.sanger.jcon.util.PreorderEnumeration;
import uk.ac.sanger.jcon.xml.jaxb.BatchJobDO;
import uk.ac.sanger.jcon.xml.jaxb.CompletionTime;
import uk.ac.sanger.jcon.xml.jaxb.ExecutableDO;
import uk.ac.sanger.jcon.xml.jaxb.StartTime;
import uk.ac.sanger.jcon.xml.jaxb.StatusDO;

/**
 * <code>BatchJobMarshallableImpl</code> is implementation of
 * <code>BatchJob</code> which may be (un)marshalled using JAXB.
 *
 * @author Keith James
 */
public class BatchJobMarshallableImpl extends BatchJobDO implements BatchJob
{
    private MutableTreeNode parent;
    private Object userObject;

    private int low;
    private int high;

    private transient PropertyChangeSupport changeSupport;

    public void setOperator(String operator)
    {
        if (! (operator.equals(Job.OPERATOR_AND) ||
               operator.equals(Job.OPERATOR_OR)))
            throw new IllegalArgumentException("Invalid operator '"
                                               + operator
                                               + "' passed to method. Valid options are: '"
                                               + Job.OPERATOR_AND
                                               + "', '"
                                               + Job.OPERATOR_OR
                                               + "'");
        super.setOperator(operator);
    }

    public Executable getExecutable()
    {
        return (Executable) getExecutableDO();
    }

    public void setExecutable(Executable executable)
    {
        if (executable instanceof ExecutableDO)
            setExecutableDO((ExecutableDO) executable);
        else
            throw new IllegalArgumentException("Potential marshalling error. Only JAXB-generated implementations may be used");
    }

    public Status getStatus()
    {
        return (Status) getStatusDO();
    }

    public void setStatus(Status status)
    {
        if (status instanceof StatusDO)
            setStatusDO((StatusDO) status);
        else
            throw new IllegalArgumentException("Potential marshalling error. Only JAXB-generated implementations may be used");
    }

    public void setMaxRetryCount(int maxRetryCount)
    {
        if (maxRetryCount < 0 || maxRetryCount > Job.MAX_MAX_RETRY_COUNT)
            throw new IllegalArgumentException("Invalid maxRetryCount '"
                                               + maxRetryCount
                                               + "' passed to method. Valid range is "
                                               + "0 >= priority <= "
                                               + Job.MAX_MAX_RETRY_COUNT);
        super.setMaxRetryCount(maxRetryCount);
    }

    public Timestamp getStartTime()
    {
        StartTime st = getStartTimeDO();

        return (st != null) ? st.getContent() : null;
    }

    public void setStartTime(Timestamp startTime)
    {
        StartTime st = getStartTimeDO();

        if (st == null)
            st = new StartTime();

        st.setContent(startTime);
        setStartTimeDO(st);
    }

    public Timestamp getCompletionTime()
    {
        CompletionTime ct = getCompletionTimeDO();

        return (ct != null) ? ct.getContent() : null;
    }

    public void setCompletionTime(Timestamp completionTime)
    {
        CompletionTime ct = getCompletionTimeDO();

        if (ct == null)
            ct = new CompletionTime();

        ct.setContent(completionTime);
        setCompletionTimeDO(ct);
    }

    public Enumeration children()
    {
        List children = getChildren();
        return new IteratorAdapter(children.iterator());
    }

    public boolean getAllowsChildren()
    {
        return true;
    }

    public TreeNode getChildAt(int childIndex)
    {
        List children = getChildren();

        if (children.size() == 0)
            throw new ArrayIndexOutOfBoundsException("Job has no children");
        else
            return (MutableTreeNode) children.get(childIndex);
    }

    public int getChildCount()
    {
       return getChildren().size();
    }

    public int getIndex(TreeNode job)
    {
        if (job == null)
            throw new IllegalArgumentException("Job argument was null");
        else if (! isOurChild((Job) job))
            return -1;
        else
            return getChildren().indexOf(job);
    }

    public TreeNode getParent()
    {
        return parent;
    }

    public void setParent(MutableTreeNode parent)
    {
        this.parent = parent;
    }

    public boolean isLeaf()
    {
        return (getChildCount() == 0);
    }

    public void add(Job newChild)
    {
        if (newChild.getParent() == this)
            throw new IllegalArgumentException("Failed to added child Job as it is already a child of this Job");
        else
            insert(newChild, getChildCount());
    }

    public void insert(MutableTreeNode newChild, int childIndex)
    {
        if (! (newChild instanceof BatchJobMarshallableImpl))
            throw new IllegalArgumentException("Potential marshalling error. Only JAXB-generated implementations may be used");
        else if (newChild == null)
            throw new IllegalArgumentException("Failed to add child Job as it was null");
        else if (isOurAncestor(newChild))
            throw new IllegalArgumentException("Failed to add child Job as it is an ancestor");

        MutableTreeNode oldParent = (MutableTreeNode) newChild.getParent();

        if (oldParent != null)
            oldParent.remove(newChild);

        newChild.setParent(this);

        getChildren().add(childIndex, newChild);
    }

    public void remove(int childIndex)
    {
        MutableTreeNode child = (MutableTreeNode) getChildAt(childIndex);

        getChildren().remove(childIndex);
        child.setParent(null);
    }

    public void remove(MutableTreeNode child)
    {
        if (child == null)
            throw new IllegalArgumentException("Failed to remove child Job as it was null");
        else if (! isOurChild(child))
            throw new IllegalArgumentException("Failed to remove child as it the Job is not its parent");
        else
            remove(getIndex(child));
    }

    public void removeFromParent()
    {
        MutableTreeNode parent = (MutableTreeNode) getParent();
        if (parent != null)
            parent.remove(this);
    }

    public Object getUserObject()
    {
        return userObject;
    }

    public void setUserObject(Object userObject)
    {
        this.userObject = userObject;
    }

    public Enumeration preorderEnumeration()
    {
        return new PreorderEnumeration(this);
    }

    public Enumeration postorderEnumeration()
    {
        return new PostorderEnumeration(this);
    }

    public int getLow()
    {
        return low;
    }

    public void setLow(int low)
    {
        this.low = low;
    }

    public int getHigh()
    {
        return high;
    }

    public void setHigh(int high)
    {
        this.high = high;
    }

    /**
     * <code>addPropertyChangeListener</code> adds a
     * <code>PropertyChangeListener</code> to the listener list. The
     * listener is registered for all properties.
     *
     * @param listener a <code>PropertyChangeListener</code>.
     */
    public void addPropertyChangeListener(PropertyChangeListener listener)
    {
        PropertyChangeSupport pcs = getPropertyChangeSupport();

        synchronized (pcs)
        {
            pcs.addPropertyChangeListener(listener);
        }
    }

    /**
     * <code>addPropertyChangeListener</code> adds a
     * <code>PropertyChangeListener</code> for a specific
     * property. The listener will be invoked only when a call on
     * firePropertyChange names that specific property.
     *
     * @param propertyName a <code>String</code>.
     *
     * @param listener a <code>PropertyChangeListener</code>.
     */
    public void addPropertyChangeListener(String                 propertyName,
                                          PropertyChangeListener listener)
    {
        PropertyChangeSupport pcs = getPropertyChangeSupport();

        synchronized (pcs)
        {
            pcs.addPropertyChangeListener(propertyName, listener);
        }
    }

    /**
     * <code>removePropertyChangeListener</code> removes a
     * <code>PropertyChangeListener</code> from the listener
     * list. This removes a <code>PropertyChangeListener</code> that
     * was registered for all properties.
     *
     * @param listener a <code>PropertyChangeListener</code>.
     */
    public void removePropertyChangeListener(PropertyChangeListener listener)
    {
        PropertyChangeSupport pcs = getPropertyChangeSupport();

        synchronized (pcs)
        {
            pcs.removePropertyChangeListener(listener);
        }
    }

    /**
     * <code>removePropertyChangeListener</code> removes a
     * <code>PropertyChangeListener</code> for a specific property.
     *
     * @param propertyName a <code>String</code>.
     *
     * @param listener a <code>PropertyChangeListener</code>.
     */
    public void removePropertyChangeListener(String                 propertyName,
                                             PropertyChangeListener listener)
    {
        PropertyChangeSupport pcs = getPropertyChangeSupport();

        synchronized (pcs)
        {
            pcs.removePropertyChangeListener(propertyName, listener);
        }
    }

    /**
     * <code>hasListeners</code> checks if there are any listeners.
     *
     * @return a <code>boolean</code>.
     */
    public boolean hasListeners()
    {
        return changeSupport != null;
    }

    /**
     * <code>hasListeners</code> checks if there are any listeners for
     * a specific property.
     *
     * @param propertyName a <code>String</code>.
     *
     * @return a <code>boolean</code>.
     */
    public boolean hasListeners(String propertyName)
    {
        return (changeSupport != null &&
                getPropertyChangeSupport().hasListeners(propertyName));
    }

    public String toString()
    {
        return "BatchJobMarshallableImpl: " + super.toString();
    }

    /**
     * <code>getPropertyChangeSupport</code>.
     *
     * @return a <code>PropertyChangeSupport</code> value.
     */
    protected PropertyChangeSupport getPropertyChangeSupport()
    {
        if (changeSupport == null)
            changeSupport = new PropertyChangeSupport(this);

        return changeSupport;
    }

    protected boolean isOurChild(TreeNode child)
    {
        if (child == null)
            return false;
        else if (getChildCount() == 0)
            return false;
        else
            return (child.getParent() == this);
    }

    protected boolean isOurAncestor(TreeNode job)
    {
        if (job == null)
            return false;

        TreeNode ancestor = this;

        while (ancestor != null)
        {
            ancestor = ancestor.getParent();
            if (ancestor == job)
                return true;
        }

        return false;
    }

    protected void writeObject(ObjectOutputStream stream) throws IOException
    {
        stream.defaultWriteObject();

        Object [] objects;

        if (userObject != null && userObject instanceof Serializable)
            objects = new Object [] { "userObject" , userObject };
        else
            objects = new Object[0];

        stream.writeObject(objects);
    }

    protected void readObject(ObjectInputStream stream) 
        throws IOException, ClassNotFoundException
    {
        stream.defaultReadObject();

        Object [] objects;

        objects = (Object []) stream.readObject();

        if (objects.length > 0 && objects[0].equals("userObject"))
            userObject = objects[1];
    }

    private static class IteratorAdapter implements Enumeration
    {
        private Iterator i;

        IteratorAdapter(Iterator i)
        {
            this.i = i;
        }

        public boolean hasMoreElements()
        {
            return i.hasNext();
        }

        public Object nextElement()
        {
            return i.next();
        }
    }
}
