/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.xml;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Iterator;
import java.util.List;
import javax.xml.bind.Dispatcher;

import org.apache.log4j.Category;

import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.Task;
import uk.ac.sanger.jcon.xml.jaxb.*;

/**
 * <p><code>TaskMarshaller</code> provides methods for marshalling and
 * unmarshalling lists of <code>Task</code>s to and from XML
 * conforming to the JobControl DTD. The objects created by
 * unmarshalling are derived from JAXB-generated classes.</p>
 *
 * <p>This class performs the necessary validation on marshalling and
 * will close the streams before exiting the method.</p>
 *
 * <p>To be successfully marshalled all objects must be derived from
 * JAXB-generated classes. There are marshallable implementations of
 * all necessary interfaces in this package.</p>
 *
 * @author Keith James
 */
public class TaskMarshaller
{
    /**
     * <code>marshal</code> marshals <code>Task</code>s to XML.
     *
     * @param tasks a <code>List</code> of <code>Task</code>s.
     * @param os an <code>OutputStream</code>.
     *
     * @exception Exception if an error occurs
     */
    public void marshal(List tasks, OutputStream os) throws Exception
    {
        TaskDOCollection tCol = new TaskDOCollection();
        tCol.getTask().addAll(tasks);
        tCol.validate();

        try
        {
            tCol.marshal(os);
        }
        finally
        {
            os.close();
        }
    }

    /**
     * <code>unmarshal</code> unmarshals <code>Task</code>s from XML.
     *
     * @param is an <code>InputStream</code>.
     *
     * @return a <code>List</code> of <code>Task</code>s.
     *
     * @exception Exception if an error occurs
     */
    public List unmarshal(InputStream is) throws Exception
    {
        TaskDOCollection tCol = new TaskDOCollection();
        Dispatcher d = tCol.newDispatcher();
        registerSubclasses(d);

        try
        {
            tCol = (TaskDOCollection) d.unmarshal(is);
        }
        finally
        {
            is.close();
        }

        ParentLinker linker = new ParentLinker();
        for (Iterator ti = tCol.getTask().iterator(); ti.hasNext();)
        {
            Task t = (Task) ti.next();
            Job [] jobs = t.getJobs();

            for (int i = 0; i < jobs.length; i++)
                linker.linkParents(jobs[i]);
        }

        return tCol.getTask();
    }

    private void registerSubclasses(Dispatcher d)
    {
        d.register(TaskDO.class, TaskMarshallableImpl.class);
        d.register(JobDO.class, JobMarshallableImpl.class);
        d.register(BatchJobDO.class, BatchJobMarshallableImpl.class);
        d.register(ExecutableDO.class, ExecutableMarshallableImpl.class);
        d.register(OwnerDO.class, OwnerMarshallableImpl.class);
        d.register(StatusDO.class, StatusMarshallableImpl.class);
    }
}
