/*
 * Copyright (c) 2002 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published 
 * by  the Free Software Foundation; either version 2 of the License or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.LIB.  If not, write to 
 * the Free Software Foundation Inc., 59 Temple Place - Suite 330, 
 * Boston, MA  02111-1307 USA
 */

package uk.ac.sanger.jcon.dao;

import java.sql.Connection;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.Iterator;

import uk.ac.sanger.jcon.JobControlSQLCase;
import uk.ac.sanger.jcon.job.Executable;
import uk.ac.sanger.jcon.job.Job;
import uk.ac.sanger.jcon.job.JobDefaultImpl;
import uk.ac.sanger.jcon.job.Owner;
import uk.ac.sanger.jcon.job.Status;
import uk.ac.sanger.jcon.job.Task;

/**
 * <code>JobSQLDAOTest</code> tests create, read, update and delete of
 * <code>Job</code>s.
 *
 * @author Keith James
 */
public class JobSQLDAOTest extends JobControlSQLCase
{
    public JobSQLDAOTest(String name) throws Exception
    {
        super(name);
    }

    public void testReadJobById() throws Exception
    {
        JobDAO jobDAO = new JobSQLDAO(datasource);

        Job job = jobDAO.readJobById(1);
        assertEquals(1, job.getId());
        assertEquals(0, job.getLow());
        assertEquals(23, job.getHigh());
        assertEquals(Job.OPERATOR_AND, job.getOperator());
        assertEquals(1, job.getExecutable().getId());
        assertEquals(Status.WAITING, job.getStatus().getId());
        assertEquals(0, job.getMaxRetryCount());
        assertEquals(0, job.getRetryCount());
    }

    public void testReadJobsByStatusAndTaskId() throws Exception
    {
        JobDAO jobDAO = new JobSQLDAO(datasource);
        StatusDAO statusDAO = new StatusSQLDAO(datasource);
        Status ready = statusDAO.readStatusById(Status.READY);

        Collection jobs = null;
        int count = 0;

        jobs = jobDAO.readJobsByStatusAndTaskId(ready, 1);

        for (Iterator ji = jobs.iterator(); ji.hasNext();)
        {
            Job job = (Job) ji.next();

            // 2, 3 are ready in Task 1
            if (job.getId() == 2 || job.getId() == 3)
                count++;
        }

        assertEquals(2, count);

        count = 0;

        jobs = jobDAO.readJobsByStatusAndTaskId(ready, 2);
        for (Iterator ji = jobs.iterator(); ji.hasNext();)
        {
            Job job = (Job) ji.next();

            // 14, 19, 23 are ready in Task 2
            if (job.getId() == 14 || job.getId() == 19 || job.getId() == 23)
                count++;
        }

        assertEquals(3, count);
    }

    public void testUpdateJob() throws Exception
    {
        JobDAO jobDAO = new JobSQLDAO(datasource);
        StatusDAO statusDAO = new StatusSQLDAO(datasource);
        Status failed = statusDAO.readStatusById(Status.FAILED);

        // Get Job 1
        Job before = jobDAO.readJobById(1);
        assertEquals(Status.WAITING, before.getStatus().getId());
        assertEquals(0, before.getRetryCount());

        // Change Job 1
        before.setStatus(failed);
        before.setRetryCount(3);
        jobDAO.updateJob(before);

        // Re-read Job 1
        Job after = jobDAO.readJobById(1);
        assertEquals(Status.FAILED, after.getStatus().getId());
        assertEquals(3, before.getRetryCount());
    }

    public void testUpdateJobReady() throws Exception
    {
        JobDAO jobDAO = new JobSQLDAO(datasource);

        // Get Job 1
        Job before = jobDAO.readJobById(1);
        assertEquals(Status.WAITING, before.getStatus().getId());

        // Change Job 1
        jobDAO.updateJobReady(before);

        // Re-read Job 1
        Job after = jobDAO.readJobById(1);
        assertEquals(Status.READY, after.getStatus().getId());
    }

    public void testUpdateJobStarted() throws Exception
    {
        JobDAO jobDAO = new JobSQLDAO(datasource);

        // Get Job 2
        Job before = jobDAO.readJobById(2);
        assertEquals(Status.READY, before.getStatus().getId());
        assertNull(before.getStartTime());

        // Set Job 2 STARTED
        jobDAO.updateJobStarted(before);

        // Re-read Job 2
        Job after = jobDAO.readJobById(2);
        assertEquals(Status.STARTED, after.getStatus().getId());
        assertNotNull(after.getStartTime());
        assertTrue(after.getStartTime() instanceof Timestamp);
    }

    public void testUpdateJobCompleted() throws Exception
    {
        JobDAO jobDAO = new JobSQLDAO(datasource);

        //
        // Testing completion within a Job tree
        //

        // Get Job 10
        Job before = jobDAO.readJobById(10);
        assertEquals(Status.STARTED, before.getStatus().getId());

        // Check its parent is WAITING
        Job parent = jobDAO.readJobById(9);
        assertEquals(Status.WAITING, parent.getStatus().getId());

        // Now we can try for COMPLETION
        jobDAO.updateJobCompleted(before);

        // Re-read Job 10
        Job after = jobDAO.readJobById(10);
        assertEquals(Status.COMPLETED, after.getStatus().getId());

        // Check that its parent, which was WAITING, is now READY
        parent = jobDAO.readJobById(9);
        assertEquals(Status.READY, parent.getStatus().getId());

        //
        // Testing completion between the root of one Job tree and the
        // leaves of the next
        //

        // Get Job 35
        Job goal = jobDAO.readJobById(35);
        assertEquals(Status.WAITING, goal.getStatus().getId());

        // Get Job 38
        before = jobDAO.readJobById(38);
        assertEquals(Status.STARTED, before.getStatus().getId());

        // Check the leaf Jobs in the next Job tree are WAITING
        Job leaf1 = jobDAO.readJobById(37);
        Job leaf2 = jobDAO.readJobById(36);
        assertEquals(Status.WAITING, leaf1.getStatus().getId());
        assertEquals(Status.WAITING, leaf2.getStatus().getId());

        // Now we can try for COMPLETION
        jobDAO.updateJobCompleted(before);

        // Re-read Job 38
        after = jobDAO.readJobById(38);
        assertEquals(Status.COMPLETED, after.getStatus().getId());

        // Check the leaf Jobs in the next Job tree, which were
        // WAITING, are now READY
        leaf1 = jobDAO.readJobById(37);
        leaf2 = jobDAO.readJobById(36);
        assertEquals(Status.READY, leaf1.getStatus().getId());
        assertEquals(Status.READY, leaf2.getStatus().getId());

        // Re-read Job 35, which should not have changed
        goal = jobDAO.readJobById(35);
        assertEquals(Status.WAITING, goal.getStatus().getId());
    }

    public void testUpdateJobRetry() throws Exception
    {
        JobDAO jobDAO = new JobSQLDAO(datasource);

        // Get Job 10
        Job before = jobDAO.readJobById(10);
        assertEquals(Status.STARTED, before.getStatus().getId());
        assertEquals(0, before.getRetryCount());

        // Now we can retry
        jobDAO.updateJobRetry(before);

        Job after = jobDAO.readJobById(10);
        assertEquals(Status.READY, after.getStatus().getId());
        assertEquals(1, after.getRetryCount());
    }

    public void testUpdateJobFailed() throws Exception
    {
        JobDAO jobDAO = new JobSQLDAO(datasource);

        // Get Job 10
        Job before = jobDAO.readJobById(10);
        assertEquals(Status.STARTED, before.getStatus().getId());

        // Max retry count is 1, so first failure becomes a retry
        jobDAO.updateJobFailed(before);

        Job inter = jobDAO.readJobById(10);
        assertEquals(Status.READY, inter.getStatus().getId());
        assertEquals(1, inter.getRetryCount());

        // Start the Job again (not strictly necessary)
        jobDAO.updateJobStarted(inter);
        inter = jobDAO.readJobById(10);
        assertEquals(Status.STARTED, inter.getStatus().getId());

        // This time we have reached max retry count so failure occurs
        jobDAO.updateJobFailed(inter);

        Job after = jobDAO.readJobById(10);
        assertEquals(Status.FAILED, after.getStatus().getId());
    }

    public void testUpdateJobFailedCascade() throws Exception
    {
        JobDAO jobDAO = new JobSQLDAO(datasource);

        // Get Job 10
        Job before = jobDAO.readJobById(10);
        assertEquals(Status.STARTED, before.getStatus().getId());

        // Max retry count is 1, so first failure becomes a retry
        jobDAO.updateJobFailed(before);

        Job inter = jobDAO.readJobById(10);
        assertEquals(Status.READY, inter.getStatus().getId());
        assertEquals(1, inter.getRetryCount());

        // Start the Job again (not strictly necessary)
        jobDAO.updateJobStarted(inter);
        inter = jobDAO.readJobById(10);
        assertEquals(Status.STARTED, inter.getStatus().getId());

        // This time we have reached max retry count so failure occurs
        // and we try to cascade upwards
        jobDAO.updateJobFailedCascade(inter);

        Job after = jobDAO.readJobById(10);
        assertEquals(Status.FAILED, after.getStatus().getId());

        // Follow cascade; all should have been cancelled
        after = jobDAO.readJobById(9);
        assertEquals(Status.CANCELLED, after.getStatus().getId());
        after = jobDAO.readJobById(1);
        assertEquals(Status.CANCELLED, after.getStatus().getId());
    }

    public void testSkipRedundantJobs() throws Exception
    {
        JobDAO jobDAO = new JobSQLDAO(datasource);

        Job job = jobDAO.readJobById(14);
        assertEquals(Status.READY, job.getStatus().getId());
        jobDAO.updateJobStarted(job);

        job = jobDAO.readJobById(14);
        jobDAO.updateJobCompleted(job);

        job = jobDAO.readJobById(13);
        assertEquals(Status.READY, job.getStatus().getId());
        jobDAO.updateJobStarted(job);

        job = jobDAO.readJobById(13);
        jobDAO.updateJobCompleted(job);

        // Jobs 18, 19, 21, 22, 23, 29, 30 and 31 should be redundant
        // and marked as SKIPPED
        int [] ids = new int [] { 18, 19, 21, 22, 23, 29, 30, 31 };

        for (int i = 0; i < ids.length; i++)
        {
            Job j = jobDAO.readJobById(ids[i]);
            assertEquals(Status.SKIPPED, j.getStatus().getId());
        }
    }
}
