/*
 * Copyright (c) 2001 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package uk.ac.sanger.psu.tigread;

import uk.ac.sanger.psu.tigrutils.*;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Collection;
import java.util.List;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.NoSuchElementException;
import java.util.Comparator;
import java.util.TreeMap;

import org.jdom.Attribute;
import org.jdom.Element;

import org.biojava.bio.Annotation;
import org.biojava.bio.SimpleAnnotation;
import org.biojava.bio.BioException;
import org.biojava.utils.ChangeVetoException;

import org.biojava.bio.Annotation;
import org.biojava.bio.SimpleAnnotation;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.Feature;
import org.biojava.bio.seq.StrandedFeature;
import org.biojava.bio.seq.impl.SimpleStrandedFeature;

import org.biojava.bio.symbol.Location;
import org.apache.regexp.RE; // Used in /color assingment
import org.apache.commons.lang.StringUtils;


/**
 * <code>EmblAnnotationFactory</code> produces various types of
 * <code>Annotation</code> objects from TIGR XML JDOM
 * <code>Elements</code>. The mapping of XML element names to EMBL
 * qualifiers is configurable through the annotation_map.properties
 * file. If this file contains a mapping from one of the recognised
 * TIGR element names to another string, that string will be used as
 * the EMBL qualifier key. Otherwise the TIGR element name will be
 * used as the qualifier key. Note that ROLE_INFO is a special case
 * where the qualifier keys contain extra encoded information. This is
 * unfortunate, but necessary, in mapping from a tree to a list and
 * back again.
 *
 *
 * @author Keith James
 */
class EmblAnnotationFactory
{
    static final String     PROPERTY_MAP_FILE = "annotation_map.properties";
    static final Properties annotationMap     = new Properties();
    private EmblLocationFactory  locFactory;
    private EmblCustomQualifiers customCode;

    EmblAnnotationFactory()
    {
        try
        {
            annotationMap.load(new FileInputStream(PROPERTY_MAP_FILE));
            // EmblLocationFactory needed to coords in /domain coords
            locFactory       = new EmblLocationFactory();
            customCode = new EmblCustomQualifiers(annotationMap);
        }
        catch (FileNotFoundException fnfe)
        {
            System.err.println("Unable to find annotation properties file: "
                               + fnfe.getMessage());
        }
        catch (IOException ioe)
        {
            System.err.println("Unable to read annotation properties file: "
                               + ioe.getMessage());
        }
    }

    /**
     * <code>makeAnnotation</code> returns the correct style of
     * annotation for the element.
     *
     * @param element an <code>Element</code>.
     * @param identifier an <code>int</code> value identifying which
     * type of annotation to make.
     *
     * @return an <code>Annotation</code>.
     */
    Annotation makeAnnotation(Element element,
                              int     identifier)
    {
        switch (identifier)
        {
            case TigrSequenceLoader.TU:
                return makeTUAnnotation(element);

            case TigrSequenceLoader.MODEL:
                return makeModelAnnotation(element);

            case TigrSequenceLoader.EXON:
                return makeExonAnnotation(element);

            case TigrSequenceLoader.RNA:
                return makeRNAAnnotation(element);

            case TigrSequenceLoader.MISC:
                return makeMiscAnnotation(element);

            case TigrSequenceLoader.REPEAT:
                return makeRepeatAnnotation(element);

            case TigrSequenceLoader.SOURCE:
                return makeSourceAnnotation(element);

            default:
                System.err.println("Requested unknown Annotation type '"
                                   + identifier
                                   + "' from EmblAnnotationFactory");
                return new SimpleAnnotation();
        }
    }

    /**
     * <code>makeAnnotation</code> returns the correct style of
     * annotation for the element.
     *
     * @param element an <code>Element</code>.
     * @param feat_name a <code>String</code> that is the /gene, so annation
     * can reference gene this features "belongs to"
     * @param identifier an <code>int</code> value identifying which
     * type of annotation to make.
     *
     * @return an <code>Annotation</code>.
     */
    Annotation makeAnnotation(Element element,
                              String  feat_name,
                              int     identifier)
    {
        switch (identifier)
        {
           case TigrSequenceLoader.SIGPTMHMM:
                return makeSigPTMHMMAnnotation(element, feat_name);

            default:
                System.err.println("Requested unknown Annotation type '"
                                   + identifier
                                   + "' from EmblAnnotationFactory");
                return new SimpleAnnotation();
        }
    }

    /**
     * <code>makeAnnotation</code> returns the correct style of
     * annotation for the element. This methid version takes a Location
     * object as an extrar arg as TMHMM and signalP methods need them
     *
     * @param element an <code>Element</code>.
     * @param identifier an <code>int</code> value identifying which
     * type of annotation to make.
     *
     * @return an <code>Annotation</code>.
     */
    Annotation makeAnnotation(Element  element,
                              int      identifier,
                              Location aaLocs,
                              String   featName)
        throws Exception
    {
       switch (identifier)
       {
           case TigrSequenceLoader.TMM:
               return makeTMHMMAnnotation(element, aaLocs, featName);

           default:
               System.err.println("Requested unknown Annotation type '"
                                  + identifier
                                  + "' from EmblAnnotationFactory");
               
               return new SimpleAnnotation();
        }
    }

    /**
     * <code>makeTUAnnotation</code> creates annotation from a TU
     * element. ROLE_INFO information is encoded into the qualifer
     * keys:
     *
     * <ul>
     * <li>_0_COMPARTMENT</li>
     * <li>_0_MAIN_ROLE</li>
     * <li>_0_SUBROLE_1</li>
     * <li>_1_COMPARTMENT</li>
     * <li>_1_MAIN_ROLE</li>
     * <li>_1_SUBROLE_1</li>
     * </ul>
     *
     * @param tu an <code>Element</code>.
     *
     * @return an <code>Annotation</code>.
     */
    private Annotation makeTUAnnotation(Element tu)
    {
        Annotation anno = new SimpleAnnotation();

        // Elements in the TU element
        String [] tuElem = new String [] { }; //"CHROMO_LINK" };

        // Elements in the GENE_INFO element
        String [] giElem = new String [] { //"LOCUS",
                                           "COM_NAME",
                                           "PUB_LOCUS",
                                           "PUB_COMMENT",
                                           "EC_NUM",
                                           "GENE_SYM" };

        // Elements in the GENE_INFO element
        String [] mElem = new String [] { }; //"FEAT_NAME" };

        // Elements in the ROLE_INFO element
        // NOTE: KEITH SAYS THIS IS NOT USED, THOUGH STILL IN DTD!!!
        String [] riElem = new String [] { "COMPARTMENT",
                                           "MAIN_ROLE",
                                           "SUBROLE_1",
                                           "SUBROLE_2",
                                           "SUBROLE_3",
                                           "SUBROLE_4" };

        Element geneInfo = tu.getChild("GENE_INFO");

        try
       {
            setColor(geneInfo, anno);
            setGO(geneInfo, anno);

            if (geneInfo.getChild("ROLE_LIST") != null)
            {
                List roleInfos = geneInfo.getChild("ROLE_LIST")
                    .getChildren("ROLE_INFO");

                List genKeys = new ArrayList();

                Object ri [] = roleInfos.toArray();

                for (int i = 0; i < ri.length; i++)
                {
                    Element r = (Element) ri[i];

                    // Get the existing Elements
                    List children = r.getChildren();
                    r.removeChildren();

                    String id = "_" + Integer.toString(i) + "_";

                    // Add generated Element names
                    for (int j = 0; j < riElem.length; j++)
                        genKeys.add(id + riElem[j]);

                    for (Iterator ci = children.iterator(); ci.hasNext();)
                    {
                        // Replace the Elements which were removed
                        Element child = (Element) ci.next();
                        Element e = new Element(id + child.getName());
                        e.addContent(child.getText());

                        r.addContent(e);
                    }

                    String [] genElem = (String []) genKeys.toArray(new String[0]);

                    // Add properties from GENE_INFO's ROLE_INFO
                    populateAnnotation(r, genElem, anno);
                }
            }

            // Add properties from TU itself
            populateAnnotation(tu, tuElem, anno);

            // Add properties from TU's GENE_INFO
            populateAnnotation(geneInfo, giElem, anno);

            // Pseudogenes are a special case
            if (geneInfo.getChild("IS_PSEUDOGENE").getText().equals("1"))
            {
                anno.setProperty("pseudo", new Boolean(true));
            }

            // Evidence is a special case
            if (geneInfo.getChild("EVIDENCE") != null)
            {
                List evidenceCodes = geneInfo.getChild("EVIDENCE").getAttributes();
                for (Iterator ei = evidenceCodes.iterator(); ei.hasNext();)
                {
                    Attribute e = (Attribute) ei.next();
                    anno.setProperty(e.getName(), e.getValue());
                }
            }

            // Assume ONE model only!
            Element   model        = tu.getChild("MODEL");

            // Add properties from MODEL itself
            populateAnnotation(model, mElem, anno);

            // get Molecular Weight, Isoelectric point, OMP count
            populateAnnotationWithPepStats(model, anno);

            // populate /psu_db_xref with link to TIGR
            customCode.addCustomQualifiers(tu, anno);
            //customCode.addCustomQualifiers(model, anno);

            ArrayList similarities = new ArrayList();
            ArrayList domains      = new ArrayList();
            ArrayList features     = new ArrayList();

            // get /domain stuff
            if( model.getChild("MODEL_EVIDENCE") != null ){
                Element evidence_type =
                    model.getChild("MODEL_EVIDENCE").getChild("EVIDENCE_TYPE");
                
                List sequence_db_matches =
                    evidence_type.getChildren("SEQUENCE_DB_MATCH");

                for(int x = 0; x < sequence_db_matches.size(); ++x){
                    List search_dbs =
                        ((Element)sequence_db_matches.get(x)).getChildren();

                    for(int y = 0; y < search_dbs.size(); ++y){
                        Element search_db = (Element) search_dbs.get(y);

                        String db_name = search_db.getAttributeValue("DB_NAME");

                        if( "Interpro".equals(db_name)     ||
                            "Pfam,TIGRFAM".equals(db_name) ||
                            "PROSITE".equals(db_name)        ){

                            if( "Interpro".equals(db_name) )
                                db_name = "InterPro"; // Use correct case

                            domains.addAll(getDomainDetails(search_db,
                                                            db_name));
                        }
                        else {
                            similarities.addAll(getSimilarityDetails(search_db,
                                                                     db_name));
                        }
                    }
                }
            }

            // Add domains and maybe psu_db_xref if Interpro
            // HACK: all Interpros are "removed" from the List
            //       cos they have no coords - 
            //
            ArrayList  interpros = new ArrayList();
            int      domainsSize = domains.size();
            ArrayList newDomains = new ArrayList();

            for(int x = 0; x < domainsSize; ++x){
                String s = ((StringBuffer) domains.get(x)).toString();

                if( s.startsWith("InterPro:") ){
                    // get all chars upto 2nd ';'
                    int  first = s.indexOf(";");
                    int second = s.indexOf(";", first + 1);
                    interpros.add(s.substring(0, second));
                }
                else {
                    newDomains.add(s);
                }
            }

            //addToProperty(anno, "domain", domains);
            addToProperty(anno, "domain", newDomains);
            addToProperty(anno, "psu_db_xref", interpros); // This is PSU ONLY!
            addToProperty(anno, "similarity", similarities);
       }
        catch (ChangeVetoException cve)
        {
            cve.printStackTrace();
        }

        return anno;
    }

    /* <code>getDomianDetails</code> builds up a <code>List</code> of
     * <code>StringBuffer</code> elements suitable for
     * one or more /domain qualifiers.
     *
     * NOTE: TIGR merge Pfam and TIGRFAM hits into one, ACCESSION being
     * the only way to tell the difference.
     *
     * @param search_db an <code>Element</code> representing a SEARCH_DB
     * @param dbName a <code>String</code> to start the domain 
     * StringBuffer eg 'Pfam'. A ':' wil be added to build up something l
     * ike 'Pfam:PF00179;Ubiquitin-conjugating enzyme'
     *
     * @return a <code>List></code> (of <code>StringBuffer</code> elements)
     */
    protected List getDomainDetails(Element search_db, String dbName){
        List      seq_elements  = search_db.getChildren("SEQ_ELEMENT");
        ArrayList domains       = new ArrayList();

        for(int x = 0; x < seq_elements.size(); ++x){
            StringBuffer  domain = new StringBuffer();

            Element  seq_element = (Element) seq_elements.get(x);
            String     accession = seq_element.getAttributeValue("ACCESSION");
            String   description = seq_element.getAttributeValue("DESCRIPTION");

            // We don't want "null" being appended to StringBuffer
            if (description == null) {
                description = "";
            }

            // Score
            String score = seq_element.getAttributeValue("SCORE");
            if( score == null ){
                score = "";
            }

            // Coords
            Element eCoords = seq_element.getChild("RELATIVE_COORDS");
            String  sCoords = "";

            if( eCoords != null ){
                // Only ever one RELATIVE_COORDS->COORDSET since one pfam
                // hit that has multiple codon hits had multiple SEARCH_DB
                // Elements
                Location l =
                    locFactory.makeLocation(eCoords.getChild("COORDSET"));
                sCoords    = "codon "+l.getMin()+"-"+l.getMax();
            }

            String dbNameAccession = 
                getDBNameFromAccession(dbName,
                                       accession);
                                                            
            // Fianlly, build the string that will be seen in the /domain
            domain.append(dbNameAccession);
            domain.append(":");
            domain.append(accession);
            domain.append("; ");
            domain.append(description);
            domain.append("; ");
            domain.append(score);
            domain.append("; ");
            domain.append(sCoords);

            domains.add(domain);
        }

        return domains;
    }

    /* <code>getDBNameFromAccession</code> returns a valid database name
     * for a given dbName or accession. This method exists because
     * TIGR merge Pfam and TIGRFAM hits into one, ACCESSION being
     * the only way to tell the difference. If passed in Interpro or Prosite 
     * it will just return that string straight back.
     *
     * @param dbName a <code>String</code> representing a DB 
     * (Interpro, Pfam,TIGRFAM)
     * @param accession a <code>String</code> holding the DB ID (PF00448,
     * TIGR00231)
     *
     * @return a <code>String</code> with the correct DB name
     */
    protected String getDBNameFromAccession(String dbName, String accession){
        if( ! "Pfam,TIGRFAM".equals(dbName) ){
            return dbName;
        }

        if( accession.startsWith("PF") ){
            return "Pfam";
        }
        else if( accession.startsWith("TIGR") ){
            return "TIGRFAM";
        }
        else {
            System.err.println("ERROR: unknown accession type for "+
                               "Pfam,TIGRFAM domain");
            throw new RuntimeException("ERROR: unknown accession type for Pfam,TIGRFAM domain");
        }
    }

    /* <code></code>
     *
     */
    /*
    class EvalueComparator implements Comparator {
        public int compare(Object o1, Object o2) {
            String e1 = (String) o1;
            String e2 = (String) o2;

            Float  f1 = null, f2 = null;

            try {
                f1 = new Float(e1);
                f2 = new Float(e2);
            }
            catch (Exception e){
                System.err.println("e1 = "+e1+", e2 = "+e2);
                e.printStackTrace();
            }

            return f1.compareTo(f2);
        }                                    

        public boolean equals(Object o) {
            return compare(this, o)==0;
        }
    } 
    */

    /* <code>getSimilarityDetails</code> builds up a <code>List</code> of
     * <code>StringBuffer</code> elements suitable for
     * one or more /similarity qualifiers.
     *
     * @param search_db an <code>Element</code> representing a SEARCH_DB
     * @param dbName a <code>String</code> to start the domain 
     * StringBuffer eg 'Pfam'. A ':' wil be added to build up something
     * like 'Pfam:PF00179;Ubiquitin-conjugating enzyme'
     *
     * @return a <code>List></code> (of <code>StringBuffer</code> elements)
     */
    protected List getSimilarityDetails(Element search_db, String dbName){
        List            seq_elements  = search_db.getChildren("SEQ_ELEMENT");
        ArrayList       similarities  = new ArrayList();
        //SortByEValueList sortedList   = new SortByEValueList();
        //EvalueComparator evc = new EvalueComparator();
        //TreeMap   sortedList = new TreeMap(evc);

        for(int x = 0; x < seq_elements.size(); ++x){
            Element seq_element = (Element) seq_elements.get(x);
            String  method      = seq_element.getAttributeValue("METHOD");

            // Has to be one of these detection methods
            if(   "BER".equalsIgnoreCase(method)
               || "BLASTP".equalsIgnoreCase(method)
               || "AUTO_BLASTP".equalsIgnoreCase(method)){
            
                StringBuffer simi = new StringBuffer();
                String       desc = seq_element.getAttributeValue("DESCRIPTION");
                String        org = "";

                if( desc == null ){
                    desc = "";
                }
                else {
                    desc = backSlashSemis(desc);
                    org  = parseOrganismFromDesc(desc);
                }

                simi.append(seq_element.getAttributeValue("METHOD"));
                simi.append("; ");

                // ACCESSIONS take the form GP|1234, convert to GP:1234
                // Also, tigr have can have 3 ids and 2 dbs...
                // Only have first 'DB:id_num'
                String   acc = seq_element.getAttributeValue("ACCESSION");
                int barIndex = acc.indexOf('|');

                if (barIndex != -1) {
                    acc = acc.substring(0, barIndex) + ":" +
                        acc.substring(barIndex + 1);
                }

                barIndex = acc.indexOf('|');

                if (barIndex != -1) {
                    acc = acc.substring(0, barIndex);
                }
                

                simi.append(acc);
                simi.append("; ");

                simi.append(org);
                simi.append("; ");
                simi.append(desc);

                // Miss out gene name, length of subject
                simi.append("; ; ; ");

                // Percentage ID
                simi.append("id=");
                simi.append(seq_element.getAttributeValue("PER_ID"));
                simi.append("%");

                // Ungapped percentage ID
                simi.append("; ; ");

                // e value                
                simi.append("E()=");
                simi.append(seq_element.getAttributeValue("E_VALUE"));
                simi.append("; ");

                // score
                String score = null;

                if( seq_element.getAttributeValue("SCORE") != null ){
                    simi.append("score=");
                    score = seq_element.getAttributeValue("SCORE");
                    simi.append(score);
                }

 
                simi.append("; ; ; ");

                //sortedList.put(score, simi); // Sort by e-value

                similarities.add(simi);
            }
        }

        return similarities;
    }

    /* <code>SortByEValueList</code>
     *
     */
    /*
    private class SortByEValueList {
        List list = new ArrayList();

        public void add(String score, StringBuffer emblString){
            if( list.size() == 0 ){
                list.add(emblString);
            }
        }
    }
    */

    /**
     * <code>parseOrganismFromDesc</code> parses text <code>String</code>
     * and returs what ever is in between the first pair of {} brackets
     * which is hopefully the name of the organism.
     *
     * @param test is a <code>String</code>.
     *
     * @return an <code>String</code>.
     */
    private String parseOrganismFromDesc(String text){
        char         last, next;
        int          length  = text.length();
        boolean      collect = false;
        StringBuffer sBuffer = new StringBuffer();

        for(int x = 0; x < length; ++x){
            next = text.charAt(x);

            if( next == '}' && collect){ // Finished collecting
                break;
            }
            else if( next == '{' ){      // Start collecting
                collect = true;
            }
            else if( collect ){          // collect
                sBuffer.append(next);
            }

            last = next;
        }

        return sBuffer.toString();
    }

    /**
     * <code>makeModelAnnotation</code> creates annotation from a
     * gene model Element.
     *
     * @param model an <code>Element</code>.
     *
     * @return an <code>Annotation</code>.
     */
    private Annotation makeModelAnnotation(Element model)
    {
        Annotation anno = new SimpleAnnotation();

        // Elements in MODEL element
        String [] modElem = new String [] { "FEAT_NAME" };

        try
        {
            populateAnnotation(model, modElem, anno);
        }
        catch (ChangeVetoException cve)
        {
            cve.printStackTrace();
        }

        return anno;
    }

    /**
     * <code>makeExonAnnotation</code> creates exon annotation, but
     * from the gene model Element.
     *
     * @param model an <code>Element</code>.
     *
     * @return an <code>Annotation</code>.
     */
    private Annotation makeExonAnnotation(Element model)
    {
        Annotation anno = new SimpleAnnotation();

        // Elements in EXON element
        String [] exonElem = new String [] { "FEAT_NAME" };

        List exons = model.getChildren("EXON");

        try
        {
            for (Iterator ei = exons.iterator(); ei.hasNext();)
            {
                Element exon = (Element) ei.next();
                populateAnnotation(exon, exonElem, anno);
            }
        }
        catch (ChangeVetoException cve)
        {
            cve.printStackTrace();
        }

        return anno;
    }

    /**
     * <code>makeRNAAnnotation</code> creates annotation from a
     * RNA_GENE element.
     *
     * @param rnaGene an <code>Element</code>.
     *
     * @return an <code>Annotation</code>.
     */
    private Annotation makeRNAAnnotation(Element rnaGene)
    {
        Annotation anno = new SimpleAnnotation();

        // Elements in the RNA_GENE element
        String [] rnaElem = new String [] { "FEAT_NAME",
                                            "COM_NAME" };

        try
        {
            populateAnnotation(rnaGene, rnaElem, anno);
        }
        catch (ChangeVetoException cve)
        {
            cve.printStackTrace();
        }

        return anno;
    }

    /**
     * <code>makeMiscAnnotation</code> creates annotation from a
     * MISC_FEATURE element *or* ATTRIBUTE_TYPE. The later is used for
     * TMHMM and SignalP.
     *
     * @param miscFeature an <code>Element</code>.
     *
     * @return an <code>Annotation</code>.
     */
    private Annotation makeMiscAnnotation(Element miscFeature)
    {
        Annotation anno = new SimpleAnnotation();

        //if( "ATTRIBUTE_TYPE".equals(miscFeature.getName()) ){
        //    System.out.println("misc_feature anno: ATTRIBUTE_TYPE");
        //}
        //else {
        //    System.out.println("misc_feature anno: !NOT! ATTRIBUTE_TYPE");
        //}

        // Elements in MISC_FEATURE
        String [] miscElem = new String [] { "FEATURE_DESC" };

        try
        {
            populateAnnotation(miscFeature, miscElem, anno);
        }
        catch (ChangeVetoException cve)
        {
            cve.printStackTrace();
        }

        return anno;
    }

    /**
     * <code>makeRepeatAnnotation</code> creates annotation from a
     * REPEAT element.
     *
     * @param repeat an <code>Element</code>.
     *
     * @return an <code>Annotation</code>.
     */
    private Annotation makeRepeatAnnotation(Element repeat)
    {
        Annotation anno = new SimpleAnnotation();

        // Elements in REPEAT
        String [] repElem = new String [] { "FEAT_NAME",
                                            "REPEAT_TYPE" };

        try
        {
            anno.setProperty("rpt_type", "direct");
            populateAnnotation(repeat, repElem, anno);
        }
        catch (ChangeVetoException cve)
        {
            cve.printStackTrace();
        }

        return anno;
    }

    private Annotation makeSourceAnnotation(Element assembly)
    {
        Annotation anno = new SimpleAnnotation();

        // Elements in ASSEMBLY
        String [] assElem = new String [] { "ASMBL_ID" };

        try
        {
            populateAnnotation(assembly, assElem, anno);
        }
        catch (ChangeVetoException cve)
        {
            cve.printStackTrace();
        }

        return anno;
    }


    /* <code>makeSigPTMHMMAnnotation</code> returns the annotation for a SignalP
     * based on a TMHMM predition. It is different to other methods here in 
     * that I have had to loop over ATT_SCORE elements and determine their
     * contents by the attribute DESC.
     *
     * @param attribute_type is an <code>Element</code> for an ATTRIBUTE_TYPE
     *
     * @return <code>Annotation</code>
     */
    private Annotation makeSigPTMHMMAnnotation(Element attribute_type,
                                               String  feat_name)
    {
        Annotation anno = new SimpleAnnotation();
        List att_scores = attribute_type.getChildren();

        try {
            for(int x = 0; x < att_scores.size(); ++x){
                Element att_score = (Element) att_scores.get(x);
                String  desc      = att_score.getAttributeValue("DESC");
                String  text      = att_score.getText();

                if( "HMM-position".equals(desc) ){
                    int pos = Integer.parseInt(text);
                    text = text+","+pos;
                    anno.setProperty("coord", text);
                }
                else if( "HMM-SPprob".equals(desc) ){
                    anno.setProperty("signal", text);
                }
                else if( "HMM-SAprob".equals(desc) ){
                    anno.setProperty("anchor", text);
                }
                else if( "HMM-cleavprob".equals(desc) ){
                    anno.setProperty("cleavage", text);
                }
                else if( "HMM-prediction".equals(desc) ){
                    anno.setProperty("note", text+" predicted for "+feat_name);
                }
            }
            
            anno.setProperty("type", "signalP");
            anno.setProperty("algorithm", TigrSequenceLoader.SIGNALPTMHMM);
        }
        catch (ChangeVetoException cve)
        {
            cve.printStackTrace();
        }

        return anno;
    }

    /* <code>makeTMHMMAnnotation</code> returns the annotation for a TMHMM
     * based on a TMHMM predition. It makes a <code>ArrayList</code> of
     * <code>Location</code> objects then sets this as the value of the 
     * coord property.
     *
     * @param attribute_type is an <code>Element</code> for an ATTRIBUTE_TYPE
     * @param aaLocs are one or more <code>Location</code> obbjects.
     * @param featName is a <code>String</code> gene name on which this TMHMM
     * has been predicted (used in the note field).
     *
     * @return <code>Annotation</code>
     */
    private Annotation makeTMHMMAnnotation(Element  attribute_type,
                                           Location aaLocs,
                                           String   featName)
        throws Exception
    {
        if (featName == null || "".equals(featName)) {
            throw new Exception("No featName passed into makeTMHMMAnnotation for Locations "+aaLocs);
        }

        Annotation anno = new SimpleAnnotation();

        try {
            Iterator    itty = aaLocs.blockIterator();
            ArrayList coords = new ArrayList();
            ArrayList   locs = new ArrayList();
            int        count = 0;

            while( itty.hasNext() ){
                Location loc = (Location) itty.next();
                coords.add(loc.getMin()+","+loc.getMax());
                locs.add(loc);
                ++count;
            }

            anno.setProperty("coord", coords);
            anno.setProperty("algorithm", TigrSequenceLoader.TMHMM);
            anno.setProperty("colour", "0");
            anno.setProperty("type", "TMHMM");

            StringBuffer note = new StringBuffer();

            if( count == 1 ){
                note.append("1 transmembrane helix predicted for ");
            }
            else {
                note.append(count+" transmembrane helices predicted for ");
            }

            note.append(featName);
            note.append(" by TMHMM2.0 at aa "); // FIX ME , TMHMM2.0 IS  HARD CODED!!!
            note.append(getStringForLocations(locs));

            anno.setProperty("note", note.toString());
        }
        catch (ChangeVetoException cve)
        {
            cve.printStackTrace();
        }

        return anno;
    }

    /* <code>getStringForLocations</code> builds a String from Locations
     * i.e.
     *  "1-2"
     *  "33-44 and 55-66"
     *  "4-23, 107-129 and 144-166"
     *
     */
    private String getStringForLocations(List coords){
        if( coords.size() == 0 ){
            return ""; // FIXME, THIS CAN'T BE RIGHT!!!
        }

        if( coords.size() == 1 ){
            Location loc = (Location) coords.get(0);
            return loc.getMin()+"-"+loc.getMax();
        }

        StringBuffer sb = new StringBuffer();

        for(int x = 0; x < coords.size(); ++x){
            Location loc = (Location) coords.get(x);

            if( x == 0 ){ // first coord
                sb.append(loc.getMin()+"-"+loc.getMax());
            }
            else if( x == (coords.size() - 1) ){ // last coord
                sb.append(" and "+loc.getMin()+"-"+loc.getMax());
            }
            else { // inbetween coord
                sb.append(", "+loc.getMin()+"-"+loc.getMax());
            }
        }

        return sb.toString();
    }

    /**
     * <code>populateAnnotationWithPepStats</code> populates an Annotation
     * object using the content of selected ATT_SCORE elements
     *
     * @param model is the <code>Element</code> that we derive
     * MODEL_ATTRIBUTE->ATTRIBUTE_TYPE->ATT_SCORE*
     * @param anno is the <code>Annotation</code> that we will add ot
     *
     */
    private void populateAnnotationWithPepStats(Element model, Annotation anno){
        try {
            Element model_attribute = model.getChild("MODEL_ATTRIBUTE");

            if( model_attribute == null ){
                return;
            }

            List attribute_types =
                model_attribute.getChildren("ATTRIBUTE_TYPE");

            for(int i = 0; i < attribute_types.size(); ++i){
                List att_scores =
                    ((Element) attribute_types.get(i)).getChildren("ATT_SCORE");

                if( att_scores.size() != 1 ){
                    continue;
                }

                Element att_score = (Element) att_scores.get(0);
                String  desc      = att_score.getAttributeValue("DESC");

                if( "Molecular Weight".equals(desc) ){
                    anno.setProperty("molecular_weight", att_score.getText()+" Da");
                }
                else if( desc.equals("Isoelectric point") ){
                    anno.setProperty("isoelectric_point", att_score.getText());
                }
                else if( desc.equals("OMP count") ){
                    anno.setProperty("OMP_count", att_score.getText());
                }
            }
        }
        catch(Exception e ){
            System.err.println("Exception trying to add pepstats;\n"+e);
        }
    }



    /**
     * <code>populateAnnotation</code> populates an Annotation object
     * using the content of selected child Elements.
     *
     * @param element an <code>Element</code> whose child Elements
     * contain the values to add to the Annotation.
     * @param elementNames a <code>String []</code> containing the
     * names of the child Elements to use.
     * @param annotation an <code>Annotation</code> object to
     * populate.
     *
     * @exception ChangeVetoException if the Annotation is locked.
     */
    private void populateAnnotation(Element    element,
                                    String []  elementNames,
                                    Annotation annotation)
        throws ChangeVetoException
    {
        for (int i = 0; i < elementNames.length; i++)
        {
            List childrenText = makeProperties(element, elementNames[i]);

            if (childrenText.size() == 0)
                continue;

            // Don't write out empty string
            if (childrenText.size() == 1 &&
                ((String)childrenText.get(0)).length() == 0){
                continue;
            }

            String mapping = null;

            try
            {
                mapping = annotationMap.getProperty(elementNames[i]);
            }
            catch (NoSuchElementException nse) { }

            // We may get a null if the Property doesn't exist
            String key = (mapping == null) ? elementNames[i] : mapping;

            Object currentVal = null;

            try
            {
                currentVal = annotation.getProperty(key);
            }
            catch (NoSuchElementException nse) { }

            // Collections of values are handled by the EMBL writing code
            if (currentVal != null && Collection.class.isInstance(currentVal))
            {
                if( useFirstQualifierOnly(key) ){
                    // Do nothing
                    //System.err.println("useFirstQualifierOnly("+key+") when trying to add "+childrenText);
                }
                else {
                    ((Collection) currentVal).addAll(childrenText);
                    annotation.setProperty(key, currentVal);
                }
            }
            else
            {
                annotation.setProperty(key, childrenText);
            }
        }
    }

    /**
     * <code>useFirstQualifierOnly</code>
     *
     * @param  qualifier is a <code>String</code> to check
     *
     * @return a boolean, true if use the first qualifier
     */
    private boolean useFirstQualifierOnly(String qualifier){

        String qualStrings = annotationMap.getProperty("use_first_qualifier");
        String []    quals = StringUtils.split(qualStrings,";");

        for (int i=0; i< quals.length ; i++) {
            String qual = quals[i].trim();

            if( qual.equals(qualifier) ){
                return true;
            }
        }

        return false;
    }



    /**
     * <code>backSlashSemis</code> is a utility function for the products/
     * description of a /similarity.
     *
     * @param text is the <code>String</code> to process
     *
     * @return a <code>String</code>
     */
    private String backSlashSemis(String text){
        StringBuffer sb = new StringBuffer();

        for(int x = 0; x < text.length(); ++x){
            if( text.charAt(x) == ';' ){
                sb.append("\\;");
            }
            else {
                sb.append(text.charAt(x));
            }
        }

        return sb.toString();
    }

    static RE putative;

    static {
        try {
            putative = new RE(",\\s*putative");
        }
        catch(org.apache.regexp.RESyntaxException exp){
            exp.printStackTrace();
        }
    }

    /**
     * <code>setColor</code> is a utility function. If the
     *    FUNCT_ANNOT_EVIDENCE TYPE="CURATED"
     * then it matches various strings to COM_NAME to set the /color accordingly.
     *
     * If the gene is a pseudo, the colour is set to 13 regardless of the above.
     *
     * FIXME: This is PSU specific, it needs to be moved out of here...
     *
     *
     * @parame geneInfo is the parent <code>Element</code> to get
     * FUNCT_ANNOT_EVIDENCE and COM_NAME from
     * @parame anno is the <code>Annotation</code> to update
     *
     */
    private void setColor(Element geneInfo, Annotation anno){

        Element eAnnoType = (Element) geneInfo.getChild("FUNCT_ANNOT_EVIDENCE");
        Element  eComName = (Element) geneInfo.getChild("COM_NAME");
        Element   ePseudo = (Element) geneInfo.getChild("IS_PSEUDOGENE");

        // Must have both of these elements to set the /color
        if( eAnnoType == null || eComName == null ){
            return;
        }

        String  sAnnoType = eAnnoType.getAttributeValue("TYPE");
        String   sComName = eComName.getText();
        String    sPseudo = ePseudo.getText();
        String      color = null;

        // Elements must contain text and be TU must be CURATED
        if( sAnnoType != null &&
            sComName  != null &&
            "CURATED".equals(sAnnoType) ){

            /*
            if( "hypothetical protein, unlikely".equals(sComName) ){
                color = "6";
            }
            else if( "hypothetical protein, conserved".equals(sComName) ){
                color = "10";
            }
            else if( "hypothetical protein".equals(sComName) ){
                color = "8";
            }

            if( putative.match(sComName) ){
                color = "7";
            }
            */
            boolean isPseudo = false;

            if (sPseudo.equals("1")) {
                isPseudo = true;
            }

            color = TigrUtils.getColourFromProduct(sComName,  isPseudo);
        }

        if( sPseudo.equals("1") ){
            color = "13";
        }

        try {
            if( color != null ){
                anno.setProperty("colour", color);
            }
        }
        catch(ChangeVetoException cve){
            // Dont care of can't set color
            System.err.println("WARNING: Could not set /color to "+color);
        }
    }

    /**
     * <code>setGO</code> is a utility function. If the GENE_INFO has a
     * GENE_ONTOLOGY element then output some GO qualifiers
     *
     * @param geneInfo is the parent <code>Element</code> to get
     * FUNCT_ANNOT_EVIDENCE and COM_NAME from
     * @param anno is the <code>Annotation</code> to update
     *
     */
    private void setGO(Element geneInfo, Annotation anno){

        Element eGeneOntology = (Element) geneInfo.getChild("GENE_ONTOLOGY");

        if( eGeneOntology == null ){
            return;
        }

        // Get zero or more GO_ID elements
        List  goIds = eGeneOntology.getChildren("GO_ID");

        //System.err.println("goIds.size() = "+goIds.size());

        for(int x = 0; x < goIds.size(); ++x){

            Element        goId = (Element) goIds.get(x);
            String goAssignment = goId.getAttributeValue("ASSIGNMENT");
            String     goAspect = goId.getChildText("GO_TYPE");
            String       goTerm = goId.getChildText("GO_TERM");
            String         date = goId.getChildText("DATE");

            Element goEvidence = null, evCode = null;
            String  sEvidence = "", sWithEvidence = "", sEvCode = "";

            // GO_EVIDENCE may not be present...
            goEvidence = goId.getChild("GO_EVIDENCE");

            if( goEvidence != null ){
                evCode        = goEvidence.getChild("EV_CODE");
                sEvCode       = evCode.getAttributeValue("CODE");
                sEvidence     = goEvidence.getChildText("EVIDENCE");
                sWithEvidence = goEvidence.getChildText("WITH_EV");
            }

            // GO qualifier wanta date in 20031231 format
            date = convertDateToNumbers(date);

            // We don't want nulls i.e 'with=null'
            goTerm        = condExpand("term",   goTerm);
            sEvCode       = condExpand("evidence", sEvCode);
            sEvidence     = condExpand("db_xref",  sEvidence);
            sWithEvidence = condExpand("with",     sWithEvidence);
            date          = condExpand("date",     date);

            // Sanity check GO aspect
            if( ! ("function".equals(goAspect) ||
                   "process".equals(goAspect)  ||
                   "component".equals(goAspect)) ){
                System.err.println("ERROR: Unknown GO aspect: "+goAspect);
		throw new RuntimeException("ERROR: Unknown GO aspect: "+goAspect);
            }


            try {
                addToProperty(anno, "GO",
                              "aspect="+goAspect+
                              "; GOid="+goAssignment+
                              goTerm+
                              sEvCode+
                              sEvidence+
                              sWithEvidence+
                              date);
            }
            catch(ChangeVetoException cve){
                System.err.println("WARNING: Could not set /GO_xxx for "+
                                   goAssignment);
            }
        }
    }

    /* <code>convertDateToNumbers</code> converts "May 24 2002 2:38PM" into 20020524
     * @param rValue is a <code>String</code>
     *
     * @return a <code>String</code> with an 8 digit number
     */ 
    private static String convertDateToNumbers(String date){
        // May 24 2002 2:38PM
        String[] sections = StringUtils.split(date," ");

        if (sections.length == 0) {
            System.err.println("Date '"+date+"' could not be parsed");
            System.exit(1);
        }

        String month = sections[0]; // This is not a number, its "Jan" etc
        String   day = sections[1];
        String  year = sections[2];

        HashMap   hm = new HashMap(12);
        hm.put("Jan", "01");
        hm.put("Feb", "02");
        hm.put("Mar", "03");
        hm.put("Apr", "04");
        hm.put("May", "05");
        hm.put("Jun", "06");
        hm.put("Jul", "07");
        hm.put("Aug", "08");
        hm.put("Sep", "09");
        hm.put("Oct", "10");
        hm.put("Nov", "11");
        hm.put("Dec", "12");

        month = (String) hm.get(month);

        if (day.length() == 1) {
            day = "0"+day;
        }

        return year+month+day;
    }


    /* <code>condExpand</code> checks to see if rValue is set to something.
     * If it is it simply returns ";"+lValue+"="+rValue i.e. "; term=transport"
     *
     * @param lValue is a <code>String</code>
     * @param rValue is a <code>String</code>
     *
     * @return a <code>String</code> being either "" or "; "+lValue+"="+rValue
     */ 
    private static String condExpand(String lValue, String rValue){
        String toReturn = "";

        if( rValue != null && ! rValue.equals("") ){
            toReturn = "; "+lValue+"="+rValue;
        }

        return toReturn;
    }

    /**
     * <code>addToProperty</code> is a wrapper around Annotatione.setProperty().
     * If no value exists for the key, it adds the value.
     * If a value does exist and is a String
     *   create a List and add 2 element to it, making this the valie
     * If a value does exist and is a List
     *   add to list
     *
     * @param anno is the <code>Annotation</code> to add to
     * @param is the key <code>String</code>
     * @param is the value <code>String</code>
     *
     */
    public static void addToProperty(Annotation anno, String key, String value)
    throws ChangeVetoException{
        try {
            Object o = anno.getProperty(key);

            if( o instanceof String ){
                ArrayList al = new ArrayList();
                al.add(o);
                al.add(value);
                anno.setProperty(key, al);
            }
            else if( o instanceof List ) {
                ((List) o).add(value);
            }
        }
        catch( NoSuchElementException nse ){
            // property not yet used, so set it here for the 1st time
            anno.setProperty(key, value);
        }
    }

    /* <code>addToProperty</code> is as above but handles List input of values
     *
     */
    public static void addToProperty(Annotation anno, String key, List value)
    throws ChangeVetoException{
        try {
            Object o = anno.getProperty(key);

            if( o instanceof String ){
                ArrayList al = new ArrayList();
                al.add(o);
                al.addAll(value);
                anno.setProperty(key, al);
            }
            else if( o instanceof List ) {
                ((List) o).addAll(value);
            }
        }
        catch( NoSuchElementException nse ){
            // property not yet used, so set it here for the 1st time
            anno.setProperty(key, value);
        }
    }

    /**
     * <code>makeProperties</code> extracts a list of values from an
     * Element, consisting of all the child Element text content.
     *
     * @param element an <code>Element</code>
     * @param key a <code>String</code> which the the name of child
     * Elements to process.
     *
     * @return a <code>List</code> of Strings.
     */
    private List makeProperties(Element element, String key)
    {
        List children = element.getChildren(key);
        List   values = new ArrayList();

        if (children.size() > 0)
        {
            for (Iterator ci = children.iterator(); ci.hasNext();)
            {
                Element child = (Element) ci.next();
                values.add(child.getText());
            }
        }
        return values;
    }
}


