/*
 * Copyright (c) 2001 Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package uk.ac.sanger.psu.tigread;

import java.util.Properties;

import org.biojava.bio.Annotation;
import org.biojava.bio.SimpleAnnotation;
import org.biojava.bio.BioException;
import org.biojava.utils.ChangeVetoException;

import org.jdom.Attribute;
import org.jdom.Element;

import org.apache.commons.lang.StringUtils;

/**
 * <code>EmblCustomQualifiers</code> is (hopefully) where all non-standard
 * embl qualifier code will live
 *
 *
 * @author Paul Mooney
 */
class EmblCustomQualifiers {
    static Properties annotationMap;
    static String[]   psuDbXrefOrder;
    static String[]   tempSysIDOrder;
    static String[]   synonymOrder;

    EmblCustomQualifiers(Properties annotationMap){
        this.annotationMap = annotationMap;

        // psu_db_xref
        String psuDBXrefHeirarchyOrder =
            annotationMap.getProperty("PSU_DB_XREF_ORDER");

        if( psuDBXrefHeirarchyOrder != null )
            psuDbXrefOrder = StringUtils.split(psuDBXrefHeirarchyOrder,";");

        // temporary_systematic_id
        String tempSysIDHeirarchyOrder =
            annotationMap.getProperty("TEMPORARY_SYSTEMATIC_ID_ORDER");

        if( tempSysIDHeirarchyOrder != null )
            tempSysIDOrder = StringUtils.split(tempSysIDHeirarchyOrder,";");

        // synonym
        String synonymOrderProp =
            annotationMap.getProperty("SYNONYM_ORDER");

        if( synonymOrderProp != null && !synonymOrderProp.equals("") ){
            synonymOrder = StringUtils.split(synonymOrderProp,";");
        }

   }

    /**
     * <code>addCustomQualifiers</code> populates an Annotation object
     * using the content of selected child Elements.
     * <P>
     * This is where the custom, non generic EMBL qualifiers are specified.
     * 
     *
     * @param element an <code>Element</code> whose child Elements
     * contain the values to add to the Annotation.
     * @param annotation an <code>Annotation</code> object to
     * populate.
     */
    protected static void addCustomQualifiers(Element element, Annotation anno)
        throws ChangeVetoException {

        if("TU".equals(element.getName())){
            addCustomTUQualifers(element, anno);
        }
                
    }

    /**
     * <code>addCustomTUQualifiers</code> populates an Annotation object
     * using the content of selected child MODEL Elements.
     *
     * @param element an <code>Element</code> whose child Elements
     * contain the values to add to the Annotation.
     * @param annotation an <code>Annotation</code> object to
     * populate.
     *
     * @exception ChangeVetoException if the Annotation is locked.
     */
    private static void addCustomTUQualifers(Element tu, Annotation anno)
        throws ChangeVetoException {

        // Can we add a /psu_db_xref
        //
        String url = annotationMap.getProperty("PSU_DB_XREF_URL");

        if( url != null ){
            String text = getTextFromHeirarchy(psuDbXrefOrder, tu);

            if( text != null ){
                EmblAnnotationFactory.addToProperty(anno,
                                                    "psu_db_xref",
                                                    url+text+";");
            }
        }


        // try and add a temporary_systematic_id
        //
        String tsi = getTextFromHeirarchy(tempSysIDOrder, tu);

        if( tsi != null ){
            EmblAnnotationFactory.addToProperty(anno,
                                                "temporary_systematic_id",
                                                tsi);
        }

        // synonym
        //
        String synonym = getTextFromHeirarchy(synonymOrder, tu);

        //System.err.println("synonym = "+synonym+
        //                   ", synonymOrder.length = "+synonymOrder.length);

        if( synonym != null ){
            EmblAnnotationFactory.addToProperty(anno, "synonym", synonym);
        }
    }

    /* <code>getTextFromHeirarchy</code> get the text from the first matching
     * hierarchy of elements (xpath).
     *
     * @param order is a <code>String</code>[]  of each xpath
     */
    private static String getTextFromHeirarchy(String[] order, Element tu){
        String toReturn = null;

        if( order == null )
            return toReturn;

        // For each XML path
        for (int i=0; i< order.length ; i++) {
            String   heirarchy = order[i].trim();
            String [] elements = StringUtils.split(heirarchy,".");

            //System.err.print(heirarchy+"\n  ");

            Element ele = null;

            // For each element on that path
            for (int j=0; j< elements.length ; j++) {
                //System.err.print(elements[j]+" . ");

                if( ele == null ){
                    ele = tu.getChild(elements[j]);
                        }
                else {
                    ele = ele.getChild(elements[j]);
                }
            }

            if( ele != null && !ele.getText().equals("") ){
                //System.err.print(" VALUE = "+ele.getText());

                toReturn = ele.getText();
 
                break;
            }

            //System.err.println();
        }

        return toReturn;
    }
}
