/*
 * Copyright (c) 2001  Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package uk.ac.sanger.psu.tigread;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

import org.jdom.Element;

import org.biojava.bio.Annotation;
import org.biojava.bio.BioException;
import org.biojava.bio.SimpleAnnotation;
import org.biojava.bio.seq.StrandedFeature;
import org.biojava.bio.seq.StrandedFeature.Strand;
//import org.biojava.bio.symbol.CompoundLocation;
import org.biojava.bio.symbol.Location;
import org.biojava.bio.symbol.LocationTools;
import org.biojava.bio.symbol.RangeLocation;
import org.biojava.utils.ChangeVetoException;

/**
 * <code>EmblFeatureTemplateFactory</code> creates templates for
 * creation of various types of EMBL feature corresponding to TIGR XML
 * elements. As the mapping is in flux, not all of these templates are
 * actually used.
 *
 * @author <a href="mailto:kdj@sanger.ac.uk">Keith James</a>
 */
class EmblFeatureTemplateFactory
{
    private EmblLocationFactory   locFactory;
    private EmblAnnotationFactory annoFactory;

    EmblFeatureTemplateFactory()
    {
        locFactory  = new EmblLocationFactory();
        annoFactory = new EmblAnnotationFactory();
    }

    /**
     * <code>makeFeatureTemplate</code> creates a template for feature
     * creation suitable for the type of <code>Element</code> passed.
     *
     * @param element an <code>Element</code>.
     * @param identifier an <code>int</code>.
     *
     * @return a <code>StrandedFeature.Template</code>.
     *
     * @exception BioException if an error occurs
     */
    StrandedFeature.Template makeFeatureTemplate(final Element element,
                                                 final int     identifier)
        throws BioException
    {
        StrandedFeature.Template strandedTemplate = null;

        switch (identifier)
        {
            case TigrSequenceLoader.TU:
                 strandedTemplate = makeTUTemplate(element);
                 break;

            case TigrSequenceLoader.MODEL:
                 strandedTemplate = makeModelTemplate(element);
                 break;

            case TigrSequenceLoader.EXON:
                 strandedTemplate = makeExonTemplate(element);
                 break;

            case TigrSequenceLoader.RNA:
                 strandedTemplate = makeRNATemplate(element);
                 break;

            case TigrSequenceLoader.MISC:
                 strandedTemplate = makeMiscTemplate(element);
                 break;

            case TigrSequenceLoader.REPEAT:
                 strandedTemplate = makeRepeatTemplate(element);
                 break;

            case TigrSequenceLoader.SOURCE:
                 strandedTemplate = makeSourceTemplate(element);
                 break;

            default:
                 System.err.println("Requested unknown StrandedFeature.Template type '"
                                    + identifier
                                    + "' from EmblFeatureTemplateFactory");
                 strandedTemplate = new StrandedFeature.Template();
        }

        /* DEBUG check */
        if( strandedTemplate.location == null ){
            System.err.println("element = "+element);
        }

        return strandedTemplate;
    }


    /**
     * <code>makeFeatureTemplate</code> creates a template for feature
     * creation suitable for the type of <code>Element</code> passed.
     *
     * NOTE: This is a modified version of the above method. MODEL is passed
     * is because the DNA COORDS are needed
     *
     * @param element an <code>Element</code>.
     * @param identifier an <code>int</code>.
     *
     * @return a <code>StrandedFeature.Template</code>.
     *
     * @exception BioException if an error occurs
     */
    StrandedFeature.Template makeFeatureTemplate(final Element model,
                                                 final Element element,
                                                 final int     identifier)
        throws BioException
    {
        StrandedFeature.Template strandedTemplate = null;

        switch (identifier)
        {
            case TigrSequenceLoader.SIGPTMHMM:
                strandedTemplate =  makeSignalPTMHMMTemplate(model, element);
                break;
            case TigrSequenceLoader.TMM:
                strandedTemplate =  makeTMHMMTemplate(model, element);
                break;
            default:
                System.err.println("Requested unknown StrandedFeature.Template type '"
                                   + identifier
                                   + "' from EmblFeatureTemplateFactory");
                strandedTemplate =  new StrandedFeature.Template();
        }

        /* DEBUG check */
        if( strandedTemplate != null &&  strandedTemplate.location == null ){
            String featName = model.getChildText("FEAT_NAME");
            System.err.println("Location is null for "+featName+", identifier = "+identifier);
        }


        return strandedTemplate;
    }

    /* <code>makeTUTemplate</code> called from <code>makeExonTemplate</code>.
     */
    private StrandedFeature.Template makeTUTemplate(final Element tu)
    {
        StrandedFeature.Template sft = new StrandedFeature.Template();

        StrandedLocation loc = (StrandedLocation)
            locFactory.makeLocation(tu.getChild("COORDSET"));

        sft.type       = "TU";
        sft.source     = "TIGR XML";
        sft.location   = new RangeLocation(loc.getMin(), loc.getMax());
        sft.strand     = loc.getStrand();
        sft.annotation =
            annoFactory.makeAnnotation(tu, TigrSequenceLoader.TU);

        return sft;
    }

    private StrandedFeature.Template makeModelTemplate(final Element model)
    {
        StrandedFeature.Template sft = new StrandedFeature.Template();
	
        StrandedLocation loc = (StrandedLocation)
            locFactory.makeLocation(model.getChild("COORDSET"));

        sft.type       = "MODEL";
        sft.source     = "TIGR XML";
        sft.location   = new RangeLocation(loc.getMin(), loc.getMax());
        sft.strand     = loc.getStrand();
        sft.annotation =
            annoFactory.makeAnnotation(model, TigrSequenceLoader.MODEL);

        return sft;
    }

    /* <code>makeExonTemplate</code> makes the CDS feature NOT an Exon
     * feature as the name would imply. This method mainly gathers the 
     * coords of all the exons, creates a CDS featue and calls the TU template
     * method to add annotation.
     */
    private StrandedFeature.Template makeExonTemplate(final Element tu)
    {
        StrandedFeature.Template sft = new StrandedFeature.Template();

        // Assumes one gene model currently
        List   exons = tu.getChild("MODEL").getChildren("EXON");
        List locList = new ArrayList();

        Strand strand = StrandedFeature.POSITIVE;

        for (Iterator ei = exons.iterator(); ei.hasNext();)
        {
            Element exon   = (Element) ei.next();
            Element coords = null;

            if( exon.getChild("CDS") != null ){
                coords = exon.getChild("CDS").getChild("COORDSET");
            }
            else {
                // Probably all UTR!
                continue;
                //System.err.println("NO CDS for "+
                //                   exon.getChildText("FEAT_NAME"));
                //coords = exon.getChild("COORDSET");
            }

            StrandedLocation loc = (StrandedLocation)
                locFactory.makeLocation(coords);

            locList.add(new RangeLocation(loc.getMin(), loc.getMax()));

            strand = loc.getStrand();
        }

        sft.type       = "CDS";
        sft.source     = "TIGR XML";
        sft.location   =
            (locList.size() == 1)     ?
            (Location) locList.get(0) :
            LocationTools.union(locList);
        sft.strand     = strand;

        // Pass the TU so we get annotation from all EXONS
        sft.annotation =
            annoFactory.makeAnnotation(tu, TigrSequenceLoader.TU);

        return sft;
    }

    private StrandedFeature.Template makeRNATemplate(final Element rna)
    {
        StrandedFeature.Template sft = new StrandedFeature.Template();

        Element exon;
        String  type = "misc_feature";
        String  name = rna.getName();

        if (name.equals("PRE-TRNA"))
        {
            exon = rna.getChild("TRNA").getChild("RNA-EXON");
            type = "tRNA";
        }
        else
        {
            exon = rna;
        }

        if (name.equals("RRNA"))
            type = "rRNA";
        else if (name.equals("SNRNA"))
            type = "snRNA";
        else if (name.equals("SNORNA"))
            type = "snoRNA";

        StrandedLocation loc = (StrandedLocation)
            locFactory.makeLocation(exon.getChild("COORDSET"));

        sft.type       = type;
        sft.source     = "TIGR XML";
        sft.location   = new RangeLocation(loc.getMin(), loc.getMax());
        sft.strand     = loc.getStrand();
        sft.annotation =
            annoFactory.makeAnnotation(rna, TigrSequenceLoader.RNA);

        if (rna.getName().equals("PRE-TRNA"))
        {
            String anticodon = rna.getChild("TRNA").getAttributeValue("ANTICODON");

            try
            {
                sft.annotation.setProperty("anticodon", anticodon);
            }
            catch (ChangeVetoException cve)
            {
                cve.printStackTrace();
            }
        }

        return sft;
    }

    /* <code>makeMiscTemplate</code> returns a misc_feature.
     *
     * Bastardized to take MISC_FEATURE or ATTRIBUTE_TYPE. The latter is for
     * SignalP and TMHMM
     *
     */
    private StrandedFeature.Template makeMiscTemplate(final Element misc)
    {
        StrandedFeature.Template sft = new StrandedFeature.Template();
        StrandedLocation         loc = null;

        if( "MISC_FEATURE".equals(misc.getName()) ){
            loc = (StrandedLocation)
                locFactory.makeLocation(misc.getChild("COORDSET"));
        }



        sft.type       = "misc_feature";
        sft.source     = "TIGR XML";
        sft.location   = new RangeLocation(loc.getMin(), loc.getMax());
        sft.strand     = loc.getStrand();
        sft.annotation =
            annoFactory.makeAnnotation(misc, TigrSequenceLoader.MISC);

        return sft;
    }

    private StrandedFeature.Template makeRepeatTemplate(final Element repeat)
    {
        StrandedFeature.Template sft = new StrandedFeature.Template();

        StrandedLocation loc = (StrandedLocation)
            locFactory.makeLocation(repeat.getChild("COORDSET"));

        sft.type       = "repeat_region";
        sft.source     = "TIGR XML";
        sft.location   = new RangeLocation(loc.getMin(), loc.getMax());
        sft.strand     = loc.getStrand();
        sft.annotation =
            annoFactory.makeAnnotation(repeat, TigrSequenceLoader.REPEAT);

        return sft;
    }

    private StrandedFeature.Template makeSourceTemplate(final Element assembly)
    {
        StrandedFeature.Template sft = new StrandedFeature.Template();

        StrandedLocation loc = (StrandedLocation)
            locFactory.makeLocation(assembly.getChild("COORDSET"));
	
        sft.type       = "source";
        sft.source     = "TIGR XML";
        sft.location   = new RangeLocation(loc.getMin(), loc.getMax());
        sft.strand     = loc.getStrand();
        sft.annotation =
            annoFactory.makeAnnotation(assembly, TigrSequenceLoader.SOURCE);

        return sft;
    }

   /**
     * <code>makeSignalPTemplate</code> creates a misc_feture template. This
     * gets the annotation first to see if it is a "Non-secretory protein"
     * prediction. If it is, ignore it as it this should just be mentioned in
     * the note field, if anywhere.
     *
     * @param model an <code>Element</code> - MODEL
     * @param att_type an <code>Element</code> - ATTRIBUTE_TYPE
     *
     * @return <code>StrandedFeature.Template</code>
     *
     */
    public StrandedFeature.Template makeSignalPTMHMMTemplate(final Element model,
                                                             final Element att_type){

        Annotation anno;
        String bestName = getBestName(model);

        anno = annoFactory.makeAnnotation(att_type,
                                          bestName,
                                          TigrSequenceLoader.SIGPTMHMM);

        String s = (String) anno.getProperty("note");

        //System.err.println("makeSignalPTMHMMTemplate() note = "+s+", bestName = "+bestName);

        // Is this a "Non-secretory protein" prediction we need to ignore?
        // NOTE: Will we always want to ignore them???
        if( s.indexOf("Non-secretory protein") >= 0 ){
            return null;
        }

        // If Yes, get details
        /*
        System.err.println("bestName    = "+bestName+"\n"+
                           "  coord     = "+((String) anno.getProperty("coord"))+"\n"+
                           "  signal    = "+((String) anno.getProperty("signal"))+"\n"+
                           "  anchor    = "+((String) anno.getProperty("anchor"))+"\n"+
                           "  cleavage  = "+((String) anno.getProperty("cleavage"))+"\n"+
                           "  type      = "+((String) anno.getProperty("type"))+"\n"+
                           "  algorithm = "+((String) anno.getProperty("algorithm"))+"\n"+
                           "  note      = "+((String) anno.getProperty("note")));
        */

        // Create Location for Exon(s) and protein
        // AND error check IT DOES NOT START AT ZERO!!!
        String coord = (String) anno.getProperty("coord");
        int    comma = coord.indexOf(",");
        String sPos  = coord.substring(0, comma);
        int    iPos  = Integer.parseInt(sPos);

        
        if( iPos == 0 ){
            System.err.println("Ignoring '"+anno.getProperty("note")+"' - it has a position of zero");
            return null;
        }

        StrandedLocation exonLocs = locFactory.getExonLocations(model);
        Location            aaLoc = new RangeLocation(1, iPos);

        // convert AA coords into DNA coords
        Location rloc = locFactory.convertAA2DNA(aaLoc,
                                                 exonLocs);


        StrandedFeature.Template sft = new StrandedFeature.Template();

        sft.type       = "misc_feature";
        sft.source     = "TIGR XML";
        sft.location   = rloc;
        sft.strand     = exonLocs.getStrand();
        sft.annotation = anno;

        return sft;
    }

   /**
     * <code>makeTMHMMupTemplate</code> creates a misc_feture template
     *
     * @param model an <code>Element</code> - MODEL
     * @param att_type an <code>Element</code> - ATTRIBUTE_TYPE
     *
     * @return <code>StrandedFeature.Template</code>
     *
     */
    public StrandedFeature.Template makeTMHMMTemplate(final Element model,
                                                      final Element att_type){

        // Get AA locations of TMHMMs
        // Covert AA coords into DNA. Need exons locations, hence;
        StrandedLocation  exonLocs = locFactory.getExonLocations(model);
        Location            aaLocs = locFactory.makeLocationForTMHMM(att_type);

        if( aaLocs == null ){
            String featName = model.getChildText("FEAT_NAME");
            System.err.println("\nERROR: no location for TMHMM "+
                               "for featname "+featName+"\n"+
                               "IGNORING TMHMM FEATURE!!!!!!!!!!!!!!!!!!\n\n");

            return null;
	    //throw new RuntimeException("No location");
        }

        //System.err.println("makeTMHMMTemplate(): aalocation = "+
        //                   locFactory.printLocations(aaLocs));



        // convert AA coords into DNA coords
        Location rloc = locFactory.convertAA2DNA(aaLocs,
                                                 exonLocs);

        // WARNING: multiple exons are not currently supported!
        //
        if( rloc == null ){
            return null;
        }

        StrandedFeature.Template sft = new StrandedFeature.Template();

        // The TMHMMs do not come with any annotation, just their locations
        Annotation anno = null;
        //String featName = model.getChildText("FEAT_NAME"); // FIX ME: THIS SHOULD BE PUB_LOCUS, LOCUS, FEAT_NAME...
        String bestName = getBestName(model);


        sft.type       = "misc_feature";
        sft.source     = "TIGR XML";
        sft.location   = rloc;
        sft.strand     = exonLocs.getStrand();

        try {
            sft.annotation = annoFactory.makeAnnotation(att_type,
                                                        TigrSequenceLoader.TMM,
                                                        aaLocs,
                                                        bestName);
        }
        catch(Exception e){
            System.err.println("ERROR: Could not create feature for TMHMM;\n"+
                               "       att_type               = "+att_type+"\n"+
                               "       TigrSequenceLoader.TMM =  "+
                               TigrSequenceLoader.TMM+"\n"+
                               "       aaLocs                 = "+aaLocs+"\n"+
                               "       bestName               = '"+bestName+"'\n");
            e.printStackTrace();
            throw new RuntimeException("Could not create feature for TMHMM");
        }

        return sft;
    }

    /* <code>getBestName</code> in this order;
     *   <LI> TU -> GENE_INFO -> PUB_LOCUS
     *   <LI> TU -> GENE_INFO -> PUB_LOCUS
     *   <LI> TU -> FEAT_NAME
     *
     * Return the "best" name from GENE_INFO in the order PUB_LOCUS, LOCUS
     * and FEAT_NAME
     *
     * @param element <code>element</code> is either a <GENE_INFO> or a <MODEL>.
     * The code will navigate to the elements required.
     * @return the best name as a <code>String</code>
     */
    private String getBestName(Element element){
        Element geneInfo = null;

        if( "GENE_INFO".equals(element.getName()) ){
            geneInfo = element;
        }
        else if( "MODEL".equals(element.getName()) ){
            geneInfo = element.getParent().getChild("GENE_INFO");
        }

        String pubLocus = geneInfo.getChildText("PUB_LOCUS");

        if( pubLocus != null && ! "".equals(pubLocus)){
            return pubLocus;
        }

        String locus = geneInfo.getChildText("LOCUS");

        if( locus != null && ! "".equals(locus)){
            return locus;
        }

        String featName = geneInfo.getParent().getChildText("FEAT_NAME");

        if( featName != null ){
            return featName;
        }

        // FIXME: some sort of proper exception is required here.
        System.err.println("WARNING: getBestname() Could not get a name!");
        return ""; // FIXME!!!!!!!!!!!!!!!!!!!
    }

    // Utility method for creating SignalP & TMHMM misc_features
    private StrandedFeature.Template makeMiscTemplate(StrandedLocation loc,
                                                      Annotation       anno){

        StrandedFeature.Template sft = new StrandedFeature.Template();

        sft.type       = "misc_feature";
        sft.source     = "TIGR XML";
        sft.location   = new RangeLocation(loc.getMin(), loc.getMax());
        sft.strand     = loc.getStrand();
        sft.annotation = anno;

        return sft;
    }

}



