/*
 * Copyright (c) 2001  Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package uk.ac.sanger.psu.tigread;

import java.util.Collection;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

import org.jdom.Element;

import org.biojava.bio.seq.StrandedFeature;
import org.biojava.bio.symbol.Location;
import org.biojava.bio.symbol.RangeLocation;
import org.biojava.bio.symbol.LocationTools;

/**
 * <code>EmblLocationFactory</code> creates <code>Location</code> and
 * strand information from a TIGR XML COORDSET element.
 *
 * it has now been altered to get a compound location from the exons.
 *
 * @author <a href="mailto:kdj@sanger.ac.uk">Keith James</a>
 */
class EmblLocationFactory
{
    /**
     * <code>makeLocation</code>.
     *
     * @param coordinateSet a COORDSET <code>Element</code>.
     *
     * @return a <code>StrandedLocation</code>.
     */
    Location makeLocation(final Element coordinateSet)
    {
        int [] coords = getCoordinates(coordinateSet);

        StrandedLocation loc;

        if (coords[0] < coords[1])
        {
            loc = new StrandedLocation(new RangeLocation(coords[0],
                                                         coords[1]));
            loc.setStrand(StrandedFeature.POSITIVE);

            return loc;
        }
        else if (coords[1] < coords[0])
        {
            loc = new StrandedLocation(new RangeLocation(coords[1],
                                                         coords[0]));
            loc.setStrand(StrandedFeature.NEGATIVE);

            return loc;                          
        }
        else
        {
            System.err.println("Warning: unable to determine strand of point feature at "
                               + coords[0]
                               + "; assuming forward strand");

            loc = new StrandedLocation(new RangeLocation(coords[0],
                                                         coords[1]));
            loc.setStrand(StrandedFeature.POSITIVE);

            return loc;
        }
    }

    /**
     * <code>getCoordinates</code> parses the coordinates from a
     * COORDSET element.
     *
     * @param coordinateSet an <code>Element</code>.
     *
     * @return an <code>int []</code> value with the 5' coordinate at
     * index 0 and the 3' coordinate at index 1.
     */
    private int [] getCoordinates(final Element coordinateSet)
    {
        return new int [] { Integer.parseInt(coordinateSet.getChildText("END5")),
                            Integer.parseInt(coordinateSet.getChildText("END3")) };
    }

    /*<code>getExonLocations</code> returns a <code>Location</code> for the
     * exons under a MODEL element
     *
     * @param model a MODEL <code>Element</code>.
     *
     * @return a <code>StrandedLocation</code>.
     *
     */
    StrandedLocation getExonLocations(final Element model){

        // Assumes one gene model currently
        List           exons = model.getChildren("EXON");
        List         locList = new ArrayList();
        StrandedLocation sloc = null;

        for (Iterator ei = exons.iterator(); ei.hasNext();)
        {
            Element exon = (Element) ei.next();

            sloc = (StrandedLocation) makeLocation(exon.getChild("COORDSET"));

            locList.add(new RangeLocation(sloc.getMin(), sloc.getMax()));
        }

        StrandedLocation newSloc = null;

        if( locList.size() == 1 ){
            newSloc = sloc;
        }
        else {
            Location locTemp = LocationTools.union(locList);
            newSloc = new StrandedLocation(locTemp);
            newSloc.setStrand(sloc.getStrand());
        }


        /*
        Iterator itty = newSloc.blockIterator();
        while(itty.hasNext()){
            Location l = (Location) itty.next();
            System.err.print(l+", ");
        }
        System.err.println();
        */

        return newSloc;
    }

    /**
     * <code>makeLocationForAT</code> returns a <code>Location</code> for
     * an ATTRIBUTE_TYPE. This is usually for a SignalP or TMHMM
     *

     * !!!!!!!!!!!!!
     * !!!!!!!!!!!!! IS THIS METHOD USED ANYWHERE ?? !!!!!!!!!!!!!!
     * !!!!!!!!!!!!!

     *
     * @param coordinateSet a COORDSET <code>Element</code>.
     *
     * @return a <code>StrandedLocation</code>.
     */
    Location makeLocationForAT(final Element at)
    {
        StrandedLocation loc = null;
        String method = at.getAttributeValue("METHOD");

        if( TigrSequenceLoader.SIGNALPTMHMM.equals(method) ){
            loc = makeLocationForSignalP(at);
        }
        else if( TigrSequenceLoader.TMHMM.equals(method) ){
            // Not yet implemented
            //loc = makeLocationForTMHMM(at);
        }
        else {
            System.err.println("ERROR: makeLocationForAT could not get "+
                               "location for method '"+method+"'");
            throw new RuntimeException("ERROR: makeLocationForAT could not get location for method '"+method+"'");
        }

        return loc;
    }

    /* 
     *
     * may return null if SignalP does not have a location
     */

     StrandedLocation makeLocationForSignalP(final Element at){
        StrandedLocation loc = null;
        List      att_scores = at.getChildren("ATT_SCORE");

        String probability = null; // These will be set in the below loop
        String prediction  = null;
        int    position    = 0;

        for(int x = 0; x < att_scores.size(); ++x){
            Element att_score  = (Element) att_scores.get(x);
            String  desc       = att_score.getAttributeValue("DESC");

            if( "HMM-position".equals(desc) ){
                String posString  = att_score.getText();
                position = Integer.parseInt(posString);

                // THIS IS AA COORDS, NOT DNA COORDS
                // NOT GOOD FOR EMBL FILE Location
                /*
                loc = new StrandedLocation(new RangeLocation(posInt,
                                                             posInt));
                loc.setStrand(StrandedFeature.POSITIVE);
                */
            }
            else if( "HMM-SPprob".equals(desc) ){
                probability = att_score.getText();
            }
            else if( "HMM-prediction".equals(desc) ){
                prediction = att_score.getText();
            }
        }

        // Collected data from ATT_SCOREs, now record them?
        if( position != 0 ){
            
        }

        return loc;
    }

    /* <code>makeLocationForTMHMM</code> extracts the location from the
     * following XML.
     *
     * <ATTRIBUTE_TYPE METHOD="GES,TMHMM2.0">
     *   <ATT_SCORE DESC="count">9</ATT_SCORE>
     *   <ATT_SCORE DESC="coords">234-253:265-287:391-413:893-915:930-952:973-995:1010-1032:1037-1059:1132-1151</ATT_SCORE>
     * </ATTRIBUTE_TYPE>
     *
     * @param att_type an ATTRIBUTE_TYPE <code>Element</code>.
     *
     * @return a <code>Location</code>.
     * 
     */
    public Location makeLocationForTMHMM(final Element att_type){
        Location aaLocation = null;

        try {
            // 1st, get the 'coords' and the 'count' of how many
            //
            List     att_scores = att_type.getChildren("ATT_SCORE");
            List countAndCoords = getCountAndCoords(att_scores);
            String       coords = (String) countAndCoords.get(1);
            int           count = -1;

            if(  coords == null ){
                return null;
            }

            try {
                count = Integer.parseInt((String) countAndCoords.get(0));
            }
            catch(NumberFormatException nfe){
                System.err.println("No 'count' provided for TMHMM - "+
                                   "return null");
                return null;
            }
            

            //System.err.println("makeLocationForTMHMM(): count  = "+count+"\n"+
            //                   "                        coords = "+coords);

            // 2nd, parse out coords
            aaLocation = getAALocation(count, coords);
        }
        catch(Exception exp){
            System.err.println("Problem with <ATTRIBUTE_TYPE> "+att_type+"\n"+
                               "makeLocationForTMHMM() could not create "+
                               "Location for TMHMM ");
            exp.printStackTrace();
        }
        
        return aaLocation;
    }

    /* <code>getCountAndCoords</code> returns in a <code>List</code>
     * 2 Strings, count and coords.
     *
     * NOTE: it will return null and null if no count or coords are
     *       supplied
     *
     * @param att_scores an <code>List</code> of objects of type
     * <code>Element</code>.
     *
     * @return a <code>List</code> containing 2 <code>String/<code> objects
     * count and coords.
     *
     */
    private List getCountAndCoords(List att_scores) throws Exception{
        List toReturn = new ArrayList();
        String  count = null;
        String coords = null;

        for(int x = 0; x < att_scores.size(); ++x){
            Element att_score = (Element) att_scores.get(x);
            String       desc = att_score.getAttributeValue("DESC");
            String       text = att_score.getText();

            if( "count".equals(desc) ){
                count = text;
            }
            else if( "coords".equals(desc) ){
                coords = text;
            }
        }

        /* Commented out - unable to check where the error lies so returns nulls
         * so code higher up can better report it.
         * 
        if( count == null || coords == null ){
            throw new Exception("ERROR: TMHMM did not have ATT_SCORE "+
                                "with DESC = to 'coords' and/or 'count'");
        }
        */
        toReturn.add(count);
        toReturn.add(coords);

        return toReturn;
    }

    /* <code>getAALocation</code> 
     *
     * @param count is the number of locations in the string coords
     * @param coords is the <code>String</code> containing the ':' seperated 
     * coords ie: '234-253' or '234-253:265-287:391-413'
     *
     * @return a <code>Location</code> which may contain sub locations
     *
     */
    private Location getAALocation(int count, String coords){

        RangeLocation rl = null; // To be returned...

        int from = 0;
        List  al = new ArrayList();

        for(int x = 0; x < count; ++x){
            int     colon = coords.indexOf(':', from);
            int     minus = coords.indexOf('-', from);
            String  first = null, second = null;

            if( colon == -1 ){
                first  = coords.substring(from, minus);
                second = coords.substring(minus+1, coords.length());
            }
            else {
                first  = coords.substring(from, minus);
                second = coords.substring(minus+1, colon);
            }

            rl = new RangeLocation(Integer.parseInt(first),
                                   Integer.parseInt(second));
            
            al.add(rl); 
            from = colon + 1; // for next loop iteration
        }

        Location loc = LocationTools.union(al);
        return loc;
    }

    /* <code>convertAA2DNA</code> take DNA and AA <code>Location</code>
     * and return the AA coords mapped into DNA space.
     *
     * The theory is we multiply the AA loc by 3 to get NA coords.
     * Mirror the coords if on negative strand
     * Then translate it by the start min location of the exon.
     *
     * Exons       -------- --------------      -----------
     * AA          -----------------
     * Results     -------- --------
     *
     * @param aaLocs is a compound <code>Location</code> with zero or more
     * sublocations
     * @param exonLocs is a compound <code>Location</code> with zero or more
     * sublocations
     *
     * @return a <code>Location</code>
     *
     */
    public Location convertAA2DNA(Location aaLocs,
                                  StrandedLocation exonLocs){
        // Assume 1 exon
        // assume no UTR!

        Iterator exonItty = exonLocs.blockIterator();
        Location     exon = (Location) exonItty.next();

        if( exonItty.hasNext() ){
            System.err.println("WARNING: 2+ exons found in convertAA2DNA "+
                               "- Ignoring and return null");
            return null;
        }

        // Build List of NA coords from AA coords
        Iterator aaItty = aaLocs.blockIterator();
        List         al = new ArrayList();

        while( aaItty.hasNext() ){
            Location   aaLoc = (Location) aaItty.next();
            RangeLocation rl = null;

            int naMin = (aaLoc.getMin() * 3);
            int naMax = (aaLoc.getMax() * 3);

            // For negative strand we have to mirror the coords.
            // The distance from END5 will become the distance from END3
            //
            if( exonLocs.getStrand().equals(StrandedFeature.NEGATIVE) ){
                int length = exon.getMax() - exon.getMin() + 1;

                //System.err.println("NEGATIVE: BEFORE: length = "+length+
                //                   ", naMin = "+naMin+", naMax = "+naMax);

                int tempMin = naMin;
                naMin       = length - naMax;
                naMax       = length - tempMin;

                //System.err.println("NEGATIVE: AFTER : length = "+length+
                //                   ", naMin = "+naMin+", naMax = "+naMax);
            }

            rl = new RangeLocation(naMin + exon.getMin(),
                                   naMax + exon.getMin());

            al.add(rl); 
        }


        Location loc = LocationTools.union(al);
        return loc;
    }

    /* <code>printLocations</code> 
     *
     * @param loc is a <code>Location</code> with zero or more sublocations
     *
     * @return a <code>String</code>
     *
     */
    public String printLocations(Location loc){
        Iterator   itty = loc.blockIterator();
        StringBuffer sb = new StringBuffer();

        while( itty.hasNext() ){
            Location l = (Location) itty.next();
            sb.append(loc.getMin()+"-"+loc.getMax()+",");
        }

        return sb.toString();
    }
}
