/*
 * Copyright (c) 2001  Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package uk.ac.sanger.psu.tigread;

import java.io.File;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.PrintStream;
import java.io.FileOutputStream;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.List;


import org.jdom.Document;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.EmblLikeFormat;
import org.biojava.bio.seq.io.SequenceFormat;
import org.biojava.bio.seq.io.SimpleSequenceBuilder;

import uk.ac.sanger.psu.tigread.*;

/**
 * <code>TigrMultiRead</code> reads in a list of filenames, expected to be XML files 
 * conforming to the TIGR DTD. The XML is validated on reading and a corresponding 
 * EMBL file is written into an output directory with a modified name. The code 
 * checks the size of output files in order to provide a warning to the curator 
 * of possible problems. 
 *
 * @author <a href="mailto:kdj@sanger.ac.uk">Keith James</a>
 * @author <a href="mailto:art@sanger.ac.uk">Adrian Tivey</a>
 */
public class TigrMultiRead {
   private File inFile = null;
   private PrintStream printStream = System.out;
   
   
   private static void usage() {
      System.out.println("fasttigr2embl output_dir");
      System.out.println("  with the files specified on stdin eg");
      System.out.println("find . -name \"*.xml\" -print | fasttigr2embl EMBL");
   }
   
   
   public static void main(String [] argv) {
      int count = 0;
      List problems = new ArrayList();
      
      if ( argv.length != 1) {
	 System.err.println("No output dir specified");
	 usage();
	 System.exit(-1);
      }
      
      
      try {
	 File f = new File(argv[0]);
	 if ( !f.exists()) {
	    f.mkdir();
	 }
	 
	 BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
	 
	 String fname;
	 while ((fname = in.readLine()) != null) {
	    count++;
	    
	    File f1 = new File(fname);
	    fname = fname.substring(0, fname.lastIndexOf(".")+1);
	    String tmpFname = fname;
	    File o = new File(f, fname+"embl");
	    PrintStream ps = new PrintStream(new FileOutputStream(o));
	    TigrMultiRead tr = new TigrMultiRead(f1, ps);
	    tr.go();
	    ps.close();
	    
	    if (!o.exists() || o.length() == 0) { 
		   problems.add(tmpFname);
	    }
	    
	    
	 }
      }
      catch(Exception e){
	   e.printStackTrace();
      }
      
      
      if (count ==0) {
	 System.err.println("No files processed");
	   usage();
	 System.exit(-1);
      }
      
      System.err.println("Processed "+count+" files");
	if (problems.size() > 0) {
	   for (int i=0; i < problems.size(); i++) {
	      System.err.println("WARNING: "+problems.get(i)+" is zero-size. Translation probably failed");
	   }					  
	}
   }
   
   
   public TigrMultiRead(File file, PrintStream ps) throws IOException {
      this.inFile = file;
      this.printStream = ps;
   }
   

   public void go() {
      try {
	 SAXBuilder builder =
	   new SAXBuilder("org.apache.xerces.parsers.SAXParser", true);
	 
	 // JDOM document built from the SAX2 parse
	 Document document = builder.build(inFile);
	 
	 TigrSequenceLoader loader = new TigrSequenceLoader();
	 // Extract elements from the JDOM tree into the SequenceBuilder
	 // 
	 Sequence seq = loader.loadSequence(document,
					    new SimpleSequenceBuilder());
	 
	 SequenceFormat eFormat = new EmblLikeFormat();
	 eFormat.writeSequence(seq, printStream);
      }
      catch (Exception exp) {
	 exp.printStackTrace();
      }
   }

}
