/*
 * Copyright (c) 2001  Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package uk.ac.sanger.psu.tigread;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;

import org.jdom.Document;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.EmblLikeFormat;
import org.biojava.bio.seq.io.SequenceFormat;
import org.biojava.bio.seq.io.SimpleSequenceBuilder;

/**
 * <code>TigrRead</code> reads in sequence and annotation as XML
 * conforming to the TIGR DTD. The XML is validated on reading and the
 * read will abort if the parser encounters invalid data. The XML is
 * temorarily converted to a JDOM tree before writing to an EMBL
 * flatfile.
 *
 * @author <a href="mailto:kdj@sanger.ac.uk">Keith James</a>
 */
public class TigrRead
{
    File             inFile = null;       // contructor sets this
    PrintStream printStream = System.out; // default

    public static void main(String [] argv)
    {
        if (argv.length != 1)
        {
            System.err.println("Usage: TigrRead <XML URI>");
            System.exit(0);
        }

        try {
            TigrRead tr = new TigrRead(argv[0]);
            tr.go();
        }
        catch(Exception e){
            e.printStackTrace();
        }
    }

    public TigrRead(String fileName, PrintStream ps) throws IOException
    {
        this(fileName);
        this.printStream = ps;
    }

    /*
     */
    public TigrRead(String fileName) throws IOException
    {
        File inFile = new File(fileName);

        if (!inFile.exists() || inFile.length() == 0) {
            throw new IOException("XML file '"+fileName+"' not found "+
                                  "or empty");
        }

        this.inFile = inFile;
    }

    /*
     */
    public void go()
    {
        try {
            SAXBuilder builder =
                new SAXBuilder("org.apache.xerces.parsers.SAXParser", true);

            // JDOM document built from the SAX2 parse
            Document document = builder.build(inFile);

            TigrSequenceLoader loader = new TigrSequenceLoader();
            // Extract elements from the JDOM tree into the SequenceBuilder

            Sequence seq = loader.loadSequence(document,
                                               new SimpleSequenceBuilder());

            SequenceFormat eFormat = new EmblLikeFormat();
            eFormat.writeSequence(seq, printStream);
        }
        catch (JDOMException jde)
        {
            jde.printStackTrace();
        }
        catch (IOException ioe)
        {
            ioe.printStackTrace();
        }
        catch (BioException be)
        {
            be.printStackTrace();
        }
    }
}
