/*
 * Copyright (c) 2001  Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package uk.ac.sanger.psu.tigread;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;

import org.jdom.Document;
import org.jdom.DocType;
import org.jdom.Element;
import org.jdom.JDOMException;

import org.biojava.bio.BioException;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.StrandedFeature;
import org.biojava.bio.seq.io.EmblLikeFormat;
import org.biojava.bio.seq.io.ParseException;
import org.biojava.bio.seq.io.SeqIOListener;
import org.biojava.bio.seq.io.SequenceBuilder;
import org.biojava.bio.seq.io.SequenceFormat;
import org.biojava.bio.seq.io.StreamParser;
import org.biojava.bio.seq.io.SymbolTokenization;
import org.biojava.bio.symbol.Alphabet;
import org.biojava.bio.symbol.FiniteAlphabet;

import org.biojava.utils.ChangeVetoException;

/**
 * <code>TigrSequenceLoader</code> extracts specific sequence and
 * annotation data from a JDOM <code>Document</code> which has been
 * generated from XML conforming to the TIGR DTD.
 *
 * @author <a href="mailto:kdj@fes1.sanger.ac.uk">Keith James</a>
 */
public class TigrSequenceLoader
{
    public static final int          TU = 0;
    public static final int       MODEL = 1;
    public static final int        EXON = 2;
    public static final int         RNA = 3;
    public static final int        MISC = 4;
    public static final int      REPEAT = 5;
    public static final int      SOURCE = 6;
    public static final int   SIGPTMHMM = 7; // SigP predited by TMHMM
    public static final int      SIGPNN = 8; // SigP predicted by neural net
    public static final int         TMM = 9;
    public static final boolean verbose = false;

    public static final String     SIGNALPNN = "SP-NN,signalP-2.0";
    public static final String  SIGNALPTMHMM = "SP-HMM,signalP-2.0";
    public static final String         TMHMM = "GES,TMHMM2.0";

    private EmblFeatureTemplateFactory templateFactory;

    public TigrSequenceLoader()
    {
        templateFactory = new EmblFeatureTemplateFactory();
    }

    /**
     * The <code>loadSequence</code> method extracts data from a JDOM
     * <code>Document</code> which has been generated from XML
     * conforming to the TIGR DTD and passes it to a
     * <code>SequenceBuilder</code> which returns a
     * <code>Sequence</code>.
     *
     * Assumption: only one <model> per TU.
     *
     * @param document a <code>Document</code>.
     * @param builder a <code>SequenceBuilder</code>.
     *
     * @return a <code>Sequence</code>.
     *
     * @exception BioException if an error occurs.
     */
    public Sequence loadSequence(final Document        document,
                                 final SequenceBuilder builder)
        throws BioException
    {
        DocType docType = document.getDocType();

        if( verbose ){
            System.err.println("DocType: "   + docType.getElementName());
            System.err.println("Public ID: " + docType.getPublicID());
            System.err.println("System ID: " + docType.getSystemID());
        }

        Element    root = document.getRootElement();
        List assemblies = null;

        assemblies = root.getChildren("ASSEMBLY");

        if( verbose ){
            System.err.println("Document contains "
                               + assemblies.size()
                               + " top level assemblies");
        }

        if (assemblies.size() == 0)
        {
            List pseudoChromosomes = root.getChildren("PSEUDOCHROMOSOME");

            if( verbose ){
                System.err.println("Document contains "
                                   + pseudoChromosomes.size()
                                   + " pseudochromosomes");
            }

            if (pseudoChromosomes.size() == 0)
            {
                System.out.println("Document is empty!");
                throw new RuntimeException("Document is empty!");
            }

            Element pseudoChromosome = (Element) pseudoChromosomes.get(0);
            assemblies = pseudoChromosome.getChildren("ASSEMBLY");
        }

        if( verbose ){
            System.err.println("Pseudochromosome contains "
                               + assemblies.size()
                               + " assemblies");
        }

        Element assembly = (Element) assemblies.get(0);

        // WARNING: IS THIS VARIABLE EVER USED?????????????????????????
        List rnaGeneList = assembly.getChild("GENE_LIST")
            .getChild("RNA_GENES")
            .getChildren();

        // Add ID direct to sequence
        //
        // FIXME: ID is hardcoded to be 'Tbrucei.chrX'
        //
        String organism =
            assembly.getChild("HEADER").getChild("ORGANISM").getText();
        String  asmblId = assembly.getChildText("ASMBL_ID");
        String      chr = assembly.getAttributeValue("CHROMOSOME");
        String       id = null;

        if (chr == null || "".equals(chr)) {
            chr = "unknown";
        }

        if ("Trypanosoma brucei".equals(organism)) {
            id = "Tbrucei.chr"+chr+"."+asmblId;
        }
        else if ("Trypanosoma cruzi".equals(organism)) {
            id = "Tcruzi.chr"+chr+"."+asmblId;
        }
        else {
            id = "chr"+chr+"."+asmblId;
        }

        builder.addSequenceProperty("ID", id);
        builder.addSequenceProperty("AC", "unknown");
        builder.addSequenceProperty("FH", "Key             Location/Qualifiers");

        try
        {
            // Inform builder of start of sequence
            builder.startSequence();

            FiniteAlphabet alpha = DNATools.getDNA();

            parseSequenceTokens(assembly.getChildText("ASSEMBLY_SEQUENCE"),
                                alpha,
                                builder);

            // Put in a fake source feature to hold the Assembly ID
            StrandedFeature.Template st =
                templateFactory.makeFeatureTemplate(assembly, SOURCE);

            // Start source
            builder.startFeature(st);

            // End source
            builder.endFeature();

            // Protein coding genes
            List protCodingList = assembly.getChild("GENE_LIST")
                .getChild("PROTEIN_CODING")
                .getChildren("TU");
	    
            for (Iterator pci = protCodingList.iterator(); pci.hasNext();)
            {
                Element tu = (Element) pci.next();

                // The TU and MODEL features are currently not shown
                // explicitly as the result became too confusing when
                // viewed in Artemis. Their presence is implied by the
                // EXON feature positions. As such it is essential
                // that there is only one gene model.

                StrandedFeature.Template et =
                    templateFactory.makeFeatureTemplate(tu, EXON);

                builder.startFeature(et); // Start EXON

                builder.endFeature();     // End EXON

                // The EXON features in the XML also have UTRs
                //
                // The PSU may not be interested in UTRs so this code maybe
                // removed, it is here as an exmaple/reminder only if
                // it needs to be implemented.
                /*
                List utrList = templateFactory.makeFeatureTemplate(tu, UTR);

                for (Iterator utri = utrList.iterator(); utri.hasNext();)
                {
                    StrandedFeature.Template utrT =
                        (StrandedFeature.Template) utri.next();
                    
                    builder.startFeature(utrT);
                    builder.endFeature(utrT);
                }
                */
            }

            // RNA coding genes
            List rnaCodingList = assembly.getChild("GENE_LIST")
                .getChild("RNA_GENES")
                .getChildren();

            for (Iterator rci = rnaCodingList.iterator(); rci.hasNext();)
            {
                Element rnaGene = (Element) rci.next();

                StrandedFeature.Template rt =
                    templateFactory.makeFeatureTemplate(rnaGene, RNA);

                // Start RNA feature
                builder.startFeature(rt);

                builder.endFeature();
            }

            // misc_features are optional
            // NOTE: no MISC_INFO elements in last TIGR XML
            Element miscInfo = assembly.getChild("MISC_INFO");

            if (miscInfo != null)
            {
                List miscFeatureList = miscInfo.getChildren();

                for (Iterator mfi = miscFeatureList.iterator(); mfi.hasNext();)
                {
                    Element miscFeature = (Element) mfi.next();

                    StrandedFeature.Template mft =
                        templateFactory.makeFeatureTemplate(miscFeature, MISC);

                    // Start misc_feature
                    builder.startFeature(mft);

                    builder.endFeature();
                }
            }

            // repeats are optional
            Element repeatList = assembly.getChild("REPEAT_LIST");

            if (repeatList != null)
            {
                List repeats = repeatList.getChildren();

                for (Iterator ri = repeats.iterator(); ri.hasNext();)
                {
                    Element repeat = (Element) ri.next();

                    StrandedFeature.Template rt =
                        templateFactory.makeFeatureTemplate(repeat, REPEAT);

                    // Start the repeat
                    builder.startFeature(rt);

                    builder.endFeature();
                }
            }

            /* SignalP & TMHMM
             * Loop over same protCodingList as above that made exons (CDS)
             */            
            for (Iterator pci = protCodingList.iterator(); pci.hasNext(); ){
                Element tu = (Element) pci.next();

                Element   model        = tu.getChild("MODEL");
                ArrayList similarities = new ArrayList();
                ArrayList domains      = new ArrayList();
                ArrayList features     = new ArrayList();

                if( model.getChild("MODEL_ATTRIBUTE") != null ){
                    Element model_attribute =
                        model.getChild("MODEL_ATTRIBUTE");
                    List    attribute_types =
                        model_attribute.getChildren("ATTRIBUTE_TYPE");

                    for(int x = 0; x < attribute_types.size(); ++x){
                        Element attribute_type =
                            ((Element)attribute_types.get(x));
                        String  method         =
                            attribute_type.getAttributeValue("METHOD");

                        StrandedFeature.Template feat = null;

                        if( SIGNALPTMHMM.equals(method) ){
                            feat =
                            templateFactory.makeFeatureTemplate(model,
                                                                attribute_type,
                                                                SIGPTMHMM);
                        }
                        else if( TMHMM.equals(method) ){
                            feat = 
                            templateFactory.makeFeatureTemplate(model,
                                                                attribute_type,
                                                                TMM);
                       }

                        if( feat != null ){
                            builder.startFeature(feat);
                            builder.endFeature();
                        }                   
                    }
                }
            }

            // Inform builder of end of sequence
            builder.endSequence();

        }
        catch (ParseException pe)
        {
            pe.printStackTrace();
        }

        return builder.makeSequence();
    }

    /**
     * <code>parseSequenceTokens</code> converts the sequence string
     * obtained from the XML document into symbols and informs the
     * builder.
     *
     * @param tokenString a <code>String</code>.
     * @param alpha a <code>FiniteAlphabet</code>.
     * @param builder a <code>SeqIOListener</code>.
     *
     * @exception BioExceptionException if an error occurs.
     */
    private void parseSequenceTokens(final String         tokenString,
                                     final FiniteAlphabet alpha,
                                     final SeqIOListener  builder)
        throws BioException
    {
        SymbolTokenization tokenParser = alpha.getTokenization("token");

        StreamParser strParser = tokenParser.parseStream(builder);

        // NOTE: A NEW LINE CHARACTER IS OFTEN AT THE START OF THE SEQUENCE
        //       HENCE THE USE OF TRIM BEFORE BIOJAVA COMPLAINS
        char [] tokenChars = tokenString.trim().toCharArray();


        strParser.characters(tokenChars, 0, tokenChars.length);
    }
}
