/*
 * Copyright (c) 2003  Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package uk.ac.sanger.psu.tigrmerge;

import uk.ac.sanger.psu.tigrutils.*;

import java.util.Iterator;
import java.util.Set;
import java.util.Collection;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Properties;
import java.util.NoSuchElementException;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import org.biojava.bio.Annotation;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.StrandedFeature;
import org.biojava.bio.seq.FeatureFilter;
import org.biojava.bio.seq.FeatureFilter.Not;
import org.biojava.bio.seq.FeatureHolder;
import org.biojava.bio.seq.impl.SimpleSequence;
import org.biojava.utils.ChangeVetoException;
import org.biojava.bio.symbol.Location;
import org.biojava.bio.symbol.Edit; // TO BE REMOVED?

import org.biojava.bio.symbol.SymbolList;

import org.apache.commons.lang.StringUtils;

import java.lang.Exception;


/**
 * <code>TigrMerge</code> reads in 2 EMBL files, the original EMBL file with
 * added annotation and the new EMBL file. The theory is that the features
 * can be matched up via gene names in the following order PUB_LOCUS, LOCUS,
 * FEAT_NAME before doing and sequence comparisons.
 *
 * TO BE DONE;
 *
 * doing an exact sequence match for features that have no matching feature
 * from the manual annotation (orig) file. Of course check that there are
 * features left in orig to "copy over".
 *
 * Comments need to have params added with the return types documented.
 *
 * @author <a href="mailto:pjm@sanger.ac.uk">Paul Mooney</a>
 */
public class TigrEmblMerger
{
    static final String     PROPERTY_MAP_FILE = "annotation_map.properties";
    static final Properties annotationMap     = new Properties();
    private Sequence origSeq, newSeq, mergedSeq;
    private boolean addCDSFeatures = false;

    private List  reportAdditions = new ArrayList();
    private List  reportDeletions = new ArrayList();
    private List reportLocChanges = new ArrayList();
    

    Map origFeatNameHash = null;
    Map origPubLocusHash = null;
    Map    origLocusHash = null;
    Map  newFeatNameHash = null;
    Map  newPubLocusHash = null;
    Map     newLocusHash = null;
    static final String PROCESSED_KEY = "feature_is_processed";
    static final int    HEADER_WIDTH  = 80;

    TigrEmblMerger (Sequence origSeq, Sequence newSeq, boolean addCDSFeatures)
    {
        this.origSeq        = origSeq;
        this.newSeq         = newSeq;
        this.addCDSFeatures = addCDSFeatures;

        try
        {
            annotationMap.load(new FileInputStream(PROPERTY_MAP_FILE));
        }
        catch (FileNotFoundException fnfe)
        {
            System.err.println("**********************************************"+
                               "**********************************");
            System.err.println("ERROR: Unable to find annotation properties file: "
                               + fnfe.getMessage());
            System.exit(1);
        }
        catch (IOException ioe)
        {
            System.err.println("**********************************************"+
                               "**********************************");
            System.err.println("ERROR: Unable to read annotation properties file: "
                               + ioe.getMessage());
            System.err.println("**********************************************"+
                               "**********************************");
            System.exit(1);
        }
    }

    /* <code>merge</code>
     *
     * <OL>
     * <LI> Setup the hash tables for the IDs
     * <LI> Make the new merged sequence
     * <LI> Loop through all features on merged sequence
     * <OL>
     *   <LI> Find "old" feature
     *   <LI> compare annotation
     *   <LI> transfer from old to new added qualifiers
     *   <LI> Report all conflicts in annotation 
     * </OL>
     * </OL>
     */
    public Sequence merge ()
    {
        extractIds();

        /* Make new mergeSeq from copy of newSeq.
         * newSeq may have no DNA sequence (a sparse EMBL file as generated
         * in data exchanges with TIGR), use oldSeqs instead
         *
         * THIS NEEDS TO BE DONE IN TigrMerge.java, Sequences ARE IMUTABLE... 
         */
        SymbolList symbolList = (SymbolList) newSeq;

        if (newSeq.length() == 0) {
            symbolList = (SymbolList) origSeq;
        }

        mergedSeq = new SimpleSequence(symbolList,
                                       "URN",
                                       "Merged_Sequence",
                                       newSeq.getAnnotation());

        //copySequenceIfNeeded(); // Sparse EMBL files may not have sequence, use old seqs sequence

        Iterator itty = newSeq.features();

        try {
            while(itty.hasNext()){
                StrandedFeature sf = (StrandedFeature) itty.next();
                
                //dumpFeature(f);

                mergedSeq.createFeature(sf.makeTemplate());
            }

            mergeFeatures();

            addExtraSangerFeatures();
        }
        catch(BioException be){
            System.err.println("**********************************************"+
                               "**********************************");
            be.printStackTrace();
            System.err.println("**********************************************"+
                               "**********************************");
            System.exit(1);
        }
        catch(ChangeVetoException cve){
            System.err.println("**********************************************"+
                               "**********************************");
            cve.printStackTrace();
            System.err.println("**********************************************"+
                               "**********************************");
            System.exit(1);
        }

        displayReports();
 
        return mergedSeq;
    }

    /* <code>displayReports</code> displays gene boundry changes, addtions and
     * deletions
     *
     */
    private void displayReports() {
        printHeader(HEADER_WIDTH, "REPORTS");

        // Boundry Change
        System.err.println("Gene Boundry Changes\n--------------------");

        for(int i = 0; i < reportLocChanges.size(); ++i){
            System.err.println("  "+reportLocChanges.get(i));
        }

        // Deletions
        System.err.println("\nDeletions\n--------------------");

        for(int i = 0; i < reportDeletions.size(); ++i){
            System.err.println("  "+reportDeletions.get(i));
        }

        // Additions
        System.err.println("\nAdditions\n--------------------");

        for(int i = 0; i < reportAdditions.size(); ++i){
            System.err.println("  "+reportAdditions.get(i));
        }

        // Report seq length and feature count for the 3 files
        System.err.println("\nSummary\n-------");
        System.err.println(displayNice("", 12, true)+
                           displayNice("Original", 10, false)+
                           displayNice("New",      10, false)+
                           displayNice("Merged",   10, false));

        System.err.println(displayNice("Length:", 12, true) +
                           displayNice(origSeq.length(),   10)+
                           displayNice(newSeq.length(),    10)+
                           displayNice(mergedSeq.length(), 10));

        System.err.println(displayNice("Features:", 12, true) +
                           displayNice(origSeq.countFeatures(),   10)+
                           displayNice(newSeq.countFeatures(),    10)+
                           displayNice(mergedSeq.countFeatures(), 10));

    }

    /* <code>printHeader</code> prints something like this;
     *
     * **********************
     *         text
     * *********************
     *
     * @param width holds the character width of the whole header
     * @param text is a <code>String</code>  - the one line of text to be 
     * centered
     */
    private void printHeader(int width, String text) {
        System.err.println("");
        System.err.println(StringUtils.repeat("*", width));
        System.err.println(StringUtils.center(text, width, " "));
        System.err.println(StringUtils.repeat("*", width));
        System.err.println("");
    }


    /* <code>dislayNice</code> is a noddy way of doing sprintf("%-10s", arg);
     *
     * @param s is the String to output
     * @param size - the number of characters to take up on the screen
     * @param b is a boolean flag, true is left justified, false is right
     * justified
     */
    private String displayNice(String s, int size, boolean b){
        StringBuffer sb = new StringBuffer(s);

        if( size < s.length() ){
            return s;
        }

        int numSpaces = size - s.length();

        for(int i = 0; i < numSpaces; ++i){
            if( b ){
                sb.append(' ');
            }
            else {
                sb.insert(0, ' ');
            }

            //System.err.println("i = "+i+", "+sb);
        }

        return sb.toString();
    }
    /* <code>dislayNice</code> is a noddy way of doing sprintf("%-10d", arg);
     */
    private String displayNice(int inty, int size){
        StringBuffer sb = new StringBuffer();
        sb.append(inty);

        if( size < sb.length() ){
            return sb.toString();
        }


        int numSpaces = size - sb.length();

        for(int i = 0; i < numSpaces; ++i){
            sb.insert(0, " ");
            //System.err.println("i = "+i+", "+sb);
        }

        return sb.toString();
    }


    /* <code>extractIds</code> simply sets up the hashes that identify
     * each feature.
     */
    private void extractIds(){
        System.err.println(displayNice("Parsing Sanger EMBL file (parameter 1)", 40, true));
        IdExtractor ide1 = new IdExtractor(origSeq);
        ide1.extract();


        /*
         *
         * DO I NEED TO GET FEAT NAMES FROM NEW, SINCE IT WILL BE COPIED
         * INTO MERGED VERSION ANYWAY???????????????????????????????????
         *
         */


        origPubLocusHash = ide1.getPubLocusHash();
        origLocusHash    = ide1.getLocusHash();
        origFeatNameHash = ide1.getFeatNameHash();

        ide1 = null; // Object no longer needed.

        System.err.println(displayNice("Parsing TIGR EMBL file (parameter 2)", 40, true));

        IdExtractor ide2 = new IdExtractor(newSeq);
        ide2.extract();

        newPubLocusHash = ide2.getPubLocusHash();
        newLocusHash    = ide2.getLocusHash();
        newFeatNameHash = ide2.getFeatNameHash();

        ide2 = null; // Object no longer needed.

    }

    public class IdExtractor
    {
        Sequence seq;
        HashMap featNameHash = new HashMap();
        HashMap pubLocusHash = new HashMap();
        HashMap    locusHash = new HashMap();

        public IdExtractor(Sequence seq) {
            this.seq = seq;
        }

        public void extract(){
            FeatureHolder features = getChangableFeatures(seq);
            Iterator        featIt = features.features();

            while(featIt.hasNext()){
                //System.err.println("*** Feature ***");

                StrandedFeature feat = (StrandedFeature) featIt.next();

                //String pubLocus = getUniqueAnno4TigrElement("PUB_LOCUS", feat);
                //String    locus = getUniqueAnno4TigrElement("LOCUS",     feat);
                //String   featName = getUniqueAnno4TigrElement("FEAT_NAME", feat);
                String[] pubLocus = getManyNames4TigrElement("PUB_LOCUS", feat);
                String[]    locus = getManyNames4TigrElement("LOCUS",     feat);
                String[] featName = getManyNames4TigrElement("FEAT_NAME", feat);

                /*
                if( pubLocus != null ){
                    pubLocusHash.put(pubLocus, feat);
                }

                if( locus != null ){
                    locusHash.put(locus, feat);
                }

                if( featName != null ){
                    featNameHash.put(featName, feat);
                }
                */

                if( pubLocus.length > 0 ){
                    for(int x = 0; x < pubLocus.length; ++x){
                        if( pubLocusHash.get(pubLocus[x]) != null ){
                            System.err.println("WARNING: PUB_LOCUS '"+pubLocus[x]+"' "+
                                               "appears in more than one feature;"+
                                               pubLocusHash.get(pubLocus[x])+";");
                            dumpFeature(feat);
                        }
                        else {
                            pubLocusHash.put(pubLocus[x], feat);
                        }
                    }
                }

                if( locus.length > 0 ){
                    for(int x = 0; x < locus.length; ++x){

                        //System.err.println("MSG: LOCUS = '"+locus[x]+"'");

                        if( locusHash.get(locus[x]) != null ){
                            System.err.println("WARNING: LOCUS '"+locus[x]+"' "+
                                               "appears in more than one feature;"+
                                               pubLocusHash.get(locus[x])+";");
                            dumpFeature(feat);
                        }
                        else {
                            locusHash.put(locus[x], feat);
                        }
                    }
                }

                if( featName.length > 0 ){
                    for(int x = 0; x < featName.length; ++x){
                        if( featNameHash.get(featName[x]) != null ){
                            System.err.println("WARNING: FEAT_NAME '"+featName[x]+"' "+
                                               "appears in more than one feature;"+
                                               featNameHash.get(featName[x])+";");
                            dumpFeature(feat);
                        }
                        else {
                            featNameHash.put(featName[x], feat);
                        }
                    }
                }

                if( featName.length == 0 &&
                        pubLocus.length == 0 &&
                        locus.length == 0 ){

                    if (! "variation".equals(feat.getType())) {

                        System.err.println("Could not get a name for the "+
                                           "feature: "+feat.getType()+ " "+feat.getLocation());
                        //dumpFeature(feat);
                    }
                }
            }  
        }

        public Map getPubLocusHash(){
            return (Map) pubLocusHash;
        }

        public Map getLocusHash(){
            return (Map) locusHash;
        }

        public Map getFeatNameHash(){
            return (Map) featNameHash;
        }
    }

    /* <code>getUniqueAnno4TigrElement</code> takes the XML element name, 
     * maps it into the qualifier specified in the annotation_map.properties
     * file and returns a String value.
     *
     * It will spit out an error and exit if more than one qualifier exists
     * in the EMBL file.
     */
    private String getUniqueAnno4TigrElement(String tigrElement,
                                             StrandedFeature f)
    {
        String emblQualifier = null;

        try
        {
            emblQualifier = annotationMap.getProperty(tigrElement);
        }
        catch (NoSuchElementException nse) { }

        // We may get a null if the Property doesn't exist
        String key = (emblQualifier == null) ? tigrElement : emblQualifier;

        String [] vals = getAnnotationValue(emblQualifier,
                                            f.getAnnotation());

        if( vals.length > 1 ){
            System.err.println("Error: Feature has more than one key "+
                               "qualifier in original file");
            dumpFeature(f);
            System.exit(1);
        }

        String toReturn = null;
        
        if (vals.length == 1){
            toReturn = vals[0];
        }
    
        return toReturn;
    }

    /* <code>getBestNameForFeature</code> tries PUB_LOCUS, LOCUS and FEAT_NAME
     * and returns a _single_ name, the first.
     *
     */
    private String getBestNameForFeature(StrandedFeature f)
    {
        String[] names = getManyNames4TigrElement("PUB_LOCUS", f);

        if (names.length == 0) {
            names = getManyNames4TigrElement("LOCUS", f);

            if (names.length == 0) {
                names = getManyNames4TigrElement("FEAT_NAME", f);
            }
        }

        if (names == null) {
            return null;
        }

        if (names.length == 0) {
            return null;
        }

        return names[0];
    }

    /* <code>getManyNames4TigrElement</code> takes the XML element name, 
     * maps it into the qualifier specified in the annotation_map.properties
     * file and returns a String[] of values.
     *
     *
     *
     * @return String[] - which maybe zero length but not null.
     */
    private String[] getManyNames4TigrElement(String tigrElement,
                                              StrandedFeature f)
    {
        String emblQualifier = null;

        try
        {
            emblQualifier = annotationMap.getProperty(tigrElement);
        }
        catch (NoSuchElementException nse) { }

        // We may get a null if the Property doesn't exist
        String key = (emblQualifier == null) ? tigrElement : emblQualifier;

        String [] vals = getAnnotationValue(emblQualifier,
                                            f.getAnnotation());
        return vals;
    }


    /**
     * <code>getAnnotationValue</code> extracts the annotation
     * value(s) for a particular Annotation key and converts them to
     * an array of Strings.
     *
     * @param key a <code>String</code>.
     * @param a an <code>Annotation</code>.
     *
     * @return a <code>String []</code> array.
     */
    private String [] getAnnotationValue(final String key, final Annotation a)
    {
        List matches = new ArrayList();

        for (Iterator ai = a.keys().iterator(); ai.hasNext();)
        {
            Object thisKey = ai.next();

            // From EMBL, all values are either Strings, Collections
            // of Strings or Boolean
            if (thisKey.equals(key))
            {
                Object value = a.getProperty(thisKey);

                // To cope with /pseudo, /partial etc.
                if (Boolean.class.isInstance(value))
                {
                    matches.add(value.toString());
                }
                else if (Collection.class.isInstance(value))
                    matches.addAll((Collection) value);
                else
                    matches.add(value);
            }
        }

        String [] toReturn = null;

        try {
            toReturn = (String []) matches.toArray(new String [0]);
        }
        catch(java.lang.ArrayStoreException ase){
            System.err.println("ERROR: tried to convert to String[] for key '"+
                               key+"'");
        }

        return toReturn;
    }

    /* <code>copySequenceIfNeeded</code> copies teh actual DNA sequence from
     * the old to the new sequence objects.
     *
     */
    private void copySequenceIfNeeded() {
        if (mergedSeq.length() == 0) {
            try {
                Edit ed = new Edit(1, 0, origSeq);
                mergedSeq.edit(ed);
                System.out.println("Sequence copied over.");
            }
            catch (Exception e) {
                e.printStackTrace();
                System.exit(1);
            }
        }
        else {
            System.out.println("Sequence NOT copied over.");
        }
    }



    /* <code>mergeFeatures</code>
     * Interate through all features on merged sequence
     * <OL>
     *   <LI> Remove feature if it has a DELETED qualifier (remove from both 
     *        data sources!)
     *   <LI> Find "old" feature
     *   <LI> compare annotation
     *   <LI> transfer from old to new added qualifiers
     *   <LI> Report all conflicts in annotation 
     * </OL>
     *
     * NOTE: TIGR may have added a feature, say a CDS, that Sanger does not yet
     * know about. This is fine. If Sanger have generated a new feature it
     * needs to be found and added. This is necessary because only certain
     * features make the round trip. This is done after mergefeatures() is called
     * and is done by addExtraSangerFeatures()
     */
    private void mergeFeatures(){

        FeatureHolder features = getChangableFeatures(mergedSeq);
        Iterator          itty = features.features();
        int          col1Width = 30;

        printHeader(HEADER_WIDTH, "MERGE FEATURES");

        System.err.println(displayNice("New (from TIGR)", col1Width, true)+
                           "Curated EMBL");
        System.err.println("-------------------------------------------------");

        try {
            while(itty.hasNext()){
                StrandedFeature sfMerged = (StrandedFeature) itty.next();
                System.err.print(displayNice(sfMerged.getType()+" "+
                                 sfMerged.getLocation(), col1Width, true));

                // get old feature
                StrandedFeature sfOld = getOldFeatureFor(sfMerged);

                if( sfOld != null ){
                    System.err.println(sfOld.getType()+" "+
                                       sfOld.getLocation());
                }

                if (featureIsDeleted(sfMerged, sfOld)) {
                    System.err.println("    *** DELETED ***");
                    reportDeletions.add(getBestNameForFeature(sfMerged));
                    continue;
                }
                else {
                    if( sfOld != null ){
                        transferAnnotation(sfMerged, sfOld);
                        postProcessAnnoTransfer(sfMerged, sfOld);
                        markAsProcessed(sfOld);
                    }
                    else {
                        // LOOK FOR ADDTIONAL CDSs here!!!!!!!!!!!!!!!!!!!!!!!!!

                        String bestName = getBestNameForFeature(sfMerged);

                        reportAdditions.add(sfMerged.getType()+" : "+bestName);

                        System.err.println("New feature or one wihout a "+
                                           "PUB_LOCUS, LOCUS, FEAT_NAME etc");
                        
                    }
                }
            }
        }
        catch(Exception e){
            System.err.println("**********************************************"+
                               "**********************************");
            System.err.println("ERROR: mergefeatures() caught an exception.");
            e.printStackTrace();
            System.err.println("**********************************************"+
                               "**********************************");
            System.exit(1);
        }
    }

    /* <code>featureIsDeleted</code> returns true if the feature has a /DELETED
     * qualifier. It also calls removefeature() on both sequences.
     *
     */

    private boolean featureIsDeleted(StrandedFeature sf, StrandedFeature oldSf){
        Annotation anno = sf.getAnnotation();

        if (anno.containsProperty("DELETED")) {
            try {
                FeatureHolder holder = sf.getParent();
                holder.removeFeature(sf);

                if (oldSf != null) {
                    FeatureHolder oldHolder = oldSf.getParent();
                    oldHolder.removeFeature(oldSf);

                    deleteAssociatedFeats(oldHolder, anno);
                }

                return true;
            }
            catch(ChangeVetoException cve) {
                System.err.println("**********************************************"+
                                   "**********************************");
                System.err.println("\nERROR: The Feature could not be delete;");
                cve.printStackTrace();
                System.err.println("**********************************************"+
                                   "**********************************");
                System.exit(1);
            }
        }

        return false;
    }

    /* <code>deleteAssociatedFeats</code> looks for misc_features and tries
     * to find ones that have "predicted for <feat_name>". It them removes
     * them.
     *
     *
     * @param holder is a <code>FeatureHolder</code> on which the misc features 
     * reside.
     * @param anno is an <code>Annotation</code> object of the feature that
     * has/will be deleted. Use it to find out is name.
     */

    private void deleteAssociatedFeats(FeatureHolder holder, Annotation anno){

        String sysId = "", tempSysId = "";

        if (anno.containsProperty("systematic_id")){
            sysId = (String) anno.getProperty("systematic_id");
        }

        if (anno.containsProperty("temporary_systematic_id")){
            tempSysId = (String) anno.getProperty("temporary_systematic_id");
        }

        // A TMHMM or singalP will have a note field with
        // something like 'predicted for SPAC212.08c'
        //    OR          'predicted for SPAC212.08c by TMHMM2.0....'
        // on the old sequence.
        boolean look = true;

        try {
            // Each time we remove something from the holder, the iterator
            // will become "bad" and we will get an exception
            // ConcurrentModificationException doing a next() on the underlying
            // List.
            while (look) {
                Iterator    itty = holder.features();
                boolean notFound = true;

                while(itty.hasNext()){
                    StrandedFeature sf2 = (StrandedFeature) itty.next();

                    if (!"misc_feature".equals(sf2.getType()))
                        continue;

                    Annotation        a = sf2.getAnnotation();

                    if (a.containsProperty("note")){
                        String  note = "";
                        Object value = a.getProperty("note");



                        // To cope with /pseudo, /partial etc.
                        if (Boolean.class.isInstance(value))
                            value = "";
                        else if (Collection.class.isInstance(value)) {
                            System.err.println("\nCollection for note: feature"+sf2);

                            Collection     c = (Collection) value;
                            Iterator   noteI = c.iterator();
                            StringBuffer  sb = new StringBuffer();

                            while (noteI.hasNext()) {
                                sb.append((String) noteI.next());
                                sb.append(" ");
                            }

                            note = sb.toString();

                            System.err.println("note now equals: '"+note+"'");
                        }
                        else
                            note = (String) value;




                        /*
                        if (a.getProperty("note") instanceof String) {
                            note = (String) a.getProperty("note"); // not always a String...
                        }
                        else {

                        }
                        */

                        int lastIndex = 
                            StringUtils.lastIndexOf(note, "predicted for ");

                        if (lastIndex != -1) {
                            String geneName =
                                StringUtils.substringBetween(note,
                                                             "predicted for ",
                                                             " ");
                            /*
                            System.err.println("BEFORE genename = "+geneName+
                                               ", note = "+note);
                            */

                            if(geneName == null || "".equals(geneName)) {
                                geneName =
                                    StringUtils.substringAfter(note,
                                                               "predicted for ");
                            }

                            /*
                            System.err.println("AFTER  genename = "+geneName+
                                               ", sysId = "+sysId+
                                               ", tempSysID = "+tempSysId+
                                               ", note = "+note);
                            */
                            if (sysId.equals(geneName) || tempSysId.equals(geneName)) {
                                System.err.println("DELETING "+sf2.getType()+
                                                   " predicted for "+ geneName+
                                                   ", note = "+note);
                                holder.removeFeature(sf2);

                                notFound = false;
                                break;
                            }
                        }
                    }
                }

                if (notFound == true) {
                    look = false;
                }
            }
        }
        catch(Exception e){
            System.err.println("**********************************************"+
                               "**********************************");
            System.err.println("ERROR;");
            e.printStackTrace();
            System.err.println("**********************************************"+
                               "**********************************");
            System.exit(1);
        }
    }

    /* <code>getChangableFeatures</code> returns all features that might
     * change because manual annotation has occurred.
     * Also, some features do not have PUB_LOCUS, LOCUS or FEAT_NAME fields
     * so they can not be mapped well (or we just don't care about them)
     *
     */
    private FeatureHolder getChangableFeatures(Sequence seq){
        FeatureFilter         miscF = new FeatureFilter.ByType("misc_feature");
        FeatureFilter       repeatF = new FeatureFilter.ByType("repeat_region");
        FeatureFilter       sourceF = new FeatureFilter.ByType("source");

        FeatureFilter miscAndRepeatF = new FeatureFilter.Or(miscF, repeatF);
        FeatureFilter  sourceAndAllF = new FeatureFilter.Or(sourceF,
                                                            miscAndRepeatF);
        FeatureFilter           notF = new FeatureFilter.Not(sourceAndAllF);
        FeatureHolder       features = seq.filter(notF);

        return features;
    }

    /* <code>getOldFeatureFor</code> the newer merged feature. It looks up
     * PUB_LOCUS, LOCUS and FEAT_NAME, returning the StrandedFeature.
     *
     * @param sfMerged is the merged (newest) version of the
     * <code>StrandedFeature</code>.
     *
     * @return a <code>StrandedFeature</code> that is the older version of
     * feature passed in.
     */
    private StrandedFeature getOldFeatureFor(StrandedFeature sfMerged){
        StrandedFeature oldFeat = null;

        oldFeat = getFeatureFromHash("PUB_LOCUS", sfMerged, origPubLocusHash);

        if( oldFeat == null ){
            oldFeat = getFeatureFromHash("LOCUS", sfMerged, origLocusHash);

            if( oldFeat == null ){
                oldFeat =
                    getFeatureFromHash("FEAT_NAME", sfMerged, origFeatNameHash);
            }
        }

        return oldFeat;
    }

    /* <code>getFeatureFromHash</code> given a XML string such as PUB_LOCUS or
     * LOCUS, it gets the zero or more qualifiers from the EMBL file looks
     * for a match against the key in the hash. It returns the value
     * (StrandedFeature) if there is a match.
     *
     * @param qualifier is a <code>String</code>, probably one of PUB_LOCUS,
     * LOCUS and FEAT_NAME
     * @param mergedFeature is the (newest) <code>StrandedFeature</code> which
     * contains the values for the qualfier (i.e. Tb927.1.100)
     * @param hash is a <code>Map</code> of the PUB_LOCUS, LOCUS and FEAT_NAME
     *
     * @return the matching <code>StrandedFeature</code> to the param
     * mergedFeature or null
     */
    private StrandedFeature getFeatureFromHash(String qualifier,
                                               StrandedFeature mergedFeature,
                                               Map hash){
        String[] names = getManyNames4TigrElement(qualifier, mergedFeature);

        StrandedFeature feat = null; // stores value to be returned

        for(int x = 0; x < names.length; ++x){
            if( hash.containsKey(names[x]) ){
                feat = (StrandedFeature) hash.get(names[x]);
                break;
            }
        }

        return feat;
    }

    /* <code>transferAnnotation</code> does what it says on the tin - it 
     * transfers qualifiers from the curated (orig) file to the new (merged)
     * file.
     * BUT, it also checks to see if any qualifiers have changed and reports
     * this to the user. It leaves these qualifiers as is.
     *
     * If the qualifier to be transfered is null or "" then the qualifier
     * is removed from the merged file completely.
     *
     * It also checks the locations haven't changed...
     */
    private void transferAnnotation(StrandedFeature sfMerged,
                                    StrandedFeature sfOld) throws Exception {

        // Check location hasn't changed
        //
        Location mergedLoc = sfMerged.getLocation();
        Location    oldLoc = sfOld.getLocation();

        if( ! mergedLoc.equals(oldLoc) ){
            report("    *** The location has changed ***");

            String name = getBestNameForFeature(sfMerged);
            reportLocChanges.add(name);
        }

        // Iterate through keys of merged Anno looking for matching
        // old version. This is more about finding changes than
        // transfering qualifiers over
        // - remove qualifier that is now "empty"
        Annotation mergedAnno = sfMerged.getAnnotation();
        Annotation    oldAnno = sfOld.getAnnotation();

        Set keys = mergedAnno.keys();

        if( ! keys.isEmpty() ){
            Iterator    itty = keys.iterator();
            List toBeDeleted = new ArrayList();

            while(itty.hasNext()){
                String mergedKey = (String) itty.next();

                // Ignore Biojava specail
                if ("internal_data".equals(mergedKey)) {
                    continue;
                }

                if( oldAnno.containsProperty(mergedKey) ){
                    // need to compare Strings
                    // and String [] - order could be different
                    // Booleans (no value, just key)
                    String [] mergedValues =
                        getAnnotationValue(mergedKey, mergedAnno);
                    String []    oldValues =
                        getAnnotationValue(mergedKey, oldAnno);

                    if( ! equal(mergedValues, oldValues) ){
                        report("    '"+mergedKey+"' has changed");

                        // Delete "empty" qualifier or boolean that needs to go
                        //
                        if (   mergedValues != null
                            && mergedValues.length == 1
                           ){

                            if (   "pseudo".equals(mergedKey)
                                && "not a pseudogene".equals(mergedValues[0])
                               ){
                                toBeDeleted.add(mergedKey);
                            }
                            else if ("partial".equals(mergedKey)
                                && "not a partial".equals(mergedValues[0])
                               ){
                                toBeDeleted.add(mergedKey);
                            }
                            else if ("".equals(mergedValues[0])) {
                                toBeDeleted.add(mergedKey);
                            }
                        }
                    }
                }
            }

            // Delete now outside the Iterator for the qualifier keys, can't
            // delete while loop over the keys - ConcurrentModificationException
            // Remove from merged and old - the code below will add
            // back in any qualifier not present in mergedAnno!!!
            //
            for (int x = 0; x < toBeDeleted.size(); ++x) { 
                String annoKey = (String) toBeDeleted.get(x);
                mergedAnno.removeProperty(annoKey);
                oldAnno.removeProperty(annoKey);
                System.err.println("    Deleteing "+annoKey+" for "+
                                   getBestNameForFeature(sfMerged));

            }
        }


        // Iterate through keys of old Anno looking for qualifiers
        // that are not present in the merged and need to be added
        keys = oldAnno.keys();

        try {
            if( ! keys.isEmpty() ){
                Iterator itty = keys.iterator();

                while(itty.hasNext()){
                    String oldKey = (String) itty.next();

                    // This is temp only, to be removed
                    if( "GO_unknown".equals(oldKey) ){
                        continue;
                    }

                    if( ! mergedAnno.containsProperty(oldKey) ){
                        mergedAnno.setProperty(oldKey,
                                               oldAnno.getProperty(oldKey));
                    }
                }
            }
        }
        catch(ChangeVetoException cve){
            System.err.println("**********************************************"+
                               "**********************************");
            System.err.println("ERROR;");
            cve.printStackTrace();
            System.err.println("**********************************************"+
                               "**********************************");
            System.exit(1);
        }
    }


   /* <code>postProcessAnntransfer</code> deals with any "weird" stuff after
    * the annotation has been moved from the old feature to the new
    * improved one.
    *
    * /colour will need to be updated if /product has changed!
    */
    private void postProcessAnnoTransfer(StrandedFeature sfMerged,
                                         StrandedFeature sfOld)
        throws Exception {

        Annotation mergedAnno = sfMerged.getAnnotation();
        Annotation    oldAnno = sfOld.getAnnotation();

        // If /product has change update it and hence /colour
        if (    mergedAnno.containsProperty("product")
            && oldAnno.containsProperty("product")
            ) {


            /* EXPLODE - 2+ /PRODUCTS NOT GO INTO ONE STRING :)
            String    oldProduct = (String) oldAnno.getProperty("product");
            String mergedProduct = (String) mergedAnno.getProperty("product");

            if (oldProduct.equals(mergedProduct)) {
                // product did not change
                return;
            }
            */
           // EXPLODE



            // need to compare Strings and String [], order could be different.
            // Booleans have no value, just key.
            String [] mergedValues =
                getAnnotationValue("product", mergedAnno);
            String []    oldValues =
                getAnnotationValue("product", oldAnno);

            // Show "WARNING" if more thn 1 product
            if (mergedValues.length > 1){
                System.err.println("WARNING: 2+ /product qualfiers for "+
                                   getBestNameForFeature(sfMerged)+
                                   " from sparse file (2nd argument)");

                for (int i = 0; i < mergedValues.length; ++i) {
                    System.err.println("     /product=\""+mergedValues[i]+"\"");
                }
           }

            if (oldValues.length > 1){
                System.err.println("    *** WARNING: 2+ /product qualfiers for "+
                                   getBestNameForFeature(sfMerged)+
                                   " from cvs file (1st argument)");

                for (int i = 0; i < oldValues.length; ++i) {
                    System.err.println("     /product=\""+oldValues[i]+"\"");
                }
            }


            if( ! equal(mergedValues, oldValues) ){

                boolean isPseudo = false;

                if (mergedAnno.containsProperty("pseudo")) {
                    isPseudo = true;
                }

                String firstProduct = mergedValues[0];
                String colour = TigrUtils.getColourFromProduct(firstProduct,
                                                               isPseudo);

                if (colour != null) {
                    // DEBUG
                    String oldColour = (String) oldAnno.getProperty("colour");
                    System.err.println("    Changing colour from "+oldColour+
                                       " to "+colour);
                    // END DEBUG
                    
                    try {
                        mergedAnno.setProperty("colour", colour);
                    }
                    catch(ChangeVetoException cve) {
                        System.err.println("**********************************************"+
                               "**********************************");
                        System.err.println("ERROR;");
                        cve.printStackTrace();
                        System.err.println("**********************************************"+
                               "**********************************");
                        System.exit(1);
                    }
                }
            }
        }
    }



    /* <code>markAsProcessed</code> simply adds a annotation value to the feature
     * so we know the feature has been processed. This is useful after we have
     * copied Sanger annotation over the top of the TIGR data and need to
     * know what features have not yet been added.
     *
     * @param sf is a <code>StrandedFeature</code>
     */
    private void markAsProcessed(StrandedFeature sf){

        try {
            Annotation anno = sf.getAnnotation();
            anno.setProperty(PROCESSED_KEY, "1");
        }
        catch(ChangeVetoException cve){
            System.err.println("**********************************************"+
                               "**********************************");
            System.err.println("ERROR;");
            cve.printStackTrace();
            System.err.println("**********************************************"+
                               "**********************************");
            System.exit(1);
        }
    }


    /* <code>isProcessed</code> returned true if the feature has a flag set
     * see above method for more details).
     *
     * @param sf is a <code>StrandedFeature</code>
     * @return true if it has been processed, false otherwise
     */
    private boolean isProcessed(StrandedFeature sf){

        Annotation anno = sf.getAnnotation();

        if (anno.containsProperty(PROCESSED_KEY)) {
            return true;
        }

        return false;
    }

    /* <code>addExtraSangerFeatures</code> looks for all non-marked features
     * in the origSeq and adds them onto the new mergedSeq. This will be
     * things like unsure features or anything that does not make the
     * roundtrip to TIGR
     *
     * It uses the flag this.addCDSFeatures to see if it should ignore adding
     * CDS features from the origSeq. This was an issue if the CDS had been
     * deleted but this code added it back in again.
     *
     * Assumes TMHMMs and signalPs come after CDS features. If they come before
     * this coded won't exclude them.
     */
    private void addExtraSangerFeatures() {
        
       Map cdsToIgnore = new HashMap();
       Iterator   itty = origSeq.features();

        try {
            while(itty.hasNext()){
                StrandedFeature sf = (StrandedFeature) itty.next();

                if (isProcessed(sf)) {
                    continue;
                }

                // Ignore CDS' and associated features
                //
                if (!this.addCDSFeatures) {
                    if ("CDS".equals(sf.getType())) {
                    
                        Annotation a = sf.getAnnotation();
                        String sysId = "", tempSysId = "";

                        if (a.containsProperty("systematic_id")){
                            sysId = (String) a.getProperty("systematic_id");
                        }

                        if (a.containsProperty("temporary_systematic_id")){
                            tempSysId = (String) a.getProperty("temporary_systematic_id");
                        }

                        if (sysId != null)
                            cdsToIgnore.put(sysId,     null);

                        if (tempSysId != null)
                            cdsToIgnore.put(tempSysId, null);

                        // Reporting
                        String name = sysId;

                        if (name == null)
                            name = tempSysId;

                        System.err.println("Ignoring CDS: "+name+" at " +sf.getLocation());

                        continue;
                    }
                    else {
                        // A TMHMM or singalP will have a note field with
                        // something like 'predicted for SPAC212.08c'
                        Annotation a = sf.getAnnotation();

                        if (a.containsProperty("note")){
                            String   note = (String) a.getProperty("note");
                            int lastIndex = 
                                StringUtils.lastIndexOf(note, "predicted for ");

                            if (lastIndex != -1) {
                                String geneName =
                                    StringUtils.substringBetween(note,
                                                                 "predicted for ",
                                                                 " ");

                                System.err.println("BEFORE genename = "+geneName+", note = "+note);

                                if(geneName == null || "".equals(geneName)) {
                                    geneName =
                                        StringUtils.substringAfter(note,
                                                                   "predicted for ");
                                }

                                System.err.println("AFTER  genename = "+geneName+", note = "+note);

                                if (cdsToIgnore.containsKey(geneName)) {
                                    System.err.println("Ignoring "+sf.getType()+
                                                       " for "+ geneName);
                                    continue;
                                }
                            }
                        }
                    }
                }

                //System.err.println("Adding unmarked feature to mergedSeq;");
                //dumpFeature(sf);

                // Copy feature to mergedSeq
                mergedSeq.createFeature(sf.makeTemplate());
            }
        }
        catch(BioException be){
            System.err.println("**********************************************"+
                               "**********************************");
            System.err.println("ERROR;");

            be.printStackTrace();
            System.err.println("**********************************************"+
                               "**********************************");
            System.exit(1);
        }
        catch(ChangeVetoException cve){
            System.err.println("**********************************************"+
                               "**********************************");
            System.err.println("ERROR;");
            cve.printStackTrace();
            System.exit(1);
            System.err.println("**********************************************"+
                               "**********************************");
        }
    }


    /* <code>equal</code> compares 2 String arrays and returns true
     * if they are indentical, although the order can be different.
     */
    private boolean equal(String [] mergedValues, String [] oldValues) throws Exception {

        if (mergedValues == null && oldValues == null) {
            throw new Exception("ERROR: equal() was passed 2 null arrays");
        }

        for(int i = 0; i < mergedValues.length; ++i){
            boolean theSame = false;

            for(int j = 0; j < oldValues.length; ++j){
                if( mergedValues[i].equals(oldValues[j]) ){
                    theSame = true;
                    break;
                }
            }

            if( ! theSame ){
                return false;
            }
        }

        return true;
    }


    /* <code>report</code>
     */
    private void report(String msg){
        System.err.println(msg);
    }

    /* <code>dumpFeature</code> simply prints out all the annotation for
     * a feature.
     *
     */
    private void dumpFeature(StrandedFeature f)
    {
        System.err.println("  "+f.getType() + "\t" + f.getLocation());

        Annotation a = f.getAnnotation();
        Set     keys = a.keys();

        if( ! keys.isEmpty() ){
            Iterator itty = keys.iterator();

            while(itty.hasNext()){
                String       key = (String) itty.next();

                if (f.PROPERTY_DATA_KEY.equals(key)) { // "internal_data"
                    continue;
                }

                String [] values = getAnnotationValue(key, a);

                for(int x = 0; x < values.length; ++x){
                    System.err.println("  /"+key+"\t\t"+values[x]);
                }
            }
        }
    }
}
