/*
 * Copyright (c) 2001  Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package uk.ac.sanger.psu.tigrmerge;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;

import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.io.EmblLikeFormat;
import org.biojava.bio.seq.io.SequenceFormat;
import org.biojava.bio.seq.io.SeqIOTools;
import org.biojava.bio.seq.io.SimpleSequenceBuilder;

/**
 * <code>TigrMerge</code> reads in 2 EMBL files, the original EMBL file with
 * added annotation and the new EMBL file. The theory is that the features
 * can be matched up via gene names in the following order PUB_LOCUS, LOCUS,
 * FEAT_NAME before doing and sequence comparisons.
 *
 * This class handles all file IO, the feature merging happens in
 * TigrEmblMerger.
 *
 * @author <a href="mailto:pjm@sanger.ac.uk">Paul Mooney</a>
 */
public class TigrMerge
{
    public static void main(String [] argv)
    {
        if (argv.length < 3 || argv.length > 4)
        {
            System.err.println("\nUsage: TigrMerge "+
                               "<original EMBL file> <new EMBL file> "+
                               "<merged file> [add_CDS_Features]\n");
            System.err.println(" Think oldest to newest going left to right\n");
            System.err.println(" Original file = manually annotated EMBL file");
            System.err.println(" New file      = latest EMBL file generated "+
                               "from TIGR XML");
            System.err.println(" Merged file   = what to call the new EMBL "+
                               "file (it will over write the old one!)\n");
            System.err.println("\n It will overlay annotation from the first "+
                               "file on top of the second file and \n"+
                               " create the 3rd.");
            System.err.println(" If you wish all CDS features from the first "+
                               "file to be added on to the second, put \n"+
                               " anything as the forth argument");
            System.exit(0);
        }

        String    origFileName = argv[0];
        String    newFileName  = argv[1];
        String mergedFileName  = argv[2];
        boolean addCDSFeatures = false;

        if (argv.length == 4) {
            addCDSFeatures = true;
        }

        try
        {
            File origFile = new File(origFileName);
            
            if (!origFile.exists() || origFile.length() == 0) {
                System.err.println("EMBL file '"+origFileName+"' not found "+
                                   "or empty");
                System.exit(1);
            }

            File newFile = new File(newFileName);
            
            if (!newFile.exists() || newFile.length() == 0) {
                System.err.println("EMBL file '"+newFileName+"' not found "+
                                   "or empty");
                System.exit(1);
            }

            // open orig
            //
            BufferedReader origReader = new BufferedReader(new
                InputStreamReader(new FileInputStream(origFile)));

            SequenceIterator origSeqIt = SeqIOTools.readEmbl(origReader);
            Sequence           origSeq = null;

            if (origSeqIt.hasNext()) {
                origSeq = origSeqIt.nextSequence();
            }
            else {
                System.err.println("ERROR: the old EMBL file has no sequence!");
                System.exit(1);
            }

            // open new
            //
            BufferedReader newReader = new BufferedReader(new
                InputStreamReader(new FileInputStream(newFile)));

            SequenceIterator newSeqIt = SeqIOTools.readEmbl(newReader);
            Sequence           newSeq = null;


            if (newSeqIt.hasNext()) {
                newSeq = newSeqIt.nextSequence();
            }
            else {
                System.err.println("ERROR: the new EMBL file has no sequence!");
                System.exit(1);
            }

            // merge and get back new sequence
            TigrEmblMerger tem = new TigrEmblMerger(origSeq, newSeq, addCDSFeatures);
            Sequence mergedSeq = tem.merge();




            // write out sequence to STDOUT?

            /* Open merged file - this is the one that will be created
             * by this program
             */
            File mergedFile = new File(mergedFileName);

            try {
                SeqIOTools.writeEmbl(new FileOutputStream(mergedFile),
                                     mergedSeq);
                //SeqIOTools.writeEmbl(System.out, mergedSeq);            
            }
            //catch(FileNotFoundException fnfe){
            //    fnfe.printStackTrace();
            //}
            catch(IOException ioe){
                ioe.printStackTrace();
            }

            
        }
        catch (IOException ioe)
        {
            ioe.printStackTrace();
        }
        catch (BioException be)
        {
            be.printStackTrace();
        }
    }
}
