/*
 * Copyright (c) 2001  Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package uk.ac.sanger.psu.tigrwrite;

import java.io.File;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.PrintStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.List;


import org.jdom.Document;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.EmblLikeFormat;
import org.biojava.bio.seq.io.SequenceFormat;
import org.biojava.bio.seq.io.SimpleSequenceBuilder;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.io.SeqIOTools;

import org.jdom.Document;
import org.jdom.output.XMLOutputter;

//import uk.ac.sanger.psu.tigrwrite.*;


/**
 * <code>TigrMultiWrite</code> is passed a directory where the EMBL
 * files this code generates should go. If the directory does not exist
 * it will be created. It then loops STDIN, each line being an EMBL file
 * to process into XML.
 * 
 * Example use;
 *
 * cd /nfs/pathdb/prod/data/input/tryp/
 * find . -name '*.embl' | grep -v '/hide/' | grep -v 'chrintermediate-sized-chromosome' | fastembl2tigr /tmp/EMBL
 *
 * @author <a href="mailto:kdj@sanger.ac.uk">Keith James</a>
 * @author <a href="mailto:pjm@sanger.ac.uk">Paul Mooney</a>
 */
public class TigrMultiWrite
{
    File             inFile = null;       // contructor sets this
    PrintStream printStream = System.out; // default
    String       chromosome = null;
   
    private static void usage() {
        System.out.println("fastembl2tigr output_dir");
        System.out.println("  with the files specified on stdin eg");
        System.out.println("find . -name '*.embl' | "+
                           "grep -v '/hide/' | "+
                           "grep -v 'chrintermediate-sized-chromosome' | "+
                           "fastembl2tigr /tmp/EMBL");
    }
   
    /* <code>main</code> loops over STDIN, each line being one file path.
     *
     * IMPORTANT: The lowest level directory name is used as the chromosome
     * name. eg ./thing/thang/chr7/ will use chr7. It will string off 'chr'
     * if present. Currently this is a good thing.
     *
     * Each EMBL file is processed by TigrXMLDocBuilder as normal.
     *
     * @param argv <code>String</code>[], is the directory to put the XML files
     * into.
     */
    public static void main(String [] argv)
    {
	int     count = 0;
	List problems = new ArrayList();
	
	if ( argv.length != 1) {
            System.err.println("No output dir specified");
            usage();
            System.exit(-1);
	}

        try {
            File dir = new File(argv[0]);

            if ( !dir.exists()) {
                dir.mkdir();
            }
	   
            BufferedReader in =
                new BufferedReader(new InputStreamReader(System.in));
	   
            String fname;

            while ((fname = in.readLine()) != null) {
                count++;

                System.err.println("Processing "+fname);

                File           f1 = new File(fname);
                String   tmpFname =
                    fname.substring(0, fname.lastIndexOf(".")+1);

                int         lastSlash = tmpFname.lastIndexOf("/");
                String           path = tmpFname.substring(0, lastSlash);
                int secondToLastSlash = path.lastIndexOf("/");
                String         chromo =
                    fname.substring(secondToLastSlash + 1, lastSlash);

                if (chromo.indexOf("chr") != -1) {
                    chromo = chromo.substring(chromo.indexOf("chr") + 3);
                }

                String   basename = tmpFname.substring(lastSlash + 1);
                File            o = new File(dir, basename+"xml");
                PrintStream    ps = new PrintStream(new FileOutputStream(o));

                System.err.println("fname = "+fname+
                                   ", basename = "+basename+
                                   ", chromo = "+chromo);

                TigrMultiWrite tr = new TigrMultiWrite();
                tr.write(f1, ps, chromo);
                ps.close();


                if (!o.exists() || o.length() == 0) { 
                    problems.add(tmpFname);
                }
            }
        }
        catch(Exception e){
            e.printStackTrace();
        }

	if (count == 0) {
            System.err.println("No files processed");
            usage();
            System.exit(-1);
	}

	System.err.println("Processed "+count+" files");

	if (problems.size() > 0) {
            for (int i=0; i < problems.size(); i++) {
                System.err.println("WARNING: "+problems.get(i)+
                                   " is zero-size. "+
                                   "Translation probably failed");
            }					  
	}
    }
   
    public TigrMultiWrite()
        throws IOException
    {

    }


    /**
     * <code>write</code> writes an EMBL file to a
     * <code>PrintStream</code> as XML.
     *
     * @param emblFile a <code>File</code> to write.
     * @param ps a <code>PrintStream</code> to write to.
     * @param chromosome is a <code>String</code> 
     *
     * @exception IOException if an error reading or writing occurs.
     * @exception BioException if an internal data error occurs.
     */
    public void write(File emblFile, PrintStream ps, String chromosome)
        throws IOException, BioException
    {
        BufferedReader reader = new BufferedReader(new
            InputStreamReader(new FileInputStream(emblFile)));

        TigrXMLDocBuilder builder = null;

        if (chromosome != null) {
            builder = new TigrXMLDocBuilder(chromosome);
        }
        else {
            builder = new TigrXMLDocBuilder();
        }

        XMLOutputter  outputter = new XMLOutputter("    ", true);
        SequenceIterator   seqI = SeqIOTools.readEmbl(reader);

        while (seqI.hasNext())
        {
            Sequence seq = seqI.nextSequence();

            Document doc = builder.makeDocument(seq, emblFile.getName());

            outputter.output(doc, ps);
        }
    }
}
