/*
 * Copyright (c) 2001  Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package uk.ac.sanger.psu.tigrwrite;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;

import org.jdom.Document;
import org.jdom.output.XMLOutputter;

import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.io.SeqIOTools;

/**
 * <code>TigrWrite</code> reads an EMBL flat file and writes the
 * sequence and annotation as XML conforming to the TIGR DTD.
 *
 * @author <a href="mailto:kdj@fes1.sanger.ac.uk">Keith James</a>
 */
public class TigrWrite
{
    public static void main(String [] argv)
    {
        if (argv.length < 1 || argv.length > 2)
        {
            System.err.println("Usage: TigrWrite <EMBL file> [chromosome]\n"+
                               "Chromosome is optional");
            System.exit(0);
        }

        // Arg 1 - EMBL FILE
        //
        File emblFile = new File(argv[0]);

        if (!emblFile.exists() || emblFile.length() == 0)
        {
            System.err.println("EMBL file '"+emblFile+"' not found or empty");
            System.exit(1);
        }

        // Arg 2 - CHROMOSOME
        //
        String chromosome = null;

        if (argv.length == 2 && argv[1] != null && argv[1] != "")
        {
            chromosome = argv[1];
        }


        // 
        TigrWrite writer = new TigrWrite();

        try
        {
            writer.write(emblFile, System.out, chromosome);
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }

    /**
     * <code>write</code> writes an EMBL file to a
     * <code>PrintStream</code> as XML.
     *
     * @param emblFile a <code>File</code> to write.
     * @param ps a <code>PrintStream</code> to write to.
     * @param organism a <code>String</code> repre
     *
     * @exception IOException if an error reading or writing occurs.
     * @exception BioException if an internal data error occurs.
     */
    public void write(File emblFile, PrintStream ps, String chromosome)
        throws IOException, BioException
    {
        BufferedReader reader = new BufferedReader(new
            InputStreamReader(new FileInputStream(emblFile)));

        TigrXMLDocBuilder builder = new TigrXMLDocBuilder(chromosome);
        XMLOutputter    outputter = new XMLOutputter("    ", true);

        SequenceIterator seqI = SeqIOTools.readEmbl(reader);

        while (seqI.hasNext())
        {
            Sequence seq = seqI.nextSequence();

            Document doc = builder.makeDocument(seq, emblFile.getName());

            outputter.output(doc, ps);
        }
    }
}


