/*
 * Copyright (c) 2001  Genome Research Limited.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package uk.ac.sanger.psu.tigrwrite;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.Calendar;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParsePosition;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Properties;
import java.util.Set;

//import org.apache.regexp.RE;
//import org.apache.regexp.RESyntaxException;

import org.jdom.Attribute;
import org.jdom.CDATA;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;

import org.biojava.bio.Annotation;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.seq.RNATools;
import org.biojava.bio.seq.Feature;
import org.biojava.bio.seq.FeatureFilter;
import org.biojava.bio.seq.FeatureHolder;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.StrandedFeature;
import org.biojava.bio.seq.io.EmblProcessor;
import org.biojava.bio.symbol.Location;
import org.biojava.bio.symbol.Symbol;
import org.biojava.bio.symbol.SymbolList;
import org.biojava.bio.symbol.SymbolListViews;
import org.biojava.bio.seq.FeatureFilter;
import org.biojava.bio.AnnotationType;
import org.biojava.bio.CollectionConstraint;
import org.biojava.bio.CardinalityConstraint;
import org.biojava.bio.PropertyConstraint;

import org.apache.commons.lang.StringUtils;

/**
 * <p><code>TigrXMLDocBuilder</code> creates a JDOM <code>Document</code>
 * from a <code>Sequence</code>. The <code>Document</code> complies
 * with the TIGR XML DTD.</p>
 *
 * <p>This isn't very elegant - it has to contain several hacks to
 * cope with inconsistent use of EMBL qualifiers.</p>
 *
 * @author <a href="mailto:kdj@sanger.ac.uk">Keith James</a>
 */
class TigrXMLDocBuilder
{
    private DocType docType;
    private Element rootElement;
    private String  currentDate;
    //private RE      locusRE;
    //private RE      pubLocusRE;
    //private RE      geneSymRE;
    //private RE      modelFeatNameRE;
    
    private static final String      PROPERTY_UNKNOWN = "null";//TIGR like "null"
    private static final String        PROPERTY_EMPTY = "";
    private static final String PROPERTY_XMLPROP_FILE = "xml.properties";
    private static final Properties          xmlProps = new Properties();

    private static String  propertySeqGroup = PROPERTY_UNKNOWN;
    private static String   propertyContact = PROPERTY_UNKNOWN;
    private static String   propertyLineage = PROPERTY_UNKNOWN;
    private static String  propertyOrganism = PROPERTY_UNKNOWN;
    private static String   propertyDTDFile = "tigrxml.dtd";
    private static boolean propertyCloneNameIsFilename = false;

    private static String  propertyPubLocusQualifier = PROPERTY_UNKNOWN;
    private static String     propertyLocusQualifier = PROPERTY_UNKNOWN;
    private static String  propertyAltLocusQualifier = PROPERTY_UNKNOWN;
    private static String   propertyGeneSymQualifier = PROPERTY_UNKNOWN;
    //private static String  propertyFeatNameQualifier = PROPERTY_UNKNOWN;

    private String cloneName = null;

    public static final int          TU = 0;  // Arrya indexes for featNameArray
    public static final int       MODEL = 1;
    public static final int        EXON = 2;
    public static final int         RNA = 3;
    public static final int         CDS = 4;
    public static final int      REPEAT = 5;

    private int[] featNameArray = {0,0,0,0,0,0}; // 1 for each of the above

    // RE to match String used for LOCUS from gene qualifiers
    //private static String    propertyLocusREString = ".*";
    // RE to match String used for PUB_LOCUS from gene qualifiers
    //private static String propertyPubLocusREString = ".*";
    // RE to match String used for GENE_SYM from gene qualifiers
    //private static String  propertyGeneSymREString = ".*";
    //RE to match String used for FEAT_NAME within <MODEL>
    private static String   propertyModelFeatNameString;

    // FIXME
    //
    // *** NOTE: These will be source from the xml.properties file
    //           in the fullness of time :)
    //
    // WARNING: THEY ARE SETUP FOR TRYP ONLY!!!!!!!!!!!!!!!11
    //
    //
    /*
    //private static String pubLocusQualifier = "systematic_id";
    //private static String    locusQualifier = "temporary_systematic_id";
    //private static String featNameQualifier = "temporary_systematic_id";
    */

    private String chromosome = null;

    TigrXMLDocBuilder(String chromosome)
    {
        this();
        this.chromosome = chromosome;
    }

    TigrXMLDocBuilder()
    {
        try
        {
            xmlProps.load(new FileInputStream(PROPERTY_XMLPROP_FILE));
        }
        catch (FileNotFoundException fnfe)
        {
            System.err.println("Unable to find XML properties file: "
                               + fnfe.getMessage());
        }
        catch (IOException ioe)
        {
            System.err.println("Unable to read XML properties file: "
                               + ioe.getMessage());
        }

        docType = new DocType("TIGR");
    }

    /**
     * The <code>makeDocument</code> method builds the
     * <code>Document</code>. Some new private methods could be
     * factored out of this code at some point. However, while the DTD
     * is changing this probably isn't worth it.
     *
     * @param seq a <code>Sequence</code> containing the data.
     * @param emblFileName a <code>String</code> which is the name of the file
     * currently being read - used to populate CLONE_NAME
     *
     * @return a <code>Document</code>.
     */
    Document makeDocument(final Sequence seq, String emblFileName)
    {
        //System.err.println("seq.getURN() = '"+seq.getURN()+"'");

        // Set the date and time
        currentDate = calculateDate();

        // Set properties from file
        try
        {
            String fileName = xmlProps.getProperty("DTD_FILE");
            if (fileName != null)
                propertyDTDFile = fileName;

            String seqGroup = xmlProps.getProperty("SEQUENCING_GROUP");
            if (seqGroup != null)
                propertySeqGroup = seqGroup;

            String contact = xmlProps.getProperty("CONTACT_PERSON");
            if (contact != null)
                propertyContact = contact;

            String lineage = xmlProps.getProperty("ORGANISM_TAXON");
            if (lineage != null)
                propertyLineage = lineage;

            String organism = xmlProps.getProperty("ORGANISM");
            if (organism != null)
                propertyOrganism = organism;

            String pubLocusQualifier =
                xmlProps.getProperty("PUB_LOCUS_QUALIFIER");
            String    locusQualifier =
                xmlProps.getProperty("LOCUS_QUALIFIER");
            String    altLocusQualifier =
                xmlProps.getProperty("ALT_LOCUS_QUALIFIER");
            String  geneSymQualifier =
                xmlProps.getProperty("GENE_SYM_QUALIFIER");

            if (pubLocusQualifier != null)
                propertyPubLocusQualifier = pubLocusQualifier;

            if (locusQualifier != null)
                propertyLocusQualifier = locusQualifier;

            if (altLocusQualifier != null)
                propertyAltLocusQualifier = altLocusQualifier;

            if (geneSymQualifier != null)
                propertyGeneSymQualifier = geneSymQualifier;

            if (locusQualifier == null && pubLocusQualifier == null) {
                System.err.println("\nWARNING: PUB_LOCUS_QUALIFIER and "+
                                   "LOCUS_QUALIFIER in the xml.properties "+
                                   "file have not been set!\n");
            }

            if (altLocusQualifier == null) {
                System.err.println("\nWARNING: ALT_LOCUS_QUALIFIER "+
                                   " in the xml.properties "+
                                   "file have not been set!\n");
            }

            if (geneSymQualifier == null) {
                System.err.println("\nWARNING: GENE_SYM_QUALIFIER "+
                                   " in the xml.properties "+
                                   "file have not been set!\n");
            }

            String cloneNameIsFileName =
                xmlProps.getProperty("CLONE_NAME_IS_FILE_NAME");

            if( cloneNameIsFileName != null &&
                cloneNameIsFileName.equals("true") ){

                propertyCloneNameIsFilename = true;
            } 
        }
        catch (NoSuchElementException nse) { }


        docType.setSystemID(propertyDTDFile);
        /*
         * Root ELEMENT 
         */
        rootElement = new Element("TIGR");
        Document doc = new Document(rootElement, docType);

        // Create ASSEMBLY
        Element assembly = new Element("ASSEMBLY");

        assembly.setAttribute("CLONE_ID",     PROPERTY_UNKNOWN);
        assembly.setAttribute("DATABASE",     getDataBase(seq));
        assembly.setAttribute("CHROMOSOME",   getChromosome(seq));
        assembly.setAttribute("CURRENT_DATE", currentDate);

        // Add ASSEMBLY
        rootElement.addContent(assembly);

        /*
         * Entities in ASSEMBLY
         */
        List assemblyElements = new ArrayList();

        assemblyElements.add(makeAsmblId(seq));
        assemblyElements.add(makeCoordSet(1, seq.length()));
        assemblyElements.add(makeHeader(seq, emblFileName));

        assemblyElements.add(makeGeneList(seq));

        Element misc_info = makeMiscInfo(seq);

        if (misc_info != null)
            assemblyElements.add(makeMiscInfo(seq));

        assemblyElements.add(makeRepeatList(seq));
        assemblyElements.add(new Element("ASSEMBLY_SEQUENCE").addContent(seq.seqString()));

        // Add entities to ASSEMBLY
        addElements(assembly, assemblyElements);

        return doc;
    }

    /**
     * <code>getDataBase</code> sets the same of the source database.
     *
     * @param seq a <code>Sequence</code>.
     *
     * @return a database name <code>String</code>.
     */
    private String getDataBase(final Sequence seq)
    {
        return "EMBL";
    }

    /**
     * <code>getChromosome</code> extracts the chromosome number from
     * the 'source' feature, if present.
     *
     * *** HACK: chromsome can be passed into a contructor, hence no need
     *           to get it from the source feature.
     *           I WOULD LIKE TO REMOVE THIS FEATURE
     *
     * @param seq a <code>Sequence</code>.
     *
     * @return a chromosome number <code>String</code>.
     */
    private String getChromosome(final Sequence seq)
    {
        if (this.chromosome != null && this.chromosome != "")
        {
            return this.chromosome;
        }

        // Now do it properly, get chromo from the source feature
        //
        String chromosome = PROPERTY_UNKNOWN;

        FeatureHolder subSet =
            seq.filter(new FeatureFilter.ByType("source"), false);

        for (Iterator fi = subSet.features(); fi.hasNext();)
        {
            Feature f = (Feature) fi.next();

            String [] values =
                getAnnotationValue("chromosome", f.getAnnotation());

            if (values.length > 0)
            {
                // There is only one chromosome qualifier
                chromosome = values[0];
                continue;
            }
        }

        return chromosome;
    }

    /**
     * <code>makeAsmblId</code> extracts the TIGR assembly ID from the
     * 'source feature, if present and created a new Element to hold
     * it.
     *
     * @param seq a <code>Sequence</code>.
     *
     * @return an <code>Element</code>.
     */
    private Element makeAsmblId(final Sequence seq)
    {
        String id = PROPERTY_UNKNOWN;

        FeatureHolder subSet =
            seq.filter(new FeatureFilter.ByType("source"), false);

        for (Iterator fi = subSet.features(); fi.hasNext();)
        {
            Feature f = (Feature) fi.next();

            String [] values = getAnnotationValue("db_xref", f.getAnnotation());

            if (values.length > 0)
            {
                // There is only one assembly ID
                id = values[0];
                continue;
            }
        }

        return new Element("ASMBL_ID").addContent(id);
    }

    /**
     * <code>makeHeader</code> builds the HEADER element.
     *
     * @param seq a <code>Sequence</code>.
     *
     * @return an <code>Element</code>.
     */
    private Element makeHeader(final Sequence seq, String emblFileName)
    {
        FeatureHolder subSet;

        Element header = new Element("HEADER");

        List headerElements = new ArrayList();

        // Add CLONE_NAME
        this.cloneName = makeGoodCloneName(emblFileName);
        headerElements.add(new Element("CLONE_NAME").addContent(this.cloneName));

        // Add SEQ_LAST_TOUCHED
        headerElements.add(new Element("SEQ_LAST_TOUCHED")
            .addContent(new Element("DATE").addContent(PROPERTY_UNKNOWN)));

        // Add GB_ACCESSION
        String accNumber = PROPERTY_UNKNOWN;
        String [] accessions =
            getAnnotationValue(EmblProcessor.PROPERTY_EMBL_ACCESSIONS,
                               seq.getAnnotation());
	
        if (accessions.length > 0)
            // Get primary accession only
            accNumber = accessions[0];

        headerElements.add(new Element("GB_ACCESSION")
            .addContent(accNumber));

        // Add ORGANISM - get from properties file or EMBL "source"
        String organism = propertyOrganism;

	if( organism == null ){
	    subSet = seq.filter(new FeatureFilter.ByType("source"), false);

	    for (Iterator fi = subSet.features(); fi.hasNext();){
		Feature f = (Feature) fi.next();
		organism = getFirstQualifier("organism", f);

		if (organism != PROPERTY_UNKNOWN)
		    // There is only one organism qualifier
		    continue;
	    }
	}

        headerElements.add(new Element("ORGANISM")
            .addContent(organism));

        // Add LINEAGE
        headerElements.add(new Element("LINEAGE")
            .addContent(propertyLineage));

        // Add SEQ_GROUP
        headerElements.add(new Element("SEQ_GROUP")
            .addContent(propertySeqGroup));

        // Add AUTHOR_LIST
        headerElements.add(new Element("AUTHOR_LIST")
            .setAttribute("CONTACT", propertyContact));

        // Add entities to HEADER
        addElements(header, headerElements);

        return header;
    }

     /**
     * <code>makeGoodCloneName</code> returns a <code>String</code>
     * that is the file name minus the .xml, .embl and .artemis suffixes.
     * Currently it can cope with ".embl.xml" due to the order of the code.
     *
     * @param seq a <code>Sequence</code>.
     *
     * @return an <code>Element</code>.
     */
    private String makeGoodCloneName(String emblFileName){
        int     dotXML  = emblFileName.indexOf(".xml");
        int     dotEmbl = emblFileName.indexOf(".embl");
        int  dotArtemis = emblFileName.indexOf(".artemis");
        String toReturn = PROPERTY_UNKNOWN;

        if( propertyCloneNameIsFilename ){
            if( dotEmbl >= 0 ){
                toReturn = emblFileName.substring(0, dotEmbl);
            }
            else if( dotXML >= 0 ){
                toReturn = emblFileName.substring(0, dotXML);
            }
            else if( dotArtemis >= 0 ){
                toReturn = emblFileName.substring(0, dotArtemis);
            }
            else {
                System.err.println("WARNING: can not CLONE_NAME FROM file name"+
                                   ", it does not end with .xml, .embl or "+
                                   ".artemis");
            }
        }

        return toReturn;
    }

    /**
     * <code>makeGeneList</code> makes a GENE_LIST
     * <code>Element</code> which contains all protein and RNA genes.
     *
     * @param seq a <code>Sequence</code>.
     *
     * @return an <code>Element</code>.
     */
    private Element makeGeneList(final Sequence seq)
    {
        Element geneList = new Element("GENE_LIST");

        geneList.addContent(makeProteinCoding(seq));
        geneList.addContent(makeRNACoding(seq));

        return geneList;
    }

    /**
     * <code>makeMiscInfo</code> makes a MISC_INFO
     * <code>Element</code> which contains all MISC_FEATUREs.
     *
     * @param seq a <code>Sequence</code>.
     *
     * @return an <code>Element</code> or null. The schema allows for 0 or 1, 
     * if one MISC_INFO is present then MISC_FEATUREs mustbe present.
     */
    private Element makeMiscInfo(final Sequence seq)
    {
        Element miscInfo = new Element("MISC_INFO");

        FeatureHolder subSet =
            seq.filter(new FeatureFilter.ByType("misc_feature"), false);

        List               sorted = makeSortedFeatureList(subSet);
        boolean miscFeaturesExist = false;

        for (Iterator fi = sorted.iterator(); fi.hasNext();)
        {
            miscFeaturesExist = true;
            StrandedFeature f = (StrandedFeature) fi.next();

            Element miscFeature = new Element("MISC_FEATURE");

            // Add COORDSET
            try
            {
                miscFeature.addContent(makeCoordSet(f));

                // Add DATE
                miscFeature.addContent(new Element("DATE")
                    .addContent(currentDate));

                // Add FEATURE_DESC
                String featDesc = getFirstQualifier("note", f);

                miscFeature.addContent(new Element("FEATURE_DESC")
                    .addContent(featDesc));

                // Add MISC_FEATURE
                miscInfo.addContent(miscFeature);
            }
            catch (BioException be)
            {
                System.err.println(be.getMessage());
            }
        }

        if (miscFeaturesExist)
            return miscInfo;

        return null;
    }

    /**
     * <code>makeRepeatList</code> makes a REPEAT_LIST
     * <code>Element</code> containing all repeats.
     *
     * @param seq a <code>Sequence</code>.
     *
     * @return an <code>Element</code>.
     */
    private Element makeRepeatList(final Sequence seq)
    {
        Element repeatList = new Element("REPEAT_LIST");

        FeatureFilter regionFilter = new FeatureFilter.ByType("repeat_region");
        FeatureFilter   unitFilter = new FeatureFilter.ByType("repeat_unit");

        FeatureHolder subSet = seq.filter(new FeatureFilter.Or(regionFilter,
                                                               unitFilter), false);

        List sorted = makeSortedFeatureList(subSet);

        for (Iterator fi = sorted.iterator(); fi.hasNext();)
        {
            StrandedFeature f = (StrandedFeature) fi.next();

            Element repeat = new Element("REPEAT");

            try
            {
                // Add FEAT_NAME - not sure what old logic was about...
                /*
                String featName = PROPERTY_UNKNOWN;
                String [] featNames =
                    getAnnotationValue("note", f.getAnnotation());

                if (featNames.length > 0)
                    featName = featNames[0];

                */
                String featName = getFeatNameFor(REPEAT);

                repeat.addContent(new Element("FEAT_NAME").addContent(featName));

                // Add DATE
                repeat.addContent(new Element("DATE").addContent(currentDate));

                // Add COORDSET
                repeat.addContent(makeCoordSet(f));

                // Add REPEAT_TYPE
                repeat.addContent(new Element("REPEAT_TYPE")
                    .addContent(f.getType()));

                // Add REPEAT
                repeatList.addContent(repeat);
            }
            catch (BioException be)
            {
                System.err.println(be.getMessage());
            }
        }

        return repeatList;
    }

    /**
     * <code>makeProteinCoding</code> makes a PROTEIN_CODING
     * <code>Element</code> representing all protein-coding genes in
     * the Sequence.
     *
     * @param seq a <code>Sequence</code> object.
     *
     * @return an <code>Element</code> object.
     */
    private Element makeProteinCoding(final Sequence seq)
    {
        Element proteinCoding = new Element("PROTEIN_CODING");

        // WARNING:
        // this FeatureFilter is set to recurse (true) because some of the 
        // qualifers are in unquoted strings, confusing the Biojava EMBL parser
        FeatureHolder subSet =
            seq.filter(new FeatureFilter.ByType("CDS"), false);

        List sorted = makeSortedFeatureList(subSet);


        int count = 0;

        for (Iterator fi = sorted.iterator(); fi.hasNext();)
        {
            ++count;

            StrandedFeature f = (StrandedFeature) fi.next();
            int strand = f.getStrand().getValue();

            Element tu = new Element("TU");

            List tuElements = new ArrayList();

            try
            {	
		// Add TU FEAT_NAME
                String tuFeatName = getFeatNameFor(TU);

                //String tuFeatName = getFirstQualifier(featNameQualifier, f);
                tuElements.add(new Element("FEAT_NAME").addContent(tuFeatName));

                // Add DATE
                tuElements.add(new Element("DATE").addContent(currentDate));

                // Add GENE_INFO
                tuElements.add(makeGeneInfo(f));

                // Add COORDSET
                tuElements.add(makeCoordSet(f));

                // Get MODEL FEAT_NAME
		String modelFeatName =
                    getFeatNameFor(MODEL);
                    //getFirstQualifier(featNameQuaifier, f);

                // Make MODEL
                Element model = new Element("MODEL");
                model.addContent(new Element("FEAT_NAME").addContent(modelFeatName));
                model.addContent(new Element("DATE").addContent(currentDate));
                model.addContent(makeCoordSet(f));

                model.setAttribute("COMMENT",
                                   "Generated from EMBL annotation");


                // Make MODEL_ATTRIBUTE
                Element model_attribute = getModelAttribute(f);

                if( model_attribute != null ){
                    model.addContent(model_attribute);
                }

                // Make MODEL_EVIDENCE
                Element model_evidence = getModelEvidence(f);

                if( model_evidence != null ){
                    model.addContent(model_evidence);
                }

                // Make EXON(s)
                for (Iterator bi = f.getLocation().blockIterator(); bi.hasNext();)
                {
                    Location  el = (Location) bi.next();
                    Element exon = new Element("EXON");
                    String mFeatName = getFeatNameFor(EXON);
                    exon.addContent(new Element("FEAT_NAME").addContent(mFeatName));
                    exon.addContent(new Element("DATE").addContent(currentDate));

                    // Although contained within a feature which has
                    // strand defined, exons redundantly define their
                    // strand
                    Element exonCoords = null; // Can only have one parent
                    Element  cdsCoords = null;

                    if (strand == -1) {
                        exonCoords = makeCoordSet(el.getMax(), el.getMin());
                        cdsCoords  = makeCoordSet(el.getMax(), el.getMin());
                    }
                    else {
                        exonCoords = makeCoordSet(el.getMin(), el.getMax());
                        cdsCoords  = makeCoordSet(el.getMin(), el.getMax());
                    }

                    exon.addContent(exonCoords);

                    // Now for CDS part of EXON, reuse coords since we do not
                    // store UTR in our locations
                    Element        cds = new Element("CDS");
                    String cdsFeatName = getFeatNameFor(CDS);
                    cds.addContent(new Element("FEAT_NAME").addContent(cdsFeatName));
                    cds.addContent(new Element("DATE").addContent(currentDate));

                    cds.addContent(cdsCoords);
                    exon.addContent(cds);
                    model.addContent(exon);
                }

                // Add CDS_SEQUENCE
                // Something bad happened within Biojava, probably down to bad EMBL
                // files, perhaps qualifiers not quoted?
                /*
                try {
                    SymbolList symbols = f.getSymbols();
                    System.err.println("Got symbols -------");
                    System.err.println("symbols.length() = "+symbols.length());
                    System.err.println("f.Location()() = "+f.getLocation());


                    SymbolList symbols2 = symbols.subList(1, symbols.length());

                    System.err.println("symbols2 = "+symbols2); // BANG

                    String   seqString = symbols.seqString(); // BANG
                    System.err.println("Got seqString");
                    String      string = f.getSymbols().seqString();

                }
                catch (Exception e) {
                    System.err.println("SOMETHING BAD HAPENED FOR FEATURE: "+f);

                    e.printStackTrace();

                    System.exit(1);
                }
                */

                model.addContent(new Element("CDS_SEQUENCE")
                    .addContent(f.getSymbols().seqString()));

                Annotation   anno = f.getAnnotation();
                boolean  isPseudo = anno.containsProperty("pseudo");
                boolean isPartial = anno.containsProperty("partial");

                //String pseudo  = getFirstQualifier("pseudo", f);
                //String partial = getFirstQualifier("partial", f);

                //if (! (pseudo.equals(PROPERTY_UNKNOWN) ||
                //       partial.equals(PROPERTY_UNKNOWN)))
                if (! (isPseudo || isPartial) )
                {
                    // Add PROTEIN_SEQUENCE
                    SymbolList rna = RNATools.transcribe(f.getSymbols());

                    try
                    {
                        SymbolList      codons =
                            SymbolListViews.windowedSymbolList(rna, 3);
                        SymbolList translation =
                            SymbolListViews.translate(codons,
                                                      RNATools.getGeneticCode("UNIVERSAL"));

                        // Remove the stop codon symbol so that it
                        // won't show up in the sequence string
                        if (translation.symbolAt(translation.length())
                            .getName().equals("TER"))
                        {
                            translation =
                                translation.subList(1, translation.length() - 1);
                        }

                        model.addContent(new Element("PROTEIN_SEQUENCE")
                            .addContent(translation.seqString()));
                    }
                    catch (IllegalArgumentException iae)
                    {
                        System.err.println("Unable to translate: "
                                           + f
                                           + " "
                                           + iae.getMessage());
                    }
                }

                // Add MODEL
                tuElements.add(model);

                // Add TRANSCRIPT_SEQUENCE
                Location tl = f.getLocation();
                SymbolList subList = seq.subList(tl.getMin(), tl.getMax());

                if (strand == -1)
                    subList = DNATools.reverseComplement(subList);

                tuElements.add(new Element("TRANSCRIPT_SEQUENCE")
                    .addContent(subList.seqString()));

                // Add entities to TU
                addElements(tu, tuElements);

                proteinCoding.addContent(tu);
            }
            catch (BioException be)
            {
                System.err.println(be.getMessage());
            }
        }

        return proteinCoding;
    }

    /**
     * <code>makeGeneInfo</code> makes a GENE_INFO
     * <code>Element</code> for a single CDS feature.
     *
     * @param f a <code>Feature</code>.
     *
     * @return an <code>Element</code>.
     */
    private Element makeGeneInfo(Feature f)
    {
        Element geneInfo = new Element("GENE_INFO");

        // Add LOCUS, which must be present (the DTD only allows
        // one). The RE matches the systematic gene name assigned by
        // Sanger, which should be unique and therefore satisfies the
        // TIGR requirement for a unique LOCUS. It's the best we can
        // do, I think.

        /*
        String locus = getFirstQualifier("gene", locusRE, f);
        geneInfo.addContent(new Element("LOCUS").addContent(locus));

        // Add PUB_LOCUS, if present (the DTD only allows one)
        String pubLocus = getFirstQualifier("label", pubLocusRE, f);
        if (pubLocus != PROPERTY_UNKNOWN)
            geneInfo.addContent(new Element("PUB_LOCUS").addContent(pubLocus));
        */


        // LOCUS is (was??) a required field, PUB_LOCUS is not
        // Even if we do not need such a thing in the EMBL file
        //
        String locus    = getFirstQualifier(propertyLocusQualifier, f);
        String pubLocus = getFirstQualifier(propertyPubLocusQualifier, f);

        if( locus != PROPERTY_UNKNOWN )
            geneInfo.addContent(new Element("LOCUS").addContent(locus));
        else if (pubLocus != PROPERTY_UNKNOWN )
            geneInfo.addContent(new Element("LOCUS").addContent(pubLocus));
        else {
            System.err.println("ERROR;\n");
            System.err.println("\nNo locus or publocus available for this feature at "+f.getLocation());
            System.err.println("No XML will be generated till this is fixed");
            System.err.println("Have the PUB_LOCUS and LOCUS lines been set correctly in the xml.properties file?");
            System.err.println("Or are the qualifiers not used in the EMBL file?");
            System.err.println("  PUB_LOCUS = "+propertyPubLocusQualifier);
            System.err.println("  LOCUS     = "+propertyLocusQualifier+"\n");
            System.exit(1);
        }

        if (pubLocus != PROPERTY_UNKNOWN)
            geneInfo.addContent(new Element("PUB_LOCUS").addContent(pubLocus));

        // GENE_SYM (probably /primary_name, set in xml.properties file)
        // Should only ever be 1 /primary_name, right?
        //
        if (propertyGeneSymQualifier != PROPERTY_UNKNOWN) {
            String [] geneSyms =
                getAnnotationValue(propertyGeneSymQualifier, f.getAnnotation());

            if (geneSyms != null) {
                for (int i = 0; i < geneSyms.length; ++i) {
                    String  geneSymString  = geneSyms[i];
                    Element geneSymElement =
                        new Element("GENE_SYM").addContent(geneSymString);

                    if (i == 0) {
                        geneSymElement.setAttribute("IS_PRIMARY", "1");
                    }
                    else {
                        geneSymElement.setAttribute("IS_PRIMARY", "0");
                    }

                    geneInfo.addContent(geneSymElement);
                }
            }
        }
          

        // Add in any ALT_LOCUS tags (/previous_systematic_ids at Sanger).
        //
        if (propertyAltLocusQualifier != PROPERTY_UNKNOWN) {
            String [] previous_ids =
                getAnnotationValue(propertyAltLocusQualifier,
                                   f.getAnnotation());

            if (previous_ids != null) {
                for (int i = 0; i < previous_ids.length; ++i) {
                    String   old_id   = previous_ids[i];
                    Element alt_locus =
                        new Element("ALT_LOCUS").addContent(old_id);

                    geneInfo.addContent(alt_locus);
                }
            }
        }


        // Add COM_NAME, one of which must have IS_PRIMARY="1"
        // Can gave several /product qualifiers each of which
        // maybe made up of several products, ';'seperated
        //
        String [] products = getAnnotationValue("product",
                                                f.getAnnotation());

        if (products != null) {
            for (int i = 0; i < products.length; ++i) {
                String productQualifier = products[i];
                String[]       sections = StringUtils.split(productQualifier,";");
                    
                for (int j = 0; j < sections.length ; j++) {
                    String   product = sections[j].trim();
                    Element com_name = new Element("COM_NAME").addContent(product);

                    if (i == 0 && j == 0) { // if first /product, first section
                        com_name.setAttribute("IS_PRIMARY", "1");
                    }
                    else {
                        com_name.setAttribute("IS_PRIMARY", "0");
                    }

                    geneInfo.addContent(com_name);
                }
            }
        }

        // if no /product or empty /product, at least leave an empty
        // <COM_NAME></COM_NAME>
        if (products == null || products.length == 0){
            Element com_name = new Element("COM_NAME").addContent("");
            com_name.setAttribute("IS_PRIMARY", "1");
            geneInfo.addContent(com_name);
        }
        

        // Add PUB_COMMENT, if present (the DTD only allows one)
        String pubComment = getFirstQualifier("note", f);
        if (pubComment != PROPERTY_UNKNOWN)
            geneInfo.addContent(new Element("PUB_COMMENT").addContent(pubComment));

        // Add EC_NUM, if present (the DTD only allows one)
        //String ecNum = getFirstQualifier("EC_number", f);
        //if (ecNum != PROPERTY_UNKNOWN)
        //    geneInfo.addContent(new Element("EC_NUM").addContent(ecNum));

        String [] ecNums = getAnnotationValue("EC_number", f.getAnnotation());

        if (ecNums != null) {
            for (int i = 0; i < ecNums.length; ++i) {
                String   ecString = ecNums[i];
                Element ecElement = new Element("EC_NUM").addContent(ecString);

                if (i == 0) {
                    ecElement.setAttribute("IS_PRIMARY", "1");
                }
                else {
                    ecElement.setAttribute("IS_PRIMARY", "0");
                }

                geneInfo.addContent(ecElement);
            }
        }



        // Add IS_PSEUDOGENE, which must be present
        //String pseudo = getFirstQualifier("pseudo", f);

        if (f.getAnnotation().containsProperty("pseudo")) {
            geneInfo.addContent(new Element("IS_PSEUDOGENE").addContent("1"));
        }
        else {
            // Annotator might have forgotten /pseudo, check all product lines
            // (products used above from COM_NAME)
            //)
            boolean isPeeudoGeneAdded = false;

            if (products != null) {
                for (int i = 0; i < products.length; ++i) {
                    String product = products[i];

                    if (product.indexOf(", pseudogene") >= 0 ||
                        product.indexOf(",pseudogene")  >= 0) {

                        System.err.println("WARNING: No /pseudo but /product contains"+
                                           "', pseudogene' for locus: "+locus+
                                           ", pubLocus: "+pubLocus);
                        geneInfo.addContent(new Element("IS_PSEUDOGENE").addContent("1"));

                        isPeeudoGeneAdded = true;
                    }
                }
            }

            if (!isPeeudoGeneAdded) {
                geneInfo.addContent(new Element("IS_PSEUDOGENE").addContent("0"));
            }
        }

        // Add evidence codes
        List evidenceCodes = decodeEvidence(f);
        if (evidenceCodes.size() > 0)
        {
            Element evidence = new Element("EVIDENCE");
            Attribute a;
            for (Iterator ei = evidenceCodes.iterator(); ei.hasNext();)
            {
                a = (Attribute) ei.next();
                evidence.setAttribute(a.getName(), a.getValue());
            }

            geneInfo.addContent(evidence);
        }

        // Add DATE
        geneInfo.addContent(new Element("DATE").addContent(currentDate));

        // Test for encoded role information
        Collection roleInfos = decodeRoleInfo(f);

        if (roleInfos.size() > 0)
        {
            Element roleList = new Element("ROLE_LIST");
            addElements(roleList, roleInfos);

            geneInfo.addContent(roleList);
        }

        // Add GENE_ONTOLOGY
        Element geneOntology = makeGeneOntology(f);

        if( geneOntology != null ){
            geneInfo.addContent(geneOntology);
        }

        return geneInfo;
    }

    /**
     * <code>makeGeneOntology</code> makes a GENE_ONTOLOGY
     * <code>Element</code> if any GO IDs exist from the EMBL feature that
     * is passed in. It will return null if none exist.
     *
     * @param f a <code>Feature</code>.
     *
     * @return an <code>Element</code> representing GENE_ONTOLOGY
     */
    private Element makeGeneOntology(Feature f)
    {
        Annotation  anno = f.getAnnotation();
        String [] goFunc = getAnnotationValue("GO_function",  anno);
        String [] goComp = getAnnotationValue("GO_component", anno);
        String [] goProc = getAnnotationValue("GO_process",   anno);

        Element geneInfo = new Element("GENE_ONTOLOGY");

        makeGOID_OLD_VERSION(geneInfo, "function",  goFunc);
        makeGOID_OLD_VERSION(geneInfo, "component", goComp);
        makeGOID_OLD_VERSION(geneInfo, "process",   goProc);

        String []   goes = getAnnotationValue("GO",   anno);
        //Element geneInfo = new Element("GENE_ONTOLOGY");

        makeGOID(geneInfo, goes);

        if( geneInfo.hasChildren() ){
            return geneInfo;
        }

        return null;
    }

    /**
     * <code>makeGOID</code> makes a GO_ID <code>Element</code> if the String
     * array passed in is not null.
     * It will return null if none exist.
     *
     *
     * @param geneInfo is the parent <code>Element</code>
     * @param goStrings a <code>String</code>[]
     *
     */
    private void makeGOID(Element geneInfo, String [] goStrings)
    {
        if( goStrings.length == 0 ){
            return;
        }

        for(int i = 0; i < goStrings.length; ++i){
            if( goStrings[i].length() <= 0 ){
                System.err.println("Empty GO String");
                continue;
            }

            GoQualifier go = parseGoQualifier(goStrings[i]);

            if( go == null ){
                continue;
            }

            // Make the GO_ID Element structure.
            Element goID = new Element("GO_ID");
            goID.setAttribute("ASSIGNMENT", go.id);
            geneInfo.addContent(goID);

            Element date = new Element("DATE");
            date.setText(go.date);
            goID.addContent(date);

            Element eGoTerm = new Element("GO_TERM");
            eGoTerm.setText(go.term);
            goID.addContent(eGoTerm);

            Element eGoType = new Element("GO_TYPE"); // The aspect
            eGoType.setText(go.aspect);
            goID.addContent(eGoType);

            Element goEvidence = new Element("GO_EVIDENCE");
            goID.addContent(goEvidence);

            Element     evCode = new Element("EV_CODE");
            String        code = "";

            if (go.evidence != null)
                code = go.evidence;

            evCode.setAttribute("CODE", code);
            goEvidence.addContent(evCode);

            Element   evidence = new Element("EVIDENCE");
            evidence.setText(go.db_xref);
            goEvidence.addContent(evidence);
        }
    }

    /* <code>GoQualifier</code> is a glorified struct, no more. It holds
     * information to be passed back to makeGoID() when parseGoQualifier()
     * is called.
     *
     */
    private class GoQualifier {
        public String  aspect;
        public String  id;
        public String  db_xref;
        public String  with;
        public String  evidence;
        public String  type;
        public String  term;
        public String  from;
        public String  date;
        public boolean not = false;

        GoQualifier() {
            date = calculateDate();
        }
    }

    /* <code>parseGoQualifier</code> parses the GO qualifier and populates
     * a <code>GoQualifier</code> object which it returns on sucess. If some
     * key data is missing it will return null.
     *
     * @param goString is a <code>String</code>
     */
    public GoQualifier parseGoQualifier(String goString){
        boolean    fromSet = false;
        boolean    withSet = false;
        String    realName = null;
        String curatorName = null;
        String    evidence = null;

        int               valid = 0;
        GoQualifier goQualifier = new GoQualifier();

        try {
            String[] sections = StringUtils.split(goString,";");

            for (int j=0; j< sections.length ; j++) {
                String expression = sections[j].trim();
                int         index = expression.indexOf("=");

                if ( index == -1) {
                    // Maybe NOT - check
                    if ("not".equals(expression) || "NOT".equals(expression)) {
                        goQualifier.not = true;
                        continue;
                    } else {
                        System.err.println("WARN: can't parse GO expression:" +
                                           expression+goString);
                        break;
                    }
                }

                String   key = expression.substring(0,index).trim();
                String value = expression.substring(index+1).trim();
                
                if ("aspect".equals(key)) {
                    if ( value.startsWith("f") || value.startsWith("F")) {
                        goQualifier.aspect = "function";
                    }

                    if ( value.startsWith("c") || value.startsWith("C")) {
                        goQualifier.aspect = "component";
                    }

                    if ( value.startsWith("p") || value.startsWith("P")) {
                        goQualifier.aspect = "process";
                    }

                    valid+=1;
                }
		   
                if ("GOid".equals(key)) {
                    int bracket =  value.indexOf("(");

                    if ( bracket != -1) {
                        value = value.substring(0,index).trim();
                    }

                    if ( value.length()!=10) {
                        System.err.println("WARN: GO id looks wrong: "+value);
                        continue;
                    }

                    goQualifier.id = value.substring(3);                
                    valid         += 2;
                }

                if ("term".equals(key)) {
                    goQualifier.term = value.trim();		    
                }
		   
                if ("date".equals(key)) {
                    String date = psuDate2TIGRDate(value);
                    goQualifier.date = date;
                }
		 
                if ("evidence".equals(key)) {
                    goQualifier.evidence = value;
                    evidence=value;
                }
  
                if ("from".equals(key)) {
                    goQualifier.from = value;
                    fromSet = true;
                }
  
                if ("with".equals(key)) {
                    goQualifier.with = value;
                    withSet = true;
                }

                if ("db_xref".equals(key)) {
                    goQualifier.db_xref = value;		    
                }
            }
        }
        catch (IllegalArgumentException exp) {
            System.err.println("WARN: Ignoring invalid GO field due to illegal"+
                               " argument  - "+ exp.getMessage());
            return null;
        }
   
        if ( fromSet && withSet) {
            System.err.println("WARN: Ignoring invalid GO field due to both "+
                               "with & from fields in "+goString);
            return null;
        }

        if ("IC".equals(evidence) && !fromSet) {
            System.err.println("WARN: Ignoring invalid GO field due to missing"+
                               " 'from' field "+
                               "with IC evidence code.");
            return null;
        }
        if (   ("ISS".equals(evidence)
            || "IGI".equals(evidence)
            || "IP".equals(evidence) )   
            && !withSet) {
            System.err.println("WARN: Ignoring invalid GO field due to "+
                               "missing with field with "
                               + evidence + " evidence code");
            return null;
        }

        if (valid != 3) {
            System.err.println("WARN: Ignoring incomplete or invalid GO field:"
                               +goString);
            return null;
        }

        return goQualifier;
    }

    /**
     * <code>makeGOID_OLD_VERSION</code> makes a GO_ID <code>Element</code> 
     * if the String array passed in is not null.
     * It will return null if none exist.
     *
     * NOTE: This parses the old style PSU GO_[component|function|process]
     *       style qualifier. When all traces of it are gone this function can
     *       be removed!
     *
     * @param geneInfo is the parent <code>Element</code>
     * @param goType is a <code>String</code> containing either GO_fuction, 
     * Go_process or GO_component.
     * @param goStrings a <code>String</code>[]
     *
     */
    private void makeGOID_OLD_VERSION (Element   geneInfo,
                                       String    goType,
                                       String [] goStrings)
    {
        if( goStrings.length == 0 ){
            return;
        }

        for(int i = 0; i < goStrings.length; ++i){
            if( goStrings[i].length() <= 0 ){
                System.err.println("Empty GO String");
                break;
            }

	    int semi1 = -1, semi2 = -1, semi3 = -1;

            semi1 = goStrings[i].indexOf(";");

	    if( semi1 > 0 ){
		semi2 = goStrings[i].indexOf(";", semi1 + 1);

		if( semi2 > 0 ){
		    semi3 = goStrings[i].indexOf(";", semi2 + 1);
		}
	    }
            else {
                semi1 = goStrings[i].length(); // No ; at all

                if( goStrings[i].trim().length() <= 0 ){
                    System.err.println("ERROR: No text for GO qualifier");
                    System.exit(1);
                }
            }

            //System.err.println("GO STRING= '"+goStrings[i]+"' semi 2 = "+semi2);

            // String out ; seperated strings
            String       id = goStrings[i].substring(0, semi1);
            String evi_code = "";
            String      evi = "";

	    if( semi2 <= 0 ){
		if( goStrings[i].length() > semi1 ){
		    evi_code = goStrings[i].substring(semi1 + 1,
                                                      goStrings[i].length());
		}
	    }
	    else {
		evi_code = goStrings[i].substring(semi1 + 1, semi2);
	    }

            if( semi3 > 0 ){
                evi = goStrings[i].substring(semi2 + 1, semi3);
            }
	    else {
		if( semi2 != -1 && goStrings[i].length() > semi2 ){
		    evi = goStrings[i].substring(semi2 + 1,
                                                 goStrings[i].length());
		}
	    }

            String goTerm = new String(id);

            // id will need stuff in brackets 2 B removed;
            // i.e. "GO:0006696 (ergosterol biosynthesis)" -> "GO:0006696"
            if( id.indexOf("(") > 0 ){
                id = id.substring(0, id.indexOf("("));
            }

            if( goTerm.indexOf("(") > 0 ){
                goTerm = goTerm.substring(goTerm.indexOf("(")+1, goTerm.length()-1);
            }
            else {
                goTerm = null;
            }

            id       = id.trim();
            evi_code = evi_code.trim();
            evi      = evi.trim();            

            // make sure we have enough info we create the GO_ID plus children
            if( evi_code == null || evi_code.equals("") ){
                System.err.println("GO_"+goType+" of "+id+
                                   " has no evidence code - IGNORING");
                continue;
            }

            // Make the GO_ID Element structure.
            Element goID = new Element("GO_ID");
            goID.setAttribute("ASSIGNMENT", id);
            geneInfo.addContent(goID);

            Element date = new Element("DATE");
            date.setText(currentDate);
            goID.addContent(date);

            Element eGoTerm = new Element("GO_TERM");
            eGoTerm.setText(goTerm);
            goID.addContent(eGoTerm);

            Element eGoType = new Element("GO_TYPE");
            eGoType.setText(goType);
            goID.addContent(eGoType);

            Element goEvidence = new Element("GO_EVIDENCE");
            goID.addContent(goEvidence);

            Element     evCode = new Element("EV_CODE");
            evCode.setAttribute("CODE", evi_code);
            goEvidence.addContent(evCode);

            Element   evidence = new Element("EVIDENCE");
            evidence.setText(evi);
            goEvidence.addContent(evidence);
        }
    }

    /**
     * <code>makeRNACoding</code> makes an RNA_GENES
     * <code>Element</code> containing all RNA genes in the Sequence.
     *
     * @param seq a <code>Sequence</code>.
     *
     * @return an <code>Element</code>.
     */
    private Element makeRNACoding(Sequence seq)
    {
        Element rnaGenes = new Element("RNA_GENES");

        FeatureHolder subSet;
        List          sorted;
        // tRNA
        subSet = seq.filter(new FeatureFilter.ByType("tRNA"), false);
        sorted = makeSortedFeatureList(subSet);

        for (Iterator fi = sorted.iterator(); fi.hasNext();)
        {
            StrandedFeature f = (StrandedFeature) fi.next();

            Element pretRNA = new Element("PRE-TRNA");

            try
            {
                // Add FEAT_NAME
                // RNA
                //
                String featName = getFeatNameFor(RNA);
                //String featName = getFirstQualifier(featNameQualifier, f);
                //if (featName == PROPERTY_UNKNOWN)
                //    featName = getFirstQualifier("gene", f);

                pretRNA.addContent(new Element("FEAT_NAME").addContent(featName));

                // Add DATE
                pretRNA.addContent(new Element("DATE").addContent(currentDate));

                // Add COORDSET
                pretRNA.addContent(makeCoordSet(f));

                // Make TRNA
                //
                Element tRNA = new Element("TRNA");

                tRNA.addContent(new Element("FEAT_NAME").addContent(featName));
                tRNA.addContent(new Element("DATE").addContent(currentDate));
                tRNA.addContent(makeCoordSet(f));
                tRNA.addContent(new Element("COM_NAME").addContent(featName));

                String anticodon = getFirstQualifier("anticodon", f);
                tRNA.setAttribute("ANTICODON", anticodon);

                // Make RNA_EXON (assumes no introns)
                //
                Element rnaExon = new Element("RNA-EXON");
                rnaExon.addContent(new Element("FEAT_NAME").addContent(featName));
                rnaExon.addContent(new Element("DATE").addContent(currentDate));

                Location l = f.getLocation();
                rnaExon.addContent(makeCoordSet(l.getMin(), l.getMax()));

                // Add RNA_EXON to TRNA
                tRNA.addContent(rnaExon);

                // Add TRNA
                pretRNA.addContent(tRNA);

                // Add PRE-TRNA
                rnaGenes.addContent(pretRNA);
            }
            catch (BioException be)
            {
                System.err.println(be.getMessage());
            }
        }

        // snRNA
        subSet = seq.filter(new FeatureFilter.ByType("snRNA"), false);
        sorted = makeSortedFeatureList(subSet);

        for (Iterator fi = sorted.iterator(); fi.hasNext();)
        {
            StrandedFeature f = (StrandedFeature) fi.next();
            //int strand = f.getStrand().getValue();

            Element snRNA = new Element("SNRNA");

            try
            {
                addCommonRNAContent(snRNA, f);
                rnaGenes.addContent(snRNA);
            }
            catch (BioException be)
            {
                System.err.println(be.getMessage());
            }
        }

        // rRNA
        subSet = seq.filter(new FeatureFilter.ByType("rRNA"), false);
        sorted = makeSortedFeatureList(subSet);

        for (Iterator fi = sorted.iterator(); fi.hasNext();)
        {
            StrandedFeature f = (StrandedFeature) fi.next();
            //int strand = f.getStrand().getValue();

            Element rRNA = new Element("RRNA");

            try
            {
                addCommonRNAContent(rRNA, f);
                rnaGenes.addContent(rRNA);
            }
            catch (BioException be)
            {
                System.err.println(be.getMessage());
            }
        }

        return rnaGenes;
    }

    /**
     * <code>addCommonRNAContent</code> adds the elements that are common.
     *
     * @param rna an <code>Element</code> - could be snRNA, rRNA etc.
     * @param f a <code>StrandedFeature</code>.
     */
    private void addCommonRNAContent(Element rna, StrandedFeature f)
    throws BioException
    {
        // Add FEAT_NAME
        String featName = getFeatNameFor(RNA);

        //String featName = getFirstQualifier(featNameQualifier, f);
        //if (featName == PROPERTY_UNKNOWN)
        //    featName = getFirstQualifier("gene", f);

        rna.addContent(new Element("FEAT_NAME").addContent(featName));

        // Add DATE
        rna.addContent(new Element("DATE").addContent(currentDate));

        // Add COORDSET
        rna.addContent(makeCoordSet(f));

        // Add COM_NAME
        String comName = getFirstQualifier("note", f);
        rna.addContent(new Element("COM_NAME").addContent(comName));
    }

    /**
     * <code>addElements</code> adds <code>Element</code>s in a Collection
     * to another <code>Element</code>.
     *
     * @param element an <code>Element</code>.
     * @param elements a <code>Collection</code>.
     */
    private void addElements(final Element element, final Collection elements)
    {
        for (Iterator ei = elements.iterator(); ei.hasNext();)
        {
            Object e = ei.next();

            element.addContent((Element) e);
        }
    }

    /**
     * <code>getAnnotationValue</code> extracts the annotation
     * value(s) for a particular Annotation key and converts them to
     * an array of Strings.
     *
     * @param key a <code>String</code>.
     * @param a an <code>Annotation</code>.
     *
     * @return a <code>String []</code> array.
     */
    private String [] getAnnotationValue(final String key, final Annotation a)
    {
        List matches = new ArrayList();

        for (Iterator ai = a.keys().iterator(); ai.hasNext();)
        {
            Object thisKey = ai.next();

            // From EMBL, all values are either Strings, Collections
            // of Strings or Boolean
            if (thisKey.equals(key))
            {
                Object value = a.getProperty(thisKey);

                // To cope with /pseudo, /partial etc.
                if (Boolean.class.isInstance(value))
                {
                    matches.add(value.toString());
                }
                else if (Collection.class.isInstance(value))
                    matches.addAll((Collection) value);
                else
                    matches.add(value);
            }
        }

        return (String []) matches.toArray(new String [0]);
    }

    /**
     * <code>makeCoordSet</code> creates a COORDSET
     * <code>Element</code> representing a
     * <code>StrandedFeature</code>.
     *
     * @param feature a <code>StrandedFeature</code>.
     *
     * @return an <code>Element</code>.
     *
     * @exception BioException if the Feature has unknown strand.
     */
    private Element makeCoordSet(final StrandedFeature feature)
        throws BioException
    {
        int  start = feature.getLocation().getMin();
        int    end = feature.getLocation().getMax();
        int strand = feature.getStrand().getValue();

        if (strand == 1)
        {
            Element end5 = new Element("END5").addContent(Integer.toString(start));
            Element end3 = new Element("END3").addContent(Integer.toString(end));

            return new Element("COORDSET").addContent(end5).addContent(end3);
        }
        else if (strand == -1)
        {
            Element end5 = new Element("END5").addContent(Integer.toString(end));
            Element end3 = new Element("END3").addContent(Integer.toString(start));

            return new Element("COORDSET").addContent(end5).addContent(end3);
        }
        else
            throw new BioException("Encountered a feature with no strand set: "
                                   + feature
                                   + " (not supported by TIGR DTD)");
    }

    /**
     * <code>makeCoordSet</code> creates COORDSET <code>Element</code>
     * from a pair of coordinates.
     *
     * @param start an <code>int</code>.
     * @param end an <code>int</code>.
     *
     * @return an <code>Element</code>.
     */
    private Element makeCoordSet(final int start, final int end)
    {
        Element end5 = new Element("END5").addContent(Integer.toString(start));
        Element end3 = new Element("END3").addContent(Integer.toString(end));

        return new Element("COORDSET").addContent(end5).addContent(end3);
    }

    /**
     * <code>makeSortedFeatureList</code> returns a <code>List</code>
     * of Features in a FeatureHolder, sorted by their Location order.
     *
     * @param holder a <code>FeatureHolder</code> of Features to sort.
     *
     * @return a <code>List</code>.
     */
    private List makeSortedFeatureList(final FeatureHolder holder)
    {
        List features = new ArrayList(holder.countFeatures());

        for (Iterator fi = holder.features(); fi.hasNext();)
        {
            features.add(fi.next());
        }

        Collections.sort(features, Feature.byLocationOrder);

        return features;
    }

    /**
     * <code>getFirstQualifier</code> gets the first instance of a
     * qualifier (used when there is known to be only one
     * instance). Returns the UNKNOWN property if no value is present.
     *
     * @param key a <code>String</code>.
     * @param f a <code>Feature</code>.
     *
     * @return a <code>String</code>.
     */
    private String getFirstQualifier(String key, Feature f)
    {
        String [] vals = getAnnotationValue(key, f.getAnnotation());

        if (vals.length > 0)
            return vals[0];
        else
            return PROPERTY_UNKNOWN;
    }



    /**
     * <code>getFirstQualifier</code> gets the first instance of a
     * qualifier matching a regex. Returns the UNKNOWN property if no
     * value is present.
     *
     * @param key a <code>String</code>.
     * @param regex a <code>RE</code>.
     * @param f a <code>Feature</code>.
     *
     * @return a <code>String</code>.
     */
    /*
    private String getFirstQualifier(String key, RE regex, Feature f)
    {
        String [] vals = getAnnotationValue(key, f.getAnnotation());

        for (int i = 0; i < vals.length; i++)
            if (regex.match(vals[i]))
                return vals[i];

        return PROPERTY_UNKNOWN;
    }
    */

    /**
     * <code>decodeRoleInfo</code> decodes role infos stored as EMBL
     * qualifiers.
     *
     * @param f a <code>Feature</code>.
     *
     * @return a <code>Collection</code> of <code>Element</code>s.
     */
    private Collection decodeRoleInfo(Feature f)
    {
        Annotation a = f.getAnnotation();

        Map roleInfos = new HashMap();

        for (Iterator ai = a.keys().iterator(); ai.hasNext();)
        {
            String thisKey = (String) ai.next();

            // Encoded as _0_, _1_ etc.
            if (thisKey.startsWith("_"))
            {
                try
                {
                    Integer index = Integer.decode(thisKey.substring(1, 2));
                    if (! roleInfos.containsKey(index))
                        roleInfos.put(index, new Element("ROLE_INFO"));

                    Element ri = (Element) roleInfos.get(index);

                    String name = thisKey.substring(3);
                    String text = a.getProperty(thisKey).toString();

                    ri.addContent(new Element(name).addContent(text));
                }
                catch (NumberFormatException nfe)
                {
                    nfe.printStackTrace();
                }
            }
        }

        return roleInfos.values();
    }

    /**
     * <code>decodeEvidence</code> decodes evidence attributes stored
     * as EMBL qualifiers.
     *
     * @param f a <code>Feature</code>.
     *
     * @return a <code>List</code> of <code>Attribute</code>s.
     */
    private List decodeEvidence(Feature f)
    {
        Annotation a = f.getAnnotation();

        List attrs = new ArrayList();

        for (Iterator ai = a.keys().iterator(); ai.hasNext();)
        {
            String thisKey = (String) ai.next();
            if (thisKey.equals("GENE_PREDICTIONS") ||
                thisKey.equals("PROTEIN_MATCHES")  ||
                thisKey.equals("GENE_INDEX_MATCHES"))
                attrs.add(new Attribute(thisKey, a.getProperty(thisKey).toString()));
        }

        return attrs;
    }

    /**
     * <code>calculateDate</code> calculates the current date
     * Current format = "October 7, 2003 2:30:51 PM"
     *
     * @return a formatted date <code>String</code>.
     */
    private String calculateDate()
    {
        return DateFormat.getDateTimeInstance(DateFormat.LONG,
                                              DateFormat.MEDIUM).format(new Date());
    }

    /* <code>psuDate2TIGRDate</code>
     * Read in "20000310" and return "to October 7, 2003 2:30:51 PM"
     *
     * @param psuDate is a <code>String</code>
     */
    private String psuDate2TIGRDate(String psuDateString){

        if( psuDateString == null || psuDateString.length() < 8 ){
            System.err.println("Date "+psuDateString+" was not in expected format - yyyymmdd");
            return "";
        }

        SimpleDateFormat psuFormatter
            = new SimpleDateFormat ("yyyyMMdd");
        
        SimpleDateFormat tigrFormatter
            = new SimpleDateFormat ("EEE MMM d hh:mm:ss z yyyy");

        ParsePosition  pos = new ParsePosition(0);
        Date          date = psuFormatter.parse(psuDateString, pos);

        String dateString = tigrFormatter.format(date);

        return dateString;
    }

    /**
     * <code>getModelAttribute</code> gets misc_features that are contained
     * within the CDS coords
     *
     * @return a <code>Element</code>.
     */
    private Element getModelAttribute(Feature f){
        Element []  sigPs = getSignalPs(f);
        Element []   tmms = getTMMs(f);
        Element []   gpis = getGPIAnchors(f);

        Element model_att = null;

        // Don't create MODEL_ATTRIBUTE unless children will be created
        if( sigPs != null || tmms != null || gpis != null){
            model_att = new Element("MODEL_ATTRIBUTE");

            if( sigPs.length > 0 ){
                for(int i = 0; i < sigPs.length; ++i){
                    model_att.addContent(sigPs[i]);
                }
            }

            if( tmms.length > 0 ){
                for(int i = 0; i < tmms.length; ++i){
                    model_att.addContent(tmms[i]);
                }
            }

            if (gpis.length > 0) {
                for(int i = 0; i < gpis.length; ++i){
                    model_att.addContent(gpis[i]);
                }
            }
        }

        return model_att;
    }

    /**
     * <code>getSignalPs</code> returns an array of <ATTRIBUTE_TYPE> elements,
     * one for each signalp contained within the location of the feature being
     * passed in.
     *
     * @param f is a <code>Feature</code> whos <code>Location</code> is used
     * to find signalps. ITS PARENT MUST BE A <code>SEQUENCE</code>
     *
     * @return a <code>Element</code>[]
     */
    private Element[] getSignalPs(Feature f){

        FeatureHolder features = getMiscFeatsByAnnoType(f, "signalp");

        Element [] sigPs = null;
        Map     attTypes = new HashMap();

        for(Iterator itty = features.features(); itty.hasNext();){
            Feature    sigp = (Feature) itty.next();
            Annotation anno = sigp.getAnnotation();

            Map map = new HashMap(); // This will be the ATT_SCOREs

            if( anno.containsProperty("coord") ){
                String      coords = (String) anno.getProperty("coord");
                int          comma = coords.indexOf(',');
                String hmmPosition = coords.substring(0, comma);
                map.put("HMM-position", hmmPosition);
            }

            if( anno.containsProperty("note") ){
                String          note = (String) anno.getProperty("note");
                int            first = note.indexOf(' ');
                int           second = note.indexOf(' ', first+1);
                String hmmPrediction = note.substring(0, second);
                map.put("HMM-prediction", hmmPrediction);
            }

            if( anno.containsProperty("signal") ){
                map.put("HMM-SPprob", (String) anno.getProperty("signal"));
            }

            if( anno.containsProperty("anchor")){
                map.put("HMM-SAprob", (String) anno.getProperty("anchor"));
            }

            if( anno.containsProperty("cleavage") ){
                map.put("HMM-cleavprob", (String) anno.getProperty("cleavage"));
            }

            String algorithm = "unknown";

            if( anno.containsProperty("algorithm") ){
                algorithm = (String) anno.getProperty("algorithm");
            }

            attTypes.put(algorithm, map);
        }

        sigPs = createAttTypeElements(attTypes);

        return sigPs;
    }
    /**
     * <code>getTMMs</code> returns an array of Elements, each one a
     * ATTRIBUTE_TYPE. It is very unlikely that more than one element will
     * be present (I think).
     *
     * @return a <code>Element</code>[]
     */
    private Element[] getTMMs(Feature f){

        FeatureHolder features = getMiscFeatsByAnnoType(f, "TMHMM");

        Element [] tmms = null;
        Map    attTypes = new HashMap();

        for(Iterator itty = features.features(); itty.hasNext();){
            Feature     tmm = (Feature) itty.next();
            Annotation anno = tmm.getAnnotation();

            //System.err.println("getTMMs: "+anno.getProperty("note"));

            Map map = new HashMap(); // This will be the ATT_SCOREs

            if( anno.containsProperty("coord") ){
                // Can be 1 or more coord qualifiers
                // Convert 1,2 to 1-2
                String []      coords = getAnnotationValue("coord", anno);
                StringBuffer sbCoords = new StringBuffer();

                for(int i = 0; i < coords.length; ++i){
                    if( sbCoords.length() > 0){
                        sbCoords.append(":");
                    }

                    int     space = coords[i].indexOf(',');
                    String  first = coords[i].substring(0, space);
                    String second = coords[i].substring(space + 1);

                    sbCoords.append(first);
                    sbCoords.append('-');
                    sbCoords.append(second);
                }

                map.put("coords", sbCoords.toString());
            }

            if( anno.containsProperty("note") ){
                String          note = (String) anno.getProperty("note");
                int            first = note.indexOf(' ');
                String count = note.substring(0, first);
                map.put("count", count);
            }

            String algorithm = "unknown";

            if (anno.containsProperty("algorithm")) {
                algorithm = (String) anno.getProperty("algorithm");
            }

            attTypes.put(algorithm, map);
        }

        tmms = createAttTypeElements(attTypes);

        return tmms;
    }
    /**
     * <code>getGPIAnchors</code> returns an array of <ATTRIBUTE_TYPE> elements,
     * one for each GPI contained within the location of the feature being
     * passed in.
     *
     * @param f is a <code>Feature</code> whos <code>Location</code> is used
     * to find signalps. ITS PARENT MUST BE A <code>SEQUENCE</code>
     *
     * @return a <code>Element</code>[]
     */
    private Element[] getGPIAnchors(Feature f){

        FeatureHolder features = getMiscFeatsByAnnoType(f, "GPI-Anchor");

        Element [] gpis = null;
        Map     attTypes = new HashMap();

        for(Iterator itty = features.features(); itty.hasNext();){
            Feature     gpi = (Feature) itty.next();
            Annotation anno = gpi.getAnnotation();

            //System.err.println("getSignalPs: "+anno.getProperty("note"));

            Map map = new HashMap(); // This will be the ATT_SCOREs

            if( anno.getProperty("coord") != null ){
                String      coords = (String) anno.getProperty("coord");
                int          comma = coords.indexOf(',');
                String hmmPosition = coords.substring(0, comma);
                map.put("GPI-position", hmmPosition);
            }

            if( anno.getProperty("note") != null ){
                String          note = (String) anno.getProperty("note");
                int            first = note.indexOf(' ');
                int           second = note.indexOf(' ', first+1);
                String hmmPrediction = note.substring(0, second);
                map.put("GPI-prediction", hmmPrediction);
                
            }

            if( anno.getProperty("cleavage") != null ){
                map.put("GPI-cleavprob", (String) anno.getProperty("cleavage"));
            }

            String algorithm = (String) anno.getProperty("algorithm");
            attTypes.put(algorithm, map);
        }

        gpis = createAttTypeElements(attTypes);

        return gpis;
    }

    /**
     * <code>createAttTypeElements</code> takes a hash that contains a key of
     * algorithm (i.e. "GES,TMHMM2.0", "SP-HMM,signalP-2.0") and a value of
     * another hash. This hash is the DESC and text value of a ATT_SCORE
     * element.
     *
     * @return a <code>Element</code>[] of ATTRIBUTE_TYPE elements
     */
    private Element[] createAttTypeElements(Map attTypes){
        List attTypeList = new ArrayList();

        // for each ATTRIBUTE_TYPE required (key is algorithm)
        for(Iterator itty = attTypes.keySet().iterator(); itty.hasNext();){
            String   method = (String) itty.next();
            Element attType = new Element("ATTRIBUTE_TYPE");
            attType.setAttribute("METHOD", method);

            Map scores = (Map) attTypes.get(method);

            for(Iterator i = scores.keySet().iterator(); i.hasNext();){
                String desc  = (String) i.next();
                String value = (String) scores.get(desc);

                Element attScore = new Element("ATT_SCORE").addContent(value);
                attScore.setAttribute("DESC", desc);

                attType.addContent(attScore);
            }

            attTypeList.add(attType);
        }

        Element [] toReturn = null;

        if( attTypeList != null ){
            Element [] a = new Element[attTypeList.size()];
            toReturn = (Element []) attTypeList.toArray(a);
        }

        return toReturn;
    }

    /**
     * <code>getMiscFeatsByAnnoType</code> returns all misc_features that
     * are contained within the Location of the feature (normally a CDS).
     * The annotation of the misc_feature must also have /type=x, where
     * x is the second argument to this method
     *
     * @param the <code>Feature</code> whos Location is used. Its parent must
     * be a Seqence.
     * @param <code>String</code> What /type= must have as a value.
     *
     * @return a <code>FeatureHolder</code>
     */
    private FeatureHolder getMiscFeatsByAnnoType(Feature f, String type)
    {
        // the CDS feature passed in has to be used to get to its parent, a
        // sequence, and then find all signalPs over lapping the CDS location
        FeatureFilter  byLoc =
            new FeatureFilter.ContainedByLocation(f.getLocation());
        FeatureFilter byType = new FeatureFilter.ByType("misc_feature");
        FeatureFilter byBoth = new FeatureFilter.And(byLoc, byType);

        AnnotationType at = new AnnotationType.Impl();
        at.setConstraint("type",
              new CollectionConstraint.Contains(
                  new PropertyConstraint.ExactValue(type),
                  CardinalityConstraint.ONE
                  ));
        FeatureFilter byAnno =
            new FeatureFilter.ByAnnotationType(at);
        FeatureFilter  byAll = new FeatureFilter.And(byBoth, byAnno);


        Sequence           seq = (Sequence) f.getParent();
        FeatureHolder features = seq.filter(byAll);

        return features;
    }

    /**
     * <code>getModelEvidence</code> creates <MODEL_EVIDENCE> and its
     * children.
     *
     * @param the <code>Feature</code> to extract domains from
     * @return a <code>Element</code>.
     */
    private Element getModelEvidence(Feature f){
        String [] domains    = getAnnotationValue("domain",
                                                  f.getAnnotation());
        String [] similarity = getAnnotationValue("similarity",
                                                  f.getAnnotation());
        Element model_evi = null;

        //        if( domains.length > 0 || similarity.length > 0 ){
        if( domains.length > 0 ){
            model_evi            = new Element("MODEL_EVIDENCE");
            Element evi_type     = new Element("EVIDENCE_TYPE");
            model_evi.addContent(evi_type);
            Element seq_db_match = new Element("SEQUENCE_DB_MATCH");
            evi_type.addContent(seq_db_match);

            if( domains.length > 0 ){
                // get Pfam, then PROSITE, then Interpro, since they are
                // grouped like this in the XML
                String [] pfams   = getDomainsStartingWith("Pfam", domains);
                String [] prosite = getDomainsStartingWith("Prosite", domains);
                String [] smart   = getDomainsStartingWith("smart", domains);
                String [] prints  = getDomainsStartingWith("prints", domains);

                makeSearchDBElement("pfam",    pfams,   seq_db_match);
                makeSearchDBElement("prosite", prosite, seq_db_match);
                makeSearchDBElement("smart",   smart,   seq_db_match);
                makeSearchDBElement("prints",  prints,  seq_db_match);
            }
        }

        return model_evi;
    }

    /* <code>makeSearchDBElement</code> creates a <SEARCH_DB> element plus
     * its children.
     *
     * @param db is ad<code>String</code>
     * @param domains <code>String</code>[] 
     * @param parent is a <code>Element</code> has a <SEARCH_DB> element 
     * added to its content.
     */
    private void makeSearchDBElement(String    db,
                                     String [] domains,
                                     Element   parent){
        if( domains != null && domains.length > 0 ){

            Element search_db = new Element("SEARCH_DB");
            search_db.setAttribute("DB_NAME", db);
            search_db.setAttribute("DB_TYPE", getDbType(db));
            parent.addContent(search_db);

            for(int x = 0; x < domains.length; ++x){
                String []  info = getDomainInfo(domains[x]);
                Element seq_ele = new Element("SEQ_ELEMENT");
                seq_ele.setAttribute("FEAT_TYPE",   "db_match");
                seq_ele.setAttribute("METHOD", PROPERTY_UNKNOWN); //program name

                if( info[1] != null ){
                    seq_ele.setAttribute("ACCESSION",   info[1]);
                }

                if( info[2] != null ){
                    seq_ele.setAttribute("DESCRIPTION", info[2]);
                }

                // NOTE: TIGR XML can not except protein coords unless you use
                //       the END5 and END3 and pretend they are N & C

                //if( info[3] != null ){
                    // create RELATIVE_COORDS -> COORDSET
                //}

                search_db.addContent(seq_ele);
            }
        }
    }

    /* <code>getDbType</code> returns a nice name for a DB.
     *
     * FIXME: This should be enhanced in the future so this info comes from the
     *        xml.properties file rather than be hard coded here.
     */
    private String getDbType(String dbName){
        if( dbName.equals("Pfam") ){
            return "Protein families database of alignments and HMMs";
        }
        else if( dbName.equals("prosite") ){
            return "Database of protein families and domains";
        }
        else if( dbName.equals("smart") ){
            return "Simple Modular Architecture Research Tool";
        }
        else if( dbName.equals("prints") ){
            return "Protein fingerprints database";
        }

        return "";
    }

    /* <code>getDomainInfo</code> returns a 5 element <code>String []</code>.
     * This is made from the domain string;
     *
     *   DB:accession;description;evalue;location
     *
     * evalue and location can contain null.
     */
    private String [] getDomainInfo(String domain){
        int colon = domain.indexOf(":");
        String db = domain.substring(0, colon);

        int    semi1     = domain.indexOf(";");
        String accession = domain.substring(colon+1, semi1);

        int    semi2     = domain.indexOf(";", semi1+1);
        String description = null;
        String evalue      = null;
        String location    = null;

        // Might only be Pfam:123;description.....
        if( semi2 <= 0 ){
            description = domain.substring(semi1+1);
            evalue      = null;
            location    = null;
        }
        else {
            description = domain.substring(semi1+1, semi2);
            int   semi3 = domain.indexOf(";", semi2+1);

            if( semi3 <= 0 ){
                evalue   = domain.substring(semi2+1);
                location = null;
            }
            else {
                evalue   = domain.substring(semi2+1, semi3);
                location = domain.substring(semi3+1);
            }
        }


        return (String []) new String [] {db, accession, description,
                                          evalue, location};
        
    }

    /* <code>getDomainsStartingWith</code>
     * Cycle through the String [] and build a List of everything 
     * that starts with startsWith
     *
     * It is case insensitive
     */
    private String [] getDomainsStartingWith(String startsWith,
                                             String [] domains){
        List list = new ArrayList();

        for(int i = 0; i < domains.length; ++i){
            
            if( domains[i].toLowerCase().startsWith(startsWith.toLowerCase()) ){
                list.add(domains[i]);
            }
        }
            
        return (String []) list.toArray(new String [0]);
    }

   /* <code>getFeatNameFor</code> just cycles through the most stable to the
     * least stable qualifiers for identifying a feature.
     *
     * @param f is the <code>Feature</code>.
     */
    private String getFeatNameFor(int type)
    {
        String featName = PROPERTY_UNKNOWN;


        switch (type)
        {
            case TU:
                featNameArray[type]++;
                featName = cloneName+".t."+featNameArray[type];
                break;

            case MODEL:
                 featNameArray[type]++;
                 featName = cloneName+".m."+featNameArray[type];
                 break;

            case EXON:
                 featNameArray[type]++;
                 featName = cloneName+".e."+featNameArray[type];
                 break;

            case RNA:
                 featNameArray[type]++;
                 featName = cloneName+".r."+featNameArray[type];
                 break;

            case CDS:
                 featNameArray[type]++;
                 featName = cloneName+".c."+featNameArray[type];
                 break;
                 
            case REPEAT:
                 featNameArray[type]++;
                 featName = cloneName+".repeat."+featNameArray[type];
                 break;
 
            default:
                 System.err.println("Requested unknown ELEMENT type '"
                                    + type
                                    + "' from TigrXMLDocBuilder");
        }

        return featName;
    }

}
