<?xml version="1.0"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                version="1.0">

  <!--

       This stylesheet transforms TIGR XML into HTML. It gives a
       summary of protein-coding gene information. RNA genes,
       miscellaneous features, repeats and sequences are
       excluded. Sequence DB hits are filtered to exclude those with a
       score of less than 100.

       Author: Keith James <kdj@sanger.ac.uk>

       -->

  <xsl:output method="html" indent="no" />
  <xsl:strip-space elements="*" />

  <!-- Template for TIGR. Top level template. -->
  <xsl:template match="TIGR">
    <xsl:apply-templates />
  </xsl:template>

  <!-- Template for PSEUDOCHROMOSOME. Top level display template. -->
  <xsl:template match="PSEUDOCHROMOSOME">
    <html>
      <head>
        <link rel="stylesheet" type="text/css" href="tigr.css" />
      </head>
      <body>
        <h1>Pseudochromosome:</h1>
      </body>
    </html>
    <xsl:apply-templates />
  </xsl:template>

  <!-- Template for SCAFFOLD. Marker template. -->
  <xsl:template match="SCAFFOLD">
    <h2>Scaffold:</h2>
  </xsl:template>

  <!-- Template for ASSEMBLY. Displays values. -->
  <xsl:template match="ASSEMBLY">
    <h3>Assembly:</h3>
    <p>
      <xsl:text>Clone ID: </xsl:text>
      <xsl:value-of select="@CLONE_ID" />
      <xsl:text>, Database: </xsl:text>
      <xsl:value-of select="@DATABASE" />
      <xsl:text>, Chromosome: </xsl:text>
      <xsl:value-of select="@CHROMOSOME" />
      <xsl:text>, Date:  </xsl:text>
      <xsl:value-of select="@CURRENT_DATE" />
    </p>
    <xsl:apply-templates select="GENE_LIST"/>
  </xsl:template>

  <!-- Template for GENE_LIST. Marker template. -->
  <xsl:template match="GENE_LIST">
    <xsl:apply-templates />
  </xsl:template>

  <!-- Template for MISC_INFO. Marker template. -->
  <xsl:template match="MISC_INFO">
    <h2>Misc info.</h2>
    <xsl:text>Data not shown</xsl:text>
  </xsl:template>

  <!-- Template for REPEAT_LIST. Marker template. -->
  <xsl:template match="REPEAT_LIST">
    <h2>Repeat list</h2>
    <xsl:text>Data not shown</xsl:text>
  </xsl:template>

  <!-- Template for PROTEIN_CODING. Marker template. -->
  <xsl:template match="PROTEIN_CODING">
    <h2>Protein Coding</h2>
    <xsl:apply-templates>
      <xsl:sort order="ascending" select="./FEAT_NAME" />
    </xsl:apply-templates>
  </xsl:template>

  <!-- Template for RNA_GENES. Marker template. -->
  <xsl:template match="RNA_GENES">
    <h2>RNA genes</h2>
    <xsl:text>Data not shown</xsl:text>
  </xsl:template>

  <!-- Template for TU. Displays values and applies selected templates. -->
  <xsl:template match="TU">
    <hr></hr>
    <h3>
      <xsl:text>Transcription unit </xsl:text>
      <xsl:value-of select="./FEAT_NAME" />
      <xsl:text> [</xsl:text>
      <xsl:value-of select="./COORDSET/END5" />
      <xsl:text>..</xsl:text>
      <xsl:value-of select="./COORDSET/END3" />]
    </h3>
    <xsl:apply-templates select="GENE_INFO | MODEL | GENE_EVIDENCE"/>
  </xsl:template>

  <!-- Template for GENE_INFO. Displays values. Applies ROLE_LIST
       template. -->
  <xsl:template match="GENE_INFO">
    <table>
      <tr class="header">
        <th>Locus</th>
        <th>Alt. locus</th>
        <th>Pub. locus</th>
        <th>Com. name</th>
        <th>Pub. comment</th>
        <th>EC number</th>
        <th>Gene sym.</th>
        <th>Pseudogene</th>
      </tr>
      <tr class="bg-normal">
        <td align="center"><xsl:value-of select="LOCUS" /></td>

        <td align="center">
            <xsl:choose>
            <xsl:when test="ALT_LOCUS">
              <xsl:value-of select="ALT_LOCUS" />
            </xsl:when>
            <xsl:otherwise>
              <xsl:text>-</xsl:text>
            </xsl:otherwise>
          </xsl:choose>
        </td>
        <td align="center">
            <xsl:choose>
            <xsl:when test="PUB_LOCUS">
              <xsl:value-of select="PUB_LOCUS" />
            </xsl:when>
            <xsl:otherwise>
              <xsl:text>-</xsl:text>
            </xsl:otherwise>
          </xsl:choose>
        </td>

        <td align="center"><xsl:value-of select="COM_NAME" /></td>

        <td align="center">
            <xsl:choose>
            <xsl:when test="PUB_COMMENT">
              <xsl:value-of select="PUB_COMMENT" />
            </xsl:when>
            <xsl:otherwise>
              <xsl:text>-</xsl:text>
            </xsl:otherwise>
          </xsl:choose>
        </td>
        <td align="center">
            <xsl:choose>
            <xsl:when test="EC_NUMBER">
              <xsl:value-of select="EC_NUMBER" />
            </xsl:when>
            <xsl:otherwise>
              <xsl:text>-</xsl:text>
            </xsl:otherwise>
          </xsl:choose>
        </td>
        <td align="center">
            <xsl:choose>
            <xsl:when test="GENE_SYM">
              <xsl:value-of select="GENE_SYM" />
            </xsl:when>
            <xsl:otherwise>
              <xsl:text>-</xsl:text>
            </xsl:otherwise>
          </xsl:choose>
        </td>
        <td align="center">
          <xsl:choose>
            <xsl:when test="IS_PSEUDOGENE = 1">
              <xsl:text>Yes</xsl:text>
            </xsl:when>
            <xsl:otherwise>
              <xsl:text>No</xsl:text>
            </xsl:otherwise>
          </xsl:choose>
        </td>
      </tr>
    </table>

    <h4>Functional annotation evidence</h4>
    <xsl:choose>
      <xsl:when test="child::FUNCT_ANNOT_EVIDENCE">
        <ul>
          <xsl:apply-templates select="FUNCT_ANNOT_EVIDENCE" />
        </ul>
      </xsl:when>
      <xsl:otherwise>
        <xsl:text>None given</xsl:text>
      </xsl:otherwise>
    </xsl:choose>   

    <h4>Role list</h4>
    <xsl:choose>
      <xsl:when test="child::ROLE_LIST">
        <xsl:apply-templates select="ROLE_LIST" />
      </xsl:when>
      <xsl:otherwise>
        <xsl:text>None given</xsl:text>
      </xsl:otherwise>
    </xsl:choose>

    <h4>Comment</h4>
    <xsl:choose>
      <xsl:when test="child::COMMENT">
        <xsl:value-of select="COMMENT" />
      </xsl:when>
      <xsl:otherwise>
        <xsl:text>None given</xsl:text>
      </xsl:otherwise>
    </xsl:choose>
    
  </xsl:template>

  <!-- Template for FUNCT_ANNOT_EVIDENCE. -->
  <xsl:template match="FUNCT_ANNOT_EVIDENCE">
    <li>
      <xsl:text>Evidence type: </xsl:text>
      <xsl:value-of select="@TYPE" />
      <ul>
       <xsl:apply-templates select="ASSIGN_ACC" />
      </ul>
    </li>
  </xsl:template>

  <!-- Template for ASSIGN_ACC. -->
  <xsl:template match="ASSIGN_ACC">
    <li>
      <xsl:text>Method: </xsl:text>
      <xsl:value-of select="@ASSIGN_METHOD" />
      <p>
      <xsl:value-of select="." />
      </p>
    </li>
  </xsl:template>

  <!-- Template for ROLE_LIST. Displays table header. -->
  <xsl:template match="ROLE_LIST">
    <table>
      <tr class="header">
        <th>Compartment</th>
        <th>Main role</th>
        <th>Subrole 1</th>
        <th>Subrole 2</th>
        <th>Subrole 3</th>
        <th>Subrole 4</th>
      </tr>
      <xsl:apply-templates />
    </table>
  </xsl:template>

  <!-- Template for ROLE_INFO. Displays values. -->
  <xsl:template match="ROLE_INFO">
    <tr class="bg-normal">
      <td align="center"><xsl:value-of select="COMPARTMENT" /></td>
      <td align="center"><xsl:value-of select="MAIN_ROLE" /></td>

      <td align="center">
        <xsl:choose>
          <xsl:when test="SUBROLE_1">
            <xsl:value-of select="SUBROLE_1" />
          </xsl:when>
          <xsl:otherwise>
            <xsl:text>-</xsl:text>
          </xsl:otherwise>
        </xsl:choose>
      </td>
      <td align="center">
        <xsl:choose>
          <xsl:when test="SUBROLE_2">
            <xsl:value-of select="SUBROLE_2" />
          </xsl:when>
          <xsl:otherwise>
            <xsl:text>-</xsl:text>
          </xsl:otherwise>
        </xsl:choose>
      </td>
      <td align="center">
        <xsl:choose>
          <xsl:when test="SUBROLE_3">
            <xsl:value-of select="SUBROLE_3" />
          </xsl:when>
          <xsl:otherwise>
            <xsl:text>-</xsl:text>
          </xsl:otherwise>
        </xsl:choose>
      </td>
      <td align="center">
        <xsl:choose>
          <xsl:when test="SUBROLE_4">
            <xsl:value-of select="SUBROLE_4" />
          </xsl:when>
          <xsl:otherwise>
            <xsl:text>-</xsl:text>
          </xsl:otherwise>
        </xsl:choose>
      </td>
    </tr>
  </xsl:template>

  <!-- Template for MODEL. Displays table header. -->
  <xsl:template match="MODEL">
    <h4>Model</h4>
    <table>
      <tr class="header">
        <th>Exon</th>
        <th>Asm. coords.</th>
        <th>Left UTR</th>
        <th>Right UTR</th>
        <th>Extended UTR</th>
      </tr>
      <xsl:apply-templates select="EXON" />
    </table>
    <xsl:apply-templates select="MODEL_EVIDENCE" />
  </xsl:template>

  <!-- Template for odd EXON. Sets row class and calls named exon
       template -->
  <xsl:template match="EXON">
    <tr class="bg-normal">
      <xsl:call-template name="exon" />
    </tr>
  </xsl:template>

  <!-- Template for even EXON. Sets row class and calls named exon
       template -->
  <xsl:template match="EXON[position() mod 2 = 0]">
    <tr class="bg-shaded">
      <xsl:call-template name="exon" />
    </tr>
  </xsl:template>

  <!-- Named exon template. -->
  <xsl:template name="exon">
    <td align="center">
      <xsl:value-of select="FEAT_NAME" />
    </td>
    <td align="center">
      <xsl:value-of select="COORDSET/END5" />
      <xsl:text>..</xsl:text>
      <xsl:value-of select="COORDSET/END3" />
    </td>
    <td align="center">
      <xsl:choose>
        <xsl:when test="child::UTRS/LEFT_UTR">
          <xsl:value-of select="UTRS/LEFT_UTR/COORDSET/END5" />
          <xsl:text>..</xsl:text>
          <xsl:value-of select="UTRS/LEFT_UTR/COORDSET/END3" />
        </xsl:when>
        <xsl:otherwise>
          <xsl:text>-</xsl:text>
        </xsl:otherwise>
      </xsl:choose>
    </td>
    <td align="center">
      <xsl:choose>
        <xsl:when test="child::UTRS/RIGHT_UTR">
          <xsl:value-of select="UTRS/RIGHT_UTR/COORDSET/END5" />
          <xsl:text>..</xsl:text>
          <xsl:value-of select="UTRS/RIGHT_UTR/COORDSET/END3" />
        </xsl:when>
        <xsl:otherwise>
          <xsl:text>-</xsl:text>
        </xsl:otherwise>
      </xsl:choose>
    </td>
    <td align="center">
      <xsl:choose>
        <xsl:when test="child::UTRS/EXTENDED_UTR">
          <xsl:value-of select="UTRS/EXTENDED_UTR/COORDSET/END5" />
          <xsl:text>..</xsl:text>
          <xsl:value-of select="UTRS/EXTENDED_UTR/COORDSET/END3" />
        </xsl:when>
        <xsl:otherwise>
          <xsl:text>-</xsl:text>
        </xsl:otherwise>
      </xsl:choose>
    </td>
  </xsl:template>

  <!-- Template for GENE_EVIDENCE. Applies selected templates. -->
  <xsl:template match="GENE_EVIDENCE">
    <h4>Gene evidence</h4>

    <h5>Sequence DB match</h5>
    <xsl:choose>
      <xsl:when test="child::EVIDENCE_TYPE/SEQUENCE_DB_MATCH/SEARCH_DB">
        <xsl:apply-templates select="EVIDENCE_TYPE/SEQUENCE_DB_MATCH" />
      </xsl:when>
      <xsl:otherwise>
        <xsl:text>None found</xsl:text>
      </xsl:otherwise>
    </xsl:choose>

   <h5>Computer prediction</h5>
    <xsl:choose>
      <xsl:when test="child::EVIDENCE_TYPE/COMPUT_PREDICTION/PREDICTION_SET">
        <xsl:apply-templates select="EVIDENCE_TYPE/COMPUT_PREDICTION" />
      </xsl:when>
      <xsl:otherwise>
        <xsl:text>None predicted</xsl:text>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <!-- Template for MODEL_EVIDENCE. Applies selected templates. -->
  <xsl:template match="MODEL_EVIDENCE">
    <h4>Model evidence</h4>

    <h5>Sequence DB match</h5>
    <xsl:choose>
      <xsl:when test="child::EVIDENCE_TYPE/SEQUENCE_DB_MATCH/SEARCH_DB">
        <xsl:apply-templates select="EVIDENCE_TYPE/SEQUENCE_DB_MATCH" />
      </xsl:when>
      <xsl:otherwise>
        <xsl:text>None found</xsl:text>
      </xsl:otherwise>
    </xsl:choose>

   <h5>Computer prediction</h5>
    <xsl:choose>
      <xsl:when test="child::EVIDENCE_TYPE/COMPUT_PREDICTION/PREDICTION_SET">
        <xsl:apply-templates select="EVIDENCE_TYPE/COMPUT_PREDICTION" />
      </xsl:when>
      <xsl:otherwise>
        <xsl:text>None predicted</xsl:text>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <!-- Template for SEARCH_DB. Displays table header. Filters hits
       with score less than 100. Sorts by score and percent id. Calls
       named seq_element template. -->
  <xsl:template match="SEARCH_DB">
    <h6>
      <xsl:text>Search DB - DB name: </xsl:text>
      <xsl:value-of select="@DB_NAME" />
      <xsl:text>, DB type: </xsl:text>
      <xsl:value-of select="@DB_TYPE" />
    </h6>

    <table>
      <tr class="header">
        <th>Assm. coords.</th>
        <th>Match coords.</th>
        <th>Rel. coords.</th>
        <th>Type</th>
        <th>Method</th>
        <th>Accession</th>
        <th>Description</th>
        <th>% id.</th>
        <th>Score</th>
        <th>E-value</th>
        <th>Element no.</th>
      </tr>

      <!-- Commented out as InterPro hits have no score recorded and
      are always filtered -->
      <!-- xsl:for-each select="SEQ_ELEMENT[@SCORE &gt; 99]" -->
      <xsl:for-each select="SEQ_ELEMENT">
        <xsl:sort data-type="number" order="descending" select="@SCORE" />
        <xsl:sort data-type="number" order="descending" select="@PER_ID" />
        <xsl:call-template name="seq_element" />
      </xsl:for-each>
    </table>
  </xsl:template>

  <!-- Template for PREDICTION_SET. Currently marker template. Display
       code is commented out. -->
  <xsl:template match="PREDICTION_SET">
    <h6>
      <xsl:text>Prediction set - Prediction tool: </xsl:text>
      <xsl:value-of select="@PREDICTION_TOOL" />
      <xsl:text>, Prediction type: </xsl:text>
      <xsl:value-of select="@PREDICTION_TYPE" />
    </h6>

    <xsl:text>Data not shown</xsl:text>

    <!-- 

    <table>
      <tr class="header">
        <th>Assm. coords.</th>
        <th>Match coords.</th>
        <th>Rel. coords.</th>
        <th>Type</th>
        <th>Method</th>
        <th>Accession</th>
        <th>Description</th>
        <th>% id.</th>
        <th>Score</th>
        <th>E value</th>
        <th>Element no.</th>
      </tr>

      <xsl:apply-templates>
        <xsl:sort data-type="number" order="descending" select="@SCORE" />
        <xsl:sort data-type="number" order="descending" select="@PER_ID" />
      </xsl:apply-templates>
    </table>

    -->
  </xsl:template>

  <!-- Named seq_element template. -->
  <xsl:template name="seq_element">
     <tr class="bg-shaded">
       <td align="center">
         <xsl:choose>
           <xsl:when test="child::ASMBL_COORDS">
             <xsl:value-of select="ASMBL_COORDS/COORDSET/END5" />
             <xsl:text>..</xsl:text>
             <xsl:value-of select="ASMBL_COORDS/COORDSET/END3" />
           </xsl:when>
           <xsl:otherwise>
             <xsl:text>-</xsl:text>
           </xsl:otherwise>
         </xsl:choose>
       </td>
       <td align="center">
         <xsl:choose>
           <xsl:when test="child::MATCH_COORDS">
             <xsl:value-of select="MATCH_COORDS/COORDSET/END5" />
             <xsl:text>..</xsl:text>
             <xsl:value-of select="MATCH_COORDS/COORDSET/END3" />
           </xsl:when>
           <xsl:otherwise>
             <xsl:text>-</xsl:text>
           </xsl:otherwise>
         </xsl:choose>
       </td>
       <td align="center">
         <xsl:choose>
           <xsl:when test="child::RELATIVE_COORDS">
             <xsl:value-of select="RELATIVE_COORDS/COORDSET/END5" />
             <xsl:text>..</xsl:text>
             <xsl:value-of select="RELATIVE_COORDS/COORDSET/END3" />
           </xsl:when>
           <xsl:otherwise>
             <xsl:text>-</xsl:text>
           </xsl:otherwise>
         </xsl:choose>
       </td>

       <td align="center"><xsl:value-of select="@FEAT_TYPE" /></td>

       <td align="center">
         <xsl:choose>
           <xsl:when test="@METHOD">
             <xsl:value-of select="@METHOD" />
           </xsl:when>
           <xsl:otherwise>
             <xsl:text>-</xsl:text>
           </xsl:otherwise>
         </xsl:choose>
       </td>
       <td align="center">
         <xsl:choose>
           <xsl:when test="@ACCESSION">
             <xsl:value-of select="@ACCESSION" />
           </xsl:when>
           <xsl:otherwise>
             <xsl:text>-</xsl:text>
           </xsl:otherwise>
         </xsl:choose>
       </td>
       <td align="center">
         <xsl:choose>
           <xsl:when test="@DESCRIPTION">
             <xsl:value-of select="@DESCRIPTION" />
           </xsl:when>
           <xsl:otherwise>
             <xsl:text>-</xsl:text>
           </xsl:otherwise>
         </xsl:choose>
       </td>
       <td align="center">
         <xsl:choose>
           <xsl:when test="@PER_ID">
             <xsl:value-of select="format-number(@PER_ID, '##.##')" />
           </xsl:when>
           <xsl:otherwise>
             <xsl:text>-</xsl:text>
           </xsl:otherwise>
         </xsl:choose>
       </td>
       <td align="center">
         <xsl:choose>
           <xsl:when test="@SCORE">
             <xsl:value-of select="@SCORE" />
           </xsl:when>
           <xsl:otherwise>
             <xsl:text>-</xsl:text>
           </xsl:otherwise>
         </xsl:choose>
       </td>
       <td align="center">
         <xsl:choose>
           <xsl:when test="@E_VALUE">
             <xsl:value-of select="@E_VALUE" />
           </xsl:when>
           <xsl:otherwise>
             <xsl:text>-</xsl:text>
           </xsl:otherwise>
         </xsl:choose>
       </td>
       <td align="center">
         <xsl:choose>
           <xsl:when test="@ELEMENT_NUM">
             <xsl:value-of select="@ELEMENT_NUM" />
           </xsl:when>
           <xsl:otherwise>
             <xsl:text>-</xsl:text>
           </xsl:otherwise>
         </xsl:choose>
       </td>
     </tr>
  </xsl:template>

  <!-- Template for FEAT_NAME. Displays value -->
  <xsl:template match="FEAT_NAME">
    <xsl:text>Feature name: </xsl:text>
    <xsl:value-of select="." />
  </xsl:template>

  <!-- Template for TRANSCRIPT_SEQUENCE. Marker template. Not used. -->
  <xsl:template match="TRANSCRIPT_SEQUENCE">
    <xsl:text>Transcript sequence: </xsl:text>
    <xsl:variable name="sequence" select="." />
    <xsl:value-of select="substring($sequence, 1, 11)" />
    <xsl:text>...</xsl:text>
  </xsl:template>

  <!-- Template for CDS_SEQUENCE. Marker template. Not used. -->
  <xsl:template match="CDS_SEQUENCE">
    <xsl:text>CDS sequence: </xsl:text>
    <xsl:variable name="sequence" select="." />
    <xsl:value-of select="substring($sequence, 1, 11)" />
    <xsl:text>...</xsl:text>
  </xsl:template>

  <!-- Template for PROTEIN_SEQUENCE. Marker template. Not used. -->
  <xsl:template match="PROTEIN_SEQUENCE">
    <xsl:text>Protein sequence: </xsl:text>
    <xsl:variable name="sequence" select="." />
    <xsl:value-of select="substring($sequence, 1, 11)" />
    <xsl:text>...</xsl:text>
  </xsl:template>

  <!-- Template for ASSEMBLY_SEQUENCE. Marker template. Not used. -->
  <xsl:template match="ASSEMBLY_SEQUENCE">
    <xsl:text>Assembly sequence: </xsl:text>
    <xsl:variable name="sequence" select="." />
    <xsl:value-of select="substring($sequence, 1, 11)" />
    <xsl:text>...</xsl:text>
  </xsl:template>
</xsl:stylesheet>
